Package: tidyprompt
Title: Prompt Large Language Models and Enhance Their Functionality
Version: 0.4.0
Authors@R: 
  c(person(given = "Luka",
           family = "Koning",
           role = c("aut", "cre", "cph"),
           email = "koningluka@gmail.com"),
    person(given = "Tjark",
           family = "Van de Merwe",
           role = c("aut", "cph"),
           email = "t.vandemerwe@kennispunttwente.nl"),
    person(given = "Kennispunt Twente",
           role = "fnd",
           email = "info@kennispunttwente.nl"))
Description: 
  Easily construct prompts and associated logic for interacting with 
  large language models (LLMs). 
  'tidyprompt' introduces the concept of prompt wraps, which are building
  blocks that you can use to quickly turn a simple prompt into a complex
  one. Prompt wraps do not just modify the prompt text, but also add
  extraction and validation functions that will be applied to the response
  of the LLM. This ensures that the user gets the desired output.
  'tidyprompt' can add various features to prompts and their evaluation
  by LLMs, such as structured output, automatic feedback, retries, reasoning
  modes, autonomous R function calling, and R code generation and evaluation.
  It is designed to be compatible with any LLM provider that offers
  chat completion.
License: GPL (>= 3) | file LICENSE
Encoding: UTF-8
RoxygenNote: 7.3.3
URL: https://github.com/KennispuntTwente/tidyprompt,
        https://KennispuntTwente.github.io/tidyprompt/
BugReports: https://github.com/KennispuntTwente/tidyprompt/issues
Depends: R (>= 4.1.0)
Suggests: knitr, rmarkdown, testthat (>= 3.0.0), withr, here, callr,
        skimr, jsonvalidate, DBI, ellmer (>= 0.3.0), coro, ggplot2,
        mcptools, grid
VignetteBuilder: knitr
Imports: dplyr, glue, httr2, jsonlite, stringr, utils, cli, R6, rlang,
        S7
Config/testthat/edition: 3
NeedsCompilation: no
Packaged: 2026-04-20 20:25:50 UTC; dhrko
Author: Luka Koning [aut, cre, cph],
  Tjark Van de Merwe [aut, cph],
  Kennispunt Twente [fnd]
Maintainer: Luka Koning <koningluka@gmail.com>
Repository: CRAN
Date/Publication: 2026-04-21 08:30:03 UTC
Built: R 4.5.2; ; 2026-04-21 11:46:57 UTC; unix

Package: tidyprompt
Title: Prompt Large Language Models and Enhance Their Functionality
Version: 0.4.0
Authors@R: 
  c(person(given = "Luka",
           family = "Koning",
           role = c("aut", "cre", "cph"),
           email = "koningluka@gmail.com"),
    person(given = "Tjark",
           family = "Van de Merwe",
           role = c("aut", "cph"),
           email = "t.vandemerwe@kennispunttwente.nl"),
    person(given = "Kennispunt Twente",
           role = "fnd",
           email = "info@kennispunttwente.nl"))
Description: 
  Easily construct prompts and associated logic for interacting with 
  large language models (LLMs). 
  'tidyprompt' introduces the concept of prompt wraps, which are building
  blocks that you can use to quickly turn a simple prompt into a complex
  one. Prompt wraps do not just modify the prompt text, but also add
  extraction and validation functions that will be applied to the response
  of the LLM. This ensures that the user gets the desired output.
  'tidyprompt' can add various features to prompts and their evaluation
  by LLMs, such as structured output, automatic feedback, retries, reasoning
  modes, autonomous R function calling, and R code generation and evaluation.
  It is designed to be compatible with any LLM provider that offers
  chat completion.
License: GPL (>= 3) | file LICENSE
Encoding: UTF-8
RoxygenNote: 7.3.3
URL: https://github.com/KennispuntTwente/tidyprompt,
        https://KennispuntTwente.github.io/tidyprompt/
BugReports: https://github.com/KennispuntTwente/tidyprompt/issues
Depends: R (>= 4.1.0)
Suggests: knitr, rmarkdown, testthat (>= 3.0.0), withr, here, callr,
        skimr, jsonvalidate, DBI, ellmer (>= 0.3.0), coro, ggplot2,
        mcptools, grid
VignetteBuilder: knitr
Imports: dplyr, glue, httr2, jsonlite, stringr, utils, cli, R6, rlang,
        S7
Config/testthat/edition: 3
NeedsCompilation: no
Packaged: 2026-04-20 20:25:50 UTC; dhrko
Author: Luka Koning [aut, cre, cph],
  Tjark Van de Merwe [aut, cph],
  Kennispunt Twente [fnd]
Maintainer: Luka Koning <koningluka@gmail.com>
Repository: CRAN
Date/Publication: 2026-04-21 08:30:03 UTC
Built: R 4.6.0; ; 2026-04-29 04:27:46 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method(chat_history,character)
S3method(chat_history,data.frame)
S3method(chat_history,default)
S3method(print,Tidyprompt)
S3method(prompt_wrap,default)
S3method(prompt_wrap,tidyprompt)
export("llm_provider-class")
export("persistent_chat-class")
export("tidyprompt-class")
export(add_image)
export(add_msg_to_chat_history)
export(add_text)
export(answer_as_boolean)
export(answer_as_category)
export(answer_as_dataframe)
export(answer_as_integer)
export(answer_as_json)
export(answer_as_key_value)
export(answer_as_list)
export(answer_as_multi_category)
export(answer_as_named_list)
export(answer_as_numeric)
export(answer_as_regex_match)
export(answer_as_text)
export(answer_by_chain_of_thought)
export(answer_by_react)
export(answer_using_r)
export(answer_using_sql)
export(answer_using_tools)
export(chat_history)
export(construct_prompt_text)
export(df_to_string)
export(extract_from_return_list)
export(get_chat_history)
export(get_prompt_wraps)
export(is_tidyprompt)
export(llm_break)
export(llm_break_soft)
export(llm_feedback)
export(llm_provider_ellmer)
export(llm_provider_google_gemini)
export(llm_provider_groq)
export(llm_provider_mistral)
export(llm_provider_ollama)
export(llm_provider_openai)
export(llm_provider_openrouter)
export(llm_provider_xai)
export(llm_verify)
export(prompt_wrap)
export(provider_prompt_wrap)
export(quit_if)
export(r_json_schema_to_example)
export(send_prompt)
export(set_chat_history)
export(set_system_prompt)
export(skim_with_labels_and_levels)
export(tidyprompt)
export(tools_add_docs)
export(tools_get_docs)
export(user_verify)
export(vector_list_to_string)
importFrom(R6,R6Class)
importFrom(rlang,"%||%")
importFrom(rlang,set_names)

<!-- README.md is generated from README.Rmd. Please edit that file -->

# tidyprompt

<!-- badges: start -->

[![R-CMD-check](https://github.com/KennispuntTwente/tidyprompt/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/KennispuntTwente/tidyprompt/actions/workflows/R-CMD-check.yaml)
[![](https://www.r-pkg.org/badges/version/tidyprompt)](https://cran.r-project.org/package=tidyprompt)
[![](https://img.shields.io/badge/devel%20version-0.3.0.9000-blue.svg)](https://github.com/KennispuntTwente/tidyprompt)
<!-- badges: end -->

‘tidyprompt’ is an R package to easily construct prompts and associated
logic for interacting with large language models (‘LLMs’).

Think of ‘tidyprompt’ as the ‘ggplot2’ package for creating prompts and
handling LLM interactions. ‘tidyprompt’ introduces the concept of prompt
wraps, which are building blocks that you can use to quickly turn a
simple prompt into an advanced one. Prompt wraps do not just modify the
prompt text, but also add extraction and validation functions that will
be applied to the response of a LLM. Moreover, these functions can send
feedback to the LLM.

With ‘tidyprompt’ and prompt wraps, you can add various features to your
prompts and define how they are evaluated by LLMs. For example:

- **structured output**: Obtain structured output from a LLM, adhering
  to a specific type and/or format. Use pre-built prompt wraps or your
  own R code to validate.

- **feedback & retries**: Automatically provide feedback to a LLM when
  the output is not as expected, allowing the LLM to retry their answer.

- **reasoning modes**: Make a LLM answer a prompt in a specific mode,
  such as chain-of-thought or ReAct (Reasoning and Acting).

- **function calling**: Give a LLM the ability to autonomously call R
  functions (‘tools’). With this, the LLM can retrieve information or
  take other actions. ‘tidyprompt’ also supports R code generation and
  evaluation, allowing LLMs to run R code. Tools from Model Context
  Protocol (MCP) servers are also supported

With its features, ‘tidyprompt’ extends the functionality of LLMs beyond
what is natively offered by LLM APIs, and you can elegantly design
complex, robust interactions with LLMs.

*‘tidyprompt’ is compatible with the
[‘ellmer’](https://ellmer.tidyverse.org/index.html) R package, which has
gained popularity for interfacing with LLM APIs. ‘tidyprompt’ supports
connecting to LLM providers from ‘ellmer’ chat objects, and ‘ellmer’
definitions for structured output and tools (see [more information
below](#how-does-tidyprompt-relate-to-ellmer-and-tidyllm)).*

## Installation

Install the development version from GitHub:

``` r
# install.packages("remotes")
remotes::install_github("KennispuntTwente/tidyprompt")
```

Or install from CRAN (0.3.0):

``` r
install.packages("tidyprompt")
```

## Getting started

See the [‘Getting
started’](https://KennispuntTwente.github.io/tidyprompt/articles/getting_started.html)
vignette for a detailed introduction to using ‘tidyprompt’.

## Examples

Here are some quick examples of what you can do with ‘tidyprompt’:

``` r
"What is 5+5?" |>
  answer_as_integer() |>
  send_prompt(llm_provider_ollama())
#> [1] 10
```

``` r
"Are you a large language model?" |>
  answer_as_boolean() |>
  send_prompt(llm_provider_ollama())
#> [1] TRUE
```

``` r
"What animal is the biggest?" |>
  answer_as_regex_match("^(cat|dog|elephant)$") |>
  send_prompt(llm_provider_ollama())
#> [1] "elephant"
```

``` r
# Make LLM use a function from an R package to search Wikipedia for the answer
"What is something fun that happened in November 2024?" |>
  answer_as_text(max_words = 25) |>
  answer_using_tools(getwiki::search_wiki) |>
  send_prompt(llm_provider_ollama())
#> [1] "The 2024 ARIA Music Awards ceremony, a vibrant celebration of Australian music,
#> took place on November 20, 2024."
```

``` r
# From prompt to linear model object in R
model <- paste0(
  "Using my data, create a statistical model",
  " investigating the relationship between two variables."
) |>
  answer_using_r(
    objects_to_use = list(data = cars),
    evaluate_code = TRUE,
    return_mode = "object"
  ) |>
  prompt_wrap(
    validation_fn = function(x) {
      if (!inherits(x, "lm"))
        return(llm_feedback("The output should be a linear model object."))
      return(TRUE)
    }
  ) |>
  send_prompt(llm_provider_ollama())
summary(model)
#> Call:
#> lm(formula = speed ~ dist, data = data)
#> 
#> Residuals:
#>     Min      1Q  Median      3Q     Max 
#> -7.5293 -2.1550  0.3615  2.4377  6.4179 
#> 
#> Coefficients:
#>             Estimate Std. Error t value Pr(>|t|)    
#> (Intercept)  8.28391    0.87438   9.474 1.44e-12 ***
#> dist         0.16557    0.01749   9.464 1.49e-12 ***
#> ---
#> Signif. codes:  0 '***' 0.001 '**' 0.01 '*' 0.05 '.' 0.1 ' ' 1
#> 
#> Residual standard error: 3.156 on 48 degrees of freedom
#> Multiple R-squared:  0.6511, Adjusted R-squared:  0.6438 
#> F-statistic: 89.57 on 1 and 48 DF,  p-value: 1.49e-12
```

``` r
# Escape validation on questions that cannot be answered
"How many years old is my neighbour's dog?" |>
  answer_as_integer() |>
  quit_if() |>
  send_prompt(llm_provider_ollama())
#> NULL
```

``` r
# LLM in the loop; 
#   LLM verifies answer of LLM and can provide feedback
"What is the capital of France?" |>
  llm_verify() |>
  send_prompt(llm_provider_ollama())
#> ...
  
# Human in the loop; 
#   user verifies answer of LLM and can provide feedback
"What is the capital of France?" |>
  user_verify() |>
  send_prompt(llm_provider_ollama())
#> ...
```

## More information and contributing

‘tidyprompt’ is developed by Luka Koning
(<l.koning@kennispunttwente.nl>) and Tjark van de Merwe
(<t.vandemerwe@kennispunttwente.nl>).

If you encounter issues, have questions, or have suggestions, please
open an issue in the GitHub repository. You are also welcome to
contribute to the package by opening a pull request.

### Why ‘tidyprompt’?

We designed ‘tidyprompt’ because we found ourselves writing code
repeatedly to both construct prompts and handle the associated output of
LLMs; these tasks were intertwined. Often times, we also wanted to add
features to our prompts, or take them away, which required us to rewrite
a lot of code. Thus, we wanted to have building blocks with which we
could easily construct prompts and simultaneously add code to handle the
output of LLMs. This led us to a design inspired by piping syntax, as
popularized by the ‘tidyverse’ and familiar to many R users.

‘tidyprompt’ should be seen as a tool which can be used to enhance the
functionality of LLMs beyond what APIs natively offer. It is designed to
be flexible and provider-agnostic, so that its features can be used with
a wide range of LLM providers and models. It is primarily focused on
‘text-based’ handling of LLMs, where textual output is parsed to achieve
structured output and other functionalities.

Several LLM providers and models also offers forms of ‘native’ handling,
where the LLM is directly controlled by the LLM provider to provide
output in a certain manner. Where appropriate, ‘tidyprompt’ supports
such native configuration of specific APIs. Currently,
`answer_as_json()` and `answer_using_tools()` offer native support for
adhering to JSON schemas and calling functions.

The functions introduced by ‘tidyprompt’ can go beyond what is enforced
by native handling, for example by adding additional validation or
feedback as defined by your R code (i.e., logic and actions which cannot
be captured in just a structured output schema).

### How does ‘tidyprompt’ relate to ‘ellmer’ and ‘tidyllm’?

‘tidyprompt’ is less focused on interfacing with the APIs of various LLM
providers, like R packages ‘ellmer’ and ‘tidyllm’ do. Instead,
‘tidyprompt’ is primarily focused on offering a framework for
constructing prompts and associated logic for interactions with LLMs.

We aim to design ‘tidyprompt’ in such a way that it can be compatible
with ‘ellmer’, ‘tidyllm’, and any other packages offering an interface
to LLM APIs.

‘ellmer’ specifically has surfaced as the most popular R package for
interfacing with LLM APIs. Therefore, ‘tidyprompt’ now supports ‘ellmer’
chat objects as LLM providers. Users can pass an ‘ellmer’ chat object
directly to `send_prompt()`, or use the `llm_provider_ellmer()` function
to create an LLM provider from an ‘ellmer’ chat object. This allows
users to use any LLM provider that can be configured with ‘ellmer’,
including the respective configuration and features from the ‘ellmer’
package.

Furthermore, `answer_as_json()` and `answer_using_tools()` support
‘ellmer’ definitions for structured output and tools. When using an
‘ellmer’ LLM provider, these functions will also call the native
‘ellmer’ functions to obtain structured output and register tools. (And
because `mcptools::mcp_tools()` returns ‘ellmer’ tool definitions,
`answer_using_tools()` also supports tools from MCP servers.)

This means that ‘tidyprompt’ extends what is possible with ‘ellmer’ (and
vice versa), and the two packages can be used together to elegantly
design complex interactions with LLMs.

if (requireNamespace("skimr", quietly = TRUE)) {
  # First add some labels to 'mtcars':
  mtcars$car <- rownames(mtcars)
  mtcars$car <- factor(mtcars$car, levels = rownames(mtcars))
  attr(mtcars$car, "label") <- "Name of the car"

  # Then skim the data:
  mtcars |>
    skim_with_labels_and_levels()
}

\dontrun{
  # Create an in-memory SQLite database
  conn <- DBI::dbConnect(RSQLite::SQLite(), ":memory:")

  # Create a sample table of customers
  DBI::dbExecute(conn, "
  CREATE TABLE
    customers (
      id INTEGER PRIMARY KEY,
      name TEXT,
      email TEXT,
      country TEXT
    );
  ")

  # Insert some sample customer data
  DBI::dbExecute(conn, "
  INSERT INTO
    customers (name, email, country)
  VALUES
    ('Alice', 'alice@example.com', 'USA'),
    ('Bob', 'bob@example.com', 'Canada'),
    ('Charlie', 'charlie@example.com', 'UK'),
    ('Diana', 'diana@example.com', 'USA');
  ")

  # Create another sample table for orders
  DBI::dbExecute(conn, "
  CREATE TABLE orders (
    order_id INTEGER PRIMARY KEY,
    customer_id INTEGER,
    product TEXT,
    amount REAL,
    order_date TEXT,
    FOREIGN KEY(customer_id) REFERENCES customers(id)
  );
  ")

  # Insert some sample orders
  DBI::dbExecute(conn, "
  INSERT INTO
    orders (customer_id, product, amount, order_date)
  VALUES
    (1, 'Widget', 19.99, '2024-01-15'),
    (1, 'Gadget', 29.99, '2024-01-17'),
    (2, 'Widget', 19.99, '2024-02-10'),
    (3, 'SuperWidget', 49.99, '2024-03-05'),
    (4, 'Gadget', 29.99, '2024-04-01'),
    (1, 'Thingamajig', 9.99, '2024-04-02');
  ")

  # Ask LLM a question which it will answer using the SQL database:
  "Where are my customers from?" |>
    answer_using_sql(
      conn = conn,
      evaluate_code = TRUE,
      output_as_tool = TRUE
    ) |>
    send_prompt(llm_provider_openai())
  # --- Sending request to LLM provider (gpt-4o-mini): ---
  # Where are my customers from?
  #
  # You must code in SQL to answer this prompt. You must provide all SQL code
  # between ```sql and ```.
  #
  # Never make assumptions about the possible values in the tables.
  # Instead, execute SQL queries to retrieve information you need.
  #
  # These tables are available in the database:
  #   customers, orders
  #
  # Table descriptions:
  #   - customers
  # Columns: id, name, email, country
  #
  # - orders
  # Columns: order_id, customer_id, product, amount, order_date
  #
  # Your SQL query will be executed on the database. The results will be sent back
  # to you. After seeing the results, you can either provide a final answer or try
  # another SQL query. When you provide your final answer, do not include any SQL code.
  # --- Receiving response from LLM provider: ---
  # ```sql
  # SELECT DISTINCT country FROM customers;
  # ```
  # --- Sending request to LLM provider (gpt-4o-mini): ---
  # --- SQL code: ---
  # SELECT DISTINCT country FROM customers;
  #
  # --- Query results: ---
  #   country
  # 1     USA
  # 2  Canada
  # 3      UK
  # --- Receiving response from LLM provider: ---
  # Based on the query results, your customers are from the following countries:
  # USA, Canada, and UK.
  # [1] "Based on the query results, your customers are from the following countries:
  # USA, Canada, and UK."
}

\dontrun{
  response <- "Hi!" |>
    send_prompt(llm_provider_ollama(), return_mode = "full") |>
    extract_from_return_list("response")
  response
  # [1] "It's nice to meet you. Is there something I can help you with,
  # or would you like to chat?"
}

# Quitting when total token count is exceeded (Google Gemini API example)
\dontrun{
  "How are you?" |>
    # Forcing multi-response via initial error, for demonstration purposes
    answer_as_integer(add_instruction_to_prompt = FALSE) |>
    # Validation function to check total token count
    prompt_wrap(validation_fn = function(response, llm_provider, http_list) {
      total_tokens <- purrr::map_dbl(
        http_list$responses,
        ~ .x$body |>
          rawToChar() |>
          jsonlite::fromJSON() |>
          purrr::pluck("usageMetadata", "totalTokenCount")
      ) |> sum()
      if (total_tokens > 50) {
        warning("Token count exceeded; preventing further interactions")
        # Using llm_break_soft() to prevent further interactions
        return(llm_break_soft(response))
      }
    }) |>
    send_prompt(llm_provider_google_gemini(), return_mode = "full")
}

\dontrun{
  "What is 5 + 5?" |>
    answer_as_integer() |>
    send_prompt()
  # --- Sending request to LLM provider (llama3.1:8b): ---
  #   What is 5 + 5?
  #
  #   You must answer with only an integer (use no other characters).
  # --- Receiving response from LLM provider: ---
  #   10
  # [1] 10
}

ollama <- llm_provider_ollama()

# Add a "short answer" mode (provider-level post prompt wrap)
ollama$add_prompt_wrap(
  provider_prompt_wrap(
    modify_fn = \(txt) paste0(
      txt,
      "\n\nPlease answer concisely (< 2 sentences)."
    )
  ),
  position = "post"
)

# Use as usual: wraps are applied automatically
\dontrun{
"What's a vignette in R?" |> send_prompt(ollama)
}

\dontrun{
  # Prompt to value calculated with R
  avg_miles_per_gallon <- paste0(
    "Using my data,",
    " calculate the average miles per gallon (mpg) for cars with 6 cylinders."
  ) |>
    answer_as_integer() |>
    answer_using_r(
      pkgs_to_use = c("dplyr"),
      objects_to_use = list(mtcars = mtcars),
      evaluate_code = TRUE,
      output_as_tool = TRUE
    ) |>
    send_prompt()
  avg_miles_per_gallon

  # Prompt to linear model object in R
  model <- paste0(
    "Using my data, create a statistical model",
    " investigating the relationship between two variables."
  ) |>
    answer_using_r(
      objects_to_use = list(data = mtcars),
      evaluate_code = TRUE,
      return_mode = "object"
    ) |>
    prompt_wrap(
      validation_fn = function(x) {
        if (!inherits(x, "lm"))
          return(llm_feedback("The output should be a linear model object."))
        return(x)
      }
    ) |>
    send_prompt()
  summary(model)

  # Prompt to plot object in R
  plot <- paste0(
    "Create a scatter plot of miles per gallon (mpg) versus",
    " horsepower (hp) for the cars in my data.",
    " Use different colors to represent the number of cylinders (cyl).",
    " Be very creative and make the plot look nice but also a little crazy!"
  ) |>
    answer_using_r(
      pkgs_to_use = c("ggplot2"),
      objects_to_use = list(mtcars = mtcars),
      evaluate_code = TRUE,
      return_mode = "object"
    ) |>
    send_prompt()
  plot
}

\dontrun{
  "What are some delicious fruits?" |>
    answer_as_list(item_name = "fruit", n_unique_items = 5) |>
    send_prompt()
  # --- Sending request to LLM provider (llama3.1:8b): ---
  # What are some delicious fruits?
  #
  # Respond with a list, like so:
  #   -- <<fruit 1>>
  #   -- <<fruit 2>>
  #   etc.
  # The list should contain 5 unique items.
  # --- Receiving response from LLM provider: ---
  # Here's the list of delicious fruits:
  #   -- Strawberries
  #   -- Pineapples
  #   -- Mangoes
  #   -- Blueberries
  #   -- Pomegranates
  # [[1]]
  # [1] "Strawberries"
  #
  # [[2]]
  # [1] "Pineapples"
  #
  # [[3]]
  # [1] "Mangoes"
  #
  # [[4]]
  # [1] "Blueberries"
  #
  # [[5]]
  # [1] "Pomegranates"
}

# A custom prompt_wrap may be created during piping
prompt <- "Hi there!" |>
  prompt_wrap(
    modify_fn = function(base_prompt) {
      paste(base_prompt, "How are you?", sep = "\n\n")
    }
  )
prompt

# (Shorter notation of the above:)
prompt <- "Hi there!" |>
  prompt_wrap(\(x) paste(x, "How are you?", sep = "\n\n"))

# It may often be preferred to make a function which takes a prompt and
#   returns a wrapped prompt:
my_prompt_wrap <- function(prompt) {
  modify_fn <- function(base_prompt) {
    paste(base_prompt, "How are you?", sep = "\n\n")
  }

  prompt_wrap(prompt, modify_fn)
}
prompt <- "Hi there!" |>
  my_prompt_wrap()

# For more advanced examples, take a look at the source code of the
#   pre-built prompt wraps in the tidyprompt package, like
#   answer_as_boolean, answer_as_integer, add_tools, answer_as_code, etc.
# Below is the source code for the 'answer_as_integer' prompt wrap function:

\dontrun{
  "What is a large language model?" |>
    answer_as_text(max_words = 10) |>
    send_prompt()
  # --- Sending request to LLM provider (llama3.1:8b): ---
  # What is a large language model?
  #
  # You must provide a text response. The response must be at most 10 words.
  # --- Receiving response from LLM provider: ---
  # A type of AI that processes and generates human-like text.
  # [1] "A type of AI that processes and generates human-like text."
}

prompt <- tidyprompt("Hi!")
print(prompt)

# Add to a tidyprompt using a prompt wrap:
prompt <- tidyprompt("Hi!") |>
  add_text("How are you?")
print(prompt)

# Strings can be input for prompt wraps; therefore,
#   a call to tidyprompt() is not necessary:
prompt <- "Hi" |>
  add_text("How are you?")

# Example of adding extraction & validation with a prompt_wrap():
prompt <- "Hi" |>
  add_text("What is 5 + 5?") |>
  answer_as_integer()

\dontrun{
  # tidyprompt objects are evaluated by send_prompt(), which will
  #   handle construct the prompt text, send it to the LLM provider,
  #   and apply the extraction and validation functions from the tidyprompt object
  prompt |>
    send_prompt(llm_provider_ollama())
  # --- Sending request to LLM provider (llama3.1:8b): ---
  #   Hi
  #
  #   What is 5 + 5?
  #
  #   You must answer with only an integer (use no other characters).
  # --- Receiving response from LLM provider: ---
  #   10
  # [1] 10

  # See prompt_wrap() and send_prompt() for more details
}

# `tidyprompt` objects may be validated with these helpers:
is_tidyprompt(prompt) # Returns TRUE if input is a valid tidyprompt object

# Get base prompt text
base_prompt <- prompt$base_prompt

# Get all prompt wraps
prompt_wraps <- prompt$get_prompt_wraps()
# Alternative:
prompt_wraps <- get_prompt_wraps(prompt)

# Construct prompt text
prompt_text <- prompt$construct_prompt_text()
# Alternative:
prompt_text <- construct_prompt_text(prompt)

# Set chat history (affecting also the base prompt)
chat_history <- data.frame(
  role = c("user", "assistant", "user"),
  content = c("What is 5 + 5?", "10", "And what is 5 + 6?")
)
prompt$set_chat_history(chat_history)

# Get chat history
chat_history <- prompt$get_chat_history()

# `answer_as_dataframe()` accepts multiple schema shapes.
# Prefer an ellmer row schema when possible, because it is concise and maps
# cleanly to native ellmer structured output.
# These ellmer schema definitions also work with non-ellmer LLM providers,
# because tidyprompt converts between ellmer schemas and JSON schemas for you.

if (requireNamespace("ellmer", quietly = TRUE)) {
  person_row_schema_ellmer <- ellmer::type_object(
    name = ellmer::type_string(),
    age = ellmer::type_integer(),
    city = ellmer::type_string()
  )

  # Also accepted: an array of row objects.
  person_array_schema_ellmer <- ellmer::type_array(person_row_schema_ellmer)
}

# Also accepted: a JSON schema describing one row.
person_row_schema_json <- list(
  type = "object",
  properties = list(
    name = list(type = "string"),
    age = list(type = "integer"),
    city = list(type = "string")
  ),
  required = c("name", "age", "city"),
  additionalProperties = FALSE
)

# Also accepted: a wrapper object with a `rows` array.
person_wrapper_schema_json <- list(
  type = "object",
  properties = list(
    rows = list(
      type = "array",
      items = person_row_schema_json
    )
  ),
  required = "rows",
  additionalProperties = FALSE
)

\dontrun{
  prompt <- paste(
    "Extract the people in the following notes as a table:",
    "Alice (32, Berlin), Bob (28, Utrecht)."
  )

  # Preferred: ellmer row schema.
  # This works both with ellmer-backed providers and with regular tidyprompt
  # providers, because tidyprompt converts the schema when needed.
  if (requireNamespace("ellmer", quietly = TRUE)) {
    prompt |>
      answer_as_dataframe(person_row_schema_ellmer) |>
      send_prompt()
    #    name age    city
    # 1 Alice  32  Berlin
    # 2   Bob  28 Utrecht

    # Also works: ellmer array-of-rows schema.
    prompt |>
      answer_as_dataframe(person_array_schema_ellmer) |>
      send_prompt()
  }

  # Also works: JSON schema for one row.
  prompt |>
    answer_as_dataframe(person_row_schema_json, type = "text-based") |>
    send_prompt()

  # Also works: JSON wrapper schema with a `rows` array.
  prompt |>
    answer_as_dataframe(person_wrapper_schema_json, type = "text-based") |>
    send_prompt()
}

chat <- "Hi there!" |>
  chat_history()
chat

chat_from_df <- data.frame(
  role = c("user", "assistant"),
  content = c("Hi there!", "Hello! How can I help you today?")
) |>
  chat_history()
chat_from_df

# `add_msg_to_chat_history()` may be used to add messages to a chat history
chat_from_df <- chat_from_df |>
  add_msg_to_chat_history("Calculate 2+2 for me, please!")
chat_from_df

# You can also continue conversations which originate from `send_prompt()`:
\dontrun{
  result <- "Hi there!" |>
    send_prompt(return_mode = "full")
  # --- Sending request to LLM provider (llama3.1:8b): ---
  # Hi there!
  # --- Receiving response from LLM provider: ---
  # It's nice to meet you. Is there something I can help you with, or would you
  # like to chat?

  # Access the chat history from the result:
  chat_from_send_prompt <- result$chat_history

  # Add a message to the chat history:
  chat_history_with_new_message <- chat_from_send_prompt |>
    add_msg_to_chat_history("Let's chat!")

  # The new chat history can be input for a new tidyprompt:
  prompt <- tidyprompt(chat_history_with_new_message)

  # You can also take an existing tidyprompt and add the new chat history to it;
  #   this way, you can continue a conversation using the same prompt wraps
  prompt$set_chat_history(chat_history_with_new_message)

  # send_prompt() also accepts a chat history as input:
  new_result <- chat_history_with_new_message |>
    send_prompt(return_mode = "full")

  # You can also create a persistent chat history object from
  #   a chat history data frame; see ?`persistent_chat-class`
  chat <- `persistent_chat-class`$new(llm_provider_ollama(), chat_from_send_prompt)
  chat$chat("Let's chat!")
}

prompt <- "Hi there!" |>
  set_system_prompt("You are an assistant who always answers in very short poems.")
prompt$system_prompt

\dontrun{
  prompt |>
    send_prompt(llm_provider_ollama())
  # --- Sending request to LLM provider (llama3.1:8b): ---
  #   Hi there!
  # --- Receiving response from LLM provider: ---
  #   Hello to you, I say,
  #   Welcome here, come what may!
  #   How can I assist today?
  # [1] "Hello to you, I say,\nWelcome here, come what may!\nHow can I assist today?"
}

# Example usage within a validation function similar to the one in 'answer_as_integer()':
validation_fn <- function(x, min = 0, max = 100) {
  if (x != floor(x)) {
    # Not a whole number
    return(llm_feedback(
      "You must answer with only an integer (use no other characters)."
    ))
  }
  if (!is.null(min) && x < min) {
    return(llm_feedback(glue::glue(
      "The number should be greater than or equal to {min}."
    )))
  }
  if (!is.null(max) && x > max) {
    return(llm_feedback(glue::glue(
      "The number should be less than or equal to {max}."
    )))
  }
  return(TRUE)
}

# This validation_fn would be part of a prompt_wrap();
#   see the `answer_as_integer()` function for an example of how to use it

\dontrun{
  "What is 5 + 5?" |>
    answer_by_react() |>
    answer_as_integer() |>
    send_prompt()
  # --- Sending request to LLM provider (llama3.1:8b): ---
  # You are given a user's prompt.
  # To answer the user's prompt, you need to think step by step,
  # take an action if needed, and then return a final answer.
  #
  # ----- START OF USER'S PROMPT -----
  # What is 5 + 5?
  #
  # You must answer with only an integer (use no other characters).
  # ----- END OF USER'S PROMPT -----
  #
  # Use the following structure:
  #   Thought: <describe your thought process>
  #   Action: <if needed, describe the action you take (e.g., look up information)>
  #   Observation: <describe the result or observation from the action>
  # (Repeat Thought -> Action -> Observation as necessary)
  #
  # When you are done, you must type:
  #   FINISH[<put here your final answer to the user's prompt>]
  #
  # Ensure your final answer aligns with your reasoning and observations.
  # --- Receiving response from LLM provider: ---
  # Thought: The problem is asking for the sum of two numbers, 5 and 5.
  #
  # Action: None needed, as this is a simple arithmetic operation that can be performed mentally.
  #
  # Observation: I can easily add these two numbers together in my mind to get the result.
  #
  # Thought: To find the sum, I will simply add the two numbers together: 5 + 5 = ?
  #
  # Action: Perform the addition.
  #
  # Observation: The result of adding 5 and 5 is 10.
  #
  # FINISH[10]
  # [1] 10
}

# Various providers:
ollama <- llm_provider_ollama()
openai <- llm_provider_openai()
openrouter <- llm_provider_openrouter()
mistral <- llm_provider_mistral()
groq <- llm_provider_groq()
xai <- llm_provider_xai()
gemini <- llm_provider_google_gemini()

# From an `ellmer::chat()` (e.g., `ellmer::chat_openai()`, ...):
\dontrun{
ellmer <- llm_provider_ellmer(ellmer::chat_openai())
}

# Initialize with settings:
ollama <- llm_provider_ollama(
  parameters = list(
    model = "llama3.2:3b",
    stream = TRUE
  ),
  verbose = TRUE,
  url = "http://localhost:11434/api/chat"
)

# Change settings:
ollama$verbose <- FALSE
ollama$parameters$stream <- FALSE
ollama$parameters$model <- "llama3.1:8b"

\dontrun{
# Try a simple chat message with '$complete_chat()':
response <- ollama$complete_chat("Hi!")
response
# $role
# [1] "assistant"
#
# $content
# [1] "How's it going? Is there something I can help you with or would you like
# to chat?"
#
# $http
# Response [http://localhost:11434/api/chat]
# Date: 2024-11-18 14:21
# Status: 200
# Content-Type: application/json; charset=utf-8
# Size: 375 B

# Use with send_prompt():
"Hi" |>
  send_prompt(ollama)
# [1] "How's your day going so far? Is there something I can help you with or
# would you like to chat?"
}

\dontrun{
  "What would be a suitable e-mail address for cupcake company?" |>
    answer_as_regex_match("^[a-zA-Z0-9._%+-]+@[a-zA-Z0-9.-]+\\.[a-zA-Z]{2,}$") |>
    send_prompt(llm_provider_ollama())
  # --- Sending request to LLM provider (llama3.1:8b): ---
  #   What would be a suitable e-mail address for cupcake company?
  #
  #   You must answer with a response that matches this regex format:
  #     ^[a-zA-Z0-9._%+-]+@[a-zA-Z0-9.-]+\.[a-zA-Z]{2,}$
  #     (use no other characters)
  # --- Receiving response from LLM provider: ---
  #   sweet.treats.cupcakes@gmail.com
  # [1] "sweet.treats.cupcakes@gmail.com"

  "What would be a suitable e-mail address for cupcake company?" |>
    add_text("Give three ideas.") |>
    answer_as_regex_match(
      "[a-zA-Z0-9._%+-]+@[a-zA-Z0-9.-]+\\.[a-zA-Z]{2,}",
      mode = "extract_matches"
    ) |>
    send_prompt(llm_provider_ollama())
  # --- Sending request to LLM provider (llama3.1:8b): ---
  #   What would be a suitable e-mail address for cupcake company?
  #
  #   Give three ideas.
  #
  #   You must answer with a response that matches this regex format:
  #     [a-zA-Z0-9._%+-]+@[a-zA-Z0-9.-]+\.[a-zA-Z]{2,}
  # --- Receiving response from LLM provider: ---
  #   Here are three potential email addresses for a cupcake company:
  #
  #   1. sweettreats.cupcakes@yummail.com
  #   2. cupcakes.and.love@flourpower.net
  #   3. thecupcakery@gmail.com
  # [1] "sweettreats.cupcakes@yummail.com" "cupcakes.and.love@flourpower.net"
  # "thecupcakery@gmail.com"
}

# Create a prompt with a remote image (web URL)
image_prompt <- "What is shown in this image?" |>
  add_image("https://upload.wikimedia.org/wikipedia/commons/3/3a/Cat03.jpg")

# Create a prompt with a local image (file path)
# First save an image to a temporary file
cat_img_file <- tempfile(fileext = ".jpg")
download.file(
  "https://upload.wikimedia.org/wikipedia/commons/3/3a/Cat03.jpg",
  destfile = cat_img_file,
  mode = "wb"
)
# Then build prompt with local image
local_image_prompt <- "What is shown in this image?" |>
  add_image(cat_img_file)

# Send prompt to different LLM providers
# (example is not run because it requires configured LLM providers)
\dontrun{
  # OpenAI-compatible
  send_prompt(image_prompt, llm_provider_openai(parameters = list(model = "gpt-4o-mini")))
  # --- Sending request to LLM provider (gpt-4o-mini): ---
  # What is shown in this image?
  # --- Receiving response from LLM provider: ---
  # The image shows a close-up of an orange tabby cat, characterized by its
  # striped fur and distinctive golden eyes. The background appears blurred,
  # suggesting a softly focused environment.

  # Ollama-compatible
  send_prompt(image_prompt, llm_provider_ollama(parameters = list(model = "qwen3-vl:2b")))
  # ...

  # 'ellmer'-compatible
  send_prompt(image_prompt, llm_provider_ellmer(ellmer::chat_openai(model = "gpt-4o-mini")))
  # ...
}

# Create a prompt with a plot (e.g., 'ggplot2' plot)
\dontrun{
if (requireNamespace("ggplot2", quietly = TRUE)) {
  plot <- ggplot2::ggplot(mtcars, ggplot2::aes(mpg, disp)) +
    ggplot2::geom_point()

  plot_prompt <- "Describe this plot" |>
    add_image(plot)

  send_prompt(plot_prompt, llm_provider_openai())
  # --- Sending request to LLM provider (gpt-4o-mini): ---
  # Describe this plot
  # --- Receiving response from LLM provider: ---
  # The plot is a scatter plot depicting the relationship between two variables:
  # "mpg" (miles per gallon) on the x-axis and "disp" (displacement) on
  # the y-axis. (...)
}
}

# Example usage within an extraction function similar to the one in 'quit_if()':
extraction_fn <- function(x) {
  quit_detect_regex <- "NO ANSWER"

  if (grepl(quit_detect_regex, x)) {
      return(llm_break(
        object_to_return = NULL,
        success = TRUE
      ))
  }

  return(x)
}

\dontrun{
  result <- "How many months old is the cat of my uncle?" |>
    answer_as_integer() |>
    prompt_wrap(
      modify_fn = function(prompt) {
        paste0(
          prompt, "\n\n",
          "Type only 'NO ANSWER' if you do not know."
        )
      },
      extraction_fn = extraction_fn,
      type = "break"
    ) |>
    send_prompt()
  result
  # NULL
}

base_prompt <- "How can I solve 8x + 7 = -23?"

# This example will show how to enforce JSON format in the response,
#   with and without a schema, using the `answer_as_json()` prompt wrap.

# If you use type = 'auto', the function will automatically detect the
#   best way to enforce JSON based on the LLM provider you are using.

# `answer_as_json()` supports two ways of supplying a schema for structured output:
#   - 1) an 'ellmer' definition (e.g., `ellmer::type_object()`;
#         see https://ellmer.tidyverse.org/articles/structured-data.html)
#   - 2) a R list object representing a JSON schema
# `answer_as_json()` will convert the schema type which you supply to any
#   LLM provider; so, whether you use an ellmer LLM provider or another type,
#   you can supply either a R list object or an ellmer definition, and don't
#   have to worry about compatibility
# Supplying a schema as an ellmer definition is likely the easiest

# Below, we will show:
#  - 1) enforcing JSON with a schema; 'ellmer' definition
#  - 2) enforcing JSON with a schema; R list object
# -  3) enforcing JSON without a schema

#### Enforcing JSON with a schema (ellmer definition): ####

# Make an ellmer definition of structured output
#   For instance, a persona:
ellmer_schema <- ellmer::type_object(
  name = ellmer::type_string(),
  age = ellmer::type_integer(),
  hobbies = ellmer::type_array(ellmer::type_string())
)

\dontrun{
  # Example Ellmer LLM provider
  ellmer_openai <- llm_provider_ellmer(ellmer::chat_openai(
    model = "gpt-4.1-mini"
  ))

  # Example regular LLM provider
  tidyprompt_openai <- llm_provider_openai()$set_parameters(
    list(model = "gpt-4.1-mini")
  )

  # You can supply the ellmer definition to both types of LLM provider
  #   to generate an R list object adhering to the schema
  result_ellmer_x_ellmer <- "Create a persona" |>
    answer_as_json(ellmer_schema) |>
    send_prompt(ellmer_openai)

  result_tidyrpompt_x_ellmer <- "Create a persona" |>
    answer_as_json(ellmer_schema) |>
    send_prompt(tidyprompt_openai)
}


#### Enforcing JSON with a schema (R list object definition): ####

# Make a list representing a JSON schema,
#   which the LLM response must adhere to:
json_schema <- list(
  name = "steps_to_solve", # Required for OpenAI API
  description = NULL, # Optional for OpenAI API
  schema = list(
    type = "object",
    properties = list(
      steps = list(
        type = "array",
        items = list(
          type = "object",
          properties = list(
            explanation = list(type = "string"),
            output = list(type = "string")
          ),
          required = c("explanation", "output"),
          additionalProperties = FALSE
        )
      ),
      final_answer = list(type = "string")
    ),
    required = c("steps", "final_answer"),
    additionalProperties = FALSE
  )
  # 'strict' parameter is set as argument 'answer_as_json()'
)
# Note: when you are not using an OpenAI API, you can also pass just the
#   internal 'schema' list object to 'answer_as_json()' instead of the full
#   'json_schema' list object

# Generate example R object based on schema:
r_json_schema_to_example(json_schema)

\dontrun{
  ## Text-based with schema (works for any provider/model):
  #   - Adds request to prompt for a JSON object
  #   - Adds schema to prompt
  #   - Extracts JSON from textual response (feedback for retry if no JSON received)
  #   - Validates JSON against schema with 'jsonvalidate' package (feedback for retry if invalid)
  #   - Parses JSON to R object
  json_4 <- base_prompt |>
    answer_as_json(schema = json_schema) |>
    send_prompt(llm_provider_ollama())
  # --- Sending request to LLM provider (llama3.1:8b): ---
  # How can I solve 8x + 7 = -23?
  #
  # Your must format your response as a JSON object.
  #
  # Your JSON object should match this example JSON object:
  #   {
  #     "steps": [
  #       {
  #         "explanation": "...",
  #         "output": "..."
  #       }
  #     ],
  #     "final_answer": "..."
  #   }
  # --- Receiving response from LLM provider: ---
  # Here is the solution to the equation:
  #
  # ```
  # {
  #   "steps": [
  #     {
  #       "explanation": "First, we want to isolate the term with 'x' by
  #       subtracting 7 from both sides of the equation.",
  #       "output": "8x + 7 - 7 = -23 - 7"
  #     },
  #     {
  #       "explanation": "This simplifies to: 8x = -30",
  #       "output": "8x = -30"
  #     },
  #     {
  #       "explanation": "Next, we want to get rid of the coefficient '8' by
  #       dividing both sides of the equation by 8.",
  #       "output": "(8x) / 8 = (-30) / 8"
  #     },
  #     {
  #       "explanation": "This simplifies to: x = -3.75",
  #       "output": "x = -3.75"
  #     }
  #   ],
  #   "final_answer": "-3.75"
  # }
  # ```

  ## Ollama with schema:
  #   - Sets 'format' parameter to 'json', enforcing JSON
  #   - Adds request to prompt for a JSON object, as is recommended by the docs
  #   - Adds schema to prompt
  #   - Validates JSON against schema with 'jsonvalidate' package (feedback for retry if invalid)
  json_5 <- base_prompt |>
    answer_as_json(json_schema, type = "auto") |>
    send_prompt(llm_provider_ollama())
  # --- Sending request to LLM provider (llama3.1:8b): ---
  # How can I solve 8x + 7 = -23?
  #
  # Your must format your response as a JSON object.
  #
  # Your JSON object should match this example JSON object:
  # {
  #   "steps": [
  #     {
  #       "explanation": "...",
  #       "output": "..."
  #     }
  #   ],
  #   "final_answer": "..."
  # }
  # --- Receiving response from LLM provider: ---
  # {
  #   "steps": [
  #     {
  #       "explanation": "First, subtract 7 from both sides of the equation to
  #       isolate the term with x.",
  #       "output": "8x = -23 - 7"
  #     },
  #     {
  #       "explanation": "Simplify the right-hand side of the equation.",
  #       "output": "8x = -30"
  #     },
  #     {
  #       "explanation": "Next, divide both sides of the equation by 8 to solve for x.",
  #       "output": "x = -30 / 8"
  #     },
  #     {
  #       "explanation": "Simplify the right-hand side of the equation.",
  #       "output": "x = -3.75"
  #     }
  #   ],
  #   "final_answer": "-3.75"
  # }

  ## OpenAI with schema:
  #   - Sets 'response_format' parameter to 'json_object', enforcing JSON
  #   - Adds json_schema to the API request, API enforces JSON adhering schema
  #   - Parses JSON to R object
  json_6 <- base_prompt |>
    answer_as_json(json_schema, type = "auto") |>
    send_prompt(llm_provider_openai())
  # --- Sending request to LLM provider (gpt-4o-mini): ---
  # How can I solve 8x + 7 = -23?
  # --- Receiving response from LLM provider: ---
  # {"steps":[
  # {"explanation":"Start with the original equation.",
  # "output":"8x + 7 = -23"},
  # {"explanation":"Subtract 7 from both sides to isolate the term with x.",
  # "output":"8x + 7 - 7 = -23 - 7"},
  # {"explanation":"Simplify the left side and the right side of the equation.",
  # "output":"8x = -30"},
  # {"explanation":"Now, divide both sides by 8 to solve for x.",
  # "output":"x = -30 / 8"},
  # {"explanation":"Simplify the fraction by dividing both the numerator and the
  # denominator by 2.",
  # "output":"x = -15 / 4"}
  # ], "final_answer":"x = -15/4"}

  # You can also use the R list object schema definition with an
  #  ellmer LLM provider; `answer_as_json()` will do the conversion for you
  json_7 <- base_prompt |>
    answer_as_json(json_schema) |>
    send_prompt(ellmer_openai)
}


#### Enforcing JSON without a schema: ####

\dontrun{
  ## Text-based (works for any provider/model):
  #   Adds request to prompt for a JSON object
  #   Extracts JSON from textual response (feedback for retry if no JSON received)
  #   Parses JSON to R object
  json_1 <- base_prompt |>
    answer_as_json() |>
    send_prompt(llm_provider_ollama())
  # --- Sending request to LLM provider (llama3.1:8b): ---
  # How can I solve 8x + 7 = -23?
  #
  # Your must format your response as a JSON object.
  # --- Receiving response from LLM provider: ---
  # Here is the solution to the equation formatted as a JSON object:
  #
  # ```
  # {
  #   "equation": "8x + 7 = -23",
  #   "steps": [
  #     {
  #       "step": "Subtract 7 from both sides of the equation",
  #       "expression": "-23 - 7"
  #     },
  #     {
  #       "step": "Simplify the expression on the left side",
  #       "result": "-30"
  #     },
  #     {
  #       "step": "Divide both sides by -8 to solve for x",
  #       "expression": "-30 / -8"
  #     },
  #     {
  #       "step": "Simplify the expression on the right side",
  #       "result": "3.75"
  #     }
  #   ],
  #   "solution": {
  #     "x": 3.75
  #   }
  # }
  # ```


  ## Ollama:
  #   - Sets 'format' parameter to 'json', enforcing JSON
  #   - Adds request to prompt for a JSON object, as is recommended by the docs
  #   - Parses JSON to R object
  json_2 <- base_prompt |>
    answer_as_json(type = "auto") |>
    send_prompt(llm_provider_ollama())
  # --- Sending request to LLM provider (llama3.1:8b): ---
  # How can I solve 8x + 7 = -23?
  #
  # Your must format your response as a JSON object.
  # --- Receiving response from LLM provider: ---
  # {"steps": [
  #   "Subtract 7 from both sides to get 8x = -30",
  #   "Simplify the right side of the equation to get 8x = -30",
  #   "Divide both sides by 8 to solve for x, resulting in x = -30/8",
  #   "Simplify the fraction to find the value of x"
  # ],
  # "value_of_x": "-3.75"}


  ## OpenAI-type API without schema:
  #   - Sets 'response_format' parameter to 'json_object', enforcing JSON
  #   - Adds request to prompt for a JSON object, as is required by the API
  #   - Parses JSON to R object
  json_3 <- base_prompt |>
    answer_as_json(type = "auto") |>
    send_prompt(llm_provider_openai())
  # --- Sending request to LLM provider (gpt-4o-mini): ---
  # How can I solve 8x + 7 = -23?
  #
  # Your must format your response as a JSON object.
  # --- Receiving response from LLM provider: ---
  # {
  #   "solution_steps": [
  #     {
  #       "step": 1,
  #       "operation": "Subtract 7 from both sides",
  #       "equation": "8x + 7 - 7 = -23 - 7",
  #       "result": "8x = -30"
  #     },
  #     {
  #       "step": 2,
  #       "operation": "Divide both sides by 8",
  #       "equation": "8x / 8 = -30 / 8",
  #       "result": "x = -3.75"
  #     }
  #   ],
  #   "solution": {
  #     "x": -3.75
  #   }
  # }
}

\dontrun{
  persona <- "Create a persona for me, please." |>
    answer_as_named_list(
      item_names = c("name", "age", "occupation"),
      item_instructions = list(
        name = "The name of the persona",
        age = "The age of the persona",
        occupation = "The occupation of the persona"
      )
    ) |> send_prompt(llm_provider_ollama())
  # --- Sending request to LLM provider (llama3.1:8b): ---
  #   Create a persona for me, please.
  #
  #   Respond with a named list like so:
  #     -- name: <<value>> (The name of the persona)
  #     -- age: <<value>> (The age of the persona)
  #     -- occupation: <<value>> (The occupation of the persona)
  #   Each name must correspond to: name, age, occupation
  # --- Receiving response from LLM provider: ---
  #   Here is your persona:
  #
  #   -- name: Astrid Welles
  #   -- age: 32
  #   -- occupation: Museum Curator
  persona$name
  # [1] "Astrid Welles"
  persona$age
  # [1] "32"
  persona$occupation
  # [1] "Museum Curator"
}

\dontrun{
  "It is sunny, that makes me happy." |>
    answer_as_category(categories = c("environment", "weather", "work")) |>
    send_prompt()
  # --- Sending request to LLM provider (llama3.1:8b): ---
  #   You need to categorize a text.
  #
  # Text:
  #   'It is sunny, that makes me happy.'
  #
  # Possible categories:
  #   1. environment
  #   2. weather
  #   3. work
  #
  # Respond with the number of the category that best describes the text.
  # (Use no other words or characters.)
  # --- Receiving response from LLM provider: ---
  #   2
  # [1] "weather"
}

## ----eval=FALSE---------------------------------------------------------------
# prompt <- "Hi there!" |>
#   prompt_wrap(
#     modify_fn = function(base_prompt) {
#       paste(base_prompt, "How are you?", sep = "\n\n")
#     }
#   )

## ----eval=FALSE---------------------------------------------------------------
# prompt <- "Hi there!" |>
#   prompt_wrap(\(x) paste(x, "How are you?", sep = "\n\n"))

## ----eval=FALSE---------------------------------------------------------------
# my_prompt_wrap <- function(prompt) {
#   modify_fn <- function(base_prompt) {
#     paste(base_prompt, "How are you?", sep = "\n\n")
#   }
# 
#   prompt_wrap(prompt, modify_fn)
# }
# prompt <- "Hi there!" |>
#   my_prompt_wrap()

## ----eval=FALSE---------------------------------------------------------------
# answer_as_boolean <- function(
#     prompt,
#     true_definition = NULL,
#     false_definition = NULL,
#     add_instruction_to_prompt = TRUE
# ) {
#   instruction <- "You must answer with only TRUE or FALSE (use no other characters)."
#   if (!is.null(true_definition))
#     instruction <- paste(instruction, glue::glue("TRUE means: {true_definition}."))
#   if (!is.null(false_definition))
#     instruction <- paste(instruction, glue::glue("FALSE means: {false_definition}."))
# 
#   modify_fn <- function(original_prompt_text) {
#     if (!add_instruction_to_prompt) {
#       return(original_prompt_text)
#     }
# 
#     glue::glue("{original_prompt_text}\n\n{instruction}")
#   }
# 
#   extraction_fn <- function(x) {
#     normalized <- tolower(trimws(x))
#     if (normalized %in% c("true", "false")) {
#       return(as.logical(normalized))
#     }
#     return(llm_feedback(instruction))
#   }
# 
#   prompt_wrap(prompt, modify_fn, extraction_fn)
# }

\dontrun{
  "What the favourite food of my cat on Thursday mornings?" |>
    quit_if() |>
    send_prompt(llm_provider_ollama())
  # --- Sending request to LLM provider (llama3.1:8b): ---
  #   What the favourite food of my cat on Thursday mornings?
  #
  #   If you think that you cannot provide a valid answer, you must type:
  #   'NO ANSWER' (use no other characters)
  # --- Receiving response from LLM provider: ---
  #   NO ANSWER
  # NULL
}

\dontrun{
  "Are you a large language model?" |>
    answer_as_boolean() |>
    send_prompt()
  # --- Sending request to LLM provider (llama3.1:8b): ---
  #   Are you a large language model?
  #
  #   You must answer with only TRUE or FALSE (use no other characters).
  # --- Receiving response from LLM provider: ---
  #   TRUE
  # [1] TRUE
}

# When using functions from base R or R packages,
#   documentation is automatically extracted from help files:
prompt_with_dir_function <- "What are the files in my current directory?" |>
  answer_using_tools(dir) # The 'dir' function is from base R
\dontrun{
  send_prompt(prompt_with_dir_function)
  # --- Sending request to LLM provider (llama3.1:8b): ---
  #   What are the files in my current directory?
  #   --- Receiving response from LLM provider: ---
  #   Calling function 'nm' with arguments:
  #   {
  #     "all.files": true,
  #     "full.names": false,
  #     "ignore.case": false,
  #     "include.dirs": false,
  #     "no..": false,
  #     "path": "./",
  #     "pattern": "*",
  #     "recursive": false
  #   }
  # Result:
  #   .git, .github, .gitignore, .Rbuildignore, .Rhistory, ...
  # The files in your current directory are:
  #   .git, .github, .gitignore, .Rbuildignore, .Rhistory, ...
  # [1] "The files in your current directory are:\n\n .git, .github, ..."
}

# Users may provide custom functions in two ways:
#   1) as a function object, optionally documented with `tools_get_docs()`, or
#   2) as an 'ellmer' tool definition, using `ellmer::tool()`

# Take this fake weather function as an example:
temperature_in_location <- function(
  location = c("Amsterdam", "Utrecht", "Enschede"),
  unit = c("Celcius", "Fahrenheit")
) {
  location <- match.arg(location)
  unit <- match.arg(unit)

  temperature_celcius <- switch(
    location,
    "Amsterdam" = 32.5,
    "Utrecht" = 19.8,
    "Enschede" = 22.7
  )

  if (unit == "Celcius") {
    return(temperature_celcius)
  } else {
    return(temperature_celcius * 9/5 + 32)
  }
}


# 1: `tools_add_docs()` --------------------------------------------------------

# Generate documentation for a function, based on formals & help file
docs <- tools_get_docs(temperature_in_location)

# The types get inferred from the function's formals
# However, descriptions are still missing as the function is not from a package
# We can modify the documentation object to add descriptions:
docs$description <- "Get the temperature in a location"
docs$arguments$unit$description <- "Unit in which to return the temperature"
docs$arguments$location$description <- "Location for which to return the temperature"
docs$return$description <- "The temperature in the specified location and unit"
# (See `?tools_add_docs` for more details on the structure of the documentation)

# When we are satisfied with the documentation, we can add it to the function:
temperature_in_location <- tools_add_docs(temperature_in_location, docs)

prompt_with_weather_function <-
  "What is the weather in Enschede? Give me Celcius degrees" |>
  answer_using_tools(temperature_in_location)
\dontrun{
  send_prompt(prompt_with_weather_function)
  # --- Sending request to LLM provider (llama3.1:8b): ---
  #   What is the weather in Enschede? Give me Celcius degrees
  #   --- Receiving response from LLM provider: ---
  #   Calling function 'temperature_in_location' with arguments:
  #   {
  #     "location": "Enschede",
  #     "unit": "Celcius"
  #   }
  # Result:
  #   22.7
  # The temperature in Enschede is 22.7 Celcius degrees.
  # [1] "The temperature in Enschede is 22.7 Celcius degrees."
}

# 2: `ellmer::tool()` -----------------------------------------------

# Alternatively, we can define the function as an 'ellmer' tool

temperature_in_location_ellmer <- ellmer::tool(
  temperature_in_location,
  name = "get_temperature",
  description = "Get the temperature in a location",
  arguments = list(
    location = ellmer::type_string(
      "Location for which to return the temperature", required = TRUE
    ),
    unit = ellmer::type_string(
      "Unit in which to return the temperature", required = TRUE
    )
  )
)

prompt_with_weather_function_ellmer <-
  "What is the weather in Utrecht? Give me Fahrenheit degrees" |>
  answer_using_tools(temperature_in_location_ellmer)
\dontrun{
  send_prompt(prompt_with_weather_function_ellmer)
  # ...
}

# Because `mcptools::mcp_tools()` returns a list of `ellmer:tool()` tools,
#   you can also use Model Context Protocol (MCP) server tools with
#   `answer_using_tools()`:
\dontrun{
  prompt_using_mcp_tools <- mcptools::mcp_tools()
  "Push my latest commit to GitHub" |>
    answer_using_tools(mcp_tools)
  send_prompt(prompt_using_mcp_tools)
}

# `answer_using_tools()` will automatically attempt to use the most appropriate
#   way of sending the tool to the LLM

# If you use a LLM provider of type 'ollama' or 'openai',
#   it will automatically convert the tool definition to parameters
#   appropriate for those APIs
# If you use a LLM provider of type 'ellmer', it will call the appropriate
#   ellmer function directly which will handle the tool call for various
#   providers
# Note that both tool definitions from `tools_add_docs()` and `ellmer::tool()`
#   will work with any LLM provider; `answer_using_tools()` can convert
#   the two types of tool definitions to each other when needed
\dontrun{
  ollama <- llm_provider_ollama()
  # Ollama LLM provider:
  "What is the weather in Amsterdam? Give me Fahrenheit degrees" |>
    answer_using_tools(temperature_in_location) |>
    send_prompt(ollama)

  # Ollama LLM provider also works with `ellmer::tool()` definitions:
  "What is the weather in Amsterdam? Give me Celcius degrees" |>
    answer_using_tools(temperature_in_location_ellmer) |>
    send_prompt(ollama)

  # Similar for OpenAI API:
  openai <- llm_provider_openai()
  "What is the weather in Amsterdam? Give me Celcius degrees" |>
    answer_using_tools(temperature_in_location) |>
    send_prompt(openai)
  # ...

  # Ellmer LLM provider:
  ellmer <- llm_provider_ellmer(ellmer::chat_openai())
  "What is the weather in Amsterdam? Give me Celcius degrees" |>
    answer_using_tools(temperature_in_location_ellmer) |>
    send_prompt(ellmer)

  # Also works with `tools_add_docs()` definition:
  "What is the weather in Amsterdam? Give me Celcius degrees" |>
    answer_using_tools(temperature_in_location) |>
    send_prompt(ellmer)
}

\dontrun{
  "Tell me a fun fact about yourself!" |>
    user_verify() |>
    send_prompt()
  # --- Sending request to LLM provider (gpt-4o-mini): ---
  # Tell me a fun fact about yourself!
  # --- Receiving response from LLM provider: ---
  # I don't have personal experiences or feelings, but a fun fact about me is that
  # I can generate text in multiple languages! From English to Spanish, French, and
  # more, I'm here to help with diverse linguistic needs.
  #
  # --- Evaluation of tidyprompt resulted in:
  # [1] "I don't have personal experiences or feelings, but a fun fact about me is
  # that I can generate text in multiple languages! From English to Spanish, French,
  # and more, I'm here to help with diverse linguistic needs."
  #
  # --- Accept or decline
  # * If satisfied, type nothing
  # * If not satisfied, type feedback to the LLM
  # Type: Needs to be funnier!
  # --- Sending request to LLM provider (gpt-4o-mini): ---
  # Needs to be funnier!
  # --- Receiving response from LLM provider: ---
  # Alright, how about this: I once tried to tell a joke, but my punchline got lost
  # in translation! Now, I just stick to delivering 'byte-sized' humor!
  #
  # --- Evaluation of tidyprompt resulted in:
  # [1] "Alright, how about this: I once tried to tell a joke, but my punchline got
  # lost in translation! Now, I just stick to delivering 'byte-sized' humor!"
  #
  # --- Accept or decline
  # * If satisfied, type nothing
  # * If not satisfied, type feedback to the LLM
  # Type:
  # * Result accepted
  # [1] "Alright, how about this: I once tried to tell a joke, but my punchline got
  # lost in translation! Now, I just stick to delivering 'byte-sized' humor!"
}

\dontrun{
  "It is sunny, that makes me happy." |>
    answer_as_multi_category(
      categories = c("environment", "weather", "work", "positive", "negative")
    ) |>
    send_prompt()
  # --- Sending request to LLM provider (llama3.1:8b): ---
  #   You need to categorize a text.
  #
  # Text:
  #   'It is sunny, that makes me happy.'
  #
  # Possible categories:
  #   1. environment
  #   2. weather
  #   3. work
  #   4. positive
  #   5. negative
  #
  # Respond with the numbers of all categories that apply to this text, separated by commas.
  # (Use no other words or characters.)
  # --- Receiving response from LLM provider: ---
  #   2, 4
  # ["weather","positive"]
}

\dontrun{
  "What is 5 + 5?" |>
    answer_by_chain_of_thought() |>
    answer_as_integer() |>
    send_prompt()
  # --- Sending request to LLM provider (llama3.1:8b): ---
  # You are given a user's prompt.
  # To answer the user's prompt, you need to think step by step to arrive at a final answer.
  #
  # ----- START OF USER'S PROMPT -----
  # What is 5 + 5?
  #
  # You must answer with only an integer (use no other characters).
  # ----- END OF USER'S PROMPT -----
  #
  # What are the steps you would take to answer the user's prompt?
  # Describe your thought process in the following format:
  #   >> step 1: <step 1 description>
  #   >> step 2: <step 2 description>
  #   (etc.)
  #
  # When you are done, you must type:
  #   FINISH[<put here your final answer to the user's prompt>]
  #
  # Make sure your final answer follows the logical conclusion of your thought process.
  # --- Receiving response from LLM provider: ---
  # >> step 1: Recognize that the user's prompt is a simple arithmetic question asking for
  # the sum of two numbers.
  #
  # >> step 2: Identify the specific numbers mentioned in the prompt, which are 5 and 5.
  #
  # >> step 3: Recall basic addition facts from memory or knowledge.
  # In this case, knowing that 5 + 0 = 5, 5 + 1 = 6, 5 + 2 = 7, and so on up to 5 + 4 = 9,
  # it is clear that the next step in adding 5 to itself would be to keep the same sum.
  #
  # >> step 4: Apply the addition principle by considering the original number (5) as being
  # added to itself once more. Since 5 + 1 = 6 and the process of adding 1 increases the total
  # sum, then 5 + 2 must equal 7, thus indicating a direct progression in the sum as the additive
  # amount increases.
  #
  # >> step 5: From steps 3 and 4, conclude that since 5 added to itself (once) would be 10,
  # then when it is added again (to itself), it would continue from the last result which was
  # achieved by adding one more than the original number. Therefore, it must follow the pattern
  # established in step 3 where each additional amount increases the total sum.
  #
  # >> step 6: Now knowing that the operation of adding a second 5 to the first one results in
  # increasing the sum to the next integer after 10 (based on steps 3 and 4), determine what the
  # next integer is after 10. Knowing that integers are whole numbers without fractions, it's
  # clear that if you add another unit to 10, it becomes 11.
  #
  # FINISH[10]
  # [1] 10
}

## ----provider_wraps, cache=FALSE, eval=FALSE----------------------------------
# ollama <- llm_provider_ollama()
# 
# # Example: add a "short answer" mode (provider-level post prompt wrap)
# ollama$add_prompt_wrap(
#   provider_prompt_wrap(
#     modify_fn = \(txt) paste0(
#       txt,
#       "\n\nPlease answer concisely (< 2 sentences)."
#     )
#   ),
#   position = "post"
# )
# 
# # Use as usual: wraps are applied automatically
# "What's a vignette in R?" |> send_prompt(ollama)

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## ----setup, eval = FALSE------------------------------------------------------
# # Install & load required packages
# packages <- c("shiny", "ipc", "future", "promises", "tidyprompt")
# for (pkg in packages) {
#   if (!requireNamespace(pkg, quietly = TRUE)) {
#     install.packages(pkg)
#   }
#   library(pkg, character.only = TRUE)
# }
# 
# # Enable asynchronous processing
# future::plan(future::multisession)
# 
# # Base provider (OpenAI; has streaming enabled by default)
# base_provider <- llm_provider_openai()
# 
# ui <- fluidPage(
#   titlePanel("tidyprompt streaming demo"),
# 
#   sidebarLayout(
#     sidebarPanel(
#       textInput(
#         "prompt",
#         "Prompt",
#         value = "Tell me a short story about a cat and a robot."
#       ),
#       actionButton("run", "Ask model"),
#       helpText("Tokens will appear below as they stream in.")
#     ),
# 
#     mainPanel(
#       verbatimTextOutput("partial_response")
#     )
#   )
# )
# 
# server <- function(input, output, session) {
#   # Queue to bridge async future back into Shiny
#   queue <- shinyQueue()
#   queue$consumer$start(100) # process queue every 100 ms
# 
#   # Reactive that holds the accumulated streamed text
#   partial_response <- reactiveVal("")
# 
#   # Streaming callback run inside the provider
#   stream_cb <- function(token, meta) {
#     # meta$partial_response is the accumulated text so far
#     queue$producer$fireAssignReactive(
#       "partial_response",
#       meta$partial_response
#     )
#     invisible(TRUE)
#   }
# 
#   # Clone provider for this session and attach callback + streaming
#   provider <- base_provider$clone()
#   provider$parameters$stream <- TRUE
#   provider$stream_callback <- stream_cb
# 
#   # Expose the reactive value to the UI
#   output$partial_response <- renderText({
#     req(partial_response())
#     partial_response()
#   })
# 
#   observeEvent(input$run, {
#     # Reset current streamed text on each run
#     partial_response("")
# 
#     user_prompt <- input$prompt
# 
#     future_promise(
#       {
#         tidyprompt::send_prompt(
#           prompt = user_prompt,
#           llm_provider = provider,
#           return_mode = "only_response"
#         )
#       },
#       globals = list(
#         user_prompt = user_prompt,
#         provider = provider
#       )
#     ) |>
#       then(
#         onFulfilled = function(value) {
#           # Final response once streaming finishes
#           partial_response(value)
#         },
#         onRejected = function(error) {
#           showNotification(
#             paste("Error:", error$message),
#             type = "error"
#           )
#           print(error)
#         }
#       )
# 
#     NULL
#   })
# }
# 
# shinyApp(ui = ui, server = server)

\dontrun{
  "What are a few capital cities around the world?" |>
    answer_as_key_value(
      key_name = "country",
      value_name = "capital"
    ) |>
    send_prompt()
  # --- Sending request to LLM provider (llama3.1:8b): ---
  # What are a few capital cities around the world?
  #
  # Respond with a list of key-value pairs, like so:
  #   -- <<country 1>>: <<capital 1>>
  #   -- <<country 2>>: <<capital 2>>
  #   etc.
  # --- Receiving response from LLM provider: ---
  # Here are a few:
  #   -- Australia: Canberra
  #   -- France: Paris
  #   -- United States: Washington D.C.
  #   -- Japan: Tokyo
  #   -- China: Beijing
  # $Australia
  # [1] "Canberra"
  #
  # $France
  # [1] "Paris"
  #
  # $`United States`
  # [1] "Washington D.C."
  #
  # $Japan
  # [1] "Tokyo"
  #
  # $China
  # [1] "Beijing"
}

\dontrun{
  "What is 10 divided by 4?" |>
    answer_as_numeric() |>
    send_prompt()
  # --- Sending request to LLM provider (llama3.1:8b): ---
  #   What is 10 divided by 4?
  #
  #   You must answer with only a number (use no other characters).
  # --- Receiving response from LLM provider: ---
  #   2.5
  # [1] 2.5
}

\dontrun{
  "Hi!" |>
    send_prompt(llm_provider_ollama())
  # --- Sending request to LLM provider (llama3.1:8b): ---
  #   Hi!
  # --- Receiving response from LLM provider: ---
  #   It's nice to meet you. Is there something I can help you with, or would you like to chat?
  # [1] "It's nice to meet you. Is there something I can help you with, or would you like to chat?"

  "Hi!" |>
    send_prompt(llm_provider_ollama(), return_mode = "full")
  # --- Sending request to LLM provider (llama3.1:8b): ---
  #   Hi!
  # --- Receiving response from LLM provider: ---
  #   It's nice to meet you. Is there something I can help you with, or would you like to chat?
  # $response
  # [1] "It's nice to meet you. Is there something I can help you with, or would you like to chat?"
  #
  # $chat_history
  # ...
  #
  # $chat_history_clean
  # ...
  #
  # $start_time
  # [1] "2024-11-18 15:43:12 CET"
  #
  # $end_time
  # [1] "2024-11-18 15:43:13 CET"
  #
  # $duration_seconds
  # [1] 1.13276
  #
  # $http
  # $http$responses[[1]]
  # Response [http://localhost:11434/api/chat]
  #   Date: 2024-11-18 14:43
  #   Status: 200
  #   Content-Type: application/x-ndjson
  # <EMPTY BODY>

  "Hi!" |>
    add_text("What is 5 + 5?") |>
    answer_as_integer() |>
    send_prompt(llm_provider_ollama(), verbose = FALSE)
  # [1] 10
}

# Create a persistent chat with any LLM provider
chat <- `persistent_chat-class`$new(llm_provider_ollama())

\dontrun{
  chat$chat("Hi! Tell me about Twente, in a short sentence?")
  # --- Sending request to LLM provider (llama3.1:8b): ---
  # Hi! Tell me about Twente, in a short sentence?
  # --- Receiving response from LLM provider: ---
  # Twente is a charming region in the Netherlands known for its picturesque
  # countryside and vibrant culture!

  chat$chat("How many people live there?")
  # --- Sending request to LLM provider (llama3.1:8b): ---
  # How many people live there?
  # --- Receiving response from LLM provider: ---
  # The population of Twente is approximately 650,000 inhabitants, making it one of
  # the largest regions in the Netherlands.

  # Access the chat history:
  chat$chat_history

  # Reset the chat history:
  chat$reset_chat_history()

  # Continue a chat from the result of `send_prompt()`:
  result <- "Hi there!" |>
    answer_as_integer() |>
    send_prompt(return_mode = "full")
  # --- Sending request to LLM provider (llama3.1:8b): ---
  # Hi there!
  #
  # You must answer with only an integer (use no other characters).
  # --- Receiving response from LLM provider: ---
  # 42
  chat <- `persistent_chat-class`$new(llm_provider_ollama(), result$chat_history)
  chat$chat("Why did you choose that number?")
  # --- Sending request to LLM provider (llama3.1:8b): ---
  # Why did you choose that number?
  # --- Receiving response from LLM provider: ---
  # I chose the number 42 because it's a reference to Douglas Adams' science fiction
  # series "The Hitchhiker's Guide to the Galaxy," in which a supercomputer named
  # Deep Thought is said to have calculated the "Answer to the Ultimate Question of
  # Life, the Universe, and Everything" as 42.
}

# Example creation of a llm_provider-class object:
llm_provider_openai <- function(
    parameters = list(
      model = "gpt-4o-mini",
      stream = getOption("tidyprompt.stream", TRUE)
    ),
    verbose = getOption("tidyprompt.verbose", TRUE),
    url = "https://api.openai.com/v1/chat/completions",
    api_key = Sys.getenv("OPENAI_API_KEY")
) {
  complete_chat <- function(chat_history) {
    headers <- c(
      "Content-Type" = "application/json",
      "Authorization" = paste("Bearer", self$api_key)
    )

    body <- list(
      messages = lapply(seq_len(nrow(chat_history)), function(i) {
        list(role = chat_history$role[i], content = chat_history$content[i])
      })
    )

    for (name in names(self$parameters))
      body[[name]] <- self$parameters[[name]]

    request <- httr2::request(self$url) |>
      httr2::req_body_json(body) |>
      httr2::req_headers(!!!headers)

    request_llm_provider(
      chat_history,
      request,
      stream = self$parameters$stream,
      verbose = self$verbose,
      api_type = self$api_type
    )
  }

  return(`llm_provider-class`$new(
    complete_chat_function = complete_chat,
    parameters = parameters,
    verbose = verbose,
    url = url,
    api_key = api_key,
    api_type = "openai"
  ))
}

llm_provider <- llm_provider_openai()

\dontrun{
  llm_provider$complete_chat("Hi!")
  # --- Sending request to LLM provider (gpt-4o-mini): ---
  # Hi!
  # --- Receiving response from LLM provider: ---
  # Hello! How can I assist you today?
}

\dontrun{
  "What is 'Enschede'?!" |>
    answer_as_text(max_words = 50) |>
    llm_verify() |>
    send_prompt()
#   --- Sending request to LLM provider (gpt-4o-mini): ---
#   What is 'Enschede'?!
#
#   You must provide a text response. The response must be at most 50 words.
#   --- Receiving response from LLM provider: ---
#   Enschede is a city in the Netherlands, located in the eastern part near the German border.
#   It is known for its vibrant culture, history, and universities, particularly the
#   University of Twente, as well as its textiles industry and beautiful parks.
#   --- Sending request to LLM provider (gpt-4o-mini): ---
#   You are given a user's prompt.
#   To answer the user's prompt, you need to think step by step to arrive at a final answer.
#
#   ----- START OF USER'S PROMPT -----
#   >>> An assistant was asked:
#
#   What is 'Enschede'?!
#
#   >>> The assistant answered:
#
#   [1] "Enschede is a city in the Netherlands, located in the eastern part near
#   the German border. It is known for its vibrant culture, history, and universities,
#   particularly the University of Twente, as well as its textiles industry and
#   beautiful parks."
#
#   >>> Is the answer satisfactory?
#
#   You must answer with only TRUE or FALSE (use no other characters).
#   ----- END OF USER'S PROMPT -----
#
#   What are the steps you would take to answer the user's prompt?
#   Describe your thought process in the following format:
#     >> step 1: <step 1 description>
#     >> step 2: <step 2 description>
#     (etc.)
#
#   When you are done, you must type:
#     FINISH[<put here your final answer to the user's prompt>]
#
#   Make sure your final answer follows the logical conclusion of your thought process.
#   --- Receiving response from LLM provider: ---
#   >> step 1: Read the user's prompt carefully to understand what is being asked
#   regarding the assistant's answer about Enschede.
#   >> step 2: Analyze the assistant's provided answer for accuracy and completeness,
#   ensuring it conveys relevant information about Enschede.
#   >> step 3: Assess whether the answer includes key aspects, such as the city's
#   location, cultural significance, educational institutions, and any notable
#   historical references or industries tied to the city.
#   >> step 4: Determine if the assistant's answer sufficiently addresses the
#   question "What is Enschede?" and provides a clear and informative overview of the city.
#   >> step 5: Conclude whether the assistant's answer meets the standards for being
#   satisfactory in terms of detail and correctness.
#
#   FINISH[TRUE]
#   [1] "Enschede is a city in the Netherlands, located in the eastern part near the
#   German border. It is known for its vibrant culture, history, and universities,
#   particularly the University of Twente, as well as its textiles industry and beautiful parks."
}

#' Make LLM answer as a list of items
#'
#' @param prompt A single string or a [tidyprompt()] object
#' @param item_name (optional) Name of the items in the list
#' @param item_explanation (optional) Additional explanation of what an item
#' should be. Item explanation should be a single string. It will be
#' appended after the list instruction
#' @param n_unique_items (optional) Number of unique items required in the list
#' @param list_mode (optional) Mode of the list. Either "bullet" or "comma".
#' "bullet mode expects items to be listed with "--" before each item, with a
#'  new line for each item (e.g., "-- item1\\n-- item2\\n-- item3").
#'  "comma" mode expects items to be listed with a number and a period before
#'  (e.g., "1. item1, 2. item2, 3. item3"). "comma" mode may be easier for
#'  smaller LLMs to use
#'
#' @return A [tidyprompt()] with an added [prompt_wrap()] which
#' will ensure that the LLM response is a list of items
#'
#' @export
#'
#' @example inst/examples/answer_as_list.R
#'
#' @seealso [answer_as_named_list()]
#'
#' @family pre_built_prompt_wraps
#' @family answer_as_prompt_wraps
answer_as_list <- function(
  prompt,
  item_name = "item",
  item_explanation = NULL,
  n_unique_items = NULL,
  list_mode = c("bullet", "comma")
) {
  prompt <- tidyprompt(prompt)
  stopifnot(
    is.character(item_name),
    length(item_name) == 1,
    is.null(item_explanation) ||
      (is.character(item_explanation) & length(item_explanation) == 1),
    is.null(n_unique_items) || (is.numeric(n_unique_items) & n_unique_items > 0)
  )
  list_mode <- match.arg(list_mode)

  # Adjust list instruction based on list_mode
  if (list_mode == "bullet") {
    list_instruction <- glue::glue(
      "Respond with a list, like so:\n",
      "  -- <<{item_name} 1>>",
      .trim = FALSE
    )

    if (is.null(n_unique_items) || n_unique_items > 1) {
      list_instruction <- glue::glue(
        "{list_instruction}\n",
        "  -- <<{item_name} 2>>",
        .trim = FALSE
      )
    }

    if (is.null(n_unique_items) || n_unique_items > 2) {
      list_instruction <- glue::glue(
        "{list_instruction}\n",
        "  etc.",
        .trim = FALSE
      )
    }
  } else if (list_mode == "comma") {
    list_instruction <- glue::glue(
      "Respond with a list, like so:\n",
      "  1. <<{item_name}>>",
      .trim = FALSE
    )

    if (is.null(n_unique_items) || n_unique_items > 1) {
      list_instruction <- glue::glue(
        "{list_instruction}, 2. <<{item_name}>>",
        .trim = FALSE
      )
    }

    if (is.null(n_unique_items) || n_unique_items > 2) {
      list_instruction <- glue::glue(
        "{list_instruction}, etc.",
        .trim = FALSE
      )
    }
  }

  if (!is.null(n_unique_items)) {
    item_or_items <- ifelse(n_unique_items == 1, "item", "items")

    list_instruction <- glue::glue(
      "{list_instruction}\n",
      "The list should contain {n_unique_items} unique {item_or_items}.",
      .trim = FALSE
    )
  }

  if (!is.null(item_explanation)) {
    list_instruction <- glue::glue(
      "{list_instruction}\n\n",
      "{item_explanation}",
      .trim = FALSE
    )
  }

  modify_fn <- function(original_prompt_text) {
    glue::glue(
      "{original_prompt_text}\n\n",
      "{list_instruction}",
      .trim = FALSE
    )
  }

  extraction_fn <- function(response) {
    if (!is.character(response)) {
      stop(
        paste0(
          "Response to extract a list from must be a string.",
          " Make sure that you do not apply any prompt wraps which make the",
          " response unsuitable for list extraction."
        )
      )
    }

    if (list_mode == "bullet") {
      # Existing bullet mode extraction
      items <- stringr::str_extract_all(response, "--\\s*([^\\-\\n]+)")[[1]]
      if (length(items) == 0) {
        return(
          llm_feedback(
            glue::glue(
              "Could not parse any listed items from your response.",
              "{list_instruction}",
              .trim = FALSE
            )
          )
        )
      }
      items <- stringr::str_trim(stringr::str_remove_all(items, "--\\s*"))
    } else if (list_mode == "comma") {
      # Improved comma mode extraction
      items <- stringr::str_match_all(response, "\\d+\\.\\s*([^,\\n]+)")[[1]][,
        2
      ]

      if (length(items) == 0) {
        return(
          llm_feedback(
            glue::glue(
              "Could not parse any listed items from your response.",
              "{list_instruction}",
              .trim = FALSE
            )
          )
        )
      }

      # Trim whitespace
      items <- stringr::str_trim(items)

      # Remove any leading numbering (e.g., "1. ", "2. ")
      items <- stringr::str_replace(items, "^\\d+\\.\\s*", "")

      # Remove empty strings
      items <- items[items != ""]

      if (length(items) == 0) {
        return(
          llm_feedback(
            glue::glue(
              "Could not parse any valid items from your response.",
              "{list_instruction}",
              .trim = FALSE
            )
          )
        )
      }
    }

    # Make items unique
    items <- unique(items)

    if (!is.null(n_unique_items) && length(items) != n_unique_items) {
      return(
        llm_feedback(
          glue::glue(
            "The number of unique items in your list should be {n_unique_items}.",
            "{list_instruction}",
            .trim = FALSE
          )
        )
      )
    }

    # Return the extracted items
    return(as.list(items))
  }

  prompt_wrap(
    prompt,
    modify_fn,
    extraction_fn,
    name = "answer_as_list"
  )
}

test_that("answer_as_list adds list instruction", {
  prompt <- "List the top programming languages." |>
    answer_as_list(
      item_name = "language",
      n_unique_items = 3,
      list_mode = "bullet"
    ) |>
    construct_prompt_text()

  expect_true(
    grepl(
      "Respond with a list, like so:\n  -- <<language 1>>\n  -- <<language 2>>\n  etc.",
      prompt,
      fixed = TRUE
    )
  )
  expect_true(
    grepl(
      "The list should contain 3 unique items.",
      prompt,
      fixed = TRUE
    )
  )
})

test_that("answer_as_list returns list of items", {
  skip_test_if_no_ollama()

  ollama <- llm_provider_ollama()

  response <- "List the primary colors." |>
    answer_as_list(item_name = "color", list_mode = "comma") |>
    send_prompt(ollama, verbose = FALSE)

  expect_true(is.list(response))
  expect_equal(length(response), 3)
  expect_true(all(c("red", "blue", "yellow") %in% tolower(unlist(response))))
})

#' Convert a dataframe to a string representation
#'
#' Converts a data frame to a string format, intended for sending it to a LLM
#' (or for display or logging).
#'
#' @param df A `data.frame` object to be converted to a string
#' @param how In what way the df should be converted to a string;
#' either "wide" or "long". "wide" presents column names on the first row,
#' followed by the row values on each new row. "long" presents the values
#' of each row together with the column names, repeating for every row
#' after two lines of whitespace
#'
#' @return A single string representing the df
#'
#' @export
#' @example inst/examples/df_to_string.R
#'
#' @family text_helpers
df_to_string <- function(df, how = c("wide", "long")) {
  how <- match.arg(how)

  # Check if input is a data frame
  if (!is.data.frame(df)) {
    stop("Input must be a data frame")
  }

  # Return empty string if data frame has no rows
  if (nrow(df) == 0) {
    return("")
  }

  # Get column names
  column_names <- colnames(df)

  # Generate the output based on the `wide` parameter
  if (how == "wide") {
    # Concatenate column names as the header
    header <- paste(column_names, collapse = ", ")

    # Convert each row into a single line string without column names
    rows_as_strings <- apply(df, 1, function(row) {
      paste(row, collapse = ", ")
    })

    # Combine header and rows into a single output string
    result_string <- paste(c(header, rows_as_strings), collapse = "\n")
  } else if (how == "long") {
    # Convert each row with column names for each value
    rows_as_strings <- apply(df, 1, function(row) {
      paste(paste(column_names, row, sep = ": "), collapse = "\n")
    })

    # Combine rows into a single output string
    result_string <- paste(rows_as_strings, collapse = "\n\n\n")
    result_string <- paste0(result_string, "\n\n\n")
  }

  return(result_string)
}

test_that("df_to_string correctly formats a simple data frame in wide format", {
  df <- data.frame(Name = c("Alice", "Bob"), Age = c(25, 30))
  expected_output <- "Name, Age\nAlice, 25\nBob, 30"

  expect_equal(df_to_string(df, how = "wide"), expected_output)
})

test_that("df_to_string correctly formats a simple data frame in long format", {
  df <- data.frame(Name = c("Alice", "Bob"), Age = c(25, 30))
  expected_output <- "Name: Alice\nAge: 25\n\n\nName: Bob\nAge: 30\n\n\n"

  expect_equal(df_to_string(df, how = "long"), expected_output)
})

test_that("df_to_string returns empty string for empty data frame", {
  df <- data.frame()
  expect_equal(df_to_string(df, how = "wide"), "")
  expect_equal(df_to_string(df, how = "long"), "")
})

test_that("df_to_string handles single-column data frame in wide and long formats", {
  df <- data.frame(Age = c(25, 30))

  expected_output_wide <- "Age\n25\n30"
  expect_equal(df_to_string(df, how = "wide"), expected_output_wide)

  expected_output_long <- "Age: 25\n\n\nAge: 30\n\n\n"
  expect_equal(df_to_string(df, how = "long"), expected_output_long)
})

test_that("df_to_string throws error if input is not a data frame", {
  expect_error(
    df_to_string("not_a_data_frame", how = "wide"),
    "Input must be a data frame"
  )
  expect_error(
    df_to_string(list(Name = c("Alice", "Bob")), how = "wide"),
    "Input must be a data frame"
  )
})

test_that("df_to_string handles data frame with NA values in wide and long formats", {
  df <- data.frame(Name = c("Alice", NA), Age = c(25, 30))

  expected_output_wide <- "Name, Age\nAlice, 25\nNA, 30"
  expect_equal(df_to_string(df, how = "wide"), expected_output_wide)

  expected_output_long <- "Name: Alice\nAge: 25\n\n\nName: NA\nAge: 30\n\n\n"
  expect_equal(df_to_string(df, how = "long"), expected_output_long)
})

test_that("df_to_string defaults to 'wide' format if 'how' parameter is omitted", {
  df <- data.frame(Name = c("Alice", "Bob"), Age = c(25, 30))
  expected_output <- "Name, Age\nAlice, 25\nBob, 30"

  expect_equal(df_to_string(df), expected_output)
})

# Suppress warning about undefined variables which are passed between environments
#   and are not actually undefined
utils::globalVariables(c("self", "schema_instruction", ".data", "private"))

# Print welcome message
cli::cli_h3(
  paste0(
    "Thank you for using '",
    cli::style_bold("tidyprompt"),
    "'!"
  )
)
cli::cli_alert_info(
  "You may download the latest version from {.href [GitHub](https://github.com/KennispuntTwente/tidyprompt)}, using:"
)
cli::cli_text(
  "{cli::symbol$arrow_right} {.run [remotes::install_github()](remotes::install_github(\"KennispuntTwente/tidyprompt\"))}"
)
cli::cli_alert_info("Bugs, suggestions, questions, or contributions?")
cli::cli_text(
  paste0(
    "{cli::symbol$arrow_right}  Open an",
    " {.href [issue](https://github.com/KennispuntTwente/tidyprompt/issues)}",
    " or {.href [pull request](https://github.com/KennispuntTwente/tidyprompt/pulls)}"
  )
)
cat("\n")

#' Have user check the result of a prompt (human-in-the-loop)
#'
#' This function is used to have a user check the result of a prompt.
#' After evaluation of the prompt and applying prompt wraps,
#' the user is presented with the result and asked to accept or decline.
#' If the user declines, they are asked to provide feedback to the large
#' language model (LLM) so that the LLM can retry the prompt.
#'
#' @param prompt A single string or a [tidyprompt-class] object
#'
#' @return A [tidyprompt-class] with an added [prompt_wrap()] which
#' will add a check for the user to accept or decline the result of the prompt,
#' providing feedback if the result is declined
#'
#' @export
#'
#' @example inst/examples/user_verify.R
user_verify <- function(
  prompt
) {
  prompt <- tidyprompt(prompt)

  validation_fn <- function(response) {
    # Ask user input via console

    repeat {
      cli::cli_h3("Evaluation of tidyprompt resulted in:")
      cat(utils::capture.output(print(response)))
      cat("\n")

      cli::cli_h3("Accept or decline")
      cli::cli_alert_info("If {.strong satisfied}, type nothing")
      cli::cli_alert_info(
        "If {.strong not satisfied}, type feedback to the LLM"
      )

      feedback <- readline("Type: ")

      if (feedback == "") {
        cli::cli_alert_success("Result accepted")
        return(TRUE)
      }

      return(llm_feedback(feedback))
    }
  }

  prompt_wrap(
    prompt,
    validation_fn = validation_fn,
    type = "check"
  )
}

# R/helper_ellmer_json_compatability.R

ellmer_available <- function() {
  requireNamespace("ellmer", quietly = TRUE)
}

# --- Detectors --------------------------------------------------------------

is_json_schema_list <- function(x) {
  is.list(x) &&
    (any(c("$schema", "type", "properties", "items", "enum") %in% names(x)) ||
      # Top-level "name/schema/strict" wrapper we sometimes build:
      identical(sort(names(x)), sort(c("name", "schema", "strict"))) ||
      # Loose heuristic for object schemas:
      (!is.null(x$type) && is.character(x$type)))
}

# Lazy cache for class signatures of ellmer prototypes (robust detection).
# Computed on first access so that ellmer (a Suggests dependency) can be
# loaded after tidyprompt without the cache staying permanently NULL.
.ellmer_class_sig_env <- new.env(parent = emptyenv())
.ellmer_class_sig_env$cache <- NULL
.ellmer_class_sig_env$resolved <- FALSE

.get_ellmer_class_signatures <- function() {
  if (.ellmer_class_sig_env$resolved) {
    return(.ellmer_class_sig_env$cache)
  }
  if (!ellmer_available()) {
    return(NULL)
  }
  sig <- tryCatch(
    {
      s <- list(
        string = class(ellmer::type_string()),
        number = class(ellmer::type_number()),
        integer = class(ellmer::type_integer()),
        boolean = class(ellmer::type_boolean()),
        enum = class(ellmer::type_enum(c("a", "b"))),
        array = class(ellmer::type_array(ellmer::type_string())),
        object = class(ellmer::type_object(
          .description = "probe",
          .additional_properties = TRUE
        ))
      )
      if (exists("type_ignore", envir = asNamespace("ellmer"))) {
        s$ignore <- tryCatch(
          class(get("type_ignore", envir = asNamespace("ellmer"))()),
          error = function(e) NULL
        )
      }
      if (exists("type_from_schema", envir = asNamespace("ellmer"))) {
        s$json_schema <- class(ellmer::type_from_schema(
          '{"type": "string"}'
        ))
      }
      s
    },
    error = function(e) NULL
  )
  .ellmer_class_sig_env$resolved <- TRUE
  .ellmer_class_sig_env$cache <- sig
  sig
}

has_all_classes <- function(x, cls) {
  if (is.null(cls)) {
    return(FALSE)
  }
  all(cls %in% class(x))
}

is_ellmer_type <- function(x) {
  sig <- .get_ellmer_class_signatures()
  if (is.null(sig)) {
    return(FALSE)
  }
  any(vapply(sig, has_all_classes, logical(1), x = x))
}

is_ellmer_chat <- function(x) {
  is.environment(x) &&
    is.function(x$chat)
}

ellmer_chat_clone_reset <- function(
  chat,
  context = "This ellmer chat object"
) {
  if (!is_ellmer_chat(chat)) {
    stop(
      paste0(
        context,
        " doesn't look like an ellmer chat object (no `$chat()` method)."
      )
    )
  }
  if (!is.function(chat$clone)) {
    stop(paste0(context, " must support `$clone()`."))
  }
  if (!is.function(chat$set_turns)) {
    stop(paste0(context, " must support `$set_turns()`."))
  }

  chat <- chat$clone()
  if (!is_ellmer_chat(chat)) {
    stop(paste0(context, " `$clone()` did not return a valid ellmer chat."))
  }

  chat$set_turns(list())
}

as_send_prompt_llm_provider <- function(
  llm_provider,
  verbose = NULL,
  stream = NULL
) {
  if (inherits(llm_provider, "LlmProvider") || !is_ellmer_chat(llm_provider)) {
    return(llm_provider)
  }

  stream_val <- stream %||% getOption("tidyprompt.stream", TRUE)
  provider <- llm_provider_ellmer(
    llm_provider,
    parameters = list(stream = stream_val),
    verbose = verbose %||% getOption("tidyprompt.verbose", TRUE)
  )

  provider$parameters$.reset_ellmer_chat <- TRUE

  provider
}

ellmer_type_ignore_compat <- function(description = NULL, required = FALSE) {
  stopifnot(ellmer_available())

  ellmer_ns <- asNamespace("ellmer")
  if (exists("TypeIgnore", envir = ellmer_ns, inherits = FALSE)) {
    ctor <- get("TypeIgnore", envir = ellmer_ns, inherits = FALSE)
    return(ctor(description = description, required = required))
  }

  if (!exists("type_ignore", envir = ellmer_ns, inherits = FALSE)) {
    stop("Ignored ellmer arguments require ellmer >= 0.4.0.")
  }

  obj <- get("type_ignore", envir = ellmer_ns, inherits = FALSE)()
  if (!is.null(description)) {
    attr(obj, "description") <- description
  }
  attr(obj, "required") <- required
  obj
}

# --- JSON Schema -> ellmer::type_* -----------------------------------------

json_schema_to_ellmer_type <- function(
  schema,
  required = TRUE,
  strict = FALSE
) {
  if (!ellmer_available()) {
    stop("ellmer is not installed; cannot convert JSON Schema to ellmer types.")
  }

  # Unwrap "json_schema" wrapper shape {name, schema, strict}
  if (!is.null(schema$schema) && is.list(schema$schema)) {
    schema <- schema$schema
  }

  if (isTRUE(schema$`x-tidyprompt-ignore`)) {
    if (
      exists("type_ignore", envir = asNamespace("ellmer"), inherits = FALSE)
    ) {
      return(ellmer_type_ignore_compat(
        description = schema$description %||% NULL,
        required = FALSE
      ))
    }

    return(ellmer::type_string(
      description = schema$description %||% NULL,
      required = FALSE
    ))
  }

  # Handle enum
  if (!is.null(schema$enum)) {
    return(ellmer::type_enum(
      schema$enum,
      description = schema$description %||% NULL,
      required = required
    ))
  }

  t <- schema$type %||% NULL

  if (identical(t, "string")) {
    return(ellmer::type_string(
      description = schema$description %||% NULL,
      required = required
    ))
  }
  if (identical(t, "number")) {
    return(ellmer::type_number(
      description = schema$description %||% NULL,
      required = required
    ))
  }
  if (identical(t, "integer")) {
    return(ellmer::type_integer(
      description = schema$description %||% NULL,
      required = required
    ))
  }
  if (identical(t, "boolean")) {
    return(ellmer::type_boolean(
      description = schema$description %||% NULL,
      required = required
    ))
  }
  if (identical(t, "array")) {
    item_schema <- schema$items %||% list(type = "string")
    return(ellmer::type_array(
      items = json_schema_to_ellmer_type(
        item_schema,
        required = TRUE,
        strict = strict
      ),
      description = schema$description %||% NULL,
      required = required
    ))
  }
  if (identical(t, "object") || !is.null(schema$properties)) {
    props <- schema$properties %||% list()
    req <- schema$required %||% character()

    ellmer_fields <- lapply(names(props), function(p) {
      json_schema_to_ellmer_type(
        props[[p]],
        required = isTRUE(p %in% req),
        strict = strict
      )
    })
    names(ellmer_fields) <- names(props)

    addl <- schema$additionalProperties
    addl_flag <- if (is.null(addl)) !isTRUE(strict) else isTRUE(addl)

    args <- c(
      ellmer_fields,
      list(
        .description = schema$description %||% NULL,
        .additional_properties = addl_flag,
        .required = required
      )
    )

    # Drop NULLs so nothing bogus leaks into `...`
    args <- args[!vapply(args, is.null, logical(1))]

    return(do.call(ellmer::type_object, args))
  }

  # Fallback: use type_from_schema() if available, else permissive object
  if (
    ellmer_available() &&
      exists("type_from_schema", envir = asNamespace("ellmer"))
  ) {
    json_str <- jsonlite::toJSON(schema, auto_unbox = TRUE)
    return(ellmer::type_from_schema(json_str))
  }
  ellmer::type_object(
    .description = schema$description %||% NULL,
    .additional_properties = TRUE,
    .required = required
  )
}

# --- ellmer::type_* -> JSON Schema (best-effort) ----------------------------

ellmer_type_to_json_schema <- function(x, strict = FALSE, description = NULL) {
  sig <- .get_ellmer_class_signatures()
  if (is.null(sig)) {
    stop("ellmer is not installed; cannot convert ellmer types to JSON Schema.")
  }
  desc <- description %||% attr(x, "description", exact = TRUE)

  # --- type_from_schema: already carries a JSON schema, extract it --------
  if (!is.null(sig$json_schema) && has_all_classes(x, sig$json_schema)) {
    schema <- attr(x, "json", exact = TRUE) %||%
      attr(x, "schema", exact = TRUE)
    if (is.list(schema)) {
      return(schema)
    }
  }

  if (!is.null(sig$ignore) && has_all_classes(x, sig$ignore)) {
    return(compact_list(list(
      `x-tidyprompt-ignore` = TRUE,
      description = desc
    )))
  }

  # --- Basic scalars: prefer S7 'type' property over class signatures ----
  basic_type <- attr(x, "type", exact = TRUE)
  if (
    is.character(basic_type) &&
      length(basic_type) == 1 &&
      basic_type %in% c("string", "number", "integer", "boolean")
  ) {
    return(compact_list(list(type = basic_type, description = desc)))
  }

  # --- Enum: prefer attribute-based detection for robustness --------------
  enum_vals <- attr(x, "values", exact = TRUE) %||%
    attr(x, "levels", exact = TRUE)
  if (!is.null(enum_vals) || has_all_classes(x, sig$enum)) {
    if (is.null(enum_vals)) {
      enum_vals <- character()
    }
    return(compact_list(list(enum = enum_vals, description = desc)))
  }

  # --- Array ---------------------------------------------------------------
  if (has_all_classes(x, sig$array)) {
    items <- attr(x, "items", exact = TRUE) %||% attr(x, "item", exact = TRUE)
    return(compact_list(list(
      type = "array",
      description = desc, # <-- move before items
      items = if (!is.null(items)) {
        ellmer_type_to_json_schema(items, strict = strict)
      } else {
        NULL
      }
    )))
  }

  # --- Object --------------------------------------------------------------
  if (has_all_classes(x, sig$object)) {
    addl_attr <- attr(x, ".additional_properties", exact = TRUE)
    if (is.null(addl_attr)) {
      addl_attr <- attr(x, "additional_properties", exact = TRUE)
    }
    addl <- if (is.null(addl_attr)) !isTRUE(strict) else isTRUE(addl_attr)

    props <- attr(x, "properties", exact = TRUE)
    if (is.null(props)) {
      at <- attributes(x) %||% list()
      reserved <- c(
        ".additional_properties",
        "additional_properties",
        "class",
        "description",
        "required",
        "properties"
      )
      prop_names <- setdiff(names(at), reserved)
      props <- lapply(prop_names, function(nm) at[[nm]])
      names(props) <- prop_names
    }

    properties <- list()
    required <- character()
    if (length(props)) {
      for (nm in names(props)) {
        val <- props[[nm]]
        if (is_ellmer_type(val)) {
          properties[[nm]] <- ellmer_type_to_json_schema(val, strict = strict)
          req_flag <- attr(val, "required", exact = TRUE)
          if (!identical(req_flag, FALSE)) required <- c(required, nm)
        }
      }
    }

    return(compact_list(list(
      type = "object",
      description = desc,
      properties = properties,
      required = if (length(required)) unique(required) else NULL,
      additionalProperties = isTRUE(addl)
    )))
  }

  # --- Fallback permissive -------------------------------------------------
  compact_list(list(
    type = "object",
    additionalProperties = TRUE,
    description = desc
  ))
}


compact_list <- function(x) {
  x[!vapply(x, is.null, logical(1))]
}

# --- Public: normalize a schema for a given target --------------------------

# Returns a list with both representations when possible:
# $json_schema and $ellmer_type, so callers can pick what they need.
normalize_schema_dual <- function(schema, strict = FALSE) {
  if (is.null(schema)) {
    return(list(json_schema = NULL, ellmer_type = NULL))
  }

  if (is_ellmer_type(schema)) {
    # ellmer type supplied; attempt reverse conversion for OpenAI/Ollama
    json_s <- tryCatch(
      ellmer_type_to_json_schema(schema, strict = strict),
      error = function(e) NULL
    )
    return(list(json_schema = json_s, ellmer_type = schema))
  }

  if (is_json_schema_list(schema)) {
    # JSON Schema supplied; convert forward for ellmer
    ellmer_t <- tryCatch(
      json_schema_to_ellmer_type(schema, required = TRUE, strict = strict),
      error = function(e) NULL
    )
    return(list(json_schema = schema, ellmer_type = ellmer_t))
  }

  stop(
    "`schema` must be either a JSON Schema list or an ellmer type_*() object."
  )
}

test_that("OpenAI Responses API maps image_url to scalar string", {
  # Construct a synthetic messages list replicating internal structure prior to transformation
  messages <- list(
    list(
      role = "user",
      content = list(
        list(type = "text", text = "Describe this image"),
        list(
          type = "image_url",
          image_url = list(url = "https://example.org/cat.jpg", detail = "auto")
        )
      )
    )
  )

  # Reproduce transformation logic (factored implicitly in provider code). We simulate only the part needed.
  input <- lapply(messages, function(msg) {
    role <- msg$role
    content <- msg$content
    parts <- list()
    for (part in content) {
      if (identical(part$type, "text")) {
        parts[[length(parts) + 1]] <- list(
          type = "input_text",
          text = part$text
        )
      } else if (identical(part$type, "image_url")) {
        img <- part$image_url
        url_val <- img$url
        detail_val <- img$detail
        parts[[length(parts) + 1]] <- compact_list(list(
          type = "input_image",
          image_url = url_val,
          detail = detail_val
        ))
      }
    }
    list(role = role, content = parts)
  })

  expect_true(is.character(input[[1]]$content[[2]]$image_url))
  expect_equal(input[[1]]$content[[2]]$image_url, "https://example.org/cat.jpg")
  expect_equal(input[[1]]$content[[2]]$detail, "auto")
})

# tests/testthat/test-ellmer_json_compatability.R

# This tests the helper functions in R/helper_ellmer_json_compatibility.R
#   and ensures that the conversion between JSON schema and ellmer types works correctly

# library(testthat)
# devtools::load_all()

# ---- Helpers ---------------------------------------------------------------

# Recursively remove NULLs
strip_nulls <- function(x) {
  if (is.list(x)) {
    x <- lapply(x, strip_nulls)
    x <- x[!vapply(x, is.null, logical(1))]
  }
  x
}

# Canonicalize schema for stable comparisons:
# - sort `properties` by name at every depth
# - sort `required` character vectors
canonicalize_schema <- function(x) {
  if (!is.list(x)) {
    return(x)
  }

  x <- strip_nulls(x)

  # sort nested object properties
  if (!is.null(x$properties) && is.list(x$properties)) {
    nms <- names(x$properties)
    if (!is.null(nms)) {
      ord <- order(nms)
      x$properties <- x$properties[ord]
      x$properties <- lapply(x$properties, canonicalize_schema)
    }
  }

  # arrays: canonicalize the `items` schema
  if (!is.null(x$items)) {
    x$items <- canonicalize_schema(x$items)
  }

  # sort required names
  if (!is.null(x$required) && is.atomic(x$required)) {
    x$required <- sort(unique(x$required))
  }

  # Recurse over any other nested lists
  for (nm in names(x)) {
    if (nm %in% c("properties", "items")) {
      next
    }
    if (is.list(x[[nm]])) x[[nm]] <- canonicalize_schema(x[[nm]])
  }
  x
}

expect_schema_equal <- function(a, b) {
  testthat::expect_equal(canonicalize_schema(a), canonicalize_schema(b))
}

# ---- Cache resilience -------------------------------------------------------

testthat::test_that("class sig cache retries after initial unavailability", {
  .ellmer_class_sig_env$cache <- NULL
  .ellmer_class_sig_env$resolved <- FALSE
  on.exit({
    .ellmer_class_sig_env$cache <- NULL
    .ellmer_class_sig_env$resolved <- FALSE
  })

  call_n <- 0L
  local_mocked_bindings(
    ellmer_available = function() {
      call_n <<- call_n + 1L
      call_n > 1L
    }
  )

  result1 <- .get_ellmer_class_signatures()
  testthat::expect_null(result1)
  testthat::expect_false(.ellmer_class_sig_env$resolved)

  result2 <- .get_ellmer_class_signatures()
  testthat::expect_true(.ellmer_class_sig_env$resolved)
  testthat::expect_true(is.list(result2) && length(result2) > 0)
})

# ---- Basic detectors -------------------------------------------------------

testthat::test_that("is_json_schema_list detects typical shapes", {
  plain <- list(type = "string")
  wrapped <- list(
    name = "Thing",
    strict = TRUE,
    schema = list(type = "object", properties = list(x = list(type = "number")))
  )
  not_schema <- list(a = 1, b = 2)

  testthat::expect_true(is_json_schema_list(plain))
  testthat::expect_true(is_json_schema_list(wrapped))
  testthat::expect_false(is_json_schema_list(not_schema))
})

test_that("object respects additionalProperties flag", {
  testthat::skip_if_not_installed("ellmer")

  s <- list(
    type = "object",
    properties = list(x = list(type = "string")),
    required = "x"
  )
  ty_strict <- json_schema_to_ellmer_type(s, strict = TRUE)
  ty_loose <- json_schema_to_ellmer_type(s, strict = FALSE)

  back_strict <- ellmer_type_to_json_schema(ty_strict, strict = TRUE)
  back_loose <- ellmer_type_to_json_schema(ty_loose, strict = TRUE)

  expect_false(isTRUE(back_strict$additionalProperties))
  expect_true(isTRUE(back_loose$additionalProperties))
})

testthat::test_that("nested object without description converts cleanly", {
  testthat::skip_if_not_installed("ellmer")

  schema <- list(
    type = "object",
    properties = list(
      outer = list(
        type = "object",
        properties = list(
          inner = list(type = "boolean")
        ),
        required = "inner"
        # no description here on purpose
      )
    ),
    required = "outer"
  )

  ty <- json_schema_to_ellmer_type(schema, strict = TRUE)
  testthat::expect_true(is_ellmer_type(ty))

  back <- ellmer_type_to_json_schema(ty, strict = TRUE)

  # With strict=TRUE, both levels should have additionalProperties = FALSE
  testthat::expect_false(isTRUE(back$additionalProperties))
  testthat::expect_false(isTRUE(back$properties$outer$additionalProperties))
  testthat::expect_equal(back$properties$outer$required, "inner")
})


# ---- Round-trip: JSON -> ellmer -> JSON -----------------------------------

testthat::test_that("round-trip preserves structure incl. required and strict", {
  testthat::skip_if_not_installed("ellmer")

  schema <- list(
    type = "object",
    description = "root",
    properties = list(
      name = list(type = "string", description = "Name"),
      age = list(type = "integer"), # optional
      tags = list(type = "array", items = list(type = "string")),
      role = list(enum = c("admin", "user")),
      prefs = list(
        type = "object",
        properties = list(
          notify = list(type = "boolean")
        ),
        required = c("notify")
      )
    ),
    required = c("name")
    # additionalProperties omitted on purpose
  )

  # strict=TRUE should force additionalProperties=FALSE on the object(s)
  ty <- json_schema_to_ellmer_type(schema, strict = TRUE)
  testthat::expect_true(is_ellmer_type(ty))

  back <- ellmer_type_to_json_schema(ty, strict = TRUE)

  expected <- schema
  expected$additionalProperties <- FALSE
  # nested object prefs should also have explicit additionalProperties per our conversion
  expected$properties$prefs$additionalProperties <- FALSE

  expect_schema_equal(back, expected)
})

testthat::test_that("array with missing items defaults to string items", {
  testthat::skip_if_not_installed("ellmer")

  schema <- list(
    type = "object",
    properties = list(
      things = list(type = "array") # no `items` provided
    )
  )

  ty <- json_schema_to_ellmer_type(schema, strict = FALSE)
  back <- ellmer_type_to_json_schema(ty, strict = FALSE)

  # Expect items defaulted to string
  testthat::expect_equal(back$properties$things$type, "array")
  testthat::expect_equal(back$properties$things$items$type, "string")
  # With strict=FALSE, additionalProperties should default TRUE
  testthat::expect_true(isTRUE(back$additionalProperties))
})

testthat::test_that("ignore marker converts across ellmer versions", {
  testthat::skip_if_not_installed("ellmer")

  schema <- list(
    `x-tidyprompt-ignore` = TRUE,
    description = "Hidden argument"
  )

  ty <- json_schema_to_ellmer_type(schema, strict = TRUE)
  back <- ellmer_type_to_json_schema(ty, strict = TRUE)

  if (exists("type_ignore", envir = asNamespace("ellmer"), inherits = FALSE)) {
    testthat::expect_true(isTRUE(back$`x-tidyprompt-ignore`))
  } else {
    testthat::expect_equal(back$type, "string")
  }
  testthat::expect_equal(back$description, "Hidden argument")
})

testthat::test_that("enum values survive a JSON->ellmer->JSON roundtrip", {
  testthat::skip_if_not_installed("ellmer")

  schema <- list(
    type = "object",
    properties = list(
      color = list(enum = c("red", "green", "blue"))
    ),
    required = c("color")
  )

  ty <- json_schema_to_ellmer_type(schema, strict = TRUE)
  back <- ellmer_type_to_json_schema(ty, strict = TRUE)

  testthat::expect_equal(back$properties$color$enum, c("red", "green", "blue"))
  testthat::expect_equal(sort(back$required), "color")
  testthat::expect_false(isTRUE(back$additionalProperties))
})

# ---- Wrapper unwrapping ----------------------------------------------------

testthat::test_that("json_schema_to_ellmer_type unwraps {name, schema, strict}", {
  testthat::skip_if_not_installed("ellmer")

  inner <- list(
    type = "object",
    properties = list(x = list(type = "string")),
    required = "x"
  )
  wrapper <- list(name = "X", strict = TRUE, schema = inner)

  ty <- json_schema_to_ellmer_type(wrapper, strict = TRUE)
  testthat::expect_true(is_ellmer_type(ty))

  back <- ellmer_type_to_json_schema(ty, strict = TRUE)

  expected <- inner
  expected$additionalProperties <- FALSE
  expect_schema_equal(back, expected)
})

# ---- normalize_schema_dual -------------------------------------------------

testthat::test_that("normalize_schema_dual returns both sides from JSON input", {
  testthat::skip_if_not_installed("ellmer")

  schema <- list(type = "string", description = "a string")
  out <- normalize_schema_dual(schema, strict = FALSE)

  testthat::expect_type(out, "list")
  testthat::expect_true(is.list(out$json_schema))
  testthat::expect_true(is_ellmer_type(out$ellmer_type))
})

testthat::test_that("normalize_schema_dual returns both sides from ellmer input", {
  testthat::skip_if_not_installed("ellmer")

  ty <- ellmer::type_object(
    id = ellmer::type_integer(required = TRUE),
    note = ellmer::type_string(required = FALSE)
  )
  out <- normalize_schema_dual(ty, strict = TRUE)

  testthat::expect_true(is.list(out$json_schema))
  testthat::expect_true(is_ellmer_type(out$ellmer_type))

  # Required should include only id; strict=TRUE -> no additional properties
  testthat::expect_equal(out$json_schema$required, "id")
  testthat::expect_false(isTRUE(out$json_schema$additionalProperties))
})

# ---- Ellmer attributes: required flags ------------------------------------

testthat::test_that("optional vs required fields preserved across roundtrip", {
  testthat::skip_if_not_installed("ellmer")

  schema <- list(
    type = "object",
    properties = list(
      a = list(type = "string"), # optional
      b = list(type = "number") # optional
    ),
    required = "b"
  )

  ty <- json_schema_to_ellmer_type(schema, strict = FALSE)

  # Inspect required attributes on child ellmer nodes
  atts <- attributes(ty)
  atts_a <- atts$properties$a |> attributes()
  a_req <- atts_a$required
  atts_b <- atts$properties$b |> attributes()
  b_req <- atts_b$required

  testthat::expect_identical(a_req, FALSE)
  testthat::expect_true(isTRUE(b_req))

  # Back to JSON: only "b" should be required
  back <- ellmer_type_to_json_schema(ty, strict = FALSE)
  testthat::expect_equal(sort(back$required), "b")
  testthat::expect_true(isTRUE(back$additionalProperties))
})

testthat::test_that("is_ellmer_type detects type_from_schema objects", {
  testthat::skip_if_not_installed("ellmer")
  ellmer_ns <- asNamespace("ellmer")
  testthat::skip_if_not(
    exists("type_from_schema", envir = ellmer_ns),
    "type_from_schema not available in this ellmer version"
  )

  schema <- list(type = "object", properties = list(x = list(type = "string")))
  tfs <- ellmer::type_from_schema(jsonlite::toJSON(schema, auto_unbox = TRUE))
  testthat::expect_true(is_ellmer_type(tfs))
})

testthat::test_that("ellmer_type_to_json_schema handles type_from_schema", {
  testthat::skip_if_not_installed("ellmer")
  ellmer_ns <- asNamespace("ellmer")
  testthat::skip_if_not(
    exists("type_from_schema", envir = ellmer_ns),
    "type_from_schema not available in this ellmer version"
  )

  schema <- list(
    type = "object",
    properties = list(
      name = list(type = "string"),
      age = list(type = "integer")
    ),
    required = c("name")
  )
  tfs <- ellmer::type_from_schema(jsonlite::toJSON(schema, auto_unbox = TRUE))
  result <- ellmer_type_to_json_schema(tfs)

  testthat::expect_equal(result$type, "object")
  testthat::expect_true("name" %in% names(result$properties))
})

testthat::test_that("normalize_schema_dual works with type_from_schema", {
  testthat::skip_if_not_installed("ellmer")
  ellmer_ns <- asNamespace("ellmer")
  testthat::skip_if_not(
    exists("type_from_schema", envir = ellmer_ns),
    "type_from_schema not available in this ellmer version"
  )

  schema <- list(type = "string")
  tfs <- ellmer::type_from_schema(jsonlite::toJSON(schema, auto_unbox = TRUE))
  dual <- normalize_schema_dual(tfs)

  testthat::expect_true(!is.null(dual$json_schema))
  testthat::expect_true(!is.null(dual$ellmer_type))
})

# TODO: take $chat_history from tidyprompt object & add when sending to provider

#' Send a prompt to a LLM provider
#'
#' This function is responsible for sending prompts to a LLM provider for evaluation.
#' The function will interact with the LLM provider until a successful response
#' is received or the maximum number of interactions is reached. The function will
#' apply extraction and validation functions to the LLM response, as specified
#' in the prompt wraps (see [prompt_wrap()]). If the maximum number of interactions
#'
#' @param prompt A string or a [tidyprompt-class] object
#'
#' @param llm_provider [llm_provider-class] object
#'  (default is [llm_provider_ollama()]).
#' This object and its settings will be used to evaluate the prompt.
#' You may also pass an `ellmer::chat()` object directly; in that case,
#' [send_prompt()] will clone it, clear any existing turns, and wrap the clean
#' clone with [llm_provider_ellmer()] before evaluation.
#' When using [llm_provider_ellmer()] or a raw `ellmer::chat()`, the working
#' `ellmer` turns are rebuilt from tidyprompt's own chat history for each call.
#' Note that the 'verbose' and 'stream' settings in the LLM provider will be
#'  overruled by the 'verbose' and 'stream' arguments in this function
#'  when those are not NULL.
#' Furthermore, advanced [tidyprompt-class] objects may carry '$parameter_fn'
#'  functions which can set parameters in the llm_provider object
#'  (see [prompt_wrap()] and [llm_provider-class] for more ).
#'
#' @param max_interactions Maximum number of interactions allowed with the
#' LLM provider. Default is 10. If the maximum number of interactions is reached
#' without a successful response, 'NULL' is returned as the response (see return
#' value). The first interaction is the initial chat completion
#' @param clean_chat_history If the chat history should be cleaned after each
#' interaction. Cleaning the chat history means that only the
#' first and last message from the user, the last message from the assistant,
#' all messages from the system, and all tool results are kept in a 'clean'
#' chat history. This clean chat history is used when requesting a new chat completion.
#' Rows marked as non-replayable are excluded from new requests regardless of
#' this setting, so the returned transcript may contain more rows than the model
#' actually sees on a retry or follow-up call.
#' (i.e., if a LLM repeatedly fails to provide a correct response, only its last failed response
#' will included in the context window). This may increase the LLM performance
#' on the next interaction
#' @param verbose If the interaction with the LLM provider should be printed
#' to the console. This will overrule the 'verbose' setting in the LLM provider
#' @param stream If the interaction with the LLM provider should be streamed.
#' This setting will only be used if the LLM provider already has a
#' 'stream' parameter (which indicates there is support for streaming). Note
#' that when 'verbose' is set to FALSE, the 'stream' setting will be ignored
#' @param return_mode One of 'full' or 'only_response'. See return value
#' @return \itemize{
#'  \item If return mode 'only_response', the function will return only the LLM response
#' after extraction and validation functions have been applied (NULL is returned
#' when unsuccessful after the maximum number of interactions).
#'  \item If return mode 'full', the function will return a list with the following elements:
#'  \itemize{
#'    \item 'response' (the LLM response after extraction and validation functions have been applied;
#'  NULL is returned when unsuccessful after the maximum number of interactions),
#'    \item 'interactions' (the number of interactions with the LLM provider),
#'    \item 'chat_history' (a dataframe with the full chat history which led to the final response.
#'    This may include rows that are retained for inspection but not re-sent to the model;
#'    such rows are marked with column `hidden_from_llm = TRUE`),
#'    \item 'chat_history_clean' (a dataframe with the cleaned chat history which led to
#' the final response; here, only the first and last message from the user, the
#' last message from the assistant, and all messages from the system are kept,
#' after excluding any non-replayable rows),
#'    \item 'start_time' (the time when the function was called),
#'    \item 'end_time' (the time when the function ended),
#'    \item 'duration_seconds' (the duration of the function in seconds),
#'    \item 'http' (a list with all HTTP requests and responses made during the interactions;
#'    as returned by `llm_provider$complete_chat()`),
#'    \item 'ellmer_chat' (if [llm_provider_ellmer()] or a raw `ellmer::chat()`
#'    object was used, this will be
#'    the updated 'ellmer' chat object, containing for instance
#'    the turns and possible tool calls. (As this function
#'    uses a clone of the provided LLM provider, the 'ellmer' chat object in the
#'    LLM provider will not be updated; use this return value when you need the
#'    updated native `ellmer` turns from the current evaluation. Note that turns in the 'ellmer'
#'    chat object may not contain the full chat history when `clean_chat_history = TRUE`
#'    was used.)
#'  }
#' }
#'
#'
#' @export
#' @example inst/examples/send_prompt.R
#'
#' @seealso [tidyprompt-class], [prompt_wrap()], [llm_provider-class], [llm_provider_ollama()],
#' [llm_provider_openai()]
#'
#' @family prompt_evaluation
send_prompt <- function(
  prompt,
  llm_provider = llm_provider_ollama(),
  max_interactions = 10,
  clean_chat_history = FALSE,
  verbose = NULL,
  stream = NULL,
  return_mode = c("only_response", "full")
) {
  ## 1 Validate arguments

  # Basic validation
  prompt <- tidyprompt(prompt)
  return_mode <- match.arg(return_mode)
  llm_provider <- as_send_prompt_llm_provider(llm_provider, verbose, stream)
  stopifnot(
    inherits(llm_provider, "LlmProvider"),
    max_interactions > 0,
    max_interactions == floor(max_interactions),
    is.logical(clean_chat_history),
    is.null(verbose) | is.logical(verbose),
    is.null(stream) | is.logical(stream)
  )

  # Add provider-specific pre/post prompt wraps
  if (isTRUE(is.function(llm_provider$apply_prompt_wraps))) {
    prompt <- llm_provider$apply_prompt_wraps(prompt)
  }

  # Verify and configure llm_provider
  llm_provider <- llm_provider$clone()
  # Deep-clone the ellmer chat object so mutations (register_tool, set_turns)
  # don't affect the caller's original provider. When send_prompt() received a
  # raw ellmer chat directly, reset the working clone before evaluation.
  if (!is.null(llm_provider[["ellmer_chat"]])) {
    if (isTRUE(llm_provider$parameters$.reset_ellmer_chat)) {
      llm_provider$ellmer_chat <- ellmer_chat_clone_reset(
        llm_provider$ellmer_chat,
        context = paste0(
          "When passing an ellmer chat directly to `send_prompt()`, `llm_provider`"
        )
      )
      llm_provider$parameters$.reset_ellmer_chat <- NULL
    } else if (is.function(llm_provider$ellmer_chat$clone)) {
      llm_provider$ellmer_chat <- llm_provider$ellmer_chat$clone()
    }
  }
  if (!is.null(verbose)) {
    llm_provider$verbose <- verbose
  }
  if (
    !is.null(stream) && !is.null(llm_provider$parameters$stream) # This means the provider supports streaming
  ) {
    llm_provider$parameters$stream <- stream
  }
  # Apply parameter_fn's to the llm_provider
  for (prompt_wrap in get_prompt_wraps(prompt)) {
    if (!is.null(prompt_wrap$parameter_fn)) {
      parameter_fn <- prompt_wrap$parameter_fn
      llm_provider$set_parameters(parameter_fn(llm_provider))
    }
  }
  # Add handler_fn's to the llm_provider
  for (prompt_wrap in get_prompt_wraps(prompt)) {
    if (!is.null(prompt_wrap$handler_fn)) {
      llm_provider$add_handler_fn(prompt_wrap$handler_fn)
    }
  }

  # Initialize variables which keep track of the process
  if (return_mode == "full") {
    start_time <- Sys.time()
  }
  http <- list(requests = list(), responses = list())
  # Object which keeps ellmer chat object, turns, structured output;
  #   when using an ellmer LLM provider:
  ellmer_chat <- NULL

  ## 2 Chat history, send_chat, handler_fns

  chat_history <- prompt$get_chat_history(llm_provider)

  # Internal function to send chat messages
  send_chat <- function(
    message,
    role = "user",
    tool_result = FALSE
  ) {
    merge_completed_chat_history <- function(
      original_chat_history,
      completed_chat_history,
      request_rows
    ) {
      updated_chat_history <- dplyr::bind_rows(
        original_chat_history,
        completed_chat_history[0, , drop = FALSE]
      )
      updated_chat_history <- updated_chat_history[
        seq_len(nrow(original_chat_history)),
        ,
        drop = FALSE
      ]

      request_rows <- as.integer(request_rows %||% integer())
      request_rows <- request_rows[
        !is.na(request_rows) &
          request_rows >= 1L &
          request_rows <= nrow(updated_chat_history)
      ]

      sent_rows <- min(length(request_rows), nrow(completed_chat_history))
      if (sent_rows > 0L) {
        updated_chat_history[
          request_rows[seq_len(sent_rows)],
          names(completed_chat_history)
        ] <- completed_chat_history[seq_len(sent_rows), , drop = FALSE]
      }

      if (nrow(completed_chat_history) > sent_rows) {
        updated_chat_history <- dplyr::bind_rows(
          updated_chat_history,
          completed_chat_history[
            (sent_rows + 1L):nrow(completed_chat_history),
            ,
            drop = FALSE
          ]
        )
      }

      updated_chat_history
    }

    if (!is.null(message)) {
      message <- as.character(message)
      chat_history <<- chat_history |>
        add_msg_to_chat_history(message, role, tool_result)
    }

    if (clean_chat_history) {
      cleaned_chat_history <- clean_chat_history(chat_history)
      response <- llm_provider$complete_chat(
        list(chat_history = cleaned_chat_history)
      )
      chat_history <<- merge_completed_chat_history(
        chat_history,
        response$completed,
        attr(cleaned_chat_history, "source_rows") %||%
          as.integer(rownames(cleaned_chat_history))
      )
    } else {
      response <- llm_provider$complete_chat(list(chat_history = chat_history))
      chat_history <<- merge_completed_chat_history(
        chat_history,
        response$completed,
        seq_len(nrow(chat_history))
      )
    }

    for (http_response in response$http$response) {
      http$responses[[length(http$responses) + 1]] <<- http_response
    }
    for (http_request in response$http$request) {
      http$requests[[length(http$requests) + 1]] <<- http_request
    }

    if (!is.null(response$ellmer_chat)) {
      ellmer_chat <<- response$ellmer_chat
    }

    utils::tail(chat_history$content, 1)
  }

  ## 3 Retrieve initial response

  response <- send_chat(NULL)
  # (NULL as initial message is already included in the chat_history)

  ## 4 Apply extractions and validations

  prompt_wraps <- get_prompt_wraps(prompt, order = "evaluation")
  # (Tools, then modes, then unspecified prompt_wraps)

  interactions <- 1
  success <- FALSE
  while (interactions < max_interactions && !success) {
    interactions <- interactions + 1

    if (length(prompt_wraps) == 0) {
      success <- TRUE
    }

    # Initialize variables for the loop
    any_prompt_wrap_not_done <- FALSE
    llm_break <- FALSE

    for (pw_index in seq_along(prompt_wraps)) {
      prompt_wrap <- prompt_wraps[[pw_index]]

      # Apply extraction function
      if (!is.null(prompt_wrap$extraction_fn)) {
        extraction_function <- prompt_wrap$extraction_fn

        # Set the environment of the extraction function, if an
        #   environment was given as attribute. At the moment, this is used
        #   for passing 'tool_functions' from add_tools() to execution here
        if (!is.null(attr(extraction_function, "environment"))) {
          environment <- attr(extraction_function, "environment")
          environment(extraction_function) <- environment
        }
        extraction_result <-
          extraction_function(response, llm_provider, http)

        # If it inherits llm_feedback,
        #   send the feedback to the LLM & get new response
        if (
          inherits(extraction_result, "llm_feedback") ||
            inherits(extraction_result, "llm_feedback_tool_result")
        ) {
          if (inherits(extraction_result, "llm_feedback_tool_result")) {
            # This ensures tool results are not filtered out when cleaning
            #   the context window in send_prompt()
            response <- send_chat(extraction_result, tool_result = TRUE)
          } else {
            response <- send_chat(extraction_result, tool_result = FALSE)
          }
          any_prompt_wrap_not_done <- TRUE
          break
        }

        if (inherits(extraction_result, "llm_break_soft")) {
          interactions <- max_interactions
        }

        if (inherits(extraction_result, "llm_break")) {
          # Still apply remaining prompt wraps of type 'check'
          #   (this may block the break when feedback is returned)
          if (pw_index < length(prompt_wraps)) {
            prompt_wraps_remaining <- prompt_wraps[
              (pw_index + 1):length(prompt_wraps)
            ]
            for (prompt_wrap_remaining in prompt_wraps_remaining) {
              if (prompt_wrap_remaining$type == "check") {
                check_result <-
                  prompt_wrap_remaining$validation_fn(
                    response,
                    llm_provider,
                    http
                  )
                if (inherits(check_result, "llm_feedback")) {
                  response <- send_chat(check_result, tool_result = TRUE)
                  any_prompt_wrap_not_done <- TRUE
                  break
                }
              }
            }
          }

          if (!extraction_result$success) {
            any_prompt_wrap_not_done <- TRUE # Will result in no success
          } else {
            any_prompt_wrap_not_done <- FALSE # Will result in success
          }
          response <- extraction_result$object_to_return
          llm_break <- TRUE
          break
        }

        # If no llm_feedback or break, extraction was succesful
        response <- extraction_result
      }

      # Apply validation function
      if (!is.null(prompt_wrap$validation_fn)) {
        validation_result <-
          prompt_wrap$validation_fn(response, llm_provider, http)

        # If it inherits llm_feedback, send the feedback to the LLM & get new response
        if (inherits(validation_result, "llm_feedback")) {
          # Set as 'tool_result' to not clear user feedback when type = 'check'
          tool_result <- FALSE
          if (prompt_wrap$type == "check") {
            tool_result <- TRUE
          }

          response <- send_chat(validation_result, tool_result = tool_result)
          any_prompt_wrap_not_done <- TRUE
          break
        }

        if (inherits(validation_result, "llm_break_soft")) {
          interactions <- max_interactions
        }

        if (inherits(validation_result, "llm_break")) {
          # Still apply remaining prompt wraps of type 'check'
          #   (this may block the break when feedback is returned)
          if (pw_index < length(prompt_wraps)) {
            prompt_wraps_remaining <- prompt_wraps[
              (pw_index + 1):length(prompt_wraps)
            ]
            for (prompt_wrap_remaining in prompt_wraps_remaining) {
              if (prompt_wrap_remaining$type == "check") {
                check_result <-
                  prompt_wrap_remaining$validation_fn(
                    response,
                    llm_provider,
                    http
                  )
                if (inherits(check_result, "llm_feedback")) {
                  response <- send_chat(check_result, tool_result = TRUE)
                  any_prompt_wrap_not_done <- TRUE
                  break
                }
              }
            }
          }

          if (!validation_result$success) {
            any_prompt_wrap_not_done <- TRUE # Will result in no success
          } else {
            any_prompt_wrap_not_done <- FALSE # Will result in success
          }
          response <- validation_result$object_to_return
          llm_break <- TRUE
          break
        }
      }
    }

    if (!any_prompt_wrap_not_done) {
      success <- TRUE
    }

    if (llm_break) break
  }

  ## 5 Final evaluation

  if (!success) {
    warning(
      paste0(
        "Failed to reach a valid answer after ",
        interactions,
        " interactions"
      )
    )
    response <- NULL
  }

  if (return_mode == "only_response") {
    return(response)
  }

  if (return_mode == "full") {
    return_list <- list()

    chat_history <- normalize_chat_history_metadata(chat_history)

    return_list$response <- response
    return_list$interactions <- interactions
    return_list$chat_history <- chat_history

    if (clean_chat_history) {
      return_list$chat_history_clean <- clean_chat_history(chat_history)
    }

    return_list$start_time <- start_time
    return_list$end_time <- Sys.time()

    return_list$duration_seconds <-
      as.numeric(
        difftime(
          return_list$end_time,
          return_list$start_time,
          units = "secs"
        )
      )

    return_list$http <- http
    return_list$ellmer_chat <- ellmer_chat

    return(return_list)
  }
}

#' Create chat history dataframe
#'
#' @description Internal function used by [send_prompt()] to create a chat history
#' dataframe. This dataframe is used to keep track of the chat history during
#' the interactions with the LLM provider.
#
#' @param role Role of the message (e.g. "user", "assistant", "system")
#' @param content Content of the message
#' @param tool_result Logical indicating whether the message is a tool result.
#' This will ensure it is not removed when the chat history context window is
#' cleaned by `clean_chat_history()` (which is an internal function which will keep only the
#' first and last message from the user, the last message from the assistant,
#' all messages from the system, and all tool_result messages)
#'
#' @return A dataframe with the chat history
#'
#' @noRd
#' @keywords internal
create_chat_df <- function(
  role = character(),
  content = character(),
  tool_result = logical()
) {
  data.frame(role = role, content = content, tool_result = tool_result)
}

#' Clean chat history
#'
#' @description Internal function used by [send_prompt()] to clean the chat history
#' context window after each interaction with the LLM provider.
#'
#' This function will keep only the first and last message from the user, the last
#' message from the assistant, all messages from the system, and all tool_result messages.
#'
#' This may be useful to keep the context window clean and increase the LLM's performance.
#'
#' @param chat_history A dataframe with the full chat history
#'
#' @return A dataframe with the cleaned chat history
#'
#' @noRd
#' @keywords internal
clean_chat_history <- function(chat_history) {
  filtered <- chat_history_to_send(chat_history)
  # source_rows maps positions in `filtered` back to the original
  # chat_history rows (accounting for hidden / tool_call filtering).
  original_source_rows <- attr(filtered, "source_rows") %||%
    seq_len(nrow(filtered))

  # Keep only first and last message from user;
  # keep only last message from assistant;
  # keep all messages from system;
  # keep all tool_result messages
  user_rows <- which(filtered$role == "user")
  assistant_rows <- which(filtered$role == "assistant")
  system_rows <- which(filtered$role == "system")
  tool_result_rows <- which(
    filtered$tool_result | filtered$role == "tool"
  )

  keep_rows <- c(
    system_rows,
    user_rows[c(1, length(user_rows))],
    utils::tail(assistant_rows, 1),
    tool_result_rows
  )

  keep_rows <- sort(unique(keep_rows))

  # Subset the dataframe with these rows
  cleaned_chat_history <- filtered[keep_rows, ]
  # Compose the two mappings: keep_rows indexes into `filtered`, whose

  # positions map back to the original frame via original_source_rows.
  attr(cleaned_chat_history, "source_rows") <- original_source_rows[keep_rows]
  # (sort(unique()) is used to ensure that the rows are in order)

  return(cleaned_chat_history)
}

test_that("answer_as_boolean works correctly", {
  skip_test_if_no_ollama()

  response <- "Is the sky blue?" |>
    answer_as_boolean() |>
    send_prompt(llm_provider_ollama(), verbose = FALSE)

  expect_true(is.logical(response))
  expect_true(response)
})

test_that("answer_as_boolean handles true_definition and false_definition", {
  prompt <- "Is water wet?" |>
    answer_as_boolean(
      true_definition = "Water is wet",
      false_definition = "Water is not wet"
    ) |>
    construct_prompt_text()

  expect_true(grepl("TRUE means: Water is wet", prompt))
  expect_true(grepl("FALSE means: Water is not wet", prompt))
})

testthat::test_that("send_prompt deep-clones the ellmer chat object", {
  fake_chat <- fake_ellmer_chat()
  provider <- llm_provider_ellmer(
    fake_chat,
    parameters = list(stream = FALSE),
    verbose = FALSE
  )

  # Remember the original chat identity

  original_chat <- provider$ellmer_chat

  result <- "Hello" |>
    send_prompt(provider, return_mode = "full", verbose = FALSE)

  # The provider's chat should still be the original (untouched)
  testthat::expect_identical(provider$ellmer_chat, original_chat)

  # The returned ellmer_chat should be a different object
  testthat::expect_false(identical(result$ellmer_chat, original_chat))
})

testthat::test_that("direct raw ellmer chats are reset before parameter_fn runs", {
  withr::local_options(list(tidyprompt.stream = FALSE))

  raw_env <- new.env(parent = emptyenv())
  provider_env <- new.env(parent = emptyenv())

  capture_turns <- function(prompt, store_env) {
    prompt |>
      prompt_wrap(parameter_fn = function(llm_provider) {
        store_env$turns <- llm_provider$ellmer_chat$get_turns()
        list()
      })
  }

  raw_chat <- fake_ellmer_chat(turns = list("existing-turn"))
  "Hello" |>
    capture_turns(raw_env) |>
    send_prompt(raw_chat, verbose = FALSE)

  explicit_provider <- llm_provider_ellmer(
    fake_ellmer_chat(turns = list("existing-turn")),
    parameters = list(stream = FALSE),
    verbose = FALSE
  )
  "Hello" |>
    capture_turns(provider_env) |>
    send_prompt(explicit_provider, verbose = FALSE)

  testthat::expect_length(raw_env$turns, 0)
  testthat::expect_length(provider_env$turns, 1)
})

test_that("answer_as_int returns int", {
  skip_test_if_no_ollama()

  ollama <- llm_provider_ollama()

  response <- "How much is 2+2?" |>
    answer_as_integer() |>
    send_prompt(ollama, verbose = FALSE)

  expect_true(is.numeric(response))
})

test_that("answer_as_integer adds instruction", {
  prompt <- "What is 5+5?" |>
    answer_as_integer() |>
    construct_prompt_text()

  expect_true(
    grepl(
      "You must answer with only an integer (use no other characters).",
      prompt,
      fixed = TRUE
    )
  )
})

test_that("LLM can exit evaluation via quit_if", {
  skip_test_if_no_ollama()

  response <- "what is the favorite number of the dog of my step father who moved to the north pole?" |>
    set_system_prompt("you are an assistant who does not guess things") |>
    answer_as_integer() |>
    quit_if() |>
    send_prompt(llm_provider_ollama(), verbose = FALSE)

  expect_true(is.null(response))
})

# Tests for tool scope isolation across successive ellmer-backed calls

test_that("prompt-level tools do not leak across successive send_prompt calls", {
  fake_chat <- fake_ellmer_chat()
  provider <- llm_provider_ellmer(
    fake_chat,
    parameters = list(stream = FALSE),
    verbose = FALSE
  )

  tool_a <- list(name = "tool_a")
  tool_b <- list(name = "tool_b")

  # First call: send_prompt with tool_a via .ellmer_tools parameter
  prompt_with_tool_a <- "Use tool A" |>
    prompt_wrap(parameter_fn = function(llm_provider) {
      list(.ellmer_tools = list(tool_a))
    })

  res1 <- send_prompt(
    prompt_with_tool_a,
    provider,
    verbose = FALSE,
    return_mode = "full"
  )

  # The returned ellmer_chat should only have tool_a registered
  expect_length(res1$ellmer_chat$get_tools(), 1)

  expect_identical(res1$ellmer_chat$get_tools()[[1]]$name, "tool_a")

  # Second call: plain prompt with no tools
  res2 <- "No tools" |>
    send_prompt(provider, verbose = FALSE, return_mode = "full")

  # The returned ellmer_chat should have NO tools registered
  expect_length(res2$ellmer_chat$get_tools(), 0)

  # Third call: prompt with only tool_b
  prompt_with_tool_b <- "Use tool B" |>
    prompt_wrap(parameter_fn = function(llm_provider) {
      list(.ellmer_tools = list(tool_b))
    })

  res3 <- send_prompt(
    prompt_with_tool_b,
    provider,
    verbose = FALSE,
    return_mode = "full"
  )

  # Should have ONLY tool_b, not tool_a + tool_b
  expect_length(res3$ellmer_chat$get_tools(), 1)
  expect_identical(res3$ellmer_chat$get_tools()[[1]]$name, "tool_b")
})

test_that("user-pre-registered tools are preserved across calls", {
  fake_chat <- fake_ellmer_chat()
  # User pre-registers a tool on the chat before creating the provider
  fake_chat$set_tools(list(list(name = "user_tool")))

  provider <- llm_provider_ellmer(
    fake_chat,
    parameters = list(stream = FALSE),
    verbose = FALSE
  )

  # Plain call with no prompt tools -- base tool stays
  res1 <- "Hello" |>
    send_prompt(provider, verbose = FALSE, return_mode = "full")

  expect_length(res1$ellmer_chat$get_tools(), 1)
  expect_identical(res1$ellmer_chat$get_tools()[[1]]$name, "user_tool")

  # Call with a prompt tool -- base tool + prompt tool
  prompt_with_tool_a <- "Use tool A" |>
    prompt_wrap(parameter_fn = function(llm_provider) {
      list(.ellmer_tools = list(list(name = "tool_a")))
    })

  res2 <- send_prompt(
    prompt_with_tool_a,
    provider,
    verbose = FALSE,
    return_mode = "full"
  )

  expect_length(res2$ellmer_chat$get_tools(), 2)
  tool_names <- vapply(
    res2$ellmer_chat$get_tools(),
    `[[`,
    character(1),
    "name"
  )
  expect_true("user_tool" %in% tool_names)
  expect_true("tool_a" %in% tool_names)

  # Follow-up plain call -- back to just the base tool
  res3 <- "Again" |>
    send_prompt(provider, verbose = FALSE, return_mode = "full")

  expect_length(res3$ellmer_chat$get_tools(), 1)
  expect_identical(res3$ellmer_chat$get_tools()[[1]]$name, "user_tool")
})

test_that("prompt-level tools do not leak through persistent_chat", {
  fake_chat <- fake_ellmer_chat()
  # User pre-registers a base tool
  fake_chat$set_tools(list(list(name = "base_tool")))

  provider <- llm_provider_ellmer(
    fake_chat,
    parameters = list(stream = FALSE),
    verbose = FALSE
  )

  # First call with prompt tool_a
  prompt_with_tool_a <- "Use tool A" |>
    prompt_wrap(parameter_fn = function(llm_provider) {
      list(.ellmer_tools = list(list(name = "tool_a")))
    })

  res1 <- send_prompt(
    prompt_with_tool_a,
    provider,
    verbose = FALSE,
    return_mode = "full"
  )

  # Should have base_tool + tool_a
  expect_length(res1$ellmer_chat$get_tools(), 2)

  # Second call with no prompt tools -- only base_tool should remain
  res2 <- "No tools" |>
    send_prompt(provider, verbose = FALSE, return_mode = "full")

  expect_length(res2$ellmer_chat$get_tools(), 1)
  expect_identical(res2$ellmer_chat$get_tools()[[1]]$name, "base_tool")
})

test_that("tools registered after provider construction are seen", {
  fake_chat <- fake_ellmer_chat()

  provider <- llm_provider_ellmer(
    fake_chat,
    parameters = list(stream = FALSE),
    verbose = FALSE
  )

  # Register a tool AFTER construction
  provider$ellmer_chat$set_tools(list(list(name = "late_tool")))

  # Plain call -- should see the late tool
  res1 <- "Hello" |>
    send_prompt(provider, verbose = FALSE, return_mode = "full")

  expect_length(res1$ellmer_chat$get_tools(), 1)
  expect_identical(res1$ellmer_chat$get_tools()[[1]]$name, "late_tool")

  # Call with prompt tool -- should see late_tool + prompt_tool
  prompt_with_tool <- "Use tool" |>
    prompt_wrap(parameter_fn = function(llm_provider) {
      list(.ellmer_tools = list(list(name = "prompt_tool")))
    })

  res2 <- send_prompt(
    prompt_with_tool,
    provider,
    verbose = FALSE,
    return_mode = "full"
  )

  expect_length(res2$ellmer_chat$get_tools(), 2)
  tool_names <- vapply(
    res2$ellmer_chat$get_tools(),
    `[[`,
    character(1),
    "name"
  )
  expect_true("late_tool" %in% tool_names)
  expect_true("prompt_tool" %in% tool_names)

  # Follow-up plain call -- back to late_tool only
  res3 <- "Again" |>
    send_prompt(provider, verbose = FALSE, return_mode = "full")

  expect_length(res3$ellmer_chat$get_tools(), 1)
  expect_identical(res3$ellmer_chat$get_tools()[[1]]$name, "late_tool")
})

test_that("prompt-level tools are not duplicated on retries within send_prompt", {
  fake_chat <- fake_ellmer_chat()
  provider <- llm_provider_ellmer(
    fake_chat,
    parameters = list(stream = FALSE),
    verbose = FALSE
  )

  # Counter to trigger a retry on the first call
  state <- new.env(parent = emptyenv())
  state$call_n <- 0L

  prompt_with_retry <- "Retry me" |>
    prompt_wrap(parameter_fn = function(llm_provider) {
      list(.ellmer_tools = list(list(name = "tool_a")))
    }) |>
    prompt_wrap(
      validation_fn = function(x) {
        state$call_n <- state$call_n + 1L
        if (state$call_n == 1L) {
          return(llm_feedback("try again"))
        }
        x
      },
      type = "check"
    )

  res <- send_prompt(
    prompt_with_retry,
    provider,
    verbose = FALSE,
    return_mode = "full"
  )

  # Provider was called at least twice (first feedback, then success)
  expect_true(state$call_n >= 2L)

  # Despite multiple invocations, tool_a should appear exactly once
  tool_names <- vapply(
    res$ellmer_chat$get_tools(),
    `[[`,
    character(1),
    "name"
  )
  expect_equal(sum(tool_names == "tool_a"), 1L)
})

test_that("can create prompt_wrap", {
  prompt <- prompt_wrap("aaa", modify_fn = \(x) "bbb")
  expect_s3_class(prompt, "Tidyprompt")
})

test_that("can add prompt wraps of all types and they are ordered correctly", {
  prompt <- "Test prompt" |>
    prompt_wrap(extraction_fn = \(x) x, type = "unspecified") |>
    prompt_wrap(extraction_fn = \(x) x, type = "mode") |>
    prompt_wrap(modify_fn = \(x) x, type = "tool") |>
    prompt_wrap(modify_fn = \(x) x, type = "break") |>
    prompt_wrap(validation_fn = \(x) x, type = "check")

  wraps_mod <- prompt$get_prompt_wraps(order = "modification")
  types_in_order_mod <- sapply(wraps_mod, function(x) x$type)
  expect_equal(
    types_in_order_mod,
    c("check", "unspecified", "break", "mode", "tool")
  )

  wraps_eval <- prompt$get_prompt_wraps(order = "evaluation")
  types_in_order_eval <- sapply(wraps_eval, function(x) x$type)
  expect_equal(
    types_in_order_eval,
    c("tool", "mode", "break", "unspecified", "check")
  )
})

test_that("handler fn works", {
  fake <- llm_provider_fake()

  result <- "Hi" |>
    prompt_wrap(
      handler_fn = function(response, llm_provider) {
        if (!inherits(llm_provider, "LlmProvider")) {
          stop("llm_provider passed by handler_fn is not an LlmProvider")
        }

        response$completed$content[
          nrow(response$completed)
        ] <- "beepido boop ba"
        return(response)
      }
    ) |>
    send_prompt(fake)

  expect_identical(result, "beepido boop ba")
})

test_that("parameter_fn throws no error", {
  fake <- llm_provider_fake()

  expect_no_error(
    "Hi" |>
      prompt_wrap(
        parameter_fn = function(llm_provider) {
          if (!inherits(llm_provider, "LlmProvider")) {
            stop("llm_provider passed by parameter_fn is not an LlmProvider")
          }

          return(
            list(
              beep = "boop"
            )
          )
        }
      ) |>
      send_prompt(fake)
  )
})

test_that("answer_as_key_value adds key-value pair instruction", {
  prompt <- "List the capitals of countries." |>
    answer_as_key_value(
      key_name = "country",
      value_name = "capital",
      n_unique_items = 2,
      list_mode = "bullet"
    ) |>
    construct_prompt_text()

  expect_true(
    grepl(
      paste0(
        "Respond with a list of key-value pairs, like so:\n  ",
        "-- <<country 1>>: <<capital 1>>\n  ",
        "-- <<country 2>>: <<capital 2>>"
      ),
      prompt,
      fixed = TRUE
    )
  )
  expect_true(
    grepl(
      "The list should contain 2 unique pairs.",
      prompt,
      fixed = TRUE
    )
  )
})

test_that("answer_as_key_value returns key-value pairs", {
  skip_test_if_no_ollama()

  ollama <- llm_provider_ollama()

  response <- "Provide the capitals of these countries: Germany, France, Japan." |>
    answer_as_key_value(
      key_name = "country",
      value_name = "capital",
      list_mode = "comma"
    ) |>
    send_prompt(ollama)

  expect_true(is.list(response))
  expect_equal(length(response), 3)
  expect_true(all(c("Germany", "France", "Japan") %in% names(response)))
  expect_true(all(c("Berlin", "Paris", "Tokyo") %in% response))
})

test_that("extraction and validation works", {
  fake_llm <- llm_provider_fake()

  response <- "What is 2 + 2?" |>
    answer_by_chain_of_thought() |>
    answer_as_integer() |>
    send_prompt(fake_llm, verbose = TRUE)

  is_whole_number <- function(x) {
    is.numeric(x) && x == floor(x)
  }

  expect_true(is_whole_number(response))
  expect_equal(response, 4)
})

test_that("full return mode works", {
  fake_llm <- llm_provider_fake()

  response <- "hi" |>
    send_prompt(
      fake_llm,
      return_mode = "full",
      clean_chat_history = TRUE
    )

  expect_type(response$response, "character")
  expect_type(response$interactions, "double")

  expect_true(length(response$response) == 1)
  expect_true(is.data.frame(response$chat_history))
  expect_true(is.data.frame(response$chat_history_clean))
  expect_true(
    is.numeric(response$interactions) &
      response$interactions > 0 &
      response$interactions == floor(response$interactions)
  )
  expect_true(is.double(response$duration_seconds))
})

test_that("send_prompt accepts raw ellmer chats directly", {
  withr::local_options(list(tidyprompt.stream = FALSE))

  raw_chat <- fake_ellmer_chat(turns = list("old-turn"))

  result <- "Hello" |>
    send_prompt(raw_chat, return_mode = "full", verbose = FALSE)

  expect_equal(result$response, "chat-response:Hello")
  expect_identical(raw_chat$turns, list("old-turn"))
  expect_false(identical(result$ellmer_chat, raw_chat))
  expect_length(result$ellmer_chat$last_method$turns, 0)
})

test_that("send_prompt keeps llm_provider_ellmer stream defaults for raw chats", {
  skip_if_not_installed("coro")
  withr::local_options(list(tidyprompt.stream = TRUE))

  raw_chat <- fake_ellmer_chat()

  result <- "Hello" |>
    send_prompt(raw_chat, return_mode = "full", verbose = FALSE)

  expect_identical(result$ellmer_chat$last_method$method, "stream")
})

test_that("send_prompt does not resend hidden history rows on retry", {
  feedback_sent <- FALSE
  state <- new.env(parent = emptyenv())
  state$provider_calls <- list()
  state$provider_call_n <- 0L

  provider <- `llm_provider-class`$new(
    complete_chat_function = function(chat_history) {
      state <- self$parameters$.test_state
      state$provider_call_n <- state$provider_call_n + 1L
      state$provider_calls[[state$provider_call_n]] <- chat_history

      if (state$provider_call_n == 1L) {
        completed <- dplyr::bind_rows(
          chat_history,
          data.frame(
            role = c("assistant", "assistant"),
            content = c("Reasoning step", "Initial answer"),
            hidden_from_llm = c(TRUE, FALSE),
            stringsAsFactors = FALSE
          )
        )
      } else {
        completed <- dplyr::bind_rows(
          chat_history,
          data.frame(
            role = "assistant",
            content = "Fixed answer",
            stringsAsFactors = FALSE
          )
        )
      }

      list(
        completed = completed,
        http = list(request = NULL, response = NULL)
      )
    },
    parameters = list(.test_state = state),
    verbose = FALSE
  )

  prompt <- "Think through this" |>
    prompt_wrap(validation_fn = function(response) {
      if (!feedback_sent) {
        feedback_sent <<- TRUE
        return(llm_feedback("Please fix format"))
      }

      TRUE
    })

  result <- send_prompt(
    prompt,
    provider,
    return_mode = "full",
    verbose = FALSE
  )

  expect_equal(result$response, "Fixed answer")
  expect_equal(state$provider_calls[[1]]$content, "Think through this")
  expect_equal(
    state$provider_calls[[2]]$content,
    c("Think through this", "Initial answer", "Please fix format")
  )
  expect_false(any(state$provider_calls[[2]]$content == "Reasoning step"))
  expect_equal(
    result$chat_history$hidden_from_llm,
    c(FALSE, TRUE, FALSE, FALSE, FALSE)
  )
})

test_that("send_prompt preserves provider metadata updates on existing rows", {
  provider <- `llm_provider-class`$new(
    complete_chat_function = function(chat_history) {
      updated_history <- chat_history
      updated_history$native_turn_id <- rep(
        NA_character_,
        nrow(updated_history)
      )
      updated_history$native_turn_id[nrow(updated_history)] <- "turn-1"
      updated_history$native_turn_role <- rep(
        NA_character_,
        nrow(updated_history)
      )
      updated_history$native_turn_role[nrow(updated_history)] <- "user"

      list(
        completed = dplyr::bind_rows(
          updated_history,
          data.frame(
            role = "assistant",
            content = "reply",
            stringsAsFactors = FALSE
          )
        ),
        http = list(request = NULL, response = NULL)
      )
    },
    verbose = FALSE
  )

  result <- send_prompt(
    "Hello",
    provider,
    return_mode = "full",
    verbose = FALSE
  )

  expect_equal(result$response, "reply")
  expect_equal(result$chat_history$native_turn_id[1], "turn-1")
  expect_equal(result$chat_history$native_turn_role[1], "user")
})

test_that("send_prompt preserves metadata updates when hidden rows exist", {
  feedback_sent <- FALSE
  state <- new.env(parent = emptyenv())
  state$provider_calls <- list()
  state$provider_call_n <- 0L

  provider <- `llm_provider-class`$new(
    complete_chat_function = function(chat_history) {
      state <- self$parameters$.test_state
      state$provider_call_n <- state$provider_call_n + 1L
      state$provider_calls[[state$provider_call_n]] <- chat_history

      if (state$provider_call_n == 1L) {
        # Provider adds metadata to the sent user row
        chat_history$native_turn_id <- rep(
          NA_character_,
          nrow(chat_history)
        )
        chat_history$native_turn_id[nrow(chat_history)] <- "turn-1"

        completed <- dplyr::bind_rows(
          chat_history,
          data.frame(
            role = c("assistant", "assistant"),
            content = c("Hidden reasoning", "First answer"),
            hidden_from_llm = c(TRUE, FALSE),
            stringsAsFactors = FALSE
          )
        )
      } else {
        # On retry, provider adds metadata to the (re-sent) user row
        chat_history$native_turn_id <- rep(
          NA_character_,
          nrow(chat_history)
        )
        chat_history$native_turn_id[nrow(chat_history)] <- "turn-retry"

        completed <- dplyr::bind_rows(
          chat_history,
          data.frame(
            role = "assistant",
            content = "Fixed answer",
            stringsAsFactors = FALSE
          )
        )
      }

      list(
        completed = completed,
        http = list(request = NULL, response = NULL)
      )
    },
    parameters = list(.test_state = state),
    verbose = FALSE
  )

  prompt <- "Hello" |>
    prompt_wrap(validation_fn = function(response) {
      if (!feedback_sent) {
        feedback_sent <<- TRUE
        return(llm_feedback("Please fix"))
      }
      TRUE
    })

  result <- send_prompt(
    prompt,
    provider,
    return_mode = "full",
    verbose = FALSE
  )

  expect_equal(result$response, "Fixed answer")

  # The feedback user row should carry the metadata from the retry call,
  # even though a hidden row existed earlier in the transcript.  Without
  # the merge fix the provider's metadata update on sent rows is discarded
  # because complete_chat reconstructs from the original (pre-send) history.
  user_rows <- which(result$chat_history$role == "user")
  expect_true(length(user_rows) >= 2)
  # Last user row is the feedback row — provider tagged it "turn-retry"
  expect_equal(
    result$chat_history$native_turn_id[user_rows[length(user_rows)]],
    "turn-retry"
  )
})

test_that("clean_chat_history=TRUE merges metadata correctly with hidden rows", {
  feedback_sent <- FALSE
  state <- new.env(parent = emptyenv())
  state$provider_calls <- list()
  state$provider_call_n <- 0L

  provider <- `llm_provider-class`$new(
    complete_chat_function = function(chat_history) {
      state <- self$parameters$.test_state
      state$provider_call_n <- state$provider_call_n + 1L
      state$provider_calls[[state$provider_call_n]] <- chat_history

      if (state$provider_call_n == 1L) {
        completed <- dplyr::bind_rows(
          chat_history,
          data.frame(
            role = c("assistant", "assistant"),
            content = c("Hidden step", "First answer"),
            hidden_from_llm = c(TRUE, FALSE),
            stringsAsFactors = FALSE
          )
        )
      } else {
        completed <- dplyr::bind_rows(
          chat_history,
          data.frame(
            role = "assistant",
            content = "Fixed answer",
            stringsAsFactors = FALSE
          )
        )
      }

      list(
        completed = completed,
        http = list(request = NULL, response = NULL)
      )
    },
    parameters = list(.test_state = state),
    verbose = FALSE
  )

  prompt <- "Hello" |>
    prompt_wrap(validation_fn = function(response) {
      if (!feedback_sent) {
        feedback_sent <<- TRUE
        return(llm_feedback("Please fix"))
      }
      TRUE
    })

  result <- send_prompt(
    prompt,
    provider,
    return_mode = "full",
    clean_chat_history = TRUE,
    verbose = FALSE
  )

  expect_equal(result$response, "Fixed answer")

  # After first call: rows are [user, assistant(hidden), assistant]
  # After feedback: rows are [user, assistant(hidden), assistant, user(feedback)]
  # clean_chat_history filters hidden row 2 → sends rows 1,3,4
  # Bug: old code mapped these back to positions 1,2,3 instead of 1,3,4,
  # which would overwrite the hidden row and misplace content.
  # Verify the hidden row is preserved intact:
  expect_equal(result$chat_history$content[2], "Hidden step")
  expect_true(result$chat_history$hidden_from_llm[2])

  # Verify content didn't get shuffled into wrong rows
  expect_equal(result$chat_history$role[1], "user")
  expect_equal(result$chat_history$content[1], "Hello")
  expect_equal(result$chat_history$role[3], "assistant")
  expect_equal(result$chat_history$content[3], "First answer")

  # Feedback row should still be at its original position (row 4)
  expect_equal(result$chat_history$role[4], "user")
  expect_equal(result$chat_history$content[4], "Please fix")
})

test_that("send_prompt does not resend tool call rows on retry", {
  feedback_sent <- FALSE
  state <- new.env(parent = emptyenv())
  state$provider_calls <- list()
  state$provider_call_n <- 0L

  provider <- `llm_provider-class`$new(
    complete_chat_function = function(chat_history) {
      state <- self$parameters$.test_state
      state$provider_call_n <- state$provider_call_n + 1L
      state$provider_calls[[state$provider_call_n]] <- chat_history

      if (state$provider_call_n == 1L) {
        completed <- dplyr::bind_rows(
          chat_history,
          data.frame(
            role = c("assistant", "tool", "assistant"),
            content = c(
              "~>> Calling function 'tool' with arguments:\n{}",
              "~>> Result:\n42",
              "Initial answer"
            ),
            tool_call = c(TRUE, FALSE, FALSE),
            tool_call_id = c("call-1", "call-1", NA_character_),
            tool_result = c(FALSE, TRUE, FALSE),
            stringsAsFactors = FALSE
          )
        )
      } else {
        completed <- dplyr::bind_rows(
          chat_history,
          data.frame(
            role = "assistant",
            content = "Fixed answer",
            stringsAsFactors = FALSE
          )
        )
      }

      list(
        completed = completed,
        http = list(request = NULL, response = NULL)
      )
    },
    parameters = list(.test_state = state),
    verbose = FALSE
  )

  prompt <- "Think through this" |>
    prompt_wrap(validation_fn = function(response) {
      if (!feedback_sent) {
        feedback_sent <<- TRUE
        return(llm_feedback("Please fix format"))
      }

      TRUE
    })

  result <- send_prompt(
    prompt,
    provider,
    return_mode = "full",
    verbose = FALSE
  )

  expect_equal(result$response, "Fixed answer")
  expect_equal(state$provider_calls[[1]]$content, "Think through this")
  expect_equal(
    state$provider_calls[[2]]$content,
    c(
      "Think through this",
      "~>> Result:\n42",
      "Initial answer",
      "Please fix format"
    )
  )
  expect_false(any(grepl(
    "Calling function",
    state$provider_calls[[2]]$content
  )))
  expect_equal(
    result$chat_history$hidden_from_llm,
    c(FALSE, TRUE, FALSE, FALSE, FALSE, FALSE)
  )
})

test_that("send_prompt with raw ellmer chat and stream=FALSE does not require coro", {
  withr::local_options(list(tidyprompt.stream = TRUE))

  # Build a fake chat that has no $stream method (simulating missing coro)
  raw_chat <- fake_ellmer_chat()
  raw_chat$stream <- NULL

  # With stream=FALSE passed to send_prompt, the constructor should not

  # check for coro even though the global option defaults to TRUE.
  result <- "Hello" |>
    send_prompt(raw_chat, stream = FALSE, return_mode = "full", verbose = FALSE)

  expect_equal(result$response, "chat-response:Hello")
  expect_identical(result$ellmer_chat$last_method$method, "chat")
})

# 8. Blank / image-only prompts ------------------------------------------------

test_that("blank prompt does not error and preserves user row", {
  echo_provider <- `llm_provider-class`$new(
    complete_chat_function = function(chat_history) {
      list(
        completed = dplyr::bind_rows(
          chat_history,
          data.frame(
            role = "assistant",
            content = "ok",
            stringsAsFactors = FALSE
          )
        ),
        http = list(request = NULL, response = NULL)
      )
    },
    verbose = FALSE
  )

  result <- "" |>
    send_prompt(echo_provider, return_mode = "full", verbose = FALSE)

  expect_equal(result$response, "ok")
  # The user row (blank content) must be in the history
  user_rows <- result$chat_history[result$chat_history$role == "user", ]
  expect_true(nrow(user_rows) >= 1L)
})

test_that("empty-content user row is not stripped from completed history", {
  # Simulates a provider returning a user row with content ""
  # (e.g., image-only turn); should NOT be dropped by the handler loop.
  provider <- `llm_provider-class`$new(
    complete_chat_function = function(chat_history) {
      list(
        completed = chat_history |>
          dplyr::bind_rows(
            data.frame(
              role = "assistant",
              content = "I see the image",
              stringsAsFactors = FALSE
            )
          ),
        http = list(request = NULL, response = NULL)
      )
    },
    verbose = FALSE
  )

  result <- "" |>
    send_prompt(provider, return_mode = "full", verbose = FALSE)

  user_rows <- result$chat_history[result$chat_history$role == "user", ]
  expect_true(nrow(user_rows) >= 1L)
})

testthat::test_that("ellmer structured output uses native result without round-trip", {
  testthat::skip_if_not_installed("ellmer")

  # The fake chat returns list(result = "ok", type = ...) from chat_structured.
  # After the fix, answer_as_json extraction should receive this directly
  # rather than a JSON round-tripped version.

  fake_chat <- fake_ellmer_chat()
  provider <- llm_provider_ellmer(fake_chat, verbose = FALSE)

  schema <- ellmer::type_object(
    result = ellmer::type_string(),
    .additional_properties = TRUE
  )

  result <- "Return a result" |>
    answer_as_json(schema = schema, type = "ellmer") |>
    send_prompt(provider)

  # The result should be the native R list from chat_structured, not reparsed
  testthat::expect_true(is.list(result))
  testthat::expect_equal(result$result, "ok")
})

testthat::test_that("native_structured_result is cleared after use", {
  testthat::skip_if_not_installed("ellmer")

  fake_chat <- fake_ellmer_chat()
  provider <- llm_provider_ellmer(fake_chat, verbose = FALSE)

  schema <- ellmer::type_object(
    value = ellmer::type_string(),
    .additional_properties = TRUE
  )

  # send_prompt clones the provider, so the original should be unaffected
  result <- "Test" |>
    answer_as_json(schema = schema, type = "ellmer") |>
    send_prompt(provider, return_mode = "full")

  # The original provider should not have the native result stashed
  testthat::expect_null(provider$parameters$.native_structured_result)
})

testthat::test_that("ellmer warns when structured output and tools are both requested", {
  testthat::skip_if_not_installed("ellmer")

  fake_chat <- fake_ellmer_chat()
  provider <- llm_provider_ellmer(fake_chat, verbose = FALSE)

  schema <- ellmer::type_object(
    value = ellmer::type_string(),
    .additional_properties = TRUE
  )
  dummy_tool <- ellmer::tool(
    function() "ok",
    "A dummy tool"
  )

  # Combining answer_as_json (ellmer native) + answer_using_tools (ellmer)
  # should emit a warning about chat_structured suppressing tool use
  tp <- "Do something" |>
    answer_as_json(schema = schema, type = "ellmer") |>
    answer_using_tools(tools = list(dummy = dummy_tool), type = "ellmer")

  testthat::expect_message(
    send_prompt(tp, provider, verbose = FALSE),
    "suppresses tool use"
  )
})

test_that("answer_as_numeric adds instruction", {
  prompt <- "What is 5 / 2?" |>
    answer_as_numeric(min = 0, max = 10) |>
    construct_prompt_text()

  expect_true(
    grepl(
      "You must answer with only a number (use no other characters).",
      prompt,
      fixed = TRUE
    )
  )
  expect_true(
    grepl(
      "Enter a number between 0 and 10.",
      prompt,
      fixed = TRUE
    )
  )
})

test_that("answer_as_numeric extracts numeric responses", {
  provider <- `llm_provider-class`$new(
    complete_chat_function = function(chat_history) {
      list(
        completed = dplyr::bind_rows(
          chat_history,
          data.frame(
            role = "assistant",
            content = "2.5",
            stringsAsFactors = FALSE
          )
        ),
        http = list(request = NULL, response = NULL)
      )
    },
    verbose = FALSE
  )

  result <- "What is 10 divided by 4?" |>
    answer_as_numeric() |>
    send_prompt(provider, verbose = FALSE)

  expect_true(is.numeric(result))
  expect_equal(result, 2.5)
})

test_that("answer_as_numeric retries when response is outside bounds", {
  state <- new.env(parent = emptyenv())
  state$call_n <- 0L

  provider <- `llm_provider-class`$new(
    complete_chat_function = function(chat_history) {
      state <- self$parameters$.test_state
      state$call_n <- state$call_n + 1L
      reply <- c("11", "7.25")[[state$call_n]]

      list(
        completed = dplyr::bind_rows(
          chat_history,
          data.frame(
            role = "assistant",
            content = reply,
            stringsAsFactors = FALSE
          )
        ),
        http = list(request = NULL, response = NULL)
      )
    },
    parameters = list(.test_state = state),
    verbose = FALSE
  )

  result <- "Give me a number between 5 and 10" |>
    answer_as_numeric(min = 5, max = 10) |>
    send_prompt(provider, return_mode = "full", verbose = FALSE)

  expect_equal(result$response, 7.25)
  expect_equal(state$call_n, 2L)
})

#' Make LLM answer as a number (between min and max)
#'
#' @param prompt A single string or a [tidyprompt()] object
#' @param min (optional) Minimum value for the number
#' @param max (optional) Maximum value for the number
#' @param add_instruction_to_prompt (optional) Add instruction for replying
#' as a number to the prompt text. Set to FALSE for debugging if extractions/
#' validations are working as expected (without instruction the answer should fail
#' the validation function, initiating a retry)
#'
#' @return A [tidyprompt()] with an added [prompt_wrap()] which
#' will ensure that the LLM response is a number.
#'
#' @export
#'
#' @example inst/examples/answer_as_numeric.R
#'
#' @family pre_built_prompt_wraps
#' @family answer_as_prompt_wraps
answer_as_numeric <- function(
  prompt,
  min = NULL,
  max = NULL,
  add_instruction_to_prompt = TRUE
) {
  instruction <- "You must answer with only a number (use no other characters)."

  if (!is.null(min) && !is.null(max)) {
    instruction <- paste(
      instruction,
      glue::glue(
        "Enter a number between {min} and {max}."
      )
    )
  } else if (!is.null(min)) {
    instruction <- paste(
      instruction,
      glue::glue(
        "Enter a number greater than or equal to {min}."
      )
    )
  } else if (!is.null(max)) {
    instruction <- paste(
      instruction,
      glue::glue(
        "Enter a number less than or equal to {max}."
      )
    )
  }

  modify_fn <- function(original_prompt_text) {
    if (!add_instruction_to_prompt) {
      return(original_prompt_text)
    }

    glue::glue("{original_prompt_text}\n\n{instruction}")
  }

  extraction_fn <- function(x) {
    extracted <- suppressWarnings(as.numeric(x))
    if (is.na(extracted) || !is.finite(extracted)) {
      return(llm_feedback(instruction))
    }
    return(extracted)
  }

  validation_fn <- function(x) {
    if (!is.null(min) && x < min) {
      return(
        llm_feedback(
          glue::glue(
            "The number should be greater than or equal to {min}."
          )
        )
      )
    }
    if (!is.null(max) && x > max) {
      return(
        llm_feedback(
          glue::glue(
            "The number should be less than or equal to {max}."
          )
        )
      )
    }
    return(TRUE)
  }

  prompt_wrap(
    prompt,
    modify_fn,
    extraction_fn,
    validation_fn,
    name = "answer_as_numeric"
  )
}

#' Print method for Tidyprompt (R6) objects
#'
#' This function prints a Tidyprompt object. It shows:
#' - The base prompt
#' - How many prompt wraps are applied
#' - The fully constructed prompt text after modifications
#'
#' @param x A Tidyprompt R6 object
#' @param ... Additional arguments (unused)
#'
#' @return The Tidyprompt object, invisibly.
#' @exportS3Method print Tidyprompt
#'
#' @noRd
#' @keywords internal
print.Tidyprompt <- function(x, ...) {
  cat(col_blue("<tidyprompt>\n"))

  line_prefix <- col_green("> ")

  format_with_prefix <- function(text, line_prefix) {
    lines <- unlist(strsplit(text, "\n", fixed = TRUE))
    paste0(line_prefix, lines, collapse = "\n")
  }

  # Get the prompt wraps
  wraps <- x$get_prompt_wraps()
  n_wraps <- length(wraps)

  if (n_wraps == 0) {
    cat(col_silver("The base prompt is not modified by prompt wraps:\n"))
    formatted_text <- format_with_prefix(x$base_prompt, line_prefix)
    cat(formatted_text, "\n")
  } else if (n_wraps == 1) {
    cat(
      col_silver(
        "The base prompt is modified by a prompt wrap, resulting in:\n"
      )
    )
  } else if (n_wraps > 1) {
    cat(
      col_silver(
        paste(
          "The base prompt is modified by",
          n_wraps,
          "prompt wraps, resulting in:\n"
        )
      )
    )
  }

  if (n_wraps > 0) {
    full_text <- x$construct_prompt_text()
    formatted_text <- format_with_prefix(full_text, line_prefix)
    cat(formatted_text, "\n")
  }

  cat(col_silver("Use 'x$base_prompt' to show the base prompt text.\n"))
  cat(
    col_silver("Use 'x$construct_prompt_text()' to get the full prompt text.\n")
  )
  if (n_wraps > 0) {
    cat(col_silver("Use 'get_prompt_wraps(x)' to show the prompt wraps.\n"))
  }

  return(invisible(x))
}

col_blue <- function(text) {
  if (requireNamespace("cli", quietly = TRUE)) {
    return(cli::col_blue(text))
  }
  text
}

col_green <- function(text) {
  if (requireNamespace("cli", quietly = TRUE)) {
    return(cli::col_green(text))
  }
  text
}

col_silver <- function(text) {
  if (requireNamespace("cli", quietly = TRUE)) {
    return(cli::col_silver(text))
  }
  text
}

#' Make evaluation of a prompt stop if LLM gives a specific response
#'
#' This function is used to wrap a [tidyprompt()] object and ensure that the
#' evaluation will stop if the LLM says it cannot answer the prompt. This is
#' useful in scenarios where it is determined the LLM is unable to provide a
#' response to a prompt.
#'
#' @param prompt A single string or a [tidyprompt()] object
#' @param quit_detect_regex A regular expression to detect in the LLM's
#' response which will cause the evaluation to stop. The default
#' will detect the string "NO ANSWER" in the response
#' @param instruction A string to be added to the prompt to instruct the LLM
#' how to respond if it cannot answer the prompt. The default is
#' "If you think that you cannot provide a valid answer, you must type: 'NO ANSWER' (use no other characters)".
#' This parameter can be set to `NULL` if no instruction is needed in the prompt
#' @param success A logical indicating whether the [send_prompt()] loop break
#' should nonetheless be considered as a successful completion of the
#' extraction and validation process. If `FALSE`, the `object_to_return` must
#' will always be set to NULL and thus parameter 'response_result' must also
#' be set to 'null'; if `FALSE`, [send_prompt()] will also print a warning
#' about the unsuccessful evaluation. If `TRUE`, the `object_to_return` will be
#' returned as the response result of [send_prompt()] (and [send_prompt()]
#' will print no warning about unsuccessful evaluation); parameter 'response_result'
#' will then determine what is returned as the response result of [send_prompt()].
#' @param response_result A character string indicating what should be returned
#' when the quit_detect_regex is detected in the LLM's response. The default is
#' 'null', which will return NULL as the response result o f [send_prompt()].
#'  Under 'llm_response', the full LLM response will be returned as the
#' response result of [send_prompt()].
#'  Under 'regex_match', the part of the LLM response that matches the
#' quit_detect_regex will be returned as the response result of [send_prompt()]
#'
#' @return A [tidyprompt()] with an added [prompt_wrap()] which will ensure
#' that the evaluation will stop upon detection of the quit_detect_regex in the
#' LLM's response
#'
#' @export
#'
#' @example inst/examples/quit_if.R
#'
#' @family pre_built_prompt_wraps
#' @family miscellaneous_prompt_wraps
quit_if <- function(
  prompt,
  quit_detect_regex = "NO ANSWER",
  instruction = paste0(
    "If you think that you cannot provide a valid answer, you must type:\n",
    "'NO ANSWER' (use no other characters)"
  ),
  success = TRUE,
  response_result = c("null", "llm_response", "regex_match")
) {
  # Validate arguments
  if (!is.null(instruction)) {
    if (!is.character(instruction)) {
      stop("The 'instruction' argument must be of type character.")
    }
    if (length(instruction) != 1) {
      stop("The 'instruction' argument must be a single string.")
    }
  }
  if (!is.character(quit_detect_regex)) {
    stop("The 'quit_detect_regex' argument must be of type character.")
  }
  if (length(quit_detect_regex) != 1) {
    stop("The 'quit_detect_regex' argument must be a single string.")
  }
  if (!is.logical(success)) {
    stop("The 'success' argument must be of type logical.")
  }
  response_result <- match.arg(response_result)
  if (!success && response_result != "null") {
    stop("If 'success' is FALSE, 'response_result' must be of type 'null'")
  }

  modify_fn <- function(original_prompt_text) {
    if (is.null(instruction)) {
      return(original_prompt_text)
    }
    glue::glue("{original_prompt_text}\n\n{instruction}")
  }

  extraction_fn <- function(x) {
    if (grepl(quit_detect_regex, x)) {
      if (response_result == "null") {
        return(
          llm_break(
            object_to_return = NULL,
            success = success
          )
        )
      }
      if (response_result == "llm_response") {
        return(
          llm_break(
            object_to_return = x,
            success = success
          )
        )
      }
      if (response_result == "regex_match") {
        return(
          llm_break(
            object_to_return = regmatches(x, regexpr(quit_detect_regex, x))[[
              1
            ]],
            success = success
          )
        )
      }
    }
    return(x)
  }

  prompt_wrap(prompt, modify_fn, extraction_fn, type = "break")
}

#' Create or validate `chat_history` object
#'
#' This function creates and validates a `chat_history` object, ensuring that it matches
#' the expected format with 'role' and 'content' columns. It has separate methods
#' for `data.frame` and `character` inputs and includes a helper function to add a
#' system prompt to the chat history.
#'
#' @param chat_history A single string, a `data.frame` with 'role' and 'content' columns,
#' or NULL. If a `data.frame` is provided, it should contain 'role' and 'content' columns,
#' where 'role' is either 'user', 'assistant', 'system', or 'tool', and 'content' is a character string
#' representing a chat message
#'
#' @return A valid chat history `data.frame` (of class `chat_history`)
#'
#' @export
#'
#' @example inst/examples/chat_history.R
#'
#' @family chat_history
chat_history <- function(chat_history) {
  UseMethod("chat_history")
}

#' Default method for `chat_history()`
#'
#' Calls error which indicates that the input was not a `character` or `data.frame`.
#'
#' @details When input is a `character` or `data.frame`, the appropriate method will be called
#' (see `[chat_history.character()] and [chat_history.data.frame()]).
#'
#' @param chat_history Object which is not `character` or `data.frame`
#'
#' @return No return value; an error is thrown
#'
#' @exportS3Method chat_history default
#' @keywords internal
chat_history.default <- function(chat_history) {
  stop(
    "The input must be either a data frame with 'role' and 'content' columns, or a single string."
  )
}

#' Method for `chat_history()` when the input is a single string
#'
#' Creates a `chat_history` object from a single string.
#'
#' @param chat_history A single string
#'
#' @return A valid chat history `data.frame` (of class `chat_history`),
#' with the 'role' set to 'user' and the 'content' set to the input string
#'
#' @exportS3Method chat_history character
#' @keywords internal
chat_history.character <- function(chat_history) {
  if (length(chat_history) != 1) {
    stop("A single character string is expected for chat history input.")
  }
  chat_data <- data.frame(
    role = "user",
    content = chat_history,
    tool_result = FALSE,
    stringsAsFactors = FALSE
  )
  class(chat_data) <- c("chat_history", class(chat_data))
  return(chat_data)
}

#' Method for `chat_history()` when the input is a `data.frame`
#'
#' Creates a `chat_history` object from a data frame.
#'
#' @param chat_history A data frame with 'role' and 'content' columns,
#' where 'role' is either 'user', 'assistant', or 'system', and 'content' is a character string
#' representing a chat message
#'
#' @return A valid chat history `data.frame` (of class `chat_history`),
#' with the 'role' and 'content' columns as specified in the input data frame
#'
#' @exportS3Method chat_history data.frame
#' @keywords internal
chat_history.data.frame <- function(chat_history) {
  # Filter out rows with all 'NA'
  chat_history <- chat_history[
    !rowSums(is.na(chat_history)) == ncol(chat_history),
  ]

  if (!all(c("role", "content") %in% names(chat_history))) {
    stop("The data frame must contain'role' and 'content' columns.")
  }
  if (!all(chat_history$role %in% c("user", "assistant", "system", "tool"))) {
    stop(
      "The 'role' column must contain only 'user', 'assistant', 'system', or 'tool'."
    )
  }
  if (!is.character(chat_history$role)) {
    stop("The 'role' column must be of type character.")
  }
  if (!is.character(chat_history$content)) {
    stop("The 'content' column must be of type character.")
  }

  # Add tool_result column if missing
  if (!"tool_result" %in% names(chat_history)) {
    chat_history$tool_result <- FALSE
  }

  class(chat_history) <- c("chat_history", class(chat_history))
  return(chat_history)
}

#' Add a message to a chat history
#'
#' This function appends a message to a [chat_history()] object.
#' The function can automatically determine the role of the message to be
#' added based on the last message in the chat history. The role can also be
#' manually specified.
#'
#' @details The chat_history object may be of different types:
#' \itemize{
#' \item A single string: The function will create a new chat history object
#' with the string as the first message; the role of that first message will be
#' "user"
#' \item A data.frame: The function will append the message to the data.frame.
#' The data.frame must be a valid chat history; see [chat_history()]
#' \item A list: The function will extract the chat history from the list.
#' The list must contain a valid chat history under the key 'chat_history'.
#' This may typically be the result from [send_prompt()] when using
#' 'return_mode = "full"'
#' \item A Tidyprompt object ([tidyprompt-class]): The function will extract the chat history
#' from the object. This will be done by concatenating the 'system_prompt',
#' 'chat_history', and 'base_prompt' into a chat history data.frame. Note
#' that the other properties of the [tidyprompt-class] object will be lost
#' \item NULL: The function will create a new chat history object
#' with no messages; the message will be the first message
#' }
#'
#' @param chat_history A single string, a data.frame which is a valid chat history
#' (see `[chat_history()]`), a list containing a valid chat history under key
#' '$chat_history', a \link{tidyprompt-class} object, or NULL
#'
#' A [chat_history()] object
#'
#' @param message A character string representing the message to add
#'
#' @param role A character string representing the role of the message sender.
#' One of: \itemize{
#' \item "auto": The function automatically determines the role. If the last message
#' was from the user, the role will be "assistant". If the last message was from anything
#' else, the role will be "user"
#' \item "user": The message is from the user
#' \item "assistant": The message is from the assistant
#' \item "system": The message is from the system
#' \item "tool": The message is from a tool (e.g., indicating the result of a function call)
#' }
#'
#' @param tool_result A logical indicating whether the message is a tool result
#' (e.g., the result of a function call)
#'
#' @return A [chat_history()] object with the message added as the last row
#' @export
#'
#' @example inst/examples/chat_history.R
add_msg_to_chat_history <- function(
  chat_history,
  message,
  role = c("auto", "user", "assistant", "system", "tool"),
  tool_result = NULL
) {
  if (is.null(chat_history)) {
    chat_history <- data.frame(
      role = character(),
      content = character(),
      stringsAsFactors = FALSE
    )
  } else if (inherits(chat_history, "Tidyprompt")) {
    chat_history <- chat_history$get_chat_history()
  } else if (
    is.list(chat_history) &&
      !is.data.frame(chat_history) &&
      "chat_history" %in% names(chat_history)
  ) {
    chat_history <- chat_history$chat_history |> chat_history()
  } else {
    chat_history <- chat_history(chat_history)
  }

  role <- match.arg(role)

  stopifnot(
    is.character(message),
    length(message) == 1,
    is.character(role),
    length(role) == 1,
    is.null(tool_result) || is.logical(tool_result)
  )

  # Automatically determine the role if set to "auto"
  if (role == "auto") {
    if (nrow(chat_history) == 0) {
      role <- "user"
    } else {
      role <- ifelse(
        chat_history[nrow(chat_history), "role"] == "user",
        "assistant",
        "user"
      )
    }
  }

  if (is.null(tool_result)) {
    tool_result <- FALSE
  }

  chat_history <- chat_history |>
    dplyr::bind_rows(
      data.frame(
        role = role,
        content = message,
        tool_result = tool_result
      )
    )

  return(chat_history)
}

test_that("chat_history.character creates a valid chat_history data frame", {
  # Test with a single string input
  result <- chat_history("Hello, this is a test message.")
  expect_s3_class(result, "chat_history")
  expect_equal(ncol(result), 3)
  expect_equal(result$role, "user")
  expect_equal(result$content, "Hello, this is a test message.")
})

test_that("chat_history.character errors on non-single string inputs", {
  # Test with multiple strings
  expect_error(
    chat_history(c("Hello", "This should fail")),
    "A single character string is expected for chat history input."
  )
})

test_that("chat_history.data.frame accepts valid chat history data frame", {
  # Valid data frame input
  valid_df <- data.frame(
    role = c("user", "assistant", "system"),
    content = c("User message", "Assistant response", "System message"),
    stringsAsFactors = FALSE
  )
  result <- chat_history(valid_df)
  expect_s3_class(result, "chat_history")
  expect_equal(ncol(result), 3)
  expect_equal(result$role, c("user", "assistant", "system"))
  expect_equal(
    result$content,
    c("User message", "Assistant response", "System message")
  )
})

test_that("chat_history.data.frame errors on invalid role values", {
  # Data frame with invalid role value
  invalid_df <- data.frame(
    role = c("user", "bot"),
    content = c("User message", "Invalid role"),
    stringsAsFactors = FALSE
  )
  expect_error(
    chat_history(invalid_df),
    "The 'role' column must contain only 'user', 'assistant', 'system', or 'tool'."
  )
})

test_that("chat_history.data.frame accepts 'tool' role", {
  tool_df <- data.frame(
    role = c("user", "assistant", "tool"),
    content = c("Hello", "Calling tool", "Tool result"),
    stringsAsFactors = FALSE
  )
  result <- chat_history(tool_df)
  expect_s3_class(result, "chat_history")
  expect_equal(result$role, c("user", "assistant", "tool"))
})

test_that("chat_history.data.frame errors on non-character content column", {
  # Data frame with non-character content column
  invalid_df <- data.frame(
    role = c("user", "assistant"),
    content = c(1, 2),
    stringsAsFactors = FALSE
  )
  expect_error(
    chat_history(invalid_df),
    "The 'content' column must be of type character."
  )
})

test_that("chat_history.default errors on invalid input types", {
  # Test with numeric input
  expect_error(
    chat_history(42),
    "The input must be either a data frame with 'role' and 'content' columns, or a single string."
  )
})

#' Make LLM answer as a boolean (TRUE or FALSE)
#'
#' @param prompt A single string or a [tidyprompt()] object
#' @param true_definition (optional) Definition of what would constitute TRUE.
#' This will be included in the instruction to the LLM. Should be a single string
#' @param false_definition (optional) Definition of what would constitute FALSE.
#' This will be included in the instruction to the LLM. Should be a single string
#' @param add_instruction_to_prompt (optional) Add instruction for replying
#' as a boolean to the prompt text. Set to FALSE for debugging if extractions/validations
#' are working as expected (without instruction the answer should fail the
#' validation function, initiating a retry)
#'
#' @return A [tidyprompt()] with an added [prompt_wrap()] which
#' will ensure that the LLM response is a boolean
#'
#' @export
#'
#' @example inst/examples/answer_as_boolean.R
#'
#' @family pre_built_prompt_wraps
#' @family answer_as_prompt_wraps
answer_as_boolean <- function(
  prompt,
  true_definition = NULL,
  false_definition = NULL,
  add_instruction_to_prompt = TRUE
) {
  instruction <- "You must answer with only TRUE or FALSE (use no other characters)."
  if (!is.null(true_definition)) {
    instruction <- paste(
      instruction,
      glue::glue("TRUE means: {true_definition}.")
    )
  }
  if (!is.null(false_definition)) {
    instruction <- paste(
      instruction,
      glue::glue("FALSE means: {false_definition}.")
    )
  }

  modify_fn <- function(original_prompt_text) {
    if (!add_instruction_to_prompt) {
      return(original_prompt_text)
    }

    glue::glue("{original_prompt_text}\n\n{instruction}")
  }

  extraction_fn <- function(x) {
    normalized <- tolower(trimws(x))
    if (normalized %in% c("true", "false")) {
      return(as.logical(normalized))
    }
    return(llm_feedback(instruction))
  }

  prompt_wrap(
    prompt,
    modify_fn,
    extraction_fn,
    name = "answer_as_boolean"
  )
}

#' Create a new Ollama LLM provider
#'
#' This function creates a new [llm_provider-class] object that interacts with the Ollama API.
#'
#' @param parameters A named list of parameters. Currently the following parameters are required:
#'    - model: The name of the model to use
#'    - stream: A logical indicating whether the API should stream responses
#'
#'  Additional parameters may be passed by adding them to the parameters list;
#'  these parameters will be passed to the Ollama API via the body of the POST request.
#'
#'  Note that various Ollama options need to be set in a list named 'options' within
#'  the parameters list (e.g., context window size is represented in $parameters$options$num_ctx).
#'  For ease of configuration, the 'set_option' and 'set_options' functions are available
#'  (e.g., `$set_option("num_ctx", 1024)` or `$set_options(list(num_ctx = 1024, temperature = 1))`).
#   See available settings at https://github.com/ollama/ollama/blob/main/docs/api.md
#' @param verbose A logical indicating whether the interaction with the LLM provider
#' should be printed to the console
#' @param url The URL to the Ollama API endpoint for chat completion
#' (typically: "http://localhost:11434/api/chat")
#' @param num_ctx The context window size to use.
#' When NULL, the default context window size will be used. This is a function
#' argument for convenience, and will be passed to '$parameters$options$num_ctx'
#'
#' @return A new [llm_provider-class] object for use of the Ollama API
#'
#' @export
#' @example inst/examples/llm_providers.R
#'
#' @family llm_provider
llm_provider_ollama <- function(
  parameters = list(
    model = "llama3.1:8b",
    stream = getOption("tidyprompt.stream", TRUE)
  ),
  verbose = getOption("tidyprompt.verbose", TRUE),
  url = "http://localhost:11434/api/chat",
  num_ctx = NULL
) {
  complete_chat <- function(chat_history) {
    # Build messages, with optional image support for the last user message
    msgs <- lapply(seq_len(nrow(chat_history)), function(i) {
      list(role = chat_history$role[i], content = chat_history$content[i])
    })

    # If images are attached, convert them to Ollama's `images` array (base64)
    add_parts <- self$parameters$.add_image_parts %||% NULL
    if (!is.null(add_parts) && length(add_parts) > 0) {
      last_i <- nrow(chat_history)
      if (last_i > 0 && identical(chat_history$role[last_i], "user")) {
        b64s <- list()
        for (p in add_parts) {
          src <- p$source %||% NULL
          if (identical(src, "b64") && !is.null(p$data)) {
            b64s[[length(b64s) + 1]] <- as.character(p$data)
          } else if (identical(src, "url") && !is.null(p$data)) {
            # Download and base64-encode
            res <- tryCatch(
              {
                rq <- httr2::request(as.character(p$data))
                rp <- httr2::req_perform(rq)
                raw <- httr2::resp_body_raw(rp)
                jsonlite::base64_enc(raw)
              },
              error = function(e) NULL
            )
            if (!is.null(res)) b64s[[length(b64s) + 1]] <- as.character(res)
          }
        }
        if (length(b64s) > 0) {
          msgs[[last_i]]$images <- unname(b64s)
        }
      }
    }

    body <- list(
      model = self$parameters$model,
      messages = msgs
    )

    # Append user-facing parameters only; skip internal helpers (prefixed with '.')
    for (name in names(self$parameters)) {
      if (!startsWith(name, ".")) body[[name]] <- self$parameters[[name]]
    }

    request <- httr2::request(self$url) |>
      httr2::req_body_json(body)

    stream_cb <- self$stream_callback %||%
      getOption("tidyprompt.stream_callback", NULL)

    request_llm_provider(
      chat_history,
      request,
      stream = self$parameters$stream,
      verbose = self$verbose,
      api_type = self$api_type,
      stream_callback = stream_cb,
      llm_provider = self
    )
  }

  if (is.null(parameters$stream)) {
    parameters$stream <- FALSE
  }
  if (!is.null(num_ctx)) {
    parameters$options$num_ctx <- num_ctx
  }

  # Extend llm_provider-class with 'set_option' functions
  class <- R6::R6Class(
    "llm_provider_ollama-class",
    inherit = `llm_provider-class`,
    public = list(
      set_option = function(name, value) {
        stopifnot(
          is.character(name),
          is.character(value) || is.logical(value) || is.numeric(value)
        )

        self$parameters$options[[name]] <- value
        return(self)
      },
      set_options = function(options) {
        stopifnot(
          is.list(options),
          !is.null(names(options))
        )

        for (name in names(options)) {
          value <- options[[name]]
          stopifnot(
            is.character(value) || is.logical(value) || is.numeric(value)
          )

          self$parameters$options[[name]] <- value
        }
        return(self)
      }
    )
  )

  ollama <- class$new(
    complete_chat_function = complete_chat,
    parameters = parameters,
    verbose = verbose,
    url = url,
    api_type = "ollama"
  )

  return(ollama)
}

#' Create a new OpenAI LLM provider
#'
#' This function creates a new [llm_provider-class] object that interacts with the Open AI API.
#' Supports both the Chat Completions API (/v1/chat/completions) and the Responses API (/v1/responses).
#'
#' @param parameters A named list of parameters. Currently the following parameters are required:
#'    - model: The name of the model to use
#'    - api_key: The API key to use for authentication with the OpenAI API. This
#'    should be a project API key (not a user API key)
#'    - url: The URL to the OpenAI API (may also be an alternative endpoint
#'    that provides a similar API.)
#'    - stream: A logical indicating whether the API should stream responses
#'
#'  Additional parameters are appended to the request body; see the OpenAI API
#'  documentation for more information: https://platform.openai.com/docs/api-reference/chat
#' @param verbose A logical indicating whether the interaction with the LLM provider
#' should be printed to the console. Default is TRUE.
#' @param url The URL to the OpenAI API endpoint for chat completion
#' (typically: "https://api.openai.com/v1/chat/completions"
#' or "https://api.openai.com/v1/responses"; both the Chat Completions API and
#' the Responses API are supported; the Responses API is more modern)
#' @param api_key The API key to use for authentication with the OpenAI API
#'
#' @return A new [llm_provider-class] object for use of the OpenAI API
#'
#' @export
#' @example inst/examples/llm_providers.R
#'
#' @family llm_provider
llm_provider_openai <- function(
  parameters = list(
    model = "gpt-4o-mini",
    stream = getOption("tidyprompt.stream", TRUE)
  ),
  verbose = getOption("tidyprompt.verbose", TRUE),
  url = "https://api.openai.com/v1/responses",
  api_key = Sys.getenv("OPENAI_API_KEY")
) {
  complete_chat <- function(chat_history) {
    headers <- c(
      "Content-Type" = "application/json",
      "Authorization" = paste("Bearer", self$api_key)
    )

    # Build messages; support multimodal content for last user message
    add_parts <- self$parameters$.add_image_parts %||% NULL
    n <- nrow(chat_history)
    messages <- lapply(seq_len(n), function(i) {
      role <- chat_history$role[i]
      text <- chat_history$content[i]
      # For the last user message, if images are attached, convert to content parts
      if (!is.null(add_parts) && i == n && identical(role, "user")) {
        parts <- list()
        if (!is.null(text) && nzchar(text)) {
          parts[[length(parts) + 1]] <- list(type = "text", text = text)
        }
        for (p in add_parts) {
          src <- p$source %||% NULL
          if (identical(src, "url") && !is.null(p$data)) {
            parts[[length(parts) + 1]] <- list(
              type = "image_url",
              image_url = compact_list(list(
                url = as.character(p$data),
                detail = p$detail %||% NULL
              ))
            )
          } else if (identical(src, "b64") && !is.null(p$data)) {
            mime <- p$mime %||% "image/png"
            data_url <- paste0("data:", mime, ";base64,", as.character(p$data))
            parts[[length(parts) + 1]] <- list(
              type = "image_url",
              image_url = compact_list(list(
                url = data_url,
                detail = p$detail %||% NULL
              ))
            )
          }
        }
        return(list(role = role, content = parts))
      }
      list(role = role, content = text)
    })

    # Decide between Chat Completions style (messages) and Responses API (input)
    # Support both legacy Chat Completions (messages) and modern Responses API (input)
    use_responses_api <- grepl("/responses$", self$url)
    if (use_responses_api) {
      # Transform messages -> input with part types changed to input_*/output_* variants.
      # For image parts, OpenAI Responses API expects a scalar URL string in `image_url`
      # and (optionally) a separate `detail` field, not an object. The previous implementation
      # forwarded the object `{ url = ..., detail = ... }` which triggers a 400 error:
      # "Invalid type for 'input[0].content[1].image_url': expected an image URL, but got an object instead.".
      input <- lapply(messages, function(msg) {
        role <- msg$role
        content <- msg$content
        parts <- list()
        text_type <- if (identical(role, "assistant")) {
          "output_text"
        } else {
          "input_text"
        }
        image_type <- if (identical(role, "assistant")) {
          "output_image"
        } else {
          "input_image"
        }
        if (is.character(content)) {
          if (nzchar(content)) {
            parts[[length(parts) + 1]] <- list(type = text_type, text = content)
          }
        } else if (is.list(content)) {
          for (part in content) {
            if (identical(part$type, "text")) {
              if (!is.null(part$text) && nzchar(part$text)) {
                parts[[length(parts) + 1]] <- list(
                  type = text_type,
                  text = part$text
                )
              }
            } else if (identical(part$type, "image_url")) {
              img <- part$image_url
              if (is.list(img)) {
                url_val <- img$url %||% NULL
                detail_val <- img$detail %||% NULL
                if (!is.null(url_val)) {
                  parts[[length(parts) + 1]] <- compact_list(list(
                    type = image_type,
                    image_url = as.character(url_val),
                    detail = detail_val
                  ))
                }
              } else if (is.character(img) && length(img) == 1) {
                parts[[length(parts) + 1]] <- list(
                  type = image_type,
                  image_url = img
                )
              }
            } else {
              # Fallback: keep as input_text if unknown
              if (!is.null(part$text)) {
                parts[[length(parts) + 1]] <- list(
                  type = text_type,
                  text = part$text
                )
              }
            }
          }
        }
        list(role = role, content = parts)
      })
      body <- list(input = input)
    } else {
      body <- list(messages = messages)
    }

    # Append user-facing parameters only; skip internal helpers (prefixed with '.')
    # This prevents sending fields like '.add_image_parts' which cause 400 errors
    # from OpenAI ("Unknown parameter"). Internal parameters are consumed above.
    for (name in names(self$parameters)) {
      if (!startsWith(name, ".")) {
        body[[name]] <- self$parameters[[name]]
      }
    }

    request <- httr2::request(self$url) |>
      httr2::req_body_json(body) |>
      httr2::req_headers(!!!headers)

    stream_cb <- self$stream_callback %||%
      getOption("tidyprompt.stream_callback", NULL)

    request_llm_provider(
      chat_history,
      request,
      stream = self$parameters$stream,
      verbose = self$verbose,
      api_type = self$api_type,
      stream_callback = stream_cb,
      llm_provider = self
    )
  }

  return(
    `llm_provider-class`$new(
      complete_chat_function = complete_chat,
      parameters = parameters,
      verbose = verbose,
      url = url,
      api_key = api_key,
      api_type = "openai"
    )
  )
}

#' Create a new OpenRouter LLM provider
#'
#' This function creates a new [llm_provider-class] object that interacts with the OpenRouter API.
#'
#' @param parameters A named list of parameters. Currently the following parameters are required:
#'    - model: The name of the model to use
#'    - stream: A logical indicating whether the API should stream responses
#'
#'  Additional parameters are appended to the request body; see the OpenRouter API
#'  documentation for more information: https://openrouter.ai/docs/parameters
#' @param verbose A logical indicating whether the interaction with the LLM provider
#' should be printed to the console.
#' @param url The URL to the OpenRouter API endpoint for chat completion
#' @param api_key The API key to use for authentication with the OpenRouter API
#'
#' @return A new [llm_provider-class] object for use of the OpenRouter API
#'
#' @export
#' @example inst/examples/llm_providers.R
#'
#' @family llm_provider
llm_provider_openrouter <- function(
  parameters = list(
    model = "qwen/qwen-2.5-7b-instruct",
    stream = getOption("tidyprompt.stream", TRUE)
  ),
  verbose = getOption("tidyprompt.verbose", TRUE),
  url = "https://openrouter.ai/api/v1/chat/completions",
  api_key = Sys.getenv("OPENROUTER_API_KEY")
) {
  llm_provider_openai(parameters, verbose, url, api_key)
}

#' Create a new Mistral LLM provider
#'
#' This function creates a new [llm_provider-class] object that interacts with the Mistral API.
#'
#' @param parameters A named list of parameters. Currently the following parameters are required:
#'    - model: The name of the model to use
#'    - stream: A logical indicating whether the API should stream responses
#'
#'  Additional parameters are appended to the request body; see the Mistral API
#'  documentation for more information: https://docs.mistral.ai/api/#tag/chat
#' @param verbose A logical indicating whether the interaction with the LLM provider
#' should be printed to the consol
#' @param url The URL to the Mistral API endpoint for chat completion
#' @param api_key The API key to use for authentication with the Mistral API
#'
#' @return A new [llm_provider-class] object for use of the Mistral API
#'
#' @export
#' @example inst/examples/llm_providers.R
#'
#' @family llm_provider
llm_provider_mistral <- function(
  parameters = list(
    model = "ministral-3b-latest",
    stream = getOption("tidyprompt.stream", TRUE)
  ),
  verbose = getOption("tidyprompt.verbose", TRUE),
  url = "https://api.mistral.ai/v1/chat/completions",
  api_key = Sys.getenv("MISTRAL_API_KEY")
) {
  llm_provider_openai(parameters, verbose, url, api_key)
}

#' Create a new Groq LLM provider
#'
#' This function creates a new [llm_provider-class] object that interacts with the Groq API.
#'
#' @param parameters A named list of parameters. Currently the following parameters are required:
#'   - model: The name of the model to use
#'   - stream: A logical indicating whether the API should stream responses
#'
#'  Additional parameters are appended to the request body; see the Groq API
#'  documentation for more information: https://console.groq.com/docs/api-reference#chat-create
#' @param verbose A logical indicating whether the interaction with the LLM provider
#' should be printed to the console
#' @param api_key The API key to use for authentication with the Groq API
#' @param url The URL to the Groq API endpoint for chat completion
#'
#' @return A new [llm_provider-class] object for use of the Groq API
#'
#' @export
#' @example inst/examples/llm_providers.R
#'
#' @family llm_provider
llm_provider_groq <- function(
  parameters = list(
    model = "llama-3.1-8b-instant",
    stream = TRUE
  ),
  verbose = getOption("tidyprompt.verbose", TRUE),
  url = "https://api.groq.com/openai/v1/chat/completions",
  api_key = Sys.getenv("GROQ_API_KEY")
) {
  llm_provider_openai(parameters, verbose, url, api_key)
}

#' Create a new XAI (Grok) LLM provider
#'
#' This function creates a new [llm_provider-class] object that interacts with the XAI API.
#'
#' @param parameters A named list of parameters. Currently the following parameters are required:
#'   - model: The name of the model to use
#'   - stream: A logical indicating whether the API should stream responses
#'
#'  Additional parameters are appended to the request body; see the XAI API
#'  documentation for more information: https://docs.x.ai/api/endpoints#chat-completions
#' @param verbose A logical indicating whether the interaction with the LLM provider
#' should be printed to the console. Default is TRUE.
#' @param url The URL to the XAI API endpoint for chat completion
#' @param api_key The API key to use for authentication with the XAI API
#'
#' @return A new [llm_provider-class] object for use of the XAI API
#'
#' @export
#' @example inst/examples/llm_providers.R
#'
#' @family llm_provider
llm_provider_xai <- function(
  parameters = list(
    model = "grok-beta",
    stream = getOption("tidyprompt.stream", TRUE)
  ),
  verbose = getOption("tidyprompt.verbose", TRUE),
  url = "https://api.x.ai/v1/chat/completions",
  api_key = Sys.getenv("XAI_API_KEY")
) {
  llm_provider_openai(parameters, verbose, url, api_key)
}

#' Create a new Google Gemini LLM provider
#'
#' @description
#' `r lifecycle::badge("superseded")`
#' This function creates a new [llm_provider-class] object that interacts with the Google Gemini API.
#'
#' @details
#' Streaming is not yet supported in this implementation. Native functions
#' like structured output and tool calling are also not supported in this implemetation.
#' This may however be achieved through creating a [llm_provider_ellmer()] object
#' with as input a `ellmer::chat_google_gemini()` object. Therefore, this function
#' is now superseded by `llm_provider_ellmer(ellmer::chat_google_gemini())`.
#'
#' @param parameters A named list of parameters. Currently the following parameters are required:
#'    - model: The name of the model to use (see: https://ai.google.dev/gemini-api/docs/models/gemini)
#'
#'  Additional parameters are appended to the request body; see the Google AI Studio API
#'  documentation for more information: https://ai.google.dev/gemini-api/docs/text-generation
#'  and https://github.com/google/generative-ai-docs/blob/main/site/en/gemini-api/docs/get-started/rest.ipynb
#' @param verbose A logical indicating whether the interaction with the LLM provider
#' should be printed to the console
#' @param url The URL to the Google Gemini API endpoint for chat completion
#' @param api_key The API key to use for authentication with the Google Gemini API
#' (see: https://aistudio.google.com/app/apikey)
#'
#' @return A new [llm_provider-class] object for use of the Google Gemini API
#'
#' @export
#'
#' @example inst/examples/llm_providers.R
#'
#' @family llm_provider
llm_provider_google_gemini <- function(
  parameters = list(
    model = "gemini-1.5-flash"
  ),
  verbose = getOption("tidyprompt.verbose", TRUE),
  url = "https://generativelanguage.googleapis.com/v1beta/models/",
  api_key = Sys.getenv("GOOGLE_AI_STUDIO_API_KEY")
) {
  complete_chat <- function(chat_history) {
    # Construct URL for the API request
    endpoint <- paste0(
      self$url,
      self$parameters$model,
      ":generateContent"
    )

    # Format chat_history for API compatibility with the 'contents' format
    formatted_contents <- lapply(seq_len(nrow(chat_history)), function(i) {
      list(
        role = ifelse(
          chat_history$role[i] == "assistant",
          "model",
          chat_history$role[i]
        ),
        parts = list(list(text = chat_history$content[i]))
      )
    })

    # Build the request body with 'contents' field
    body <- list(
      contents = formatted_contents
    )

    # Append all other parameters to the body
    for (name in names(self$parameters)) {
      body[[name]] <- self$parameters[[name]]
    }

    # Send the POST request with httr2
    request <- httr2::request(endpoint) |>
      httr2::req_headers(
        `Content-Type` = "application/json"
      ) |>
      httr2::req_body_json(body) |>
      httr2::req_url_query(key = self$api_key)
    response <- request |> httr2::req_perform()

    # Check if the request was successful
    if (httr2::resp_status(response) == 200) {
      content <- httr2::resp_body_json(response)

      completed <- chat_history |>
        dplyr::bind_rows(
          data.frame(
            role = "assistant",
            content = content$candidates[[1]]$content$parts[[1]]$text
          )
        )

      return(
        list(
          completed = completed,
          http = list(
            request = request,
            response = response
          )
        )
      )
    } else {
      stop(
        "Error: ",
        httr2::resp_status(response),
        " - ",
        httr2::resp_body_string(response)
      )
    }
  }

  `llm_provider-class`$new(
    complete_chat_function = complete_chat,
    parameters = parameters,
    verbose = verbose,
    url = url,
    api_key = api_key,
    api_type = "gemini"
  )
}

#' Create a fake [llm_provider-class] (for development and testing purposes)
#'
#' This function creates a fake [llm_provider-class] that can be used for development
#' and testing purposes. It is hardcoded to send back specific responses to
#' specific prompts that are used in vignettes, tests, and examples.
#' This is useful for running tests and builds in environments in which an
#' actual [llm_provider-class] is not available.
#'
#' @param verbose A logical indicating whether the interaction with the [llm_provider-class]
#' should be printed to the console. Default is TRUE.
#'
#' @return A new [llm_provider-class] object for use of the fake LLM provider
#'
#' @noRd
#' @keywords internal
llm_provider_fake <- function(verbose = getOption("tidyprompt.verbose", TRUE)) {
  complete_chat <- function(chat_history) {
    last_msg <- utils::tail(chat_history$content, 1)

    answer_as_integer_input <-
      "You must answer with only an integer (use no other characters)."

    chain_of_thought_input <-
      "To answer the user's prompt, you need to think step by step to arrive at a final answer."

    if (last_msg == "Hi there!") {
      return(
        list(
          completed = chat_history |>
            dplyr::bind_rows(
              data.frame(
                role = "assistant",
                content = paste0(
                  "It's nice to meet you.",
                  " Is there something I can help you with or would you like to chat?"
                )
              )
            )
        )
      )
    }

    if (
      grepl(
        "What is a large language model? Explain in 10 words.",
        last_msg,
        fixed = TRUE
      )
    ) {
      return(
        list(
          completed = chat_history |>
            dplyr::bind_rows(
              data.frame(
                role = "assistant",
                content = "Complex computer program trained on vast texts to generate human-like responses."
              )
            )
        )
      )
    }

    if (
      grepl("What is 2 + 2?", last_msg, fixed = TRUE) &&
        grepl(answer_as_integer_input, last_msg, fixed = TRUE) &&
        !grepl(chain_of_thought_input, last_msg, fixed = TRUE)
    ) {
      return(
        list(
          completed = chat_history |>
            dplyr::bind_rows(
              data.frame(
                role = "assistant",
                content = "4"
              )
            )
        )
      )
    }

    if (
      grepl("What is 2 + 2?", last_msg, fixed = TRUE) &&
        !grepl(answer_as_integer_input, last_msg, fixed = TRUE)
    ) {
      return(
        list(
          completed = chat_history |>
            dplyr::bind_rows(
              data.frame(
                role = "assistant",
                content = glue::glue(
                  ">> step 1: Identify the mathematical operation in the prompt,
          which is a simple addition problem.

          >> step 2: Recall the basic arithmetic fact that 2 + 2 equals a specific
          numerical value.

          >> step 3: Apply this knowledge to determine the result of the addition problem,
          using the known facts about numbers and their operations.

          >> step 4: Conclude that based on this mathematical understanding, the
          solution to the prompt \"What is 2 + 2?\" is a fixed numerical quantity."
                )
              )
            )
        )
      )
    }

    if (
      any(
        grepl(
          "What is 2 + 2?",
          chat_history$content[chat_history$role == "user"],
          fixed = TRUE
        )
      ) &&
        grepl(answer_as_integer_input, last_msg, fixed = TRUE) &&
        !grepl(chain_of_thought_input, last_msg, fixed = TRUE)
    ) {
      return(
        list(
          completed = chat_history |>
            dplyr::bind_rows(
              data.frame(
                role = "assistant",
                content = "22"
              )
            )
        )
      )
    }

    if (
      grepl("What is 2 + 2?", last_msg, fixed = TRUE) &&
        grepl(chain_of_thought_input, last_msg, fixed = TRUE) &&
        grepl(answer_as_integer_input, last_msg, fixed = TRUE)
    ) {
      return(
        list(
          completed = chat_history |>
            dplyr::bind_rows(
              data.frame(
                role = "assistant",
                content = glue::glue(
                  ">> step 1: Identify the mathematical operation in the prompt,
          which is a simple addition problem.

          >> step 2: Recall the basic arithmetic fact that 2 + 2 equals a specific
          numerical value.

          >> step 3: Apply this knowledge to determine the result of the addition problem,
          using the known facts about numbers and their operations.

          >> step 4: Conclude that based on this mathematical understanding, the
          solution to the prompt \"What is 2 + 2?\" is a fixed numerical quantity.

          FINISH[4]"
                )
              )
            )
        )
      )
    }

    if (
      grepl(
        'example usage: FUNCTION[temperature_in_location]("Amsterdam", "Fahrenheit")',
        last_msg,
        fixed = TRUE
      )
    ) {
      return(
        list(
          completed = chat_history |>
            dplyr::bind_rows(
              data.frame(
                role = "assistant",
                content = glue::glue(
                  "I'll use the provided function to get the current temperature in Enschede.

          FUNCTION[temperature_in_location](\"Enschede\", \"Celcius\")"
                )
              )
            )
        )
      )
    }

    if (
      grepl(
        "function called: temperature_in_location",
        last_msg,
        fixed = TRUE
      ) &&
        grepl("arguments used: location = Enschede", last_msg, fixed = TRUE)
    ) {
      return(
        list(
          completed = chat_history |>
            dplyr::bind_rows(
              data.frame(
                role = "assistant",
                content = "22.7"
              )
            )
        )
      )
    }

    if (
      any(
        grepl(
          "So the current temperature in Enschede is 22.7 degrees Celsius.",
          chat_history$content[chat_history$role == "assistant"],
          fixed = TRUE
        )
      ) &&
        grepl(last_msg, answer_as_integer_input, fixed = TRUE)
    ) {
      return(
        list(
          completed = chat_history |>
            dplyr::bind_rows(
              data.frame(
                role = "assistant",
                content = "22.7"
              )
            )
        )
      )
    }

    return(
      list(
        completed = chat_history |>
          dplyr::bind_rows(
            data.frame(
              role = "assistant",
              content = "I'm a fake LLM! This is my default response."
            )
          )
      )
    )
  }

  `llm_provider-class`$new(
    complete_chat_function = complete_chat,
    verbose = verbose,
    parameters = list(
      model = 'llama3.1:8b'
    ),
    api_type = "fake"
  )
}

#' Create a new LLM provider from an `ellmer::chat()` object
#'
#' @description
#' `r lifecycle::badge("experimental")`
#'
#' This function creates a [llm_provider-class] from an `ellmer::chat()` object.
#' This allows the user to use the various LLM providers which are supported
#' by the 'ellmer' R package, including respective configuration and features.
#'
#' Please note that this function is experimental. This provider type may show different behavior than
#' other LLM providers, and may not function optimally.
#'
#' @details
#' **ellmer version compatibility:** tidyprompt requires ellmer >= 0.3.0.
#' Some features are only fully available with ellmer >= 0.4.0; on older
#' versions they degrade gracefully:
#' \itemize{
#'   \item Built-in tools (`ToolBuiltIn`, e.g. `claude_tool_web_search()`)
#'     require ellmer >= 0.4.0.
#'   \item `type_ignore()` for ignored tool arguments requires ellmer >= 0.4.0;
#'     on older versions, ignored arguments fall back to optional string
#'     parameters.
#' }
#'
#' Unlike other LLM provider classes,
#' most LLM provider settings need to be managed in the `ellmer::chat()` object
#' (and not in the `$parameters` list). `$get_chat()` and `$set_chat()` may be used
#' to manipulate the chat object. There are however some parameters that can be set
#' in the `$parameters` list; these are documented below.
#'
#' 1) Streaming can be controlled through via the `$parameters$stream` parameter.
#' If set to TRUE (default), streaming will be used if supported by the underlying
#' `ellmer::chat()` object. If the underlying `ellmer::chat()` object does not support streaming,
#' you may need to set this parameter to FALSE to avoid errors.
#'
#' 2) A special parameter `$.ellmer_structured_type` may also be set in the `$parameters` list;
#' this parameter is used to specify a structured output format. This should be a 'ellmer'
#' structured type (e.g., `ellmer::type_object`; see https://ellmer.tidyverse.org/articles/structured-data.html).
#' `answer_as_json()` sets this parameter to obtain structured output
#' (it is not recommended to set this parameter manually, but it is possible).
#'
#' @param chat An `ellmer::chat()` object (e.g., `ellmer::chat_openai()`)
#' @param parameters A named list of parameters. See 'details' for supported parameters
#' @param verbose A logical indicating whether the interaction with the [llm_provider-class]
#' should be printed to the console. Default is TRUE
#'
#' @return An [llm_provider-class] with api_type = "ellmer"
#'
#' @export
#'
#' @example inst/examples/llm_providers.R
#'
#' @family llm_provider
llm_provider_ellmer <- function(
  chat,
  parameters = list(
    stream = getOption("tidyprompt.stream", TRUE)
  ),
  verbose = getOption("tidyprompt.verbose", TRUE)
) {
  if (missing(chat) || is.null(chat)) {
    stop("`chat` must be an ellmer chat object (e.g., ellmer::chat_openai()).")
  }
  if (!is_ellmer_chat(chat)) {
    stop(
      "`chat` doesn't look like an ellmer chat object (no `$chat()` method)."
    )
  }

  if (isTRUE(parameters$stream) && !requireNamespace("coro", quietly = TRUE)) {
    stop(paste0(
      "The 'coro' package is required for streaming with `llm_provider_ellmer()`.\n",
      "Please install it first (e.g., `install.packages('coro')`), ",
      "or set `parameters$stream = FALSE` to disable streaming"
    ))
  }

  complete_chat <- function(chat_history) {
    private$sync_model()
    ch <- self$ellmer_chat
    params <- self$parameters

    # Known parameter names that are handled by the ellmer provider path.
    # Anything else is silently ignored because ellmer Chat objects carry their
    # API params (temperature, max_tokens, etc.) from construction and cannot
    # be modified afterwards.
    ellmer_known_params <- c(
      "model",
      "stream",
      ".ellmer_tools",
      ".ellmer_structured_type",
      ".add_image_parts",
      ".reset_ellmer_chat"
    )

    # Warn once per session about parameters that cannot be forwarded.
    extra <- setdiff(names(params), ellmer_known_params)
    extra <- extra[!startsWith(extra, ".")]
    if (length(extra)) {
      rlang::warn(
        c(
          paste0(
            "Parameters not forwarded to ellmer provider: ",
            paste0("`", extra, "`", collapse = ", "),
            "."
          ),
          i = "Set these when creating the ellmer chat, e.g., `chat_openai(params = list(...))`."
        ),
        .frequency = "once",
        .frequency_id = "tidyprompt_ellmer_extra_params"
      )
    }

    if (!all(c("role", "content") %in% names(chat_history))) {
      stop("`chat_history` must have columns `role` and `content`.")
    }

    # Seed prior turns
    hist <- if (nrow(chat_history) > 1) {
      chat_history[seq_len(nrow(chat_history) - 1), , drop = FALSE]
    } else {
      chat_history[0, , drop = FALSE]
    }

    ellmer_available <- requireNamespace("ellmer", quietly = TRUE)
    ellmer_ns <- if (ellmer_available) asNamespace("ellmer") else NULL

    as_content_text <- function(text) {
      if (is.null(text) || !nzchar(text)) {
        return(NULL)
      }
      if (
        !is.null(ellmer_ns) &&
          exists("ContentText", envir = ellmer_ns, inherits = FALSE)
      ) {
        return(ellmer::ContentText(text))
      }
      list(type = "text", text = text)
    }

    as_turn <- function(role, contents) {
      # ellmer only supports user/assistant/system roles; map "tool" rows
      # (created by openai/ollama tool handlers) to user turns so replayed
      # histories don't crash.
      if (identical(role, "tool")) {
        role <- "user"
      }

      if (!is.null(ellmer_ns)) {
        # Prefer role-specific constructors (ellmer >= 0.4.0)
        ctor <- switch(
          role,
          user = if (exists("UserTurn", envir = ellmer_ns, inherits = FALSE)) {
            ellmer_ns$UserTurn
          },
          assistant = if (
            exists("AssistantTurn", envir = ellmer_ns, inherits = FALSE)
          ) {
            ellmer_ns$AssistantTurn
          },
          system = if (
            exists("SystemTurn", envir = ellmer_ns, inherits = FALSE)
          ) {
            ellmer_ns$SystemTurn
          },
          NULL
        )
        if (!is.null(ctor)) {
          return(ctor(contents = contents))
        }
        # Fall back to Turn(role=...) for ellmer < 0.4.0
        if (exists("Turn", envir = ellmer_ns, inherits = FALSE)) {
          return(ellmer::Turn(role = role, contents = contents))
        }
      }
      list(role = role, contents = contents)
    }

    ellmer_object_props <- function(x) {
      if (is.null(x)) {
        return(list())
      }

      props <- tryCatch(
        S7::props(x),
        error = function(e) NULL
      )
      if (is.list(props) && length(props)) {
        return(props)
      }

      attributes(x) %||% list()
    }

    append_native_history <- function(
      history,
      role,
      content,
      tool_call_id = NA_character_,
      tool_call = FALSE,
      tool_result = FALSE,
      hidden_from_llm = FALSE,
      native_turn_id = NA_character_,
      native_turn_role = NA_character_,
      native_contents = NULL
    ) {
      normalize_native_contents <- function(x) {
        if (is.null(x)) {
          return(list())
        }
        if (is_ellmer_content_object(x)) {
          return(list(x))
        }
        if (!is.list(x)) {
          return(list())
        }

        Filter(Negate(is.null), x)
      }

      row <- data.frame(
        role = as.character(role %||% "assistant"),
        content = as.character(content %||% ""),
        tool_call = tool_call,
        tool_call_id = ifelse(is.null(tool_call_id), NA, tool_call_id),
        tool_result = tool_result,
        hidden_from_llm = hidden_from_llm,
        native_turn_id = as.character(native_turn_id %||% NA_character_),
        native_turn_role = as.character(native_turn_role %||% NA_character_),
        stringsAsFactors = FALSE
      )
      row$native_contents <- list(normalize_native_contents(native_contents))

      dplyr::bind_rows(
        history,
        row
      )
    }

    set_native_turn_metadata <- function(
      history,
      row,
      native_turn_id = NA_character_,
      native_turn_role = NA_character_,
      native_contents = NULL
    ) {
      if (!nrow(history) || row < 1L || row > nrow(history)) {
        return(history)
      }

      if (!"native_turn_id" %in% names(history)) {
        history$native_turn_id <- rep(NA_character_, nrow(history))
      }
      if (!"native_turn_role" %in% names(history)) {
        history$native_turn_role <- rep(NA_character_, nrow(history))
      }
      if (!"native_contents" %in% names(history)) {
        history$native_contents <- vector("list", nrow(history))
      }

      history$native_turn_id[row] <- as.character(
        native_turn_id %||% NA_character_
      )
      history$native_turn_role[row] <- as.character(
        native_turn_role %||% NA_character_
      )
      history$native_contents[[row]] <- if (is.null(native_contents)) {
        list()
      } else if (is_ellmer_content_object(native_contents)) {
        list(native_contents)
      } else if (is.list(native_contents)) {
        Filter(Negate(is.null), native_contents)
      } else {
        list()
      }

      history
    }

    is_ellmer_content_object <- function(x) {
      classes <- class(x) %||% character()
      any(grepl("^ellmer::Content", classes))
    }

    serialize_native_content <- function(x) {
      if (is.null(x)) {
        return("")
      }

      props <- ellmer_object_props(x)
      classes <- class(x) %||% character()
      content_class <- classes[grepl("^ellmer::Content", classes)][1] %||%
        classes[1] %||%
        "Content"

      if (any(grepl("ContentToolRequest|ContentToolResult", classes))) {
        return("")
      }

      if (
        any(grepl("ContentText", classes)) ||
          !is.null(props$text)
      ) {
        return(as.character(props$text %||% ""))
      }

      if (
        any(grepl("ContentThinking", classes)) ||
          !is.null(props$thinking)
      ) {
        return(as.character(props$thinking %||% ""))
      }

      if (any(grepl("ContentJson", classes))) {
        json_text <- props$string %||% NULL
        if (!is.null(json_text) && nzchar(as.character(json_text))) {
          return(as.character(json_text))
        }

        json_text <- tryCatch(
          jsonlite::toJSON(
            props$parsed %||% props$data %||% NULL,
            auto_unbox = TRUE
          ),
          error = function(e) NULL
        )
        if (!is.null(json_text)) {
          return(as.character(json_text))
        }
      }

      if (
        any(grepl("ContentImageRemote", classes)) ||
          !is.null(props$url)
      ) {
        return(paste0("[image: ", as.character(props$url %||% "remote"), "]"))
      }

      if (any(grepl("ContentImageInline", classes))) {
        return(paste0(
          "[image: inline ",
          as.character(props$type %||% "unknown"),
          "]"
        ))
      }

      if (any(grepl("ContentPDF", classes))) {
        return(paste0(
          "[pdf: ",
          as.character(props$filename %||% "document.pdf"),
          "]"
        ))
      }

      fallback <- tryCatch(
        jsonlite::toJSON(props, auto_unbox = TRUE),
        error = function(e) NULL
      )

      if (!is.null(fallback)) {
        return(paste0("[", content_class, "] ", as.character(fallback)))
      }

      paste0("[", content_class, "]")
    }

    stringify_tool_result <- function(x) {
      if (is.null(x)) {
        return("")
      }
      if (is_ellmer_content_object(x)) {
        return(serialize_native_content(x))
      }
      if (is.character(x) && length(x) == 1) {
        return(as.character(x))
      }
      if (is.atomic(x)) {
        return(paste(as.character(x), collapse = ", "))
      }
      if (
        is.list(x) &&
          length(x) > 0 &&
          all(vapply(x, is_ellmer_content_object, logical(1)))
      ) {
        return(paste(
          vapply(x, serialize_native_content, character(1)),
          collapse = "\n"
        ))
      }

      json <- tryCatch(
        jsonlite::toJSON(x, auto_unbox = TRUE),
        error = function(e) NULL
      )
      if (!is.null(json)) {
        return(as.character(json))
      }

      paste(as.character(x), collapse = ", ")
    }

    native_turns_to_history <- function(turns, start_index = 1L) {
      history <- data.frame(
        role = character(),
        content = character(),
        tool_call = logical(),
        tool_call_id = character(),
        tool_result = logical(),
        stringsAsFactors = FALSE
      )
      history$native_turn_id <- character()
      history$native_turn_role <- character()
      history$native_contents <- list()

      for (turn_i in seq_along(turns)) {
        turn <- turns[[turn_i]]
        turn_props <- ellmer_object_props(turn)
        role <- turn_props$role %||% "assistant"
        contents <- turn_props$contents %||% list()
        turn_id <- paste0("ellmer-turn-", start_index + turn_i - 1L)

        for (content_obj in contents) {
          props <- ellmer_object_props(content_obj)
          classes <- class(content_obj) %||% character()

          if (
            any(grepl("ContentToolRequest", classes)) ||
              (!is.null(props$id) &&
                !is.null(props$name) &&
                !is.null(props$arguments))
          ) {
            pretty_args <- tryCatch(
              jsonlite::toJSON(
                props$arguments %||% list(),
                auto_unbox = FALSE,
                pretty = TRUE
              ),
              error = function(e) "null"
            )
            history <- append_native_history(
              history,
              role = "assistant",
              content = paste0(
                "~>> Calling function '",
                as.character(props$name %||% "tool"),
                "' with arguments:\n",
                pretty_args
              ),
              tool_call_id = as.character(props$id %||% NA_character_),
              tool_call = TRUE,
              tool_result = FALSE,
              native_turn_id = turn_id,
              native_turn_role = role,
              native_contents = content_obj
            )
            next
          }

          if (
            any(grepl("ContentToolResult", classes)) ||
              (!is.null(props$request) &&
                ("value" %in% names(props) || "error" %in% names(props)))
          ) {
            request_props <- ellmer_object_props(props$request %||% NULL)
            result_text <- if (!is.null(props$error)) {
              paste0("Error: ", stringify_tool_result(props$error))
            } else {
              stringify_tool_result(props$value)
            }

            history <- append_native_history(
              history,
              role = "tool",
              content = paste0("~>> Result:\n", result_text),
              tool_call_id = as.character(request_props$id %||% NA_character_),
              tool_call = FALSE,
              tool_result = TRUE,
              native_turn_id = turn_id,
              native_turn_role = role,
              native_contents = content_obj
            )
            next
          }

          if (
            any(grepl("ContentThinking", classes)) ||
              !is.null(props$thinking)
          ) {
            thinking <- as.character(props$thinking %||% "")
            if (nzchar(thinking)) {
              history <- append_native_history(
                history,
                role,
                thinking,
                hidden_from_llm = TRUE,
                native_turn_id = turn_id,
                native_turn_role = role,
                native_contents = content_obj
              )
            }
            next
          }

          text <- serialize_native_content(content_obj)
          if (nzchar(text)) {
            history <- append_native_history(
              history,
              role,
              text,
              native_turn_id = turn_id,
              native_turn_role = role,
              native_contents = content_obj
            )
          }
        }
      }

      if (!nrow(history)) {
        return(NULL)
      }

      history
    }

    build_image_content <- function(p) {
      # URL-based image
      if (
        identical(p$source, "url") &&
          !is.null(p$data) &&
          !is.null(ellmer_ns) &&
          isTRUE(exists("content_image_url", ellmer_ns, inherits = FALSE))
      ) {
        return(ellmer::content_image_url(
          url = as.character(p$data),
          detail = p$detail %||% "auto"
        ))
      }

      # Base64-encoded image -> temp file + content_image_file, when available
      if (
        identical(p$source, "b64") &&
          !is.null(p$data) &&
          !is.null(ellmer_ns) &&
          isTRUE(exists("content_image_file", ellmer_ns, inherits = FALSE))
      ) {
        raw_bytes <- tryCatch(
          jsonlite::base64_dec(as.character(p$data)),
          error = function(e) raw()
        )
        if (length(raw_bytes)) {
          ext <- ".png"
          if (!is.null(p$mime) && grepl("jpeg", p$mime)) {
            ext <- ".jpg"
          }
          if (!is.null(p$mime) && grepl("gif", p$mime)) {
            ext <- ".gif"
          }
          if (!is.null(p$mime) && grepl("webp", p$mime)) {
            ext <- ".webp"
          }
          tf <- tempfile(fileext = ext)
          writeBin(raw_bytes, tf)
          on.exit(unlink(tf), add = TRUE)
          img <- tryCatch(
            ellmer::content_image_file(
              path = tf,
              content_type = p$mime %||% "auto"
            ),
            error = function(e) {
              b64 <- jsonlite::base64_enc(raw_bytes)
              mime <- p$mime %||% "image/png"

              # If ellmer can't build from a temp file (for example because
              # an image helper dependency is unavailable), fall back to an
              # inline ellmer content object instead of an OpenAI-style payload.
              if (
                exists(
                  "ContentImageInline",
                  envir = ellmer_ns,
                  inherits = FALSE
                )
              ) {
                return(ellmer_ns$ContentImageInline(
                  type = mime,
                  data = b64
                ))
              }

              if (
                exists("content_image_url", envir = ellmer_ns, inherits = FALSE)
              ) {
                return(ellmer::content_image_url(
                  url = paste0("data:", mime, ";base64,", b64),
                  detail = p$detail %||% "auto"
                ))
              }

              stop(e)
            }
          )
          return(img)
        }
      }

      # Already-constructed ellmer content (e.g., content_image_url/file/plot)
      if (identical(p$source, "ellmer") && !is.null(p$obj)) {
        return(p$obj)
      }

      # Fallback stub when ellmer helpers are unavailable (primarily for tests)
      if (!is.null(p$source) && !is.null(p$data)) {
        return(list(
          source = p$source,
          data = p$data,
          mime = p$mime %||% NA_character_
        ))
      }

      NULL
    }

    history_native_turn_id <- function(history, i) {
      if (!"native_turn_id" %in% names(history)) {
        return(NA_character_)
      }

      as.character(history$native_turn_id[i] %||% NA_character_)
    }

    history_native_turn_role <- function(history, i) {
      if (!"native_turn_role" %in% names(history)) {
        return(NA_character_)
      }

      as.character(history$native_turn_role[i] %||% NA_character_)
    }

    history_native_contents <- function(history, i) {
      if (!"native_contents" %in% names(history)) {
        return(NULL)
      }

      contents <- history$native_contents[[i]]
      if (is.null(contents)) {
        return(NULL)
      }
      if (is_ellmer_content_object(contents)) {
        return(list(contents))
      }
      if (!is.list(contents)) {
        return(NULL)
      }

      contents <- Filter(Negate(is.null), contents)
      if (!length(contents)) {
        return(NULL)
      }

      contents
    }

    # Prepare historical turns (excluding the latest message)
    prior_turns <- list()
    if (nrow(hist)) {
      i <- 1L

      while (i <= nrow(hist)) {
        turn_id <- history_native_turn_id(hist, i)

        if (!is.na(turn_id) && nzchar(turn_id)) {
          grouped_contents <- list()
          turn_role <- history_native_turn_role(hist, i)
          j <- i

          while (
            j <= nrow(hist) &&
              identical(history_native_turn_id(hist, j), turn_id)
          ) {
            grouped_contents <- c(
              grouped_contents,
              history_native_contents(hist, j) %||% list()
            )
            j <- j + 1L
          }

          grouped_contents <- Filter(Negate(is.null), grouped_contents)
          if (length(grouped_contents)) {
            prior_turns[[length(prior_turns) + 1L]] <- as_turn(
              role = if (!is.na(turn_role) && nzchar(turn_role)) {
                turn_role
              } else {
                hist$role[i]
              },
              contents = grouped_contents
            )
            i <- j
            next
          }
        }

        ct <- as_content_text(hist$content[i])
        prior_turns[[length(prior_turns) + 1L]] <- as_turn(
          role = hist$role[i],
          contents = if (is.null(ct)) list() else list(ct)
        )
        i <- i + 1L
      }
    }

    # Prompt = last message (may be converted to multimodal contents below)
    prompt <- if (nrow(chat_history) > 0L) {
      chat_history$content[nrow(chat_history)]
    } else {
      ""
    }
    prompt <- if (length(prompt) == 1L && !is.na(prompt)) {
      as.character(prompt)
    } else {
      ""
    }
    # Merge tools: the cloned chat already carries the user's base tools
    # (registered via $register_tool() / $set_tools() on the provider's
    # ellmer_chat).  Prompt-level tools are layered on top for this request
    # only.  When there are no prompt tools we leave the registry untouched.
    #
    # base_tools is captured once (before any prompt-level merge) so that
    # retries within the same send_prompt() call don't accumulate duplicates.
    prompt_tools <- params$.ellmer_tools %||% list()
    if (length(prompt_tools)) {
      if (is.null(private$.base_tools_snapshot)) {
        private$.base_tools_snapshot <- tryCatch(
          {
            gt <- ch$get_tools
            if (is.function(gt)) gt() else list()
          },
          error = function(e) list()
        )
      }
      all_tools <- c(private$.base_tools_snapshot, prompt_tools)
      if (is.function(ch$set_tools)) {
        ch$set_tools(all_tools)
      } else {
        for (td in prompt_tools) {
          ch$register_tool(td)
        }
      }
    }

    # Check if we are doing structured output
    structured_type <- params$.ellmer_structured_type %||% NULL
    use_structured <- !is.null(structured_type) &&
      is.function(ch$chat_structured)

    # Warn if both tools and structured output are requested:
    # ellmer's chat_structured() suppresses tool use, so both cannot
    # work simultaneously in a single request.
    if (use_structured && !is.null(params$.ellmer_tools)) {
      cli::cli_alert_warning(
        paste0(
          "{.strong `llm_provider_ellmer()`}:\n",
          "* Both structured output and tool use are requested;\n",
          "ellmer's {.fn chat_structured} suppresses tool use, ",
          "so tools will not be called in this request\n",
          "* Consider using {.code type = \"text-based\"} for one of them"
        )
      )
    }

    # Central streaming hook: if provided, we will stream tokens/chunks
    # through this function instead of letting ellmer print to console.
    stream_cb <- self$stream_callback %||%
      getOption("tidyprompt.stream_callback", NULL)

    # Try to detect attached images and use ellmer multimodal helpers if available.
    # `add_image()` normalizes all inputs into `.add_image_parts`; here we
    # translate those parts into ellmer content objects.
    add_parts <- self$parameters$.add_image_parts %||% NULL

    multimodal_contents <- list()
    use_multimodal <- length(add_parts) > 0

    if (use_multimodal) {
      text_content <- as_content_text(prompt)
      if (!is.null(text_content)) {
        multimodal_contents[[length(multimodal_contents) + 1]] <- text_content
      }

      for (p in add_parts) {
        ic <- build_image_content(p)
        if (!is.null(ic)) {
          multimodal_contents[[length(multimodal_contents) + 1]] <- ic
        }
      }

      if (length(multimodal_contents) == 0L) {
        use_multimodal <- FALSE
      }
    }

    turns_to_send <- prior_turns
    prompt_for_model <- prompt

    # For multimodal, pass content objects directly to the chat method
    # instead of pre-seeding the current turn and calling with empty prompt
    # (which would create a duplicate empty user turn).
    multimodal_args <- NULL
    if (use_multimodal) {
      multimodal_args <- multimodal_contents
    }

    ch <- ch$set_turns(turns_to_send)

    # Store the native structured result (if any) so downstream extraction
    # can use it directly without lossy JSON round-tripping.
    native_structured_result <- NULL

    if (use_structured) {
      if (!is.null(multimodal_args)) {
        reply_struct <- do.call(
          ch$chat_structured,
          c(multimodal_args, list(type = structured_type))
        )
      } else {
        reply_struct <- ch$chat_structured(
          prompt_for_model,
          type = structured_type
        )
      }
      native_structured_result <- reply_struct
      # Store a JSON string in the transcript (so downstream plain JSON extractors still work)
      assistant_text <- jsonlite::toJSON(reply_struct, auto_unbox = TRUE) |>
        as.character()
    } else if (isTRUE(params$stream) && is.function(ch$stream)) {
      # --- Streaming path (ellmer-style sync streaming) ----------------------
      stream_error <- NULL
      assistant_text <- NULL

      stream <- tryCatch(
        if (!is.null(multimodal_args)) {
          do.call(ch$stream, multimodal_args)
        } else {
          ch$stream(prompt_for_model)
        },
        error = function(e) {
          stream_error <<- e
          NULL
        }
      )

      if (is.null(stream) && !is.null(stream_error)) {
        # Streaming failed to initialise; fall back to non-streaming chat.
        # (Multimodal paths already did this; now also try for plain text.)
        reply_any <- tryCatch(
          if (use_multimodal) {
            do.call(ch$chat, multimodal_args)
          } else {
            ch$chat(prompt_for_model)
          },
          error = function(e) NULL
        )
        if (!is.null(reply_any)) {
          assistant_text <- as.character(reply_any)
        } else {
          stop(stream_error)
        }
      } else {
        # Create environment to hold partial response
        partial_response_env <- new.env()
        assign("partial_response", "", envir = partial_response_env)

        coro::loop(
          for (chunk in stream) {
            if (length(chunk) == 0L || all(is.na(chunk))) {
              next
            }

            chunk_str <- paste0(as.character(chunk), collapse = "")
            if (!nzchar(chunk_str)) {
              next
            }

            current_response <- get(
              "partial_response",
              envir = partial_response_env
            )
            updated_response <- paste0(current_response, chunk_str)
            assign(
              "partial_response",
              updated_response,
              envir = partial_response_env
            )

            # If a callback is provided, use it; otherwise, mirror HTTP providers
            # by cat()ing chunks when verbose = TRUE.
            if (is.function(stream_cb)) {
              latest_message <- chat_history[nrow(chat_history), , drop = FALSE]

              meta <- list(
                llm_provider = self,
                chat_history = chat_history,
                latest_message = latest_message,
                partial_response = updated_response,
                chunk = chunk_str,
                api_type = "ellmer",
                endpoint = "chat",
                verbose = self$verbose
              )

              tryCatch(
                stream_cb(chunk_str, meta),
                error = function(e) {
                  warning(
                    "stream_callback error (streaming continues): ",
                    conditionMessage(e),
                    call. = FALSE
                  )
                }
              )
            } else if (isTRUE(self$verbose)) {
              cat(chunk_str)
            }
          }
        )

        # After streaming, use accumulated partial_response as assistant text
        assistant_text <- get("partial_response", envir = partial_response_env)
        # If it leads with '\n', strip that
        assistant_text <- sub("^\n", "", assistant_text)
        # If it ends with '\n', strip that
        assistant_text <- sub("\n$", "", assistant_text)
      }
    } else {
      # Regular, non-streaming chat
      reply_any <- if (!is.null(multimodal_args)) {
        do.call(ch$chat, multimodal_args)
      } else {
        ch$chat(prompt_for_model)
      }
      assistant_text <- as.character(reply_any)
    }

    completed <- NULL

    native_turns <- tryCatch(
      {
        gt <- ch$get_turns
        if (is.function(gt)) gt() else NULL
      },
      error = function(e) NULL
    )

    prompt_turn_index <- length(prior_turns) + 1L
    if (length(native_turns) >= prompt_turn_index && nrow(chat_history) > 0) {
      prompt_turn <- native_turns[[prompt_turn_index]]
      prompt_props <- ellmer_object_props(prompt_turn)
      chat_history <- set_native_turn_metadata(
        chat_history,
        row = nrow(chat_history),
        native_turn_id = paste0("ellmer-turn-", prompt_turn_index),
        native_turn_role = prompt_props$role %||%
          chat_history$role[nrow(chat_history)],
        native_contents = prompt_props$contents %||% list()
      )
    }

    first_new_turn <- length(prior_turns) + 2L
    if (length(native_turns) >= first_new_turn) {
      native_rows <- native_turns_to_history(
        native_turns[
          first_new_turn:length(native_turns)
        ],
        start_index = first_new_turn
      )
      if (!is.null(native_rows) && nrow(native_rows)) {
        completed <- dplyr::bind_rows(chat_history, native_rows)
      }
    }

    if (is.null(completed)) {
      completed <- dplyr::bind_rows(
        chat_history,
        data.frame(
          role = "assistant",
          content = assistant_text,
          stringsAsFactors = FALSE
        )
      )
    }

    list(
      completed = completed,
      http = list(request = NULL, response = NULL),
      ellmer_chat = ch,
      native_structured_result = native_structured_result
    )
  }

  klass <- R6::R6Class(
    "llm_provider_ellmer-class",
    inherit = `llm_provider-class`,
    public = list(
      ellmer_chat = NULL,

      get_chat = function() self$ellmer_chat,

      # Replace the underlying ellmer chat object and re-sync `parameters$model`
      set_chat = function(new_chat) {
        stopifnot(is.environment(new_chat), is.function(new_chat$chat))
        self$ellmer_chat <- new_chat
        private$sync_model()
        invisible(self)
      }
    ),
    private = list(
      complete_chat_function = NULL,
      .base_tools_snapshot = NULL,

      # Keep self$parameters$model in sync with ellmer_chat$get_model()
      sync_model = function() {
        m <- tryCatch(
          {
            gm <- self$ellmer_chat$get_model
            if (is.function(gm)) self$ellmer_chat$get_model() else NULL
          },
          error = function(e) NULL
        )
        if (!is.null(m)) {
          self$parameters$model <- as.character(m)
        }
        invisible(NULL)
      }
    )
  )

  provider <- klass$new(
    complete_chat_function = complete_chat,
    parameters = list(),
    verbose = verbose,
    api_type = "ellmer"
  )

  if (length(parameters)) {
    provider$set_parameters(parameters)
  }

  provider$ellmer_chat <- chat
  # Initial sync so $parameters$model reflects the chat's current model
  provider$.__enclos_env__$private$sync_model()

  provider
}

testthat::test_that("llm_provider_ellmer supports multimodal structured output", {
  testthat::skip_if_not_installed("ellmer")

  provider <- llm_provider_ellmer(fake_ellmer_chat(), verbose = FALSE)
  provider$parameters$.ellmer_structured_type <- list(schema = "stub")
  provider$parameters$.add_image_parts <- list(
    list(
      source = "url",
      data = "https://example.com/image.png",
      detail = "low"
    )
  )

  chat_history <- data.frame(
    role = "user",
    content = "Hello world",
    stringsAsFactors = FALSE
  )

  result <- provider$complete_chat(chat_history)
  fake_chat <- provider$ellmer_chat

  testthat::expect_equal(fake_chat$last_method$method, "chat_structured")
  # Content objects passed directly (not pre-seeded as turns + empty prompt)
  testthat::expect_true(length(fake_chat$last_method$args) >= 2)
  # No prior turns should be pre-seeded (single-message history)
  testthat::expect_length(fake_chat$last_method$turns, 0)
  testthat::expect_match(
    tail(result$completed$content, 1),
    '"result":"ok"'
  )
})


testthat::test_that("llm_provider_ellmer streams multimodal prompts when supported", {
  testthat::skip_if_not_installed("coro")
  testthat::skip_if_not_installed("ellmer")

  provider <- llm_provider_ellmer(
    fake_ellmer_chat(),
    parameters = list(stream = TRUE),
    verbose = FALSE
  )
  provider$parameters$.add_image_parts <- list(
    list(
      source = "url",
      data = "https://example.com/image.png",
      detail = "low"
    )
  )

  chunks <- character()
  provider$stream_callback <- function(chunk, meta) {
    chunks <<- c(chunks, chunk)
  }

  chat_history <- data.frame(
    role = "user",
    content = "Ping",
    stringsAsFactors = FALSE
  )

  result <- provider$complete_chat(chat_history)
  fake_chat <- provider$ellmer_chat

  testthat::expect_equal(fake_chat$last_method$method, "stream")
  # Content objects passed directly (not pre-seeded + empty prompt)
  testthat::expect_true(length(fake_chat$last_method$args) >= 2)
  testthat::expect_equal(
    result$completed$content[nrow(result$completed)],
    "chunk-end"
  )
  testthat::expect_true(length(chunks) >= 2)
})

testthat::test_that("ellmer follow-up calls replay multimodal turns with native content", {
  skip_if_no_ellmer_turn_classes()

  make_multimodal_fake_chat <- function(turns = list()) {
    chat_env <- fake_ellmer_chat(turns = turns)

    chat_env$chat <- function(...) {
      args <- list(...)
      chat_env$last_method <- list(
        method = "chat",
        args = args,
        turns = chat_env$turns
      )

      user_contents <- lapply(args, function(x) {
        if (is.character(x)) {
          return(ellmer::ContentText(x))
        }
        x
      })

      chat_env$turns <- c(
        chat_env$turns,
        list(
          ellmer::UserTurn(user_contents),
          ellmer::AssistantTurn(list(
            ellmer::ContentThinking("internal reasoning"),
            ellmer::content_image_url("https://example.com/cat.png"),
            ellmer::ContentText("A cat")
          ))
        )
      )

      "A cat"
    }

    chat_env$clone <- function() {
      copy <- make_multimodal_fake_chat(turns = chat_env$turns)
      copy$last_method <- chat_env$last_method
      copy$set_turns_calls <- chat_env$set_turns_calls
      copy
    }

    chat_env
  }

  fake_chat <- make_multimodal_fake_chat()

  provider <- llm_provider_ellmer(
    fake_chat,
    parameters = list(stream = FALSE),
    verbose = FALSE
  )

  first <- tidyprompt("Describe the image") |>
    add_image("https://example.com/cat.png") |>
    send_prompt(provider, return_mode = "full", verbose = FALSE)

  provider$ellmer_chat <- first$ellmer_chat

  follow_up <- first$chat_history |>
    add_msg_to_chat_history("What image was that?") |>
    send_prompt(provider, return_mode = "full", verbose = FALSE)

  prior_turns <- follow_up$ellmer_chat$last_method$turns

  testthat::expect_length(prior_turns, 2)
  testthat::expect_equal(prior_turns[[1]]@role, "user")
  testthat::expect_length(prior_turns[[1]]@contents, 2)
  testthat::expect_true(any(grepl(
    "ContentText",
    class(prior_turns[[1]]@contents[[1]])
  )))
  testthat::expect_true(any(grepl(
    "ContentImageRemote",
    class(prior_turns[[1]]@contents[[2]])
  )))

  testthat::expect_equal(prior_turns[[2]]@role, "assistant")
  testthat::expect_length(prior_turns[[2]]@contents, 2)
  testthat::expect_false(any(vapply(
    prior_turns[[2]]@contents,
    function(x) any(grepl("ContentThinking", class(x))),
    logical(1)
  )))
  testthat::expect_true(any(grepl(
    "ContentImageRemote",
    class(prior_turns[[2]]@contents[[1]])
  )))
  testthat::expect_equal(prior_turns[[2]]@contents[[2]]@text, "A cat")
})

# Tests for hardened streaming in llm_provider_ellmer:
# 1. stream_callback errors are caught (stream continues)
# 2. Streaming-init failures fall back to non-streaming chat

test_that("stream_callback errors are caught and warned, stream continues", {
  skip_if_not_installed("ellmer")
  skip_if_not_installed("coro")

  # Build a tiny fake ellmer-like chat that supports stream().
  # stream() returns a coro generator yielding two chunks.
  fake_chat <- new.env(parent = emptyenv())
  fake_chat$chat <- function(...) "ok"
  fake_chat$clone <- function() fake_chat
  fake_chat$set_turns <- function(turns) {
    fake_chat
  }
  fake_chat$get_turns <- function() list()
  fake_chat$stream <- function(...) {
    coro::gen({
      coro::yield("hello")
      coro::yield(" world")
    })()
  }

  provider <- llm_provider_ellmer(fake_chat, verbose = FALSE)
  provider$parameters$stream <- TRUE

  bomb_count <- 0L
  provider$stream_callback <- function(chunk, meta) {
    bomb_count <<- bomb_count + 1L
    stop("callback boom")
  }

  chat_hist <- data.frame(
    role = "user",
    content = "hi",
    stringsAsFactors = FALSE
  )

  # The callback explodes on every chunk; we should get warnings but NOT an error.
  warnings_seen <- character()
  result <- withCallingHandlers(
    provider$complete_chat(chat_hist),
    warning = function(w) {
      warnings_seen <<- c(warnings_seen, conditionMessage(w))
      tryInvokeRestart("muffleWarning")
    }
  )

  expect_true(bomb_count >= 1L)
  expect_true(any(grepl("stream_callback error", warnings_seen)))
  # The assistant text should still be accumulated despite callback errors.
  assistant <- utils::tail(result$completed$content, 1)
  expect_true(grepl("hello", assistant))
})

test_that("streaming-init failure falls back to non-streaming chat", {
  skip_if_not_installed("ellmer")

  fake_chat <- new.env(parent = emptyenv())
  fake_chat$chat <- function(...) "fallback reply"
  fake_chat$clone <- function() fake_chat
  fake_chat$set_turns <- function(turns) {
    fake_chat
  }
  fake_chat$get_turns <- function() list()
  fake_chat$stream <- function(...) stop("stream not supported")

  provider <- llm_provider_ellmer(fake_chat, verbose = FALSE)
  provider$parameters$stream <- TRUE

  chat_hist <- data.frame(
    role = "user",
    content = "hi",
    stringsAsFactors = FALSE
  )

  # Should fall back to ch$chat() instead of crashing.
  result <- provider$complete_chat(chat_hist)
  assistant <- utils::tail(result$completed$content, 1)
  expect_equal(assistant, "fallback reply")
})

# Test: Initialization of llm_provider with basic parameters
test_that("llm_provider initializes with parameters", {
  parameters <- list(model = "my-llm-model", api_key = "my-api-key")
  provider <- `llm_provider-class`$new(
    complete_chat_function = function(chat_history) {
      list(role = "assistant", content = "Hello")
    },
    parameters = parameters
  )

  expect_s3_class(provider, "LlmProvider")
  expect_equal(provider$parameters, parameters)
})

# Test: Setting and updating parameters
test_that("llm_provider updates parameters correctly", {
  parameters <- list(model = "my-llm-model", api_key = "my-api-key")
  provider <- `llm_provider-class`$new(
    complete_chat_function = function(chat_history) {
      list(role = "assistant", content = "Hello")
    },
    parameters = parameters
  )

  # Update parameters
  new_parameters <- list(api_key = "new-api-key", timeout = 10)
  provider$set_parameters(new_parameters)

  updated_params <- provider$parameters
  expect_equal(updated_params$model, "my-llm-model")
  expect_equal(updated_params$api_key, "new-api-key")
  expect_equal(updated_params$timeout, 10)
})

# Test: complete_chat function with verbose on
test_that("llm_provider complete_chat prints message when verbose is TRUE", {
  test_chat_function <- function(chat_history) {
    return(
      list(
        completed = data.frame(
          role = "assistant",
          content = "Hello!"
        )
      )
    )
  }

  provider <- `llm_provider-class`$new(
    complete_chat_function = test_chat_function,
    verbose = TRUE
  )

  # Test interaction with chat history
  chat_history <- data.frame(role = "user", content = "Hello")
  expect_message(
    provider$complete_chat(list(chat_history = chat_history)),
    "--- Receiving response from LLM provider: ---"
  )
})

# Test: Fake LLM provider responses
test_that("llm_provider_fake returns expected response for known prompt", {
  provider_fake <- llm_provider_fake(verbose = FALSE)

  chat_history <- data.frame(role = "user", content = "Hi there!")
  result <- provider_fake$complete_chat(list(chat_history = chat_history))
  response <- result$completed |> utils::tail(1)

  expect_equal(response$role, "assistant")
  expect_match(response$content, "nice to meet you")
})

test_that("llm_provider normalizes sparse tool metadata in completed history", {
  provider <- `llm_provider-class`$new(
    complete_chat_function = function(chat_history) {
      list(
        completed = dplyr::bind_rows(
          chat_history,
          data.frame(
            role = "assistant",
            content = "Hello!",
            stringsAsFactors = FALSE
          )
        ),
        http = list(request = NULL, response = NULL)
      )
    },
    verbose = FALSE
  )

  result <- provider$complete_chat(list(
    chat_history = data.frame(
      role = "user",
      content = "Hello",
      tool_call = FALSE,
      tool_call_id = NA_character_,
      tool_result = FALSE,
      stringsAsFactors = FALSE
    )
  ))

  expect_equal(result$completed$tool_call, c(FALSE, FALSE))
  expect_true(all(is.na(result$completed$tool_call_id)))
  expect_equal(result$completed$tool_result, c(FALSE, FALSE))
})

test_that("llm_provider excludes hidden rows from outgoing requests", {
  state <- new.env(parent = emptyenv())
  state$seen_chat_history <- NULL

  provider <- `llm_provider-class`$new(
    complete_chat_function = function(chat_history) {
      self$parameters$.test_state$seen_chat_history <- chat_history

      list(
        completed = dplyr::bind_rows(
          chat_history,
          data.frame(
            role = "assistant",
            content = "Fixed answer",
            stringsAsFactors = FALSE
          )
        ),
        http = list(request = NULL, response = NULL)
      )
    },
    parameters = list(.test_state = state),
    verbose = FALSE
  )

  input <- data.frame(
    role = c("user", "assistant", "assistant", "user"),
    content = c("Question", "Reasoning step", "Answer", "Follow-up"),
    hidden_from_llm = c(FALSE, TRUE, FALSE, FALSE),
    stringsAsFactors = FALSE
  )

  result <- provider$complete_chat(list(chat_history = input))

  expect_equal(
    state$seen_chat_history$content,
    c("Question", "Answer", "Follow-up")
  )
  expect_false(any(state$seen_chat_history$hidden_from_llm))
  expect_equal(
    result$completed$content,
    c("Question", "Reasoning step", "Answer", "Follow-up", "Fixed answer")
  )
  expect_equal(
    result$completed$hidden_from_llm,
    c(FALSE, TRUE, FALSE, FALSE, FALSE)
  )
})

test_that("llm_provider excludes tool call rows from outgoing requests", {
  state <- new.env(parent = emptyenv())
  state$seen_chat_history <- NULL

  provider <- `llm_provider-class`$new(
    complete_chat_function = function(chat_history) {
      self$parameters$.test_state$seen_chat_history <- chat_history

      list(
        completed = dplyr::bind_rows(
          chat_history,
          data.frame(
            role = "assistant",
            content = "Fixed answer",
            stringsAsFactors = FALSE
          )
        ),
        http = list(request = NULL, response = NULL)
      )
    },
    parameters = list(.test_state = state),
    verbose = FALSE
  )

  input <- data.frame(
    role = c("user", "assistant", "tool", "assistant", "user"),
    content = c(
      "Question",
      "~>> Calling function 'tool' with arguments:\n{}",
      "~>> Result:\n42",
      "Answer",
      "Follow-up"
    ),
    tool_call = c(FALSE, TRUE, FALSE, FALSE, FALSE),
    tool_call_id = c(
      NA_character_,
      "call-1",
      "call-1",
      NA_character_,
      NA_character_
    ),
    tool_result = c(FALSE, FALSE, TRUE, FALSE, FALSE),
    stringsAsFactors = FALSE
  )

  result <- provider$complete_chat(list(chat_history = input))

  expect_equal(
    state$seen_chat_history$content,
    c("Question", "~>> Result:\n42", "Answer", "Follow-up")
  )
  expect_false(any(state$seen_chat_history$tool_call))
  expect_equal(
    result$completed$hidden_from_llm,
    c(FALSE, TRUE, FALSE, FALSE, FALSE, FALSE)
  )
})

# Test: Invalid parameters handling
test_that("llm_provider errors on invalid parameters", {
  expect_error(
    `llm_provider-class`$new(
      complete_chat_function = function(chat_history) {
        list(role = "assistant", content = "Hello")
      },
      parameters = list("unnamed")
    ),
    "parameters must be a named list"
  )
})

testthat::test_that("ellmer provider prefers role-specific Turn constructors", {
  testthat::skip_if_not_installed("ellmer")

  fake_chat <- fake_ellmer_chat()
  provider <- llm_provider_ellmer(
    fake_chat,
    parameters = list(stream = FALSE),
    verbose = FALSE
  )

  chat_history <- data.frame(
    role = c("system", "user"),
    content = c("Be helpful.", "Hello"),
    stringsAsFactors = FALSE
  )

  result <- provider$complete_chat(chat_history)
  turns <- result$ellmer_chat$turns

  # The prior turn (system) should be a Turn object with contents
  testthat::expect_true(length(turns) >= 1)
  testthat::expect_true(length(turns[[1]]@contents) >= 1)
  testthat::expect_equal(turns[[1]]@role, "system")
})

testthat::test_that("ellmer provider builds Turn objects with correct contents", {
  testthat::skip_if_not_installed("ellmer")

  fake_chat <- fake_ellmer_chat()
  provider <- llm_provider_ellmer(
    fake_chat,
    parameters = list(stream = FALSE),
    verbose = FALSE
  )

  chat_history <- data.frame(
    role = c("user", "assistant", "user"),
    content = c("First", "Reply", "Second"),
    stringsAsFactors = FALSE
  )

  result <- provider$complete_chat(chat_history)
  turns <- result$ellmer_chat$turns

  # Two prior turns should have been built (the third is the current prompt)
  testthat::expect_length(turns, 2)

  # Check contents are accessible via S7 slot
  for (turn in turns) {
    testthat::expect_true(length(turn@contents) >= 1)
  }
})

testthat::test_that("ellmer provider handles tool-role rows without crashing", {
  testthat::skip_if_not_installed("ellmer")

  fake_chat <- fake_ellmer_chat()
  provider <- llm_provider_ellmer(fake_chat, verbose = FALSE)

  # Simulate a chat history containing tool-role rows (as produced by

  # the openai/ollama handler in answer_using_tools)
  chat_history <- data.frame(
    role = c("user", "assistant", "tool", "assistant", "user"),
    content = c(
      "Call my_func(1)",
      "Calling my_func...",
      "~>> Result:\n42",
      "The result is 42.",
      "Thanks"
    ),
    tool_result = c(FALSE, FALSE, TRUE, FALSE, FALSE),
    stringsAsFactors = FALSE
  )

  # Should not error; tool rows are mapped to user turns
  result <- testthat::expect_no_error(
    provider$complete_chat(chat_history)
  )

  turns <- result$ellmer_chat$turns
  # 4 prior turns (all except last), tool row mapped to user
  testthat::expect_length(turns, 4)
  # The third turn (originally "tool") should now be a user turn
  testthat::expect_equal(turns[[3]]@role, "user")
})

testthat::test_that("ellmer provider normalizes tool metadata on plain replies", {
  testthat::skip_if_not_installed("ellmer")

  fake_chat <- fake_ellmer_chat()
  provider <- llm_provider_ellmer(
    fake_chat,
    parameters = list(stream = FALSE),
    verbose = FALSE
  )

  result <- provider$complete_chat(data.frame(
    role = "user",
    content = "what is 5+5",
    stringsAsFactors = FALSE
  ))

  testthat::expect_equal(result$completed$role, c("user", "assistant"))
  testthat::expect_equal(result$completed$tool_call, c(FALSE, FALSE))
  testthat::expect_true(all(is.na(result$completed$tool_call_id)))
  testthat::expect_equal(result$completed$tool_result, c(FALSE, FALSE))
})

testthat::test_that("ellmer provider preserves native tool history in completed rows", {
  skip_if_no_ellmer_turn_classes()

  fake_chat <- fake_ellmer_chat()
  fake_chat$chat <- function(...) {
    request <- ellmer::ContentToolRequest(
      id = "call-1",
      name = "get_secret_number",
      arguments = list(input = 123)
    )

    fake_chat$turns <- c(
      fake_chat$turns,
      list(
        ellmer::UserTurn(list(ellmer::ContentText("Call the function"))),
        ellmer::AssistantTurn(list(request)),
        ellmer::UserTurn(list(
          ellmer::ContentToolResult(value = "42", request = request)
        )),
        ellmer::AssistantTurn(list(ellmer::ContentText("The result is 42.")))
      )
    )

    "The result is 42."
  }

  provider <- llm_provider_ellmer(
    fake_chat,
    parameters = list(stream = FALSE),
    verbose = FALSE
  )

  result <- provider$complete_chat(data.frame(
    role = "user",
    content = "Call the function",
    stringsAsFactors = FALSE
  ))

  testthat::expect_equal(
    result$completed$role,
    c("user", "assistant", "tool", "assistant")
  )
  testthat::expect_equal(
    result$completed$hidden_from_llm,
    c(FALSE, TRUE, FALSE, FALSE)
  )
  testthat::expect_true(isTRUE(result$completed$tool_call[2]))
  testthat::expect_equal(result$completed$tool_call_id[2], "call-1")
  testthat::expect_true(isTRUE(result$completed$tool_result[3]))
  testthat::expect_match(
    result$completed$content[2],
    "Calling function 'get_secret_number'"
  )
  testthat::expect_match(result$completed$content[3], "~>> Result")
  testthat::expect_equal(result$completed$content[4], "The result is 42.")
})

testthat::test_that("ellmer provider preserves native thinking content in completed rows", {
  skip_if_no_ellmer_turn_classes()

  fake_chat <- fake_ellmer_chat()
  fake_chat$chat <- function(...) {
    fake_chat$turns <- c(
      fake_chat$turns,
      list(
        ellmer::UserTurn(list(ellmer::ContentText("Think through this"))),
        ellmer::AssistantTurn(list(
          ellmer::ContentThinking("Reasoning step"),
          ellmer::ContentText("Final answer")
        ))
      )
    )

    "Final answer"
  }

  provider <- llm_provider_ellmer(
    fake_chat,
    parameters = list(stream = FALSE),
    verbose = FALSE
  )

  result <- provider$complete_chat(data.frame(
    role = "user",
    content = "Think through this",
    stringsAsFactors = FALSE
  ))

  testthat::expect_equal(
    result$completed$role,
    c("user", "assistant", "assistant")
  )
  testthat::expect_equal(
    result$completed$hidden_from_llm,
    c(FALSE, TRUE, FALSE)
  )
  testthat::expect_match(result$completed$content[2], "Reasoning step")
  testthat::expect_equal(result$completed$content[3], "Final answer")
})

testthat::test_that("ellmer provider preserves native json content in completed rows", {
  skip_if_no_ellmer_turn_classes()

  ellmer_ns <- asNamespace("ellmer")
  fake_chat <- fake_ellmer_chat()
  fake_chat$chat <- function(...) {
    fake_chat$turns <- c(
      fake_chat$turns,
      list(
        ellmer::UserTurn(list(ellmer::ContentText("Return JSON"))),
        ellmer::AssistantTurn(list(
          ellmer_ns$ContentJson(data = list(answer = 42))
        ))
      )
    )

    "{\"answer\":42}"
  }

  provider <- llm_provider_ellmer(
    fake_chat,
    parameters = list(stream = FALSE),
    verbose = FALSE
  )

  result <- provider$complete_chat(data.frame(
    role = "user",
    content = "Return JSON",
    stringsAsFactors = FALSE
  ))

  testthat::expect_equal(result$completed$role, c("user", "assistant"))
  testthat::expect_equal(result$completed$content[2], '{"answer":42}')
})

testthat::test_that("ellmer provider preserves native image tool results in completed rows", {
  skip_if_no_ellmer_turn_classes()

  ellmer_ns <- asNamespace("ellmer")
  fake_chat <- fake_ellmer_chat()
  fake_chat$chat <- function(...) {
    request <- ellmer::ContentToolRequest(
      id = "call-image-1",
      name = "make_plot",
      arguments = list()
    )

    fake_chat$turns <- c(
      fake_chat$turns,
      list(
        ellmer::UserTurn(list(ellmer::ContentText("Make me a plot"))),
        ellmer::AssistantTurn(list(request)),
        ellmer::UserTurn(list(
          ellmer::ContentToolResult(
            value = ellmer_ns$ContentImageRemote(
              url = "https://example.com/plot.png"
            ),
            request = request
          )
        )),
        ellmer::AssistantTurn(list(ellmer::ContentText("Done")))
      )
    )

    "Done"
  }

  provider <- llm_provider_ellmer(
    fake_chat,
    parameters = list(stream = FALSE),
    verbose = FALSE
  )

  result <- provider$complete_chat(data.frame(
    role = "user",
    content = "Make me a plot",
    stringsAsFactors = FALSE
  ))

  testthat::expect_equal(
    result$completed$role,
    c("user", "assistant", "tool", "assistant")
  )
  testthat::expect_match(
    result$completed$content[3],
    "\\[image: https://example.com/plot.png\\]"
  )
  testthat::expect_equal(result$completed$content[4], "Done")
})

test_that("llm_provider_ellmer streaming uses stream_callback and accumulates text", {
  skip_if_not_installed("ellmer")
  skip_test_if_no_openai()

  # Build a minimal ellmer chat that supports streaming
  ch <- ellmer::chat_openai()

  # Prefer a very small model / cheap model if configured
  model <- Sys.getenv("TIDYPROMPT_OPENAI_CHAT_MODEL", unset = NA_character_)
  if (!is.na(model)) {
    ch <- ch$set_model(model)
  }

  # Force streaming on the underlying chat if such a knob exists; we rely on
  # ch$stream() being available for chat_openai() in recent ellmer versions.
  expect_true(is.function(ch$stream))

  provider <- llm_provider_ellmer(ch, verbose = FALSE)

  # Enable streaming via parameters list
  provider$parameters$stream <- TRUE

  # Tracking env for the callback
  seen <- new.env(parent = emptyenv())
  seen$n_calls <- 0L
  seen$partial_values <- character()

  provider$stream_callback <- function(chunk, meta) {
    seen$n_calls <- seen$n_calls + 1L
    seen$partial_values <- c(seen$partial_values, meta$partial_response)

    # Basic sanity on metadata
    expect_equal(meta$api_type, "ellmer")
    expect_equal(meta$endpoint, "chat")
    expect_true(is.list(meta$chat_history))
  }

  chat_history <- data.frame(
    role = "user",
    content = "Reply with exactly the single word: pong",
    stringsAsFactors = FALSE
  )

  out <- provider$complete_chat(list(chat_history = chat_history))
  expect_true(nrow(out$completed) == 2L)
  assistant <- utils::tail(out$completed$content, 1)
  reply <- tolower(trimws(assistant))

  expect_true(seen$n_calls > 0L)
  expect_true(nchar(reply) > 0)
})

# Tests for unforwardable parameter warnings on ellmer providers

test_that("ellmer provider warns about unforwardable parameters", {
  skip_if_not_installed("ellmer")

  fake_chat <- new.env(parent = emptyenv())
  fake_chat$chat <- function(...) "ok"
  fake_chat$clone <- function() fake_chat
  fake_chat$set_turns <- function(turns) {
    fake_chat
  }
  fake_chat$get_turns <- function() list()
  fake_chat$stream <- NULL

  provider <- llm_provider_ellmer(fake_chat, verbose = FALSE)
  provider$parameters$stream <- FALSE
  provider$parameters$temperature <- 0.7
  provider$parameters$max_tokens <- 100L

  chat_hist <- data.frame(
    role = "user",
    content = "hi",
    stringsAsFactors = FALSE
  )

  # Reset the rlang once-per-session guard so the warning fires in tests.
  rlang::reset_warning_verbosity("tidyprompt_ellmer_extra_params")

  expect_warning(
    provider$complete_chat(chat_hist),
    "not forwarded"
  )
})

test_that("ellmer provider does not warn with only known parameters", {
  skip_if_not_installed("ellmer")

  fake_chat <- new.env(parent = emptyenv())
  fake_chat$chat <- function(...) "ok"
  fake_chat$clone <- function() fake_chat
  fake_chat$set_turns <- function(turns) {
    fake_chat
  }
  fake_chat$get_turns <- function() list()
  fake_chat$stream <- NULL

  provider <- llm_provider_ellmer(fake_chat, verbose = FALSE)
  provider$parameters$stream <- FALSE

  chat_hist <- data.frame(
    role = "user",
    content = "hi",
    stringsAsFactors = FALSE
  )

  expect_no_warning(provider$complete_chat(chat_hist))
})

test_that("answer_as_json (text-based) works without schema", {
  skip_test_if_no_openai()

  response <- "Create a very short persona" |>
    answer_as_json(type = "text-based") |>
    send_prompt(
      llm_provider_openai(
        url = "https://api.openai.com/v1/responses"
      )$set_parameters(list(model = "gpt-4.1-mini"))
    )

  expect_true(is.list(response), info = "Response should be a list")
})

test_that("answer_as_json (text-based) works with json schema", {
  skip_test_if_no_openai()

  schema <- list(
    "$schema" = "http://json-schema.org/draft-04/schema#",
    title = "Persona",
    type = "object",
    properties = list(
      name = list(type = "string", description = "The persona's name"),
      age = list(
        type = "integer",
        minimum = 0,
        description = "The persona's age"
      ),
      gender = list(
        type = "string",
        enum = c("Male", "Female", "Non-binary", "Other"),
        description = "The persona's gender"
      ),
      hobbies = list(
        type = "array",
        items = list(type = "string"),
        description = "List of hobbies"
      ),
      pet = list(
        type = "object",
        description = "Information about the persona's pet",
        properties = list(
          name = list(type = "string", description = "The pet's name"),
          age = list(
            type = "integer",
            minimum = 0,
            description = "The pet's age"
          ),
          species = list(
            type = "string",
            enum = c("Dog", "Cat", "Fish", "Bird", "Other"),
            description = "The pet's species"
          )
        ),
        required = c("name", "age", "species")
      )
    ),
    required = c("name", "age", "gender", "hobbies", "pet"),
    additionalProperties = FALSE
  )

  r_json_schema_to_example(schema)

  response <- "Create a persona" |>
    answer_as_json(schema, schema_strict = TRUE, type = "text-based") |>
    send_prompt(
      llm_provider_openai(
        url = "https://api.openai.com/v1/responses"
      )$set_parameters(list(model = "gpt-4.1-mini"))
    )

  expect_true(is.list(response), info = "Response should be a list")

  # Validate top-level properties
  expect_true(
    all(
      c(
        "name",
        "age",
        "gender",
        "hobbies",
        "pet"
      ) %in%
        names(response)
    )
  )

  # Validate `name`
  expect_true(is.character(response$name), info = "`name` should be a string")
  expect_true(nchar(response$name) > 0, info = "`name` should not be empty")

  # Validate `age`
  expect_true(is.numeric(response$age), info = "`age` should be numeric")
  expect_true(response$age >= 0, info = "`age` should be non-negative")

  # Validate `ge  nder`
  expect_true(
    is.character(response$gender),
    info = "`gender` should be a string"
  )
  expect_true(
    response$gender %in% c("Male", "Female", "Non-binary", "Other"),
    info = "`gender` should be one of the allowed values"
  )

  # Validate `hobbies`
  expect_true(
    is.list(response$hobbies) || is.character(response$hobbies),
    info = "`hobbies` should be a list or a character vector"
  )
  expect_true(
    all(sapply(response$hobbies, is.character)),
    info = "All items in `hobbies` should be strings"
  )

  # Validate `pet` object
  expect_true(is.list(response$pet), info = "`pet` should be a list")
  expect_true(
    is.character(response$pet$name),
    info = "`pet$name` should be a string"
  )
  expect_true(
    nchar(response$pet$name) > 0,
    info = "`pet$name` should not be empty"
  )
  expect_true(
    is.numeric(response$pet$age),
    info = "`pet$age` should be numeric"
  )
  expect_true(response$pet$age >= 0, info = "`pet$age` should be non-negative")
  expect_true(
    is.character(response$pet$species),
    info = "`pet$species` should be a string"
  )
  expect_true(
    response$pet$species %in% c("Dog", "Cat", "Fish", "Bird", "Other"),
    info = "`pet$species` should be one of the allowed values"
  )
})

test_that("answer_as_json (openai via auto) works", {
  skip_test_if_no_openai()

  schema <- list(
    "$schema" = "http://json-schema.org/draft-04/schema#",
    title = "Persona",
    type = "object",
    properties = list(
      name = list(type = "string", description = "The persona's name"),
      age = list(type = "integer", description = "The persona's age"),
      gender = list(
        type = "string",
        enum = c("Male", "Female", "Non-binary", "Other"),
        description = "The persona's gender"
      ),
      hobbies = list(
        type = "array",
        items = list(type = "string"),
        description = "List of hobbies"
      )
    ),
    required = c("name", "age", "gender", "hobbies"),
    additionalProperties = FALSE
  )

  expect_no_error(
    "Create a persona" |>
      answer_as_json(schema, type = "auto") |>
      send_prompt(llm_provider_openai())
  )

  expect_no_error(
    "Create a very short persona" |>
      answer_as_json(type = "auto") |>
      send_prompt(llm_provider_openai())
  )
})

test_that("answer_as_json (ollama via auto) works", {
  skip_test_if_no_ollama()

  schema <- list(
    "$schema" = "http://json-schema.org/draft-04/schema#",
    title = "Persona",
    type = "object",
    properties = list(
      name = list(type = "string", description = "The persona's name"),
      age = list(type = "integer", description = "The persona's age"),
      gender = list(
        type = "string",
        enum = c("Male", "Female", "Non-binary", "Other"),
        description = "The persona's gender"
      ),
      hobbies = list(
        type = "array",
        items = list(type = "string"),
        description = "List of hobbies"
      )
    ),
    required = c("name", "age", "gender", "hobbies"),
    additionalProperties = FALSE
  )

  expect_no_error(
    "Create a persona" |>
      answer_as_json(schema, type = "auto") |>
      send_prompt(llm_provider_ollama())
  )

  expect_no_error(
    "Create a very short persona" |>
      answer_as_json(type = "auto") |>
      send_prompt(llm_provider_ollama())
  )
})

test_that("answer_as_json (compatability with ellmer) works", {
  skip_test_if_no_openai()
  skip_if_not_installed("ellmer")

  # Persona validation function
  is_valid_persona <- function(persona) {
    if (!is.list(persona)) {
      return(FALSE)
    }

    required_fields <- c("name", "age", "hobbies")
    if (!all(required_fields %in% names(persona))) {
      return(FALSE)
    }

    if (!is.character(persona$name) || length(persona$name) != 1) {
      return(FALSE)
    }
    if (!is.numeric(persona$age) || length(persona$age) != 1) {
      return(FALSE)
    }
    if (
      !is.vector(persona$hobbies) || !all(sapply(persona$hobbies, is.character))
    ) {
      return(FALSE)
    }

    TRUE
  }

  # Ellmer LLM provider
  ellmer_openai <- llm_provider_ellmer(ellmer::chat_openai(
    model = "gpt-4.1-mini"
  ))

  # Ellmer schema
  ellmer_schema <- ellmer::type_object(
    name = ellmer::type_string(),
    age = ellmer::type_integer(),
    hobbies = ellmer::type_array(ellmer::type_string())
  )

  # Ellmer LLM provider with Ellmer schema
  result_ellmer_x_ellmer <- "Create a persona" |>
    answer_as_json(ellmer_schema) |>
    send_prompt(ellmer_openai)
  expect_true(is_valid_persona(result_ellmer_x_ellmer))

  # Regular LLM provider with Ellmer schema
  result_tidyrpompt_x_ellmer <- "Create a persona" |>
    answer_as_json(ellmer_schema) |>
    send_prompt(
      llm_provider_openai(
        url = "https://api.openai.com/v1/responses"
      )$set_parameters(list(model = "gpt-4.1-mini"))
    )
  expect_true(is_valid_persona(result_tidyrpompt_x_ellmer))

  # Regular schema
  schema <- list(
    "$schema" = "http://json-schema.org/draft-04/schema#",
    title = "Persona",
    type = "object",
    properties = list(
      name = list(type = "string", description = "The persona's name"),
      age = list(type = "integer", description = "The persona's age"),
      hobbies = list(
        type = "array",
        items = list(type = "string"),
        description = "List of hobbies"
      )
    ),
    required = c("name", "age", "hobbies"),
    additionalProperties = FALSE
  )

  # Ellmer LLM provider with regular schema
  result_ellmer_x_regular <- "Create a persona" |>
    answer_as_json(schema) |>
    send_prompt(ellmer_openai)
  expect_true(is_valid_persona(result_ellmer_x_regular))
})

test_that("answer_as_json + llm_provider_ellmer: scalar types", {
  skip_test_if_no_openai()
  skip_if_not_installed("ellmer")

  ellmer_openai <- llm_provider_ellmer(ellmer::chat_openai(
    model = "gpt-4.1-mini"
  ))

  # type_string()
  res_str <- "Return the string 'hello' only." |>
    answer_as_json(ellmer::type_string()) |>
    send_prompt(ellmer_openai)
  expect_true(is.character(res_str) && length(res_str) == 1)

  # type_integer()
  res_int <- "Return the integer 7 only." |>
    answer_as_json(ellmer::type_integer()) |>
    send_prompt(ellmer_openai)
  expect_true(is.numeric(res_int) && length(res_int) == 1)
  expect_equal(as.integer(res_int), 7L)

  # type_number()
  res_num <- "Return the number 3.14 only." |>
    answer_as_json(ellmer::type_number()) |>
    send_prompt(ellmer_openai)
  expect_true(is.numeric(res_num) && length(res_num) == 1)
  expect_gt(res_num, 3)
  expect_lt(res_num, 3.2)

  # type_boolean()
  res_bool <- "Return the boolean true only." |>
    answer_as_json(ellmer::type_boolean()) |>
    send_prompt(ellmer_openai)
  expect_true(is.logical(res_bool) && length(res_bool) == 1)

  # type_enum()
  allowed <- c("Technology", "Sports", "Politics")
  res_enum <- "Categorize: 'AI chips surge in demand'. Choose one of Technology, Sports, Politics; return only the category." |>
    answer_as_json(ellmer::type_enum(allowed)) |>
    send_prompt(ellmer_openai)
  enum_val <- if (is.factor(res_enum)) as.character(res_enum) else res_enum
  expect_true(is.character(enum_val) && enum_val %in% allowed)
})

test_that("answer_as_json works with direct ellmer chat", {
  skip_test_if_no_openai()
  skip_if_not_installed("ellmer")

  ellmer_chat <- ellmer::chat_openai(model = "gpt-4.1-mini")

  result <- "Return the integer 7 only." |>
    answer_as_json(ellmer::type_integer()) |>
    send_prompt(ellmer_chat)

  expect_true(is.numeric(result) && length(result) == 1)
  expect_equal(as.integer(result), 7L)
  expect_length(ellmer_chat$get_turns(), 0)
})

test_that("answer_as_json + llm_provider_ellmer: arrays of scalars", {
  skip_test_if_no_openai()
  skip_if_not_installed("ellmer")

  ellmer_openai <- llm_provider_ellmer(ellmer::chat_openai(
    model = "gpt-4.1-mini"
  ))

  # vector of integers
  res_int_vec <- "Return exactly: 18, 21, 30." |>
    answer_as_json(ellmer::type_array(ellmer::type_integer())) |>
    send_prompt(ellmer_openai)
  expect_true(is.numeric(res_int_vec))
  expect_length(res_int_vec, 3)
  expect_equal(as.integer(res_int_vec), c(18L, 21L, 30L))

  # vector of numbers
  res_num_vec <- "Return exactly: 1.5, 2.0, 2.5." |>
    answer_as_json(ellmer::type_array(ellmer::type_number())) |>
    send_prompt(ellmer_openai)
  expect_true(is.numeric(res_num_vec))
  expect_equal(length(res_num_vec), 3)
  expect_true(all(res_num_vec > 0))

  # vector of enums
  allowed <- c("red", "green", "blue")
  res_enum_vec <- "Return these colours in order: red, green, blue." |>
    answer_as_json(ellmer::type_array(ellmer::type_enum(allowed))) |>
    send_prompt(ellmer_openai)
  expect_true(is.character(res_enum_vec) || is.factor(res_enum_vec))
  if (is.factor(res_enum_vec)) {
    res_enum_vec <- as.character(res_enum_vec)
  }
  expect_equal(res_enum_vec, allowed)
})

test_that("answer_as_json + llm_provider_ellmer: arrays of objects -> data.frame", {
  skip_test_if_no_openai()
  skip_if_not_installed("ellmer")

  ellmer_openai <- llm_provider_ellmer(ellmer::chat_openai(
    model = "gpt-4.1-mini"
  ))

  prompt <- r"(
    * John Smith. Age: 30. Height: 180 cm. Weight: 80 kg.
    * Jane Doe. Age: 25. Height: 165 cm. Weight: 50 kg.
    * Jose Rodriguez. Age: 40. Height: 190 cm. Weight: 90 kg.
  )"

  type_person <- ellmer::type_object(
    name = ellmer::type_string(),
    age = ellmer::type_integer(),
    height = ellmer::type_number("in cm"),
    weight = ellmer::type_number("in kg")
  )
  type_people <- ellmer::type_array(type_person)

  df <- prompt |>
    answer_as_json(type_people) |>
    send_prompt(ellmer_openai)

  expect_s3_class(df, "data.frame")
  expect_true(all(c("name", "age", "height", "weight") %in% names(df)))
  expect_equal(nrow(df), 3)
  expect_true(is.character(df$name))
  expect_true(
    is.numeric(df$age) && is.numeric(df$height) && is.numeric(df$weight)
  )
})

test_that("answer_as_json + llm_provider_ellmer: optional fields (required = FALSE)", {
  skip_test_if_no_openai()
  skip_if_not_installed("ellmer")

  ellmer_openai <- llm_provider_ellmer(ellmer::chat_openai(
    model = "gpt-4.1-mini"
  ))

  type_person_opt <- ellmer::type_object(
    name = ellmer::type_string(required = FALSE),
    age = ellmer::type_integer(required = FALSE)
  )

  # Only age present
  res_age_only <- "I'm 33 years old." |>
    answer_as_json(type_person_opt) |>
    send_prompt(ellmer_openai)
  expect_true(is.list(res_age_only))
  expect_true(all(names(res_age_only) %in% c("name", "age")))
  if ("age" %in% names(res_age_only)) {
    expect_true(is.numeric(res_age_only$age) && length(res_age_only$age) == 1)
  }

  # Only name present
  res_name_only <- "My name is Taylor." |>
    answer_as_json(type_person_opt) |>
    send_prompt(ellmer_openai)
  expect_true(is.list(res_name_only))
  expect_true(all(names(res_name_only) %in% c("name", "age")))
  if ("name" %in% names(res_name_only)) {
    expect_true(
      is.character(res_name_only$name) && nchar(res_name_only$name) > 0
    )
  }
})

test_that("answer_as_json + llm_provider_ellmer: arrays of enums and objects combined", {
  skip_test_if_no_openai()
  skip_if_not_installed("ellmer")

  ellmer_openai <- llm_provider_ellmer(ellmer::chat_openai(
    model = "gpt-4.1-mini"
  ))

  # Object with an enum and an array of enums
  type_pref <- ellmer::type_object(
    primary = ellmer::type_enum(c("red", "green", "blue")),
    fallback = ellmer::type_array(ellmer::type_enum(c("red", "green", "blue")))
  )

  res_pref <- "Primary colour is green; fallbacks are red then blue." |>
    answer_as_json(type_pref) |>
    send_prompt(ellmer_openai)

  expect_true(is.list(res_pref))
  primary <- if (is.factor(res_pref$primary)) {
    as.character(res_pref$primary)
  } else {
    res_pref$primary
  }
  expect_true(primary %in% c("red", "green", "blue"))

  fallbacks <- res_pref$fallback
  if (is.factor(fallbacks)) {
    fallbacks <- as.character(fallbacks)
  }
  expect_true(is.character(fallbacks))
  expect_true(all(fallbacks %in% c("red", "green", "blue")))
})

test_that("answer_as_json ellmer path falls back to text extraction when ellmer_type is NULL", {
  skip_if_not_installed("ellmer")
  withr::local_options(list(
    tidyprompt.stream = FALSE,
    tidyprompt.warn.auto.json = FALSE
  ))

  # Build a fake ellmer provider whose api_type is "ellmer" but where the
  # schema deliberately fails ellmer type conversion (ellmer_type = NULL).
  # The LLM returns fenced JSON, which bare jsonlite::fromJSON() would choke on.
  fenced_json <- "Here is the result:\n```json\n{\"name\":\"Alice\",\"age\":30}\n```"

  fake_chat <- fake_ellmer_chat()
  fake_chat$chat <- function(...) fenced_json
  fake_chat$clone <- function() {
    copy <- fake_ellmer_chat()
    copy$chat <- function(...) fenced_json
    copy
  }

  provider <- llm_provider_ellmer(
    fake_chat,
    verbose = FALSE,
    parameters = list(stream = FALSE)
  )

  schema <- list(
    type = "object",
    properties = list(
      name = list(type = "string"),
      age = list(type = "integer")
    ),
    required = c("name", "age")
  )

  # Force ellmer type conversion to fail so sch$ellmer_type is NULL.
  local_mocked_bindings(
    json_schema_to_ellmer_type = function(...) stop("nope"),
    .package = "tidyprompt"
  )

  result <- "Create a persona" |>
    answer_as_json(schema, type = "ellmer") |>
    send_prompt(provider)

  # Should successfully extract from fenced JSON instead of erroring
  expect_true(is.list(result))
  expect_equal(result$name, "Alice")
  expect_equal(result$age, 30)
})

testthat::test_that("send message with fake llm provider", {
  ollama <- llm_provider_fake()
  result <- ollama$complete_chat("Hi!")

  testthat::expect_equal(nrow(result$completed), 2)

  # Check that 'role' and 'response' are in the result
  testthat::expect_true(all(c("role", "content") %in% names(result$completed)))
})

testthat::test_that("send basic prompt", {
  result <- "Hi" |>
    send_prompt(llm_provider_fake())

  # Check that the result is a single character string
  testthat::expect_true(is.character(result))
  testthat::expect_length(result, 1)
})

testthat::test_that("send prompt with text added", {
  result <- "Hi" |>
    add_text("How are you?") |>
    send_prompt(llm_provider_fake())

  testthat::expect_true(is.character(result))
  testthat::expect_length(result, 1)
})

testthat::test_that("send prompt with validation & extraction added", {
  result <- "Hi" |>
    add_text("What is 2 + 2?") |>
    answer_as_integer() |>
    send_prompt(llm_provider_fake())

  is_whole_number <- function(x) {
    is.numeric(x) && x == floor(x)
  }

  testthat::expect_true(is_whole_number(result))
  testthat::expect_length(result, 1)
})

testthat::test_that("send prompt with mode added", {
  result <- "What is 2 + 2?" |>
    answer_by_chain_of_thought() |>
    answer_as_integer() |>
    send_prompt(llm_provider_fake())

  is_whole_number <- function(x) {
    is.numeric(x) && x == floor(x)
  }

  testthat::expect_true(is_whole_number(result))
  testthat::expect_length(result, 1)
})

openai_provider <- llm_provider_openai(parameters = list(model = "gpt-4o-mini"))
ellmer_provider <- llm_provider_ellmer(ellmer::chat_openai(
  model = "gpt-4o-mini"
))
ollama_provider <- llm_provider_ollama(parameters = list(model = "qwen3-vl:2b"))

if (!testthat:::on_ci()) {
  cat_img_file <- tempfile(fileext = ".jpg")

  download.file(
    "https://upload.wikimedia.org/wikipedia/commons/3/3a/Cat03.jpg",
    destfile = cat_img_file,
    mode = "wb"
  )

  img_file_prompt <- "Describe this image" |>
    add_image(cat_img_file)

  cat_plot <- ggplot2::ggplot() +
    ggplot2::annotate(
      "text",
      x = 0.5,
      y = 0.5,
      label = "cat",
      size = 50,
      hjust = 0.5,
      vjust = 0.5
    ) +
    ggplot2::theme_void()

  img_plot_prompt <- "Describe this image" |>
    add_image(cat_plot)
}

cat_mentioned <- function(text) {
  grepl("cat", text, ignore.case = TRUE)
}

testthat::test_that("openai provider - add_image works (file)", {
  skip_test_if_no_openai()
  skip_on_ci()

  result <- send_prompt(
    img_file_prompt,
    llm_provider = openai_provider
  )

  testthat::expect_true(
    cat_mentioned(result)
  )
})

testthat::test_that("openai provider - add_image works (plot)", {
  skip_test_if_no_openai()

  result <- send_prompt(
    img_plot_prompt,
    llm_provider = openai_provider
  )

  testthat::expect_true(
    cat_mentioned(result)
  )
})

testthat::test_that("ellmer provider - add_image works (file)", {
  skip_test_if_no_openai()
  skip_if_not_installed("ellmer")
  skip_if_not_installed("magick")
  skip_on_ci()

  result <- send_prompt(
    img_file_prompt,
    llm_provider = ellmer_provider
  )

  testthat::expect_true(
    cat_mentioned(result)
  )
})

testthat::test_that("ellmer provider - add_image works (plot)", {
  skip_test_if_no_openai()
  skip_if_not_installed("ellmer")
  skip_if_not_installed("magick")

  result <- send_prompt(
    img_plot_prompt,
    llm_provider = ellmer_provider
  )

  testthat::expect_true(
    cat_mentioned(result)
  )
})

testthat::test_that("ollama provider - add_image works (file)", {
  skip_test_if_no_ollama()
  skip_on_ci()

  result <- send_prompt(
    img_file_prompt,
    llm_provider = ollama_provider
  )

  testthat::expect_true(
    cat_mentioned(result)
  )
})

testthat::test_that("ollama provider - add_image works (plot)", {
  skip_test_if_no_ollama()

  result <- send_prompt(
    img_plot_prompt,
    llm_provider = ollama_provider
  )

  testthat::expect_true(
    cat_mentioned(result)
  )
})

#' Make LLM answer as a named list
#'
#' Get a named list from LLM response with optional item instructions and validations.
#'
#' @param prompt A single string or a [tidyprompt()] object
#' @param item_names A character vector specifying the expected item names
#' @param item_instructions An optional named list of additional instructions for each item
#' @param item_validations An optional named list of validation functions for each item.
#' Like validation functions for a [prompt_wrap()], these functions should return
#' [llm_feedback()] if the validation fails. If the validation
#' is successful, the function should return TRUE
#'
#' @return A [tidyprompt()] with an added [prompt_wrap()] that ensures
#' the LLM response is a named list with the specified item names, optional
#' instructions, and validations.
#'
#' @export
#'
#' @example inst/examples/answer_as_named_list.R
#'
#' @seealso [answer_as_list()] [llm_feedback()]
#'
#' @family pre_built_prompt_wraps
#' @family answer_as_prompt_wraps
answer_as_named_list <- function(
  prompt,
  item_names,
  item_instructions = NULL,
  item_validations = NULL
) {
  prompt <- tidyprompt(prompt)

  if (!is.character(item_names) || length(item_names) == 0) {
    stop("`item_names` must be a non-empty character vector.")
  }

  if (!is.null(item_instructions)) {
    if (
      !is.list(item_instructions) ||
        !all(names(item_instructions) %in% item_names)
    ) {
      stop(
        "`item_instructions` must be a named list with names matching `item_names`."
      )
    }
  }

  if (!is.null(item_validations)) {
    if (
      !is.list(item_validations) ||
        !all(names(item_validations) %in% item_names)
    ) {
      stop(
        "`item_validations` must be a named list with names matching `item_names`."
      )
    }
  }

  # Generate instructions for the expected format
  list_instruction <- glue::glue(
    "Respond with a named list like so:\n",
    paste(
      sapply(
        item_names,
        function(x) {
          instruction <- if (!is.null(item_instructions[[x]])) {
            glue::glue(" ({item_instructions[[x]]})")
          } else {
            ""
          }
          glue::glue("  -- {x}: <<value>>{instruction}")
        }
      ),
      collapse = "\n"
    ),
    "\n\nItem names in the list must correspond to:\n",
    "  {paste(item_names, collapse = ', ')}",
    .trim = FALSE
  )

  modify_fn <- function(original_prompt_text) {
    glue::glue(
      "{original_prompt_text}\n\n",
      "{list_instruction}"
    )
  }

  extraction_fn <- function(response) {
    if (!is.character(response)) {
      stop(
        paste0(
          "Response to extract a list from must be a string.",
          " Ensure no prompt wraps make the response unsuitable for list extraction."
        )
      )
    }

    # Use regex to extract named key-value pairs
    named_items <- stringr::str_match_all(response, "--\\s*([^:]+):\\s*(.+)")[[
      1
    ]]

    if (nrow(named_items) == 0) {
      return(
        llm_feedback(
          glue::glue(
            "Could not parse any named items from your response.\n\n",
            "{list_instruction}"
          )
        )
      )
    }

    # Extract names and values
    names <- stringr::str_trim(named_items[, 2])
    values <- stringr::str_trim(named_items[, 3])

    # Create named list
    named_list <- values
    names(named_list) <- names

    # Validate all expected names are present
    missing_names <- setdiff(item_names, names(named_list))
    if (length(missing_names) > 0) {
      return(
        llm_feedback(
          glue::glue(
            "The response is missing the following expected names: {paste(missing_names, collapse = ', ')}.\n",
            "{list_instruction}"
          )
        )
      )
    }

    # Return the named list
    named_list |> as.list()
  }

  validation_fn <- function(named_list) {
    # Apply validation functions if provided
    if (!is.null(item_validations)) {
      validation_results <- mapply(
        function(x, validation_fn) {
          validation_result <- validation_fn(x)
          if (isTRUE(validation_result)) {
            NULL # Validation passed
          } else {
            validation_result # Return validation error message or object
          }
        },
        named_list[names(item_validations)],
        item_validations,
        SIMPLIFY = FALSE
      )

      # Check if any validation failed
      failed_validations <- validation_results[
        !sapply(validation_results, is.null)
      ]
      if (length(failed_validations) > 0) {
        failed_messages <- sapply(
          names(failed_validations),
          function(name) paste0("-- ", name, ": ", failed_validations[[name]]),
          USE.NAMES = FALSE
        )
        return(
          llm_feedback(
            glue::glue(
              "Validation failed for the following items:\n",
              paste(failed_messages, collapse = "\n"),
              "\n{list_instruction}"
            )
          )
        )
      }
    }

    return(TRUE)
  }

  prompt_wrap(
    prompt,
    modify_fn,
    extraction_fn,
    validation_fn,
    name = "answer_as_named_list"
  )
}

#' Enable LLM to draft and execute R code
#'
#' This function adds a prompt wrap to a [tidyprompt()] that instructs the
#' LLM to answer the prompt with R code. There are various options to customize
#' the behavior of this prompt wrap, concerning the evaluation of the R code,
#' the packages that may be used, the objects that already exist in the R
#' session, and if the console output that should be sent back to the LLM.
#'
#' @details For the evaluation of the R code, the 'callr' package is required.
#' Please note: automatic evaluation of generated R code may be dangerous to your
#' system; you must use this function with caution.
#'
#' @param prompt A single string or a [tidyprompt()] object
#'
#' @param add_text Single string which will be added to the prompt text,
#' informing the LLM that they must code in R to answer the prompt
#'
#' @param pkgs_to_use A character vector of package names that may be used
#' in the R code that the LLM will generate. If evaluating the R code, these
#' packages will be pre-loaded in the R session
#'
#' @param objects_to_use A named list of objects that may be used in the R code
#' that the LLM will generate. If evaluating the R code, these objects will be pre-loaded
#' in the R session. The names of the list will be used as the object names in the
#' R session
#'
#' @param list_packages Logical indicating whether the LLM should be informed
#' about the packages that may be used in their R code (if TRUE, a list of the
#' loaded packages will be shown in the initial prompt)
#'
#' @param list_objects Logical indicating whether the LLM should be informed
#' about the existence of 'objects_to_use' (if TRUE, a list of the objects
#' plus their types will be shown in the initial prompt)
#'
#' @param skim_dataframes Logical indicating whether the LLM should be informed
#' about the structure of dataframes present in 'objects_to_use' (if TRUE,
#' a skim summary of each `data.frame` type object will be shown in the initial prompt).
#' This uses the function [skim_with_labels_and_levels()]
#'
#' @param evaluate_code Logical indicating whether the R code should be
#' evaluated. If TRUE, the R code will be evaluated in a separate R session
#' (using 'callr' to create an isolated R session via \link[callr]{r_session}).
#' Note that setting this to 'TRUE' means that code generated by the LLM will
#' run on your system; use this setting with caution
#'
#' @param r_session_options A list of options to pass to the \link[callr]{r_session}.
#' This can be used to customize the R session. See \link[callr]{r_session_options}
#' for the available options. If no options are provided, the default options
#' will be used but with 'system_profile' and 'user_profile' set to FALSE
#'
#' @param output_as_tool Logical indicating whether the console output of the
#' evaluated R code should be sent back to the LLM, meaning the LLM will use
#' R code as a tool to formulate a final answer to the prompt. If TRUE, the LLM
#' can decide if they can answer the prompt with the output, or if they need to modify
#' their R code. Once the LLM does not provide new R code (i.e., the prompt is being answered)
#' this prompt wrap will end (it will continue for as long as the LLM provides R code).
#' When this option is enabled, the resulting [prompt_wrap()] will be of type 'tool'.
#' If TRUE, the return mode will also always be set to 'llm_answer'
#'
#' @param return_mode Single string indicating the return mode. One of:
#' \itemize{
#'  \item 'full': Return a list with the final LLM answer, the extracted R code,
#'  and (if argument 'evaluate_code' is TRUE) the output of the R code
#'  \item 'code': Return only the extracted R code
#'  \item 'console': Return only the console output of the evaluated R code
#'  \item 'object': Return only the object produced by the evaluated R code
#'  \item 'formatted_output': Return a formatted string with the extracted R code
#'  and its console output, and a print of the last object (this is identical to how it would
#'  be presented to the LLM if 'output_as_tool' is TRUE)
#'  \item 'llm_answer': Return only the final LLM answer
#'  }
#'  When choosing 'console' or 'object', an additional instruction will be added to
#'  the prompt text to inform the LLM about the expected output of the R code.
#'  If 'output_as_tool' is TRUE, the return mode will always be set to 'llm_answer'
#'  (as the LLM will be using the R code as a tool to answer the prompt)
#'
#' @return A [tidyprompt()] object with the [prompt_wrap()] added to it, which
#' will handle R code generation and possibly evaluation
#'
#' @export
#'
#' @example inst/examples/answer_using_r.R
#'
#' @seealso [answer_using_tools()]
#'
#' @family pre_built_prompt_wraps
#' @family answer_using_prompt_wraps
answer_using_r <- function(
  prompt,
  add_text = "You must code in the programming language 'R' to answer this prompt.",
  pkgs_to_use = c(),
  objects_to_use = list(),
  list_packages = TRUE,
  list_objects = TRUE,
  skim_dataframes = TRUE,
  evaluate_code = FALSE,
  r_session_options = list(),
  output_as_tool = FALSE,
  return_mode = c(
    "full",
    "code",
    "console",
    "object",
    "formatted_output",
    "llm_answer"
  )
) {
  ## Validate settings

  return_mode <- match.arg(return_mode)

  stopifnot(
    is.character(add_text),
    length(add_text) == 1,
    length(pkgs_to_use) == 0 ||
      is.vector(pkgs_to_use) & all(sapply(pkgs_to_use, is.character)),
    is.list(objects_to_use),
    length(objects_to_use) == 0 || !is.null(names(objects_to_use)),
    is.logical(list_packages),
    is.logical(list_objects),
    is.logical(skim_dataframes),
    is.logical(evaluate_code),
    is.logical(output_as_tool),
    is.list(r_session_options)
  )

  if (evaluate_code && !requireNamespace("callr", quietly = TRUE)) {
    stop("The 'callr' package is required to evaluate R code.")
  }
  if (!evaluate_code && output_as_tool) {
    output_as_tool <- FALSE
  }
  if (output_as_tool) {
    return_mode <- "llm_answer"
  }
  if (
    !evaluate_code &&
      return_mode %in% c("console", "object", "formatted_output")
  ) {
    stop(
      "The return mode must be 'full', 'code', or 'llm_answer' if 'evaluate_code' is FALSE"
    )
  }

  ## Validate evaluation_session & load packages

  if (evaluate_code) {
    if (length(r_session_options) == 0) {
      r_session_options <- callr::r_session_options()
      r_session_options$system_profile <- FALSE
      r_session_options$user_profile <- FALSE
    }
    evaluation_session <- callr::r_session$new(options = r_session_options)

    # Check if packages are installed using requireNamespace
    installed_pkgs <- evaluation_session$run(
      function(pkgs_to_use) {
        # Check if each package is installed using requireNamespace and return as a named list
        sapply(
          pkgs_to_use,
          function(pkg) {
            requireNamespace(pkg, quietly = TRUE)
          },
          simplify = TRUE,
          USE.NAMES = TRUE
        )
      },
      args = list(pkgs_to_use = pkgs_to_use)
    )

    if (!all(installed_pkgs)) {
      stop(
        paste0(
          "The following packages are not installed: ",
          names(installed_pkgs)[!installed_pkgs]
        )
      )
    }

    # Load the packages
    loaded_pkgs <- evaluation_session$run(
      function(pkgs_to_use) {
        for (pkg_name in pkgs_to_use) {
          library(pkg_name, character.only = TRUE)
        }
        session_info <- utils::sessionInfo()
        loaded_pkgs <- names(session_info$otherPkgs)
        loaded_pkgs
      },
      args = list(pkgs_to_use = pkgs_to_use)
    )

    # Load the objects
    loaded_objects <- evaluation_session$run(
      function(objects_to_use) {
        for (i in seq_along(objects_to_use)) {
          obj <- objects_to_use[[i]]
          obj_name <- names(objects_to_use)[i]

          # Assign to the global environment of the r_session
          assign(obj_name, obj, envir = parent.env(environment()))
        }

        # List objects in the global environment
        ls(envir = .GlobalEnv)
      },
      args = list(objects_to_use = objects_to_use)
    )
    loaded_objects

    if (!all(loaded_objects %in% names(objects_to_use))) {
      stop(
        paste0(
          "The following objects could not be loaded: ",
          names(objects_to_use)[
            !(names(objects_to_use) %in% names(loaded_objects))
          ]
        )
      )
    }
  } else {
    loaded_pkgs <- pkgs_to_use
    loaded_objects <- names(objects_to_use)
  }

  ## Define modify_fn which will add information about the setting
  ##   in which R code can be generated

  modify_fn <- function(original_prompt_text) {
    new_text <- glue::glue(
      "{original_prompt_text}\n\n",
      "{add_text}"
    )

    if (list_packages && length(loaded_pkgs) > 0) {
      new_text <- glue::glue(
        "{new_text}\n",
        "You can use functions from these packages: ",
        "{loaded_pkgs |> paste(collapse = ', ')}."
      )
    }

    new_text <- glue::glue(
      "{new_text}\n",
      "You may not install or load any additional packages."
    )

    if (list_objects) {
      object_types <- sapply(objects_to_use, function(obj) class(obj))
      objects_df <- data.frame(
        Object_name = names(objects_to_use),
        Type = object_types
      )

      if (nrow(objects_df) > 0) {
        new_text <- glue::glue(
          "{new_text}\n",
          "These objects already exist in the R session:\n\n",
          "{objects_df |> df_to_string()}.\n\n",
          "Do not define these objects in your R code."
        )

        if (skim_dataframes) {
          dataframes <- objects_df$Object_name[objects_df$Type == "data.frame"]
          if (length(dataframes) > 0) {
            if (!requireNamespace("skimr", quietly = TRUE)) {
              warning(
                paste0(
                  "The 'skimr' package is required to skim dataframes.",
                  " Skim summary of dataframes currently not shown in prompt"
                )
              )
            } else {
              for (df_name in dataframes) {
                df <- objects_to_use[[df_name]]
                new_text <- glue::glue(
                  "{new_text}\n\n",
                  "Summary of the dataframe '{df_name}':\n",
                  "{df |> skim_with_labels_and_levels() |> df_to_string()}\n\n"
                )
              }
            }
          }
        }

        if (output_as_tool) {
          new_text <- glue::glue(
            "{new_text}\n",
            "If you need more information about these objects,",
            " you can call R functions to describe them."
          )
        }
      }
    }

    if (evaluate_code && return_mode == "console") {
      new_text <- glue::glue(
        "{new_text}\n",
        "The R code should produce console output that answers the prompt."
      )
    }
    if (evaluate_code && return_mode == "object") {
      new_text <- glue::glue(
        "{new_text}\n",
        "The R code should produce an object that answers the prompt."
      )
    }

    if (output_as_tool) {
      new_text <- glue::glue(
        "{new_text}\n",
        "The console output of your R code will be sent back to you.",
        " Use print() on all objects or values that you need to see.",
        " You can not view plots, all output must be text-based.",
        " After you get console output from me, decide if you can answer the prompt or if",
        " you need to modify your R code. When you can formulate your final answer,",
        " do not provide any R code in it."
      )
    }

    return(new_text)
  }

  ## Define extraction_fn which will extract R code from the response
  ##   and handle it according to the settings of this function

  extraction_fn <- function(x) {
    return_list <- list()
    return_list$llm_answer <- x

    extracted_code <- answer_using_r_extract_r_code(x)

    if (length(extracted_code) == 0) {
      if (output_as_tool) {
        return(x)
      }

      return(
        llm_feedback(
          paste0(
            "No R code detected. You must provide R code",
            " between ```r and ```."
          )
        )
      )
    }

    # Check if the R code is valid
    parsed_code <- tryCatch(parse(text = extracted_code), error = function(e) e)
    if (inherits(parsed_code, "error")) {
      return(
        llm_feedback(
          glue::glue(
            "Invalid R code detected:\n",
            "    {parsed_code$message}\n",
            "Please provide syntactically correct R code."
          )
        )
      )
    }
    return_list$code <- parsed_code

    if (!evaluate_code) {
      if (return_mode == "code") {
        return(parsed_code)
      }
      if (return_mode == "llm_answer") {
        return(x)
      }
      return(return_list)
    }

    clone_session <- evaluation_session$clone() # Reset the session every time
    output <- clone_session$run_with_output(
      function(r_code) {
        eval(parse(text = r_code))
      },
      args = list(parsed_code)
    )

    # Check if errors occurred during execution
    if (!is.null(output$error)) {
      return(
        llm_feedback(
          glue::glue(
            "An error occurred while executing the R code:\n",
            "    {output$error}"
          )
        )
      )
    }

    # Check if the code produced any relevant output
    if (output$stdout == "" && return_mode == "console") {
      return(
        llm_feedback(
          glue::glue(
            "The R code did not produce any console output.",
            " Please provide R code that produces console output."
          )
        )
      )
    }
    if (is.null(output$result) && return_mode == "object") {
      return(
        llm_feedback(
          glue::glue(
            "The R code did not produce an object.",
            " Please provide R code that produces an object."
          )
        )
      )
    }
    if (is.null(output$stdout) && is.null(output$result)) {
      return(
        llm_feedback(
          glue::glue(
            "The R code did not produce any output.",
            " Please provide R code that produces output."
          )
        )
      )
    }

    return_list$output <- output
    return_list$formatted_output <- glue::glue(
      "--- R code: ---\n",
      "{extracted_code |> paste(collapse = \"\\n\")}\n\n",
      "--- Console output: ---\n",
      "{
          if (is.null(output$stdout) || output$stdout == \"\") {
            \"No console output produced.\"
          } else {
            output$stdout |>
            paste(collapse = \"\\n\") |>
            stringr::str_trunc(1000)
          }
        }\n\n",
      "--- Last object: ---\n",
      "{
          if (is.null(output$result)) {
            \"No object produced.\"
          } else {
            utils::capture.output(print(output$result)) |>
            paste(collapse = \"\\n\") |>
            stringr::str_trunc(100)
          }
        }"
    )

    if (output_as_tool) {
      return(llm_feedback(return_list$formatted_output, tool_result = TRUE))
    }

    if (return_mode == "full") {
      return(return_list)
    }
    if (return_mode == "code") {
      return(return_list$code)
    }
    if (return_mode == "console") {
      return(return_list$output$stdout)
    }
    if (return_mode == "object") {
      return(return_list$output$result)
    }
    if (return_mode == "formatted_output") {
      return(return_list$formatted_output)
    }
    if (return_mode == "llm_answer") {
      return(x)
    }

    return(output$stdout)
  }

  ## If we are sending back output, we can consider this wrapper a tool

  type <- "unspecified"
  if (output_as_tool) {
    type <- "tool"
  }

  ## Finally, wrap the prompt with the new prompt wrap

  prompt_wrap(
    prompt,
    modify_fn,
    extraction_fn,
    type = type,
    name = "answer_using_r"
  )
}

#' Helper function to extract R code from a string
#'
#' This function extracts R code from a string by matching all content between
#' '```r' and '```'.
#'
#' @param input_string A string containing R code, typically a response
#' from an LLM
#'
#' @return A character vector containing the extracted R code
#'
#' @noRd
answer_using_r_extract_r_code <- function(input_string) {
  # Use regular expression to match all content between ```r and ```, with case-insensitive matching
  matches <- gregexpr("(?s)```[rR]\\s*(.*?)\\s*```", input_string, perl = TRUE)
  extracted_code <- regmatches(input_string, matches)

  # Remove the ```r and ``` wrappers (works for both lowercase and uppercase R)
  extracted_code <- lapply(extracted_code, function(x) {
    sub("(?s)```[rR]\\s*(.*?)\\s*```", "\\1", x, perl = TRUE)
  })

  return(unlist(extracted_code))
}

test_that("answer using r can create linear model with data", {
  skip_test_if_no_ollama()

  # Prompt to linear model object in R
  model <- paste0(
    "Using my data, create a statistical model",
    " investigating the relationship between two variables."
  ) |>
    answer_using_r(
      objects_to_use = list(data = mtcars),
      evaluate_code = TRUE,
      return_mode = "object"
    ) |>
    prompt_wrap(
      validation_fn = function(x) {
        if (!inherits(x, "lm")) {
          return(llm_feedback("The output should be a linear model object."))
        }
        return(x)
      }
    ) |>
    send_prompt(llm_provider_ollama(), verbose = FALSE)

  # Check if the model is a linear model object
  expect_true(inherits(model, "lm"))
})

test_that("answer_using_r handles complex objects in formatted_output", {
  # Test that complex objects (like lists, functions, etc.) can be handled
  # in the formatted_output without causing errors

  # This test doesn't require an LLM provider since we're testing the
  # extraction function directly
  skip_if_not_installed("callr")

  # Create a tidyprompt with answer_using_r wrapper
  tp <- "test" |>
    answer_using_r(
      evaluate_code = TRUE,
      return_mode = "formatted_output"
    )

  # Get the extraction function
  wraps <- get_prompt_wraps(tp)
  extraction_fn <- wraps[[1]]$extraction_fn

  # Test with R code that produces a complex object (list)
  mock_response <- '```r
complex_obj <- list(
  data = mtcars[1:3, 1:3],
  func = function(x) x + 1,
  plot_like = structure(list(x = 1:3, y = 4:6), class = "custom_plot")
)
complex_obj
```'

  # This should not throw an error
  expect_no_error({
    result <- extraction_fn(mock_response)
  })

  # The result should be a character string (formatted output)
  expect_type(result, "character")
  expect_true(length(result) == 1)

  # Should contain the expected sections
  expect_true(grepl("--- R code: ---", result, fixed = TRUE))
  expect_true(grepl("--- Console output: ---", result, fixed = TRUE))
  expect_true(grepl("--- Last object: ---", result, fixed = TRUE))
})

fake_ellmer_chat <- function(turns = list()) {
  env <- new.env(parent = emptyenv())
  env$turns <- turns
  env$last_method <- NULL
  env$set_turns_calls <- list()

  env$set_turns <- function(value) {
    env$set_turns_calls[[length(env$set_turns_calls) + 1]] <- value
    env$turns <- value
    env
  }
  env$get_turns <- function() env$turns
  env$get_model <- function() "fake-model"
  env$._tools <- list()
  env$register_tool <- function(tool) {
    env$._tools[[length(env$._tools) + 1L]] <- tool
    invisible(NULL)
  }
  env$set_tools <- function(tools) {
    env$._tools <- tools
    invisible(NULL)
  }
  env$get_tools <- function() env$._tools

  env$clone <- function() {
    copy <- fake_ellmer_chat(turns = env$turns)
    copy$turns <- env$turns
    copy$last_method <- env$last_method
    copy$set_turns_calls <- env$set_turns_calls
    copy$._tools <- env$._tools
    copy
  }

  env$chat <- function(...) {
    args <- list(...)
    env$last_method <- list(
      method = "chat",
      args = args,
      turns = env$turns
    )
    paste0(
      "chat-response:",
      paste(
        vapply(
          args,
          function(a) {
            tryCatch(
              paste(as.character(a), collapse = ","),
              error = function(e) class(a)[1]
            )
          },
          character(1)
        ),
        collapse = ""
      )
    )
  }

  env$chat_structured <- function(..., type) {
    args <- list(...)
    env$last_method <- list(
      method = "chat_structured",
      args = args,
      type = type,
      turns = env$turns
    )
    list(result = "ok")
  }

  if (requireNamespace("coro", quietly = TRUE)) {
    env$stream <- function(...) {
      args <- list(...)
      env$last_method <- list(
        method = "stream",
        args = args,
        turns = env$turns
      )
      coro::generator(function() {
        coro::yield("chunk")
        coro::yield("-end")
      })()
    }
  } else {
    env$stream <- NULL
  }

  env
}

# Helper: skip tests requiring ellmer >= 0.4.0 S7 turn constructors
skip_if_no_ellmer_turn_classes <- function() {
  testthat::skip_if_not_installed("ellmer")
  ns <- asNamespace("ellmer")
  if (!exists("UserTurn", envir = ns, inherits = FALSE)) {
    testthat::skip("ellmer >= 0.4.0 required (UserTurn not available)")
  }
}

# Tests for persistent_chat ellmer state sync

test_that("persistent_chat syncs ellmer_chat after each turn", {
  skip_if_not_installed("ellmer")

  call_count <- 0L

  fake_chat <- new.env(parent = emptyenv())
  fake_chat$chat <- function(...) {
    call_count <<- call_count + 1L
    paste0("reply-", call_count)
  }
  fake_chat$clone <- function() {
    copy <- new.env(parent = emptyenv())
    copy$chat <- fake_chat$chat
    copy$clone <- fake_chat$clone
    copy$set_turns <- fake_chat$set_turns
    copy$get_turns <- fake_chat$get_turns
    copy$stream <- NULL
    copy$.synced_marker <- call_count
    copy
  }
  fake_chat$set_turns <- function(turns) {
    fake_chat
  }
  fake_chat$get_turns <- function() list()
  fake_chat$stream <- NULL
  fake_chat$.synced_marker <- 0L

  provider <- llm_provider_ellmer(fake_chat, verbose = FALSE)
  provider$parameters$stream <- FALSE

  pc <- `persistent_chat-class`$new(
    llm_provider = provider,
    chat_history = NULL
  )

  # First turn
  res1 <- pc$chat("hello", verbose = FALSE)
  # The ellmer_chat from the response should now be stored on the provider.
  expect_false(is.null(res1$ellmer_chat))
  # The provider's ellmer_chat should have been updated to the returned one.
  expect_identical(pc$llm_provider$ellmer_chat, res1$ellmer_chat)

  # Second turn
  res2 <- pc$chat("world", verbose = FALSE)
  expect_identical(pc$llm_provider$ellmer_chat, res2$ellmer_chat)
  # Chat history should span both turns (system included or not, at least
  # user + assistant for each).
  expect_true(nrow(pc$chat_history) >= 4L)
})

test_that("persistent_chat replays native tool results on follow-up turns", {
  skip_if_no_ellmer_turn_classes()

  fake_chat <- fake_ellmer_chat()
  fake_chat$chat <- function(...) {
    args <- list(...)
    fake_chat$last_method <- list(
      method = "chat",
      args = args,
      turns = fake_chat$turns
    )

    prompt <- paste(vapply(args, as.character, character(1)), collapse = "")

    if (identical(prompt, "call tool")) {
      request <- ellmer::ContentToolRequest(
        id = "call-1",
        name = "get_secret_number",
        arguments = list(input = 123)
      )

      fake_chat$turns <- c(
        fake_chat$turns,
        list(
          ellmer::UserTurn(list(ellmer::ContentText("call tool"))),
          ellmer::AssistantTurn(list(request)),
          ellmer::UserTurn(list(
            ellmer::ContentToolResult(value = "42", request = request)
          )),
          ellmer::AssistantTurn(list(ellmer::ContentText("The result is 42.")))
        )
      )

      return("The result is 42.")
    }

    fake_chat$turns <- c(
      fake_chat$turns,
      list(
        ellmer::UserTurn(list(ellmer::ContentText(prompt))),
        ellmer::AssistantTurn(list(ellmer::ContentText("follow-up reply")))
      )
    )

    "follow-up reply"
  }

  provider <- llm_provider_ellmer(
    fake_chat,
    parameters = list(stream = FALSE),
    verbose = FALSE
  )

  pc <- `persistent_chat-class`$new(
    llm_provider = provider,
    chat_history = NULL
  )

  invisible(pc$chat("call tool", verbose = FALSE))
  res <- pc$chat("what was the tool result?", verbose = FALSE)

  prior_turns <- res$ellmer_chat$last_method$turns

  expect_length(prior_turns, 3)
  expect_equal(
    vapply(prior_turns, function(turn) turn@role, character(1)),
    c(
      "user",
      "user",
      "assistant"
    )
  )
  expect_true(any(grepl(
    "ContentToolResult",
    class(prior_turns[[2]]@contents[[1]])
  )))
  expect_equal(prior_turns[[3]]@contents[[1]]@text, "The result is 42.")
})

# R/helper_ellmer_tool_compatability.R

# Lazy cache for ellmer ToolDef class signature.
# Computed on first access so ellmer (a Suggests dependency) does not need
# to be available at tidyprompt load time.
.ellmer_tool_sig_env <- new.env(parent = emptyenv())
.ellmer_tool_sig_env$tooldef <- NULL
.ellmer_tool_sig_env$tooldef_resolved <- FALSE
.ellmer_tool_sig_env$builtin <- NULL
.ellmer_tool_sig_env$builtin_resolved <- FALSE

.get_ellmer_tooldef_sig <- function() {
  if (.ellmer_tool_sig_env$tooldef_resolved) {
    return(.ellmer_tool_sig_env$tooldef)
  }
  if (!ellmer_available()) {
    return(NULL)
  }
  sig <- tryCatch(
    {
      dummy_fun <- function(x = 1) x
      td <- ellmer::tool(
        dummy_fun,
        description = "dummy",
        arguments = list(x = ellmer::type_number())
      )
      class(td)
    },
    error = function(e) NULL
  )
  .ellmer_tool_sig_env$tooldef_resolved <- TRUE
  .ellmer_tool_sig_env$tooldef <- sig
  sig
}

.get_ellmer_builtin_sig <- function() {
  if (.ellmer_tool_sig_env$builtin_resolved) {
    return(.ellmer_tool_sig_env$builtin)
  }
  if (!ellmer_available()) {
    return(NULL)
  }
  sig <- tryCatch(
    {
      ellmer_ns <- asNamespace("ellmer")
      if (exists("ToolBuiltIn", envir = ellmer_ns, inherits = FALSE)) {
        tbi <- ellmer_ns$ToolBuiltIn(name = "dummy")
        class(tbi)
      }
    },
    error = function(e) NULL
  )
  .ellmer_tool_sig_env$builtin_resolved <- TRUE
  .ellmer_tool_sig_env$builtin <- sig
  sig
}

is_ellmer_tool <- function(x) {
  sig <- .get_ellmer_tooldef_sig()
  if (is.null(sig)) {
    return(FALSE)
  }
  has_all_classes(x, sig)
}

is_ellmer_builtin_tool <- function(x) {
  sig <- .get_ellmer_builtin_sig()
  if (is.null(sig)) {
    return(FALSE)
  }
  has_all_classes(x, sig)
}

is_ellmer_any_tool <- function(x) {
  is_ellmer_tool(x) || is_ellmer_builtin_tool(x)
}

ellmer_tool_name <- function(tool, fallback = NULL) {
  props <- tryCatch(
    S7::props(tool),
    error = function(e) NULL
  )
  name <- props$name %||%
    attr(tool, "name", exact = TRUE) %||%
    fallback

  if (is.null(name)) {
    return(NULL)
  }

  as.character(name)
}

rename_ellmer_tool <- function(tooldef, name) {
  stopifnot(is_ellmer_tool(tooldef))

  name <- as.character(name)
  if (!length(name) || !nzchar(name)) {
    stop("`name` must be a non-empty string.")
  }

  current_name <- ellmer_tool_name(tooldef)
  if (identical(current_name, name)) {
    return(tooldef)
  }

  props <- tryCatch(
    S7::props(tooldef),
    error = function(e) list()
  )

  ellmer::tool(
    tooldef,
    name = name,
    description = props$description %||%
      attr(tooldef, "description", exact = TRUE) %||%
      "",
    arguments = .ellmer_tool_properties(tooldef),
    convert = props$convert %||%
      attr(tooldef, "convert", exact = TRUE) %||%
      TRUE,
    annotations = props$annotations %||%
      attr(tooldef, "annotations", exact = TRUE) %||%
      list()
  )
}

# Internal: pull the list of <Type>s from a ToolDef's argument object
.ellmer_tool_properties <- function(tooldef) {
  # Prefer S7 slot if available, then fall back to attributes
  props <- tryCatch(tooldef@arguments@properties, error = function(e) NULL)
  if (!is.null(props)) {
    return(props)
  }

  # Fallback: robust attribute-based extraction
  at <- attributes(tooldef@arguments) %||% list()
  reserved <- c(
    ".additional_properties",
    "additional_properties",
    "class",
    "description",
    "required",
    "properties"
  )
  prop_names <- setdiff(names(at), reserved)
  props <- lapply(prop_names, function(nm) at[[nm]])
  names(props) <- prop_names
  props
}

.ellmer_tool_formals <- function(tooldef) {
  fmls <- tryCatch(formals(tooldef), error = function(e) NULL)
  if (!is.null(fmls) && length(fmls)) {
    return(fmls)
  }

  arg_names <- names(.ellmer_tool_properties(tooldef))
  if (!length(arg_names)) {
    return(alist(... = ))
  }
  blanks <- rep(list(quote(expr = )), length(arg_names))
  names(blanks) <- arg_names
  as.pairlist(blanks)
}

.tidyprompt_arg_is_ignored <- function(arg) {
  if (is.null(arg)) {
    return(FALSE)
  }

  if (is.character(arg) && length(arg) == 1) {
    return(identical(arg, "ignore"))
  }

  if (!is.list(arg)) {
    return(FALSE)
  }

  is.character(arg$type) &&
    length(arg$type) == 1 &&
    identical(arg$type, "ignore")
}

# ---- JSON Schema -> tidyprompt docs (argument shape) -----------------------

# Return just the "type descriptor" used by tidyprompt docs:
# - atomic types are strings: "string", "integer", "numeric", "logical"
# - arrays become "vector <type>" where possible
# - objects become a *named list* of child type descriptors
.json_schema_to_tidyprompt_type_only <- function(s) {
  if (is.null(s)) {
    return("unknown")
  }

  if (isTRUE(s$`x-tidyprompt-ignore`)) {
    return("ignore")
  }

  # enums can't be expressed in nested named-lists in tidyprompt's type mini-DSL
  if (!is.null(s$enum)) {
    return("string")
  }

  t <- s$type %||% NULL
  if (identical(t, "string")) {
    return("string")
  }
  if (identical(t, "integer")) {
    return("integer")
  }
  if (identical(t, "number")) {
    return("numeric")
  }
  if (identical(t, "boolean")) {
    return("logical")
  }

  if (identical(t, "array")) {
    items <- s$items %||% list()
    it <- items$type %||% NULL
    if (identical(it, "integer")) {
      return("vector integer")
    }
    if (identical(it, "number")) {
      return("vector numeric")
    }
    if (identical(it, "boolean")) {
      return("vector logical")
    }
    if (identical(it, "string")) {
      return("vector string")
    }
    return("vector unknown")
  }

  if (identical(t, "object") || !is.null(s$properties)) {
    props <- s$properties %||% list()
    out <- lapply(props, .json_schema_to_tidyprompt_type_only)
    # Preserve names; nested enums degrade to "string" (see above)
    return(out)
  }

  "unknown"
}

# Turn a per-argument JSON Schema into a tidyprompt docs entry:
# list(type=..., [default_value=... for match.arg], [description=...])
.json_schema_to_tidyprompt_arg <- function(s) {
  res <- list()

  if (!is.null(s$enum)) {
    res$type <- "match.arg"
    res$default_value <- s$enum
    if (!is.null(s$description)) {
      res$description <- s$description
    }
    return(res)
  }

  res$type <- .json_schema_to_tidyprompt_type_only(s)
  if (!is.null(s$description)) {
    res$description <- s$description
  }
  res
}

# ---- ellmer ToolDef -> tidyprompt docs / function --------------------------

# Build a tidyprompt "docs" list from an ellmer ToolDef
ellmer_tool_to_tidyprompt_docs <- function(tooldef) {
  stopifnot(is_ellmer_tool(tooldef))

  # Name + description are S7 properties
  name <- tryCatch(tooldef@name, error = function(e) NULL)
  desc <- tryCatch(tooldef@description, error = function(e) NULL)

  props <- .ellmer_tool_properties(tooldef)
  arg_names <- names(.ellmer_tool_formals(tooldef))
  if (!length(arg_names)) {
    arg_names <- names(props)
  }

  args_docs <- list()
  if (length(arg_names)) {
    for (nm in arg_names) {
      if (!nm %in% names(props)) {
        args_docs[[nm]] <- list(type = "ignore")
        next
      }

      # Best-effort: use your ellmer -> JSON Schema converter
      s <- tryCatch(
        ellmer_type_to_json_schema(props[[nm]], strict = TRUE),
        error = function(e) NULL
      )
      if (is.null(s)) {
        args_docs[[nm]] <- list(type = "unknown")
      } else {
        args_docs[[nm]] <- .json_schema_to_tidyprompt_arg(s)
      }
    }
  }

  compact_list(list(
    name = name %||% "tool",
    description = desc %||% "",
    arguments = args_docs,
    'return' = list() # we can't infer reliably; leave empty
  ))
}

# Create a plain R function wrapper around an ellmer ToolDef and attach docs
# so tidyprompt can use it directly.
ellmer_tool_to_tidyprompt <- function(tooldef) {
  stopifnot(is_ellmer_tool(tooldef))

  docs <- ellmer_tool_to_tidyprompt_docs(tooldef)

  wrapper <- function() {}
  formals(wrapper) <- .ellmer_tool_formals(tooldef)
  body(wrapper) <- quote({
    tool <- attr(sys.function(), "ellmer_tool", exact = TRUE)
    args <- as.list(match.call(expand.dots = TRUE))[-1]
    do.call(tool, args, quote = TRUE)
  })
  # Do NOT change environment(wrapper); just attach the ToolDef
  attr(wrapper, "ellmer_tool") <- tooldef

  tools_add_docs(wrapper, docs)
}

# ---- tidyprompt docs/function -> ellmer ToolDef ----------------------------

# Given tidyprompt docs + function, build an ellmer ToolDef.
# Uses: tools_docs_to_r_json_schema() -> json_schema_to_ellmer_type() -> properties -> tool()
tidyprompt_docs_to_ellmer_tool <- function(
  fun,
  docs,
  name = docs$name %||% NULL,
  convert = TRUE,
  annotations = list(),
  strict = TRUE
) {
  stopifnot(is.function(fun), is.list(docs))

  if (!ellmer_available()) {
    stop(
      "ellmer is not installed; cannot convert tidyprompt docs to ellmer tool."
    )
  }

  # Build JSON Schema for the argument object using your helper.
  # Respect which formals have defaults: only formals without defaults are required.
  fn_formals <- formals(fun)
  has_default <- vapply(
    fn_formals,
    function(x) !identical(x, quote(expr = )),
    logical(1)
  )
  required_args <- names(fn_formals)[!has_default]

  js <- tools_docs_to_r_json_schema(
    docs,
    all_required = FALSE,
    additional_properties = FALSE
  )
  # Override the required list based on formals analysis
  if (length(required_args) > 0) {
    js$required <- intersect(required_args, names(js$properties %||% list()))
  } else {
    js$required <- NULL
  }

  # Convert to a single ellmer Type (object)
  etype <- json_schema_to_ellmer_type(
    schema = js,
    required = TRUE,
    strict = isTRUE(strict)
  )

  # Extract per-argument Type list to feed to tool(arguments=)
  props <- attr(etype, "properties", exact = TRUE)
  if (is.null(props)) {
    # Attribute-based fallback (robust across ellmer versions)
    at <- attributes(etype) %||% list()
    reserved <- c(
      ".additional_properties",
      "additional_properties",
      "class",
      "description",
      "required",
      "properties"
    )
    prop_names <- setdiff(names(at), reserved)
    props <- lapply(prop_names, function(nm) at[[nm]])
    names(props) <- prop_names
  }

  # Ensure argument names match function formals (ellmer::tool() checks this)
  fn_formals <- names(formals(fun))
  # Fill any missing with permissive strings; drop extras
  missing <- setdiff(fn_formals, names(props))
  for (nm in missing) {
    arg_doc <- docs$arguments[[nm]] %||% NULL
    if (
      .tidyprompt_arg_is_ignored(arg_doc) &&
        exists("type_ignore", envir = asNamespace("ellmer"), inherits = FALSE)
    ) {
      props[[nm]] <- ellmer_type_ignore_compat(
        description = arg_doc$description %||% NULL,
        required = FALSE
      )
    } else {
      props[[nm]] <- ellmer::type_string(required = FALSE)
    }
  }
  props <- props[fn_formals]

  ellmer::tool(
    fun,
    name = name,
    description = docs$description %||% "",
    arguments = props,
    convert = convert,
    annotations = annotations
  )
}

# Convenience: take a tidyprompt-style tool function and return an ellmer ToolDef
tidyprompt_tool_to_ellmer <- function(
  fun,
  name = NULL,
  convert = TRUE,
  annotations = list(),
  strict = TRUE
) {
  stopifnot(is.function(fun))
  docs <- tools_get_docs(fun, name = name %||% NULL)
  tidyprompt_docs_to_ellmer_tool(
    fun,
    docs,
    name = name %||% docs$name %||% NULL,
    convert = convert,
    annotations = annotations,
    strict = strict
  )
}

# ---- Public: normalize a tool for a given target ---------------------------

# Returns a list with both representations when possible:
# $tidyprompt_tool (function with docs) and $ellmer_tool (ToolDef)
normalize_tool_dual <- function(
  tool,
  annotations = list(),
  convert = TRUE,
  strict = TRUE
) {
  if (is.null(tool)) {
    return(list(tidyprompt_tool = NULL, ellmer_tool = NULL))
  }

  if (is_ellmer_tool(tool)) {
    # From ellmer ToolDef -> tidyprompt
    tp_fn <- ellmer_tool_to_tidyprompt(tool)
    return(list(tidyprompt_tool = tp_fn, ellmer_tool = tool))
  }

  if (is_ellmer_builtin_tool(tool)) {
    # ToolBuiltIn can only be used in native ellmer mode; no tidyprompt equivalent
    return(list(tidyprompt_tool = NULL, ellmer_tool = tool))
  }

  if (is.function(tool)) {
    # From tidyprompt -> ellmer
    ell_tool <- tryCatch(
      tidyprompt_tool_to_ellmer(
        tool,
        convert = convert,
        annotations = annotations,
        strict = strict
      ),
      error = function(e) NULL
    )
    return(list(tidyprompt_tool = tool, ellmer_tool = ell_tool))
  }

  stop(
    "`tool` must be a function, ellmer ToolDef, or ellmer ToolBuiltIn."
  )
}

testthat::test_that("tidyprompt_docs_to_ellmer_tool respects optional args", {
  testthat::skip_if_not_installed("ellmer")

  my_tool <- function(x, y = 10) x + y
  my_tool <- tools_add_docs(
    my_tool,
    list(
      name = "add",
      description = "Add two numbers",
      arguments = list(
        x = list(type = "numeric", description = "First number"),
        y = list(type = "numeric", description = "Second number (optional)")
      )
    )
  )

  docs <- tools_get_docs(my_tool)
  td <- tidyprompt_docs_to_ellmer_tool(my_tool, docs)

  testthat::expect_true(is_ellmer_tool(td))

  # x should be required, y should be optional
  props <- .ellmer_tool_properties(td)
  x_required <- attr(props[["x"]], "required", exact = TRUE)
  y_required <- attr(props[["y"]], "required", exact = TRUE)

  testthat::expect_true(isTRUE(x_required) || is.null(x_required))
  testthat::expect_false(isTRUE(y_required))
})

testthat::test_that("tidyprompt_docs_to_ellmer_tool fills missing formals", {
  testthat::skip_if_not_installed("ellmer")

  # A function with a formal not documented
  my_tool <- function(x, extra = "default") x
  my_tool <- tools_add_docs(
    my_tool,
    list(
      name = "identity",
      description = "Return x",
      arguments = list(
        x = list(type = "string", description = "input")
      )
    )
  )

  docs <- tools_get_docs(my_tool)
  td <- tidyprompt_docs_to_ellmer_tool(my_tool, docs)

  testthat::expect_true(is_ellmer_tool(td))

  # 'extra' should be present (filled) and not required
  props <- .ellmer_tool_properties(td)
  testthat::expect_true("extra" %in% names(props))
  extra_required <- attr(props[["extra"]], "required", exact = TRUE)
  testthat::expect_false(isTRUE(extra_required))
})

# tests/testthat/test-ellmer_tool_compatability.R
#
# This tests the helper functions in R/helper_ellmer_json_compatibility.R
#   for converting between ellmer ToolDef <-> tidyprompt tools (functions+docs)

# library(testthat)
# devtools::load_all()

# ---- Local helpers (mirrors your schema helpers style) ---------------------

# Recursively remove NULLs
strip_nulls <- function(x) {
  if (is.list(x)) {
    x <- lapply(x, strip_nulls)
    x <- x[!vapply(x, is.null, logical(1))]
  }
  x
}

# Extract a minimal, stable “types-only” view from tidyprompt docs$arguments
docs_types_only <- function(docs) {
  stopifnot(is.list(docs), is.list(docs$arguments))
  arg_names <- sort(names(docs$arguments))
  out <- list()
  for (nm in arg_names) {
    a <- docs$arguments[[nm]]
    # If nested named-list: recurse on names
    if (is.list(a$type)) {
      sub <- a$type
      # ensure simple types-only without descriptions
      for (sn in names(sub)) {
        s <- sub[[sn]]
        if (is.list(s)) {
          # support deeper nesting if present
          out[[nm]][[sn]] <- if (is.list(s$type)) {
            s$type
          } else {
            (s$type %||% "unknown")
          }
        } else {
          out[[nm]][[sn]] <- s %||% "unknown"
        }
      }
    } else if (!is.null(a$enum) || identical(a$type, "match.arg")) {
      out[[nm]] <- list(
        type = "match.arg",
        values = sort(as.character(a$default_value %||% character()))
      )
    } else {
      out[[nm]] <- a$type %||% "unknown"
    }
  }
  strip_nulls(out)
}

expect_docs_types_equal <- function(a, b) {
  testthat::expect_equal(docs_types_only(a), docs_types_only(b))
}

# Extract per-argument JSON Schemas from an ellmer ToolDef
ellmer_tool_prop_schemas <- function(tooldef, strict = TRUE) {
  props <- tryCatch(tooldef@arguments@properties, error = function(e) NULL)
  if (is.null(props)) {
    # attribute fallback
    at <- attributes(tooldef@arguments) %||% list()
    reserved <- c(
      ".additional_properties",
      "additional_properties",
      "class",
      "description",
      "required",
      "properties"
    )
    prop_names <- setdiff(names(at), reserved)
    props <- lapply(prop_names, function(nm) at[[nm]])
    names(props) <- prop_names
  }
  lapply(props, function(p) ellmer_type_to_json_schema(p, strict = strict))
}

# ---- Cache resilience -------------------------------------------------------

testthat::test_that("tooldef cache retries after initial unavailability", {
  # Reset the cache to simulate a fresh session
  .ellmer_tool_sig_env$tooldef <- NULL
  .ellmer_tool_sig_env$tooldef_resolved <- FALSE
  on.exit({
    .ellmer_tool_sig_env$tooldef <- NULL
    .ellmer_tool_sig_env$tooldef_resolved <- FALSE
  })

  # Stub ellmer_available to return FALSE on first call
  call_n <- 0L
  local_mocked_bindings(
    ellmer_available = function() {
      call_n <<- call_n + 1L
      call_n > 1L
    }
  )

  # First call: ellmer unavailable -> NULL, but cache should NOT be locked
  result1 <- .get_ellmer_tooldef_sig()
  testthat::expect_null(result1)
  testthat::expect_false(.ellmer_tool_sig_env$tooldef_resolved)

  # Second call: ellmer now available -> should resolve
  result2 <- .get_ellmer_tooldef_sig()
  testthat::expect_true(.ellmer_tool_sig_env$tooldef_resolved)
  testthat::expect_true(is.character(result2) && length(result2) > 0)
})

testthat::test_that("builtin cache retries after initial unavailability", {
  .ellmer_tool_sig_env$builtin <- NULL
  .ellmer_tool_sig_env$builtin_resolved <- FALSE
  on.exit({
    .ellmer_tool_sig_env$builtin <- NULL
    .ellmer_tool_sig_env$builtin_resolved <- FALSE
  })

  call_n <- 0L
  local_mocked_bindings(
    ellmer_available = function() {
      call_n <<- call_n + 1L
      call_n > 1L
    }
  )

  result1 <- .get_ellmer_builtin_sig()
  testthat::expect_null(result1)
  testthat::expect_false(.ellmer_tool_sig_env$builtin_resolved)

  result2 <- .get_ellmer_builtin_sig()
  # ToolBuiltIn may or may not exist in the installed ellmer version;
  # either way resolved should be TRUE after a successful probe
  testthat::expect_true(.ellmer_tool_sig_env$builtin_resolved)
})

# ---- Detection -------------------------------------------------------------

testthat::test_that("is_ellmer_tool detects ToolDef and not plain functions", {
  testthat::skip_if_not_installed("ellmer")

  add <- function(x, y) x + y
  td <- ellmer::tool(
    add,
    description = "Add two numbers",
    arguments = list(
      x = ellmer::type_number(),
      y = ellmer::type_number()
    )
  )

  testthat::expect_true(is_ellmer_tool(td))
  testthat::expect_false(is_ellmer_tool(add))
})

# ---- ellmer -> tidyprompt: docs -------------------------------------------

testthat::test_that("ellmer_tool_to_tidyprompt_docs maps common types correctly", {
  testthat::skip_if_not_installed("ellmer")

  f <- function(s, n, i, b, c, vs, obj) {
    paste(s, n, i, b, c[1], paste(vs, collapse = ","), obj$x)
  }

  td <- ellmer::tool(
    f,
    description = "Mixed args",
    arguments = list(
      s = ellmer::type_string("S"),
      n = ellmer::type_number(),
      i = ellmer::type_integer(),
      b = ellmer::type_boolean(),
      c = ellmer::type_enum(c("a", "b")),
      vs = ellmer::type_array(ellmer::type_string()),
      obj = ellmer::type_object(
        x = ellmer::type_string(),
        y = ellmer::type_integer(required = FALSE)
      )
    )
  )

  docs <- ellmer_tool_to_tidyprompt_docs(td)
  # Types-only assertions
  types <- docs_types_only(docs)
  testthat::expect_equal(types$s, "string")
  testthat::expect_equal(types$n, "numeric")
  testthat::expect_equal(types$i, "integer")
  testthat::expect_equal(types$b, "logical")
  testthat::expect_equal(types$vs, "vector string")
  testthat::expect_true(is.list(types$c))
  testthat::expect_equal(unname(types$c$type), "match.arg")
  testthat::expect_equal(unname(types$c$values), c("a", "b"))
  testthat::expect_true(is.list(types$obj))
  testthat::expect_equal(types$obj$x, "string")
  testthat::expect_equal(types$obj$y, "integer")
})

testthat::test_that("ellmer_tool_to_tidyprompt_docs preserves ignored args", {
  testthat::skip_if_not_installed("ellmer")
  testthat::skip_if_not(
    exists("type_ignore", envir = asNamespace("ellmer"), inherits = FALSE),
    "type_ignore not available in this ellmer version"
  )

  f <- function(x, hidden = 1) x + hidden

  td <- ellmer::tool(
    f,
    description = "Uses a hidden default",
    arguments = list(
      x = ellmer::type_number(),
      hidden = ellmer::type_ignore()
    )
  )

  docs <- ellmer_tool_to_tidyprompt_docs(td)

  testthat::expect_equal(docs$arguments$x$type, "numeric")
  testthat::expect_equal(docs$arguments$hidden$type, "ignore")
})

# ---- ellmer -> tidyprompt: function wrapper --------------------------------

testthat::test_that("ellmer_tool_to_tidyprompt creates a working wrapper with docs", {
  testthat::skip_if_not_installed("ellmer")

  add <- function(x, y) x + y
  td <- ellmer::tool(
    add,
    description = "Add two numbers",
    arguments = list(
      x = ellmer::type_number(),
      y = ellmer::type_number()
    )
  )

  tp_fun <- ellmer_tool_to_tidyprompt(td)

  # Same behavior
  testthat::expect_identical(tp_fun(2, 3), add(2, 3))

  # Docs attached & name preserved
  tp_docs <- tools_get_docs(tp_fun)
  testthat::expect_true(is.list(tp_docs))
  testthat::expect_identical(tp_docs$name, td@name)
})

# ---- tidyprompt -> ellmer: from docs + function ----------------------------

testthat::test_that("tidyprompt_docs_to_ellmer_tool builds a ToolDef that matches formals and types", {
  testthat::skip_if_not_installed("ellmer")

  get_weather_like <- function(
    city,
    units = c("metric", "imperial"),
    opts = list(detail = TRUE)
  ) {
    paste(city, units[1], if (isTRUE(opts$detail)) "(detail)" else "")
  }

  docs <- list(
    name = "get_weather_like",
    description = "Demo weather-like tool",
    arguments = list(
      city = list(type = "string", description = "City name"),
      units = list(type = "match.arg", default_value = c("metric", "imperial")),
      opts = list(type = list(detail = "logical"))
    ),
    'return' = list(description = "A short string")
  )

  td <- tidyprompt_docs_to_ellmer_tool(get_weather_like, docs, strict = TRUE)
  testthat::expect_true(is_ellmer_tool(td))

  # Names must match function formals
  testthat::expect_identical(
    names(formals(get_weather_like)),
    names(td@arguments@properties)
  )

  # Per-argument schema checks
  ps <- ellmer_tool_prop_schemas(td, strict = TRUE)
  testthat::expect_equal(ps$city$type, "string")
  testthat::expect_equal(sort(ps$units$enum), c("imperial", "metric"))

  # opts should be an object with a boolean subproperty "detail"
  testthat::expect_equal(ps$opts$type, "object")
  testthat::expect_equal(ps$opts$properties$detail$type, "boolean")

  # Callable and returns expected string
  res <- td(city = "London", units = "metric", opts = list(detail = TRUE))
  testthat::expect_match(res, "London metric")
})

# ---- tidyprompt -> ellmer: fills missing type defs to satisfy formals ------

testthat::test_that("tidyprompt_docs_to_ellmer_tool fills missing argument types", {
  testthat::skip_if_not_installed("ellmer")

  f <- function(a, b) a
  docs <- list(
    name = "f",
    description = "Returns a, ignores b",
    arguments = list(
      a = list(type = "string")
      # b intentionally missing
    )
  )

  td <- tidyprompt_docs_to_ellmer_tool(f, docs, strict = TRUE)
  testthat::expect_true(is_ellmer_tool(td))
  # Both names present and ordered as formals
  testthat::expect_identical(names(td@arguments@properties), c("a", "b"))

  # b gets a permissive string type
  ps <- ellmer_tool_prop_schemas(td, strict = TRUE)
  testthat::expect_equal(ps$b$type, "string")
})

testthat::test_that("tidyprompt_docs_to_ellmer_tool rebuilds ignored args", {
  testthat::skip_if_not_installed("ellmer")
  testthat::skip_if_not(
    exists("type_ignore", envir = asNamespace("ellmer"), inherits = FALSE),
    "type_ignore not available in this ellmer version"
  )

  f <- function(x, hidden = 1) x + hidden
  docs <- list(
    name = "f",
    description = "Uses a hidden default",
    arguments = list(
      x = list(type = "numeric"),
      hidden = list(type = "ignore", description = "Use the default")
    )
  )

  td <- tidyprompt_docs_to_ellmer_tool(f, docs, strict = TRUE)

  testthat::expect_true(is_ellmer_tool(td))
  testthat::expect_identical(names(formals(td)), c("x", "hidden"))
  testthat::expect_identical(names(.ellmer_tool_properties(td)), "x")
  testthat::expect_identical(td(x = 2), 3)
})

# ---- Round-trip behavior ---------------------------------------------------

testthat::test_that("ellmer ToolDef -> tidyprompt fn -> ellmer ToolDef keeps behavior", {
  testthat::skip_if_not_installed("ellmer")

  greet <- function(name, excited = FALSE) {
    paste0("Hello, ", name, if (isTRUE(excited)) "!")
  }

  td <- ellmer::tool(
    greet,
    description = "Greet a person",
    arguments = list(
      name = ellmer::type_string(),
      excited = ellmer::type_boolean()
    )
  )

  # Convert to tidyprompt wrapper
  tp_fun <- ellmer_tool_to_tidyprompt(td)

  # Back to ellmer ToolDef from the tidyprompt function
  td2 <- tidyprompt_tool_to_ellmer(tp_fun, strict = TRUE)

  # Same call results on both tooldefs
  r1 <- td(name = "Ada", excited = TRUE)
  r2 <- td2(name = "Ada", excited = TRUE)
  testthat::expect_identical(r1, r2)
})

testthat::test_that("ignored args stay hidden across tool roundtrips", {
  testthat::skip_if_not_installed("ellmer")
  testthat::skip_if_not(
    exists("type_ignore", envir = asNamespace("ellmer"), inherits = FALSE),
    "type_ignore not available in this ellmer version"
  )

  add_hidden <- function(x, hidden = 1) x + hidden

  td <- ellmer::tool(
    add_hidden,
    description = "Add with a hidden default",
    arguments = list(
      x = ellmer::type_number(),
      hidden = ellmer::type_ignore()
    )
  )

  tp_fun <- ellmer_tool_to_tidyprompt(td)
  tp_docs <- tools_get_docs(tp_fun)
  td2 <- tidyprompt_tool_to_ellmer(tp_fun, strict = TRUE)

  testthat::expect_equal(tp_docs$arguments$hidden$type, "ignore")
  testthat::expect_false("hidden" %in% names(.ellmer_tool_properties(td2)))
  testthat::expect_identical(td2(x = 2), 3)
})

testthat::test_that("ignored args are omitted from emitted schemas and text", {
  docs <- list(
    name = "f",
    description = "Uses a hidden default",
    arguments = list(
      x = list(type = "numeric", description = "Visible"),
      hidden = list(type = "ignore", description = "Hidden")
    )
  )

  schema <- tools_docs_to_r_json_schema(docs, all_required = FALSE)
  text <- tools_docs_to_text(docs)

  testthat::expect_identical(names(schema$properties), "x")
  testthat::expect_identical(as.character(schema$required), "x")
  testthat::expect_false(grepl("- hidden:", text, fixed = TRUE))
})

# ---- normalize_tool_dual ---------------------------------------------------

testthat::test_that("normalize_tool_dual returns both sides from ellmer input", {
  testthat::skip_if_not_installed("ellmer")

  adder <- function(x, y) x + y
  td <- ellmer::tool(
    adder,
    description = "Add",
    arguments = list(x = ellmer::type_number(), y = ellmer::type_number())
  )

  out <- normalize_tool_dual(td)
  testthat::expect_true(is.function(out$tidyprompt_tool))
  testthat::expect_true(is_ellmer_tool(out$ellmer_tool))

  # Both callable
  testthat::expect_equal(out$tidyprompt_tool(10, 5), 15)
  testthat::expect_equal(out$ellmer_tool(x = 10, y = 5), 15)
})

testthat::test_that("normalize_tool_dual returns both sides from tidyprompt function input", {
  testthat::skip_if_not_installed("ellmer")

  # a small documented function
  say <- function(text, shout = FALSE) {
    if (isTRUE(shout)) toupper(text) else text
  }
  say <- tools_add_docs(
    say,
    list(
      name = "say",
      description = "Echo text",
      arguments = list(
        text = list(type = "string"),
        shout = list(type = "logical")
      )
    )
  )

  out <- normalize_tool_dual(say)
  testthat::expect_true(is.function(out$tidyprompt_tool))
  testthat::expect_true(is_ellmer_tool(out$ellmer_tool))

  testthat::expect_identical(out$tidyprompt_tool("ok", FALSE), "ok")
  testthat::expect_identical(out$ellmer_tool(text = "ok", shout = FALSE), "ok")
})

testthat::test_that("is_ellmer_builtin_tool detects ToolBuiltIn objects", {
  testthat::skip_if_not_installed("ellmer")
  ellmer_ns <- asNamespace("ellmer")
  testthat::skip_if_not(
    exists("ToolBuiltIn", envir = ellmer_ns),
    "ToolBuiltIn not available in this ellmer version"
  )

  tbi <- ellmer_ns$ToolBuiltIn(
    name = "web_search"
  )
  testthat::expect_true(is_ellmer_builtin_tool(tbi))
  testthat::expect_true(is_ellmer_any_tool(tbi))
  testthat::expect_false(is_ellmer_tool(tbi))
})

testthat::test_that("normalize_tool_dual handles ToolBuiltIn", {
  testthat::skip_if_not_installed("ellmer")
  ellmer_ns <- asNamespace("ellmer")
  testthat::skip_if_not(
    exists("ToolBuiltIn", envir = ellmer_ns),
    "ToolBuiltIn not available in this ellmer version"
  )

  tbi <- ellmer_ns$ToolBuiltIn(
    name = "web_search"
  )
  dual <- normalize_tool_dual(tbi)

  # ToolBuiltIn has no tidyprompt equivalent
  testthat::expect_null(dual$tidyprompt_tool)
  # But should be passed through as the ellmer tool
  testthat::expect_identical(dual$ellmer_tool, tbi)
})

testthat::test_that("answer_using_tools accepts ToolBuiltIn in tool list", {
  testthat::skip_if_not_installed("ellmer")
  ellmer_ns <- asNamespace("ellmer")
  testthat::skip_if_not(
    exists("ToolBuiltIn", envir = ellmer_ns),
    "ToolBuiltIn not available in this ellmer version"
  )

  my_fun <- function(x) x
  my_fun <- tools_add_docs(
    my_fun,
    list(
      name = "echo",
      description = "Echo input",
      arguments = list(x = list(type = "string"))
    )
  )

  tbi <- ellmer_ns$ToolBuiltIn(
    name = "web_search"
  )

  # Should not error when mixed ToolDef/ToolBuiltIn/function tools are passed
  result <- testthat::expect_no_error(
    "test" |>
      answer_using_tools(tools = list(echo = my_fun, search = tbi))
  )
})

testthat::test_that("answer_using_tools errors for ToolBuiltIn on non-ellmer providers", {
  testthat::skip_if_not_installed("ellmer")
  ellmer_ns <- asNamespace("ellmer")
  testthat::skip_if_not(
    exists("ToolBuiltIn", envir = ellmer_ns),
    "ToolBuiltIn not available in this ellmer version"
  )

  tbi <- ellmer_ns$ToolBuiltIn(
    name = "web_search"
  )

  testthat::expect_error(
    "test" |>
      answer_using_tools(tools = list(search = tbi), type = "text-based"),
    "built-in tools can only be used with an ellmer-backed provider"
  )

  prompt <- "test" |>
    answer_using_tools(tools = list(search = tbi))

  testthat::expect_error(
    send_prompt(prompt, llm_provider_fake(verbose = FALSE), verbose = FALSE),
    "built-in tools can only be used with an ellmer-backed provider"
  )
})

#' @title Skip Tests if Ollama Server is Unavailable
#'
#' @description Helper function for unit tests that skips the test if the local
#' Ollama server is not running.
#'
#' @details The function sends a request to the Ollama server at
#' `http://localhost:11434`. If the server is not running or the request fails,
#'  the test is skipped with an appropriate message.
#'
#' @param do_not_run Logical indicating whether to skip the test even if the
#' Ollama server is running. Temporarily set to `TRUE` to skip the tests
#'
#' @return testthat skip object or NULL
#'
#' @noRd
#' @keywords internal
skip_test_if_no_ollama <- function(do_not_run = FALSE) {
  if (do_not_run) {
    return(
      testthat::skip("skip test: Skipping test even if Ollama is available")
    )
  }

  if (!ollama_available()) {
    return(testthat::skip("skip test: Ollama not available"))
  }

  return(invisible(NULL))
}

ollama_available <- function(
  url = Sys.getenv("OLLAMA_URL", unset = "http://localhost:11434")
) {
  req <- httr2::request(url)

  # Try performing the request and catch any errors
  ollama_running <- tryCatch(
    {
      resp <- httr2::req_perform(req)
      if (resp$status_code == 200) {
        TRUE
      } else {
        FALSE
      }
    },
    error = function(e) {
      FALSE # If there's an error, assume the server is not running
    }
  )

  return(ollama_running)
}

#' Skip tests if OpenAI model is unavailable
#'
#' Helper function for unit tests that skips the test if the OpenAI API is not available.
#' The function sends a request to the OpenAI API to check available models.
#' If there is no valid API key or the request fails, the test is skipped with an appropriate message.
#'
#' @param do_not_run Logical indicating whether to skip the test even if the
#' OpenAI API is available. Temporarily set to `TRUE` to skip the tests
#'
#' @return testthat skip object or NULL
#'
#' @noRd
#' @keywords internal
skip_test_if_no_openai <- function(do_not_run = FALSE) {
  if (do_not_run) {
    return(
      testthat::skip("skip test: Skipping test even if OpenAI is available")
    )
  }

  if (!openai_available()) {
    return(testthat::skip("skip test: OpenAI not available"))
  }

  return(invisible(NULL))
}

openai_available <- function(
  url = Sys.getenv("OPENAI_URL", unset = "https://api.openai.com"),
  api_key = Sys.getenv("OPENAI_API_KEY")
) {
  models_endpoint <- paste0(url, "/v1/models")

  if (is.null(api_key) || api_key == "") {
    return(testthat::skip("skip test: OpenAI API key not found"))
  }

  req <- httr2::request(models_endpoint) |>
    httr2::req_headers(
      "Content-Type" = "application/json",
      "Authorization" = paste("Bearer", api_key)
    )

  openai_available <- tryCatch(
    {
      resp <- httr2::req_perform(req)
      if (resp$status_code == 200) {
        TRUE
      } else {
        FALSE
      }
    },
    error = function(e) {
      FALSE
    }
  )

  return(openai_available)
}

#' Function to extract a specific element from a list
#'
#' This function is intended as a helper function for piping with output from
#' [send_prompt()] when using `return_mode = "full"`. It allows to
#' extract a specific element from the list returned by [send_prompt()], which
#' can be useful for further piping.
#'
#' @param list A list, typically the output from [send_prompt()] with `return_mode = "full"`
#' @param name_of_element A character string with the name of the element to extract from the list
#'
#' @return The extracted element from the list
#'
#' @export
#'
#' @example inst/examples/extract_from_return_list.R
#'
#' @family miscellaneous_helpers
extract_from_return_list <- function(list, name_of_element = "response") {
  if (!is.list(list)) {
    stop("Input must be a list")
  }

  if (!name_of_element %in% names(list)) {
    stop("Element not found in list")
  }

  return(list[[name_of_element]])
}

#' @title PersistentChat R6 class
#' @name persistent_chat-class
#'
#' @description
#' A class for managing a persistent chat with a large language model (LLM).
#'
#' While 'tidyprompt' is primariy focused on automatic interactions with
#' LLMs through [send_prompt()] using a [tidyprompt-class] object with
#' [prompt_wrap()], this class may be useful for having a manual conversation
#' with an LLM. (It may specifically be used to continue a chat history which was
#' returned by [send_prompt()] with `return_mode = "full"`.)
#'
#' @example inst/examples/persistent_chat.R
#'
#' @seealso [llm_provider-class] [chat_history()]
NULL

#' @rdname persistent_chat-class
#' @export
`persistent_chat-class` <- R6::R6Class(
  "PersistentChat",
  public = list(
    #' @field chat_history A [chat_history()] object
    chat_history = NULL,
    #' @field llm_provider A [llm_provider-class] object
    llm_provider = NULL,

    #' @description Initialize the PersistentChat object
    #'
    #' @param llm_provider A [llm_provider-class] object
    #' @param chat_history (optional) A [chat_history()] object
    #'
    #' @return The initialized PersistentChat object
    initialize = function(
      llm_provider,
      chat_history = NULL
    ) {
      if (
        is.list(chat_history) &&
          !is.data.frame(chat_history) &&
          "chat_history" %in% names(chat_history)
      ) {
        chat_history <- chat_history$chat_history |> chat_history()
      }

      if (!is.null(chat_history)) {
        chat_history <- chat_history(chat_history)
        self$chat_history <- chat_history
      }

      if (!inherits(llm_provider, "LlmProvider")) {
        stop("The 'llm_provider' argument must be of class 'LlmProvider'")
      }

      self$llm_provider <- llm_provider
      self
    },

    #' @description Add a message to the chat history and get a response from the LLM
    #'
    #' @param msg Message to add to the chat history
    #' @param role Role of the message
    #' @param verbose Whether to print the interaction to the console
    #'
    #' @return The response from the LLM
    chat = function(
      msg,
      role = "user",
      verbose = TRUE
    ) {
      self$chat_history <- add_msg_to_chat_history(
        self$chat_history,
        msg,
        role = role
      )
      stopifnot(
        is.character(msg),
        length(msg) == 1,
        is.logical(verbose)
      )

      llm_provider <- self$llm_provider$clone()
      llm_provider$verbose <- verbose

      response <- llm_provider$complete_chat(self$chat_history)
      self$chat_history <- response$completed

      # Sync the updated ellmer chat back so native state (tool calls,
      # thinking turns, etc.) accumulates across persistent_chat turns.
      if (!is.null(response$ellmer_chat)) {
        self$llm_provider$ellmer_chat <- response$ellmer_chat
      }

      if (verbose) {
        return(invisible(response))
      }

      return(response)
    },

    #' @description Reset the chat history
    #'
    #' @return NULL
    reset_chat_history = function() {
      self$chat_history <- NULL
    }
  )
)

#' @title
#' Wrap a prompt with functions for modification and handling the LLM response
#'
#' @description
#' This function takes a single string or a [tidyprompt-class] object and
#'  adds a new prompt wrap to it.
#'
#' A prompt wrap is a set of functions that modify the prompt text,
#'  extract a value from the LLM response, and validate the extracted value.
#'
#' The functions are used to ensure that the prompt and LLM response are in the
#'  correct format and meet the specified criteria; they may also be used to
#'  provide the LLM with feedback or additional information,
#'  like the result of a tool call or some evaluated code.
#'
#' Advanced prompt wraps may also include functions that directly handle
#' the response from a LLM API or configure API parameters.
#'
#' @details
#' For advanced use, modify_fn, extraction_fn, and validation_fn
#'  may take the [llm_provider-class] object (as used with [send_prompt()]) as
#'  second argument, and the 'http_list' (a list of all HTTP requests
#'  and responses made during [send_prompt()]) as third argument.
#' Use of these arguments is not required, but can be useful for more complex
#'  prompt wraps which require additional information about the LLM provider
#'  or requests made so far.
#' The functions (including parameter_fn) also have access to
#'  the object `self` (not a function argument; it is attached to the environment
#'  of the function) which contains the [tidyprompt-class] object that the prompt wrap
#'  is a part of. This can be used to access other prompt wraps, or to access the
#'  prompt text or other information about the prompt. For instance,
#'  other prompt wraps can be accessed through `self$get_prompt_wraps()`.
#'
#' @param prompt A string or a [tidyprompt-class] object
#'
#' @param modify_fn A function that takes the previous prompt text (as
#'  first argument) and returns the new prompt text
#'
#' @param extraction_fn A function that takes the LLM response (as first argument)
#'  and attempts to extract a value from it.
#' Upon succesful extraction, the function should return the extracted value.
#'  If the extraction fails, the function should return a [llm_feedback()] message
#'  to initiate a retry.
#' A [llm_break()] can be returned to break the extraction and validation loop,
#'  ending [send_prompt()]
#'
#' @param validation_fn A function that takes the (extracted) LLM response
#'  (as first argument) and attempts to validate it.
#' Upon succesful validation, the function should return TRUE. If the validation
#'  fails, the function should return a [llm_feedback()] message to initiate a retry.
#' A [llm_break()] can be returned to break the extraction and validation loop,
#'  ending [send_prompt()]
#'
#' @param handler_fn A function that takes a 'completion' object (a result
#'  of a request to a LLM, as returned by `$complete_chat()` of a [llm_provider-class]
#'  object) as first argument and the [llm_provider-class] object as second argument.
#' The function should return a (modified or identical) completion object.
#' This can be used for advanced side effects, like logging, or native tool calling,
#'  or keeping track of token usage. See [llm_provider-class] for more information;
#'  handler_fn is attached to the [llm_provider-class] object that is being used.
#' When using an [llm_provider_ellmer()], the up-to-date ellmer_chat is synced
#'  onto the provider before handlers run. This allows handlers to access,
#'  for instance, the current cost of the conversation, and, for instance,
#'  to stop the conversation if a certain budget is exceeded.
#' For example usage, see source code of [answer_using_tools()]

#' @param parameter_fn A function that takes the [llm_provider-class] object
#'  which is being used with [send_prompt()] and returns a named list of parameters
#'  to be set in the [llm_provider-class] object via its `$set_parameters()` method.
#' This can be used to configure specific parameters of the [llm_provider-class]
#'  object when evaluating the prompt.
#' For example, [answer_as_json()] may set different parameters for different APIs
#'  related to JSON output.
#' This function is typically only used with advanced prompt wraps that require
#'  specific settings in the [llm_provider-class] object
#'
#' @param type The type of prompt wrap. Must be one of:
#' \itemize{
#' \item "unspecified": The default type, typically used for prompt wraps
#' which request a specific format of the LLM response, like [answer_as_integer()]
#'
#' \item "mode": For prompt wraps that change how the LLM should answer the prompt,
#' like [answer_by_chain_of_thought()] or [answer_by_react()]
#'
#' \item "tool": For prompt wraps that enable the LLM to use tools, like [answer_using_tools()]
#' or [answer_using_r()] when 'output_as_tool' = TRUE
#'
#' \item "break": For prompt wraps that may break the extraction and validation loop,
#'  like [quit_if()]. These are applied before type "unspecified" as they may
#'  instruct the LLM to not answer the prompt in the manner specified by those
#'  prompt wraps
#'
#' \item "check": For prompt wraps that apply a last check to the final answer,
#'  after all other prompt wraps have been evaluated.
#' These prompt wraps may only contain a validation function, and are applied
#'  after all other prompt wraps have been evaluated. These prompt wraps are
#'  even applied after an earlier prompt wrap has broken the extraction and validation loop
#'  with [llm_break()]
#' }
#' Types are used to determine the order in which prompt wraps are applied.
#' When constructing the prompt text, prompt wraps are applied to the base prompt
#'  in the following order: 'check', 'unspecified', 'break', 'mode', 'tool'.
#' When evaluating the LLM response and applying extraction and validation functions,
#'  prompt wraps are applied in the reverse order: 'tool', 'mode', 'break',
#'  'unspecified', 'check'.
#' Order among the same type is preserved in the order they were added to the prompt.

#' @param name An optional name for the prompt wrap.
#'  This can be used to identify the prompt wrap in the [tidyprompt-class] object
#'
#' @return A [tidyprompt-class] object with the [prompt_wrap()] appended to it
#'
#' @export
#'
#' @example inst/examples/prompt_wrap.R
#' @example R/answer_as_integer.R
#'
#' @family prompt_wrap
#' @family pre_built_prompt_wraps
#'
#' @seealso [tidyprompt-class] [send_prompt()]
prompt_wrap <- function(
  prompt,
  modify_fn = NULL,
  extraction_fn = NULL,
  validation_fn = NULL,
  handler_fn = NULL,
  parameter_fn = NULL,
  type = c("unspecified", "mode", "tool", "break", "check"),
  name = NULL
) {
  UseMethod("prompt_wrap")
}

#' [prompt_wrap()] method for when a [tidyprompt-class] object is supplied
#'
#' Calls the internal function to append the prompt wrap.
#'
#' @param prompt A single string or a [tidyprompt-class] object
#' @param ... Additional arguments
#'
#' @return A [tidyprompt-class] object with the [prompt_wrap()] appended to it
#'
#' @exportS3Method prompt_wrap tidyprompt
#'
#' @noRd
#' @keywords internal
prompt_wrap.tidyprompt <- function(prompt, ...) {
  prompt_wrap_internal(prompt, ...)
}

#' Default method for [prompt_wrap()]
#'
#' Attempts to create a [tidyprompt-class] object from whatever is passed as 'prompt';
#' then calls the internal function to append the [prompt_wrap()].
#'
#' @param prompt A single string or a [tidyprompt-class] object
#' @param ... Additional arguments
#'
#' @return A [tidyprompt-class] object with the [prompt_wrap()] appended to it
#'
#' @exportS3Method prompt_wrap default
#'
#' @noRd
#' @keywords internal
prompt_wrap.default <- function(prompt, ...) {
  prompt <- tidyprompt(prompt)
  prompt_wrap_internal(prompt, ...)
}

#' Internal function to append a [prompt_wrap()] to a [tidyprompt-class] object
#'
#' @param prompt See [prompt_wrap()]
#' @param modify_fn See [prompt_wrap()]
#' @param extraction_fn See [prompt_wrap()]
#' @param validation_fn See [prompt_wrap()]
#' @param handler_fn See [prompt_wrap()]
#' @param parameter_fn See [prompt_wrap()]
#' @param type See [prompt_wrap()]
#' @param name See [prompt_wrap()]
#'
#' @return A [tidyprompt-class] object with the [prompt_wrap()] appended to it
#'
#' @noRd
#' @keywords internal
prompt_wrap_internal <- function(
  prompt,
  modify_fn = NULL,
  extraction_fn = NULL,
  validation_fn = NULL,
  handler_fn = NULL,
  parameter_fn = NULL,
  type = c("unspecified", "mode", "tool", "break", "check"),
  name = NULL
) {
  type <- match.arg(type)

  stopifnot(
    is.null(modify_fn) || is.function(modify_fn),
    is.null(validation_fn) || is.function(validation_fn),
    is.null(extraction_fn) || is.function(extraction_fn),
    is.null(handler_fn) || is.function(handler_fn),
    is.null(parameter_fn) || is.function(parameter_fn)
  )

  if (type == "check") {
    if (is.null(validation_fn)) {
      stop("When using type 'check', validation_fn is required")
    }
    if (
      !all(
        sapply(
          list(modify_fn, extraction_fn, handler_fn, parameter_fn),
          is.null
        )
      )
    ) {
      stop("When using type 'check', only validation_fn is allowed")
    }
  }

  if (
    all(
      sapply(
        list(
          modify_fn,
          extraction_fn,
          validation_fn,
          handler_fn,
          parameter_fn
        ),
        is.null
      )
    )
  ) {
    stop(
      paste0(
        "At least one of modify_fn, extraction_fn, validation_fn, handler_fn, or parameter_fn",
        " must be provided"
      )
    )
  }

  ensure_three_arguments <- function(func) {
    if (is.null(func)) {
      return(NULL)
    }

    # Get the original arguments of the function
    original_args <- formals(func)

    # Define the required arguments in the desired order, without overwriting the first argument's name
    required_args <- c("llm_provider", "http_list")

    # Combine the original arguments with the missing required arguments
    combined_args <- c(
      original_args,
      structure(
        rep(list(quote(expr = NULL)), length(required_args)),
        names = required_args
      )
    )

    # Remove excess arguments beyond the first three
    final_args <- combined_args[1:3]

    # Ensure the non-first required arguments are named appropriately
    if (length(final_args) > 1) {
      names(final_args)[-1] <- required_args[1:(length(final_args) - 1)]
    }

    # Update the function's arguments
    formals(func) <- final_args
    func
  }
  modify_fn <- ensure_three_arguments(modify_fn)
  validation_fn <- ensure_three_arguments(validation_fn)
  extraction_fn <- ensure_three_arguments(extraction_fn)

  if (!is.null(parameter_fn) && length(formals(parameter_fn)) != 1) {
    stop(
      paste0(
        "Parameter_fn should be a function that takes one argument,",
        " which is the llm_provider"
      )
    )
  }

  type <- match.arg(type)

  # Create prompt wrap
  prompt_wrap <- structure(
    list(
      type = type,
      modify_fn = modify_fn,
      extraction_fn = extraction_fn,
      validation_fn = validation_fn,
      handler_fn = handler_fn,
      parameter_fn = parameter_fn,
      name = name
    ),
    class = "prompt_wrap"
  )

  # Append to prompt
  prompt$add_prompt_wrap(prompt_wrap)

  return(prompt)
}

#' Create a provider-level prompt wrap
#'
#' @description
#' `r lifecycle::badge("experimental")`
#' Build a provider-specific prompt wrap, to store on an [llm_provider-class] object
#' (with `$add_prompt_wrap()`). These prompt wraps can be applied before or
#' after any prompt-specific prompt wraps. In this way, you can ensure that
#' certain prompt wraps are always applied when using a specific LLM provider.
#'
#' @inheritParams prompt_wrap
#'
#' @export
#'
#' @return A `provider_prompt_wrap` object, to be stored on an [llm_provider-class] object
#'
#' @example inst/examples/provider_prompt_wrap.R
provider_prompt_wrap <- function(
  modify_fn = NULL,
  extraction_fn = NULL,
  validation_fn = NULL,
  handler_fn = NULL,
  parameter_fn = NULL,
  type = c("unspecified", "mode", "tool", "break", "check"),
  name = NULL
) {
  type <- match.arg(type)

  stopifnot(
    is.null(modify_fn) || is.function(modify_fn),
    is.null(extraction_fn) || is.function(extraction_fn),
    is.null(validation_fn) || is.function(validation_fn),
    is.null(handler_fn) || is.function(handler_fn),
    is.null(parameter_fn) || is.function(parameter_fn)
  )

  if (type == "check") {
    if (is.null(validation_fn)) {
      stop("When using type 'check', validation_fn is required")
    }
    if (
      !all(sapply(
        list(modify_fn, extraction_fn, handler_fn, parameter_fn),
        is.null
      ))
    ) {
      stop("When using type 'check', only validation_fn is allowed")
    }
  }

  if (
    all(sapply(
      list(modify_fn, extraction_fn, validation_fn, handler_fn, parameter_fn),
      is.null
    ))
  ) {
    stop(
      "At least one of modify_fn, extraction_fn, validation_fn, handler_fn, or parameter_fn must be provided"
    )
  }

  if (!is.null(parameter_fn) && length(formals(parameter_fn)) != 1) {
    stop("parameter_fn must take one argument: llm_provider")
  }

  structure(
    list(
      type = type,
      modify_fn = modify_fn,
      extraction_fn = extraction_fn,
      validation_fn = validation_fn,
      handler_fn = handler_fn,
      parameter_fn = parameter_fn,
      name = name
    ),
    class = c("provider_prompt_wrap", "prompt_wrap_spec")
  )
}

#' Add text to a tidyprompt
#'
#' Add text to a prompt by adding a [prompt_wrap()] which will append the text to
#' the before or after the current prompt text.
#'
#' @param prompt A single string or a [tidyprompt()] object
#' @param text Text to be added to the current prompt text
#' @param sep Separator to be used between the current prompt text and the text to be added
#' @param position Where to add the text; either "after" or "before".
#
#' @return A [tidyprompt()] with an added [prompt_wrap()] which
#' will append the text to the end of the current prompt text
#'
#' @export
#'
#' @example inst/examples/add_text.R
#'
#' @family pre_built_prompt_wraps
#' @family miscellaneous_prompt_wraps
add_text <- function(
  prompt,
  text,
  position = c("after", "before"),
  sep = "\n\n"
) {
  position <- match.arg(position)

  modify_fn <- function(original_prompt_text) {
    if (position == "after") {
      paste(original_prompt_text, text, sep = sep)
    } else {
      paste(text, original_prompt_text, sep = sep)
    }
  }

  prompt_wrap(prompt, modify_fn, name = "add_text")
}

test_that("add text to prompt", {
  prompt <- "Prompt" |>
    add_text("Text after") |>
    add_text("Text before", position = "before")

  constructed_text <- prompt |> construct_prompt_text()

  expect_equal(
    constructed_text,
    "Text before\n\nPrompt\n\nText after"
  )
})

#' Enable LLM to draft and execute SQL queries on a database
#'
#' @param prompt A single string or a [tidyprompt()] object
#'
#' @param add_text Single string which will be added to the prompt text,
#' informing the LLM that they must use SQL to answer the prompt
#'
#' @param conn A DBIConnection object to the SQL database
#'
#' @param list_tables Logical indicating whether to list tables available in the database
#' in the prompt text
#'
#' @param describe_tables Logical indicating whether to describe the tables available in the database
#' in the prompt text. If TRUE, the columns of each table will be listed
#'
#' @param evaluate_code Logical indicating whether to evaluate the SQL code.
#' If TRUE, the SQL code will be executed on the database and the results will be returned.
#' Use with caution, as this allows the LLM to execute arbitrary SQL code
#'
#' @param output_as_tool Logical indicating whether to return the output as a tool result.
#' If TRUE, the output of the SQL query will be sent back to the LLM as a tool result.
#' The LLM can then provide a final answer or try another query. This can
#' continue until the LLM provides a final answer without any SQL code
#'
#' @param return_mode Character string indicating the return mode. Options are:
#' \itemize{
#'  \item "full": Return a list containing the SQL code, output, and formatted output
#'  \item "code": Return only the SQL code
#'  \item "object": Return only the query result object
#'  \item "formatted_output": Return the formatted output: a string detailing the SQL code
#'  and query result object.This is identical to how the LLM would see the output when
#'  output_as_tool is TRUE
#'  \item "llm_answer": Return the LLM answer. If output as tool is TRUE,
#'  the return mode will always be "llm_answer" (since the LLM uses SQL to
#'  provide a final answer)
#'  }
#'
#' @return A [tidyprompt()] with an added [prompt_wrap()] which will ensure
#' that the LLM will use SQL to answer the prompt
#'
#' @export
#'
#' @example inst/examples/answer_using_sql.R
#'
#' @family pre_built_prompt_wraps
#' @family answer_using_prompt_wraps
answer_using_sql <- function(
  prompt,
  add_text = paste0(
    "You must code in SQL to answer this prompt.",
    " You must provide all SQL code between ```sql and ```.",
    "\n\n",
    "Never make assumptions about the possible values in the tables.\n",
    "Instead, execute SQL queries to retrieve information you need."
  ),
  conn,
  list_tables = TRUE,
  describe_tables = TRUE,
  evaluate_code = FALSE,
  output_as_tool = FALSE,
  return_mode = c(
    "full",
    "code",
    "object",
    "formatted_output",
    "llm_answer"
  )
) {
  # Validate settings
  return_mode <- match.arg(return_mode)

  stopifnot(
    is.character(add_text),
    length(add_text) == 1,
    inherits(conn, "DBIConnection"),
    is.logical(list_tables),
    is.logical(describe_tables),
    is.logical(evaluate_code),
    is.logical(output_as_tool)
  )

  if (!requireNamespace("DBI", quietly = TRUE)) {
    stop(
      "'DBI' package is required; please install it using 'install.packages(\"DBI\")"
    )
  }

  if (!evaluate_code && output_as_tool) {
    output_as_tool <- FALSE
  }
  if (output_as_tool) {
    return_mode <- "llm_answer"
  }
  if (
    !evaluate_code &&
      return_mode %in% c("console", "object", "formatted_output")
  ) {
    stop(
      "The return mode must be 'full', 'code', or 'llm_answer' if 'evaluate_code' is FALSE"
    )
  }

  # Retrieve list of tables if needed
  tables <- character(0)
  if (list_tables) {
    tables <- DBI::dbListTables(conn)
  }

  # If describe_tables is TRUE, we will attempt to fetch their columns
  describe_text <- ""
  if (list_tables && describe_tables && length(tables) > 0) {
    describe_info <- lapply(tables, function(tbl) {
      fields <- tryCatch(
        DBI::dbListFields(conn, tbl),
        error = function(e) character(0)
      )
      fields_str <- if (length(fields) > 0) {
        paste0("  Columns: ", paste(fields, collapse = ", "))
      } else {
        "  No columns could be retrieved."
      }
      paste0("- ", tbl, "\n", fields_str)
    })
    describe_text <- paste0(
      "\n\nTable descriptions:\n",
      paste(describe_info, collapse = "\n\n")
    )
  }

  # Define modify_fn which adds info about the SQL environment
  modify_fn <- function(original_prompt_text) {
    new_text <- glue::glue(
      "{original_prompt_text}\n\n",
      "{add_text}"
    )

    if (list_tables && length(tables) > 0) {
      new_text <- glue::glue(
        "{new_text}\n\n",
        "These tables are available in the database:\n",
        "  {paste(tables, collapse = ', ')}"
      )
      # Append describe info if available
      if (nzchar(describe_text)) {
        new_text <- glue::glue(
          "{new_text}{describe_text}"
        )
      }
    }

    if (evaluate_code && return_mode == "object") {
      new_text <- glue::glue(
        "{new_text}\n\n",
        "The SQL code should produce a query result that answers the prompt."
      )
    }

    if (output_as_tool) {
      new_text <- glue::glue(
        "{new_text}\n\n",
        "Your SQL query will be executed on the database. The results will be",
        " sent back to you. After seeing the results, you can either provide a",
        " final answer or try another SQL query. When you provide your final answer,",
        " do not include any SQL code."
      )
    }

    return(new_text)
  }

  # Define extraction function to handle returned LLM responses
  extraction_fn <- function(x) {
    return_list <- list()
    return_list$llm_answer <- x

    # Extract SQL code
    extracted_code <- answer_using_sql_extract_sql_code(x)

    if (length(extracted_code) == 0) {
      if (output_as_tool) {
        return(x)
      }

      return(
        llm_feedback(
          paste0(
            "No SQL code detected. You must provide SQL code ",
            "between ```sql and ```."
          )
        )
      )
    }

    # For simplicity, assume one SQL code block; if multiple, use the first
    sql_code <- extracted_code[1]
    return_list$code <- sql_code

    if (!evaluate_code) {
      # Just return the code or final answer
      if (return_mode == "code") {
        return(sql_code)
      }
      if (return_mode == "llm_answer") {
        return(x)
      }
      # 'full' without evaluation just returns code + llm_answer
      if (return_mode == "full") {
        return(return_list)
      }
      return(sql_code)
    }

    # Evaluate the SQL code
    query_res <- tryCatch(
      DBI::dbGetQuery(conn, sql_code),
      error = function(e) e
    )

    if (inherits(query_res, "error")) {
      return(
        llm_feedback(
          glue::glue(
            "An error occurred while executing the SQL:\n",
            "    {query_res$message}\n",
            "Please provide a valid SQL query."
          )
        )
      )
    }

    # query_res should now be a data frame
    return_list$output <- query_res

    formatted_output <- glue::glue(
      "--- SQL code: ---\n",
      "{sql_code}\n\n",
      "--- Query results: ---\n",
      "{utils::capture.output(print(query_res)) |> paste(collapse = \"\\n\")}"
    )
    return_list$formatted_output <- formatted_output

    if (output_as_tool) {
      # Send formatted output back as tool result
      return(llm_feedback(formatted_output, tool_result = TRUE))
    }

    # Return according to return_mode
    if (return_mode == "full") {
      return(return_list)
    }
    if (return_mode == "code") {
      return(sql_code)
    }
    if (return_mode == "console") {
      return(utils::capture.output(print(query_res)) |> paste(collapse = "\n"))
    }
    if (return_mode == "object") {
      return(query_res)
    }
    if (return_mode == "formatted_output") {
      return(formatted_output)
    }
    if (return_mode == "llm_answer") {
      return(x)
    }

    return(return_list)
  }

  # If we are sending back output as a tool
  type <- "unspecified"
  if (output_as_tool) {
    type <- "tool"
  }

  # Wrap the prompt
  prompt_wrap(
    prompt,
    modify_fn,
    extraction_fn,
    type = type,
    name = "answer_using_sql"
  )
}

#' Helper function to extract SQL code from a string
#'
#' This function extracts SQL code from a string by matching all content between
#' sql tags.
#'
#' @param input_string A string containing SQL code, typically a response
#' from an LLM
#'
#' @return A character vector containing the extracted SQL code
#'
#' @noRd
#' @keywords internal
answer_using_sql_extract_sql_code <- function(input_string) {
  # Use regular expression to match all content between ```sql and ```, with case-insensitive matching
  matches <- gregexpr(
    "(?s)```[sS][qQ][lL]\\s*(.*?)\\s*```",
    input_string,
    perl = TRUE
  )
  extracted_code <- regmatches(input_string, matches)

  # Remove the ```sql and ``` wrappers
  extracted_code <- lapply(extracted_code, function(x) {
    sub("(?s)```[sS][qQ][lL]\\s*(.*?)\\s*```", "\\1", x, perl = TRUE)
  })

  return(unlist(extracted_code))
}

#' Make LLM answer as a constrained text response
#'
#' @param prompt A single string or a [tidyprompt()] object
#' @param max_words (optional) Maximum number of words allowed in the response.
#' If specified, responses exceeding this limit will fail validation
#' @param max_characters (optional) Maximum number of characters allowed in the response.
#' If specified, responses exceeding this limit will fail validation
#' @param add_instruction_to_prompt (optional) Add instruction for replying
#' within the constraints to the prompt text. Set to FALSE for debugging if
#' extractions/validations are working as expected (without instruction the
#' answer should fail the validation function, initiating a retry)
#'
#' @return A [tidyprompt()] with an added [prompt_wrap()] which
#' will ensure that the LLM response conforms to the specified constraints
#'
#' @export
#'
#' @example inst/examples/answer_as_text.R
#'
#' @family pre_built_prompt_wraps
#' @family answer_as_prompt_wraps
answer_as_text <- function(
  prompt,
  max_words = NULL,
  max_characters = NULL,
  add_instruction_to_prompt = TRUE
) {
  instruction <- "You must provide a text response."

  if (!is.null(max_words)) {
    instruction <- paste(
      instruction,
      glue::glue("The response must be at most {max_words} words.")
    )
  }

  if (!is.null(max_characters)) {
    instruction <- paste(
      instruction,
      glue::glue("The response must be at most {max_characters} characters.")
    )
  }

  modify_fn <- function(original_prompt_text) {
    if (!add_instruction_to_prompt) {
      return(original_prompt_text)
    }

    glue::glue("{original_prompt_text}\n\n{instruction}")
  }

  extraction_fn <- function(x) {
    trimmed <- trimws(x)
    word_count <- strsplit(trimmed, "\\s+")[[1]] |> length()
    char_count <- nchar(trimmed)

    if (!is.null(max_words) && word_count > max_words) {
      return(
        llm_feedback(
          glue::glue("Response exceeds the maximum word limit of {max_words}.")
        )
      )
    }

    if (!is.null(max_characters) && char_count > max_characters) {
      return(
        llm_feedback(
          glue::glue(
            "Response exceeds the maximum character limit of {max_characters}."
          )
        )
      )
    }

    return(trimmed)
  }

  prompt_wrap(
    prompt,
    modify_fn,
    extraction_fn,
    name = "answer_as_text"
  )
}

#' Convert a named or unnamed list/vector to a string representation
#'
#' Converts a named or unnamed list/vector to a string format, intended for
#' sending it to an LLM (or for display or logging).
#'
#' @param obj A list or vector (named or unnamed) to be converted to a string.
#' @param how In what way the object should be converted to a string;
#' either "inline" or "expanded". "inline" presents all key-value pairs
#' or values as a single line. "expanded" presents each key-value pair or value
#' on a separate line.
#'
#' @return A single string representing the list/vector.
#'
#' @export
#'
#' @example inst/examples/vector_list_to_string.R
#'
#' @family text_helpers
vector_list_to_string <- function(obj, how = c("inline", "expanded")) {
  how <- match.arg(how)

  # Check if input is a list or vector
  if (!is.list(obj) && !is.vector(obj)) {
    stop("Input must be a list or vector")
  }

  # Return empty string if object is empty
  if (length(obj) == 0) {
    return("")
  }

  # Handle named and unnamed cases
  if (!is.null(names(obj)) && any(names(obj) != "")) {
    # Named case: Generate output based on the `how` parameter
    if (how == "inline") {
      result_string <- paste(
        paste(names(obj), obj, sep = ": "),
        collapse = ", "
      )
    } else if (how == "expanded") {
      result_string <- paste(
        paste(names(obj), obj, sep = ": "),
        collapse = "\n"
      )
    }
  } else {
    # Unnamed case: Handle values only
    if (how == "inline") {
      result_string <- paste(obj, collapse = ", ")
    } else if (how == "expanded") {
      result_string <- paste(obj, collapse = "\n")
    }
  }

  return(result_string)
}

#' @title Set chain of thought mode for a prompt
#'
#' @description This function enables chain of thought mode for evaluation of a prompt
#' or a [tidyprompt()]. In chain of thought mode, the large language model (LLM)
#' In chain of thought mode, the large language model (LLM) is asked to think
#' step by step to arrive at a final answer. It is hypothesized that this may
#' increase LLM performance at solving complex tasks. Chain of thought mode
#' is inspired by the method described in Wei et al. (2022).
#'
#' @param prompt A single string or a [tidyprompt()] object
#' @param extract_from_finish_brackets A logical indicating whether the final answer
#' should be extracted from the text inside the "FINISH[...]" brackets.
#' @param extraction_lenience A logical indcating whether the extraction function should be lenient.
#' If TRUE, the extraction function will attempt to extract the final answer
#' even if it cannot be extracted from within the brackets, by extracting
#' everything after the final occurence of 'FINISH' (if present). This may
#' be useful for smaller LLMs which may not follow the output format
#' as strictly
#'
#' @return A [tidyprompt()] with an added [prompt_wrap()] which will ensure
#' that the LLM follows the chain of thought mode in answering the prompt
#'
#' @export
#' @example inst/examples/answer_by_chain_of_thought.R
#'
#' @references
#' Wei, J., Wang, X., Schuurmans, D., Bosma, M., Ichter, B., Xia, F., Chi, E., Le, Q., & Zhou, D. (2022).
#' Chain-of-Thought Prompting Elicits Reasoning in Large Language Models.
#' <doi:10.48550/arXiv.2201.11903>
#'
#' @family pre_built_prompt_wraps
#' @family answer_by_prompt_wraps
answer_by_chain_of_thought <- function(
  prompt,
  extract_from_finish_brackets = TRUE,
  extraction_lenience = TRUE
) {
  # Define modification/extraction/validation functions:
  modify_fn <- function(original_prompt_text) {
    new_prompt <- glue::glue(
      "You are given a user's prompt.
      To answer the user's prompt, you need to think step by step to arrive at a final answer.

      ----- START OF USER'S PROMPT -----
      {original_prompt_text}
      ----- END OF USER'S PROMPT -----

      What are the steps you would take to answer the user's prompt?
      Describe your thought process in the following format:
        >> step 1: <step 1 description>
        >> step 2: <step 2 description>
        (etc.)"
    )

    if (extract_from_finish_brackets) {
      new_prompt <- glue::glue(
        "{new_prompt}

        When you are done, you must type:
          FINISH[<put here your final answer to the user's prompt>]

        Make sure your final answer follows the logical conclusion of your thought process."
      )
    }

    return(new_prompt)
  }

  extraction_fn <- function(llm_response) {
    if (!extract_from_finish_brackets) {
      return(llm_response)
    }
    extraction_fn_finish(llm_response, extraction_lenience)
  }

  prompt_wrap(
    prompt,
    modify_fn,
    extraction_fn,
    type = "mode",
    name = "answer_by_chain_of_thought"
  )
}

#' @title Make a request to an LLM provider
#'
#' @description This is a helper function which facilitates making requests to LLM
#' providers which follow the structure of the OpenAI API or the Ollama
#' API. It handles both streaming and non-streaming requests.
#'
#' This function is part of the internal API and is not intended to be called directly by
#' users. It is used in some of the pre-built [llm_provider-class] objects
#' included in 'tidyprompt' (e.g., [llm_provider_openai()) and
#' [llm_provider_ollama()]).
#'
#' @param chat_history A data frame with 'role' and 'content' columns
#' (see [chat_history()])
#' @param request A 'httr2' request object with the URL, headers, and body
#' @param stream Logical indicating whether the API should stream responses
#' @param verbose Logical indicating whether interactions should be printed to the console
#' @param api_type API type, one of "openai" or "ollama"
#'
#' @return A list with the completed chat history and the HTTP request and response
#' objects
#'
#' @keywords internal
#' @noRd
normalize_chat_history_metadata <- function(chat_history) {
  stopifnot(is.data.frame(chat_history))

  if (!"tool_call" %in% names(chat_history)) {
    chat_history$tool_call <- FALSE
  }
  if (!"tool_call_id" %in% names(chat_history)) {
    chat_history$tool_call_id <- NA_character_
  }
  if (!"tool_result" %in% names(chat_history)) {
    chat_history$tool_result <- FALSE
  }
  if (!"hidden_from_llm" %in% names(chat_history)) {
    chat_history$hidden_from_llm <- FALSE
  }

  chat_history$tool_call <- as.logical(chat_history$tool_call)
  chat_history$tool_call[is.na(chat_history$tool_call)] <- FALSE
  chat_history$tool_call_id <- as.character(chat_history$tool_call_id)
  chat_history$tool_result <- as.logical(chat_history$tool_result)
  chat_history$tool_result[is.na(chat_history$tool_result)] <- FALSE
  chat_history$hidden_from_llm <- as.logical(chat_history$hidden_from_llm)
  chat_history$hidden_from_llm[is.na(chat_history$hidden_from_llm)] <- FALSE
  chat_history$hidden_from_llm[chat_history$tool_call] <- TRUE

  chat_history
}

chat_history_to_send <- function(chat_history) {
  stopifnot(is.data.frame(chat_history))

  keep_rows <- rep(TRUE, nrow(chat_history))

  if ("hidden_from_llm" %in% names(chat_history)) {
    hidden_rows <- as.logical(chat_history$hidden_from_llm)
    hidden_rows[is.na(hidden_rows)] <- FALSE
    keep_rows <- keep_rows & !hidden_rows
  }

  if ("tool_call" %in% names(chat_history)) {
    tool_call_rows <- as.logical(chat_history$tool_call)
    tool_call_rows[is.na(tool_call_rows)] <- FALSE
    keep_rows <- keep_rows & !tool_call_rows
  }

  source_rows <- which(keep_rows)
  chat_history <- chat_history[keep_rows, , drop = FALSE]
  if ("hidden_from_llm" %in% names(chat_history)) {
    chat_history$hidden_from_llm <- NULL
  }

  attr(chat_history, "source_rows") <- source_rows
  chat_history
}

request_llm_provider <- function(
  chat_history,
  request,
  stream = NULL,
  verbose = getOption("tidyprompt.verbose", TRUE),
  api_type = c("openai", "ollama"),
  stream_callback = NULL,
  llm_provider = NULL
) {
  api_type <- match.arg(api_type)
  request <- normalize_openai_request(request, api_type)

  if (!is.null(stream) && stream) {
    req_result <- req_llm_stream(
      req = request,
      api_type = api_type,
      verbose = verbose,
      stream_callback = stream_callback,
      chat_history = chat_history,
      llm_provider = llm_provider
    )
  } else {
    req_result <- req_llm_non_stream(request, api_type, verbose)
  }

  stopifnot(
    is.list(req_result),
    "new" %in% names(req_result),
    is.data.frame(req_result$new)
  )

  completed <- dplyr::bind_rows(chat_history, req_result$new)
  completed <- normalize_chat_history_metadata(completed)

  return(
    list(
      completed = completed,
      http = list(
        request = request,
        response = req_result$httr2_response
      )
    )
  )
}

req_llm_handle_error <- function(e) {
  msg <- paste0(e$message)

  # Try to parse and include JSON body if available
  body_msg <- tryCatch(
    {
      body <- e$resp |>
        httr2::resp_body_string() |>
        jsonlite::fromJSON()
      paste0(
        "\nResponse body: ",
        jsonlite::toJSON(body, pretty = TRUE, auto_unbox = TRUE)
      )
    },
    error = function(e) "\n(Could not parse JSON body from response)"
  )

  msg <- paste0(
    msg,
    body_msg,
    "\nUse 'httr2::last_response()' and 'httr2::last_request()' for more information"
  )

  stop(msg, call. = FALSE)
}

req_llm_stream <- function(
  req,
  api_type,
  verbose,
  stream_callback = NULL,
  chat_history = NULL,
  llm_provider = NULL
) {
  role <- NULL
  message_accumulator <- ""
  tool_calls <- list()

  using_responses <- (api_type == "openai") && is_openai_responses_endpoint(req)

  # Resolve a central stream callback (provider-level overrides may already
  # have been applied before calling this function)
  stream_callback <- stream_callback %||%
    getOption("tidyprompt.stream_callback", NULL)

  partial_response <- ""

  emit_chunk <- function(text) {
    if (is.null(text) || !nzchar(text)) {
      return(invisible(NULL))
    }

    partial_response <<- paste0(partial_response, as.character(text))

    if (is.function(stream_callback)) {
      latest_message <- NULL
      if (is.data.frame(chat_history) && nrow(chat_history) > 0) {
        latest_message <- chat_history[nrow(chat_history), , drop = FALSE]
      }

      meta <- list(
        llm_provider = llm_provider,
        chat_history = chat_history,
        latest_message = latest_message,
        partial_response = partial_response,
        chunk = as.character(text),
        api_type = api_type,
        endpoint = if (using_responses) "responses" else "chat",
        verbose = verbose
      )
      stream_callback(as.character(text), meta)
    } else if (isTRUE(verbose)) {
      cat(text)
    }
    invisible(NULL)
  }

  # No low-level HTTP verbosity; we'll print only tokens ourselves or via callback
  resp <- tryCatch(
    httr2::req_perform_connection(req, verbosity = 0),
    error = function(e) req_llm_handle_error(e)
  )

  if (api_type == "ollama") {
    while (!httr2::resp_stream_is_complete(resp)) {
      lines <- httr2::resp_stream_lines(resp, lines = 1, warn = FALSE)
      if (length(lines) == 0) {
        next
      }
      for (line in lines) {
        if (!nzchar(line)) {
          next
        }
        data <- tryCatch(jsonlite::fromJSON(line), error = function(e) NULL)
        if (is.null(data)) {
          next
        }

        if (is.null(role) && !is.null(data$message$role)) {
          role <- data$message$role
        }
        if (!is.null(data$message$content)) {
          message_accumulator <- paste0(
            message_accumulator,
            data$message$content
          )
          emit_chunk(data$message$content)
        }
      }
    }
    emit_chunk("\n")
  } else if (!using_responses) {
    # OpenAI Chat Completions (SSE) — use resp_stream_sse for robust parsing
    if (is.null(role)) {
      role <- "assistant"
    }

    repeat {
      ev <- httr2::resp_stream_sse(resp)
      if (!nzchar(ev$data)) {
        next
      } # keepalive / empty
      if (identical(ev$data, "[DONE]")) {
        cat("\n")
        break
      }

      payload <- tryCatch(jsonlite::fromJSON(ev$data), error = function(e) NULL)
      if (is.null(payload)) {
        next
      }

      # tool calls (function calling) support
      tc <- payload$choices[1, ]$delta$tool_calls
      if (length(tc) > 0) {
        tool_calls <- append_or_update_tool_calls(
          tool_calls,
          tc,
          verbose
        )
        next
      }

      d <- tryCatch(
        payload$choices[1, ]$delta,
        error = function(e) NULL
      )
      if (is.null(d)) {
        next
      }

      if (!is.null(d$role) && is.null(role)) {
        role <- d$role
      }

      # text tokens
      if (!is.null(d$content)) {
        add <- d$content
        if (length(add) > 0 && !is.na(add)) {
          message_accumulator <- paste0(message_accumulator, add)
          emit_chunk(add)
        }
      }
    }
  } else {
    # OpenAI Responses API (SSE)
    if (is.null(role)) {
      role <- "assistant"
    }

    response_id <- NULL
    tool_calls <- list()
    args_buf <- new.env(parent = emptyenv()) # per-tool_call_id argument accumulator

    repeat {
      ev <- httr2::resp_stream_sse(resp)
      if (is.null(ev)) {
        emit_chunk("\n")
        break
      }
      if (!nzchar(ev$data) || identical(ev$data, "[DONE]")) {
        next
      }

      payload <- tryCatch(
        jsonlite::fromJSON(ev$data, simplifyVector = FALSE),
        error = function(e) NULL
      )
      if (is.null(payload)) {
        next
      }

      tp <- payload$type %||% ""

      # capture id once
      if (is.null(response_id) && !is.null(payload$id)) {
        response_id <- payload$id
      }

      # 1) normal text deltas
      if (
        grepl("response\\.(output_)?text\\.delta$", tp) &&
          !is.null(payload$delta)
      ) {
        message_accumulator <- paste0(
          message_accumulator,
          as.character(payload$delta)
        )
        emit_chunk(as.character(payload$delta))
        next
      }

      # 2) listen for tool calls
      if (identical(tp, "response.output_item.done")) {
        if (
          !is.null(payload$item) &&
            is.list(payload$item) &&
            identical(payload$item$type, "function_call")
        ) {
          tool_calls <- append_or_update_tool_calls(
            tool_calls,
            list(
              list(
                id = payload$item$id,
                type = "function",
                `function` = list(
                  name = payload$item$name,
                  arguments = payload$item$arguments |>
                    jsonlite::fromJSON() |>
                    jsonlite::toJSON(auto_unbox = TRUE, pretty = TRUE) |>
                    as.character()
                )
              )
            ),
            verbose
          )

          break
        }
        next
      }
    }
  }

  # attach tool_calls and response_id so the caller can act
  if (!is.list(resp$body)) {
    resp$body <- list()
  }
  resp$body$tool_calls <- tool_calls
  if (exists("response_id")) {
    resp$body$response_id <- response_id
  }

  list(
    new = data.frame(
      role = if (is.null(role)) "assistant" else role,
      content = message_accumulator,
      stringsAsFactors = FALSE
    ),
    httr2_response = resp
  )
}

req_llm_non_stream <- function(req, api_type, verbose) {
  response <- tryCatch(
    httr2::req_perform(req),
    error = function(e) req_llm_handle_error(e)
  )

  content <- httr2::resp_body_json(response)
  using_responses <- (api_type == "openai") && is_openai_responses_endpoint(req)

  if (api_type == "ollama") {
    new <- tryCatch(
      data.frame(
        role = content$message$role,
        content = content$message$content,
        stringsAsFactors = FALSE
      ),
      error = function(e) data.frame()
    )
  } else if (using_responses) {
    # Prefer convenience field if present
    txt <- NULL
    if (!is.null(content$output_text)) {
      txt <- content$output_text
    } else if (!is.null(content$output) && length(content$output) > 0) {
      # Fallback: stitch message items' text
      bits <- unlist(lapply(content$output, function(it) {
        if (!is.null(it$content)) {
          unlist(lapply(it$content, function(c) c$text %||% NULL))
        }
      }))
      txt <- paste(bits, collapse = "")
    }
    new <- data.frame(
      role = "assistant",
      content = txt %||% "",
      stringsAsFactors = FALSE
    )
  } else {
    # OpenAI Chat Completions
    new <- tryCatch(
      data.frame(
        role = content$choices[[1]]$message$role,
        content = content$choices[[1]]$message$content,
        stringsAsFactors = FALSE
      ),
      error = function(e) data.frame()
    )
  }

  list(new = new, httr2_response = response)
}

append_or_update_tool_calls <- function(tool_calls, new_tool_calls, verbose) {
  tool_call <- new_tool_calls[[1]]
  id <- tool_call$id

  last_id <- if (length(tool_calls) > 0) {
    tool_calls[[length(tool_calls)]]$id
  } else {
    NULL
  }

  if (!is.null(id) && (is.null(last_id) || (id != last_id))) {
    tool_calls <- append(
      tool_calls,
      list(
        list(
          id = tool_call$id,
          type = tool_call$type,
          `function` = list(
            name = tool_call$`function`$name,
            args = tool_call$`function`$arguments
          )
        )
      )
    )
  } else {
    # Update arguments of the last call
    arguments_current <- tool_calls[[length(tool_calls)]]$`function`$arguments
    arguments_new <- tool_call$`function`$arguments
    if (length(arguments_new) > 0) {
      tool_calls[[length(tool_calls)]]$`function`$arguments <- paste0(
        arguments_current,
        arguments_new
      )
    }
  }

  tool_calls
}

is_openai_responses_endpoint <- function(req) {
  url_str <- tryCatch(
    httr2::url_build(req$url),
    error = function(e) as.character(req$url)
  )
  grepl("/responses(\\b|/|\\?)", url_str)
}

normalize_openai_request <- function(req, api_type) {
  if (api_type != "openai") {
    return(req)
  }
  if (!is_openai_responses_endpoint(req)) {
    return(req)
  }

  body_obj <- tryCatch(req$body$data, error = function(e) NULL)
  if (!is.list(body_obj)) {
    return(req)
  }

  # Chat Completions -> Responses rename
  if (!is.null(body_obj$messages)) {
    msgs <- body_obj$messages

    msgs <- lapply(msgs, function(m) {
      m$tool_calls <- NULL
      m$function_call <- NULL
      m$name <- NULL
      if (is.null(m$content)) {
        m$content <- ""
      }

      # Convert CC tool-role messages to typed items for Responses
      if (identical(m$role, "tool") && !is.null(m$tool_call_id)) {
        return(list(
          type = "function_call_output",
          call_id = m$tool_call_id,
          output = as.character(m$content %||% "")
        ))
      }
      m
    })

    needs_fc <- function(item) {
      is.list(item) &&
        isTRUE(identical(item$type, "function_call_output")) &&
        !is.null(item$call_id)
    }
    has_fc_for <- function(call_id, items) {
      any(vapply(
        items,
        function(x) {
          is.list(x) &&
            identical(x$type, "function_call") &&
            identical(x$call_id, call_id)
        },
        logical(1)
      ))
    }

    input <- list()
    for (m in msgs) {
      if (needs_fc(m) && !has_fc_for(m$call_id, input)) {
        input[[length(input) + 1]] <- list(
          type = "function_call",
          call_id = m$call_id,
          name = "-",
          arguments = "-"
        )
      }
      input[[length(input) + 1]] <- m
    }

    body_obj$input <- input
    body_obj$messages <- NULL
  }

  # response_format -> text.format  (Responses API)
  if (!is.null(body_obj$response_format) && is.null(body_obj$text$format)) {
    rf <- body_obj$response_format
    if (is.character(rf)) {
      rf <- list(type = rf)
    }

    fmt <- NULL
    if (identical(rf$type, "json_object")) {
      fmt <- list(type = "json_object")
    } else if (identical(rf$type, "json_schema")) {
      js <- rf$json_schema %||% rf$schema %||% rf
      name <- js$name %||% js$title %||% "Schema"
      schema <- js$schema %||% js
      strict <- isTRUE(rf$strict %||% js$strict)
      fmt <- list(
        type = "json_schema",
        name = as.character(name),
        schema = schema
      )
      if (strict) fmt$strict <- TRUE
    }
    if (!is.null(fmt)) {
      body_obj$text <- body_obj$text %||% list()
      body_obj$text$format <- fmt
      body_obj$response_format <- NULL
    }
  }

  # SAFETY GUARD: Only touch list-like items with a role
  if (is.list(body_obj$input)) {
    for (i in seq_along(body_obj$input)) {
      item <- body_obj$input[[i]]

      # skip non-lists (e.g., atomic strings) or items without a role
      if (!(is.list(item) && !is.null(item$role))) {
        next
      }

      # If any lingering tool-role items slipped through, degrade to "system"
      # (You already convert tool messages above; this is just a fallback.)
      if (identical(item$role, "tool")) {
        if (!is.null(item$tool_call_id) && !is.null(item$content)) {
          body_obj$input[[i]] <- list(
            type = "function_call_output",
            call_id = item$tool_call_id,
            output = as.character(item$content %||% "")
          )
        } else {
          body_obj$input[[i]]$role <- "system"
        }
      }
    }
  }

  # Tools normalization for Responses API
  if (!is.null(body_obj$tools) && is.list(body_obj$tools)) {
    body_obj$tools <- lapply(body_obj$tools, function(t) {
      if (is.null(t$type) && !is.null(t$`function`)) {
        t$type <- "function"
      }
      if (is.null(t$name) && !is.null(t$`function`$name)) {
        t$name <- t$`function`$name
      }
      if (is.null(t$description) && !is.null(t$`function`$description)) {
        t$description <- t$`function`$description
      }
      if (is.null(t$parameters) && !is.null(t$`function`$parameters)) {
        t$parameters <- t$`function`$parameters
      }
      t
    })
  }

  # Normalize tool_choice shape
  if (!is.null(body_obj$tool_choice) && is.list(body_obj$tool_choice)) {
    if (
      !is.null(body_obj$tool_choice$`function`$name) &&
        is.null(body_obj$tool_choice$name)
    ) {
      body_obj$tool_choice$name <- body_obj$tool_choice$`function`$name
    }
  }

  httr2::req_body_json(req, body_obj, auto_unbox = TRUE)
}

# tests/testthat/test-openai-real.R

# Here we test some functions from R/internal_request_llm_provider.R,
#   to verify that the OpenAI API can be used correctly with both
#   the responses and the chat completions endpoints

testthat::skip_if_not_installed("httr2")

suppressWarnings(library(testthat))
suppressWarnings(library(httr2))

# ---- Config ------------------------------------------------------------------
openai_key <- Sys.getenv("OPENAI_API_KEY", unset = "")
skip_if(openai_key == "", "OPENAI_API_KEY not set; skipping OpenAI tests")

model_chat <- Sys.getenv("TIDYPROMPT_OPENAI_CHAT_MODEL", unset = "gpt-4o-mini")
model_resp <- Sys.getenv(
  "TIDYPROMPT_OPENAI_RESPONSES_MODEL",
  unset = "gpt-4o-mini"
)

dummy_history <- function() {
  data.frame(role = "user", content = "ping", stringsAsFactors = FALSE)
}

openai_headers <- function(req) {
  req |>
    req_headers(
      Authorization = paste("Bearer", openai_key),
      `Content-Type` = "application/json"
    )
}

# ---- Chat Completions (non-stream) -------------------------------------------
test_that("OpenAI Chat Completions (non-stream) returns content", {
  req <- request("https://api.openai.com/v1/chat/completions") |>
    openai_headers() |>
    req_body_json(
      list(
        model = model_chat,
        temperature = 0,
        max_tokens = 10,
        messages = list(
          list(
            role = "system",
            content = "Reply with exactly the single word: pong"
          ),
          list(role = "user", content = "Say it.")
        )
      ),
      auto_unbox = TRUE
    )

  out <- request_llm_provider(
    chat_history = dummy_history(),
    request = req,
    stream = FALSE,
    api_type = "openai"
  )

  expect_s3_class(out$completed, "data.frame")
  assistant <- tail(out$completed$content, 1)
  expect_true(is.character(assistant) && nchar(assistant) > 0)
  expect_true(grepl("pong", assistant, ignore.case = TRUE))
})

# ---- Responses API (non-stream) ----------------------------------------------
test_that("OpenAI Responses API (non-stream) returns content", {
  req <- request("https://api.openai.com/v1/responses") |>
    openai_headers() |>
    req_body_json(
      list(
        model = model_resp,
        temperature = 0,
        max_output_tokens = 32,
        input = "Reply with exactly the single word: pong"
      ),
      auto_unbox = TRUE
    )

  out <- request_llm_provider(
    chat_history = dummy_history(),
    request = req,
    stream = FALSE,
    api_type = "openai"
  )

  assistant <- tail(out$completed$content, 1)
  expect_true(is.character(assistant) && nchar(assistant) > 0)
  expect_true(grepl("pong", assistant, ignore.case = TRUE))
})

# ---- Chat Completions (stream) -----------------------------------------------
test_that("OpenAI Chat Completions (stream) accumulates deltas", {
  req <- request("https://api.openai.com/v1/chat/completions") |>
    openai_headers() |>
    req_body_json(
      list(
        model = model_chat,
        temperature = 0,
        max_tokens = 32,
        stream = TRUE,
        messages = list(
          list(
            role = "system",
            content = "Reply with exactly the single word: pong"
          ),
          list(role = "user", content = "Say it.")
        )
      ),
      auto_unbox = TRUE
    )

  out <- request_llm_provider(
    chat_history = dummy_history(),
    request = req,
    stream = TRUE, # ensure SSE path
    verbose = FALSE, # silence token-by-token printing in tests
    api_type = "openai"
  )

  assistant <- tail(out$completed$content, 1)
  reply <- tolower(trimws(assistant))

  expect_true(nchar(reply) > 0)
  expect_true(grepl("\\bpong\\b", reply, ignore.case = FALSE))

  # If you want to be strict:
  # expect_equal(reply, "pong")
})

# ---- Responses API (stream) ---------------------------------------------------
test_that("OpenAI Responses API (stream) accumulates output_text deltas", {
  req <- request("https://api.openai.com/v1/responses") |>
    openai_headers() |>
    req_body_json(
      list(
        model = model_resp,
        temperature = 0,
        max_output_tokens = 32,
        stream = TRUE,
        input = "Reply with exactly the single word: pong"
      ),
      auto_unbox = TRUE
    )

  out <- request_llm_provider(
    chat_history = dummy_history(),
    request = req,
    stream = TRUE, # ensure we take the SSE path
    verbose = FALSE, # the new impl can print token deltas; silence in tests
    api_type = "openai"
  )

  # Grab the assistant's final message from the completed history
  assistant <- tail(out$completed$content, 1)
  reply <- tolower(trimws(assistant))

  # Basic sanity checks
  expect_true(nchar(reply) > 0)
  expect_true(grepl("\\bpong\\b", reply, ignore.case = FALSE))

  # (Optional) If you want to be stricter, uncomment:
  # expect_equal(reply, "pong")
})

test_that("request_llm_provider stream_callback is invoked with metadata", {
  skip_if_not_installed("httr2")

  library(httr2)

  dummy_history <- data.frame(
    role = "user",
    content = "ping",
    stringsAsFactors = FALSE
  )

  # We'll hit a cheap, fast OpenAI endpoint if configured; otherwise skip.
  openai_key <- Sys.getenv("OPENAI_API_KEY", unset = "")
  skip_if(
    openai_key == "",
    "OPENAI_API_KEY not set; skipping stream_callback test"
  )

  model_chat <- Sys.getenv(
    "TIDYPROMPT_OPENAI_CHAT_MODEL",
    unset = "gpt-4o-mini"
  )

  req <- request("https://api.openai.com/v1/chat/completions") |>
    req_headers(
      Authorization = paste("Bearer", openai_key),
      `Content-Type` = "application/json"
    ) |>
    req_body_json(
      list(
        model = model_chat,
        temperature = 0,
        max_tokens = 8,
        stream = TRUE,
        messages = list(
          list(
            role = "system",
            content = "Reply with exactly the single word: pong"
          ),
          list(role = "user", content = "Say it.")
        )
      ),
      auto_unbox = TRUE
    )

  seen <- new.env(parent = emptyenv())
  seen$n_calls <- 0L
  seen$last_chunk <- NULL
  seen$meta_samples <- list()

  cb <- function(chunk, meta) {
    seen$n_calls <- seen$n_calls + 1L
    seen$last_chunk <- chunk
    # Store a tiny sample of metadata for later checks
    seen$meta_samples[[length(seen$meta_samples) + 1L]] <- list(
      api_type = meta$api_type,
      endpoint = meta$endpoint,
      verbose = meta$verbose,
      has_partial = !is.null(meta$partial_response),
      has_latest = !is.null(meta$latest_message)
    )
  }

  out <- request_llm_provider(
    chat_history = dummy_history,
    request = req,
    stream = TRUE,
    verbose = FALSE,
    api_type = "openai",
    stream_callback = cb
  )

  expect_true(nrow(out$completed) == 2L)
  assistant <- tail(out$completed$content, 1)

  expect_true(seen$n_calls > 0L)
  expect_true(is.character(seen$last_chunk))
  expect_true(nchar(seen$last_chunk) > 0)

  # At least one metadata sample has the expected shape
  expect_true(length(seen$meta_samples) > 0L)
  m <- seen$meta_samples[[1L]]
  expect_equal(m$api_type, "openai")
  expect_true(m$has_partial)
  expect_true(m$has_latest)
  expect_true(is.character(assistant))
  expect_true(nchar(assistant) > 0)
})

testthat::test_that("request_llm_provider normalizes sparse tool metadata", {
  testthat::skip_if_not_installed("httr2")

  testthat::local_mocked_bindings(
    req_llm_non_stream = function(req, api_type, verbose) {
      list(
        new = data.frame(
          role = "assistant",
          content = "pong",
          stringsAsFactors = FALSE
        ),
        httr2_response = NULL
      )
    },
    .package = "tidyprompt"
  )

  history <- data.frame(
    role = "user",
    content = "ping",
    tool_call = FALSE,
    tool_call_id = NA_character_,
    tool_result = FALSE,
    stringsAsFactors = FALSE
  )

  result <- request_llm_provider(
    chat_history = history,
    request = httr2::request("https://example.com/v1/chat/completions"),
    stream = FALSE,
    verbose = FALSE,
    api_type = "openai"
  )

  testthat::expect_equal(result$completed$tool_call, c(FALSE, FALSE))
  testthat::expect_true(all(is.na(result$completed$tool_call_id)))
  testthat::expect_equal(result$completed$tool_result, c(FALSE, FALSE))
})

#' @title Tidyprompt R6 Class
#' @name tidyprompt-class
#'
#' @description
#' A [tidyprompt-class] object contains a base prompt and a list
#' of [prompt_wrap()] objects. It provides structured methods to modify the prompt
#' while simultaneously adding logic to extract from and validate the LLM response.
#' Besides a base prompt, a [tidyprompt-class] object may contain a system prompt
#' and a chat history which precede the base prompt.
#'
#' @example inst/examples/tidyprompt.R
#'
#' @family tidyprompt
NULL

#' @rdname tidyprompt-class
#' @export
`tidyprompt-class` <- R6::R6Class(
  "Tidyprompt",
  private = list(
    # Internal validator
    validate_tidyprompt = function() {
      if (!is.character(self$base_prompt) || length(self$base_prompt) != 1) {
        stop("The base prompt must be a single string.", call. = FALSE)
      }

      if (!is.list(private$prompt_wraps)) {
        stop("$prompt_wraps must be a list.", call. = FALSE)
      }

      if (
        length(private$prompt_wraps) > 0 &&
          !all(
            sapply(private$prompt_wraps, function(x) inherits(x, "prompt_wrap"))
          )
      ) {
        stop(
          "All elements of $prompt_wraps must be of class `prompt_wrap`.",
          call. = FALSE
        )
      }

      if (!is.null(private$chat_history)) {
        tryCatch(
          chat_history(private$chat_history),
          error = function(e) {
            stop(
              paste0(
                "The chat history is not valid.\n",
                "Error in `chat_history(private$chat_history)`:\n",
                e$message
              )
            )
          }
        )
      }
    },

    # A list of prompt_wrap objects
    prompt_wraps = list(),

    # A chat history object
    chat_history = NULL
  ),
  public = list(
    #' @field base_prompt
    #' The base prompt string.
    #' The base prompt be modified by prompt wraps during [construct_prompt_text()];
    #' the modified prompt text will be used as the final message of role 'user'
    #' during [send_prompt()]
    base_prompt = NULL,

    #' @field system_prompt
    #' A system prompt string.
    #' This will be added at the start of the chat history as role 'system'
    #' during [send_prompt()]
    system_prompt = NULL,

    #' @description
    #' Initialize a [tidyprompt-class] object
    #'
    #' @details Different types of input are accepted for initialization of
    #' a [tidyprompt-class] object:
    #' \itemize{
    #'  \item A single character string. This will be used as the base prompt
    #'
    #'  \item A dataframe which is a valid chat history (see [chat_history()])
    #'
    #'  \item A list containing a valid chat history under '$chat_history'
    #'  (e.g., a result from [send_prompt()] when using 'return_mode' = "full")
    #'
    #'  \item A [tidyprompt-class] object. This will be checked for validity and, if valid,
    #'   the fields are copied to the object which is returned from this method
    #' }
    #' When passing a dataframe or list with a chat history, the last row of the
    #' chat history must have role 'user'; this row will be used as the base prompt.
    #' If the first row of the chat history has role 'system', it will be used
    #' as the system prompt.
    #'
    #' @param input A string, a chat history, a list containing
    #' a chat history under key '$chat_history', or a [tidyprompt-class] object
    #'
    #' @return A [tidyprompt-class] object
    initialize = function(input) {
      input_must_be <- paste0(
        "Input must be:",
        " a string,",
        " a dataframe which is a valid chat history (see `?chat_history`),",
        " a list containing a valid chat history under key '$chat_history',",
        " or a `tidyprompt-class` object"
      )
      # Turn single string input into base prompt
      if (is.character(input) && length(input) == 1) {
        self$base_prompt <- input
        return(self)
      }

      # Take chat_history from list input
      if (is.list(input) && !is.data.frame(input)) {
        if (is.null(input$chat_history)) {
          stop(input_must_be)
        }
        input <- input$chat_history
      }

      # Take relevant variables from dataframe input
      if (is.data.frame(input)) {
        chat_history <- tryCatch(
          chat_history(input),
          error = function(e) {
            stop(
              paste0(
                "Input for `tidyprompt-class` is a dataframe, but dataframe is not",
                " a valid chat history.\n",
                "Error in `chat_history(input)`:\n",
                e$message,
                "\n",
                "(see `?chat_history`)"
              )
            )
            NULL
          }
        )

        if (is.null(chat_history)) {
          stop(input_must_be)
        }

        # Last row of chat history must be user
        if (tail(chat_history$role, 1) != "user") {
          stop(
            paste0(
              "The last row of the chat history must have role 'user'.\n",
              "Add a message to the chat history with `chat_history_add_msg()`"
            )
          )
        }

        # Extract base prompt from chat history
        self$base_prompt <- tail(chat_history$content, 1)
        chat_history <- chat_history[-nrow(chat_history), ]

        # Extract system prompt from chat history
        if (nrow(chat_history) > 0 && head(chat_history$role, 1) == "system") {
          self$system_prompt <- head(chat_history$content, 1)
          chat_history <- chat_history[-1, ]
        }

        # Add the rest of chat history to field
        private$chat_history <- chat_history

        return(self)
      }

      # Validate a pre-existing tidyprompt object
      if (inherits(input, "Tidyprompt")) {
        if (!input$is_valid()) {
          stop("The input `tidyprompt-class` object is not valid")
        }

        # Copy fields
        self$base_prompt <- input$base_prompt
        self$system_prompt <- input$system_prompt
        private$chat_history <- input$.__enclos_env__$private$chat_history
        private$prompt_wraps <- input$.__enclos_env__$private$prompt_wraps

        return(self)
      }

      stop(input_must_be)
    },

    #' @description
    #' Check if the tidyprompt object is valid.
    #'
    #' @return `TRUE` if valid, otherwise `FALSE`
    is_valid = function() {
      tryCatch(
        {
          private$validate_tidyprompt()
          TRUE
        },
        error = function(e) FALSE
      )
    },

    #' @description
    #' Add a [prompt_wrap()] to the [tidyprompt-class] object.
    #'
    #' @param prompt_wrap A [prompt_wrap()] object
    #'
    #' @return The updated [tidyprompt-class] object
    add_prompt_wrap = function(prompt_wrap) {
      if (!inherits(prompt_wrap, "prompt_wrap")) {
        stop("`prompt_wrap` must be of class `prompt_wrap`", call. = FALSE)
      }
      private$prompt_wraps <- c(private$prompt_wraps, list(prompt_wrap))
      private$validate_tidyprompt()
    },

    #' @description
    #' Get list of [prompt_wrap()] objects from the [tidyprompt-class] object.
    #'
    #' @param order The order to return the wraps. Options are:
    #'   - "default": as originally added to the object
    #'
    #'   - "modification": as ordered for modification of the base prompt;
    #'   ordered by type: check, unspecified, mode, tool, break. This is the order
    #'   in which prompt wraps are applied during [construct_prompt_text()]
    #'
    #'   - "evaluation": ordered for evaluation of the LLM response;
    #'   ordered by type: tool, mode, break, unspecified, check. This is the order
    #'   in which wraps are applied to the LLM output during [send_prompt()]
    #'
    #' @return A list of [prompt_wrap()] objects.
    get_prompt_wraps = function(
      order = c("default", "modification", "evaluation")
    ) {
      private$validate_tidyprompt()
      order <- match.arg(order)
      wraps <- private$prompt_wraps

      # Update the environment of functions in the `prompt_wrap` to include `self`
      functions_to_update <- c(
        "modify_fn",
        "extraction_fn",
        "validation_fn",
        "handler_fn",
        "parameter_fn"
      )
      i <- 0
      for (wrap in wraps) {
        i <- i + 1
        for (fn_name in functions_to_update) {
          if (!is.null(wrap[[fn_name]]) && is.function(wrap[[fn_name]])) {
            original_env <- environment(wrap[[fn_name]])
            new_env <- list2env(list(self = self), parent = original_env)
            fn <- wrap[[fn_name]]
            environment(fn) <- new_env
            wraps[[i]][[fn_name]] <- fn
          }
        }
      }
      rm(i)

      if (length(wraps) == 0) {
        return(list())
      }
      if (order == "default") {
        return(wraps)
      }

      # Categorize wraps
      t_check <- wraps[sapply(wraps, function(x) x$type == "check")]
      t_unspecified <- wraps[sapply(wraps, function(x) x$type == "unspecified")]
      t_mode <- wraps[sapply(wraps, function(x) x$type == "mode")]
      t_tool <- wraps[sapply(wraps, function(x) x$type == "tool")]
      t_break <- wraps[sapply(wraps, function(x) x$type == "break")]

      if (order == "modification") {
        return(c(t_check, t_unspecified, t_break, t_mode, t_tool))
      }
      if (order == "evaluation") {
        return(c(t_tool, t_mode, t_break, t_unspecified, t_check))
      }
    },

    #' @description
    #' Construct the complete prompt text.
    #'
    #' @param llm_provider Optional [llm_provider-class] object.
    #' This may sometimes affect the prompt text construction
    #' @param apply_provider_prompt_wraps Logical. Whether to apply
    #' provider-specific pre/post prompt wraps when constructing the prompt text
    #'
    #' @return A string representing the constructed prompt text
    construct_prompt_text = function(
      llm_provider = NULL,
      apply_provider_prompt_wraps = FALSE
    ) {
      private$validate_tidyprompt()
      prompt_text <- self$base_prompt

      # Add provider-specific pre/post prompt wraps
      if (
        isTRUE(is.function(llm_provider$apply_prompt_wraps)) &&
          isTRUE(apply_provider_prompt_wraps)
      ) {
        modified_self <- llm_provider$apply_prompt_wraps(self)
      } else {
        modified_self <- self
      }

      wraps <- modified_self$get_prompt_wraps(order = "modification")

      for (wrap in wraps) {
        if (!is.null(wrap$modify_fn)) {
          prompt_text <- wrap$modify_fn(prompt_text, llm_provider)
        }
      }

      prompt_text
    },

    #' @description This function sets the chat history for the tidyprompt object.
    #' The chat history will also set the base prompt and system prompt
    #' (the last message of the chat history should be of role 'user' and
    #' will be used as the base prompt; the first message of the chat history
    #' may be of the role 'system' and will then be used as the system prompt).
    #' This may be useful when one wants to change the base prompt, system prompt,
    #' and chat history of a [tidyprompt-class] object while retaining other fields like
    #' the prompt wraps.
    #'
    #' @param chat_history A valid chat history (see [chat_history()])
    #'
    #' @return The updated [tidyprompt-class] object
    set_chat_history = function(chat_history) {
      chat_history <- chat_history(chat_history)

      # Last row of chat history must be user
      if (tail(chat_history$role, 1) != "user") {
        stop(
          paste0(
            "The last row of the chat history must have role 'user'.\n",
            "Add a message to the chat history using `chat_history_add_msg()`"
          )
        )
      }

      self$base_prompt <- tail(chat_history$content, 1)
      # Remove base prompt from chat history
      chat_history <- chat_history[-nrow(chat_history), ]

      # If first row is system message, we will set it as the system prompt
      if (nrow(chat_history) > 0 && head(chat_history$role, 1) == "system") {
        self$system_prompt <- head(chat_history$content, 1)
        # Remove system prompt from chat history
        chat_history <- chat_history[-1, ]
      }

      private$chat_history <- chat_history
    },

    #' @description This function gets the chat history of the [tidyprompt-class] object.
    #' The chat history is constructed from the base prompt, system prompt, and chat
    #' history field. The returned object will be the chat history
    #' with the system prompt as the first message with role 'system' and the
    #' the base prompt as the last message with role 'user'.
    #'
    #' @param llm_provider An optional [llm_provider-class] object.
    #' This may sometimes affect the prompt text construction
    #'
    #' @return A dataframe containing the chat history
    get_chat_history = function(llm_provider = NULL) {
      chat_history_construction <- c(
        role = "system",
        content = self$system_prompt
      ) |>
        dplyr::bind_rows(
          private$chat_history
        ) |>
        dplyr::bind_rows(
          c(
            role = "user",
            content = self$construct_prompt_text(llm_provider)
          )
        )

      # Remove system prompt row when it has no content; keep everything
      # else (including blank user rows needed for image-only prompts).
      chat_history_construction <- chat_history_construction |>
        dplyr::filter(
          .data$role != "system" |
            (.data$content != "" & !is.na(.data$content))
        )

      chat_history <- chat_history(chat_history_construction)
      chat_history
    }
  )
)

#' Create a [tidyprompt-class] object
#'
#' This is a wrapper around the [tidyprompt-class] constructor.
#'
#' @details Different types of input are accepted for initialization of
#' a [tidyprompt-class] object:
#' \itemize{
#'  \item A single character string. This will be used as the base prompt
#'
#'  \item A dataframe which is a valid chat history (see [chat_history()])
#'
#'  \item A list containing a valid chat history under '$chat_history'
#'  (e.g., a result from [send_prompt()] when using 'return_mode' = "full")
#'
#'  \item A [tidyprompt-class] object. This will be checked for validity and, if valid,
#'   the fields are copied to the object which is returned from this method
#' }
#' When passing a dataframe or list with a chat history, the last row of the
#' chat history must have role 'user'; this row will be used as the base prompt.
#' If the first row of the chat history has role 'system', it will be used
#' as the system prompt.
#'
#' @param input A string, a chat history, a list containing
#' a chat history under key '$chat_history', or a [tidyprompt-class] object
#'
#' @return A [tidyprompt-class] object
#'
#' @export
#'
#' @family tidyprompt
#' @example inst/examples/tidyprompt.R
tidyprompt <- function(input) {
  `tidyprompt-class`$new(input)
}

#' Check if object is a [tidyprompt-class] object
#'
#' @param x An object to check
#'
#' @return TRUE if the object is a valid [tidyprompt-class] object, otherwise FALSE
#'
#' @export
#'
#' @family tidyprompt
#' @example inst/examples/tidyprompt.R
is_tidyprompt <- function(x) {
  inherits(x, "Tidyprompt") && x$is_valid()
}

#' Get prompt wraps from a [tidyprompt-class] object
#'
#' @param x A [tidyprompt-class] object
#'
#' @param order The order to return the wraps. Options are:
#'   - "default": as originally added to the object
#'
#'   - "modification": as ordered for modification of the base prompt;
#'   ordered by type: check, unspecified, mode, tool, break. This is the order
#'   in which prompt wraps are applied during [construct_prompt_text()]
#'
#'   - "evaluation": ordered for evaluation of the LLM response;
#'   ordered by type: tool, mode, break, unspecified, check. This is the order
#'   in which wraps are applied to the LLM output during [send_prompt()]
#'
#' @return A list of prompt wrap objects (see [prompt_wrap()])
#'
#' @export
#'
#' @family tidyprompt
#' @example inst/examples/tidyprompt.R
get_prompt_wraps <- function(
  x,
  order = c("default", "modification", "evaluation")
) {
  x <- tidyprompt(x)
  x$get_prompt_wraps(order = order)
}

#' Construct prompt text from a [tidyprompt-class] object
#'
#' @param x A [tidyprompt-class] object
#' @param llm_provider An optional [llm_provider-class] object.
#' This may sometimes affect the prompt text construction
#'
#' @return The constructed prompt text
#'
#' @export
#'
#' @family tidyprompt
#' @example inst/examples/tidyprompt.R
construct_prompt_text <- function(x, llm_provider = NULL) {
  x <- tidyprompt(x)
  x$construct_prompt_text(llm_provider)
}

#' @title
#' Set the chat history of a [tidyprompt-class] object
#'
#' @description
#' This function sets the chat history for a [tidyprompt-class] object.
#' The chat history will also set the base prompt and system prompt
#' (the last message of the chat history should be of role 'user' and
#' will be used as the base prompt; the first message of the chat history
#' may be of the role 'system' and will then be used as the system prompt).
#'
#' This may be useful when one wants to change the base prompt, system prompt,
#' and chat history of a [tidyprompt-class] object while retaining other fields like
#' the list of prompt wraps.
#'
#' @param x A [tidyprompt-class] object
#' @param chat_history A valid chat history (see [chat_history()])
#'
#' @return The updated [tidyprompt-class] object
#'
#' @export
#'
#' @family tidyprompt
#' @seealso [chat_history()]
#'
#' @example inst/examples/tidyprompt.R
set_chat_history <- function(x, chat_history) {
  x <- tidyprompt(x)
  x$set_chat_history(chat_history)
}

#' Get the chat history of a [tidyprompt-class] object
#'
#' This function gets the chat history of the [tidyprompt-class] object.
#' The chat history is constructed from the base prompt, system prompt, and chat
#' history field. The returned object will be the chat history
#' with the system prompt as the first message with role 'system' and the
#' the base prompt as the last message with role 'user'.
#'
#' @param x A [tidyprompt-class] object
#'
#' @return A dataframe containing the chat history
#'
#' @export
#'
#' @family tidyprompt
#' @example inst/examples/tidyprompt.R
#'
#' @seealso [chat_history()]
get_chat_history <- function(x) {
  x <- tidyprompt(x)
  x$get_chat_history()
}

#' Set system prompt of a [tidyprompt-class] object
#'
#' Set the system prompt for a prompt. The system prompt will be added
#' as a message with role 'system' at the start of the chat history when
#' this prompt is evaluated by [send_prompt()].
#'
#' @details The system prompt will be stored in the [tidyprompt()] object
#' as '$system_prompt'.
#'
#' @param prompt A single string or a [tidyprompt()] object
#' @param system_prompt A single character string representing the system prompt
#'
#' @return A [tidyprompt()] with the system prompt set
#'
#' @export
#'
#' @example inst/examples/set_system_prompt.R
#'
#' @family pre_built_prompt_wraps
#' @family miscellaneous_prompt_wraps
set_system_prompt <- function(prompt, system_prompt) {
  prompt <- tidyprompt(prompt)

  if (
    !is.character(system_prompt) ||
      length(system_prompt) != 1
  ) {
    stop("system_prompt must be a single character string")
  }

  prompt$system_prompt <- system_prompt

  return(prompt)
}

test_that("provider-level pre/post wraps modify initial user message via send_prompt()", {
  fake <- llm_provider_fake(verbose = FALSE)

  # Ensure a clean slate
  fake$pre_prompt_wraps <- list()
  fake$post_prompt_wraps <- list()

  # Add provider-level wraps
  fake$add_prompt_wrap(
    provider_prompt_wrap(
      modify_fn = \(txt, llm, http) paste0("PRE(", txt, ")"),
      type = "unspecified",
      name = "prov-pre"
    ),
    position = "pre"
  )
  fake$add_prompt_wrap(
    provider_prompt_wrap(
      modify_fn = \(txt, llm, http) paste0("POST(", txt, ")"),
      type = "unspecified",
      name = "prov-post"
    ),
    position = "post"
  )

  # Prompt-level wrap too, to check ordering (pre -> prompt -> post)
  pr <- prompt_wrap(
    "Hi",
    modify_fn = \(txt, llm, http) paste0("PROMPT(", txt, ")"),
    type = "unspecified",
    name = "prompt-wrap"
  )

  out <- send_prompt(pr, fake, return_mode = "full")

  first_user <- subset(out$chat_history, role == "user")
  first_user <- first_user[1, , drop = FALSE]

  expect_equal(first_user$content, "POST(PROMPT(PRE(Hi)))")
})

test_that("provider-level handler_fn runs during completion", {
  fake <- llm_provider_fake(verbose = FALSE)

  fake$pre_prompt_wraps <- list()
  fake$post_prompt_wraps <- list()

  fake$add_prompt_wrap(
    provider_prompt_wrap(
      handler_fn = function(response, llm_provider) {
        response$completed$content[nrow(
          response$completed
        )] <- "provider handled"
        response
      },
      name = "prov-handler"
    ),
    position = "pre"
  )

  res <- send_prompt("Hi there", fake)
  expect_identical(res, "provider handled")
})

test_that("provider-level parameter_fn is applied before first completion", {
  prov <- `llm_provider-class`$new(
    complete_chat_function = function(chat_history) {
      # 'self' is the provider; parameter_fn should have set self$parameters$marker
      add <- data.frame(
        role = "assistant",
        content = paste0("marker:", self$parameters$marker),
        tool_result = FALSE
      )
      completed <- rbind(chat_history, add)
      list(
        completed = completed,
        http = list(request = NULL, response = NULL)
      )
    },
    verbose = FALSE
  )

  prov$add_prompt_wrap(
    provider_prompt_wrap(
      parameter_fn = function(llm_provider) list(marker = "XYZ"),
      name = "prov-params"
    ),
    position = "pre"
  )

  res <- send_prompt("Hello", prov)
  expect_identical(res, "marker:XYZ")
})

test_that("provider-level pre and post handlers both run", {
  fake <- llm_provider_fake(verbose = FALSE)

  fake$pre_prompt_wraps <- list()
  fake$post_prompt_wraps <- list()

  fake$add_prompt_wrap(
    provider_prompt_wrap(
      handler_fn = function(response, llm_provider) {
        response$completed$content[nrow(response$completed)] <-
          paste0(response$completed$content[nrow(response$completed)], " +PREH")
        response
      },
      name = "prov-pre-handler"
    ),
    position = "pre"
  )

  fake$add_prompt_wrap(
    provider_prompt_wrap(
      handler_fn = function(response, llm_provider) {
        response$completed$content[nrow(response$completed)] <-
          paste0(
            response$completed$content[nrow(response$completed)],
            " +POSTH"
          )
        response
      },
      name = "prov-post-handler"
    ),
    position = "post"
  )

  res <- send_prompt("Ping", fake)
  expect_true(grepl("\\+PREH", res))
  expect_true(grepl("\\+POSTH", res))
})

test_that("provider_prompt_wrap enforces 'check' type constraints", {
  expect_error(
    provider_prompt_wrap(
      type = "check",
      modify_fn = \(x) x,
      validation_fn = \(x, llm, http) TRUE
    ),
    "only validation_fn is allowed"
  )

  expect_no_error(
    provider_prompt_wrap(
      type = "check",
      validation_fn = \(x, llm, http) TRUE
    )
  )
})

test_that("apply_prompt_wraps attaches named provider wraps to a tidyprompt", {
  fake <- llm_provider_fake(verbose = FALSE)

  fake$pre_prompt_wraps <- list()
  fake$post_prompt_wraps <- list()

  fake$add_prompt_wrap(
    provider_prompt_wrap(
      modify_fn = \(txt, llm, http) txt,
      type = "unspecified",
      name = "prov-pre-a"
    ),
    position = "pre"
  )
  fake$add_prompt_wrap(
    provider_prompt_wrap(
      modify_fn = \(txt, llm, http) txt,
      type = "unspecified",
      name = "prov-post-b"
    ),
    position = "post"
  )

  tp <- fake$apply_prompt_wraps(tidyprompt("X"))
  wraps <- tp$get_prompt_wraps(order = "modification")
  wrap_names <- vapply(wraps, \(w) w$name %||% NA_character_, character(1))

  expect_true(any(wrap_names == "prov-pre-a"))
  expect_true(any(wrap_names == "prov-post-b"))
})


test_that("construct_prompt_text applies provider pre/post wraps around prompt wraps in correct type order", {
  fake <- llm_provider_fake(verbose = FALSE)
  fake$pre_prompt_wraps <- list()
  fake$post_prompt_wraps <- list()

  # Provider-level wraps:
  # order for modification: check, unspecified, break, mode, tool
  fake$add_prompt_wrap(
    provider_prompt_wrap(
      modify_fn = \(txt, llm, http) paste(txt, "[preU]"),
      type = "unspecified",
      name = "prov-pre-unspecified"
    ),
    position = "pre"
  )
  fake$add_prompt_wrap(
    provider_prompt_wrap(
      modify_fn = \(txt, llm, http) paste(txt, "[preB]"),
      type = "break",
      name = "prov-pre-break"
    ),
    position = "pre"
  )
  fake$add_prompt_wrap(
    provider_prompt_wrap(
      modify_fn = \(txt, llm, http) paste(txt, "[postM]"),
      type = "mode",
      name = "prov-post-mode"
    ),
    position = "post"
  )
  fake$add_prompt_wrap(
    provider_prompt_wrap(
      modify_fn = \(txt, llm, http) paste(txt, "[postT]"),
      type = "tool",
      name = "prov-post-tool"
    ),
    position = "post"
  )

  # Prompt-level wraps
  pr <- "Hi" |>
    prompt_wrap(
      modify_fn = \(txt, llm, http) paste(txt, "[promptU]"),
      type = "unspecified",
      name = "prompt-unspecified"
    ) |>
    prompt_wrap(
      modify_fn = \(txt, llm, http) paste(txt, "[promptB]"),
      type = "break",
      name = "prompt-break"
    )

  # Apply provider wraps, then build text
  pr <- fake$apply_prompt_wraps(pr)
  built <- construct_prompt_text(pr, llm_provider = fake)

  # Expected order:
  # unspecified: prov-pre -> prompt
  # break:       prov-pre -> prompt
  # mode:        prov-post
  # tool:        prov-post
  expect_identical(
    built,
    "Hi [preU] [promptU] [preB] [promptB] [postM] [postT]"
  )
})

test_that("construct_prompt_text: unspecified wraps keep relative order: pre -> prompt -> post", {
  fake <- llm_provider_fake(verbose = FALSE)
  fake$pre_prompt_wraps <- list()
  fake$post_prompt_wraps <- list()

  fake$add_prompt_wrap(
    provider_prompt_wrap(
      modify_fn = \(txt, llm, http) paste(txt, "Apre"),
      type = "unspecified",
      name = "Apre"
    ),
    position = "pre"
  )
  pr <- "Hi" |>
    prompt_wrap(
      modify_fn = \(txt, llm, http) paste(txt, "Bprompt"),
      type = "unspecified",
      name = "Bprompt"
    )
  fake$add_prompt_wrap(
    provider_prompt_wrap(
      modify_fn = \(txt, llm, http) paste(txt, "Cpost"),
      type = "unspecified",
      name = "Cpost"
    ),
    position = "post"
  )

  pr <- fake$apply_prompt_wraps(pr)
  built <- construct_prompt_text(pr, llm_provider = fake)
  expect_identical(built, "Hi Apre Bprompt Cpost")
})

test_that("construct_prompt_text provider modify_fn receives llm_provider", {
  fake <- llm_provider_fake(verbose = FALSE)
  fake$pre_prompt_wraps <- list()
  fake$post_prompt_wraps <- list()

  fake$add_prompt_wrap(
    provider_prompt_wrap(
      modify_fn = function(txt, llm_provider, http) {
        if (inherits(llm_provider, "LlmProvider")) {
          return("text for fake provider via provider wrap")
        }
        return("other")
      },
      type = "unspecified"
    ),
    position = "pre"
  )

  pr <- tidyprompt("ignored")
  pr <- fake$apply_prompt_wraps(pr)
  built <- construct_prompt_text(pr, llm_provider = fake)
  expect_identical(built, "text for fake provider via provider wrap")
})

test_that("multiple provider wraps of same type preserve insertion order", {
  fake <- llm_provider_fake(verbose = FALSE)
  fake$pre_prompt_wraps <- list()
  fake$post_prompt_wraps <- list()

  fake$add_prompt_wrap(
    provider_prompt_wrap(
      modify_fn = \(txt, llm, http) paste(txt, "one"),
      type = "unspecified",
      name = "one"
    ),
    position = "pre"
  )
  fake$add_prompt_wrap(
    provider_prompt_wrap(
      modify_fn = \(txt, llm, http) paste(txt, "two"),
      type = "unspecified",
      name = "two"
    ),
    position = "pre"
  )

  pr <- tidyprompt("Hi")
  pr <- fake$apply_prompt_wraps(pr)
  built <- construct_prompt_text(pr, llm_provider = fake)
  expect_identical(built, "Hi one two")
})

test_that("apply_prompt_wraps preserves system prompt and prior chat history", {
  fake <- llm_provider_fake(verbose = FALSE)
  fake$pre_prompt_wraps <- list()
  fake$post_prompt_wraps <- list()
  fake$add_prompt_wrap(
    provider_prompt_wrap(modify_fn = \(txt, llm, http) paste0(txt, " [prov]")),
    position = "pre"
  )

  ch <- create_chat_df(
    role = c("system", "assistant", "user", "assistant", "user"),
    content = c("SYS", "A1", "U1", "A2", "U2"),
    tool_result = c(FALSE, FALSE, FALSE, FALSE, FALSE)
  )

  orig <- tidyprompt(ch)
  wrapped <- fake$apply_prompt_wraps(orig)

  # System prompt untouched
  expect_identical(orig$system_prompt, "SYS")
  expect_identical(wrapped$system_prompt, "SYS")

  # Non-final rows identical
  gh_o <- orig$get_chat_history()
  gh_w <- wrapped$get_chat_history(fake) # last user line will include [prov]
  expect_identical(gh_o[-nrow(gh_o), ], gh_w[-nrow(gh_w), ])

  # Only the last user message is modified by provider wraps
  expect_match(utils::tail(gh_w$content, 1), "\\[prov\\]$")
})

test_that("handler_fns order and loop semantics work", {
  prov <- `llm_provider-class`$new(
    complete_chat_function = function(chat_history) {
      add <- data.frame(role = "assistant", content = "ok", tool_result = FALSE)
      list(
        completed = rbind(chat_history, add),
        http = list(request = NULL, response = NULL)
      )
    },
    verbose = FALSE
  )

  seq_vec <- character(0)
  # provider-pre
  prov$add_prompt_wrap(
    provider_prompt_wrap(
      handler_fn = function(resp, llm) {
        seq_vec <<- c(seq_vec, "pre")
        resp$done <- FALSE # force another loop
        resp
      }
    ),
    position = "pre"
  )

  # prompt-level
  pr <- prompt_wrap(
    "X",
    handler_fn = function(resp, llm) {
      seq_vec <<- c(seq_vec, "prompt")
      resp
    },
    type = "unspecified"
  )

  # provider-post that ends the loop on 2nd pass
  pass <- 0L
  prov$add_prompt_wrap(
    provider_prompt_wrap(
      handler_fn = function(resp, llm) {
        seq_vec <<- c(seq_vec, "post")
        pass <<- pass + 1L
        resp$done <- pass >= 2L
        resp
      }
    ),
    position = "post"
  )

  send_prompt(pr, prov)
  expect_identical(seq_vec, c("pre", "prompt", "post", "pre", "prompt", "post"))
})

test_that("evaluation order across types is tool→mode→break→unspecified→check", {
  order <- character()
  # Fake provider that just echoes
  prov <- llm_provider_fake(verbose = FALSE)

  pr <- "base" |>
    prompt_wrap(
      extraction_fn = \(x, llm, http) {
        order <<- c(order, "tool")
        x
      },
      type = "tool"
    ) |>
    prompt_wrap(
      extraction_fn = \(x, llm, http) {
        order <<- c(order, "mode")
        x
      },
      type = "mode"
    ) |>
    prompt_wrap(
      extraction_fn = \(x, llm, http) {
        order <<- c(order, "break")
        x
      },
      type = "break"
    ) |>
    prompt_wrap(extraction_fn = \(x, llm, http) {
      order <<- c(order, "unspecified")
      x
    }) |>
    prompt_wrap(
      validation_fn = \(x, llm, http) {
        order <<- c(order, "check")
        TRUE
      },
      type = "check"
    )

  send_prompt(pr, prov)
  expect_identical(order, c("tool", "mode", "break", "unspecified", "check"))
})


test_that("ensure_three_arguments works for 1-arg funcs", {
  prov <- llm_provider_fake(verbose = FALSE)
  pr <- prompt_wrap(
    "Z",
    extraction_fn = function(resp) resp, # only 1 arg
    validation_fn = function(resp) TRUE # only 1 arg
  )
  expect_no_error(send_prompt(pr, prov))
})

test_that("parameter_fn must take exactly one argument", {
  expect_error(
    prompt_wrap("x", parameter_fn = function(a, b) list(foo = 1)),
    "takes one argument"
  )
  expect_error(
    provider_prompt_wrap(parameter_fn = function(a, b) list(bar = 2)),
    "must take one argument"
  )
})

test_that("apply_prompt_wraps returns a new prompt; original unchanged", {
  fake <- llm_provider_fake(verbose = FALSE)
  fake$pre_prompt_wraps <- list()
  fake$add_prompt_wrap(
    provider_prompt_wrap(name = "prov", modify_fn = function(x) {
      return(x)
    }),
    position = "pre"
  )

  orig <- tidyprompt("hello")
  wrapped <- fake$apply_prompt_wraps(orig)

  orig_names <- vapply(
    get_prompt_wraps(orig),
    \(w) w$name %||% NA_character_,
    ""
  )
  wrap_names <- vapply(
    get_prompt_wraps(wrapped),
    \(w) w$name %||% NA_character_,
    ""
  )

  expect_false("prov" %in% orig_names)
  expect_true("prov" %in% wrap_names)
})

test_that("can set sytem prompt", {
  prompt <- "hi" |>
    set_system_prompt("you only say 'beep boop'")

  expect_equal(prompt$system_prompt, "you only say 'beep boop'")

  chat_history <- prompt$get_chat_history()
  expect_equal(chat_history$role[1], "system")
  expect_equal(chat_history$content[1], "you only say 'beep boop'")
})

test_that("create tidyprompt", {
  prompt <- "Hi" |> tidyprompt()
  expect_s3_class(prompt, "Tidyprompt")
})

test_that("construct simple prompt text", {
  prompt <- "Hi" |> tidyprompt()
  expect_equal(construct_prompt_text(prompt), "Hi")
})

test_that("construct more complex prompt text", {
  prompt <- "Hi" |> add_text("How are you?", sep = "\n\n", position = "after")
  expect_equal(
    construct_prompt_text(prompt),
    "Hi\n\nHow are you?"
  )
})

test_that("construct prompt text based on llm_provider", {
  fake <- llm_provider_fake()

  prompt <- "Hi" |>
    prompt_wrap(
      modify_fn = function(x) {
        if (!is.null(llm_provider)) {
          if (!inherits(llm_provider, "LlmProvider")) {
            stop("llm_provider passed to modify_fn is not LlmProvider object")
          }
        }

        if (isTRUE(llm_provider$api_type == "fake")) {
          return("text for fake llm provider")
        }

        return("text for other llm provider")
      }
    )

  text_fake <- construct_prompt_text(prompt, llm_provider = fake)
  text_other <- construct_prompt_text(prompt)

  expect_identical(text_fake, "text for fake llm provider")
  expect_identical(text_other, "text for other llm provider")
})

test_that("create tidyprompt from chat_history object", {
  prompt <- "Hi" |> add_text("How are you?", sep = "\n\n", position = "after")

  chat_history <- prompt$get_chat_history()

  # Expected tidyprompt behavior: construct prompt from chat history
  prompt2 <- tidyprompt(chat_history)

  # Check class
  expect_s3_class(prompt2, "Tidyprompt")

  # Check text
  expect_equal(
    construct_prompt_text(prompt2),
    "Hi\n\nHow are you?"
  )
})

test_that("setting and getting chat history returns consistent results", {
  # Example chat history
  chat_history <- dplyr::tibble(
    role = c("user", "assistant", "user"),
    content = c("What is 5 + 5?", "10", "And what is 5 + 6?"),
    tool_result = c(FALSE, FALSE, NA)
  ) |>
    chat_history()

  # Create Tidyprompt object and set chat history
  # Initial prompt should not show up.
  prompt <- tidyprompt("Initial prompt")
  prompt$set_chat_history(chat_history)

  # Get chat history and compare
  retrieved_chat_history <- prompt$get_chat_history()
  expect_equal(retrieved_chat_history, chat_history)
})

test_that("can create tidyprompt from tidyprompt", {
  prompt <- tidyprompt("hi") |>
    answer_as_integer() |>
    quit_if()
  prompt2 <- tidyprompt(prompt)

  expect_identical(prompt, prompt2)
})

#' Make LLM answer as a list of key-value pairs
#'
#' This function is similar to `answer_as_list()` but instead of returning
#' a list of items, it instructs the LLM to return a list of key-value pairs.
#'
#' @param prompt A single string or a [tidyprompt()] object
#' @param key_name (optional) A name or placeholder describing the "key" part
#'   of each pair
#' @param value_name (optional) A name or placeholder describing the "value" part
#'   of each pair
#' @param pair_explanation (optional) Additional explanation of what a pair should
#'   be. It should be a single string. It will be appended after the list instruction.
#' @param n_unique_items (optional) Number of unique key-value pairs required in the list
#' @param list_mode (optional) Mode of the list: "bullet" or "comma".
#'   - "bullet" mode expects pairs like:
#'     ```
#'     -- key1: value1
#'     -- key2: value2
#'     ```
#'   - "comma" mode expects pairs like:
#'     ```
#'     1. key: value, 2. key: value, etc.
#'     ```
#'
#' @return A [tidyprompt()] with an added [prompt_wrap()] which
#'   will ensure that the LLM response is a list of key-value pairs.
#'
#' @export
#'
#' @example inst/examples/answer_as_key_value.R
answer_as_key_value <- function(
  prompt,
  key_name = "key",
  value_name = "value",
  pair_explanation = NULL,
  n_unique_items = NULL,
  list_mode = c("bullet", "comma")
) {
  prompt <- tidyprompt(prompt)
  stopifnot(
    is.character(key_name),
    length(key_name) == 1,
    is.character(value_name),
    length(value_name) == 1,
    is.null(pair_explanation) ||
      (is.character(pair_explanation) & length(pair_explanation) == 1),
    is.null(n_unique_items) || (is.numeric(n_unique_items) & n_unique_items > 0)
  )
  list_mode <- match.arg(list_mode)

  # Construct the example instructions
  if (list_mode == "bullet") {
    list_instruction <- glue::glue(
      "Respond with a list of key-value pairs, like so:\n",
      "  -- <<{key_name} 1>>: <<{value_name} 1>>",
      .trim = FALSE
    )

    if (is.null(n_unique_items) || n_unique_items > 1) {
      list_instruction <- glue::glue(
        "{list_instruction}\n",
        "  -- <<{key_name} 2>>: <<{value_name} 2>>",
        .trim = FALSE
      )
    }

    if (is.null(n_unique_items) || n_unique_items > 2) {
      list_instruction <- glue::glue(
        "{list_instruction}\n",
        "  etc.",
        .trim = FALSE
      )
    }
  } else if (list_mode == "comma") {
    list_instruction <- glue::glue(
      "Respond with a list of key-value pairs, like so:\n",
      "  1. <<{key_name}>>: <<{value_name}>>, 2. <<{key_name}>>: <<{value_name}>>, etc.",
      .trim = FALSE
    )
  }

  if (!is.null(n_unique_items)) {
    pair_or_pairs <- ifelse(n_unique_items == 1, "pair", "pairs")
    list_instruction <- glue::glue(
      "{list_instruction}\n",
      "The list should contain {n_unique_items} unique {pair_or_pairs}.",
      .trim = FALSE
    )
  }

  if (!is.null(pair_explanation)) {
    list_instruction <- glue::glue(
      "{list_instruction}\n\n",
      "{pair_explanation}",
      .trim = FALSE
    )
  }

  modify_fn <- function(original_prompt_text) {
    glue::glue(
      "{original_prompt_text}\n\n",
      "{list_instruction}",
      .trim = FALSE
    )
  }

  extraction_fn <- function(response) {
    if (!is.character(response)) {
      stop(
        paste0(
          "Response to extract must be a string. ",
          "Ensure that your LLM response is suitable for key-value pair extraction."
        )
      )
    }

    # Extract pairs depending on list_mode
    if (list_mode == "bullet") {
      # Pattern to capture lines starting with `-- ` followed by `key: value`
      pairs <- stringr::str_extract_all(
        response,
        "--\\s*([^:]+):\\s*([^\\n]+)"
      )[[1]]
      if (length(pairs) == 0) {
        return(
          llm_feedback(
            glue::glue(
              "Could not parse any key-value pairs from your response.\n\n",
              "{list_instruction}"
            )
          )
        )
      }

      # Split each match into key and value
      kv_pattern <- "^--\\s*([^:]+):\\s*(.*)$"
      keys <- stringr::str_replace(pairs, kv_pattern, "\\1") |>
        stringr::str_trim()
      vals <- stringr::str_replace(pairs, kv_pattern, "\\2") |>
        stringr::str_trim()
    } else if (list_mode == "comma") {
      # For comma mode, we need to handle cases where there might be commas or periods
      # Adjusted regex pattern to allow commas and periods at the end of the line
      matches <- stringr::str_match_all(
        response,
        "\\d+\\.\\s*([^:]+):\\s*([^,\\n.]+)"
      )[[1]]
      if (nrow(matches) == 0) {
        return(
          llm_feedback(
            glue::glue(
              "Could not parse any key-value pairs from your response.\n\n",
              "{list_instruction}"
            )
          )
        )
      }

      # Extract keys and values
      keys <- stringr::str_trim(matches[, 2])
      vals <- stringr::str_trim(matches[, 3])
    }

    # Create a data frame for the key-value pairs
    df <- data.frame(key = keys, value = vals, stringsAsFactors = FALSE)
    # Ensure uniqueness by key
    df <- df[!duplicated(df$key), ]

    # Check if the number of unique items matches the expected count
    if (!is.null(n_unique_items) && nrow(df) != n_unique_items) {
      return(
        llm_feedback(
          glue::glue(
            "The number of unique key-value pairs should be {n_unique_items}.\n\n",
            "{list_instruction}"
          )
        )
      )
    }

    # Return as a named list
    result_list <- as.list(df$value)
    names(result_list) <- df$key
    return(result_list)
  }

  prompt_wrap(
    prompt,
    modify_fn,
    extraction_fn,
    name = "answer_as_key_value_list"
  )
}

#' @title Make LLM answer as JSON (with optional schema; structured output)
#'
#' @description This functions wraps a prompt with settings that ensure the LLM response
#' is a valid JSON object, optionally matching a given JSON schema
#' (also known as 'structured output'/'structured data'). Users
#' may provide either an 'ellmer' type (e.g. [ellmer::type_object()];
#' see ['ellmer' documentation](https://ellmer.tidyverse.org/articles/structured-data.html)) or
#' a JSON schema (as R list object) to enforce structure on the response.
#'
#' The function can work with all models and providers through text-based
#' handling, but also supports native settings for the OpenAI, Ollama,
#' and various 'ellmer' types of LLM providers. (See argument 'type'.)
#' This means that it is possible to easily switch between providers with
#' different levels of structured output support, while always ensuring
#' the response will be in the desired format.
#'
#' @param prompt A single string or a [tidyprompt()] object
#'
#' @param schema Either a R list object which represents
#' a JSON schema that the response should match, or an 'ellmer' definition of
#' structured data (e.g., [ellmer::type_object()]; see ['ellmer' documentation](https://ellmer.tidyverse.org/articles/structured-data.html))
#'
#' @param schema_strict If TRUE, the provided schema will be strictly enforced.
#' This option is passed as part of the schema when using type
#' "openai", "ollama", or "ellmer", and when using "ollama_oo", "openai_oo", or "text-based"
#' it is passed to [jsonvalidate::json_validate()] when validating the response
#'
#' @param schema_in_prompt_as If providing a schema and
#' when using type "text-based", "openai_oo", or "ollama_oo", this argument specifies
#' how the schema should be included in the prompt:
#' \itemize{
#' \item "example" (default): The schema will be included as an example JSON object
#' (tends to work best). [r_json_schema_to_example()] is used to generate the example object
#' from the schema
#' \item "schema": The schema will be included as a JSON schema
#' }
#' @param type The way that JSON response should be enforced:
#' \itemize{
#' \item "auto": Automatically determine the type based on 'llm_provider$api_type'
#' or 'llm_provider$json_type' (if set; 'json_type' overrides 'api_type' determination).
#' This may not always consider model compatibility and could lead to errors;
#' set 'type' manually if errors occur; use 'text-based' if unsure
#' \item "text-based": Instruction will be added to the prompt
#' asking for JSON; when a schema is provided, this will also be included
#' in the prompt (see argument 'schema_in_prompt_as'). JSON will be parsed
#' from the LLM response and, when a schema is provided, it will be validated
#' against the schema with [jsonvalidate::json_validate()]. Feedback is sent to the
#' LLM when the response is not valid. This option always works, but may in some
#' cases may be less powerful than the other native JSON options
#' \item "openai" and "ollama": The response format will be set via the relevant API parameters,
#' making the API enforce a valid JSON response. If a schema is provided,
#' it will also be included in the API parameters and also be enforced by the API.
#' When no schema is provided, a request for JSON is added to the prompt (as required
#' by the APIs). Note that these JSON options may not be available for all models
#' of your provider; consult their documentation for more information.
#' If you are unsure or encounter errors, use "text-based"
#' \item "openai_oo" and "ollama_oo": Similar to "openai" and "ollama", but if a
#' schema is provided it is not included in the API parameters. Schema validation
#' will be done in R with [jsonvalidate::json_validate()]. This can be useful if
#' you want to use the API's JSON support, but their schema support is limited
#' \item "ellmer": A parameter will be added to the LLM provider, indicating
#' that the response should be structured according to the provided schema.
#' The native 'ellmer' `chat$chat_structured()` function will then be used to obtain
#' the response. This type only useful when using an `llm_provider_ellmer()` object.
#' When `type` is set to "auto" and the `llm_provider` is an 'ellmer' LLM provider,
#' this type will be automatically selected (so it should not be necessary to set
#' this option manually)
#' }
#' Note that the "openai" and "ollama" types may also work for other APIs with a
#'  similar structure.
#' Note furthermore that the "ellmer" type is still experimental and conversion
#'  between 'ellmer' schemas and R list schemas might contain bugs.
#'
#' @return A [tidyprompt()] with an added [prompt_wrap()] which will ensure
#' that the LLM response is a valid JSON object. Note that the prompt wrap
#' will parse the JSON response and return it as an R object (usually a list)
#'
#' @export
#'
#' @example inst/examples/answer_as_json.R
#'
#' @family pre_built_prompt_wraps
#' @family answer_as_prompt_wraps
#' @family json
answer_as_json <- function(
  prompt,
  schema = NULL,
  schema_strict = FALSE,
  schema_in_prompt_as = c("example", "schema"),
  type = c(
    "auto",
    "text-based",
    "openai",
    "ollama",
    "openai_oo",
    "ollama_oo",
    "ellmer"
  )
) {
  prompt <- tidyprompt(prompt)
  schema_in_prompt_as <- match.arg(schema_in_prompt_as)
  type <- match.arg(type)

  # Accept either JSON Schema list or ellmer::type_* object
  if (!is.null(schema) && !(is.list(schema) || is_ellmer_type(schema))) {
    stop("The 'schema' must be a JSON Schema list or an ellmer::type_* object.")
  }
  if (is.list(schema) && length(schema) == 0) {
    stop("The 'schema' list must be non-empty.")
  }

  original_schema <- schema
  if (is.list(schema)) {
    # Inherit 'strict' if present in a wrapper and user didn't set TRUE explicitly
    if (
      isFALSE(schema_strict) &&
        !is.null(schema$strict) &&
        is.logical(schema$strict)
    ) {
      schema_strict <- isTRUE(schema$strict)
    }
    schema <- unwrap_json_schema(schema)
  }

  if (type == "auto" && getOption("tidyprompt.warn.auto.json", TRUE)) {
    cli::cli_alert_warning(
      paste0(
        "{.strong `tidyprompt::answer_as_json()`}:\n",
        "* Automatically determining type based on 'llm_provider$api_type' ",
        "(or 'llm_provider$json_type' if set); this may not consider model compatibility\n",
        "* Manually set 'type' or 'llm_provider$json_type' if errors occur ",
        "(\"text-based\" always works)\n",
        "* Use `options(tidyprompt.warn.auto.json = FALSE)` to suppress this warning"
      )
    )
  }

  determine_type <- function(llm_provider = NULL) {
    if (type != "auto") {
      return(type)
    }
    valid_types <- c(
      "text-based",
      "openai",
      "ollama",
      "openai_oo",
      "ollama_oo",
      "ellmer"
    )
    provider_json_type <- llm_provider[["json_type"]]
    if (
      !is.null(provider_json_type) &&
        !identical(provider_json_type, "auto") &&
        provider_json_type %in% valid_types
    ) {
      return(provider_json_type)
    }
    provider_api_type <- llm_provider[["api_type"]]
    if (isTRUE(provider_api_type == "openai")) {
      return("openai")
    }
    if (isTRUE(provider_api_type == "ollama")) {
      return("ollama")
    }
    if (isTRUE(provider_api_type == "ellmer")) {
      return("ellmer")
    }
    "text-based"
  }

  # Normalize once so we can use either representation everywhere
  sch <- tryCatch(
    normalize_schema_dual(schema, strict = schema_strict),
    error = function(e) list(json_schema = NULL, ellmer_type = NULL)
  )

  # Will be filled by modify_fn if we inject schema details into the prompt
  schema_instruction <- NULL

  parameter_fn <- function(llm_provider) {
    t <- determine_type(llm_provider)

    if (t == "ellmer") {
      if (!is.null(sch$ellmer_type)) {
        return(list(.ellmer_structured_type = sch$ellmer_type))
      } else {
        cli::cli_alert_warning(
          "{.strong `answer_as_json()`}: ellmer type not available; falling back."
        )
        return(NULL)
      }
    }

    if (t == "ollama") {
      # Native JSON mode; include JSON Schema if we have it
      if (is.null(schema)) {
        return(list(format = "json"))
      }
      if (!is.null(sch$json_schema)) {
        js <- sch$json_schema
        js$strict <- schema_strict
        return(list(format = js))
      }
      return(list(format = "json"))
    }

    if (t == "openai") {
      # Native JSON mode; include JSON Schema if we have it
      if (is.null(schema)) {
        return(list(response_format = list(type = "json_object")))
      }
      if (!is.null(sch$json_schema)) {
        json_schema <- list(
          name = "schema",
          schema = sch$json_schema,
          strict = schema_strict
        )
        return(list(
          response_format = list(
            type = "json_schema",
            json_schema = json_schema
          )
        ))
      }
      return(list(response_format = list(type = "json_object")))
    }

    if (t == "ollama_oo") {
      return(list(format = "json"))
    }
    if (t == "openai_oo") {
      return(list(response_format = list(type = "json_object")))
    }

    NULL
  }

  modify_fn <- function(prompt_text, llm_provider) {
    t <- determine_type(llm_provider)

    # If native schema support is in use AND we actually have the right form,
    # don't inject schema into the prompt.
    if (!is.null(schema)) {
      if (t == "ellmer" && !is.null(sch$ellmer_type)) {
        return(prompt_text)
      }
      if (t == "openai" && !is.null(sch$json_schema)) {
        return(prompt_text)
      }
      if (t == "ollama" && !is.null(sch$json_schema)) return(prompt_text)
    }

    # Otherwise, enforce JSON via instructions (text-based style),
    # optionally embedding a JSON example or schema (requires JSON Schema form).
    prompt_text <- glue::glue(
      "{prompt_text}\n\nYou must format your response as a JSON object."
    )

    if (!is.null(schema) && !is.null(sch$json_schema)) {
      jsonvalidate_installed() # warn if missing when we later validate

      if (schema_in_prompt_as == "example") {
        schema_instruction <<- paste0(
          "Your JSON object should match this example JSON object:\n",
          jsonlite::toJSON(
            r_json_schema_to_example(sch$json_schema),
            auto_unbox = TRUE,
            pretty = TRUE
          )
        )
      } else {
        # "schema"
        schema_instruction <<- paste0(
          "Your JSON object should match this JSON schema:\n",
          jsonlite::toJSON(sch$json_schema, auto_unbox = TRUE, pretty = TRUE)
        )
      }
      prompt_text <- paste0(prompt_text, "\n\n", schema_instruction)
    }

    prompt_text
  }

  extraction_fn <- function(llm_response, llm_provider) {
    t <- determine_type(llm_provider)

    # In ellmer structured mode, use the native result directly if available,
    # avoiding lossy JSON round-tripping.
    if (
      isTRUE(t == "ellmer") && !is.null(schema) && !is.null(sch$ellmer_type)
    ) {
      native <- llm_provider$parameters$.native_structured_result
      if (!is.null(native)) {
        # Clear it so it's not reused on retries
        llm_provider$parameters$.native_structured_result <- NULL
        return(native)
      }
      # Fallback: parse from the transcript text
      return(llm_response |> jsonlite::fromJSON())
    }

    jsons <- extraction_fn_json(llm_response)
    if (length(jsons) == 0) {
      return(llm_feedback("You must respond as a valid JSON object."))
    }
    if (length(jsons) == 1) {
      jsons <- jsons[[1]]
    }

    # If we're not using provider-side schema enforcement, and we have a JSON Schema,
    # validate locally.
    if (
      !is.null(schema) &&
        !isTRUE(t %in% c("openai", "ollama")) &&
        !is.null(sch$json_schema) &&
        jsonvalidate_installed()
    ) {
      answer_json <- jsonlite::toJSON(jsons, auto_unbox = TRUE, pretty = TRUE)
      schema_json <- jsonlite::toJSON(
        sch$json_schema,
        auto_unbox = TRUE,
        pretty = TRUE
      )

      validation_result <- jsonvalidate::json_validate(
        answer_json,
        schema_json,
        strict = schema_strict,
        verbose = TRUE
      )

      if (!validation_result) {
        error_details <- attr(validation_result, "errors")
        return(
          llm_feedback(
            paste0(
              "Your response did not match the expected JSON schema.\n\n",
              df_to_string(error_details),
              if (!is.null(schema_instruction)) {
                paste0("\n\n", schema_instruction)
              } else {
                ""
              }
            )
          )
        )
      }
    }

    jsons
  }

  prompt_wrap(
    prompt,
    modify_fn,
    extraction_fn,
    NULL,
    NULL,
    parameter_fn,
    name = "answer_as_json"
  )
}

jsonvalidate_installed <- function() {
  if (!requireNamespace("jsonvalidate", quietly = TRUE)) {
    cli::cli_alert_warning(
      paste0(
        "{.strong `answer_as_json()`}:\n",
        "* When using type \"text-based\" and providing a schema,\n",
        " the 'jsonvalidate' package must be installed to validate the response\n",
        " against the schema\n",
        "* The 'jsonvalidate' package is not installed;\n",
        " the LLM response will not be validated against the schema"
      )
    )

    return(invisible(FALSE))
  }
  return(invisible(TRUE))
}

# Unwrap common JSON schema wrappers (OpenAI/our own)
unwrap_json_schema <- function(x) {
  if (!is.list(x)) {
    return(x)
  }

  # Case 1: OpenAI-style response_format wrapper
  if (
    !is.null(x$response_format) &&
      is.list(x$response_format) &&
      identical(x$response_format$type, "json_schema") &&
      is.list(x$response_format$json_schema)
  ) {
    js <- x$response_format$json_schema
    return(unwrap_json_schema(js))
  }

  # Case 2: top-level wrapper with json_schema field
  if (!is.null(x$json_schema) && is.list(x$json_schema)) {
    return(unwrap_json_schema(x$json_schema))
  }

  # Case 3: wrapper like list(name=..., schema=..., strict=...)
  if (!is.null(x$schema) && is.list(x$schema)) {
    return(x$schema)
  }

  # Already a bare JSON Schema (has typical keywords)
  x
}

#' @title
#' Add an image to a tidyprompt (multimodal)
#'
#' @description
#' `r lifecycle::badge("experimental")`
#'
#' Attach an image to a [tidyprompt()] for use with multimodal LLMs.
#'
#' Supports 'ollama', 'openai' (completions & responses) and 'ellmer'-backed providers.
#' Can convert from and to 'ellmer' content image objects as needed.
#'
#' @param prompt A single string or a [tidyprompt()] object
#' @param image An image reference. One of:
#'   - a local file path (e.g., "path/to/image.png")
#'   - a URL (e.g., "https://.../image.jpg")
#'   - a base64 string (optionally with data URL prefix)
#'   - a raw vector of bytes
#'   - a plot object (e.g., base `recordedplot`, `ggplot`, or grid grob) to be
#'     rasterized automatically
#'   - an 'ellmer' content object created by `ellmer::content_image_url()`,
#'     `ellmer::content_image_file()`, or `ellmer::content_image_plot()`
#'     (this will work with both regular providers and 'ellmer'-backed providers)
#'
#' For OpenAI Responses API, URLs must point directly to an image resource (not an HTML
#' page) and are transmitted as a scalar string `image_url` with optional `detail`.
#' Supplying a webpage URL (e.g. a Wikipedia media viewer link) will result in a
#' provider 400 error expecting an image URL string
#'
#' @param alt Optional alternative text/alt description
#' @param detail Detail hint for some providers (OpenAI): one of "auto", "low",
#' "high"
#' @param mime Optional mime-type if providing raw/base64 without data URL
#' (e.g., "image/png")
#'
#' @return A [tidyprompt()] with an added [prompt_wrap()] which will
#' attach an image to the prompt for use with multimodal LLMs
#'
#' @export
#'
#' @family pre_built_prompt_wraps
#' @family miscellaneous_prompt_wraps
#'
#' @example inst/examples/add_image.R
add_image <- function(
  prompt,
  image,
  alt = NULL,
  detail = c("auto", "low", "high"),
  mime = NULL
) {
  detail <- match.arg(detail)
  prompt <- tidyprompt(prompt)

  part <- .tp_normalize_image_input(
    image,
    mime = mime,
    alt = alt,
    detail = detail
  )

  parameter_fn <- function(llm_provider) {
    existing <- llm_provider$parameters$.add_image_parts %||% list()
    new_parts <- c(existing, list(part))
    llm_provider$parameters$.add_image_parts <- new_parts
    list(.add_image_parts = new_parts)
  }

  prompt_wrap(
    prompt = prompt,
    modify_fn = NULL,
    parameter_fn = parameter_fn,
    name = "add_image"
  )
}

# Internal: normalize input into a provider-agnostic image part
.tp_normalize_image_input <- function(
  image,
  mime = NULL,
  alt = NULL,
  detail = "auto"
) {
  stopifnot(!missing(image))

  # Pass-through for ellmer content objects (content_image_url/file/plot)
  if (isTRUE(requireNamespace("ellmer", quietly = TRUE))) {
    if (.tp_is_ellmer_content(image)) {
      if (!isTRUE(requireNamespace("S7", quietly = TRUE))) {
        stop(
          "The S7 package is required to handle ellmer content image objects, but is not installed."
        )
      }
      props <- S7::props(image)

      # Retrieve detail, prioritizing props over argument
      detail <- props[["detail"]] %||% detail

      # Retrieve url/data and set as 'image' to normalize with regular input
      image <- if (!is.null(props[["url"]])) {
        props[["url"]]
      } else if (!is.null(props[["data"]])) {
        props[["data"]]
      } else {
        stop(
          "'ellmer' content image must have either '@url' or '@data' property"
        )
      }

      # Preserve MIME type from ellmer ContentImageInline (e.g., image/jpeg)
      mime <- props[["type"]] %||% mime
    }
  }

  # Plot objects (base recorded plots, ggplot, grid grobs, etc.)
  if (.tp_is_plot_object(image)) {
    plot_raw <- .tp_plot_to_png_raw(image)
    b64 <- jsonlite::base64_enc(plot_raw)
    return(list(
      kind = "image",
      source = "b64",
      data = as.character(b64),
      mime = "image/png",
      alt = alt,
      detail = detail
    ))
  }

  # raw vector
  if (is.raw(image)) {
    b64 <- jsonlite::base64_enc(image)
    return(list(
      kind = "image",
      source = "b64",
      data = as.character(b64),
      mime = mime %||% "image/png",
      alt = alt,
      detail = detail
    ))
  }

  # base64 data URL string
  if (
    is.character(image) && length(image) == 1 && grepl("^data:image/", image)
  ) {
    # data:image/<type>;base64,<payload>
    parts <- strsplit(image, ",", fixed = TRUE)[[1]]
    header <- parts[1]
    payload <- parts[length(parts)]
    mm <- sub("^data:(.*?);base64$", "\\1", header)
    return(list(
      kind = "image",
      source = "b64",
      data = as.character(payload),
      mime = mm %||% mime %||% "image/png",
      alt = alt,
      detail = detail
    ))
  }

  # bare base64 (heuristic: contains only base64 chars and is long)
  if (
    is.character(image) &&
      length(image) == 1 &&
      grepl("^[A-Za-z0-9+/=\n\r]+$", image) &&
      nchar(image) > 128
  ) {
    return(list(
      kind = "image",
      source = "b64",
      data = gsub("[\n\r]", "", image),
      mime = mime %||% "image/png",
      alt = alt,
      detail = detail
    ))
  }

  # URL
  if (is.character(image) && length(image) == 1 && grepl("^https?://", image)) {
    return(list(
      kind = "image",
      source = "url",
      data = as.character(image),
      mime = NULL,
      alt = alt,
      detail = detail
    ))
  }

  # File path
  if (is.character(image) && length(image) == 1 && file.exists(image)) {
    raw <- readBin(image, what = "raw", n = file.info(image)$size)
    b64 <- jsonlite::base64_enc(raw)
    # Try to infer mime from extension
    mm <- mime %||% .tp_guess_mime_from_path(image) %||% "image/png"
    return(list(
      kind = "image",
      source = "b64",
      data = as.character(b64),
      mime = mm,
      alt = alt,
      detail = detail
    ))
  }

  stop(sprintf(
    "Unsupported `image` input (class: %s); provide a url, file path, base64 string, or raw bytes.",
    paste(class(image), collapse = "/")
  ))
}

.tp_is_plot_object <- function(x) {
  if (is.null(x)) {
    return(FALSE)
  }
  .tp_is_ggplot_object(x) ||
    inherits(x, "recordedplot") ||
    inherits(x, "grob") ||
    inherits(x, "gTree") ||
    inherits(x, "gtable") ||
    inherits(x, "trellis")
}

.tp_is_ggplot_object <- function(x) {
  inherits(x, "ggplot") ||
    inherits(x, "ggplot2::ggplot") ||
    .tp_is_s7_ggplot(x)
}

.tp_is_s7_ggplot <- function(x) {
  if (!isTRUE(requireNamespace("ggplot2", quietly = TRUE))) {
    return(FALSE)
  }

  ns <- asNamespace("ggplot2")
  is_ggplot <- get0("is_ggplot", envir = ns, inherits = FALSE)
  if (is.null(is_ggplot)) {
    is_ggplot <- get0("is.ggplot", envir = ns, inherits = FALSE)
  }
  if (!is.function(is_ggplot)) {
    return(FALSE)
  }

  isTRUE(is_ggplot(x))
}

.tp_plot_to_png_raw <- function(
  plot_obj,
  width = 800,
  height = 600,
  res = 96,
  bg = "white"
) {
  file <- tempfile(fileext = ".png")
  on.exit(unlink(file), add = TRUE)

  device_open <- TRUE
  grDevices::png(
    filename = file,
    width = width,
    height = height,
    res = res,
    units = "px",
    bg = bg
  )
  on.exit(
    {
      if (device_open) {
        try(grDevices::dev.off(), silent = TRUE)
      }
    },
    add = TRUE
  )

  .tp_draw_plot_object(plot_obj)

  grDevices::dev.off()
  device_open <- FALSE

  size <- file.info(file)$size
  if (is.na(size) || size <= 0) {
    stop("Unable to convert plot to image; rendered file is empty.")
  }

  readBin(file, what = "raw", n = size)
}

.tp_draw_plot_object <- function(plot_obj) {
  if (inherits(plot_obj, "recordedplot")) {
    grDevices::replayPlot(plot_obj)
    return(invisible(NULL))
  }

  if (
    inherits(plot_obj, "grob") ||
      inherits(plot_obj, "gTree") ||
      inherits(plot_obj, "gtable")
  ) {
    if (!requireNamespace("grid", quietly = TRUE)) {
      stop(
        "'grid' package is required to handle 'grob'/'gTree'/'gtable' plot objects; please install it"
      )
    }
    grid::grid.newpage()
    grid::grid.draw(plot_obj)
    return(invisible(NULL))
  }

  if (inherits(plot_obj, "ggplot")) {
    print(plot_obj)
    return(invisible(NULL))
  }

  if (inherits(plot_obj, "trellis")) {
    print(plot_obj)
    return(invisible(NULL))
  }

  print(plot_obj)
  invisible(NULL)
}

.tp_guess_mime_from_path <- function(path) {
  ext <- tolower(tools::file_ext(path))
  if (identical(ext, "png")) {
    return("image/png")
  }
  if (identical(ext, "jpg") || identical(ext, "jpeg")) {
    return("image/jpeg")
  }
  if (identical(ext, "gif")) {
    return("image/gif")
  }
  if (identical(ext, "webp")) {
    return("image/webp")
  }
  if (identical(ext, "bmp")) {
    return("image/bmp")
  }
  NULL
}

# Heuristic: detect ellmer content objects (e.g., content_image_*())
.tp_is_ellmer_content <- function(x) {
  cl <- class(x)
  if (length(cl) == 0L) {
    return(FALSE)
  }
  any(grepl("^Content", cl)) ||
    any(grepl("^ellmer", cl))
}

test_that("add_image() accumulates image parts via parameter_fn", {
  tp <- tidyprompt("Describe the picture")
  tp <- add_image(tp, image = "data:image/png;base64,AAAA")
  tp <- add_image(tp, image = "https://example.com/cat.jpg")

  wraps <- get_prompt_wraps(tp, order = "default")
  expect_true(length(wraps) >= 1)

  prov <- llm_provider_fake()
  # Apply parameter_fns as send_prompt would
  for (pw in wraps) {
    if (!is.null(pw$parameter_fn)) {
      prov$set_parameters(pw$parameter_fn(prov))
    }
  }

  parts <- prov$parameters$.add_image_parts
  expect_true(is.list(parts))
  expect_equal(length(parts), 2)
  expect_setequal(
    unique(vapply(parts, function(p) p$source, character(1))),
    c("b64", "url")
  )
})

test_that("add_image() converts recordedplot objects to images", {
  skip_on_cran()
  skip_on_ci()

  tmp_img <- tempfile(fileext = ".png")
  on.exit(unlink(tmp_img), add = TRUE)

  grDevices::png(tmp_img)
  recorded <- tryCatch(
    {
      plot(1:5, 1:5)
      grDevices::recordPlot()
    },
    error = function(e) e
  )
  grDevices::dev.off()

  skip_if(inherits(recorded, "error"), "Current device cannot record plots")
  skip_if_not(inherits(recorded, "recordedplot"))
  # In headless CI environments, png() might produce empty files or fail to record properly
  if (file.exists(tmp_img) && file.size(tmp_img) == 0) {
    skip("png() device produced empty file; skipping plot test")
  }

  tp <- tidyprompt("Describe the chart")
  tp <- add_image(tp, image = recorded)

  wraps <- get_prompt_wraps(tp, order = "default")
  prov <- llm_provider_fake()
  for (pw in wraps) {
    if (!is.null(pw$parameter_fn)) {
      prov$set_parameters(pw$parameter_fn(prov))
    }
  }

  parts <- prov$parameters$.add_image_parts
  expect_true(length(parts) >= 1)
  last_part <- parts[[length(parts)]]
  expect_identical(last_part$mime, "image/png")
  expect_identical(last_part$source, "b64")
  expect_true(nchar(last_part$data) > 100)
})

test_that("add_image() accepts ggplot objects when available", {
  skip_if_not_installed("ggplot2")
  skip_on_ci()
  skip_on_cran()

  plt <- ggplot2::ggplot(mtcars, ggplot2::aes(mpg, disp)) +
    ggplot2::geom_point()

  skip_if_not(
    inherits(plt, "ggplot"),
    "ggplot object creation failed to produce 'ggplot' class"
  )

  # Check if png device is working (required for rasterizing ggplot)
  tmp_check <- tempfile()
  png_works <- tryCatch(
    {
      grDevices::png(tmp_check)
      grDevices::dev.off()
      file.exists(tmp_check)
    },
    error = function(e) FALSE
  )
  unlink(tmp_check)
  skip_if_not(png_works, "png() device is not available or working")

  tp <- tidyprompt("Describe the scatterplot")
  tp <- add_image(tp, image = plt)

  wraps <- get_prompt_wraps(tp, order = "default")
  prov <- llm_provider_fake()
  for (pw in wraps) {
    if (!is.null(pw$parameter_fn)) {
      prov$set_parameters(pw$parameter_fn(prov))
    }
  }

  parts <- prov$parameters$.add_image_parts
  expect_true(length(parts) >= 1)
  last_part <- parts[[length(parts)]]
  expect_identical(last_part$mime, "image/png")
  expect_identical(last_part$source, "b64")
  expect_true(nchar(last_part$data) > 100)
})

test_that("ggplot objects are not misclassified as ellmer content", {
  skip_if_not_installed("ggplot2")

  plt <- ggplot2::ggplot(mtcars, ggplot2::aes(mpg, disp)) +
    ggplot2::geom_point()

  expect_false(tidyprompt:::.tp_is_ellmer_content(plt))
  expect_true(tidyprompt:::.tp_is_plot_object(plt))
})

test_that("llm_provider_ellmer forwards add_image URL parts via ellmer helpers", {
  skip_if_not_installed("ellmer")

  rec_env <- new.env(parent = emptyenv())
  rec_env$urls <- character()

  original_content_image_url <- ellmer::content_image_url

  testthat::local_mocked_bindings(
    content_image_url = function(url, detail = "auto", ...) {
      rec_env$urls <- c(rec_env$urls, url)
      original_content_image_url(url = url, detail = detail, ...)
    },
    .package = "ellmer"
  )

  chat <- fake_ellmer_chat()
  prov <- llm_provider_ellmer(
    chat,
    parameters = list(stream = FALSE),
    verbose = FALSE
  )

  tp <- tidyprompt("Describe the image") |>
    add_image("https://example.com/cat.jpg")

  result <- send_prompt(tp, prov, return_mode = "full", verbose = FALSE)
  expect_true("https://example.com/cat.jpg" %in% rec_env$urls)

  # After multimodal fix, content objects are passed as args to chat(), not pre-seeded as turns.
  # Access the cloned chat's last_method via the returned ellmer_chat.
  used_chat <- result$ellmer_chat
  expect_true(length(used_chat$last_method$args) >= 2)
})

test_that("llm_provider_ellmer falls back to ellmer content when content_image_file fails", {
  skip_if_not_installed("ellmer")

  testthat::local_mocked_bindings(
    content_image_file = function(...) {
      stop("simulated content_image_file failure")
    },
    .package = "ellmer"
  )

  chat <- fake_ellmer_chat()
  prov <- llm_provider_ellmer(
    chat,
    parameters = list(stream = FALSE),
    verbose = FALSE
  )

  img <- tempfile(fileext = ".png")
  grDevices::png(img)
  plot(1:2, 1:2)
  grDevices::dev.off()
  on.exit(unlink(img), add = TRUE)

  result <- tidyprompt("Describe the image") |>
    add_image(img) |>
    send_prompt(prov, return_mode = "full", verbose = FALSE)

  used_chat <- result$ellmer_chat
  image_arg <- used_chat$last_method$args[[2]]

  expect_true(.tp_is_ellmer_content(image_arg))
})

test_that("add_image preserves MIME type from ellmer ContentImageInline", {
  skip_if_not_installed("ellmer")
  skip_if_not_installed("S7")
  skip_if_not(
    exists(
      "ContentImageInline",
      envir = asNamespace("ellmer"),
      inherits = FALSE
    ),
    "ContentImageInline not available in this ellmer version"
  )

  # Create an inline JPEG content object via ellmer
  # ContentImageInline has props: type (MIME), data (base64)
  inline <- ellmer::ContentImageInline(
    type = "image/jpeg",
    data = paste0(rep("A", 200), collapse = "")
  )

  part <- .tp_normalize_image_input(inline)
  expect_equal(part$mime, "image/jpeg")
  expect_equal(part$source, "b64")
})

test_that("build_image_content cleans up temp files for base64 images", {
  skip_if_not_installed("ellmer")

  chat <- fake_ellmer_chat()
  prov <- llm_provider_ellmer(
    chat,
    parameters = list(stream = FALSE),
    verbose = FALSE
  )

  # Create a real PNG image file to use as input
  img <- tempfile(fileext = ".png")
  grDevices::png(img)
  plot(1:2, 1:2)
  grDevices::dev.off()
  on.exit(unlink(img), add = TRUE)

  before <- list.files(tempdir(), full.names = TRUE)

  tp <- tidyprompt("Describe the image") |>
    add_image(img)

  send_prompt(tp, prov, verbose = FALSE)

  after <- list.files(tempdir(), full.names = TRUE)
  leaked <- setdiff(after, before)

  expect_length(leaked, 0L)
})

#' Make LLM answer match a specific regex
#'
#' @param prompt A single string or a [tidyprompt()] object
#' @param regex A character string specifying the regular expression the response must match
#' @param mode A character string specifying the mode of the regex match. Options are
#' "exact_match" (default) and "extract_all_matches".
#'  Under "full_match", the full LLM response must match the regex. If
#'  it does not, the LLM will be sent feedback to retry. The full LLM response
#'  will be returned if the regex is matched.
#'  Under "extract_matches", all matches of the regex in the LLM response will be returned
#'  (if present). If the regex is not matched at all, the LLM will be sent feedback to retry.
#'  If there is at least one match, the matches will be returned as a character vector
#'
#' @return A [tidyprompt()] with an added [prompt_wrap()] which will ensure
#' that the LLM response matches the specified regex
#'
#' @export
#'
#' @example inst/examples/answer_as_regex_match.R
#'
#' @family pre_built_prompt_wraps
#' @family answer_as_prompt_wraps
answer_as_regex_match <- function(
  prompt,
  regex,
  mode = c("full_match", "extract_matches")
) {
  mode <- match.arg(mode)

  # Default instruction for LLM if regex is provided
  instruction <- glue::glue(
    "You must answer with a response that matches this regex format:\n",
    "  {regex}",
    .trim = FALSE
  )
  if (mode == "full_match") {
    instruction <- glue::glue(
      "{instruction}\n",
      "  (use no other characters)",
      .trim = FALSE
    )
  }

  modify_fn <- function(original_prompt_text) {
    glue::glue("{original_prompt_text}\n\n{instruction}", .trim = FALSE)
  }

  extraction_fn <- function(x) {
    if (mode == "full_match") {
      if (!grepl(regex, x)) {
        return(llm_feedback(instruction))
      }
      return(x)
    }

    if (mode == "extract_matches") {
      # Extract all matches using gregexpr
      matches <- unlist(regmatches(x, gregexpr(regex, x)))
      # Remove empty matches
      matches <- matches[matches != ""]
      if (length(matches) == 0) {
        return(llm_feedback(instruction))
      }
      return(matches)
    }
  }

  prompt_wrap(prompt, modify_fn, extraction_fn, name = "answer_as_regex")
}

# `answer_using_tools()` prompt wrap --------------------------------------

#' @title
#' Enable LLM to call R functions (and/or MCP server tools)
#'
#' @description
#' This function adds the ability for the a LLM to call R functions.
#' Users can specify a list of functions that the LLM can call, and the
#' prompt will be modified to include information, as well as an
#' accompanying extraction function to call the functions (handled by
#' [send_prompt()]). Documentation for the functions is extracted from
#' the help file (if available), or from documentation added by
#' [tools_add_docs()]. Users can also provide an 'ellmer' tool definition
#' (see [ellmer::tool()]; ['ellmer' documentation](https://ellmer.tidyverse.org/articles/tool-calling.html)).
#' Model Context Protocol (MCP) tools from MCP servers, as returned from
#' [mcptools::mcp_tools()], may also be used. Regular [ellmer::tool()]
#' definitions are converted as needed and can work with both 'ellmer' and
#' regular LLM providers. Provider-specific 'ellmer' built-in tools
#' (i.e., 'ToolBuiltIn' objects) only work with 'ellmer'-backed providers.
#'
#' @details
#' Note that conversion between 'tidyprompt' and 'ellmer' tool definitions
#' is experimntal and might contain bugs.
#' Provider-specific 'ellmer' built-in tools are only supported when using an
#' [llm_provider_ellmer()] object (or an 'ellmer' chat passed directly to
#' [send_prompt()]).
#'
#' @param prompt A single string or a [tidyprompt()] object
#'
#' @param tools An R function, an 'ellmer' tool definition (from [ellmer::tool()]),
#' or a list of either, which the LLM will be able to call. If an R function is passed
#' which has been documented in a help file (e.g., because it is part of a package),
#' the documentation will be parsed from the help file. If it is a custom function,
#' documentation should be added with [tools_add_docs()] or with [ellmer::tool()].
#' Note that you can also provide Model Context Protocol (MCP) tools from MCP servers
#' as returned from [mcptools::mcp_tools()].
#' Provider-specific 'ellmer' built-in tools are only supported with
#' 'ellmer'-backed providers.
#'
#' @param type (optional) The way that tool calling should be enabled.
#' "auto" will automatically determine the type based on `llm_provider$api_type`
#' or 'llm_provider$tool_type' (if set; 'tool_type' overrides 'api_type' determination)
#' (note that this may not consider model compatibility, and could lead to errors;
#' set 'type' manually if errors occur). "openai" and "ollama" will set the
#' relevant API parameters. "ellmer" will register the tool in the 'ellmer' chat
#' object of the LLM provider; note that this will only work for an [llm_provider_ellmer()]
#' ("auto" will always set the type to "ellmer" if you are using an 'ellmer' LLM provider).
#' "text-based" will provide function definitions in the prompt, extract function
#' calls from the LLM response, and call the
#' functions, providing the results back via [llm_feedback()]. "text-based"
#' always works, but may be inefficient for APIs that support tool calling
#' natively. Note that when using "openai", "ollama", or "ellmer", tool calls are not counted
#' as interactions by [send_prompt()] and may continue indefinitely unless restricted
#' by other means
#'
#' @return A [tidyprompt()] with an added [prompt_wrap()] which
#' will allow the LLM to call the given R functions when evaluating
#' the prompt with [send_prompt()]
#'
#' @export
#'
#' @example inst/examples/answer_using_tools.R
#'
#' @seealso [answer_using_r()] [tools_get_docs()]
#'
#' @family pre_built_prompt_wraps
#' @family answer_using_prompt_wraps
#' @family tools
answer_using_tools <- function(
  prompt,
  tools = list(),
  type = c("auto", "openai", "ollama", "ellmer", "text-based")
) {
  prompt <- tidyprompt(prompt)
  type <- match.arg(type)

  if (type == "auto" && getOption("tidyprompt.warn.auto.tools", TRUE)) {
    cli::cli_alert_warning(
      paste0(
        "{.strong `answer_using_tools()`}:\n",
        "* Automatically determining type based on 'llm_provider$api_type';\n",
        "(or 'llm_provider$tool_type' if set); this does not consider model compatability\n",
        "* Manually set argument 'type' if errors occur ",
        "(\"text-based\" always works)\n",
        "* Use `options(tidyprompt.warn.auto.tools = FALSE)` to suppress this warning"
      )
    )
  }

  # -- Accept function, ellmer ToolDef, ToolBuiltIn, or list of mixed ---------
  is_toolish <- function(x) is.function(x) || is_ellmer_any_tool(x)

  sanitize_name <- function(x) {
    x <- as.character(x)
    x <- gsub("[^A-Za-z0-9_]", "_", x)
    x[!nzchar(x)] <- "tool"
    x
  }

  # Normalize to a named list while trying to keep stable names
  user_named <- logical(0) # track which positions had explicit names
  if (is_toolish(tools)) {
    nm <- sanitize_name(deparse(substitute(tools)))
    tools <- rlang::set_names(list(tools), nm)
    user_named <- FALSE
  } else {
    stopifnot(
      is.list(tools),
      length(tools) > 0,
      all(vapply(tools, is_toolish, logical(1)))
    )
    nms <- names(tools) %||% rep("", length(tools))
    user_named <- nzchar(nms)

    if (!all(user_named)) {
      # Try to derive names for unnamed positions from the call
      call_elems <- tryCatch(
        as.list(substitute(tools))[-1],
        error = function(e) NULL
      )
      for (i in which(!user_named)) {
        if (!is.null(call_elems) && length(call_elems) == length(tools)) {
          nms[i] <- sanitize_name(deparse(call_elems[[i]]))
        } else {
          nms[i] <- paste0("tool", i)
        }
      }
    }

    names(tools) <- sanitize_name(nms)
  }

  # For ellmer ToolDefs that were NOT explicitly named by the user,

  # prefer the tool's own declared @name so text prompts, API payloads,
  # and dispatch all agree on the same identifier.
  for (i in seq_along(tools)) {
    if (!user_named[i] && is_ellmer_tool(tools[[i]])) {
      declared <- tryCatch(tools[[i]]@name, error = function(e) NULL)
      if (is.character(declared) && nzchar(declared)) {
        names(tools)[i] <- sanitize_name(declared)
      }
    }
  }

  # Detect name collisions that would silently discard tools
  tool_nms <- names(tools)
  dup_nms <- unique(tool_nms[duplicated(tool_nms)])
  if (length(dup_nms) > 0L) {
    stop(
      "Tool name collision after sanitization: ",
      paste0("'", dup_nms, "'", collapse = ", "),
      ". Provide distinct names to avoid tools being silently dropped.",
      call. = FALSE
    )
  }

  has_builtin_tools <- any(vapply(tools, is_ellmer_builtin_tool, logical(1)))
  builtin_tool_error <- paste0(
    "Provider-specific ellmer built-in tools can only be used with an ",
    "ellmer-backed provider (`type = \"ellmer\"` or `type = \"auto\"` ",
    "resolving to an ellmer provider)."
  )

  if (has_builtin_tools && !(type %in% c("auto", "ellmer"))) {
    stop(builtin_tool_error, call. = FALSE)
  }

  # -- Build dual representations ---------------------------------------------
  # For each supplied tool, we want:
  # - a tidyprompt-style function (with docs) for text-based/OpenAI/Ollama paths
  # - an ellmer ToolDef for native ellmer path
  tp_tools <- list()
  ellmer_tools <- list()
  for (nm in names(tools)) {
    dual <- normalize_tool_dual(tools[[nm]])
    if (!is.null(dual$tidyprompt_tool)) {
      tp_tools[[nm]] <- dual$tidyprompt_tool
    }
    if (!is.null(dual$ellmer_tool)) ellmer_tools[[nm]] <- dual$ellmer_tool
  }

  determine_type <- function(llm_provider = NULL) {
    if (type != "auto") {
      return(type)
    }
    valid_types <- c(
      "text-based",
      "openai",
      "ollama",
      "ellmer"
    )
    provider_json_type <- llm_provider[["tool_type"]]
    if (
      !is.null(provider_json_type) &&
        !identical(provider_json_type, "auto") &&
        provider_json_type %in% valid_types
    ) {
      return(provider_json_type)
    }
    if (isTRUE(llm_provider$api_type == "openai")) {
      return("openai")
    }
    if (isTRUE(llm_provider$api_type == "ollama")) {
      return("ollama")
    }
    if (isTRUE(llm_provider$api_type == "ellmer")) {
      return("ellmer")
    }
    "text-based"
  }

  parameter_fn <- function(llm_provider) {
    t <- determine_type(llm_provider)
    parameters <- list()

    if (has_builtin_tools && t != "ellmer") {
      stop(builtin_tool_error, call. = FALSE)
    }

    if (t == "ellmer" && !isTRUE(llm_provider$api_type == "ellmer")) {
      stop(
        paste0(
          "`answer_using_tools(type = \"ellmer\")` requires an ellmer-backed ",
          "provider created with `llm_provider_ellmer()` or a direct ellmer chat."
        ),
        call. = FALSE
      )
    }

    if (t %in% c("openai", "ollama")) {
      tools_openai <- list()
      for (tool_name in names(tp_tools)) {
        tool <- tp_tools[[tool_name]]
        docs <- tools_get_docs(tool, tool_name)

        tool_openai <- list()
        tool_openai$type <- "function"
        tool_openai[["function"]]$name <- tool_name
        if (!is.null(docs$description)) {
          tool_openai[["function"]]$description <- docs$description
        }
        tool_openai[["function"]]$parameters <- tools_docs_to_r_json_schema(
          docs
        )
        tool_openai[["function"]]$strict <- TRUE

        tools_openai[[length(tools_openai) + 1]] <- tool_openai
      }
      parameters$tools <- tools_openai
    }

    if (t == "ollama" && isTRUE(llm_provider$parameters$stream)) {
      cli::cli_alert_warning(
        paste0(
          "{.strong `answer_using_tools()`}:\n",
          "* 'ollama' does not support streaming when using tools;\n",
          "'stream' parameter was set to FALSE"
        )
      )
      parameters$stream <- FALSE
    }

    if (t == "ellmer") {
      # Ensure we have ToolDefs for every tool; convert if needed
      missing_defs <- setdiff(names(tp_tools), names(ellmer_tools))
      if (length(missing_defs)) {
        for (nm in missing_defs) {
          # convert tidyprompt tool -> ellmer ToolDef
          ell_tool <- tryCatch(
            tidyprompt_tool_to_ellmer(tp_tools[[nm]], name = nm),
            error = function(e) NULL
          )
          if (!is.null(ell_tool)) ellmer_tools[[nm]] <- ell_tool
        }
      }

      if (length(ellmer_tools)) {
        for (nm in names(ellmer_tools)) {
          if (is_ellmer_tool(ellmer_tools[[nm]])) {
            ellmer_tools[[nm]] <- rename_ellmer_tool(ellmer_tools[[nm]], nm)
          }
        }

        effective_names <- vapply(
          names(ellmer_tools),
          function(nm) {
            tool <- ellmer_tools[[nm]]
            if (is_ellmer_builtin_tool(tool)) {
              return(ellmer_tool_name(tool, fallback = nm))
            }
            nm
          },
          character(1)
        )
        dup_native_names <- unique(effective_names[duplicated(effective_names)])
        if (length(dup_native_names) > 0L) {
          stop(
            "Tool name collision on native ellmer path: ",
            paste0("'", dup_native_names, "'", collapse = ", "),
            ". Built-in ellmer tools keep their provider-defined names.",
            call. = FALSE
          )
        }
      }

      # Pass ToolDefs for provider to register natively.
      # NOTE for provider wrapper:
      #   if (!is.null(params$.ellmer_tools)) {
      #     for (td in params$.ellmer_tools) chat$register_tool(td)
      #   }
      if (length(ellmer_tools)) {
        parameters$.ellmer_tools <- unname(ellmer_tools)
      }
    }

    parameters
  }

  handler_fn <- function(response, llm_provider) {
    t <- determine_type(llm_provider)
    # Native tool handling is done by the provider/model itself
    if (t == "ellmer") {
      return(response)
    }
    if (!(t %in% c("openai", "ollama"))) {
      return(response)
    }

    repeat {
      # 1) Normalize body
      body <- .parse_http_body(response$http$response$body)
      if (is.null(body)) {
        break
      }

      # 2) Extract tool calls for this provider type
      tool_calls <- .get_tool_calls(body, t)
      if (length(tool_calls) == 0) {
        break
      }

      # 3) Rebuild messages from completed chat history, then add assistant tool_calls
      chat_history <- response$completed
      messages <- .messages_from_history(chat_history)

      if (t == "openai") {
        messages[[length(messages) + 1]] <- list(
          role = "assistant",
          tool_calls = tool_calls
        )
      } else if (t == "ollama") {
        # Keep original Ollama structure if available
        messages[[length(messages) + 1]] <- list(
          role = "assistant",
          tool_calls = body$message$tool_calls
        )
      }

      # 4) Iterate over tool calls: parse args, call tool, record result
      for (tool_call in tool_calls) {
        fn_info <- tool_call[["function"]] %||% list()
        tool_name <- fn_info$name %||% tool_call$name %||% NA_character_
        if (is.na(tool_name) || !nzchar(tool_name)) {
          next
        }
        tool <- tp_tools[[tool_name]]

        # Parse arguments robustly
        arguments <- .parse_arguments(tool_call, t)

        # Record the call in chat history
        pretty_args <- tryCatch(
          jsonlite::toJSON(arguments, auto_unbox = FALSE, pretty = TRUE),
          error = function(e) "null"
        )
        chat_history <- .append_history(
          chat_history,
          role = "assistant",
          content = paste0(
            "~>> Calling function '",
            tool_name,
            "' with arguments:\n",
            pretty_args
          ),
          tool_call_id = tool_call$id,
          tool_call = TRUE,
          tool_result = FALSE
        )
        .maybe_message(
          dplyr::slice_tail(chat_history, n = 1)$content,
          llm_provider
        )

        # Call the tool safely
        if (is.null(tool)) {
          result <- glue::glue("Error: tool '{tool_name}' not registered")
        } else {
          result <- tryCatch(
            do.call(tool, arguments),
            error = function(e) glue::glue("Error: {e$message}")
          )
        }

        # Normalize result to character
        if (length(result) > 0) {
          result <- paste(result, collapse = ", ")
        }
        if (length(result) == 0) {
          result <- ""
        }
        result <- as.character(result)

        # Add tool result to next request messages
        message_addition <- list(role = "tool", content = result)
        if (t == "openai") {
          message_addition$tool_call_id <- tool_call$id
        }
        messages[[length(messages) + 1]] <- message_addition

        # Record result in chat history
        chat_history <- .append_history(
          chat_history,
          role = "tool",
          content = paste0("~>> Result:\n", result),
          tool_call_id = tool_call$id,
          tool_call = FALSE,
          tool_result = TRUE
        )
        .maybe_message(paste0("~>> Result:\n", result), llm_provider)
      }

      # 5) Send follow-up request including tool outputs, keep loop if model asks again
      new_request <- response$http$request
      new_request$body$data$messages <- messages

      response <- request_llm_provider(
        chat_history,
        new_request,
        stream = llm_provider$parameters$stream,
        verbose = llm_provider$verbose,
        api_type = llm_provider$api_type
      )
    }

    response
  }

  modify_fn <- function(original_prompt_text, llm_provider) {
    t <- determine_type(llm_provider)
    if (t %in% c("openai", "ollama", "ellmer")) {
      return(original_prompt_text)
    }

    new_prompt <- glue::glue(
      "{original_prompt_text}\n\n",
      "If you need more information, you can call functions to help you."
    )

    new_prompt <- glue::glue(
      "{new_prompt}\n\n",
      "To call a function, output a JSON object with the following format:\n",
      "  {{\n",
      "    \"function\": \"<function name>\",\n",
      "    \"arguments\": {{\n",
      "      \"<argument_name>\": <argument_value>,\n",
      "      # ...\n",
      "    }}\n",
      "  }}\n",
      "  (Note: you may not provide function calls as function arguments.)"
    )

    new_prompt <- glue::glue(
      "{new_prompt}\n\n",
      "The following functions are available:"
    )

    for (tool_name in names(tp_tools)) {
      tool <- tp_tools[[tool_name]]
      docs <- tools_get_docs(tool, tool_name)

      tool_llm_text <- tools_docs_to_text(docs, with_arguments = TRUE)
      new_prompt <- glue::glue("{new_prompt}\n\n{tool_llm_text}", .trim = FALSE)
    }

    new_prompt <- glue::glue(
      "{new_prompt}\n\n",
      "After you call a function, wait until you receive more information.\n",
      "Use the information to decide your next steps or provide a final response.",
      .trim = FALSE
    )

    new_prompt
  }

  extraction_fn <- function(llm_response, llm_provider) {
    t <- determine_type(llm_provider)
    if (t %in% c("openai", "ollama", "ellmer")) {
      return(llm_response)
    }

    jsons <- extraction_fn_json(llm_response)

    fn_results <- lapply(jsons, function(json) {
      if (is.null(json[["function"]])) {
        return(NULL)
      }

      if (json[["function"]] %in% names(tp_tools)) {
        tool_function <- tp_tools[[json[["function"]]]]
        arguments <- json[["arguments"]]

        result <- tryCatch(
          {
            do.call(tool_function, arguments)
          },
          error = function(e) {
            glue::glue("Error in {e$message}")
          }
        )

        if (length(result) > 0) {
          result <- paste(result, collapse = ", ")
        }
        if (length(result) == 0) {
          result <- ""
        }

        string_of_named_arguments <-
          paste(names(arguments), arguments, sep = " = ") |>
          paste(collapse = ", ")

        glue::glue(
          "function called: {json[[\"function\"]]}\n",
          "arguments used: {string_of_named_arguments}\n",
          "result: {result}"
        )
      } else {
        glue::glue("Error: Function '{json[[\"function\"]]}' not found.")
      }
    })
    fn_results <- fn_results[!vapply(fn_results, is.null, logical(1))]

    if (length(fn_results) == 0) {
      return(llm_response)
    }

    llm_feedback(paste(fn_results, collapse = "\n\n"), tool_result = TRUE)
  }

  # Attach the (tidyprompt) tool functions for text-based/OpenAI/Ollama paths
  environment_with_tools <- new.env()
  environment_with_tools$tools <- tp_tools
  attr(extraction_fn, "environment") <- environment_with_tools
  attr(handler_fn, "environment") <- environment_with_tools

  prompt_wrap(
    prompt,
    modify_fn,
    extraction_fn,
    NULL,
    handler_fn,
    parameter_fn,
    type = "tool",
    name = "answer_using_tools"
  )
}


# Helpers for handler_fn in `answer_using_tools()` ------------------------

# These functions are used by the handler_fn in `answer_using_tools()`,
#  to parse tool calls from LLM responses of various API providers,
#  normalize them, call the relevant R functions, and provide the results
#  back to the LLM
# Used for text-based, openai, and ollama types; when using an 'ellmer'
#   LLM provider, tool calling is handled natively by 'ellmer'

.parse_http_body <- function(resp_body) {
  # If it's already a list with tool_calls (some wrappers pre-normalize), return it
  if (is.list(resp_body) && ("tool_calls" %in% names(resp_body))) {
    return(resp_body)
  }
  # Try to decode raw -> JSON -> list
  if (is.raw(resp_body)) {
    return(tryCatch(
      jsonlite::fromJSON(rawToChar(resp_body), simplifyVector = FALSE),
      error = function(e) NULL
    ))
  }
  # If it's character JSON, try to parse
  if (is.character(resp_body) && length(resp_body) == 1) {
    return(tryCatch(
      jsonlite::fromJSON(resp_body, simplifyVector = FALSE),
      error = function(e) NULL
    ))
  }
  # If it's already a list but without tool_calls, just return it
  if (is.list(resp_body)) {
    return(resp_body)
  }
  NULL
}

.normalize_openai_tool_calls <- function(body) {
  calls <- list()

  format_args <- function(args) {
    if (is.null(args)) {
      return("{}")
    }
    if (is.character(args) && length(args) == 1) {
      return(args)
    }
    safe <- tryCatch(
      jsonlite::toJSON(args, auto_unbox = TRUE),
      error = function(e) NULL
    )
    if (is.null(safe)) {
      return("{}")
    }
    as.character(safe)
  }

  row_to_list <- function(df_row) {
    if (!is.data.frame(df_row)) {
      return(df_row)
    }
    lapply(df_row, function(cell) {
      if (is.list(cell) && length(cell) == 1) {
        return(cell[[1]])
      }
      cell
    })
  }

  # --- Chat Completions style -------------------------------------------------
  if (!is.null(body$choices)) {
    choices <- body$choices
    if (is.data.frame(choices)) {
      choices <- lapply(seq_len(nrow(choices)), function(i) {
        row_to_list(choices[i, , drop = FALSE])
      })
    }
    if (is.list(choices)) {
      for (choice in choices) {
        msg <- choice$message %||% choice[["message"]]
        if (is.null(msg)) {
          next
        }
        if (is.data.frame(msg)) {
          msg <- row_to_list(msg[1, , drop = FALSE])
        }
        tcs <- msg$tool_calls %||% msg[["tool_calls"]]
        if (is.null(tcs) || length(tcs) == 0) {
          next
        }

        if (is.data.frame(tcs)) {
          tcs <- lapply(seq_len(nrow(tcs)), function(i) {
            row_to_list(tcs[i, , drop = FALSE])
          })
        }

        for (tc in tcs) {
          fn <- tc[["function"]]
          if (is.null(fn)) {
            next
          }
          calls[[length(calls) + 1]] <- list(
            id = tc$id %||% NA_character_,
            type = tc$type %||% "function",
            "function" = list(
              name = fn$name %||% tc$name %||% "",
              arguments = format_args(
                fn$arguments %||% tc$arguments %||% fn$args
              )
            )
          )
        }
      }
      if (length(calls) > 0) return(calls)
    }
  }

  # --- Responses API style ----------------------------------------------------
  if (!is.null(body$output)) {
    output <- body$output

    if (is.data.frame(output)) {
      output <- lapply(seq_len(nrow(output)), function(i) {
        row_to_list(output[i, , drop = FALSE])
      })
    }

    if (is.list(output)) {
      for (item in output) {
        if (!is.list(item)) {
          next
        }
        type <- item$type %||% NULL

        if (identical(type, "function_call")) {
          fn_obj <- item[["function"]]
          calls[[length(calls) + 1]] <- list(
            id = item$call_id %||% item$id %||% NA_character_,
            type = "function",
            "function" = list(
              name = item$name %||% fn_obj$name %||% "",
              arguments = format_args(
                item$arguments %||%
                  fn_obj$arguments %||%
                  item$args
              )
            )
          )
        } else if (!is.null(item$content) && length(item$content) > 0) {
          for (content_part in item$content) {
            if (is.data.frame(content_part)) {
              content_part <- row_to_list(content_part[1, , drop = FALSE])
            }
            tc <- content_part$tool_calls %||% content_part[["tool_calls"]]
            if (is.null(tc)) {
              next
            }
            if (is.data.frame(tc)) {
              tc <- lapply(seq_len(nrow(tc)), function(i) {
                row_to_list(tc[i, , drop = FALSE])
              })
            }
            for (tool_call in tc) {
              fn <- tool_call[["function"]]
              if (is.null(fn)) {
                next
              }
              calls[[length(calls) + 1]] <- list(
                id = tool_call$id %||% NA_character_,
                type = tool_call$type %||% "function",
                "function" = list(
                  name = fn$name %||% tool_call$name %||% "",
                  arguments = format_args(
                    fn$arguments %||%
                      tool_call$arguments %||%
                      fn$args
                  )
                )
              )
            }
          }
        }
      }
    }
  }

  calls
}

.normalize_ollama_tool_calls <- function(body) {
  calls <- list()
  if (
    !is.null(body$message) &&
      length(body$message$tool_calls) > 0
  ) {
    tc_list <- body$message$tool_calls
    if (is.data.frame(tc_list)) {
      tc_list <- lapply(seq_len(nrow(tc_list)), function(i) {
        row <- tc_list[i, , drop = FALSE]
        lapply(row, function(x) if (is.list(x) && length(x) == 1) x[[1]] else x)
      })
    } else if (!is.list(tc_list)) {
      tc_list <- list(tc_list)
    }

    for (tc in tc_list) {
      fn <- NULL
      if (is.list(tc)) {
        fn <- tc[["function"]] %||% tc[["function"]] %||% NULL
      }
      if (is.null(fn)) {
        next
      }
      calls[[length(calls) + 1]] <- list(
        "function" = as.list(fn)
      )
    }
  }
  calls
}

.get_tool_calls <- function(body, t) {
  # Case where upstream wrapped a tool_calls list directly in body
  if (!is.null(body$tool_calls) && length(body$tool_calls) > 0) {
    return(body$tool_calls)
  }
  if (t == "openai") {
    return(.normalize_openai_tool_calls(body))
  }
  if (t == "ollama") {
    return(.normalize_ollama_tool_calls(body))
  }
  list()
}

.parse_arguments <- function(tool_call, t) {
  if (t == "ollama") {
    args_raw <- NULL
    fn <- tool_call[["function"]]
    if (is.list(fn)) {
      args_raw <- fn$arguments %||% fn$args %||% NULL
    }
    if (is.null(args_raw)) {
      return(list())
    }
    if (is.character(args_raw) && length(args_raw) == 1) {
      parsed <- tryCatch(
        jsonlite::fromJSON(args_raw, simplifyVector = FALSE),
        error = function(e) NULL
      )
      if (!is.null(parsed)) return(parsed)
    }
    if (is.list(args_raw)) {
      return(args_raw)
    }
    return(as.list(args_raw))
  }
  # OpenAI: try "arguments" then "args", both JSON
  args <- NULL
  try(
    {
      args <- jsonlite::fromJSON(tool_call[["function"]]$arguments)
    },
    silent = TRUE
  )
  if (is.null(args)) {
    try(
      {
        args <- jsonlite::fromJSON(tool_call[["function"]]$args)
      },
      silent = TRUE
    )
  }
  args
}

.append_history <- function(
  history,
  role,
  content,
  tool_call_id = NA,
  tool_call = FALSE,
  tool_result = FALSE
) {
  dplyr::bind_rows(
    history,
    data.frame(
      role = role,
      content = content,
      tool_call = tool_call,
      tool_call_id = ifelse(is.null(tool_call_id), NA, tool_call_id),
      tool_result = tool_result
    )
  )
}

.maybe_message <- function(text, llm_provider) {
  if (isTRUE(llm_provider$parameters$stream) || isTRUE(llm_provider$verbose)) {
    message(text)
    message("")
  }
}

.messages_from_history <- function(history_df) {
  history_df <- chat_history_to_send(history_df)

  lapply(seq_len(nrow(history_df)), function(i) {
    list(role = history_df$role[i], content = history_df$content[i])
  })
}


# Add documentation to functions ------------------------------------------

# These functions are used to extract & add documentation to/from R functions,
#   so that LLMs can be provided with information about the functions
# They are sort of superseded by the 'ellmer' integration and `ellmer::tool()`

#' Add tidyprompt function documentation to a function
#'
#' @description This function adds documentation to a custom function. This documentation
#' is used to extract information about the function's name, description, arguments,
#' and return value. This information is used to provide an LLM with information
#' about the functions, so that the LLM can call R functions. The intended
#' use of this function is to add documentation to custom functions that do not
#' have help files; [tools_get_docs()] may generate documentation from a
#' help file when the function is part of base R or a package.
#'
#' If a function already has documentation, the documentation added by this
#' function may overwrite it. If you wish to modify existing documentation,
#' you may make a call to [tools_get_docs()] to extract the existing documentation,
#' modify it, and then call [tools_add_docs()] to add the modified documentation.
#'
#' @param func A function object
#' @param docs A list with the following elements:
#' \itemize{
#' \item 'name': (optional) The name of the function. If not provided, the function
#' name will be extracted from the function object. Use this parameter to override
#' the function name if necessary
#' \item 'description': A description of the function and its purpose
#' \item 'arguments': A named list of arguments with descriptions. Each argument is a list
#' which may contain: \itemize{
#' \item 'description': A description of the argument and its purpose. Not
#' required or used for native function calling (e.g., with OpenAI), but recommended
#' for text-based function calling
#' \item 'type': The type of the argument. This should be one of:
#' 'integer', 'numeric', 'logical', 'string', 'ignore', 'match.arg',
#' 'vector integer', 'vector numeric', 'vector logical', 'vector string'.
#'  For arguments which are named lists, 'type' should be a named list
#'  which contains the types of the elements. For type 'match.arg', the
#'  possible values should be passed as a vector under 'default_value'.
#'  Use 'ignore' to keep an argument out of generated tool prompts/schemas and
#'  rely on the R function itself to supply it (for example via a default value).
#'  'type' is required for native function calling (with, e.g., OpenAI) but may
#'  also be useful to provide for text-based function calling, in which it will
#'  be added to the prompt introducing the function
#' \item 'default_value': The default value of the argument. This is only required
#' when 'type' is set to 'match.arg'. It should then be a vector of possible values
#' for the argument. In other cases, it is not required; for native function calling,
#' it is not used in other cases; for text-based function calling, it may be useful
#' to provide the default value, which will be added to the prompt introducing the
#' function
#' }
#' \item 'return': A list with the following elements:
#' \itemize{
#' \item 'description': A description of the return value or the side effects of the function
#' }
#' }
#' @return The function object with the documentation added as an attribute
#' ('tidyprompt_tool_docs')
#'
#' @export
#'
#' @example inst/examples/answer_using_tools.R
#' @family tools
tools_add_docs <- function(
  func,
  docs
) {
  stopifnot(
    is.function(func),
    is.list(docs),
    length(docs) > 0,
    !is.null(names(docs)),
    (is.null(docs$name) | is.character(docs$name) & length(docs$name) == 1),
    is.character(docs$description) & length(docs$description) == 1,
    is.list(docs$arguments),
    length(docs$arguments) == 0 | !is.null(names(docs$arguments)),
    is.null(docs$return$description) |
      is.character(docs$return$description) &
        length(docs$return$description) == 1
  )

  if (is.null(docs$name)) {
    docs$name <- deparse(substitute(func))
  }

  attr(func, "tidyprompt_tool_docs") <- docs

  return(func)
}

#' Extract documentation from a function
#'
#' This function extracts documentation from a help file (if available,
#' i.e., when the function is part of a package) or from documentation added
#' by [tools_add_docs()]. The extracted documentation includes
#' the function's name, description, arguments, and return value.
#' This information is used to provide an LLM with information about the functions,
#' so that the LLM can call R functions.
#'
#' @details This function will prioritize documentation added by
#' [tools_add_docs()] over documentation from a help file.
#' Thus, it is possible to override the help file documentation by adding
#' custom documentation
#'
#' @param func A function object. The function should belong to a package
#' and have documentation available in a help file, or it should
#' have documentation added by [tools_add_docs()]
#' @param name The name of the function if already known (optional).
#' If not provided it will be extracted from the documentation or the
#' function object's name
#'
#' @return A list with documentation for the function. See [tools_add_docs()]
#' for more information on the contents
#'
#' @export
#'
#' @example inst/examples/answer_using_tools.R
#'
#' @family tools
tools_get_docs <- function(func, name = NULL) {
  stopifnot(
    is.function(func),
    is.null(name) || (is.character(name) & length(name) == 1)
  )

  docs <- list()
  caller_name <- name
  if (is.null(name)) {
    name <- deparse(substitute(func))
  }

  if (!is.null(attr(func, "tidyprompt_tool_docs"))) {
    docs <- attr(func, "tidyprompt_tool_docs")

    stopifnot(
      is.list(docs),
      !is.null(names(docs)),
      all(c("name", "description", "arguments") %in% names(docs)),
      is.character(docs$name) & length(docs$name) == 1,
      is.character(docs$description) & length(docs$description) == 1,
      is.list(docs$arguments),
      !is.null(names(docs$arguments)),
      is.null(docs$return$description) ||
        (is.character(docs$return$description) &
          length(docs$return$description) == 1)
    )

    # When the caller provides an explicit name, use it so the advertised
    # name always matches the dispatch key (e.g., in answer_using_tools).
    if (!is.null(caller_name)) {
      docs$name <- caller_name
    }
  } else {
    # Resolve the real function name for help-file lookup.  The caller
    # may pass a sanitized alias (e.g., "list_files" for list.files);
    # looking up help with the alias would fail.  We search the function's
    # home namespace for the original symbol.
    help_name <- resolve_function_name(func, name)
    docs <- tools_generate_docs(name, func, help_name = help_name)
  }

  # Check that all formal arguments have documentation
  if (!all(names(formals(func)) %in% names(docs$arguments))) {
    # If not, modify args with what is known from formals
    args <- docs$arguments
    true_args <- gd_get_args_defaults_types(func)

    # Add true_args to args only where not present
    for (arg_name in names(true_args)) {
      if (!arg_name %in% names(args)) {
        args[[arg_name]] <- true_args[[arg_name]]
      }
    }

    # Add type, default value where not present
    for (arg_name in names(args)) {
      if (!"type" %in% names(args[[arg_name]])) {
        args[[arg_name]]$type <- true_args[[arg_name]]$type
      }
      if (!"default_value" %in% names(args[[arg_name]])) {
        args[[arg_name]]$default_value <- true_args[[arg_name]]$default_value
      }
    }

    docs$arguments <- args
  }

  # Remove args not present in formals
  for (arg_name in names(docs$arguments)) {
    if (!arg_name %in% names(formals(func))) {
      docs$arguments[[arg_name]] <- NULL
      warning(
        paste0(
          "Argument '",
          arg_name,
          "' not found in function formals. Removing from documentation"
        )
      )
    }
  }

  docs
}

#' Resolve the real symbol name of a function in its home namespace
#'
#' When a sanitized alias (e.g., "list_files") is passed as the display name,
#' help-file lookup needs the original symbol ("list.files"). This helper
#' searches the function's package namespace for a matching binding.
#'
#' @param func A function object
#' @param fallback The name to return when resolution fails
#'
#' @return A single character string: the real symbol name, or `fallback`.
#' @noRd
#' @keywords internal
resolve_function_name <- function(func, fallback) {
  pkg_env <- environment(func)
  if (is.null(pkg_env)) {
    return(fallback)
  }
  pkg_name <- environmentName(pkg_env)
  if (pkg_name == "" || pkg_name == "R_GlobalEnv") {
    return(fallback)
  }
  # Search the namespace for the exact function object
  ns <- tryCatch(asNamespace(pkg_name), error = function(e) NULL)
  if (is.null(ns)) {
    return(fallback)
  }
  exports <- tryCatch(getNamespaceExports(ns), error = function(e) NULL)
  if (is.null(exports)) {
    return(fallback)
  }
  for (sym in exports) {
    obj <- tryCatch(
      get(sym, envir = ns, inherits = FALSE),
      error = function(e) NULL
    )
    if (identical(obj, func)) {
      return(sym)
    }
  }
  fallback
}

#' Generate function documentation from formals and help file
#'
#' This function generates documentation for a function based on its formals
#' and help file (if available). The documentation includes the function's
#' name, description, arguments, and return value. This function is called internally
#' when there is no documentation.#'
#'
#' @param name (string) A name of a function (e.g., 'list.files')
#' @param func_object (function) A function object. If provided, the function
#' object will be used instead of the function with the name 'name', when
#' that cannot be obtained with 'get()'
#'
#' @return A list with documentation for the function
#'
#' @noRd
#' @keywords internal
tools_generate_docs <- function(name, func_object = NULL, help_name = NULL) {
  # Get the package name and function
  lookup <- help_name %||% name
  func <- tryCatch(
    get(lookup, mode = "function"),
    error = function(e) {
      NULL
    }
  )
  if (is.null(func) && !is.null(func_object)) {
    func <- func_object
  }
  package_env <- environment(func)
  package_name <- environmentName(package_env)

  # Get arguments, defaults, and likely types based on formals
  args <- gd_get_args_defaults_types(func)

  if (package_name == "R_GlobalEnv" || package_name == "") {
    help_file <- NULL
  } else {
    # Get the function's help file using the real symbol name
    help_file <- utils::help(lookup, package = as.character(package_name))
  }
  if (length(help_file) == 0) {
    # No help file found
    # Return without function description, arg descriptions, or return description
    return(
      list(
        name = name,
        arguments = args
      )
    )
  }

  # Extract the text from the help file
  help_text <- tools::Rd2txt(gd_get_help_file(as.character(help_file))) |>
    utils::capture.output()

  # Parse the help text
  parsed_help <- gd_parse_help_text(help_text)

  # Add descriptions to arguments
  for (arg_name in names(args)) {
    description <- parsed_help[["Arguments"]][[arg_name]]
    if (!is.null(description)) {
      args[[arg_name]]$description <- description
    }
  }

  # Return list with function name, description, arguments, and return value
  return(
    list(
      name = name,
      description = paste0(parsed_help$Title, ": ", parsed_help$Description),
      arguments = args,
      'return' = list(description = parsed_help$Value)
    )
  )
}

#' Get argument names, default values, and types from a function's formals
#'
#' This function extracts the argument names, default values, and types from
#' a function's formals. The types are inferred from the default values.
#' Is called internally to generate function documentation when there is none.
#'
#' @param func A function object
#'
#' @return A named list of arguments, where each argument is a list containing:
#' - 'default_value': The default value of the argument. This is not
#'  included if there is no default value. Note that when the default value
#'  is included and is NULL, the default value is NULL (and not missing)
#'  - 'type': see 'gd_infer_type_from_default()'. If there is no default value,
#'  the type is set to 'unknown'
#'
#' @noRd
#' @keywords internal
gd_get_args_defaults_types <- function(func) {
  args_formals <- formals(func)
  arg_names <- names(args_formals)

  arguments <- list()
  for (arg_name in arg_names) {
    default_value <- args_formals[[arg_name]]
    if (missing(default_value)) {
      # If missing default value, "default_value" will not be added to list
      # Note: this is different from adding 'NULL' as "default_value"
      #   (In that case, the default value is NULL and not missing)
      arguments[[arg_name]] <- list(
        type = "unknown"
      )
      next
    }

    arguments[[arg_name]] <- list(
      default_value = default_value,
      type = gd_infer_type_from_default(default_value)
    )
  }

  return(arguments)
}

#' Infer the type of an argument from its default value
#'
#' This function infers the type of an argument from its default value.
#'
#' @param default_value The default value of the argument
#'
#' @return A string indicating the type of the argument. The possible types are:
#' - "integer": The argument is an integer value
#' - "numeric": The argument is a numeric value
#' - "logical": The argument is a logical value
#' - "string": The argument is a character value
#' - "match.arg": The argument is a match.arg-type argument
#' - "vector \[<type>\]": The argument is a vector of the specified type
#'   (e.g. "integer", "numeric", "logical", "string", or "unknown")
#' - "list": The argument is a list
#' - "call": The argument is a call to a function
#' - "unknown": The type of the argument is unknown
#'
#' @noRd
#' @keywords internal
gd_infer_type_from_default <- function(default_value) {
  is_whole_number <- function(x) {
    tryCatch(
      x %% 1 == 0,
      error = function(e) FALSE
    )
  }

  infer_list_types <- function(lst) {
    # Recursively infer types for each element in a named list
    sapply(lst, gd_infer_type_from_default, simplify = FALSE)
  }

  if (is.null(default_value)) {
    return("unknown")
  }
  if (is.numeric(default_value) && length(default_value) == 1) {
    if (is_whole_number(default_value)) {
      return("integer")
    }
    return("numeric")
  } else if (is.logical(default_value) && length(default_value) == 1) {
    return("logical")
  } else if (is.character(default_value) && length(default_value) == 1) {
    return("string")
  } else if (is.call(default_value)) {
    func_name <- as.character(default_value[[1]])
    if (
      func_name == "c" ||
        (func_name == "list" && is.null(names(default_value[-1])))
    ) {
      # Treat list like c only if it has no names
      values <- default_value[-1]

      if (length(values) == 0) {
        return("vector unknown")
      }

      if (length(values) > 1 && all(sapply(values, is.character))) {
        # Assuming it's a match.arg-type vector
        return("match.arg")
      } else {
        if (all(sapply(values, is.numeric))) {
          if (all(sapply(values, is_whole_number))) {
            return("vector integer")
          }
          return("vector numeric")
        } else if (all(sapply(values, is.logical))) {
          return("vector logical")
        } else if (all(sapply(values, is.character))) {
          return("vector string")
        }

        return("vector unknown")
      }
    } else if (func_name == "list") {
      # Handle named lists
      if (!is.null(names(default_value[-1]))) {
        return(infer_list_types(as.list(default_value[-1])))
      } else {
        return("list")
      }
    } else {
      return("call")
    }
  } else if (is.list(default_value)) {
    if (!is.null(names(default_value))) {
      # Named list
      return(infer_list_types(default_value))
    } else {
      # Unnamed list
      return("list")
    }
  } else {
    return("unknown")
  }
}

#' Extract help file function
#'
#' @param file String containing name of file
#'
#' @noRd
#' @keywords internal
#'
#' @source Function adapted from the CRAN package
#' \href{https://cran.r-project.org/web/packages/devtoolbox/index.html}{'devtoolbox'},
#' which based this function on the 'utils:::.getHelpFile' function
gd_get_help_file <- function(file) {
  path <- dirname(file)
  dirpath <- dirname(path)

  if (!file.exists(dirpath)) {
    stop(
      gettextf("invalid %s argument", sQuote("file")),
      domain = NA
    )
  }

  pkgname <- basename(dirpath)
  RdDB <- file.path(path, pkgname)

  if (!file.exists(paste0(RdDB, ".rdx"))) {
    stop(
      gettextf(
        "package %s exists but was not installed under R >= 2.10.0 so help cannot be accessed",
        sQuote(pkgname)
      ),
      domain = NA
    )
  }

  # internal function from tools
  fetchRdDB <- function(filebase, key = NULL) {
    fun <- function(db) {
      vals <- db$vals
      vars <- db$vars
      datafile <- db$datafile
      compressed <- db$compressed
      envhook <- db$envhook

      fetch <- function(key) {
        lazyLoadDBfetch(vals[key][[1L]], datafile, compressed, envhook)
      }

      if (length(key)) {
        if (!key %in% vars) {
          stop(
            gettextf(
              "No help on %s found in RdDB %s",
              sQuote(key),
              sQuote(filebase)
            ),
            domain = NA
          )
        }

        fetch(key)
      } else {
        res <- lapply(vars, fetch)
        names(res) <- vars
        res
      }
    }

    res <- lazyLoadDBexec(filebase, fun)

    if (length(key)) res else invisible(res)
  }

  fetchRdDB(RdDB, basename(file))
}

#' Function to parse help text from a function's help file
#'
#' @param help_text The text extracted from a function's help file
#'
#' @return A list with the following elements:
#' - 'Title': The title of the help file
#' - 'Description': The description of the function
#' - 'Arguments': A named list of arguments with descriptions
#' - 'Value': The return value or side effects of the function
#'
#' @noRd
#' @keywords internal
gd_parse_help_text <- function(help_text) {
  # Helper function to remove overstruck sequences
  remove_overstrike <- function(text) {
    repeat {
      # Replace overstruck sequences
      new_text <- gsub("(.)(\\x08)(.)", "\\3", text, perl = TRUE)
      if (identical(new_text, text)) {
        break
      }
      text <- new_text
    }
    return(text)
  }

  # Clean up and remove formatting artifacts from all lines
  clean_text <- remove_overstrike(help_text)
  clean_text <- gsub("\\\\b", "", clean_text) # Remove remaining \b escape sequences
  clean_text <- trimws(clean_text) # Remove leading/trailing whitespace

  # Extract the first non-empty line as the title
  title <- clean_text[nzchar(clean_text)][1]
  title_index <- which(clean_text == title)[1]
  clean_text <- clean_text[-title_index] # Remove the title line

  # Find headers dynamically by looking for lines that match header pattern
  header_pattern <- "^[[:space:]]*[[:alpha:][:space:]\\(\\)]+:\\s*$"
  header_indices <- grep(header_pattern, clean_text)
  headers <- clean_text[header_indices]
  headers <- gsub(":", "", headers) # Remove colon for easier matching
  headers <- trimws(headers)

  # Initialize result list
  parsed_result <- list(Title = title)

  # Loop through headers and extract content
  for (i in seq_along(headers)) {
    current_header <- headers[i]
    current_start <- header_indices[i]
    next_start <- ifelse(
      i < length(header_indices),
      header_indices[i + 1] - 1,
      length(clean_text)
    )

    # Extract content for the current section
    section_content <- clean_text[(current_start + 1):next_start]
    # Remove empty lines
    section_content <- section_content[nzchar(section_content)]

    # For "Arguments" section, process differently
    if (current_header == "Arguments") {
      parsed_result[["Arguments"]] <- section_content
    } else {
      # For other sections, join the lines
      section_text <- paste(section_content, collapse = "\n")
      parsed_result[[current_header]] <- section_text
    }
  }

  # Process Arguments section to extract argument names and descriptions
  if ("Arguments" %in% names(parsed_result)) {
    args_content <- parsed_result[["Arguments"]]
    args_list <- list()
    arg_name <- NULL
    arg_desc_lines <- c()

    for (line in args_content) {
      # Check if the line is an argument name
      arg_line_pattern <- "^\\s*([a-zA-Z0-9_\\.]+):\\s*(.*)$"
      matches <- regexec(arg_line_pattern, line)
      match <- regmatches(line, matches)[[1]]

      if (length(match) >= 2 && nzchar(match[2])) {
        # Save previous argument description if exists
        if (!is.null(arg_name)) {
          arg_desc <- paste(arg_desc_lines, collapse = " ")
          arg_desc <- trimws(arg_desc)
          args_list[[arg_name]] <- arg_desc
        }
        # Start new argument
        arg_name <- match[2]
        arg_desc_lines <- if (nzchar(match[3])) match[3] else c()
      } else if (!is.null(arg_name)) {
        # Continuation of argument description
        arg_desc_lines <- c(arg_desc_lines, line)
      }
    }
    # Save the last argument description
    if (!is.null(arg_name)) {
      arg_desc <- paste(arg_desc_lines, collapse = " ")
      arg_desc <- trimws(arg_desc)
      args_list[[arg_name]] <- arg_desc
    }
    parsed_result[["Arguments"]] <- args_list
  }

  # Return parsed result
  return(parsed_result)
}

#' Convert function documentation to a JSON schema
#'
#' This function converts function documentation to an R list that represents
#' a JSON schema. The JSON schema can be used as input LLM providers
#' which support native function calling and require a JSON schema to
#' describe function and its arguments.
#'
#' @param docs Function documentation as returned by [tools_get_docs()]
#' @param all_required Logical indicating whether all arguments should be required.
#' 'TRUE' currently required for the OpenAI API
#' @param additional_properties Logical indicating whether additional properties
#' are allowed. See your LLM provider's documentation to know if this is
#' supported
#'
#' @return A list (R object) representing a JSON schema for the function
#'
#' @noRd
#' @keywords internal
tools_docs_to_r_json_schema <- function(
  docs,
  all_required = TRUE,
  additional_properties = FALSE
) {
  # Helper function to process each argument recursively
  process_argument <- function(arg) {
    if (.tidyprompt_arg_is_ignored(arg)) {
      return(NULL)
    }

    prop <- list()
    r_type <- arg$type

    normalize_r_type <- function(x) {
      if (is.list(x) || is.null(x)) {
        return(x)
      }
      x <- tolower(trimws(as.character(x)))
      switch(
        x,
        "character" = "string",
        "str" = "string",
        "bool" = "logical",
        "boolean" = "logical",
        "double" = "numeric",
        "float" = "numeric",
        "int" = "integer",
        x
      )
    }
    r_type <- normalize_r_type(arg$type)

    if (is.list(r_type)) {
      # Handle named lists (objects)
      prop$type <- "object"
      prop$additionalProperties <- FALSE # Set additionalProperties to FALSE
      prop$properties <- list()
      for (name in names(r_type)) {
        sub_arg <- list(
          type = r_type[[name]],
          default_value = if (!is.null(arg$default_value[[name]])) {
            arg$default_value[[name]]
          } else {
            NULL
          }
        )
        sub_prop <- process_argument(sub_arg)
        if (!is.null(sub_prop)) {
          prop$properties[[name]] <- sub_prop
        }
      }
    } else if (is.null(r_type) || r_type == "unknown") {
      prop$type <- "string"
    } else if (r_type == "string") {
      prop$type <- "string"
    } else if (r_type == "integer") {
      prop$type <- "integer"
    } else if (r_type == "numeric") {
      prop$type <- "number"
    } else if (r_type == "logical") {
      prop$type <- "boolean"
    } else if (r_type == "match.arg") {
      prop$type <- "string"
      # Check if default_value is a call, e.g. c("Val1", "Val2", ...)
      if (
        is.call(arg$default_value) &&
          identical(arg$default_value[[1]], as.name("c"))
      ) {
        # Evaluate the call to get a standard R vector
        prop$enum <- eval(arg$default_value)
      } else {
        prop$enum <- arg$default_value
      }
    } else if (grepl("^vector ", r_type)) {
      # Handle vector types
      item_type <- sub("^vector ", "", r_type)
      prop$type <- "array"
      if (item_type == "integer") {
        prop$items <- list(type = "integer")
      } else if (item_type == "numeric") {
        prop$items <- list(type = "number")
      } else if (item_type %in% c("string", "character")) {
        prop$items <- list(type = "string")
      } else if (item_type == "logical") {
        prop$items <- list(type = "boolean")
      } else {
        prop$items <- list()
      }
    } else if (r_type == "list") {
      # Handle unnamed lists
      default_value <- arg$default_value
      if (is.null(default_value)) {
        prop$type <- "array"
        prop$items <- list()
      } else if (is.list(default_value)) {
        if (is.null(names(default_value))) {
          # Unnamed list (array)
          prop$type <- "array"
          if (length(default_value) > 0) {
            # Assume homogeneous items
            sub_arg <- list(
              type = arg$type,
              default_value = default_value[[1]]
            )
            prop$items <- process_argument(sub_arg)
          } else {
            prop$items <- list()
          }
        } else {
          # Named list (object)
          prop$type <- "object"
          prop$additionalProperties <- FALSE # Set additionalProperties to FALSE
          prop$properties <- list()
          for (name in names(default_value)) {
            sub_arg <- list(
              type = arg$type[[name]],
              default_value = default_value[[name]]
            )
            sub_prop <- process_argument(sub_arg)
            if (!is.null(sub_prop)) {
              prop$properties[[name]] <- sub_prop
            }
          }
        }
      } else {
        prop$type <- "array"
        prop$items <- list()
      }
    } else if (r_type == "call") {
      cli::cli_alert_warning(
        paste0(
          "{.strong `answer_using_tools()`, `tools_docs_to_r_json_schema()`}:\n",
          "* Function calls are not possible as argument type; defaulting to 'string'"
        )
      )
      prop$type <- "string"
    } else {
      cli::cli_alert_warning(
        paste0(
          "{.strong `answer_using_tools()`, `tools_docs_to_r_json_schema()`}:\n",
          "* Unknown argument type ('",
          r_type,
          "'); ",
          "defaulting to 'string'"
        )
      )
      prop$type <- "string"
    }
    return(prop)
  }

  # Process arguments
  args <- docs$arguments
  properties <- list()
  required_args <- character(0)

  for (arg_name in names(args)) {
    arg <- args[[arg_name]]
    prop <- process_argument(arg)
    if (is.null(prop)) {
      next
    }

    # Add to required arguments if there's no default value
    if (!"default_value" %in% names(arg)) {
      required_args <- c(required_args, arg_name)
    }

    # Add property to properties list
    properties[[arg_name]] <- prop
  }

  if (all_required) {
    required_args <- names(properties)
  }

  list(
    type = "object",
    properties = properties,
    required = I(required_args),
    additionalProperties = additional_properties
  )
}

#' Create text description of function from documentation
#'
#' @param docs See [tools_get_docs()]
#' @param with_arguments Logical indicating whether to include arguments in the text
#'
#' @return A string with the arguments formatted as text
#'
#' @noRd
#' @keywords internal
tools_docs_to_text <- function(docs, with_arguments = TRUE) {
  # Internal helper function to process arguments
  tools_docs_to_text_arguments <- function(docs, line_prefix = "  ") {
    process_argument <- function(arg_name, arg_info, prefix = "  ") {
      if (.tidyprompt_arg_is_ignored(arg_info)) {
        return(NULL)
      }

      # Warning for unknown type
      if (arg_info$type == "unknown") {
        cli::cli_alert_warning(
          paste0(
            "{.strong `answer_using_tools()`, `tools_docs_to_text()`}:\n",
            "* Argument '",
            arg_name,
            "' has an unknown type. Defaulting to 'string'"
          )
        )
      }

      # Start argument text
      arg_text <- glue::glue("{prefix}- {arg_name}:", .trim = FALSE)

      # Add description if available
      if (!is.null(arg_info$description)) {
        arg_text <- glue::glue(
          "{arg_text} {arg_info$description}",
          .trim = FALSE
        )
      }

      # Handle match.arg type
      if (arg_info$type == "match.arg") {
        arg_options <- paste(
          paste0('"', arg_info$default_value, '"'),
          collapse = ", "
        )
        arg_text <- glue::glue(
          "{arg_text} [Type: string (one of: {arg_options})]",
          .trim = FALSE
        )
      } else {
        arg_text <- glue::glue(
          "{arg_text} [Type: {arg_info$type}]",
          .trim = FALSE
        )
      }

      return(arg_text)
    }

    process_nested_list <- function(arg_list, prefix = "") {
      nested_text <- ""
      for (arg_name in names(arg_list)) {
        arg_info <- arg_list[[arg_name]]

        if (.tidyprompt_arg_is_ignored(arg_info)) {
          next
        }

        # If arg_info is not a list, assume it represents the type
        if (!is.list(arg_info)) {
          arg_info <- list(type = arg_info)
        }

        # Check if we have a nested structure
        if (is.list(arg_info$type) || is.null(arg_info$type)) {
          # Include description if available
          if (!is.null(arg_info$description)) {
            nested_line <- glue::glue(
              "{prefix}- {arg_name}: {arg_info$description} [Type: named list]",
              .trim = FALSE
            )
          } else {
            nested_line <- glue::glue(
              "{prefix}- {arg_name}: [Type: named list]",
              .trim = FALSE
            )
          }

          # Use arg_info$type if it's a list, else use arg_info
          next_level <- if (is.list(arg_info$type)) arg_info$type else arg_info
          nested_subtext <- process_nested_list(
            next_level,
            paste0(prefix, "  ")
          )
          nested_text <- paste(
            nested_text,
            nested_line,
            nested_subtext,
            sep = "\n"
          )
        } else {
          # For simple arguments, handle normally
          arg_line <- process_argument(arg_name, arg_info, prefix)
          if (is.null(arg_line)) {
            next
          }
          nested_text <- paste(
            nested_text,
            arg_line,
            sep = "\n"
          )
        }
      }
      return(nested_text)
    }

    tool_llm_text <- process_nested_list(docs$arguments)

    remove_empty_lines <- function(text) {
      # Split into individual lines
      lines <- unlist(strsplit(text, "\n", fixed = TRUE))

      # Keep only lines that are not empty (after trimming whitespace)
      non_empty_lines <- lines[!grepl("^\\s*$", lines)]

      # Rejoin the cleaned lines into a single string
      cleaned_text <- paste(non_empty_lines, collapse = "\n")

      return(cleaned_text)
    }

    # Apply line prefix to each line after cleaning
    cleaned_result <- remove_empty_lines(tool_llm_text)
    if (nzchar(line_prefix)) {
      lines <- strsplit(cleaned_result, "\n")[[1]]
      lines <- paste0(line_prefix, lines)
      cleaned_result <- paste(lines, collapse = "\n")
    }

    cleaned_result
  }

  # Define text
  tool_llm_text <- glue::glue(
    "  function name: {docs$name}",
    .trim = FALSE
  )
  if (length(docs$description) > 0) {
    tool_llm_text <- glue::glue(
      "{tool_llm_text}\n  description: {docs$description}",
      .trim = FALSE
    )
  }
  if (length(docs$arguments) > 0 && with_arguments) {
    tool_llm_text <- glue::glue(
      "{tool_llm_text}\n  arguments:",
      "\n",
      tools_docs_to_text_arguments(docs, line_prefix = "    "),
      .trim = FALSE
    )
  }
  if (length(docs$return$description) > 0) {
    tool_llm_text <- glue::glue(
      "{tool_llm_text}\n  return value: {docs$return$description}",
      .trim = FALSE
    )
  }

  return(tool_llm_text)
}

# Fake function to test; requires some numeric input, gives back random number
# To verify tool call worked we'll verify if random number is in LLM response
secret_number <- sample(1000000:9999999, 1)
get_secret_number <- function(input = 12345) {
  return(secret_number)
}

# Check if a number is present in a string, ignoring commas, dots, or spaces
number_present_in_string <- function(string, number) {
  if (is.null(string) || is.null(number)) {
    return(FALSE)
  }

  # extract just the digits from the number
  digits <- gsub("\\D", "", as.character(number))
  if (digits == "") {
    return(FALSE)
  }

  # build a regex that allows commas, dots, or spaces between digits
  pieces <- strsplit(digits, "")[[1]]
  pattern <- paste0("(?<!\\d)", paste(pieces, collapse = "[,\\. ]?"), "(?!\\d)")

  grepl(pattern, string, perl = TRUE)
}

base_prompt <- "Call the function and tell me the output number"

test_that("text-based function calling works with base r function", {
  skip_test_if_no_openai()

  result <- "What are the files in my current directory?" |>
    answer_using_tools(dir, type = "text-based") |>
    send_prompt(llm_provider_openai()$set_parameters(list(
      model = "gpt-4.1-mini"
    )))

  expect_true(!is.null(result))
})

test_that("text-based function calling works with custom function", {
  skip_test_if_no_openai()

  # Example fake weather function to add to the prompt:
  temperature_in_location <- function(
    location = c("Amsterdam", "Utrecht", "Enschede"),
    unit = c("Celcius", "Fahrenheit")
  ) {
    location <- match.arg(location)
    unit <- match.arg(unit)

    temperature_celcius <- switch(
      location,
      "Amsterdam" = 32.5,
      "Utrecht" = 19.8,
      "Enschede" = 22.7
    )

    if (unit == "Celcius") {
      return(temperature_celcius)
    } else {
      return(temperature_celcius * 9 / 5 + 32)
    }
  }

  # Generate documentation for a function
  #   (based on formals, & help file if available)
  docs <- tools_get_docs(temperature_in_location)
  docs$description <- "Get the temperature in a location"
  docs$arguments$unit$description <- "Unit in which to return the temperature"
  docs$arguments$location$description <- "Location for which to return the temperature"
  docs$
  return$description <- "The temperature in the specified location and unit"
  temperature_in_location <- tools_add_docs(temperature_in_location, docs)

  prompt <- "Hi, what is the weather in Enschede? Give me Celcius degrees" |>
    answer_using_tools(temperature_in_location, type = "text-based")

  result <- prompt |>
    send_prompt(llm_provider_openai()$set_parameters(list(
      model = "gpt-4.1-mini"
    )))

  expect_true(!is.null(result))
})

test_that("native ollama function calling works", {
  skip_test_if_no_ollama()

  result <- base_prompt |>
    answer_using_tools(get_secret_number) |>
    send_prompt(llm_provider_ollama())

  expect_true(number_present_in_string(result, secret_number))
})

test_that("native openai function calling works (chat completions API, stream)", {
  skip_test_if_no_openai()

  result <- base_prompt |>
    answer_using_tools(get_secret_number, type = "openai") |>
    send_prompt(
      llm_provider_openai(
        url = "https://api.openai.com/v1/chat/completions"
      )$set_parameters(list(model = "gpt-4.1-mini")),
      stream = TRUE
    )

  expect_true(number_present_in_string(result, secret_number))
})

test_that("native openai function calling works (chat completions API, non-stream)", {
  skip_test_if_no_openai()

  result <- base_prompt |>
    answer_using_tools(get_secret_number, type = "openai") |>
    send_prompt(
      llm_provider_openai(
        url = "https://api.openai.com/v1/chat/completions"
      )$set_parameters(list(model = "gpt-4.1-mini")),
      stream = FALSE
    )

  expect_true(number_present_in_string(result, secret_number))
})

test_that("native openai function calling works (responses API, stream)", {
  skip_test_if_no_openai()

  prompt <- "What files are in my current directory?" |>
    answer_using_tools(dir, type = "openai")

  result <- base_prompt |>
    answer_using_tools(get_secret_number, type = "openai") |>
    send_prompt(
      llm_provider_openai(
        url = "https://api.openai.com/v1/responses"
      )$set_parameters(list(model = "gpt-4.1-mini")),
      stream = TRUE
    )

  expect_true(number_present_in_string(result, secret_number))
})

test_that("native openai function calling works (responses API, non-stream)", {
  skip_test_if_no_openai()

  result <- base_prompt |>
    answer_using_tools(get_secret_number, type = "openai") |>
    send_prompt(
      llm_provider_openai(
        url = "https://api.openai.com/v1/responses"
      )$set_parameters(list(model = "gpt-4.1-mini")),
      stream = FALSE
    )

  expect_true(number_present_in_string(result, secret_number))
})

test_that("function calling works with ellmer llm provider", {
  skip_test_if_no_openai()
  skip_if_not_installed("ellmer")

  ellmer_openai <- llm_provider_ellmer(ellmer::chat_openai(
    model = "gpt-4.1-mini"
  ))

  result <- base_prompt |>
    answer_using_tools(get_secret_number) |>
    send_prompt(ellmer_openai)

  expect_true(number_present_in_string(result, secret_number))
})

test_that("function calling works with direct ellmer chat", {
  skip_test_if_no_openai()
  skip_if_not_installed("ellmer")

  ellmer_chat <- ellmer::chat_openai(model = "gpt-4.1-mini")

  result <- base_prompt |>
    answer_using_tools(get_secret_number) |>
    send_prompt(ellmer_chat)

  expect_true(number_present_in_string(result, secret_number))
  expect_length(ellmer_chat$get_turns(), 0)
})

test_that("function calling works with ellmer tool definition", {
  skip_test_if_no_openai()
  skip_if_not_installed("ellmer")

  ellmer_openai <- llm_provider_ellmer(ellmer::chat_openai(
    model = "gpt-4.1-mini"
  ))

  tool_def <- ellmer::tool(
    get_secret_number,
    name = "get_secret_number",
    description = "Get a secret number",
    arguments = list(
      input = ellmer::type_number()
    )
  )

  # Using ellmer LLM provider
  result_ellmer_llm_provider <- base_prompt |>
    answer_using_tools(tool_def) |>
    send_prompt(ellmer_openai)
  expect_true(number_present_in_string(
    result_ellmer_llm_provider,
    secret_number
  ))

  # Using tidyprompt LLM provider
  result_tidyprompt_llm_provider <- base_prompt |>
    answer_using_tools(tool_def) |>
    send_prompt(
      llm_provider_openai(
        url = "https://api.openai.com/v1/responses"
      )$set_parameters(list(model = "gpt-4.1-mini"))
    )
  expect_true(number_present_in_string(
    result_tidyprompt_llm_provider,
    secret_number
  ))
})

test_that("messages_from_history omits non-replayable rows", {
  history <- data.frame(
    role = c("user", "assistant", "assistant", "tool", "assistant"),
    content = c(
      "Question",
      "Reasoning step",
      "~>> Calling function 'tool' with arguments:\n{}",
      "~>> Result:\n42",
      "Answer"
    ),
    hidden_from_llm = c(FALSE, TRUE, FALSE, FALSE, FALSE),
    tool_call = c(FALSE, FALSE, TRUE, FALSE, FALSE),
    tool_result = c(FALSE, FALSE, FALSE, TRUE, FALSE),
    stringsAsFactors = FALSE
  )

  messages <- .messages_from_history(history)

  expect_equal(
    vapply(messages, `[[`, character(1), "content"),
    c("Question", "~>> Result:\n42", "Answer")
  )
  expect_equal(
    vapply(messages, `[[`, character(1), "role"),
    c("user", "tool", "assistant")
  )
})

test_that("text-based dispatch uses ellmer tool's declared name, not R symbol", {
  skip_if_not_installed("ellmer")

  add_xy <- function(x, y) x + y

  td <- ellmer::tool(
    add_xy,
    name = "sum_two",
    description = "Add two numbers",
    arguments = list(
      x = ellmer::type_number(),
      y = ellmer::type_number()
    )
  )

  state <- new.env(parent = emptyenv())
  state$call_n <- 0L

  provider <- `llm_provider-class`$new(
    complete_chat_function = function(chat_history) {
      state <- self$parameters$.test_state
      state$call_n <- state$call_n + 1L

      if (state$call_n == 1L) {
        # Model calls the tool by its declared name
        reply <- '{"function": "sum_two", "arguments": {"x": 3, "y": 4}}'
      } else {
        reply <- "The answer is 7"
      }

      list(
        completed = dplyr::bind_rows(
          chat_history,
          data.frame(
            role = "assistant",
            content = reply,
            stringsAsFactors = FALSE
          )
        ),
        http = list(request = NULL, response = NULL)
      )
    },
    parameters = list(.test_state = state),
    verbose = FALSE
  )

  result <- "What is 3 + 4?" |>
    answer_using_tools(td, type = "text-based") |>
    send_prompt(provider, verbose = FALSE)

  # If dispatch failed, the tool call would error and we'd never reach call 2
  expect_equal(state$call_n, 2L)
  expect_equal(result, "The answer is 7")
})

test_that("openai path uses ellmer tool's declared name in API payload", {
  skip_if_not_installed("ellmer")

  add_xy <- function(x, y) x + y

  td <- ellmer::tool(
    add_xy,
    name = "sum_two",
    description = "Add two numbers",
    arguments = list(
      x = ellmer::type_number(),
      y = ellmer::type_number()
    )
  )

  # Build prompt and extract the parameter_fn to inspect the OpenAI tool payload
  prompt <- "What is 3 + 4?" |>
    answer_using_tools(td, type = "openai")

  wraps <- get_prompt_wraps(prompt)
  # Find the wrap with a parameter_fn
  param_wrap <- NULL
  for (w in wraps) {
    if (!is.null(w$parameter_fn)) {
      param_wrap <- w
      break
    }
  }
  expect_false(is.null(param_wrap))

  # Call parameter_fn with a mock provider
  mock_provider <- list(tool_type = "openai", api_type = "openai")
  params <- param_wrap$parameter_fn(mock_provider)

  # The tools payload should have function.name = "sum_two"
  expect_true(!is.null(params$tools))
  expect_equal(params$tools[[1]][["function"]]$name, "sum_two")
})

test_that("tools_get_docs overrides attribute name when caller provides one", {
  my_fn <- function(x) x + 1
  docs_attr <- list(
    name = "original_name",
    description = "A test function",
    arguments = list(x = list(type = "number", description = "A number"))
  )
  attr(my_fn, "tidyprompt_tool_docs") <- docs_attr

  # Without explicit name, docs$name stays as attribute value
  docs_default <- tools_get_docs(my_fn)
  expect_equal(docs_default$name, "original_name")

  # With explicit name, docs$name is overridden

  docs_override <- tools_get_docs(my_fn, name = "dispatch_key")
  expect_equal(docs_override$name, "dispatch_key")
})

test_that("partially named tool list preserves explicit aliases", {
  skip_if_not_installed("ellmer")

  add_xy <- function(x, y) x + y

  td <- ellmer::tool(
    add_xy,
    name = "sum_two",
    description = "Add two numbers",
    arguments = list(
      x = ellmer::type_number(),
      y = ellmer::type_number()
    )
  )

  h <- function(x) x + 1

  state <- new.env(parent = emptyenv())
  state$call_n <- 0L

  provider <- `llm_provider-class`$new(
    complete_chat_function = function(chat_history) {
      state <- self$parameters$.test_state
      state$call_n <- state$call_n + 1L

      if (state$call_n == 1L) {
        # Model calls the tool by the user-supplied alias
        reply <- '{"function": "custom_alias", "arguments": {"x": 3, "y": 4}}'
      } else {
        reply <- "The answer is 7"
      }

      list(
        completed = dplyr::bind_rows(
          chat_history,
          data.frame(
            role = "assistant",
            content = reply,
            stringsAsFactors = FALSE
          )
        ),
        http = list(request = NULL, response = NULL)
      )
    },
    parameters = list(.test_state = state),
    verbose = FALSE
  )

  result <- "What is 3 + 4?" |>
    answer_using_tools(
      tools = list(custom_alias = td, h = h),
      type = "text-based"
    ) |>
    send_prompt(provider, verbose = FALSE)

  # Dispatch must honour the user alias "custom_alias", not the declared "sum_two"
  expect_equal(state$call_n, 2L)
  expect_equal(result, "The answer is 7")
})

test_that("partially named list: unnamed ellmer tool gets declared @name", {
  skip_if_not_installed("ellmer")

  add_xy <- function(x, y) x + y

  td <- ellmer::tool(
    add_xy,
    name = "sum_two",
    description = "Add two numbers",
    arguments = list(
      x = ellmer::type_number(),
      y = ellmer::type_number()
    )
  )

  h <- function(x) x + 1

  # Build prompt and inspect tool names in the text-based instruction
  prompt <- "What is 3 + 4?" |>
    answer_using_tools(
      tools = list(my_alias = h, td),
      type = "text-based"
    )

  prompt_text <- construct_prompt_text(prompt)
  # The explicit alias must appear

  expect_true(grepl("my_alias", prompt_text))
  # The unnamed ellmer tool should use its declared @name, not the symbol "td"
  expect_true(grepl("sum_two", prompt_text))
  expect_false(grepl("\\btd\\b", prompt_text))
})

test_that("ellmer path preserves explicit aliases for ToolDefs", {
  skip_if_not_installed("ellmer")

  add1 <- function(x) x + 1
  add2 <- function(x) x + 2

  td1 <- ellmer::tool(
    add1,
    name = "shared_name",
    description = "First tool",
    arguments = list(x = ellmer::type_number())
  )
  td2 <- ellmer::tool(
    add2,
    name = "shared_name",
    description = "Second tool",
    arguments = list(x = ellmer::type_number())
  )

  provider <- llm_provider_ellmer(
    fake_ellmer_chat(),
    parameters = list(stream = FALSE),
    verbose = FALSE
  )

  result <- "Use the tools" |>
    answer_using_tools(
      tools = list(alias_one = td1, alias_two = td2),
      type = "ellmer"
    ) |>
    send_prompt(provider, verbose = FALSE, return_mode = "full")

  registered_tools <- result$ellmer_chat$get_tools()
  registered_names <- vapply(
    registered_tools,
    ellmer_tool_name,
    character(1)
  )

  expect_length(registered_tools, 2)
  expect_equal(registered_names, c("alias_one", "alias_two"))
})

test_that("ellmer path preserves explicit aliases for regular functions", {
  skip_if_not_installed("ellmer")

  increment <- function(x) x + 1
  increment <- tools_add_docs(
    increment,
    list(
      name = "increment",
      description = "Increment a number",
      arguments = list(x = list(type = "numeric"))
    )
  )

  provider <- llm_provider_ellmer(
    fake_ellmer_chat(),
    parameters = list(stream = FALSE),
    verbose = FALSE
  )

  result <- "Use the tool" |>
    answer_using_tools(
      tools = list(custom_alias = increment),
      type = "ellmer"
    ) |>
    send_prompt(provider, verbose = FALSE, return_mode = "full")

  registered_tools <- result$ellmer_chat$get_tools()

  expect_length(registered_tools, 1)
  expect_equal(ellmer_tool_name(registered_tools[[1]]), "custom_alias")
})

test_that("tool name collision from sanitization raises error", {
  f1 <- function(x) x + 1
  f2 <- function(x) x + 2

  expect_error(
    "test" |>
      answer_using_tools(
        tools = list("a-b" = f1, a_b = f2),
        type = "text-based"
      ),
    "Tool name collision"
  )
})

test_that("tool name collision from ellmer @name raises error", {
  skip_if_not_installed("ellmer")

  add1 <- function(x) x + 1
  add2 <- function(x) x + 2

  td1 <- ellmer::tool(
    add1,
    name = "my_tool",
    description = "First tool",
    arguments = list(x = ellmer::type_number())
  )
  td2 <- ellmer::tool(
    add2,
    name = "my_tool",
    description = "Second tool",
    arguments = list(x = ellmer::type_number())
  )

  expect_error(
    "test" |>
      answer_using_tools(tools = list(td1, td2), type = "text-based"),
    "Tool name collision"
  )
})

test_that("tools_get_docs finds help for dotted function names under sanitized alias", {
  # list.files is a base R function; sanitize_name() turns it into "list_files"
  docs <- tools_get_docs(list.files, "list_files")

  expect_equal(docs$name, "list_files")
  # Should have a description from the help file, not just NULL
  expect_true(
    is.character(docs$description) && nzchar(docs$description)
  )
  # Should have documented arguments
  expect_true(length(docs$arguments) > 0)
})

test_that("tools_get_docs finds help for namespaced dotted names", {
  docs <- tools_get_docs(utils::download.file, "download_file")

  expect_equal(docs$name, "download_file")
  expect_true(
    is.character(docs$description) && nzchar(docs$description)
  )
})

#' Create an `llm_feedback` object
#'
#' This object is used to send feedback to a LLM when a LLM reply
#' does not succesfully pass an extraction or validation function
#' (as handled by [send_prompt()] and defined using [prompt_wrap()]).
#' The feedback text is sent back to the LLM. The extraction or validation function
#' should then return this object with the feedback text that should be sent to the LLM.
#'
#' @param text A character string containing the feedback text. This will
#' be sent back to the LLM after not passing an extractor or validator function
#'
#' @param tool_result A logical indicating whether the feedback is a tool result.
#' If TRUE, [send_prompt()] will not remove it from the chat history when
#' cleaning the context window during repeated interactions
#'
#' @return An object of class "llm_feedback" (or "llm_feedback_tool_result")
#' containing the feedback text to send back to the LLM
#'
#' @export
#'
#' @example inst/examples/llm_feedback.R
#'
#' @family prompt_wrap
#' @family prompt_evaluation
#' @seealso [llm_break()]
llm_feedback <- function(text, tool_result = FALSE) {
  if (!tool_result) {
    return(
      structure(
        text,
        class = "llm_feedback"
      )
    )
  } else {
    return(
      structure(
        list(
          text = text
        ),
        class = "llm_feedback_tool_result"
      )
    )
  }
}

#' Create an `llm_break` object
#'
#' This object is used to break a extraction and validation loop defined in a [prompt_wrap()]
#' as evaluated by [send_prompt()]. When an extraction or validation function returns
#' this object, the loop will be broken and no further extraction or validation
#' functions are applied; instead, [send_prompt()] will be able to return
#' the result at that point. This may be useful in scenarios where
#' it is determined the LLM is unable to provide a response to a prompt.
#'
#' @param object_to_return The object to return as the response result
#' from [send_prompt()] when this object is returned from an extraction or
#' validation function
#' @param success A logical indicating whether the [send_prompt()] loop break
#' should nonetheless be considered as a successful completion of the
#' extraction and validation process.
#'  If `FALSE`, the `object_to_return` must
#' be `NULL` (as the response result of [send_prompt()] will always be 'NULL'
#' when the evaluation was unsuccessful); if `FALSE`, [send_prompt()] will also
#' print a warning about the unsuccessful evaluation.
#'  If `TRUE`, the `object_to_return`
#' will be returned as the response result of [send_prompt()] (and [send_prompt()])
#' will print no warning about unsuccessful evaluation)
#'
#' @return An list of class "llm_break" containing the object to return and
#' a logical indicating whether the evaluation was successful
#'
#' @export
#'
#' @example inst/examples/llm_break.R
#'
#' @family prompt_wrap
#' @family prompt_evaluation
#' @seealso [llm_feedback()]
llm_break <- function(
  object_to_return = NULL,
  success = FALSE
) {
  if (!is.logical(success)) {
    stop("The 'success' argument must be of type logical.")
  }
  if (!success && !is.null(object_to_return)) {
    object_to_return <- NULL
    warning("If 'success' is FALSE, 'object_to_return' is set to NULL")
  }

  return(
    structure(
      list(
        object_to_return = object_to_return,
        success = success
      ),
      class = "llm_break"
    )
  )
}

#' @title Create an `llm_break_soft` object
#'
#' @description
#' This object is used to break a extraction and validation loop defined in a [prompt_wrap()],
#' as evaluated by [send_prompt()]. When an extraction or validation function returns
#' this object, it will prevent any future interactions with the LLM provider for the
#' current prompt. Remaining extraction and validation functions will still be applied
#' and it will still be possible to pass these with the current response from the LLM
#' provider; only, no more new tries will be made if the current response is not
#' satisfactory.
#'
#' This is useful when, e.g., the token limit for the LLM provider has been reached,
#' but the final response that we got may still be satisfactory. In this case,
#' `llm_break()` cannot be used, as it would instantly return the current response as the final
#' result, which is not what we want. Instead, `llm_break_soft()` can be used to
#' prevent any further interactions with the LLM provider, but still allow the
#' remaining extraction and validation functions to be applied (and have those decide
#' the success of the current response).
#'
#' @param object_to_return The object to return as the response result
#' from [send_prompt()] when this object is returned from an extraction or
#' validation function
#'
#' @return An list of class "llm_break_soft" containing the object to return
#' @export llm_break_soft
#'
#' @example inst/examples/llm_break_soft.R
llm_break_soft <- function(
  object_to_return = NULL
) {
  return(
    structure(
      class = "llm_break_soft",
      list(
        object_to_return = object_to_return
      )
    )
  )
}

#' @title LlmProvider R6 Class
#' @name llm_provider-class
#'
#' @description This class provides a structure for creating [llm_provider-class]
#'  objects with different implementations of `$complete_chat()`.
#' Using this class, you can create an [llm_provider-class] object that interacts
#'  with different LLM providers, such Ollama, OpenAI, or other custom providers.
#'
#' @example inst/examples/llm_provider.R
#'
#' @family llm_provider
NULL

#' @rdname llm_provider-class
#' @export
`llm_provider-class` <- R6::R6Class(
  "LlmProvider",
  public = list(
    #' @field parameters
    #' A named list of parameters to configure the [llm_provider-class].
    #' Parameters may be appended to the request body when interacting with the
    #'  LLM provider API
    parameters = list(),

    #' @field verbose
    #' A logical indicating whether interaction with the LLM provider should be
    #'  printed to the console
    verbose = getOption("tidyprompt.verbose", TRUE),

    #' @field url
    #' The URL to the LLM provider API endpoint for chat completion
    url = NULL,

    #' @field api_key
    #' The API key to use for authentication with the LLM provider API
    api_key = NULL,

    #' @field api_type
    #' The type of API to use (e.g., "openai", "ollama", "ellmer").
    #'  This is used to determine certain specific behaviors for different APIs,
    #'  for instance, as is done in the [answer_as_json()] function
    api_type = "unspecified",

    #' @field json_type
    #' The type of JSON mode to use (e.g., 'auto', 'openai', 'ollama', 'ellmer', or 'text-based').
    #'  Using 'auto' or having this field not set, the api_type field will be used to
    #'  determine the JSON mode during the [answer_as_json()] function. If this field
    #'  is set, this will override the api_type field for JSON mode determination.
    #'  (Note: this determination only happens when the 'type' argument in
    #'  [answer_as_json()] is also set to 'auto'.)
    json_type = "auto",

    #' @field tool_type
    #' The type of tool use mode to use (e.g., 'auto', 'openai', 'ollama', 'ellmer', or 'text-based').
    #'  Using 'auto' or having this field not set, the api_type field will be used to
    #'  determine the tool use mode during the [answer_using_tools()] function. If this field
    #'  is set, this will override the api_type field for tool use mode determination
    #'  (Note: this determination only happens when the 'type' argument in
    #'  [answer_using_tools()] is also set to 'auto'.)
    tool_type = "auto",

    #' @field handler_fns
    #' A list of functions that will be called after the completion of a chat.
    #'  See `$add_handler_fn()`
    handler_fns = list(),

    #' @field stream_callback
    #' Optional callback function for streaming tokens/chunks. If set, this function will be
    #' called with arguments `(chunk, meta)` where `chunk` is the latest
    #' streamed text chunk and `meta` is a list with fields including `llm_provider`,
    #' `chat_history`, and `partial_response` to provide more context about the current
    #' prompt that is being replied to. You may use this to implement custom streaming behaviour;
    #' see `vignette(`"streaming_shiny_ipc", "tidyprompt")` for an example of how this
    #' function is used to stream a non-blocking async LLM response to a Shiny app using the
    #' 'ipc' package.
    stream_callback = NULL,

    #' @field pre_prompt_wraps
    #' A list of prompt wraps that will be applied to any prompt evaluated
    #' by this [llm_provider-class] object, before any prompt-specific
    #' prompt wraps are applied. See `$add_prompt_wrap()`.
    #' This can be used to set default behavior for all prompts
    #' evaluated by this [llm_provider-class] object.
    pre_prompt_wraps = list(),

    #' @field post_prompt_wraps
    #' A list of prompt wraps that will be applied to any prompt evaluated
    #' by this [llm_provider-class] object, after any prompt-specific
    #' prompt wraps are applied. See `$add_prompt_wrap()`.
    #' This can be used to set default behavior for all prompts
    #' evaluated by this [llm_provider-class] object.
    post_prompt_wraps = list(),

    #' @description
    #' Create a new [llm_provider-class] object
    #'
    #' @param complete_chat_function
    #' Function that will be called by the [llm_provider-class] to complete a chat.
    #' This function should take a list containing at least '$chat_history'
    #'  (a data frame with 'role' and 'content' columns) and return a response
    #'  object, which contains:
    #'  \itemize{
    #'  \item 'completed': A dataframe with 'role' and 'content' columns,
    #'  containing the completed chat history
    #'
    #'  \item 'http': A list containing a list 'requests' and a list 'responses',
    #'  containing the HTTP requests and responses made during the chat completion
    #'  }
    #'
    #' @param parameters
    #' A named list of parameters to configure the [llm_provider-class].
    #'  These parameters may be appended to the request body when interacting with
    #'  the LLM provider.
    #' For example, the `model` parameter may often be required.
    #'  The 'stream' parameter may be used to indicate that the API should stream.
    #' Parameters should not include the chat_history, or 'api_key' or 'url', which
    #'  are handled separately by the [llm_provider-class] and '$complete_chat()'.
    #' Parameters should also not be set when they are handled by prompt wraps
    #'
    #' @param verbose
    #' A logical indicating whether interaction with the LLM
    #'  provider should be printed to the console
    #'
    #' @param url
    #' The URL to the LLM provider API endpoint for chat completion
    #'  (typically required, but may be left NULL in some cases, for instance
    #'  when creating a fake LLM provider)
    #'
    #' @param api_key
    #' The API key to use for authentication with the LLM
    #'  provider API (optional, not required for, for instance, Ollama)
    #'
    #' @param api_type
    #' The type of API to use (e.g., "openai", "ollama").
    #'  This is used to determine certain specific behaviors for different APIs
    #'  (see for example the [answer_as_json()] function)
    #'
    #' @return
    #' A new [llm_provider-class] R6 object
    initialize = function(
      complete_chat_function,
      parameters = list(),
      verbose = TRUE,
      url = NULL,
      api_key = NULL,
      api_type = "unspecified"
    ) {
      if (length(parameters) > 0 && is.null(names(parameters))) {
        stop("parameters must be a named list")
      }

      private$complete_chat_function <- complete_chat_function
      self$parameters <- parameters
      self$verbose <- verbose
      self$url <- url
      self$api_key <- api_key
      self$api_type <- api_type
    },

    #' @description
    #' Helper function to set the parameters of the [llm_provider-class]
    #'  object.
    #' This function appends new parameters to the existing parameters
    #'  list.
    #'
    #' @param new_parameters
    #' A named list of new parameters to append to the
    #'  existing parameters list
    #'
    #' @return The modified [llm_provider-class] object
    set_parameters = function(new_parameters) {
      if (length(new_parameters) == 0) {
        return(self)
      }

      stopifnot(
        is.list(new_parameters),
        length(new_parameters) > 0,
        !is.null(names(new_parameters))
      )
      self$parameters <- utils::modifyList(self$parameters, new_parameters)
      return(self)
    },

    #' @description Sends a chat history (see [chat_history()]
    #'  for details) to the LLM provider using the configured `$complete_chat()`.
    #' This function is typically called by [send_prompt()] to interact with the LLM
    #'  provider, but it can also be called directly.
    #'
    #' @param input A string, a data frame which is a valid chat history
    #'  (see [chat_history()]), or a list containing a valid chat history under key
    #'  '$chat_history'
    #'
    #' @return The response from the LLM provider
    complete_chat = function(input) {
      if (length(input) == 1 && is.character(input)) {
        chat_history <- chat_history(input)
        input <- list(chat_history = chat_history)
      } else if (is.data.frame(input)) {
        chat_history <- chat_history(input)
        input <- list(chat_history = chat_history)
      }

      stopifnot(
        is.list(input),
        "chat_history" %in% names(input)
      )

      chat_history <- chat_history(input$chat_history)
      chat_history_request <- chat_history_to_send(chat_history)
      if (self$verbose) {
        message(
          crayon::bold(
            glue::glue(
              "--- Sending request to LLM provider",
              " ({
              if (!is.null(self$parameters$model)) {
                self$parameters$model
              } else {
                'no model specified'
              }
          }):",
              " ---"
            )
          )
        )

        message(chat_history$content[nrow(chat_history)])
      }

      if (self$verbose) {
        message(
          crayon::bold(
            glue::glue(
              "--- Receiving response from LLM provider: ---"
            )
          )
        )
      }

      environment(private$complete_chat_function) <- environment()
      response <- private$complete_chat_function(chat_history_request)

      # If this is an ellmer provider, sync the chat object
      #   This ensures handler_fn can access the current state of the chat
      #   object (e.g., to monitor cost and cancel if over budget)
      if (!is.null(response$ellmer_chat) && "ellmer_chat" %in% names(self)) {
        self$ellmer_chat <- response$ellmer_chat
      }

      # Store native structured output (from ellmer chat_structured) if present,
      # so extraction functions can use it without JSON round-tripping
      if (!is.null(response$native_structured_result)) {
        self$parameters$.native_structured_result <- response$native_structured_result
      }

      if (nrow(chat_history_request) < nrow(chat_history)) {
        source_rows <- attr(chat_history_request, "source_rows") %||%
          seq_len(nrow(chat_history_request))

        new_rows_start <- nrow(chat_history_request) + 1L
        new_rows <- if (nrow(response$completed) >= new_rows_start) {
          response$completed[
            new_rows_start:nrow(response$completed),
            ,
            drop = FALSE
          ]
        } else {
          response$completed[0, , drop = FALSE]
        }

        # Merge provider-side metadata updates on sent rows back into
        # the full history so that native_turn_id / native_turn_role /
        # native_contents etc. are not discarded.
        sent_count <- min(length(source_rows), nrow(response$completed))
        if (sent_count > 0L) {
          updated <- dplyr::bind_rows(
            chat_history,
            response$completed[0, , drop = FALSE]
          )
          updated <- updated[seq_len(nrow(chat_history)), , drop = FALSE]
          resp_cols <- names(response$completed)
          updated[
            source_rows[seq_len(sent_count)],
            resp_cols
          ] <- response$completed[seq_len(sent_count), , drop = FALSE]
          response$completed <- dplyr::bind_rows(updated, new_rows)
        } else {
          response$completed <- dplyr::bind_rows(chat_history, new_rows)
        }
      }

      response$completed <- normalize_chat_history_metadata(response$completed)

      # Filter empty-content tool_call rows (Ollama sends these); keep
      # user/assistant/system rows so image-only turns are preserved.
      is_empty <- response$completed$content == ""
      is_tool_call <- if ("tool_call" %in% names(response$completed)) {
        as.logical(response$completed$tool_call)
      } else {
        rep(FALSE, nrow(response$completed))
      }
      response$completed <- response$completed[
        !(is_empty & is_tool_call),
      ]

      http <- list()
      http$requests[[1]] <- response$http$request
      http$responses[[1]] <- response$http$response

      repeat {
        for (handler_fn in self$handler_fns) {
          response <- handler_fn(response, self)
          http$requests[[length(http$requests) + 1]] <- response$http$request
          http$responses[[length(http$responses) + 1]] <- response$http$response

          stopifnot(
            is.list(response),
            "completed" %in% names(response),
            is.data.frame(response$completed),
            all(c("role", "content") %in% names(response$completed))
          )

          response$completed <- normalize_chat_history_metadata(
            response$completed
          )

          if (isTRUE(response$`break`)) break
        }

        if (!isFALSE(response$done) || isTRUE(response$`break`)) {
          break
        }
      }

      # Update http list
      response$http <- http

      # Print difference between chat_history and completed
      if (
        self$verbose &&
          (is.null(self$parameters$stream) || !self$parameters$stream)
      ) {
        chat_history_new <- response$completed[
          (nrow(chat_history) + 1):nrow(response$completed),
        ]

        # Filter out rows with 'tool_result' == TRUE
        # That's already being printed in the handler function of
        #   `answer_using_tools()`
        if (
          "tool_result" %in%
            names(chat_history_new) &&
            "tool_call" %in% names(chat_history_new)
        ) {
          chat_history_new_print <- chat_history_new |>
            dplyr::filter(
              (is.na(tool_result) | !tool_result),
              (is.na(tool_call) | !tool_call)
            )
        } else {
          chat_history_new_print <- chat_history_new
        }

        for (i in seq_len(nrow(chat_history_new_print))) {
          message(chat_history_new_print$content[i])
        }
      }

      if (isTRUE(response$`break`)) {
        warning(
          paste0(
            "Chat completion was interrupted by a handler break"
          )
        )
      }

      if (self$verbose) {
        return(invisible(response))
      }

      return(response)
    },

    #' @description
    #' Helper function to add a handler function to the
    #'  [llm_provider-class] object.
    #' Handler functions are called after the completion of a chat and can be
    #'  used to modify the response before it is returned by the [llm_provider-class].
    #' Each handler function should take the response object
    #'  as input (first argument) as well as 'self' (the [llm_provider-class]
    #'  object) and return a modified response object.
    #' The functions will be called in the order they are added to the list.
    #'
    #' @details
    #' If a handler function returns a list with a 'break' field set to `TRUE`,
    #'  the chat completion will be interrupted and the response will be returned
    #'  at that point.
    #' If a handler function returns a list with a 'done' field set to `FALSE`,
    #'  the handler functions will continue to be called in a loop until the 'done'
    #'  field is not set to `FALSE`.
    #'
    #' @param handler_fn A function that takes the response object plus
    #'  'self' (the [llm_provider-class] object) as input and
    #'  returns a modified response object
    add_handler_fn = function(handler_fn) {
      stopifnot(is.function(handler_fn))
      self$handler_fns <- c(self$handler_fns, list(handler_fn))
      return(self)
    },

    #' @description
    #' Helper function to set the handler functions of the
    #'  [llm_provider-class] object.
    #' This function replaces the existing
    #'  handler functions list with a new list of handler functions.
    #' See `$add_handler_fn()` for more information
    #'
    #' @param handler_fns A list of handler functions to set
    set_handler_fns = function(handler_fns) {
      stopifnot(is.list(handler_fns))
      self$handler_fns <- handler_fns
      return(self)
    },

    #' @description
    #' Add a provider-level prompt wrap template to be applied to all prompts.
    #' @param prompt_wrap A list created by [provider_prompt_wrap()]
    #' @param position One of "pre" or "post" (applied before/after prompt-specific wraps)
    add_prompt_wrap = function(prompt_wrap, position = c("pre", "post")) {
      position <- match.arg(position)
      stopifnot(is.list(prompt_wrap))

      # Normalize fields so downstream code can rely on them
      needed <- c(
        "type",
        "modify_fn",
        "extraction_fn",
        "validation_fn",
        "handler_fn",
        "parameter_fn",
        "name"
      )
      missing <- setdiff(needed, names(prompt_wrap))
      for (nm in missing) {
        prompt_wrap[[nm]] <- NULL
      }

      class(prompt_wrap) <- unique(c(
        "provider_prompt_wrap",
        class(prompt_wrap)
      ))

      if (identical(position, "pre")) {
        self$pre_prompt_wraps <- c(self$pre_prompt_wraps, list(prompt_wrap))
      } else {
        self$post_prompt_wraps <- c(self$post_prompt_wraps, list(prompt_wrap))
      }
      invisible(self)
    },

    #' @description
    #' Apply all provider-level wraps to a prompt (character or tidyprompt)
    #' and return a tidyprompt with wraps attached.
    #' This is typically called inside `send_prompt()` before evaluation of
    #' the prompt.
    #' @param prompt A string, a chat history, a list containing
    #' a chat history under key '$chat_history', or a [tidyprompt-class] object
    apply_prompt_wraps = function(prompt) {
      if (!inherits(prompt, "Tidyprompt")) {
        prompt <- tidyprompt(prompt)
      }

      # Fresh Tidyprompt with same base/system/chat_history but no wraps yet
      new_prompt <- `tidyprompt-class`$new(prompt$base_prompt)
      new_prompt$system_prompt <- prompt$system_prompt
      # Preserve chat history without constructing prompt text
      new_prompt$.__enclos_env__$private$chat_history <-
        prompt$.__enclos_env__$private$chat_history

      # 1) Provider pre wraps
      if (length(self$pre_prompt_wraps)) {
        for (pw in self$pre_prompt_wraps) {
          new_prompt <- prompt_wrap_internal(
            new_prompt,
            modify_fn = pw$modify_fn,
            extraction_fn = pw$extraction_fn,
            validation_fn = pw$validation_fn,
            handler_fn = pw$handler_fn,
            parameter_fn = pw$parameter_fn,
            type = pw$type,
            name = pw$name
          )
        }
      }

      # 2) Existing prompt wraps in their original order
      if (length(prompt$get_prompt_wraps(order = "default"))) {
        for (pw in prompt$get_prompt_wraps(order = "default")) {
          new_prompt <- prompt_wrap_internal(
            new_prompt,
            modify_fn = pw$modify_fn,
            extraction_fn = pw$extraction_fn,
            validation_fn = pw$validation_fn,
            handler_fn = pw$handler_fn,
            parameter_fn = pw$parameter_fn,
            type = pw$type,
            name = pw$name
          )
        }
      }

      # 3) Provider post wraps
      if (length(self$post_prompt_wraps)) {
        for (pw in self$post_prompt_wraps) {
          new_prompt <- prompt_wrap_internal(
            new_prompt,
            modify_fn = pw$modify_fn,
            extraction_fn = pw$extraction_fn,
            validation_fn = pw$validation_fn,
            handler_fn = pw$handler_fn,
            parameter_fn = pw$parameter_fn,
            type = pw$type,
            name = pw$name
          )
        }
      }

      new_prompt
    }
  ),
  private = list(
    complete_chat_function = NULL
  )
)

#' Generate an example object from a JSON schema
#'
#' This function generates an example JSON object from a JSON schema.
#' This is used when enforcing a JSON schema through text-based handling
#' (requiring an example to be added to the prompt text).
#'
#' @param schema A list (R object) representing a JSON schema
#'
#' @return A list (R object) which matches the JSON schema definition
#'
#' @export
#'
#' @example inst/examples/answer_as_json.R
#' @family json
r_json_schema_to_example <- function(schema) {
  if ("schema" %in% names(schema)) {
    schema <- schema$schema
  }

  if (is.null(schema$type)) {
    return(NULL)
  }
  switch(
    schema$type,
    object = {
      # Generate an object with properties
      result <- list()
      if (!is.null(schema$properties)) {
        for (name in names(schema$properties)) {
          result[[name]] <- r_json_schema_to_example(schema$properties[[name]])
        }
      }
      return(result)
    },
    array = {
      # Generate an array with one example item
      example_item <- r_json_schema_to_example(schema$items)
      return(list(example_item))
    },
    string = {
      if (!is.null(schema$enum)) {
        return(schema$enum) # Include all possible values
      }
      return(ifelse(!is.null(schema$example), schema$example, "..."))
    },
    number = {
      return(ifelse(!is.null(schema$example), schema$example, 123.45))
    },
    integer = {
      return(ifelse(!is.null(schema$example), schema$example, 123))
    },
    boolean = {
      return(ifelse(!is.null(schema$example), schema$example, TRUE))
    },
    NULL
  ) # For unsupported or missing types, return NULL
}

#' @title Set ReAct mode for a prompt
#'
#' @description This function enables ReAct mode for the evaluation of a prompt
#' or a [tidyprompt()]. In ReAct mode, the large language model (LLM) is asked to think step by step,
#' each time detailing a thought, action, and observation, to eventually arrive at
#' a final answer. It is hypothesized that this may increase LLM performance
#' at solving complex tasks. ReAct mode is inspired by the method described in
#' Yao et al. (2022).
#'
#' @details Please note that ReAct mode may be most useful if in combination
#' with tools that the LLM can use. See, for example, 'add_tools()' for
#' enabling R function calling, or, for example, 'answer_as_code()' with
#' 'output_as_tool = TRUE' for enabling R code evaluation as a tool.
#'
#' @param prompt A single string or a [tidyprompt()] object
#' @param extract_from_finish_brackets A logical indicating whether the final answer
#' should be extracted from the text inside the "FINISH[...]" brackets
#' @param extraction_lenience A logical indcating whether the extraction function should be lenient.
#' If TRUE, the extraction function will attempt to extract the final answer
#' even if it cannot be extracted from within the brackets, by extracting
#' everything after the final occurence of 'FINISH' (if present). This may
#' be useful for smaller LLMs which may not follow the output format
#' as strictly
#'
#' @return A [tidyprompt()] with an added [prompt_wrap()] which will ensure
#' that the LLM follows the ReAct mode in answering the prompt
#'
#' @export
#'
#' @example inst/examples/answer_by_react.R
#'
#' @references
#' Yao, S., Wu, Y., Cheung, W., Wang, Z., Narasimhan, K., & Kong, L. (2022).
#' ReAct: Synergizing Reasoning and Acting in Language Models.
#' <doi:10.48550/arXiv.2210.03629>
#'
#' @seealso [answer_using_tools()] [answer_using_r()]
#'
#' @family pre_built_prompt_wraps
#' @family answer_by_prompt_wraps
answer_by_react <- function(
  prompt,
  extract_from_finish_brackets = TRUE,
  extraction_lenience = TRUE
) {
  # Define modification/extraction/validation functions:
  modify_fn <- function(original_prompt_text) {
    new_prompt <- glue::glue(
      "You are given a user's prompt.
      To answer the user's prompt, you need to think step by step, take an action if needed, and then return a final answer.

      ----- START OF USER'S PROMPT -----
      {original_prompt_text}
      ----- END OF USER'S PROMPT -----

      Use the following structure:
        Thought: <describe your thought process>
        Action: <if needed, describe the action you take (e.g., look up information)>
        Observation: <describe the result or observation from the action>
      (Repeat Thought -> Action -> Observation as necessary)

      When you are done, you must type:
        FINISH[<put here your final answer to the user's prompt>]

      Ensure your final answer aligns with your reasoning and observations."
    )

    return(new_prompt)
  }

  extraction_fn <- function(llm_response) {
    if (!extract_from_finish_brackets) {
      return(llm_response)
    }
    extraction_fn_finish(llm_response, extraction_lenience)
  }

  prompt_wrap(
    prompt,
    modify_fn,
    extraction_fn,
    type = "mode",
    name = "answer_by_react"
  )
}

#' Extraction function to extract final response from 'FINISH[...]'
#'
#' This is a helper function which assists in extracting the final response from
#' a large language model after it has been instructed to give this within
#' 'FINISH[...]' (as is done by [answer_by_chain_of_thought()] and
#' [answer_by_react()]). If something cannot be extracted from within the brackets,
#' it falls back on attempting to extract everything after the final occurence
#' of a 'FINISH' variant.
#'
#' @param llm_response Response of a large language model
#' @param lenience Whether lenience should be applied when extracting;
#' when TRUE, extraction fallback method will be to extract anything after
#' the last occurence of a variation on 'FINISH'
#'
#' @return Extracted final answer
#'
#' @noRd
#' @keywords internal
extraction_fn_finish <- function(llm_response, lenience = TRUE) {
  error_feedback <- llm_feedback(
    glue::glue(
      "Error, could not parse your final answer.\n",
      "Please type: 'FINISH[<put here your final answer to the original prompt>]'"
    )
  )

  # Check if FINISH is mentioned in the response.
  if (!stringr::str_detect(llm_response, "(?si)FINISH")) {
    return(error_feedback)
  }

  # First attempt: Extract text between FINISH[...]
  extracted_response <- stringr::str_extract(
    llm_response,
    "(?si)(?<=FINISH\\[).+?(?=\\])"
  )

  # If extraction fails, try alternative method
  if (is.na(extracted_response) && lenience) {
    # Use regex to match 'FINISH' variants and capture text between brackets/braces or spaces
    # start with capturing everything after finish:
    extracted_response <- stringr::str_extract(
      llm_response,
      "(?si)(?<=\\bFINISH).*"
    )

    # remove whitespaces
    extracted_response <- stringr::str_trim(extracted_response)

    # if text starts with punctuation ':, =, ...' and optional space remove it
    extracted_response <- stringr::str_replace(
      extracted_response,
      "^[[:punct:]]+\\s*",
      ""
    )

    # if text begins and ends with any form of braces '[{(' remove them
    has_start_brace <- stringr::str_detect(extracted_response, "^[\\[\\{\\(]")
    has_end_brace <- stringr::str_detect(extracted_response, "[\\]\\}\\)]$")

    if (has_start_brace && has_end_brace) {
      # remove braces
      extracted_response <- stringr::str_replace_all(
        extracted_response,
        "(^[\\[\\{\\(])|([\\]\\}\\)]$)",
        ""
      )
    }
  }

  # Final check for successful extraction
  if (
    is.na(extracted_response) ||
      tolower(trimws(extracted_response)) %in% c("answer", "final answer") ||
      extracted_response |> stringr::str_trim() == ""
  ) {
    return(error_feedback)
  }

  return(extracted_response)
}

#' Extract and parse JSONs from a string (LLM response)
#'
#' This function extracts JSON blocks from a string and parses them using `jsonlite::fromJSON()`.
#' This can be used to extract all JSONs from LLM responses, immediately converting
#' them to R objects.
#'
#' @param llm_response A character string
#'
#' @return A list of parsed JSON objects
#'
#' @family text_helpers
#'
#' @noRd
#' @keywords internal
extraction_fn_json <- function(llm_response) {
  positions <- gregexpr("[{}]", llm_response)[[1]]
  matches <- regmatches(llm_response, gregexpr("[{}]", llm_response))[[1]]

  stack <- c()
  blocks <- list()
  block_start <- c()

  for (i in seq_along(positions)) {
    char <- matches[i]
    pos <- positions[i]

    if (is.na(char)) {
      next
    }

    if (char == "{") {
      if (length(stack) == 0) {
        # Start of a new block
        block_start <- c(block_start, pos)
      }
      stack <- c(stack, char)
    } else if (char == "}") {
      if (length(stack) == 0) {
        # Unmatched closing brace
        next
      }
      stack <- stack[-length(stack)]
      if (length(stack) == 0) {
        # End of a block
        start_pos <- block_start[1]
        end_pos <- pos
        block_text <- substr(llm_response, start_pos, end_pos)
        blocks <- c(blocks, block_text)
        block_start <- block_start[-1]
      }
    }
  }

  # Now parse the blocks
  parsed_jsons <- lapply(blocks, function(json_candidate) {
    tryCatch(
      {
        jsonlite::fromJSON(json_candidate, simplifyDataFrame = FALSE)
      },
      error = function(e) NULL
    )
  })

  # Remove NULL entries
  parsed_jsons <- parsed_jsons[!sapply(parsed_jsons, is.null)]
  return(parsed_jsons)
}

#' Make LLM answer as a data frame via structured output
#'
#' @description
#' This function builds on [answer_as_json()] to extract a data frame from an
#' LLM response using structured output. The supplied `schema` should describe a
#' single row of the desired data frame, or an array of such rows. Internally,
#' `answer_as_dataframe()` standardizes the schema to a JSON object with a
#' `rows` field containing an array of row objects. This shape works well with
#' both text-based JSON extraction and native structured-output backends,
#' including 'ellmer', where arrays of objects are converted to data frames.
#'
#' @details
#' Prefer supplying an 'ellmer' row schema created with
#' `ellmer::type_object(...)` when possible. This is usually the clearest way to
#' describe the columns you want, and it maps cleanly to native 'ellmer'
#' structured output. These 'ellmer' schema definitions can also be used with
#' non-'ellmer' LLM providers, because 'tidyprompt' converts between 'ellmer'
#' schema definitions and JSON-schema representations as needed.
#'
#' `answer_as_dataframe()` accepts the following schema shapes:
#' * A single row schema, such as `ellmer::type_object(...)` or a JSON schema
#'   object whose properties describe the columns of one row.
#' * An array-of-rows schema, such as `ellmer::type_array(row_schema)` or a JSON
#'   schema with `type = "array"` and row objects under `items`.
#' * A wrapper object whose **only** property is a `rows` field containing an
#'   array of row objects (matching the shape produced internally by
#'   `answer_as_dataframe()`). Schemas with additional sibling properties
#'   alongside `rows` are treated as row schemas, not wrappers.
#'
#' Regardless of which of these forms you supply, `answer_as_dataframe()`
#' normalizes it to a row-oriented structured-output schema before delegating to
#' [answer_as_json()].
#'
#' @param prompt A single string or a [tidyprompt()] object
#' @param schema A JSON schema list or an 'ellmer' type definition describing a
#' single row, an array of rows, or a wrapper object containing a `rows` array.
#' @param min_rows (optional) Minimum number of rows required in the returned
#' data frame
#' @param max_rows (optional) Maximum number of rows allowed in the returned
#' data frame
#' @param schema_strict If TRUE, the wrapped schema will be strictly enforced.
#' Passed through to [answer_as_json()]
#' @param schema_in_prompt_as Passed through to [answer_as_json()] when using a
#' text-based JSON path
#' @param type Passed through to [answer_as_json()] to control the structured
#' output backend
#'
#' @return A [tidyprompt()] with an added [prompt_wrap()] which will ensure
#' that the LLM response is returned as a data frame.
#'
#' @export
#'
#' @example inst/examples/answer_as_dataframe.R
#'
#' @family pre_built_prompt_wraps
#' @family answer_as_prompt_wraps
answer_as_dataframe <- function(
  prompt,
  schema,
  min_rows = NULL,
  max_rows = NULL,
  schema_strict = FALSE,
  schema_in_prompt_as = c("example", "schema"),
  type = c(
    "auto",
    "text-based",
    "openai",
    "ollama",
    "openai_oo",
    "ollama_oo",
    "ellmer"
  )
) {
  schema_in_prompt_as <- match.arg(schema_in_prompt_as)
  type <- match.arg(type)

  row_schema <- answer_as_dataframe_row_schema(schema, strict = schema_strict)
  wrapped_schema <- answer_as_dataframe_wrapper_schema(row_schema)
  json_wrap <- answer_as_dataframe_json_wrap(
    wrapped_schema = wrapped_schema,
    schema_strict = schema_strict,
    schema_in_prompt_as = schema_in_prompt_as,
    type = type
  )

  prompt_wrap(
    prompt,
    modify_fn = json_wrap$modify_fn,
    extraction_fn = function(x, llm_provider, http_list) {
      structured <- json_wrap$extraction_fn(x, llm_provider, http_list)

      if (
        inherits(structured, "llm_feedback") ||
          inherits(structured, "llm_feedback_tool_result") ||
          inherits(structured, "llm_break") ||
          inherits(structured, "llm_break_soft")
      ) {
        return(structured)
      }

      answer_as_dataframe_extract(
        structured,
        row_schema = row_schema,
        min_rows = min_rows,
        max_rows = max_rows
      )
    },
    validation_fn = json_wrap$validation_fn,
    handler_fn = json_wrap$handler_fn,
    parameter_fn = json_wrap$parameter_fn,
    type = json_wrap$type,
    name = "answer_as_dataframe"
  )
}

answer_as_dataframe_json_wrap <- function(
  wrapped_schema,
  schema_strict,
  schema_in_prompt_as,
  type
) {
  json_prompt <- answer_as_json(
    tidyprompt("dataframe"),
    schema = wrapped_schema,
    schema_strict = schema_strict,
    schema_in_prompt_as = schema_in_prompt_as,
    type = type
  )

  wraps <- json_prompt$get_prompt_wraps(order = "default")
  wraps[[length(wraps)]]
}

answer_as_dataframe_row_schema <- function(schema, strict = FALSE) {
  normalized <- normalize_schema_dual(schema, strict = strict)
  json_schema <- normalized$json_schema

  if (is.null(json_schema)) {
    stop(
      "The 'schema' must be convertible to a JSON schema for ",
      "`answer_as_dataframe()`."
    )
  }

  json_schema <- unwrap_json_schema(json_schema)

  # Detect a wrapper object whose `rows` field is an array of row objects.

  # We require that rows.items looks like a row schema (has properties)
  # AND that `rows` is the ONLY property -- a genuine wrapper produced by

  # answer_as_dataframe_wrapper_schema() never has sibling columns.  When

  # a row schema itself has a column named "rows" (e.g. array<object>),
  # sibling columns will be present and we leave the schema alone.
  if (
    identical(json_schema$type %||% NULL, "object") &&
      is.list(json_schema$properties) &&
      length(json_schema$properties) == 1L &&
      "rows" %in% names(json_schema$properties) &&
      identical(json_schema$properties$rows$type %||% NULL, "array") &&
      is.list(json_schema$properties$rows$items) &&
      (identical(json_schema$properties$rows$items$type %||% NULL, "object") ||
        is.list(json_schema$properties$rows$items$properties))
  ) {
    json_schema <- json_schema$properties$rows$items
  }

  if (identical(json_schema$type %||% NULL, "array")) {
    json_schema <- json_schema$items
  }

  if (
    !identical(json_schema$type %||% NULL, "object") &&
      is.null(json_schema$properties)
  ) {
    stop(
      "The 'schema' for `answer_as_dataframe()` must describe row objects, ",
      "an array of row objects, or an object with a `rows` array."
    )
  }

  json_schema
}

answer_as_dataframe_wrapper_schema <- function(row_schema) {
  list(
    type = "object",
    properties = list(
      rows = list(
        type = "array",
        description = "Rows of the data frame.",
        items = row_schema
      )
    ),
    required = list("rows"),
    additionalProperties = FALSE
  )
}

answer_as_dataframe_extract <- function(
  x,
  row_schema,
  min_rows = NULL,
  max_rows = NULL
) {
  rows <- answer_as_dataframe_rows_from_response(x)

  if (is.null(rows)) {
    return(llm_feedback(
      "You must respond with a JSON object containing a `rows` array of objects."
    ))
  }

  df <- answer_as_dataframe_to_df(rows, row_schema = row_schema)

  if (is.null(df)) {
    return(llm_feedback(
      "The `rows` field must contain an array of objects that can be converted to a data frame."
    ))
  }

  observed_names <- names(df)
  required_cols <- row_schema$required %||% character()
  missing_required <- setdiff(required_cols, observed_names)
  if (length(missing_required) > 0) {
    return(llm_feedback(
      paste0(
        "The data frame is missing required columns: ",
        paste(missing_required, collapse = ", "),
        "."
      )
    ))
  }

  df <- answer_as_dataframe_complete_columns(df, row_schema)

  if (!is.null(min_rows) && nrow(df) < min_rows) {
    return(llm_feedback(
      glue::glue("The data frame should contain at least {min_rows} rows.")
    ))
  }
  if (!is.null(max_rows) && nrow(df) > max_rows) {
    return(llm_feedback(
      glue::glue("The data frame should contain at most {max_rows} rows.")
    ))
  }

  df
}

answer_as_dataframe_rows_from_response <- function(x) {
  if (is.data.frame(x)) {
    return(x)
  }

  if (is.list(x) && "rows" %in% names(x)) {
    return(x$rows)
  }

  if (is.list(x) && length(x) > 0 && all(vapply(x, is.list, logical(1)))) {
    return(x)
  }

  NULL
}

answer_as_dataframe_to_df <- function(rows, row_schema) {
  if (is.data.frame(rows)) {
    return(rows)
  }

  if (is.list(rows) && length(rows) == 0) {
    return(answer_as_dataframe_empty_df(row_schema))
  }

  if (!is.list(rows) || !all(vapply(rows, is.list, logical(1)))) {
    return(NULL)
  }

  tryCatch(
    dplyr::bind_rows(rows),
    error = function(e) NULL
  )
}

answer_as_dataframe_complete_columns <- function(df, row_schema) {
  expected_cols <- names(row_schema$properties %||% list())
  if (length(expected_cols) == 0) {
    return(df)
  }

  missing_cols <- setdiff(expected_cols, names(df))
  for (col in missing_cols) {
    df[[col]] <- answer_as_dataframe_missing_column(
      row_schema$properties[[col]],
      nrow(df)
    )
  }

  ordered <- c(expected_cols, setdiff(names(df), expected_cols))
  df[, ordered, drop = FALSE]
}

answer_as_dataframe_empty_df <- function(row_schema) {
  props <- row_schema$properties %||% list()
  out <- lapply(props, answer_as_dataframe_missing_column, n_rows = 0)

  if (length(out) == 0) {
    return(data.frame())
  }

  out <- out[names(props)]
  data.frame(out, check.names = FALSE)
}

answer_as_dataframe_missing_column <- function(column_schema, n_rows) {
  column_type <- column_schema$type %||% NULL
  if (!is.null(column_schema$enum)) {
    column_type <- "string"
  }

  if (identical(column_type, "string")) {
    return(rep(NA_character_, n_rows))
  }
  if (identical(column_type, "number")) {
    return(rep(NA_real_, n_rows))
  }
  if (identical(column_type, "integer")) {
    return(rep(NA_integer_, n_rows))
  }
  if (identical(column_type, "boolean")) {
    return(rep(NA, n_rows))
  }

  vector("list", n_rows)
}

test_that("answer_as_dataframe extracts a data frame from structured JSON", {
  provider <- `llm_provider-class`$new(
    complete_chat_function = function(chat_history) {
      reply <- jsonlite::toJSON(
        list(
          rows = list(
            list(name = "Alice", age = 32L),
            list(name = "Bob", age = 28L)
          )
        ),
        auto_unbox = TRUE
      )
      reply <- as.character(reply)

      list(
        completed = dplyr::bind_rows(
          chat_history,
          data.frame(
            role = "assistant",
            content = reply,
            stringsAsFactors = FALSE
          )
        ),
        http = list(request = NULL, response = NULL)
      )
    },
    verbose = FALSE
  )

  schema <- list(
    type = "object",
    properties = list(
      name = list(type = "string"),
      age = list(type = "integer")
    ),
    required = c("name", "age"),
    additionalProperties = FALSE
  )

  result <- "Extract the people in the text." |>
    answer_as_dataframe(schema, type = "text-based") |>
    send_prompt(provider, verbose = FALSE)

  expect_s3_class(result, "data.frame")
  expect_equal(names(result), c("name", "age"))
  expect_equal(nrow(result), 2)
  expect_equal(result$name, c("Alice", "Bob"))
  expect_equal(result$age, c(32L, 28L))
})

test_that("answer_as_dataframe accepts array-of-rows schemas", {
  provider <- `llm_provider-class`$new(
    complete_chat_function = function(chat_history) {
      reply <- jsonlite::toJSON(
        list(
          rows = list(
            list(name = "Alice", age = 32L),
            list(name = "Bob", age = 28L)
          )
        ),
        auto_unbox = TRUE
      )
      reply <- as.character(reply)

      list(
        completed = dplyr::bind_rows(
          chat_history,
          data.frame(
            role = "assistant",
            content = reply,
            stringsAsFactors = FALSE
          )
        ),
        http = list(request = NULL, response = NULL)
      )
    },
    verbose = FALSE
  )

  schema <- list(
    type = "array",
    items = list(
      type = "object",
      properties = list(
        name = list(type = "string"),
        age = list(type = "integer")
      ),
      required = c("name", "age"),
      additionalProperties = FALSE
    )
  )

  result <- "Extract the people in the text." |>
    answer_as_dataframe(schema, type = "text-based") |>
    send_prompt(provider, verbose = FALSE)

  expect_s3_class(result, "data.frame")
  expect_equal(nrow(result), 2)
  expect_equal(result$age, c(32L, 28L))
})

test_that("answer_as_dataframe accepts wrapper-object schema (rows-only)", {
  provider <- `llm_provider-class`$new(
    complete_chat_function = function(chat_history) {
      reply <- jsonlite::toJSON(
        list(
          rows = list(
            list(name = "Alice", age = 32L),
            list(name = "Bob", age = 28L)
          )
        ),
        auto_unbox = TRUE
      )
      reply <- as.character(reply)

      list(
        completed = dplyr::bind_rows(
          chat_history,
          data.frame(
            role = "assistant",
            content = reply,
            stringsAsFactors = FALSE
          )
        ),
        http = list(request = NULL, response = NULL)
      )
    },
    verbose = FALSE
  )

  # Wrapper schema with rows as the sole property
  schema <- list(
    type = "object",
    properties = list(
      rows = list(
        type = "array",
        items = list(
          type = "object",
          properties = list(
            name = list(type = "string"),
            age = list(type = "integer")
          ),
          required = c("name", "age"),
          additionalProperties = FALSE
        )
      )
    ),
    required = "rows"
  )

  result <- "Extract the people." |>
    answer_as_dataframe(schema, type = "text-based") |>
    send_prompt(provider, verbose = FALSE)

  expect_s3_class(result, "data.frame")
  expect_equal(nrow(result), 2)
  expect_equal(names(result), c("name", "age"))
  expect_equal(result$name, c("Alice", "Bob"))
})

test_that("answer_as_dataframe treats row schema with a 'rows' column correctly", {
  provider <- `llm_provider-class`$new(
    complete_chat_function = function(chat_history) {
      reply <- jsonlite::toJSON(
        list(
          rows = list(
            list(
              id = 1L,
              rows = 10L
            ),
            list(
              id = 2L,
              rows = 20L
            )
          )
        ),
        auto_unbox = TRUE
      )
      reply <- as.character(reply)

      list(
        completed = dplyr::bind_rows(
          chat_history,
          data.frame(
            role = "assistant",
            content = reply,
            stringsAsFactors = FALSE
          )
        ),
        http = list(request = NULL, response = NULL)
      )
    },
    verbose = FALSE
  )

  # Row schema where "rows" is a regular integer column, not a wrapper
  schema <- list(
    type = "object",
    properties = list(
      id = list(type = "integer"),
      rows = list(type = "integer")
    ),
    required = c("id", "rows")
  )

  result <- "Extract data." |>
    answer_as_dataframe(schema, type = "text-based") |>
    send_prompt(provider, verbose = FALSE)

  expect_s3_class(result, "data.frame")
  expect_equal(nrow(result), 2)
  expect_true("id" %in% names(result))
  expect_true("rows" %in% names(result))
  expect_equal(result$id, c(1L, 2L))
  expect_equal(result$rows, c(10L, 20L))
})

test_that("answer_as_dataframe retries when row bounds are violated", {
  state <- new.env(parent = emptyenv())
  state$call_n <- 0L

  provider <- `llm_provider-class`$new(
    complete_chat_function = function(chat_history) {
      state <- self$parameters$.test_state
      state$call_n <- state$call_n + 1L

      rows <- if (state$call_n == 1L) {
        list(
          list(name = "Alice", age = 32L),
          list(name = "Bob", age = 28L),
          list(name = "Cara", age = 41L)
        )
      } else {
        list(
          list(name = "Alice", age = 32L),
          list(name = "Bob", age = 28L)
        )
      }

      reply <- jsonlite::toJSON(list(rows = rows), auto_unbox = TRUE)
      reply <- as.character(reply)

      list(
        completed = dplyr::bind_rows(
          chat_history,
          data.frame(
            role = "assistant",
            content = reply,
            stringsAsFactors = FALSE
          )
        ),
        http = list(request = NULL, response = NULL)
      )
    },
    parameters = list(.test_state = state),
    verbose = FALSE
  )

  schema <- list(
    type = "object",
    properties = list(
      name = list(type = "string"),
      age = list(type = "integer")
    ),
    required = c("name", "age"),
    additionalProperties = FALSE
  )

  result <- "Extract exactly two people." |>
    answer_as_dataframe(
      schema,
      min_rows = 2,
      max_rows = 2,
      type = "text-based"
    ) |>
    send_prompt(provider, verbose = FALSE)

  expect_s3_class(result, "data.frame")
  expect_equal(nrow(result), 2)
  expect_equal(state$call_n, 2L)
})

test_that("answer_as_dataframe uses native ellmer structured results", {
  skip_if_not_installed("ellmer")

  provider <- `llm_provider-class`$new(
    complete_chat_function = function(chat_history) {
      list(
        completed = dplyr::bind_rows(
          chat_history,
          data.frame(
            role = "assistant",
            content = "Structured result",
            stringsAsFactors = FALSE
          )
        ),
        http = list(request = NULL, response = NULL),
        native_structured_result = list(
          rows = data.frame(
            name = c("Alice", "Bob"),
            age = c(32, 28),
            stringsAsFactors = FALSE
          )
        )
      )
    },
    verbose = FALSE,
    api_type = "ellmer"
  )

  schema <- ellmer::type_object(
    name = ellmer::type_string(),
    age = ellmer::type_integer()
  )

  result <- "Extract the people in the text." |>
    answer_as_dataframe(schema) |>
    send_prompt(provider, verbose = FALSE)

  expect_s3_class(result, "data.frame")
  expect_equal(result$name, c("Alice", "Bob"))
  expect_equal(result$age, c(32, 28))
})

test_that("answer_as_dataframe_row_schema preserves row schema with array<object> 'rows' column", {
  # Row schema where "rows" is an array-of-objects column, not a wrapper
  schema <- list(
    type = "object",
    properties = list(
      id = list(type = "integer"),
      rows = list(
        type = "array",
        items = list(
          type = "object",
          properties = list(label = list(type = "string"))
        )
      )
    ),
    required = c("id", "rows")
  )

  row_schema <- answer_as_dataframe_row_schema(schema)

  # The heuristic must NOT unwrap; the row schema should keep both columns
  expect_equal(row_schema$type, "object")
  expect_true("id" %in% names(row_schema$properties))
  expect_true("rows" %in% names(row_schema$properties))
  expect_equal(row_schema$properties$id$type, "integer")
  expect_equal(row_schema$properties$rows$type, "array")
})

test_that("answer_as_dataframe_wrapper_schema emits required as array", {
  row_schema <- list(
    type = "object",
    properties = list(name = list(type = "string"))
  )
  wrapper <- answer_as_dataframe_wrapper_schema(row_schema)

  expect_equal(wrapper$required, list("rows"))
  # When serialized with auto_unbox, must produce a JSON array
  json <- jsonlite::toJSON(wrapper, auto_unbox = TRUE)
  parsed <- jsonlite::fromJSON(json, simplifyVector = FALSE)
  expect_true(is.list(parsed$required))
  expect_equal(parsed$required[[1]], "rows")
})

#' Skim a dataframe and include labels and levels
#'
#' This function takes a `data.frame` and returns a skim summary with variable names,
#' labels, and levels for categorical variables. It is a wrapper around the [skimr::skim()]
#' function.
#'
#' @param data A `data.frame` to be skimmed
#'
#' @return A `data.frame` with variable names, labels, levels, and a skim summary
#'
#' @export
#'
#' @example inst/examples/skim_with_labels_and_levels.R
#'
#' @family text_helpers
skim_with_labels_and_levels <- function(data) {
  if (!requireNamespace("skimr", quietly = TRUE)) {
    stop("The 'skimr' package is required to use skim_with_labels_and_levels()")
  }
  stopifnot(is.data.frame(data))

  # Extract labels and levels from data
  labels <- sapply(
    data,
    function(x) ifelse(is.null(attr(x, "label")), NA, attr(x, "label"))
  )
  levels_list <- sapply(data, function(x) if (is.factor(x)) levels(x) else NA)

  # Create a data frame with variable names, labels, and levels
  labels_df <- data.frame(
    variable = names(labels),
    description = labels,
    levels = I(levels_list), # Use I() to keep lists intact
    stringsAsFactors = FALSE
  )

  # Get skim summary
  skimmed <- skimr::skim(data)

  # Merge labels with skimmed output
  skim_with_labels <- merge(
    labels_df,
    skimmed,
    by.x = "variable",
    by.y = "skim_variable",
    all.y = TRUE
  )

  # Remove columns that are completely NA
  skim_with_labels <- skim_with_labels[,
    colSums(is.na(skim_with_labels)) < nrow(skim_with_labels)
  ]

  # Return the skimmed summary with labels and levels
  return(skim_with_labels)
}

## ----chunk_options, include = FALSE, cache=TRUE-------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## ----setup, cache=TRUE--------------------------------------------------------
library(tidyprompt)

## ----llm_provider, cache=TRUE-------------------------------------------------
# Ollama running on local PC
ollama <- create_ollama_llm_provider(
  parameters = list(
    model = "llama3.1:8b",
    url = "http://localhost:11434/api/chat"
  )
)

# OpenAI API
openai <- create_openai_llm_provider(
  parameters = list(
    model = "gpt-4o-mini",
    api_key = Sys.getenv("OPENAI_API_KEY")
  )
)

# Create your own LLM provider hook using create_llm_provider();
#   see ?create_llm_provider for more information, and take a look at
#   the source code of create_ollama_llm_provider() and create_openai_llm_provider()

#' Make LLM answer as a category
#'
#' @details
#' For multiple categories, see [answer_as_multi_category()].
#'
#' @param prompt A single string or a [tidyprompt()] object
#' @param categories A character vector of category names. Must not be empty
#'   and must not contain duplicates
#' @param descriptions An optional character vector of descriptions,
#'   corresponding to each category. If provided, its length must match
#'   the length of `categories`. Defaults to `NULL`
#'
#' @return A [tidyprompt()] with an added [prompt_wrap()] which
#' will ensure that the LLM response is the most fitting category of a text as
#' a character vector of length one
#'
#' @export
#'
#' @example inst/examples/answer_as_category.R
#'
#' @family pre_built_prompt_wraps
#' @family answer_as_prompt_wraps
answer_as_category <- function(
  prompt,
  categories,
  descriptions = NULL
) {
  prompt <- tidyprompt(prompt)

  numbered_categories_text <- .build_formatted_category_list(
    categories,
    descriptions
  )

  instruction <- paste0(
    numbered_categories_text,
    "\n\n",
    "Respond with the number of the category that best describes the text.",
    "\n",
    "(Use no other words or characters.)"
  )

  modify_fn <- function(original_prompt_text) {
    paste0(
      "You need to categorize the following text.\n\n",
      "Text:\n  '",
      original_prompt_text,
      "'\n\n",
      instruction
    )
  }

  extraction_fn <- function(response) {
    normalized <- trimws(tolower(response))
    if (normalized %in% as.character(seq_along(categories))) {
      return(categories[[as.integer(normalized)]])
    }
    match <- which(tolower(categories) == normalized)
    if (length(match) == 1) {
      return(categories[[match]])
    }
    return(tidyprompt::llm_feedback(instruction))
  }

  prompt_wrap(
    prompt,
    modify_fn,
    extraction_fn,
    name = "answer_as_category"
  )
}


#' Build prompt for categorizing a text into multiple categories
#'
#' @details
#' For a single category, see [answer_as_category()].
#'
#' @param prompt A single string or a [tidyprompt()] object
#' @param categories A character vector of category names. Must not be empty
#'   and must not contain duplicates
#' @param descriptions An optional character vector of descriptions,
#'   corresponding to each category. If provided, its length must match
#'   the length of `categories`. Defaults to `NULL`
#'
#' @return A [tidyprompt()] with an added [prompt_wrap()] which
#' will ensure that the LLM response is a vector of fitting categories of a text
#'
#' @export
#'
#' @example inst/examples/answer_as_multi_category.R
#'
#' @family pre_built_prompt_wraps
#' @family answer_as_prompt_wraps
answer_as_multi_category <- function(
  prompt,
  categories,
  descriptions = NULL
) {
  prompt <- tidyprompt(prompt)

  numbered_categories_text <- .build_formatted_category_list(
    categories,
    descriptions
  )

  instruction <- paste0(
    numbered_categories_text,
    "\n\n",
    "Respond with the numbers of all categories that apply to this text, separated by commas.",
    "\n",
    "(Use no other words or characters.)"
  )

  modify_fn <- function(original_prompt_text) {
    paste0(
      "You need to categorize the following text.\n\n",
      "Text:\n  '",
      original_prompt_text,
      "'\n\n",
      instruction
    )
  }

  extraction_fn <- function(response) {
    normalized <- trimws(tolower(response))
    numbers <- unlist(strsplit(normalized, ",\\s*"))
    valid_numbers <- numbers[
      numbers %in% as.character(seq_along(categories))
    ]
    if (length(valid_numbers) == 0) {
      return(tidyprompt::llm_feedback(
        "You must select at least one valid category number."
      ))
    }
    categories_selected <- categories[as.integer(valid_numbers)]

    return(categories_selected)
  }

  prompt_wrap(
    prompt,
    modify_fn,
    extraction_fn,
    name = "answer_as_multi_category"
  )
}

#' Build formatted categories text
#'
#' This helper function generates a text block listing categories,
#' optionally with their descriptions.
#' It also provides an appropriate introductory phrase.
#'
#' @param categories A character vector of category names. Must not be empty
#'   and must not contain duplicates
#' @param descriptions An optional character vector of descriptions,
#'   corresponding to each category. If provided, its length must match
#'   the length of `categories`. Defaults to `NULL`
#'
#' @return A string containing the formatted list of categories
#'
#'
#' @noRd
#'
#' @keywords internal
.build_formatted_category_list <- function(
  categories,
  descriptions = NULL
) {
  stopifnot(
    "'categories' must be a character vector." = is.character(categories),
    "'categories' must not be empty." = length(categories) > 0,
    "'categories' must not contain duplicates." = !anyDuplicated(categories) > 0
  )

  if (!is.null(descriptions)) {
    length_equal <- length(descriptions) == length(categories)
    stopifnot(
      "'descriptions' must be a character vector." = is.character(descriptions),
      "'descriptions' must have the same length as 'categories'." = length_equal
    )
  }

  if (is.null(descriptions)) {
    category_entries <- sprintf(
      "%d. %s",
      seq_along(categories),
      categories
    )
    instruction_intro <- "Possible categories:"
  } else {
    category_entries <- sprintf(
      "%d. %s: %s",
      seq_along(categories),
      categories,
      descriptions
    )
    instruction_intro <- "Possible categories and their descriptions:"
  }

  numbered_categories_text <- paste0(category_entries, collapse = "\n  ")

  formatted_category_list <- paste0(
    instruction_intro,
    "\n  ",
    numbered_categories_text
  )

  return(formatted_category_list)
}

#' @title
#' Have LLM check the result of a prompt (LLM-in-the-loop)
#'
#' @description
#' This function will wrap a prompt with a check for a LLM to accept or decline
#' the result of the prompt, providing feedback if the result is declined.
#' The evaluating LLM will be presented with the original prompt and the result
#' of the prompt, and will be asked to verify if the answer is satisfactory
#' (using chain of thought reasoning to arrive at a boolean decision). If
#' the result is declined, the chain of thought responsible for the decision
#' will be summarized and sent back to the original LLM that was asked to
#' evaluate the prompt, so that it may retry the prompt.
#'
#' Note that this function is experimental and, because it relies on chain of
#' thought reasoning by a LLM about the answer of another LLM, it may not
#' always provide accurate results and can increase the token cost of
#' evaluating a prompt.
#'
#' @details The original prompt text shown to the LLM is built from
#' the base prompt as well as all prompt wraps that have a modify function
#' but do not have an extraction or validation function. This is to ensure
#' that no redundant validation is performed by the evaluating LLM on
#' instructions which have already been validated by functions in those
#' prompt wraps.
#'
#' @param prompt A single string or a [tidyprompt-class] object
#'
#' @param question The question to ask the LLM to verify the result of the prompt.
#' The LLM will be presented the original prompt, its result, and this question.
#' The LLM will be asked to provide a boolean answer to this question. If TRUE,
#' the result of the prompt will be accepted; if FALSE, the result will be declined
#'
#' @param llm_provider A [llm_provider-class] object
#'  which will be used to verify the evaluation led to a satisfactory result.
#' If not provided, the same LLM provider as the prompt was originally
#'  evaluated with will be used
#'
#' @param max_words_feedback The maximum number of words allowed in the summary of
#'  why the result was declined.
#' This summary is sent back to the LLM originally asked to evaluate the prompt
#'
#' @return A [tidyprompt-class] with an added [prompt_wrap()] which
#'  will add a check for a LLM to accept or decline the result of the prompt,
#'  providing feedback if the result is declined
#'
#' @export
#'
#' @example inst/examples/llm_verify.R
llm_verify <- function(
  prompt,
  question = "Is the answer satisfactory?",
  llm_provider = NULL,
  max_words_feedback = 50
) {
  prompt <- tidyprompt(prompt)

  # Verify & install provided LLM provider
  stopifnot(
    is.null(llm_provider) || inherits(llm_provider, "LlmProvider")
  )
  super_llm_provider <- llm_provider

  validation_fn <- function(response, llm_provider) {
    if (!is.null(super_llm_provider)) {
      llm_provider <- super_llm_provider
    }
    llm_provider <- llm_provider$clone()

    prompt_text <- self$base_prompt
    prompt_wraps <- self$get_prompt_wraps("modification")
    for (wrap in prompt_wraps) {
      if (is.null(wrap$modify_fn)) {
        next
      }
      # We skip prompt wraps that have extraction or validation functions;
      #   we assume that potential instructions added by these wraps
      #   have been properly validated by these functions;
      #   adding them to the prompt text here would lead to
      #   redundant validation by the LLM
      if (!is.null(wrap$extraction_fn) || !is.null(wrap$validation_fn)) {
        next
      }

      prompt_text <- wrap$modify_fn(prompt_text)
    }

    original_prompt <- self$construct_prompt_text()
    result_as_text <- utils::capture.output(print(response))

    satisfied <- glue::glue(
      ">>> An assistant was asked:\n\n",
      "{prompt_text}\n\n",
      ">>> The assistant answered:\n\n",
      "  ",
      paste(result_as_text, collapse = "\n  "),
      "\n\n",
      ">>> {question}",
      .trim = FALSE
    ) |>
      answer_as_boolean() |>
      answer_by_chain_of_thought() |>
      send_prompt(llm_provider, return_mode = "full")

    if (satisfied$response) {
      return(TRUE)
    }

    # Create summary of why the response was declined
    feedback <- glue::glue(
      "An assistant's answer was declined by another assistant.",
      "The declining assistant provided their chain of thought:\n\n",
      satisfied$chat_history_clean |>
        dplyr::select(c("role", "content")) |>
        df_to_string(),
      "\n\n",
      "Summarize why the result was declined."
    ) |>
      answer_as_text(max_words = max_words_feedback) |>
      send_prompt(llm_provider)

    return(
      llm_feedback(
        paste0(
          "Your answer was rejected for this reason:\n\n",
          feedback
        )
      )
    )
  }

  prompt_wrap(
    prompt,
    validation_fn = validation_fn,
    type = "check"
  )
}

---
title: "Streaming LLM responses to Shiny apps"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Streaming LLM responses to Shiny apps}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

This vignette shows a minimal example of how to stream a LLM response
gathered with 'tidyprompt' to a Shiny app in real-time. 

Shiny apps run on a single R process, meaning that when you call an LLM synchronously,
the UI will be blocked until the response is complete. Therefore, when integrating LLM calls
into Shiny apps, you typically want to use `send_prompt()` asynchronously 
(e.g., with the 'future' and 'promises' packages).

If you just want to show the final LLM response after it is complete, this is 
fairly straightforward. But if you want to show the LLM response as it streams in 
token-by-token, this is a bit more complex. The asynchronous process will have
to somehow communicate back to the main Shiny R process to update the UI in real-time
as new tokens arrive.

Therefore, for LLM providers created with 'tidyprompt' that support streaming responses, 
you can set the `stream_callback` field of your `llm_provider` object. This is a
function that is called for each token (or text chunk) as it arrives from the LLM.
You can then leverage this to push the tokens into the Shiny app in real-time. 
The 'ipc' package can facilitate herein, allowing inter-process communication between
the background R process (where the LLM call runs) and the main Shiny R process
(where the UI runs). Each time a new output token arrives, the `stream_callback` can push
then push the token into a reactive value in the main Shiny process.

## Example

Below is a minimal example which shows how to achieve this. We will:

- create an `llm_provider` with streaming enabled;
- define a `stream_callback` that writes tokens into an `ipc::shinyQueue`;
- start a `future` in a separate R process (`future::plan(multisession)`) where we call `send_prompt()`;
- consume the queue from the Shiny main process to update the UI, showing a live stream of LLM output.

```{r setup, eval = FALSE}
# Install & load required packages
packages <- c("shiny", "ipc", "future", "promises", "tidyprompt")
for (pkg in packages) {
  if (!requireNamespace(pkg, quietly = TRUE)) {
    install.packages(pkg)
  }
  library(pkg, character.only = TRUE)
}

# Enable asynchronous processing
future::plan(future::multisession)

# Base provider (OpenAI; has streaming enabled by default)
base_provider <- llm_provider_openai()

ui <- fluidPage(
  titlePanel("tidyprompt streaming demo"),

  sidebarLayout(
    sidebarPanel(
      textInput(
        "prompt",
        "Prompt",
        value = "Tell me a short story about a cat and a robot."
      ),
      actionButton("run", "Ask model"),
      helpText("Tokens will appear below as they stream in.")
    ),

    mainPanel(
      verbatimTextOutput("partial_response")
    )
  )
)

server <- function(input, output, session) {
  # Queue to bridge async future back into Shiny
  queue <- shinyQueue()
  queue$consumer$start(100) # process queue every 100 ms

  # Reactive that holds the accumulated streamed text
  partial_response <- reactiveVal("")

  # Streaming callback run inside the provider
  stream_cb <- function(token, meta) {
    # meta$partial_response is the accumulated text so far
    queue$producer$fireAssignReactive(
      "partial_response",
      meta$partial_response
    )
    invisible(TRUE)
  }

  # Clone provider for this session and attach callback + streaming
  provider <- base_provider$clone()
  provider$parameters$stream <- TRUE
  provider$stream_callback <- stream_cb

  # Expose the reactive value to the UI
  output$partial_response <- renderText({
    req(partial_response())
    partial_response()
  })

  observeEvent(input$run, {
    # Reset current streamed text on each run
    partial_response("")

    user_prompt <- input$prompt

    future_promise(
      {
        tidyprompt::send_prompt(
          prompt = user_prompt,
          llm_provider = provider,
          return_mode = "only_response"
        )
      },
      globals = list(
        user_prompt = user_prompt,
        provider = provider
      )
    ) |>
      then(
        onFulfilled = function(value) {
          # Final response once streaming finishes
          partial_response(value)
        },
        onRejected = function(error) {
          showNotification(
            paste("Error:", error$message),
            type = "error"
          )
          print(error)
        }
      )

    NULL
  })
}

shinyApp(ui = ui, server = server)
```

---
title: "Creating prompt wraps"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Creating prompt wraps}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---
  

### Creating prompt wraps

Using `prompt_wrap()`, you can create your own prompt wraps.
An input for `prompt_wrap()` wrap may be string or a tidyprompt object. If you
pass a string, it will be automatically turned into a tidyprompt object.

Under the hood, a tidyprompt object is just a list with a base prompt (a string) and a series of prompt
wraps. `prompt_wrap()` adds a new prompt wrap to the list of prompt wraps. Each prompt wrap is a list
with a modification function, an extraction function, and/or a validation function (at least
one of these functions must be present). The modification function alters the prompt text,
the extraction function applies a transformation to the LLM's response, and the validation function
checks if the (transformed) LLM's response is valid.

Both extraction and validation functions can return feedback to the LLM, using
`llm_feedback()`. When an extraction or validation function returns this, a
message is sent back to the LLM, and the LLM can retry answering the prompt according
to the feedback. Feedback messages may be a reiteration of instruction or a specific error
message which occured during extraction or validation. When all extractions and validations
have been applied without resulting in feedback, the LLM's response (after transformations
by the extraction functions) will be returned. (`send_prompt()` is responsible
for executing this process.)

Below is a simple example of a prompt wrap, which just adds some text to the base prompt:

```{r, eval=FALSE}
prompt <- "Hi there!" |>
  prompt_wrap(
    modify_fn = function(base_prompt) {
      paste(base_prompt, "How are you?", sep = "\n\n")
    }
  )
```

Shorter notation of the above would be:

```{r, eval=FALSE}
prompt <- "Hi there!" |>
  prompt_wrap(\(x) paste(x, "How are you?", sep = "\n\n"))
```

Often times, it may be preferred to make a function which takes a prompt and
returns a wrapped prompt:

```{r, eval=FALSE}
my_prompt_wrap <- function(prompt) {
  modify_fn <- function(base_prompt) {
    paste(base_prompt, "How are you?", sep = "\n\n")
  }

  prompt_wrap(prompt, modify_fn)
}
prompt <- "Hi there!" |>
  my_prompt_wrap()
```

Take look at the source code of `answer_as_boolean()`, which also uses extraction:

```{r, eval=FALSE}
answer_as_boolean <- function(
    prompt,
    true_definition = NULL,
    false_definition = NULL,
    add_instruction_to_prompt = TRUE
) {
  instruction <- "You must answer with only TRUE or FALSE (use no other characters)."
  if (!is.null(true_definition))
    instruction <- paste(instruction, glue::glue("TRUE means: {true_definition}."))
  if (!is.null(false_definition))
    instruction <- paste(instruction, glue::glue("FALSE means: {false_definition}."))

  modify_fn <- function(original_prompt_text) {
    if (!add_instruction_to_prompt) {
      return(original_prompt_text)
    }

    glue::glue("{original_prompt_text}\n\n{instruction}")
  }

  extraction_fn <- function(x) {
    normalized <- tolower(trimws(x))
    if (normalized %in% c("true", "false")) {
      return(as.logical(normalized))
    }
    return(llm_feedback(instruction))
  }

  prompt_wrap(prompt, modify_fn, extraction_fn)
}
```

Take a look at the source code of, for instance, `answer_as_integer()`, `answer_by_chain_of_thought()`,
and `answer_using_tools()` for more advanced examples of prompt wraps.

#### Breaking out of the evaluation loop

In some cases, you may want to exit the extraction or validation process early. For instance,
your LLM may indicate that it is unable to answer the prompt. In such cases, you can have your
extraction or validation function return `llm_break()`. This will cause the evaluation loop
to break, forwarding to the return statement of `send_prompt()`. See `quit_if()`
for an example of this.

#### Extraction versus validation functions

Both extraction and validation functions can return `llm_break()` or `llm_feedback()`.
The difference between extraction and validation functions is only that an extraction may
transform the LLM response and pass it on to the next extraction and/or validation functions,
while a validation function only checks if the LLM response passes a logical test (without
altering the response). Thus, if you wish, you can perform validations in an
extraction function.

#### Prompt wrap types and order of application

When constructing the prompt text and when evaluating a prompt, prompt wraps are applied
prompt wrap after prompt wrap (e.g., first the extraction and validation functions of
one wrap, then of the other).

The order in which prompt wraps are applied is important. Currently, four types of prompt wraps
are distinguished: 'unspecified', 'break', 'mode', and 'tool'.

When constructing the prompt text,
prompt wraps are applied in the order of these types. Prompt wraps will be automatically reordered if
necesarry (keeping intact the order of prompt wraps of the same type).

When evaluating the prompt,
prompt wraps are applied in the reverse order of types (i.e., first 'tool', then 'mode', then 'break',
and finally 'unspecified'). This is because 'tool' prompt wraps may return a value to be used in
the final answer, 'mode' prompt wraps alter how a LLM forms a final answer,
'break' prompt wraps quit evaluation early based on a specific final answer, and 'unspecified'
prompt wraps are the most general type of prompt wraps which force a final answer
to be in a specific format.

#### Configuring a LLM provider with a prompt wrap

Advanced prompt wraps may wish to configure certain settings of a LLM provider.
For instance, `answer_as_json()` configures certain parameters of the LLM provider,
based on which LLM provider is being used (Ollama and OpenAI have different API parameters
available for JSON output). This is done by defining a `parameter_fn` within the prompt wrap;
`parameter_fn` is a function which takes the LLM provider as input and returns a list of parameters,
which will be set as the parameters of the LLM provider before sending the prompt. See
`prompt_wrap()` documentation and `answer_as_json()` for an example.

#### Configuring a prompt wrap based on the LLM provider or HTTP responses

`modify_fn`, `extraction_fn`, and `validation_fn` may take the
LLM provider as the second argument and the `http_list` (a list of all
HTTP responses made during `send_prompt()`) as the third argument. This allows for
advanced configuration of the prompt and the extraction and validation logic
based on the LLM provider and any data available the HTTP responses.

#### Configuring a prompt wrap based on other prompt wraps

`modify_fn`, `extraction_fn`, and `validation_fn` all have access to the `self`
object, which represents the `tidyprompt-class` object that they are a part of. This
allows for advanced configuration of the prompt and the extraction and validation logic
based on other prompt wraps. Inside these functions, simply use `self$` to access
the `tidyprompt-class` object. 

#### Handler functions

Prompt wraps may also include 'handler functions'. These are functions which are called
upon every received chat completion. This allows for advanced
processing of the LLM's responses, such as logging, tracking tokens, or other
custom processing. See the `prompt_wrap()` documentation for more information;
see the source code of `answer_using_tools()` for an example of a handler function.

---
title: "Getting started"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Getting started}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
resource_files:
  - ../man/figures/answer_using_r1-1.png
---




``` r
library(tidyprompt)
```

## Setup an LLM provider

'tidyprompt' can be used with any LLM provider capable of completing a chat.

At the moment, 'tidyprompt' includes pre-built functions to connect with various
LLM providers, such as Ollama, OpenAI, OpenRouter, Mistral, Groq, XAI (Grok), and Google Gemini.

With the `llm_provider_ellmer()` function, you can also create an LLM provider
from an `ellmer::chat()` object. This allows you to use any LLM provider
that you can configure with the 'ellmer' R package (including its respective
configuration and features). It is also possible to directly use an `ellmer::chat()` 
object with `send_prompt()`. When using an 'ellmer' LLM provider, 
functions like `answer_as_json()` and `answer_using_tools()` will call the native
'ellmer' functions for these features.

Furthermore, with the `llm_provider-class`, you can easily write a hook for any other LLM provider.
You could make API calls using the 'httr2' package or use another R package that already has
a hook for the LLM provider you want to use. If your API of choice is compatible with the structure
of the OpenAI API, you can call `llm_provider_openai()` and change only
the URL and the API key to make it work.

``` r
# Ollama running on local PC
ollama <- llm_provider_ollama(
  parameters = list(model = "llama3.1:8b"),
)

# OpenAI API
openai <- llm_provider_openai(
  parameters = list(model = "gpt-4o-mini")
)

# Various providers via OpenRouter (e.g., Anthropic)
openrouter <- llm_provider_openrouter(
  parameters = list(model = "anthropic/claude-4.6-sonnet")
)

# ... functions also included for Mistral, Groq, XAI (Grok), and Google Gemini

# Alternatively, you can use an `ellmer::chat()` object (e.g., via `ellmer::chat_openai()`)
# (See also: https://ellmer.tidyverse.org/index.html)
ellmer <- llm_provider_ellmer(ellmer::chat_openai(model = "gpt-5.4-nano"))

# ... or easily create your own hook for any other LLM provider;
#   see ?`llm_provider-class` for more information; also take a look at the source code of
#   `llm_provider_ollama()` and `llm_provider_openai()`. For APIs that follow the structure
#   of the OpenAI API for chat completion, you can use `llm_provider_openai()`
```

## Basic prompting

A simple string serves as the base for a prompt.

By adding prompt wraps, you can influence various aspects of how the LLM handles the prompt,
while verifying that the output is structured and valid (including retries with feedback to the LLM if it is not).


``` r
  "Hi there!" |>
    send_prompt(ollama)
#> --- Sending request to LLM provider (llama3.1:8b): ---
#> Hi there!
#> --- Receiving response from LLM provider: ---
#> How's your day going so far? Is there something I can help you with or would you like to chat?
#> [1] "How's your day going so far? Is there something I can help you with or would you like to chat?"
```
`add_text()` is a simple example of a prompt wrap. It simply adds some text at the end of the base prompt.


``` r
  "Hi there!" |>
    add_text("What is a large language model? Explain in 10 words.") |>
    send_prompt(ollama)
#> --- Sending request to LLM provider (llama3.1:8b): ---
#> Hi there!
#> 
#> What is a large language model? Explain in 10 words.
#> --- Receiving response from LLM provider: ---
#> Advanced computer program trained on vast amounts of written data.
#> [1] "Advanced computer program trained on vast amounts of written data."
```
You can also construct the final prompt text, without sending it to an LLM provider.


``` r
  "Hi there!" |>
    add_text("What is a large language model? Explain in 10 words.")
#> <tidyprompt>
#> The base prompt is modified by a prompt wrap, resulting in:
#> > Hi there!
#> > 
#> > What is a large language model? Explain in 10 words. 
#> Use 'x$base_prompt' to show the base prompt text.
#> Use 'x$construct_prompt_text()' to get the full prompt text.
#> Use 'get_prompt_wraps(x)' to show the prompt wraps.
```

## Prompt wraps

This package contains three main families of pre-built prompt wraps which affect
how a LLM handles a prompt:

- `answer_as`: specify the **format** of the output (e.g., integer, list, json)
- `answer_by`: specify a **reasoning mode** to reach the answer (e.g., chain-of-thought, ReAct)
- `answer_using`: give the LLM **tools** to reach the answer (e.g., R functions, R code, SQL)

Below, we will show examples of each type of prompt wrap.

### answer_as: Retrieving output in a specific format

Using prompt wraps, you can force the LLM to return the output in a specific format.
You can also extract the output to turn it from a character into another data type.

For instance, `answer_as_integer()` adds a prompt wrap which forces the LLM
to reply with an integer.

To achieve this, the prompt wrap will add some text to the base prompt, asking
the LLM to reply with an integer. However, the prompt wrap does more: it also
will attempt to extract and validate the integer from the LLM's response. If
extraction or validation fails, feedback is sent back to the LLM, after which
the LLM can retry answering the prompt. Because the extraction function turns the
original character response into a numeric value, the final output from
`send_prompt()` will also be a numeric type.


``` r
  "What is 2 + 2?" |>
    answer_as_integer() |>
    send_prompt(ollama)
#> --- Sending request to LLM provider (llama3.1:8b): ---
#> What is 2 + 2?
#> 
#> You must answer with only an integer (use no other characters).
#> --- Receiving response from LLM provider: ---
#> 4
#> [1] 4
```
Below is an example of a prompt which will initially fail, but will succeed after
`llm_feedback()` and a retry.


``` r
  "What is 2 + 2?" |>
    add_text("Please write out your reply in words, use no numbers.") |>
    answer_as_integer(add_instruction_to_prompt = FALSE) |>
    send_prompt(ollama)
#> --- Sending request to LLM provider (llama3.1:8b): ---
#> What is 2 + 2?
#> 
#> Please write out your reply in words, use no numbers.
#> --- Receiving response from LLM provider: ---
#> Four.
#> --- Sending request to LLM provider (llama3.1:8b): ---
#> You must answer with only an integer (use no other characters).
#> --- Receiving response from LLM provider: ---
#> 4
#> [1] 4
```

'tidyprompt' offers various other 'answer_as' functions, such as `answer_as_boolean()`, `answer_as_regex_match()`,
`answer_as_named_list()`, `answer_as_text()` and `answer_as_json()`.

##### JSON output

`answer_as_json()` may be especially powerful when your LLM provider and model have native support for
returning JSON objects and adhering to JSON schemas (e.g., OpenAI, Ollama). Text-based handling is however always
possible, also for providers which do not natively support such functions. This means that
you can always switch between providers while ensuring the results will be in the correct format.

`answer_as_json()` can take schemas defined as either a R list object representing a JSON schema,
or structured output definitions from the 'ellmer' package (e.g., `ellmer::type_object()`).
See the documentation of `answer_as_json()` for more information on how to use it.

### answer_by: Adding a reasoning mode to the LLM

Prompt wraps may also be used to add a reasoning mode to the LLM. It is hypothesized that this could
improve the LLM's performance on more complex tasks.

For instance, `answer_by_chain_of_thought()` will add chain of thought reasoning mode to the
prompt evaluation by the LLM.
The function wraps the base prompt text within a request for the LLM to reason step by step, asking
it to provide the final answer within 'FINISH[<final answer here>]'. An extraction
function then ensures only the final answer is returned.


``` r
  "What is 2 + 2?" |>
    answer_by_chain_of_thought() |>
    answer_as_integer() |>
    send_prompt(ollama)
#> --- Sending request to LLM provider (llama3.1:8b): ---
#> You are given a user's prompt.
#> To answer the user's prompt, you need to think step by step to arrive at a final answer.
#> 
#> ----- START OF USER'S PROMPT -----
#> What is 2 + 2?
#> 
#> You must answer with only an integer (use no other characters).
#> ----- END OF USER'S PROMPT -----
#> 
#> What are the steps you would take to answer the user's prompt?
#> Describe your thought process in the following format:
#>   >> step 1: <step 1 description>
#>   >> step 2: <step 2 description>
#>   (etc.)
#> 
#> When you are done, you must type:
#>   FINISH[<put here your final answer to the user's prompt>]
#> 
#> Make sure your final answer follows the logical conclusion of your thought process.
#> --- Receiving response from LLM provider: ---
#> Here are my thought steps:
#> 
#> >> step 1: Understand the question - The prompt asks for the result of a simple arithmetic operation, specifically adding 2 and 2.
#> 
#> >> step 2: Recall basic math facts - I recall that the sum of two identical numbers is equal to twice the value of each number. In this case, both numbers are 2.
#> 
#> >> step 3: Apply the math fact - Using the knowledge from step 2, I calculate the result by multiplying 2 (the number being added) by 2 (the other number), which gives 4.
#> 
#> >> step 4: Confirm the answer - Before providing a final response, I confirm that my calculation is correct. Adding 2 and 2 indeed equals 4.
#> 
#> FINISH[4]
#> [1] 4
```

### answer_using: Have the LLM work with tools and code

#### Tools (function-calling)

With `answer_using_tools()`, you can enable your LLM to call R functions.
This enables the LLM to autonomously retrieve additional information or take other actions.

`answer_using_tools()` automatically extracts documentation when it is available
for base R functions and functions from packages. Types are inferred from the
default arguments of the function. If you want to define a custom function and/or override the default documentation,
you can use `tools_add_docs()`. See example usage in the documentation of `answer_using_tools()`.

`answer_using_tools()` also accepts tool definitions from `ellmer::tool()`.
Because `answer_using_tools()` can convert tool definitions, `ellmer::tool()`
definitions will work with any LLM provider type (also non-'ellmer' LLM
providers). Similarly, when using an `llm_provider_ellmer()`, a tool definition
can also be given with `tools_add_docs()`. And, when using an `llm_provider_ellmer()`,
`answer_using_tools()` will call the native 'ellmer' functions to register
tools. Provider-specific 'ellmer' built-in tools are the exception: they only
work with 'ellmer'-backed providers. See example usage in the documentation of
`answer_using_tools()`.

Tool definitions returned by `mcptools::mcp_tools()` may also be provided,
allowing the use of tools from Model Context Protocol (MCP) servers.

`answer_using_tools()` supports both text-based function calling and native function calling (via API
parameters, currently implemented for OpenAI and Ollama API structures, and 'ellmer' LLM providers).

``` r
  "What are the files in my current directory?" |>
    answer_using_tools(list.files) |>
    send_prompt(ollama)
#> ! `answer_using_tools()`, `tools_docs_to_text()`:
#> * Argument 'pattern' has an unknown type. Defaulting to 'string'
#> --- Sending request to LLM provider (llama3.1:8b): ---
#> What are the files in my current directory?
#> 
#> If you need more information, you can call functions to help you.
#> 
#> To call a function, output a JSON object with the following format:
#>   {
#>     "function": "<function name>",
#>     "arguments": {
#>       "<argument_name>": <argument_value>,
#>       # ...
#>     }
#>   }
#>   (Note: you may not provide function calls as function arguments.)
#> 
#> The following functions are available:
#> 
#>   function name: list.files
#>   description: List the Files in a Directory/Folder: These functions produce a character vector of the names of files
#> or directories in the named directory.
#>   arguments:
#>     - path: a character vector of full path names; the default corresponds to the working directory, 'getwd()'. (...) [Type: string]
#>     - pattern: an optional regular expression.  Only file names which match the regular expression will be returned. [Type: unknown]
#>     - all.files: a logical value.  If 'FALSE', only the names of visible files are returned (following Unix-style visibility, that is files whose name does not start with a dot).  If 'TRUE', all file names will be returned. [Type: logical]
#>     - full.names: a logical value.  If 'TRUE', the directory path is prepended to the file names to give a relative file path.  If 'FALSE', the file names (rather than paths) are returned. [Type: logical]
#>     - recursive: logical.  Should the listing recurse into directories? [Type: logical]
#>     - ignore.case: logical.  Should pattern-matching be case-insensitive? [Type: logical]
#>     - (...)
#> 
#> After you call a function, wait until you receive more information.
#> Use the information to decide your next steps or provide a final response.
#> --- Receiving response from LLM provider: ---
#> To get the files in my current directory, I will call the `list.files` function with the necessary arguments.
#> 
#> Here's the JSON object:
#> 
#> ```
#> {
#>   "function": "list.files",
#>   "arguments": {
#>     "path": "."
#>   }
#> }
#> ```
#> 
#> This will list the files in the current working directory. The argument `path` is set to `"."`, which corresponds to the current working directory, and I'm not providing any pattern or other arguments as they are optional.
#> 
#> Let's see what happens next...
#> --- Sending request to LLM provider (llama3.1:8b): ---
#> function called: list.files
#> arguments used: path = .
#> result: getting_started.Rmd, getting_started.Rmd.orig, man, precompile vignettes.R, sentiment_analysis.Rmd, sentiment_analysis.Rmd.orig
#> --- Receiving response from LLM provider: ---
#> It looks like I have a mix of files in my current directory.
#> 
#> The `list.files` function returned a list of file names, which are:
#> 
#> * `getting_started.Rmd`
#> * `getting_started.Rmd.orig`
#> * `man`
#> * `precompile vignettes.R`
#> * `sentiment_analysis.Rmd`
#> * `sentiment_analysis.Rmd.orig`
#> 
#> These files seem to be related to some kind of project or documentation.
```

#### Code generation and evaluation

`answer_using_r()` provides a more advanced prompt wrap, which has various options
to enable LLM code generation. R code can be extracted, parsed
for validity, and optionally be evaluated in a dedicated R session (using the
'callr' package). The prompt wrap can also be set to 'tool mode' (with `output_as_tool = TRUE`),
where the output of R code is returned to the LLM, so that it can be used to formulate
a final answer.


``` r
# From prompt to ggplot
plot <- paste0(
  "Create a scatter plot of miles per gallon (mpg) versus",
  " horsepower (hp) for the cars in the mtcars dataset.",
  " Use different colors to represent the number of cylinders (cyl).",
  " Make the plot nice and readable,",
  " but also be creative, a little crazy, and have humour!"
) |>
  answer_using_r(
    pkgs_to_use = c("ggplot2"),
    evaluate_code = TRUE,
    return_mode = "object"
  ) |>
  send_prompt(openai)
plot
```
![plot](../man/figures/answer_using_r1-1.png)

### Creating custom prompt wraps

See `prompt_wrap()` and the '[Creating prompt wraps](creating_prompt_wraps.html)'
vignette for information on how to create your own prompt wraps.

### Provider-level prompt wraps

Provider-level prompt wraps let you set defaults for a specific LLM provider, so they apply to every prompt you send through it.
This is useful if you want consistent behavior across all prompts without repeating the same wraps every time.
You can choose to apply a provider-level prompt wrap either before or after the prompt-specific prompt wraps.

```{r provider_wraps, cache=FALSE, eval=FALSE}
ollama <- llm_provider_ollama()

# Example: add a "short answer" mode (provider-level post prompt wrap)
ollama$add_prompt_wrap(
  provider_prompt_wrap(
    modify_fn = \(txt) paste0(
      txt,
      "\n\nPlease answer concisely (< 2 sentences)."
    )
  ),
  position = "post"
)

# Use as usual: wraps are applied automatically
"What's a vignette in R?" |> send_prompt(ollama)
```

---
title: "Sentiment analysis in R with a LLM and 'tidyprompt'"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Sentiment analysis in R with a LLM and 'tidyprompt'}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
resource_files:
  - ../man/figures/plot_sentiment_analysis-1.png
---



This vignette will show you how to use the `tidyprompt` package to perform a sentiment analysis
in R with a LLM (large language model).

We will first define a simple dataset of sentences that we will create sentiment scores for.


``` r
sentences_df <- data.frame(
  sentence = c(
    "I love this product!",
    "This product is terrible",
    "The customer service was excellent",
    "I am very disappointed with this product",
    "The delivery was fast and efficient",
    "I would not recommend this product to anyone",
    "It was not bad, not great either",
    "Meh",
    "It felt like walking up a mountain",
    "I am angry!!!"
  )
)
```

First we will create a connection to a locally running LLM (using Ollama)


``` r
library(tidyprompt)

ollama <- llm_provider_ollama()
ollama$parameters$model <- "llama3.1:8b"
```

For every sentence, we will now prompt the LLM to provide a sentiment score.
To do this, we will use 'purrr' `map()` to iterate over each sentence.
Within each iteration, we use a `tidyprompt()` to prompt the LLM for the sentiment score.
The latter will force the LLM to answer as an integer and extract the integer from its response.


``` r
library(purrr)

sentences_df$sentiment_score <- map_int(
  sentences_df$sentence,
  function(sentence) {
    paste0(
      "Please provide a sentiment score for the following sentence:\n\n",
      sentence
    ) |>
      answer_as_integer(min = 1, max = 100) |>
      send_prompt(ollama, verbose = FALSE)
  }
)

sentences_df
#>                                        sentence sentiment_score
#> 1                          I love this product!              95
#> 2                      This product is terrible               3
#> 3            The customer service was excellent              95
#> 4      I am very disappointed with this product              20
#> 5           The delivery was fast and efficient              95
#> 6  I would not recommend this product to anyone               3
#> 7              It was not bad, not great either              46
#> 8                                           Meh              50
#> 9            It felt like walking up a mountain              25
#> 10                                I am angry!!!              10
```

Let's plot the results!


``` r
library(ggplot2)

ggplot(sentences_df, aes(x = sentiment_score, y = reorder(sentence, sentiment_score))) +
  geom_col(aes(fill = sentiment_score)) +
  scale_fill_gradient(low = "red", high = "green") +
  theme_minimal() +
  labs(
    title = "Sentiment scores for each sentence",
    x = "Sentiment score",
    y = "Sentence"
  )
```

![Plot of sentiment scores](../man/figures/plot_sentiment_analysis-1.png)

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/helper_text_df_to_string.R
\name{df_to_string}
\alias{df_to_string}
\title{Convert a dataframe to a string representation}
\usage{
df_to_string(df, how = c("wide", "long"))
}
\arguments{
\item{df}{A \code{data.frame} object to be converted to a string}

\item{how}{In what way the df should be converted to a string;
either "wide" or "long". "wide" presents column names on the first row,
followed by the row values on each new row. "long" presents the values
of each row together with the column names, repeating for every row
after two lines of whitespace}
}
\value{
A single string representing the df
}
\description{
Converts a data frame to a string format, intended for sending it to a LLM
(or for display or logging).
}
\examples{
cars |>
  head(5) |>
  df_to_string(how = "wide")

cars |>
  head(5) |>
  df_to_string(how = "long")
}
\seealso{
Other text_helpers: 
\code{\link{skim_with_labels_and_levels}()},
\code{\link{vector_list_to_string}()}
}
\concept{text_helpers}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/answer_using_tools.R
\name{tools_add_docs}
\alias{tools_add_docs}
\title{Add tidyprompt function documentation to a function}
\usage{
tools_add_docs(func, docs)
}
\arguments{
\item{func}{A function object}

\item{docs}{A list with the following elements:
\itemize{
\item 'name': (optional) The name of the function. If not provided, the function
name will be extracted from the function object. Use this parameter to override
the function name if necessary
\item 'description': A description of the function and its purpose
\item 'arguments': A named list of arguments with descriptions. Each argument is a list
which may contain: \itemize{
\item 'description': A description of the argument and its purpose. Not
required or used for native function calling (e.g., with OpenAI), but recommended
for text-based function calling
\item 'type': The type of the argument. This should be one of:
'integer', 'numeric', 'logical', 'string', 'ignore', 'match.arg',
'vector integer', 'vector numeric', 'vector logical', 'vector string'.
For arguments which are named lists, 'type' should be a named list
which contains the types of the elements. For type 'match.arg', the
possible values should be passed as a vector under 'default_value'.
Use 'ignore' to keep an argument out of generated tool prompts/schemas and
rely on the R function itself to supply it (for example via a default value).
'type' is required for native function calling (with, e.g., OpenAI) but may
also be useful to provide for text-based function calling, in which it will
be added to the prompt introducing the function
\item 'default_value': The default value of the argument. This is only required
when 'type' is set to 'match.arg'. It should then be a vector of possible values
for the argument. In other cases, it is not required; for native function calling,
it is not used in other cases; for text-based function calling, it may be useful
to provide the default value, which will be added to the prompt introducing the
function
}
\item 'return': A list with the following elements:
\itemize{
\item 'description': A description of the return value or the side effects of the function
}
}}
}
\value{
The function object with the documentation added as an attribute
('tidyprompt_tool_docs')
}
\description{
This function adds documentation to a custom function. This documentation
is used to extract information about the function's name, description, arguments,
and return value. This information is used to provide an LLM with information
about the functions, so that the LLM can call R functions. The intended
use of this function is to add documentation to custom functions that do not
have help files; \code{\link[=tools_get_docs]{tools_get_docs()}} may generate documentation from a
help file when the function is part of base R or a package.

If a function already has documentation, the documentation added by this
function may overwrite it. If you wish to modify existing documentation,
you may make a call to \code{\link[=tools_get_docs]{tools_get_docs()}} to extract the existing documentation,
modify it, and then call \code{\link[=tools_add_docs]{tools_add_docs()}} to add the modified documentation.
}
\examples{
# When using functions from base R or R packages,
#   documentation is automatically extracted from help files:
prompt_with_dir_function <- "What are the files in my current directory?" |>
  answer_using_tools(dir) # The 'dir' function is from base R
\dontrun{
  send_prompt(prompt_with_dir_function)
  # --- Sending request to LLM provider (llama3.1:8b): ---
  #   What are the files in my current directory?
  #   --- Receiving response from LLM provider: ---
  #   Calling function 'nm' with arguments:
  #   {
  #     "all.files": true,
  #     "full.names": false,
  #     "ignore.case": false,
  #     "include.dirs": false,
  #     "no..": false,
  #     "path": "./",
  #     "pattern": "*",
  #     "recursive": false
  #   }
  # Result:
  #   .git, .github, .gitignore, .Rbuildignore, .Rhistory, ...
  # The files in your current directory are:
  #   .git, .github, .gitignore, .Rbuildignore, .Rhistory, ...
  # [1] "The files in your current directory are:\n\n .git, .github, ..."
}

# Users may provide custom functions in two ways:
#   1) as a function object, optionally documented with `tools_get_docs()`, or
#   2) as an 'ellmer' tool definition, using `ellmer::tool()`

# Take this fake weather function as an example:
temperature_in_location <- function(
  location = c("Amsterdam", "Utrecht", "Enschede"),
  unit = c("Celcius", "Fahrenheit")
) {
  location <- match.arg(location)
  unit <- match.arg(unit)

  temperature_celcius <- switch(
    location,
    "Amsterdam" = 32.5,
    "Utrecht" = 19.8,
    "Enschede" = 22.7
  )

  if (unit == "Celcius") {
    return(temperature_celcius)
  } else {
    return(temperature_celcius * 9/5 + 32)
  }
}


# 1: `tools_add_docs()` --------------------------------------------------------

# Generate documentation for a function, based on formals & help file
docs <- tools_get_docs(temperature_in_location)

# The types get inferred from the function's formals
# However, descriptions are still missing as the function is not from a package
# We can modify the documentation object to add descriptions:
docs$description <- "Get the temperature in a location"
docs$arguments$unit$description <- "Unit in which to return the temperature"
docs$arguments$location$description <- "Location for which to return the temperature"
docs$return$description <- "The temperature in the specified location and unit"
# (See `?tools_add_docs` for more details on the structure of the documentation)

# When we are satisfied with the documentation, we can add it to the function:
temperature_in_location <- tools_add_docs(temperature_in_location, docs)

prompt_with_weather_function <-
  "What is the weather in Enschede? Give me Celcius degrees" |>
  answer_using_tools(temperature_in_location)
\dontrun{
  send_prompt(prompt_with_weather_function)
  # --- Sending request to LLM provider (llama3.1:8b): ---
  #   What is the weather in Enschede? Give me Celcius degrees
  #   --- Receiving response from LLM provider: ---
  #   Calling function 'temperature_in_location' with arguments:
  #   {
  #     "location": "Enschede",
  #     "unit": "Celcius"
  #   }
  # Result:
  #   22.7
  # The temperature in Enschede is 22.7 Celcius degrees.
  # [1] "The temperature in Enschede is 22.7 Celcius degrees."
}

# 2: `ellmer::tool()` -----------------------------------------------

# Alternatively, we can define the function as an 'ellmer' tool

temperature_in_location_ellmer <- ellmer::tool(
  temperature_in_location,
  name = "get_temperature",
  description = "Get the temperature in a location",
  arguments = list(
    location = ellmer::type_string(
      "Location for which to return the temperature", required = TRUE
    ),
    unit = ellmer::type_string(
      "Unit in which to return the temperature", required = TRUE
    )
  )
)

prompt_with_weather_function_ellmer <-
  "What is the weather in Utrecht? Give me Fahrenheit degrees" |>
  answer_using_tools(temperature_in_location_ellmer)
\dontrun{
  send_prompt(prompt_with_weather_function_ellmer)
  # ...
}

# Because `mcptools::mcp_tools()` returns a list of `ellmer:tool()` tools,
#   you can also use Model Context Protocol (MCP) server tools with
#   `answer_using_tools()`:
\dontrun{
  prompt_using_mcp_tools <- mcptools::mcp_tools()
  "Push my latest commit to GitHub" |>
    answer_using_tools(mcp_tools)
  send_prompt(prompt_using_mcp_tools)
}

# `answer_using_tools()` will automatically attempt to use the most appropriate
#   way of sending the tool to the LLM

# If you use a LLM provider of type 'ollama' or 'openai',
#   it will automatically convert the tool definition to parameters
#   appropriate for those APIs
# If you use a LLM provider of type 'ellmer', it will call the appropriate
#   ellmer function directly which will handle the tool call for various
#   providers
# Note that both tool definitions from `tools_add_docs()` and `ellmer::tool()`
#   will work with any LLM provider; `answer_using_tools()` can convert
#   the two types of tool definitions to each other when needed
\dontrun{
  ollama <- llm_provider_ollama()
  # Ollama LLM provider:
  "What is the weather in Amsterdam? Give me Fahrenheit degrees" |>
    answer_using_tools(temperature_in_location) |>
    send_prompt(ollama)

  # Ollama LLM provider also works with `ellmer::tool()` definitions:
  "What is the weather in Amsterdam? Give me Celcius degrees" |>
    answer_using_tools(temperature_in_location_ellmer) |>
    send_prompt(ollama)

  # Similar for OpenAI API:
  openai <- llm_provider_openai()
  "What is the weather in Amsterdam? Give me Celcius degrees" |>
    answer_using_tools(temperature_in_location) |>
    send_prompt(openai)
  # ...

  # Ellmer LLM provider:
  ellmer <- llm_provider_ellmer(ellmer::chat_openai())
  "What is the weather in Amsterdam? Give me Celcius degrees" |>
    answer_using_tools(temperature_in_location_ellmer) |>
    send_prompt(ellmer)

  # Also works with `tools_add_docs()` definition:
  "What is the weather in Amsterdam? Give me Celcius degrees" |>
    answer_using_tools(temperature_in_location) |>
    send_prompt(ellmer)
}
}
\seealso{
Other tools: 
\code{\link{answer_using_tools}()},
\code{\link{tools_get_docs}()}
}
\concept{tools}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/answer_using_r.R
\name{answer_using_r}
\alias{answer_using_r}
\title{Enable LLM to draft and execute R code}
\usage{
answer_using_r(
  prompt,
  add_text = "You must code in the programming language 'R' to answer this prompt.",
  pkgs_to_use = c(),
  objects_to_use = list(),
  list_packages = TRUE,
  list_objects = TRUE,
  skim_dataframes = TRUE,
  evaluate_code = FALSE,
  r_session_options = list(),
  output_as_tool = FALSE,
  return_mode = c("full", "code", "console", "object", "formatted_output", "llm_answer")
)
}
\arguments{
\item{prompt}{A single string or a \code{\link[=tidyprompt]{tidyprompt()}} object}

\item{add_text}{Single string which will be added to the prompt text,
informing the LLM that they must code in R to answer the prompt}

\item{pkgs_to_use}{A character vector of package names that may be used
in the R code that the LLM will generate. If evaluating the R code, these
packages will be pre-loaded in the R session}

\item{objects_to_use}{A named list of objects that may be used in the R code
that the LLM will generate. If evaluating the R code, these objects will be pre-loaded
in the R session. The names of the list will be used as the object names in the
R session}

\item{list_packages}{Logical indicating whether the LLM should be informed
about the packages that may be used in their R code (if TRUE, a list of the
loaded packages will be shown in the initial prompt)}

\item{list_objects}{Logical indicating whether the LLM should be informed
about the existence of 'objects_to_use' (if TRUE, a list of the objects
plus their types will be shown in the initial prompt)}

\item{skim_dataframes}{Logical indicating whether the LLM should be informed
about the structure of dataframes present in 'objects_to_use' (if TRUE,
a skim summary of each \code{data.frame} type object will be shown in the initial prompt).
This uses the function \code{\link[=skim_with_labels_and_levels]{skim_with_labels_and_levels()}}}

\item{evaluate_code}{Logical indicating whether the R code should be
evaluated. If TRUE, the R code will be evaluated in a separate R session
(using 'callr' to create an isolated R session via \link[callr]{r_session}).
Note that setting this to 'TRUE' means that code generated by the LLM will
run on your system; use this setting with caution}

\item{r_session_options}{A list of options to pass to the \link[callr]{r_session}.
This can be used to customize the R session. See \link[callr]{r_session_options}
for the available options. If no options are provided, the default options
will be used but with 'system_profile' and 'user_profile' set to FALSE}

\item{output_as_tool}{Logical indicating whether the console output of the
evaluated R code should be sent back to the LLM, meaning the LLM will use
R code as a tool to formulate a final answer to the prompt. If TRUE, the LLM
can decide if they can answer the prompt with the output, or if they need to modify
their R code. Once the LLM does not provide new R code (i.e., the prompt is being answered)
this prompt wrap will end (it will continue for as long as the LLM provides R code).
When this option is enabled, the resulting \code{\link[=prompt_wrap]{prompt_wrap()}} will be of type 'tool'.
If TRUE, the return mode will also always be set to 'llm_answer'}

\item{return_mode}{Single string indicating the return mode. One of:
\itemize{
\item 'full': Return a list with the final LLM answer, the extracted R code,
and (if argument 'evaluate_code' is TRUE) the output of the R code
\item 'code': Return only the extracted R code
\item 'console': Return only the console output of the evaluated R code
\item 'object': Return only the object produced by the evaluated R code
\item 'formatted_output': Return a formatted string with the extracted R code
and its console output, and a print of the last object (this is identical to how it would
be presented to the LLM if 'output_as_tool' is TRUE)
\item 'llm_answer': Return only the final LLM answer
}
When choosing 'console' or 'object', an additional instruction will be added to
the prompt text to inform the LLM about the expected output of the R code.
If 'output_as_tool' is TRUE, the return mode will always be set to 'llm_answer'
(as the LLM will be using the R code as a tool to answer the prompt)}
}
\value{
A \code{\link[=tidyprompt]{tidyprompt()}} object with the \code{\link[=prompt_wrap]{prompt_wrap()}} added to it, which
will handle R code generation and possibly evaluation
}
\description{
This function adds a prompt wrap to a \code{\link[=tidyprompt]{tidyprompt()}} that instructs the
LLM to answer the prompt with R code. There are various options to customize
the behavior of this prompt wrap, concerning the evaluation of the R code,
the packages that may be used, the objects that already exist in the R
session, and if the console output that should be sent back to the LLM.
}
\details{
For the evaluation of the R code, the 'callr' package is required.
Please note: automatic evaluation of generated R code may be dangerous to your
system; you must use this function with caution.
}
\examples{
\dontrun{
  # Prompt to value calculated with R
  avg_miles_per_gallon <- paste0(
    "Using my data,",
    " calculate the average miles per gallon (mpg) for cars with 6 cylinders."
  ) |>
    answer_as_integer() |>
    answer_using_r(
      pkgs_to_use = c("dplyr"),
      objects_to_use = list(mtcars = mtcars),
      evaluate_code = TRUE,
      output_as_tool = TRUE
    ) |>
    send_prompt()
  avg_miles_per_gallon

  # Prompt to linear model object in R
  model <- paste0(
    "Using my data, create a statistical model",
    " investigating the relationship between two variables."
  ) |>
    answer_using_r(
      objects_to_use = list(data = mtcars),
      evaluate_code = TRUE,
      return_mode = "object"
    ) |>
    prompt_wrap(
      validation_fn = function(x) {
        if (!inherits(x, "lm"))
          return(llm_feedback("The output should be a linear model object."))
        return(x)
      }
    ) |>
    send_prompt()
  summary(model)

  # Prompt to plot object in R
  plot <- paste0(
    "Create a scatter plot of miles per gallon (mpg) versus",
    " horsepower (hp) for the cars in my data.",
    " Use different colors to represent the number of cylinders (cyl).",
    " Be very creative and make the plot look nice but also a little crazy!"
  ) |>
    answer_using_r(
      pkgs_to_use = c("ggplot2"),
      objects_to_use = list(mtcars = mtcars),
      evaluate_code = TRUE,
      return_mode = "object"
    ) |>
    send_prompt()
  plot
}
}
\seealso{
\code{\link[=answer_using_tools]{answer_using_tools()}}

Other pre_built_prompt_wraps: 
\code{\link{add_image}()},
\code{\link{add_text}()},
\code{\link{answer_as_boolean}()},
\code{\link{answer_as_category}()},
\code{\link{answer_as_dataframe}()},
\code{\link{answer_as_integer}()},
\code{\link{answer_as_json}()},
\code{\link{answer_as_list}()},
\code{\link{answer_as_multi_category}()},
\code{\link{answer_as_named_list}()},
\code{\link{answer_as_numeric}()},
\code{\link{answer_as_regex_match}()},
\code{\link{answer_as_text}()},
\code{\link{answer_by_chain_of_thought}()},
\code{\link{answer_by_react}()},
\code{\link{answer_using_sql}()},
\code{\link{answer_using_tools}()},
\code{\link{prompt_wrap}()},
\code{\link{quit_if}()},
\code{\link{set_system_prompt}()}

Other answer_using_prompt_wraps: 
\code{\link{answer_using_sql}()},
\code{\link{answer_using_tools}()}
}
\concept{answer_using_prompt_wraps}
\concept{pre_built_prompt_wraps}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/misc_add_image.R
\name{add_image}
\alias{add_image}
\title{Add an image to a tidyprompt (multimodal)}
\usage{
add_image(
  prompt,
  image,
  alt = NULL,
  detail = c("auto", "low", "high"),
  mime = NULL
)
}
\arguments{
\item{prompt}{A single string or a \code{\link[=tidyprompt]{tidyprompt()}} object}

\item{image}{An image reference. One of:
\itemize{
\item a local file path (e.g., "path/to/image.png")
\item a URL (e.g., "https://.../image.jpg")
\item a base64 string (optionally with data URL prefix)
\item a raw vector of bytes
\item a plot object (e.g., base \code{recordedplot}, \code{ggplot}, or grid grob) to be
rasterized automatically
\item an 'ellmer' content object created by \code{ellmer::content_image_url()},
\code{ellmer::content_image_file()}, or \code{ellmer::content_image_plot()}
(this will work with both regular providers and 'ellmer'-backed providers)
}

For OpenAI Responses API, URLs must point directly to an image resource (not an HTML
page) and are transmitted as a scalar string \code{image_url} with optional \code{detail}.
Supplying a webpage URL (e.g. a Wikipedia media viewer link) will result in a
provider 400 error expecting an image URL string}

\item{alt}{Optional alternative text/alt description}

\item{detail}{Detail hint for some providers (OpenAI): one of "auto", "low",
"high"}

\item{mime}{Optional mime-type if providing raw/base64 without data URL
(e.g., "image/png")}
}
\value{
A \code{\link[=tidyprompt]{tidyprompt()}} with an added \code{\link[=prompt_wrap]{prompt_wrap()}} which will
attach an image to the prompt for use with multimodal LLMs
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}}

Attach an image to a \code{\link[=tidyprompt]{tidyprompt()}} for use with multimodal LLMs.

Supports 'ollama', 'openai' (completions & responses) and 'ellmer'-backed providers.
Can convert from and to 'ellmer' content image objects as needed.
}
\examples{
# Create a prompt with a remote image (web URL)
image_prompt <- "What is shown in this image?" |>
  add_image("https://upload.wikimedia.org/wikipedia/commons/3/3a/Cat03.jpg")

# Create a prompt with a local image (file path)
# First save an image to a temporary file
cat_img_file <- tempfile(fileext = ".jpg")
download.file(
  "https://upload.wikimedia.org/wikipedia/commons/3/3a/Cat03.jpg",
  destfile = cat_img_file,
  mode = "wb"
)
# Then build prompt with local image
local_image_prompt <- "What is shown in this image?" |>
  add_image(cat_img_file)

# Send prompt to different LLM providers
# (example is not run because it requires configured LLM providers)
\dontrun{
  # OpenAI-compatible
  send_prompt(image_prompt, llm_provider_openai(parameters = list(model = "gpt-4o-mini")))
  # --- Sending request to LLM provider (gpt-4o-mini): ---
  # What is shown in this image?
  # --- Receiving response from LLM provider: ---
  # The image shows a close-up of an orange tabby cat, characterized by its
  # striped fur and distinctive golden eyes. The background appears blurred,
  # suggesting a softly focused environment.

  # Ollama-compatible
  send_prompt(image_prompt, llm_provider_ollama(parameters = list(model = "qwen3-vl:2b")))
  # ...

  # 'ellmer'-compatible
  send_prompt(image_prompt, llm_provider_ellmer(ellmer::chat_openai(model = "gpt-4o-mini")))
  # ...
}

# Create a prompt with a plot (e.g., 'ggplot2' plot)
\dontrun{
if (requireNamespace("ggplot2", quietly = TRUE)) {
  plot <- ggplot2::ggplot(mtcars, ggplot2::aes(mpg, disp)) +
    ggplot2::geom_point()

  plot_prompt <- "Describe this plot" |>
    add_image(plot)

  send_prompt(plot_prompt, llm_provider_openai())
  # --- Sending request to LLM provider (gpt-4o-mini): ---
  # Describe this plot
  # --- Receiving response from LLM provider: ---
  # The plot is a scatter plot depicting the relationship between two variables:
  # "mpg" (miles per gallon) on the x-axis and "disp" (displacement) on
  # the y-axis. (...)
}
}
}
\seealso{
Other pre_built_prompt_wraps: 
\code{\link{add_text}()},
\code{\link{answer_as_boolean}()},
\code{\link{answer_as_category}()},
\code{\link{answer_as_dataframe}()},
\code{\link{answer_as_integer}()},
\code{\link{answer_as_json}()},
\code{\link{answer_as_list}()},
\code{\link{answer_as_multi_category}()},
\code{\link{answer_as_named_list}()},
\code{\link{answer_as_numeric}()},
\code{\link{answer_as_regex_match}()},
\code{\link{answer_as_text}()},
\code{\link{answer_by_chain_of_thought}()},
\code{\link{answer_by_react}()},
\code{\link{answer_using_r}()},
\code{\link{answer_using_sql}()},
\code{\link{answer_using_tools}()},
\code{\link{prompt_wrap}()},
\code{\link{quit_if}()},
\code{\link{set_system_prompt}()}

Other miscellaneous_prompt_wraps: 
\code{\link{add_text}()},
\code{\link{quit_if}()},
\code{\link{set_system_prompt}()}
}
\concept{miscellaneous_prompt_wraps}
\concept{pre_built_prompt_wraps}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/send_prompt.R
\name{send_prompt}
\alias{send_prompt}
\title{Send a prompt to a LLM provider}
\usage{
send_prompt(
  prompt,
  llm_provider = llm_provider_ollama(),
  max_interactions = 10,
  clean_chat_history = FALSE,
  verbose = NULL,
  stream = NULL,
  return_mode = c("only_response", "full")
)
}
\arguments{
\item{prompt}{A string or a \linkS4class{tidyprompt} object}

\item{llm_provider}{\linkS4class{llm_provider} object
(default is \code{\link[=llm_provider_ollama]{llm_provider_ollama()}}).
This object and its settings will be used to evaluate the prompt.
You may also pass an \code{ellmer::chat()} object directly; in that case,
\code{\link[=send_prompt]{send_prompt()}} will clone it, clear any existing turns, and wrap the clean
clone with \code{\link[=llm_provider_ellmer]{llm_provider_ellmer()}} before evaluation.
When using \code{\link[=llm_provider_ellmer]{llm_provider_ellmer()}} or a raw \code{ellmer::chat()}, the working
\code{ellmer} turns are rebuilt from tidyprompt's own chat history for each call.
Note that the 'verbose' and 'stream' settings in the LLM provider will be
overruled by the 'verbose' and 'stream' arguments in this function
when those are not NULL.
Furthermore, advanced \linkS4class{tidyprompt} objects may carry '$parameter_fn'
functions which can set parameters in the llm_provider object
(see \code{\link[=prompt_wrap]{prompt_wrap()}} and \linkS4class{llm_provider} for more ).}

\item{max_interactions}{Maximum number of interactions allowed with the
LLM provider. Default is 10. If the maximum number of interactions is reached
without a successful response, 'NULL' is returned as the response (see return
value). The first interaction is the initial chat completion}

\item{clean_chat_history}{If the chat history should be cleaned after each
interaction. Cleaning the chat history means that only the
first and last message from the user, the last message from the assistant,
all messages from the system, and all tool results are kept in a 'clean'
chat history. This clean chat history is used when requesting a new chat completion.
Rows marked as non-replayable are excluded from new requests regardless of
this setting, so the returned transcript may contain more rows than the model
actually sees on a retry or follow-up call.
(i.e., if a LLM repeatedly fails to provide a correct response, only its last failed response
will included in the context window). This may increase the LLM performance
on the next interaction}

\item{verbose}{If the interaction with the LLM provider should be printed
to the console. This will overrule the 'verbose' setting in the LLM provider}

\item{stream}{If the interaction with the LLM provider should be streamed.
This setting will only be used if the LLM provider already has a
'stream' parameter (which indicates there is support for streaming). Note
that when 'verbose' is set to FALSE, the 'stream' setting will be ignored}

\item{return_mode}{One of 'full' or 'only_response'. See return value}
}
\value{
\itemize{
\item If return mode 'only_response', the function will return only the LLM response
after extraction and validation functions have been applied (NULL is returned
when unsuccessful after the maximum number of interactions).
\item If return mode 'full', the function will return a list with the following elements:
\itemize{
\item 'response' (the LLM response after extraction and validation functions have been applied;
NULL is returned when unsuccessful after the maximum number of interactions),
\item 'interactions' (the number of interactions with the LLM provider),
\item 'chat_history' (a dataframe with the full chat history which led to the final response.
This may include rows that are retained for inspection but not re-sent to the model;
such rows are marked with column \code{hidden_from_llm = TRUE}),
\item 'chat_history_clean' (a dataframe with the cleaned chat history which led to
the final response; here, only the first and last message from the user, the
last message from the assistant, and all messages from the system are kept,
after excluding any non-replayable rows),
\item 'start_time' (the time when the function was called),
\item 'end_time' (the time when the function ended),
\item 'duration_seconds' (the duration of the function in seconds),
\item 'http' (a list with all HTTP requests and responses made during the interactions;
as returned by \code{llm_provider$complete_chat()}),
\item 'ellmer_chat' (if \code{\link[=llm_provider_ellmer]{llm_provider_ellmer()}} or a raw \code{ellmer::chat()}
object was used, this will be
the updated 'ellmer' chat object, containing for instance
the turns and possible tool calls. (As this function
uses a clone of the provided LLM provider, the 'ellmer' chat object in the
LLM provider will not be updated; use this return value when you need the
updated native \code{ellmer} turns from the current evaluation. Note that turns in the 'ellmer'
chat object may not contain the full chat history when \code{clean_chat_history = TRUE}
was used.)
}
}
}
\description{
This function is responsible for sending prompts to a LLM provider for evaluation.
The function will interact with the LLM provider until a successful response
is received or the maximum number of interactions is reached. The function will
apply extraction and validation functions to the LLM response, as specified
in the prompt wraps (see \code{\link[=prompt_wrap]{prompt_wrap()}}). If the maximum number of interactions
}
\examples{
\dontrun{
  "Hi!" |>
    send_prompt(llm_provider_ollama())
  # --- Sending request to LLM provider (llama3.1:8b): ---
  #   Hi!
  # --- Receiving response from LLM provider: ---
  #   It's nice to meet you. Is there something I can help you with, or would you like to chat?
  # [1] "It's nice to meet you. Is there something I can help you with, or would you like to chat?"

  "Hi!" |>
    send_prompt(llm_provider_ollama(), return_mode = "full")
  # --- Sending request to LLM provider (llama3.1:8b): ---
  #   Hi!
  # --- Receiving response from LLM provider: ---
  #   It's nice to meet you. Is there something I can help you with, or would you like to chat?
  # $response
  # [1] "It's nice to meet you. Is there something I can help you with, or would you like to chat?"
  #
  # $chat_history
  # ...
  #
  # $chat_history_clean
  # ...
  #
  # $start_time
  # [1] "2024-11-18 15:43:12 CET"
  #
  # $end_time
  # [1] "2024-11-18 15:43:13 CET"
  #
  # $duration_seconds
  # [1] 1.13276
  #
  # $http
  # $http$responses[[1]]
  # Response [http://localhost:11434/api/chat]
  #   Date: 2024-11-18 14:43
  #   Status: 200
  #   Content-Type: application/x-ndjson
  # <EMPTY BODY>

  "Hi!" |>
    add_text("What is 5 + 5?") |>
    answer_as_integer() |>
    send_prompt(llm_provider_ollama(), verbose = FALSE)
  # [1] 10
}
}
\seealso{
\linkS4class{tidyprompt}, \code{\link[=prompt_wrap]{prompt_wrap()}}, \linkS4class{llm_provider}, \code{\link[=llm_provider_ollama]{llm_provider_ollama()}},
\code{\link[=llm_provider_openai]{llm_provider_openai()}}

Other prompt_evaluation: 
\code{\link{llm_break}()},
\code{\link{llm_feedback}()}
}
\concept{prompt_evaluation}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tidyprompt.R
\name{tidyprompt-class}
\alias{tidyprompt-class}
\title{Tidyprompt R6 Class}
\description{
A \linkS4class{tidyprompt} object contains a base prompt and a list
of \code{\link[=prompt_wrap]{prompt_wrap()}} objects. It provides structured methods to modify the prompt
while simultaneously adding logic to extract from and validate the LLM response.
Besides a base prompt, a \linkS4class{tidyprompt} object may contain a system prompt
and a chat history which precede the base prompt.
}
\examples{
prompt <- tidyprompt("Hi!")
print(prompt)

# Add to a tidyprompt using a prompt wrap:
prompt <- tidyprompt("Hi!") |>
  add_text("How are you?")
print(prompt)

# Strings can be input for prompt wraps; therefore,
#   a call to tidyprompt() is not necessary:
prompt <- "Hi" |>
  add_text("How are you?")

# Example of adding extraction & validation with a prompt_wrap():
prompt <- "Hi" |>
  add_text("What is 5 + 5?") |>
  answer_as_integer()

\dontrun{
  # tidyprompt objects are evaluated by send_prompt(), which will
  #   handle construct the prompt text, send it to the LLM provider,
  #   and apply the extraction and validation functions from the tidyprompt object
  prompt |>
    send_prompt(llm_provider_ollama())
  # --- Sending request to LLM provider (llama3.1:8b): ---
  #   Hi
  #
  #   What is 5 + 5?
  #
  #   You must answer with only an integer (use no other characters).
  # --- Receiving response from LLM provider: ---
  #   10
  # [1] 10

  # See prompt_wrap() and send_prompt() for more details
}

# `tidyprompt` objects may be validated with these helpers:
is_tidyprompt(prompt) # Returns TRUE if input is a valid tidyprompt object

# Get base prompt text
base_prompt <- prompt$base_prompt

# Get all prompt wraps
prompt_wraps <- prompt$get_prompt_wraps()
# Alternative:
prompt_wraps <- get_prompt_wraps(prompt)

# Construct prompt text
prompt_text <- prompt$construct_prompt_text()
# Alternative:
prompt_text <- construct_prompt_text(prompt)

# Set chat history (affecting also the base prompt)
chat_history <- data.frame(
  role = c("user", "assistant", "user"),
  content = c("What is 5 + 5?", "10", "And what is 5 + 6?")
)
prompt$set_chat_history(chat_history)

# Get chat history
chat_history <- prompt$get_chat_history()
}
\seealso{
Other tidyprompt: 
\code{\link{construct_prompt_text}()},
\code{\link{get_chat_history}()},
\code{\link{get_prompt_wraps}()},
\code{\link{is_tidyprompt}()},
\code{\link{set_chat_history}()},
\code{\link{tidyprompt}()}
}
\concept{tidyprompt}
\section{Public fields}{
\if{html}{\out{<div class="r6-fields">}}
\describe{
\item{\code{base_prompt}}{The base prompt string.
The base prompt be modified by prompt wraps during \code{\link[=construct_prompt_text]{construct_prompt_text()}};
the modified prompt text will be used as the final message of role 'user'
during \code{\link[=send_prompt]{send_prompt()}}}

\item{\code{system_prompt}}{A system prompt string.
This will be added at the start of the chat history as role 'system'
during \code{\link[=send_prompt]{send_prompt()}}}
}
\if{html}{\out{</div>}}
}
\section{Methods}{
\subsection{Public methods}{
\itemize{
\item \href{#method-Tidyprompt-new}{\code{tidyprompt-class$new()}}
\item \href{#method-Tidyprompt-is_valid}{\code{tidyprompt-class$is_valid()}}
\item \href{#method-Tidyprompt-add_prompt_wrap}{\code{tidyprompt-class$add_prompt_wrap()}}
\item \href{#method-Tidyprompt-get_prompt_wraps}{\code{tidyprompt-class$get_prompt_wraps()}}
\item \href{#method-Tidyprompt-construct_prompt_text}{\code{tidyprompt-class$construct_prompt_text()}}
\item \href{#method-Tidyprompt-set_chat_history}{\code{tidyprompt-class$set_chat_history()}}
\item \href{#method-Tidyprompt-get_chat_history}{\code{tidyprompt-class$get_chat_history()}}
\item \href{#method-Tidyprompt-clone}{\code{tidyprompt-class$clone()}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Tidyprompt-new"></a>}}
\if{latex}{\out{\hypertarget{method-Tidyprompt-new}{}}}
\subsection{Method \code{new()}}{
Initialize a \linkS4class{tidyprompt} object
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{tidyprompt-class$new(input)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{input}}{A string, a chat history, a list containing
a chat history under key '$chat_history', or a \linkS4class{tidyprompt} object}
}
\if{html}{\out{</div>}}
}
\subsection{Details}{
Different types of input are accepted for initialization of
a \linkS4class{tidyprompt} object:
\itemize{
\item A single character string. This will be used as the base prompt

\item A dataframe which is a valid chat history (see \code{\link[=chat_history]{chat_history()}})

\item A list containing a valid chat history under '$chat_history'
(e.g., a result from \code{\link[=send_prompt]{send_prompt()}} when using 'return_mode' = "full")

\item A \linkS4class{tidyprompt} object. This will be checked for validity and, if valid,
the fields are copied to the object which is returned from this method
}
When passing a dataframe or list with a chat history, the last row of the
chat history must have role 'user'; this row will be used as the base prompt.
If the first row of the chat history has role 'system', it will be used
as the system prompt.
}

\subsection{Returns}{
A \linkS4class{tidyprompt} object
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Tidyprompt-is_valid"></a>}}
\if{latex}{\out{\hypertarget{method-Tidyprompt-is_valid}{}}}
\subsection{Method \code{is_valid()}}{
Check if the tidyprompt object is valid.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{tidyprompt-class$is_valid()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
\code{TRUE} if valid, otherwise \code{FALSE}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Tidyprompt-add_prompt_wrap"></a>}}
\if{latex}{\out{\hypertarget{method-Tidyprompt-add_prompt_wrap}{}}}
\subsection{Method \code{add_prompt_wrap()}}{
Add a \code{\link[=prompt_wrap]{prompt_wrap()}} to the \linkS4class{tidyprompt} object.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{tidyprompt-class$add_prompt_wrap(prompt_wrap)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{prompt_wrap}}{A \code{\link[=prompt_wrap]{prompt_wrap()}} object}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
The updated \linkS4class{tidyprompt} object
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Tidyprompt-get_prompt_wraps"></a>}}
\if{latex}{\out{\hypertarget{method-Tidyprompt-get_prompt_wraps}{}}}
\subsection{Method \code{get_prompt_wraps()}}{
Get list of \code{\link[=prompt_wrap]{prompt_wrap()}} objects from the \linkS4class{tidyprompt} object.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{tidyprompt-class$get_prompt_wraps(
  order = c("default", "modification", "evaluation")
)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{order}}{The order to return the wraps. Options are:
\itemize{
\item "default": as originally added to the object
\item "modification": as ordered for modification of the base prompt;
ordered by type: check, unspecified, mode, tool, break. This is the order
in which prompt wraps are applied during \code{\link[=construct_prompt_text]{construct_prompt_text()}}
\item "evaluation": ordered for evaluation of the LLM response;
ordered by type: tool, mode, break, unspecified, check. This is the order
in which wraps are applied to the LLM output during \code{\link[=send_prompt]{send_prompt()}}
}}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
A list of \code{\link[=prompt_wrap]{prompt_wrap()}} objects.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Tidyprompt-construct_prompt_text"></a>}}
\if{latex}{\out{\hypertarget{method-Tidyprompt-construct_prompt_text}{}}}
\subsection{Method \code{construct_prompt_text()}}{
Construct the complete prompt text.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{tidyprompt-class$construct_prompt_text(
  llm_provider = NULL,
  apply_provider_prompt_wraps = FALSE
)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{llm_provider}}{Optional \linkS4class{llm_provider} object.
This may sometimes affect the prompt text construction}

\item{\code{apply_provider_prompt_wraps}}{Logical. Whether to apply
provider-specific pre/post prompt wraps when constructing the prompt text}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
A string representing the constructed prompt text
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Tidyprompt-set_chat_history"></a>}}
\if{latex}{\out{\hypertarget{method-Tidyprompt-set_chat_history}{}}}
\subsection{Method \code{set_chat_history()}}{
This function sets the chat history for the tidyprompt object.
The chat history will also set the base prompt and system prompt
(the last message of the chat history should be of role 'user' and
will be used as the base prompt; the first message of the chat history
may be of the role 'system' and will then be used as the system prompt).
This may be useful when one wants to change the base prompt, system prompt,
and chat history of a \linkS4class{tidyprompt} object while retaining other fields like
the prompt wraps.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{tidyprompt-class$set_chat_history(chat_history)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{chat_history}}{A valid chat history (see \code{\link[=chat_history]{chat_history()}})}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
The updated \linkS4class{tidyprompt} object
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Tidyprompt-get_chat_history"></a>}}
\if{latex}{\out{\hypertarget{method-Tidyprompt-get_chat_history}{}}}
\subsection{Method \code{get_chat_history()}}{
This function gets the chat history of the \linkS4class{tidyprompt} object.
The chat history is constructed from the base prompt, system prompt, and chat
history field. The returned object will be the chat history
with the system prompt as the first message with role 'system' and the
the base prompt as the last message with role 'user'.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{tidyprompt-class$get_chat_history(llm_provider = NULL)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{llm_provider}}{An optional \linkS4class{llm_provider} object.
This may sometimes affect the prompt text construction}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
A dataframe containing the chat history
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Tidyprompt-clone"></a>}}
\if{latex}{\out{\hypertarget{method-Tidyprompt-clone}{}}}
\subsection{Method \code{clone()}}{
The objects of this class are cloneable with this method.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{tidyprompt-class$clone(deep = FALSE)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{deep}}{Whether to make a deep clone.}
}
\if{html}{\out{</div>}}
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/answer_as_named_list.R
\name{answer_as_named_list}
\alias{answer_as_named_list}
\title{Make LLM answer as a named list}
\usage{
answer_as_named_list(
  prompt,
  item_names,
  item_instructions = NULL,
  item_validations = NULL
)
}
\arguments{
\item{prompt}{A single string or a \code{\link[=tidyprompt]{tidyprompt()}} object}

\item{item_names}{A character vector specifying the expected item names}

\item{item_instructions}{An optional named list of additional instructions for each item}

\item{item_validations}{An optional named list of validation functions for each item.
Like validation functions for a \code{\link[=prompt_wrap]{prompt_wrap()}}, these functions should return
\code{\link[=llm_feedback]{llm_feedback()}} if the validation fails. If the validation
is successful, the function should return TRUE}
}
\value{
A \code{\link[=tidyprompt]{tidyprompt()}} with an added \code{\link[=prompt_wrap]{prompt_wrap()}} that ensures
the LLM response is a named list with the specified item names, optional
instructions, and validations.
}
\description{
Get a named list from LLM response with optional item instructions and validations.
}
\examples{
\dontrun{
  persona <- "Create a persona for me, please." |>
    answer_as_named_list(
      item_names = c("name", "age", "occupation"),
      item_instructions = list(
        name = "The name of the persona",
        age = "The age of the persona",
        occupation = "The occupation of the persona"
      )
    ) |> send_prompt(llm_provider_ollama())
  # --- Sending request to LLM provider (llama3.1:8b): ---
  #   Create a persona for me, please.
  #
  #   Respond with a named list like so:
  #     -- name: <<value>> (The name of the persona)
  #     -- age: <<value>> (The age of the persona)
  #     -- occupation: <<value>> (The occupation of the persona)
  #   Each name must correspond to: name, age, occupation
  # --- Receiving response from LLM provider: ---
  #   Here is your persona:
  #
  #   -- name: Astrid Welles
  #   -- age: 32
  #   -- occupation: Museum Curator
  persona$name
  # [1] "Astrid Welles"
  persona$age
  # [1] "32"
  persona$occupation
  # [1] "Museum Curator"
}
}
\seealso{
\code{\link[=answer_as_list]{answer_as_list()}} \code{\link[=llm_feedback]{llm_feedback()}}

Other pre_built_prompt_wraps: 
\code{\link{add_image}()},
\code{\link{add_text}()},
\code{\link{answer_as_boolean}()},
\code{\link{answer_as_category}()},
\code{\link{answer_as_dataframe}()},
\code{\link{answer_as_integer}()},
\code{\link{answer_as_json}()},
\code{\link{answer_as_list}()},
\code{\link{answer_as_multi_category}()},
\code{\link{answer_as_numeric}()},
\code{\link{answer_as_regex_match}()},
\code{\link{answer_as_text}()},
\code{\link{answer_by_chain_of_thought}()},
\code{\link{answer_by_react}()},
\code{\link{answer_using_r}()},
\code{\link{answer_using_sql}()},
\code{\link{answer_using_tools}()},
\code{\link{prompt_wrap}()},
\code{\link{quit_if}()},
\code{\link{set_system_prompt}()}

Other answer_as_prompt_wraps: 
\code{\link{answer_as_boolean}()},
\code{\link{answer_as_category}()},
\code{\link{answer_as_dataframe}()},
\code{\link{answer_as_integer}()},
\code{\link{answer_as_json}()},
\code{\link{answer_as_list}()},
\code{\link{answer_as_multi_category}()},
\code{\link{answer_as_numeric}()},
\code{\link{answer_as_regex_match}()},
\code{\link{answer_as_text}()}
}
\concept{answer_as_prompt_wraps}
\concept{pre_built_prompt_wraps}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/persistent_chat.R
\name{persistent_chat-class}
\alias{persistent_chat-class}
\title{PersistentChat R6 class}
\description{
A class for managing a persistent chat with a large language model (LLM).

While 'tidyprompt' is primariy focused on automatic interactions with
LLMs through \code{\link[=send_prompt]{send_prompt()}} using a \linkS4class{tidyprompt} object with
\code{\link[=prompt_wrap]{prompt_wrap()}}, this class may be useful for having a manual conversation
with an LLM. (It may specifically be used to continue a chat history which was
returned by \code{\link[=send_prompt]{send_prompt()}} with \code{return_mode = "full"}.)
}
\examples{
# Create a persistent chat with any LLM provider
chat <- `persistent_chat-class`$new(llm_provider_ollama())

\dontrun{
  chat$chat("Hi! Tell me about Twente, in a short sentence?")
  # --- Sending request to LLM provider (llama3.1:8b): ---
  # Hi! Tell me about Twente, in a short sentence?
  # --- Receiving response from LLM provider: ---
  # Twente is a charming region in the Netherlands known for its picturesque
  # countryside and vibrant culture!

  chat$chat("How many people live there?")
  # --- Sending request to LLM provider (llama3.1:8b): ---
  # How many people live there?
  # --- Receiving response from LLM provider: ---
  # The population of Twente is approximately 650,000 inhabitants, making it one of
  # the largest regions in the Netherlands.

  # Access the chat history:
  chat$chat_history

  # Reset the chat history:
  chat$reset_chat_history()

  # Continue a chat from the result of `send_prompt()`:
  result <- "Hi there!" |>
    answer_as_integer() |>
    send_prompt(return_mode = "full")
  # --- Sending request to LLM provider (llama3.1:8b): ---
  # Hi there!
  #
  # You must answer with only an integer (use no other characters).
  # --- Receiving response from LLM provider: ---
  # 42
  chat <- `persistent_chat-class`$new(llm_provider_ollama(), result$chat_history)
  chat$chat("Why did you choose that number?")
  # --- Sending request to LLM provider (llama3.1:8b): ---
  # Why did you choose that number?
  # --- Receiving response from LLM provider: ---
  # I chose the number 42 because it's a reference to Douglas Adams' science fiction
  # series "The Hitchhiker's Guide to the Galaxy," in which a supercomputer named
  # Deep Thought is said to have calculated the "Answer to the Ultimate Question of
  # Life, the Universe, and Everything" as 42.
}

}
\seealso{
\linkS4class{llm_provider} \code{\link[=chat_history]{chat_history()}}
}
\section{Public fields}{
\if{html}{\out{<div class="r6-fields">}}
\describe{
\item{\code{chat_history}}{A \code{\link[=chat_history]{chat_history()}} object}

\item{\code{llm_provider}}{A \linkS4class{llm_provider} object}
}
\if{html}{\out{</div>}}
}
\section{Methods}{
\subsection{Public methods}{
\itemize{
\item \href{#method-PersistentChat-new}{\code{persistent_chat-class$new()}}
\item \href{#method-PersistentChat-chat}{\code{persistent_chat-class$chat()}}
\item \href{#method-PersistentChat-reset_chat_history}{\code{persistent_chat-class$reset_chat_history()}}
\item \href{#method-PersistentChat-clone}{\code{persistent_chat-class$clone()}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-PersistentChat-new"></a>}}
\if{latex}{\out{\hypertarget{method-PersistentChat-new}{}}}
\subsection{Method \code{new()}}{
Initialize the PersistentChat object
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{persistent_chat-class$new(llm_provider, chat_history = NULL)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{llm_provider}}{A \linkS4class{llm_provider} object}

\item{\code{chat_history}}{(optional) A \code{\link[=chat_history]{chat_history()}} object}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
The initialized PersistentChat object
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-PersistentChat-chat"></a>}}
\if{latex}{\out{\hypertarget{method-PersistentChat-chat}{}}}
\subsection{Method \code{chat()}}{
Add a message to the chat history and get a response from the LLM
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{persistent_chat-class$chat(msg, role = "user", verbose = TRUE)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{msg}}{Message to add to the chat history}

\item{\code{role}}{Role of the message}

\item{\code{verbose}}{Whether to print the interaction to the console}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
The response from the LLM
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-PersistentChat-reset_chat_history"></a>}}
\if{latex}{\out{\hypertarget{method-PersistentChat-reset_chat_history}{}}}
\subsection{Method \code{reset_chat_history()}}{
Reset the chat history
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{persistent_chat-class$reset_chat_history()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
NULL
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-PersistentChat-clone"></a>}}
\if{latex}{\out{\hypertarget{method-PersistentChat-clone}{}}}
\subsection{Method \code{clone()}}{
The objects of this class are cloneable with this method.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{persistent_chat-class$clone(deep = FALSE)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{deep}}{Whether to make a deep clone.}
}
\if{html}{\out{</div>}}
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/chat_history.R
\name{chat_history.character}
\alias{chat_history.character}
\title{Method for \code{chat_history()} when the input is a single string}
\usage{
\method{chat_history}{character}(chat_history)
}
\arguments{
\item{chat_history}{A single string}
}
\value{
A valid chat history \code{data.frame} (of class \code{chat_history}),
with the 'role' set to 'user' and the 'content' set to the input string
}
\description{
Creates a \code{chat_history} object from a single string.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tidyprompt-package.R
\docType{package}
\name{tidyprompt-package}
\alias{tidyprompt-package}
\title{tidyprompt: Prompt Large Language Models and Enhance Their Functionality}
\description{
Easily construct prompts and associated logic for interacting with large language models (LLMs). 'tidyprompt' introduces the concept of prompt wraps, which are building blocks that you can use to quickly turn a simple prompt into a complex one. Prompt wraps do not just modify the prompt text, but also add extraction and validation functions that will be applied to the response of the LLM. This ensures that the user gets the desired output. 'tidyprompt' can add various features to prompts and their evaluation by LLMs, such as structured output, automatic feedback, retries, reasoning modes, autonomous R function calling, and R code generation and evaluation. It is designed to be compatible with any LLM provider that offers chat completion.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://github.com/KennispuntTwente/tidyprompt}
  \item \url{https://KennispuntTwente.github.io/tidyprompt/}
  \item Report bugs at \url{https://github.com/KennispuntTwente/tidyprompt/issues}
}

}
\author{
\strong{Maintainer}: Luka Koning \email{koningluka@gmail.com} [copyright holder]

Authors:
\itemize{
  \item Tjark Van de Merwe \email{t.vandemerwe@kennispunttwente.nl} [copyright holder]
}

Other contributors:
\itemize{
  \item Kennispunt Twente \email{info@kennispunttwente.nl} [funder]
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/llm_feedback.R
\name{llm_feedback}
\alias{llm_feedback}
\title{Create an \code{llm_feedback} object}
\usage{
llm_feedback(text, tool_result = FALSE)
}
\arguments{
\item{text}{A character string containing the feedback text. This will
be sent back to the LLM after not passing an extractor or validator function}

\item{tool_result}{A logical indicating whether the feedback is a tool result.
If TRUE, \code{\link[=send_prompt]{send_prompt()}} will not remove it from the chat history when
cleaning the context window during repeated interactions}
}
\value{
An object of class "llm_feedback" (or "llm_feedback_tool_result")
containing the feedback text to send back to the LLM
}
\description{
This object is used to send feedback to a LLM when a LLM reply
does not succesfully pass an extraction or validation function
(as handled by \code{\link[=send_prompt]{send_prompt()}} and defined using \code{\link[=prompt_wrap]{prompt_wrap()}}).
The feedback text is sent back to the LLM. The extraction or validation function
should then return this object with the feedback text that should be sent to the LLM.
}
\examples{
# Example usage within a validation function similar to the one in 'answer_as_integer()':
validation_fn <- function(x, min = 0, max = 100) {
  if (x != floor(x)) {
    # Not a whole number
    return(llm_feedback(
      "You must answer with only an integer (use no other characters)."
    ))
  }
  if (!is.null(min) && x < min) {
    return(llm_feedback(glue::glue(
      "The number should be greater than or equal to {min}."
    )))
  }
  if (!is.null(max) && x > max) {
    return(llm_feedback(glue::glue(
      "The number should be less than or equal to {max}."
    )))
  }
  return(TRUE)
}

# This validation_fn would be part of a prompt_wrap();
#   see the `answer_as_integer()` function for an example of how to use it
}
\seealso{
\code{\link[=llm_break]{llm_break()}}

Other prompt_wrap: 
\code{\link{llm_break}()},
\code{\link{prompt_wrap}()}

Other prompt_evaluation: 
\code{\link{llm_break}()},
\code{\link{send_prompt}()}
}
\concept{prompt_evaluation}
\concept{prompt_wrap}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/answer_as_list.R
\name{answer_as_list}
\alias{answer_as_list}
\title{Make LLM answer as a list of items}
\usage{
answer_as_list(
  prompt,
  item_name = "item",
  item_explanation = NULL,
  n_unique_items = NULL,
  list_mode = c("bullet", "comma")
)
}
\arguments{
\item{prompt}{A single string or a \code{\link[=tidyprompt]{tidyprompt()}} object}

\item{item_name}{(optional) Name of the items in the list}

\item{item_explanation}{(optional) Additional explanation of what an item
should be. Item explanation should be a single string. It will be
appended after the list instruction}

\item{n_unique_items}{(optional) Number of unique items required in the list}

\item{list_mode}{(optional) Mode of the list. Either "bullet" or "comma".
"bullet mode expects items to be listed with "--" before each item, with a
new line for each item (e.g., "-- item1\\n-- item2\\n-- item3").
"comma" mode expects items to be listed with a number and a period before
(e.g., "1. item1, 2. item2, 3. item3"). "comma" mode may be easier for
smaller LLMs to use}
}
\value{
A \code{\link[=tidyprompt]{tidyprompt()}} with an added \code{\link[=prompt_wrap]{prompt_wrap()}} which
will ensure that the LLM response is a list of items
}
\description{
Make LLM answer as a list of items
}
\examples{
\dontrun{
  "What are some delicious fruits?" |>
    answer_as_list(item_name = "fruit", n_unique_items = 5) |>
    send_prompt()
  # --- Sending request to LLM provider (llama3.1:8b): ---
  # What are some delicious fruits?
  #
  # Respond with a list, like so:
  #   -- <<fruit 1>>
  #   -- <<fruit 2>>
  #   etc.
  # The list should contain 5 unique items.
  # --- Receiving response from LLM provider: ---
  # Here's the list of delicious fruits:
  #   -- Strawberries
  #   -- Pineapples
  #   -- Mangoes
  #   -- Blueberries
  #   -- Pomegranates
  # [[1]]
  # [1] "Strawberries"
  #
  # [[2]]
  # [1] "Pineapples"
  #
  # [[3]]
  # [1] "Mangoes"
  #
  # [[4]]
  # [1] "Blueberries"
  #
  # [[5]]
  # [1] "Pomegranates"
}

}
\seealso{
\code{\link[=answer_as_named_list]{answer_as_named_list()}}

Other pre_built_prompt_wraps: 
\code{\link{add_image}()},
\code{\link{add_text}()},
\code{\link{answer_as_boolean}()},
\code{\link{answer_as_category}()},
\code{\link{answer_as_dataframe}()},
\code{\link{answer_as_integer}()},
\code{\link{answer_as_json}()},
\code{\link{answer_as_multi_category}()},
\code{\link{answer_as_named_list}()},
\code{\link{answer_as_numeric}()},
\code{\link{answer_as_regex_match}()},
\code{\link{answer_as_text}()},
\code{\link{answer_by_chain_of_thought}()},
\code{\link{answer_by_react}()},
\code{\link{answer_using_r}()},
\code{\link{answer_using_sql}()},
\code{\link{answer_using_tools}()},
\code{\link{prompt_wrap}()},
\code{\link{quit_if}()},
\code{\link{set_system_prompt}()}

Other answer_as_prompt_wraps: 
\code{\link{answer_as_boolean}()},
\code{\link{answer_as_category}()},
\code{\link{answer_as_dataframe}()},
\code{\link{answer_as_integer}()},
\code{\link{answer_as_json}()},
\code{\link{answer_as_multi_category}()},
\code{\link{answer_as_named_list}()},
\code{\link{answer_as_numeric}()},
\code{\link{answer_as_regex_match}()},
\code{\link{answer_as_text}()}
}
\concept{answer_as_prompt_wraps}
\concept{pre_built_prompt_wraps}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/prompt_wrap.R
\name{prompt_wrap}
\alias{prompt_wrap}
\title{Wrap a prompt with functions for modification and handling the LLM response}
\usage{
prompt_wrap(
  prompt,
  modify_fn = NULL,
  extraction_fn = NULL,
  validation_fn = NULL,
  handler_fn = NULL,
  parameter_fn = NULL,
  type = c("unspecified", "mode", "tool", "break", "check"),
  name = NULL
)
}
\arguments{
\item{prompt}{A string or a \linkS4class{tidyprompt} object}

\item{modify_fn}{A function that takes the previous prompt text (as
first argument) and returns the new prompt text}

\item{extraction_fn}{A function that takes the LLM response (as first argument)
and attempts to extract a value from it.
Upon succesful extraction, the function should return the extracted value.
If the extraction fails, the function should return a \code{\link[=llm_feedback]{llm_feedback()}} message
to initiate a retry.
A \code{\link[=llm_break]{llm_break()}} can be returned to break the extraction and validation loop,
ending \code{\link[=send_prompt]{send_prompt()}}}

\item{validation_fn}{A function that takes the (extracted) LLM response
(as first argument) and attempts to validate it.
Upon succesful validation, the function should return TRUE. If the validation
fails, the function should return a \code{\link[=llm_feedback]{llm_feedback()}} message to initiate a retry.
A \code{\link[=llm_break]{llm_break()}} can be returned to break the extraction and validation loop,
ending \code{\link[=send_prompt]{send_prompt()}}}

\item{handler_fn}{A function that takes a 'completion' object (a result
of a request to a LLM, as returned by \verb{$complete_chat()} of a \linkS4class{llm_provider}
object) as first argument and the \linkS4class{llm_provider} object as second argument.
The function should return a (modified or identical) completion object.
This can be used for advanced side effects, like logging, or native tool calling,
or keeping track of token usage. See \linkS4class{llm_provider} for more information;
handler_fn is attached to the \linkS4class{llm_provider} object that is being used.
When using an \code{\link[=llm_provider_ellmer]{llm_provider_ellmer()}}, the up-to-date ellmer_chat is synced
onto the provider before handlers run. This allows handlers to access,
for instance, the current cost of the conversation, and, for instance,
to stop the conversation if a certain budget is exceeded.
For example usage, see source code of \code{\link[=answer_using_tools]{answer_using_tools()}}}

\item{parameter_fn}{A function that takes the \linkS4class{llm_provider} object
which is being used with \code{\link[=send_prompt]{send_prompt()}} and returns a named list of parameters
to be set in the \linkS4class{llm_provider} object via its \verb{$set_parameters()} method.
This can be used to configure specific parameters of the \linkS4class{llm_provider}
object when evaluating the prompt.
For example, \code{\link[=answer_as_json]{answer_as_json()}} may set different parameters for different APIs
related to JSON output.
This function is typically only used with advanced prompt wraps that require
specific settings in the \linkS4class{llm_provider} object}

\item{type}{The type of prompt wrap. Must be one of:
\itemize{
\item "unspecified": The default type, typically used for prompt wraps
which request a specific format of the LLM response, like \code{\link[=answer_as_integer]{answer_as_integer()}}

\item "mode": For prompt wraps that change how the LLM should answer the prompt,
like \code{\link[=answer_by_chain_of_thought]{answer_by_chain_of_thought()}} or \code{\link[=answer_by_react]{answer_by_react()}}

\item "tool": For prompt wraps that enable the LLM to use tools, like \code{\link[=answer_using_tools]{answer_using_tools()}}
or \code{\link[=answer_using_r]{answer_using_r()}} when 'output_as_tool' = TRUE

\item "break": For prompt wraps that may break the extraction and validation loop,
like \code{\link[=quit_if]{quit_if()}}. These are applied before type "unspecified" as they may
instruct the LLM to not answer the prompt in the manner specified by those
prompt wraps

\item "check": For prompt wraps that apply a last check to the final answer,
after all other prompt wraps have been evaluated.
These prompt wraps may only contain a validation function, and are applied
after all other prompt wraps have been evaluated. These prompt wraps are
even applied after an earlier prompt wrap has broken the extraction and validation loop
with \code{\link[=llm_break]{llm_break()}}
}
Types are used to determine the order in which prompt wraps are applied.
When constructing the prompt text, prompt wraps are applied to the base prompt
in the following order: 'check', 'unspecified', 'break', 'mode', 'tool'.
When evaluating the LLM response and applying extraction and validation functions,
prompt wraps are applied in the reverse order: 'tool', 'mode', 'break',
'unspecified', 'check'.
Order among the same type is preserved in the order they were added to the prompt.}

\item{name}{An optional name for the prompt wrap.
This can be used to identify the prompt wrap in the \linkS4class{tidyprompt} object}
}
\value{
A \linkS4class{tidyprompt} object with the \code{\link[=prompt_wrap]{prompt_wrap()}} appended to it
}
\description{
This function takes a single string or a \linkS4class{tidyprompt} object and
adds a new prompt wrap to it.

A prompt wrap is a set of functions that modify the prompt text,
extract a value from the LLM response, and validate the extracted value.

The functions are used to ensure that the prompt and LLM response are in the
correct format and meet the specified criteria; they may also be used to
provide the LLM with feedback or additional information,
like the result of a tool call or some evaluated code.

Advanced prompt wraps may also include functions that directly handle
the response from a LLM API or configure API parameters.
}
\details{
For advanced use, modify_fn, extraction_fn, and validation_fn
may take the \linkS4class{llm_provider} object (as used with \code{\link[=send_prompt]{send_prompt()}}) as
second argument, and the 'http_list' (a list of all HTTP requests
and responses made during \code{\link[=send_prompt]{send_prompt()}}) as third argument.
Use of these arguments is not required, but can be useful for more complex
prompt wraps which require additional information about the LLM provider
or requests made so far.
The functions (including parameter_fn) also have access to
the object \code{self} (not a function argument; it is attached to the environment
of the function) which contains the \linkS4class{tidyprompt} object that the prompt wrap
is a part of. This can be used to access other prompt wraps, or to access the
prompt text or other information about the prompt. For instance,
other prompt wraps can be accessed through \code{self$get_prompt_wraps()}.
}
\examples{
# A custom prompt_wrap may be created during piping
prompt <- "Hi there!" |>
  prompt_wrap(
    modify_fn = function(base_prompt) {
      paste(base_prompt, "How are you?", sep = "\n\n")
    }
  )
prompt

# (Shorter notation of the above:)
prompt <- "Hi there!" |>
  prompt_wrap(\(x) paste(x, "How are you?", sep = "\n\n"))

# It may often be preferred to make a function which takes a prompt and
#   returns a wrapped prompt:
my_prompt_wrap <- function(prompt) {
  modify_fn <- function(base_prompt) {
    paste(base_prompt, "How are you?", sep = "\n\n")
  }

  prompt_wrap(prompt, modify_fn)
}
prompt <- "Hi there!" |>
  my_prompt_wrap()

# For more advanced examples, take a look at the source code of the
#   pre-built prompt wraps in the tidyprompt package, like
#   answer_as_boolean, answer_as_integer, add_tools, answer_as_code, etc.
# Below is the source code for the 'answer_as_integer' prompt wrap function:
#' Make LLM answer as an integer (between min and max)
#'
#' @param prompt A single string or a [tidyprompt()] object
#' @param min (optional) Minimum value for the integer
#' @param max (optional) Maximum value for the integer
#' @param add_instruction_to_prompt (optional) Add instruction for replying
#' as an integer to the prompt text. Set to FALSE for debugging if extractions/validations
#' are working as expected (without instruction the answer should fail the
#' validation function, initiating a retry)
#'
#' @return A [tidyprompt()] with an added [prompt_wrap()] which
#' will ensure that the LLM response is an integer.
#'
#' @export
#'
#' @example inst/examples/answer_as_integer.R
#'
#' @family pre_built_prompt_wraps
#' @family answer_as_prompt_wraps
answer_as_integer <- function(
  prompt,
  min = NULL,
  max = NULL,
  add_instruction_to_prompt = TRUE
) {
  instruction <- "You must answer with only an integer (use no other characters)."

  if (!is.null(min) && !is.null(max)) {
    instruction <- paste(
      instruction,
      glue::glue(
        "Enter an integer between {min} and {max}."
      )
    )
  } else if (!is.null(min)) {
    instruction <- paste(
      instruction,
      glue::glue(
        "Enter an integer greater than or equal to {min}."
      )
    )
  } else if (!is.null(max)) {
    instruction <- paste(
      instruction,
      glue::glue(
        "Enter an integer less than or equal to {max}."
      )
    )
  }

  modify_fn <- function(original_prompt_text) {
    if (!add_instruction_to_prompt) {
      return(original_prompt_text)
    }

    glue::glue("{original_prompt_text}\n\n{instruction}")
  }

  extraction_fn <- function(x) {
    extracted <- suppressWarnings(as.numeric(x))
    if (is.na(extracted)) {
      return(llm_feedback(instruction))
    }
    return(extracted)
  }

  validation_fn <- function(x) {
    if (x != floor(x)) {
      # Not a whole number
      return(llm_feedback(instruction))
    }

    if (!is.null(min) && x < min) {
      return(
        llm_feedback(
          glue::glue(
            "The number should be greater than or equal to {min}."
          )
        )
      )
    }
    if (!is.null(max) && x > max) {
      return(
        llm_feedback(
          glue::glue(
            "The number should be less than or equal to {max}."
          )
        )
      )
    }
    return(TRUE)
  }

  prompt_wrap(
    prompt,
    modify_fn,
    extraction_fn,
    validation_fn,
    name = "answer_as_integer"
  )
}
}
\seealso{
\linkS4class{tidyprompt} \code{\link[=send_prompt]{send_prompt()}}

Other prompt_wrap: 
\code{\link{llm_break}()},
\code{\link{llm_feedback}()}

Other pre_built_prompt_wraps: 
\code{\link{add_image}()},
\code{\link{add_text}()},
\code{\link{answer_as_boolean}()},
\code{\link{answer_as_category}()},
\code{\link{answer_as_dataframe}()},
\code{\link{answer_as_integer}()},
\code{\link{answer_as_json}()},
\code{\link{answer_as_list}()},
\code{\link{answer_as_multi_category}()},
\code{\link{answer_as_named_list}()},
\code{\link{answer_as_numeric}()},
\code{\link{answer_as_regex_match}()},
\code{\link{answer_as_text}()},
\code{\link{answer_by_chain_of_thought}()},
\code{\link{answer_by_react}()},
\code{\link{answer_using_r}()},
\code{\link{answer_using_sql}()},
\code{\link{answer_using_tools}()},
\code{\link{quit_if}()},
\code{\link{set_system_prompt}()}
}
\concept{pre_built_prompt_wraps}
\concept{prompt_wrap}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/chat_history.R
\name{chat_history.data.frame}
\alias{chat_history.data.frame}
\title{Method for \code{chat_history()} when the input is a \code{data.frame}}
\usage{
\method{chat_history}{data.frame}(chat_history)
}
\arguments{
\item{chat_history}{A data frame with 'role' and 'content' columns,
where 'role' is either 'user', 'assistant', or 'system', and 'content' is a character string
representing a chat message}
}
\value{
A valid chat history \code{data.frame} (of class \code{chat_history}),
with the 'role' and 'content' columns as specified in the input data frame
}
\description{
Creates a \code{chat_history} object from a data frame.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/answer_by_react.R
\name{answer_by_react}
\alias{answer_by_react}
\title{Set ReAct mode for a prompt}
\usage{
answer_by_react(
  prompt,
  extract_from_finish_brackets = TRUE,
  extraction_lenience = TRUE
)
}
\arguments{
\item{prompt}{A single string or a \code{\link[=tidyprompt]{tidyprompt()}} object}

\item{extract_from_finish_brackets}{A logical indicating whether the final answer
should be extracted from the text inside the "FINISH\link{...}" brackets}

\item{extraction_lenience}{A logical indcating whether the extraction function should be lenient.
If TRUE, the extraction function will attempt to extract the final answer
even if it cannot be extracted from within the brackets, by extracting
everything after the final occurence of 'FINISH' (if present). This may
be useful for smaller LLMs which may not follow the output format
as strictly}
}
\value{
A \code{\link[=tidyprompt]{tidyprompt()}} with an added \code{\link[=prompt_wrap]{prompt_wrap()}} which will ensure
that the LLM follows the ReAct mode in answering the prompt
}
\description{
This function enables ReAct mode for the evaluation of a prompt
or a \code{\link[=tidyprompt]{tidyprompt()}}. In ReAct mode, the large language model (LLM) is asked to think step by step,
each time detailing a thought, action, and observation, to eventually arrive at
a final answer. It is hypothesized that this may increase LLM performance
at solving complex tasks. ReAct mode is inspired by the method described in
Yao et al. (2022).
}
\details{
Please note that ReAct mode may be most useful if in combination
with tools that the LLM can use. See, for example, 'add_tools()' for
enabling R function calling, or, for example, 'answer_as_code()' with
'output_as_tool = TRUE' for enabling R code evaluation as a tool.
}
\examples{
\dontrun{
  "What is 5 + 5?" |>
    answer_by_react() |>
    answer_as_integer() |>
    send_prompt()
  # --- Sending request to LLM provider (llama3.1:8b): ---
  # You are given a user's prompt.
  # To answer the user's prompt, you need to think step by step,
  # take an action if needed, and then return a final answer.
  #
  # ----- START OF USER'S PROMPT -----
  # What is 5 + 5?
  #
  # You must answer with only an integer (use no other characters).
  # ----- END OF USER'S PROMPT -----
  #
  # Use the following structure:
  #   Thought: <describe your thought process>
  #   Action: <if needed, describe the action you take (e.g., look up information)>
  #   Observation: <describe the result or observation from the action>
  # (Repeat Thought -> Action -> Observation as necessary)
  #
  # When you are done, you must type:
  #   FINISH[<put here your final answer to the user's prompt>]
  #
  # Ensure your final answer aligns with your reasoning and observations.
  # --- Receiving response from LLM provider: ---
  # Thought: The problem is asking for the sum of two numbers, 5 and 5.
  #
  # Action: None needed, as this is a simple arithmetic operation that can be performed mentally.
  #
  # Observation: I can easily add these two numbers together in my mind to get the result.
  #
  # Thought: To find the sum, I will simply add the two numbers together: 5 + 5 = ?
  #
  # Action: Perform the addition.
  #
  # Observation: The result of adding 5 and 5 is 10.
  #
  # FINISH[10]
  # [1] 10
}
}
\references{
Yao, S., Wu, Y., Cheung, W., Wang, Z., Narasimhan, K., & Kong, L. (2022).
ReAct: Synergizing Reasoning and Acting in Language Models.
\url{doi:10.48550/arXiv.2210.03629}
}
\seealso{
\code{\link[=answer_using_tools]{answer_using_tools()}} \code{\link[=answer_using_r]{answer_using_r()}}

Other pre_built_prompt_wraps: 
\code{\link{add_image}()},
\code{\link{add_text}()},
\code{\link{answer_as_boolean}()},
\code{\link{answer_as_category}()},
\code{\link{answer_as_dataframe}()},
\code{\link{answer_as_integer}()},
\code{\link{answer_as_json}()},
\code{\link{answer_as_list}()},
\code{\link{answer_as_multi_category}()},
\code{\link{answer_as_named_list}()},
\code{\link{answer_as_numeric}()},
\code{\link{answer_as_regex_match}()},
\code{\link{answer_as_text}()},
\code{\link{answer_by_chain_of_thought}()},
\code{\link{answer_using_r}()},
\code{\link{answer_using_sql}()},
\code{\link{answer_using_tools}()},
\code{\link{prompt_wrap}()},
\code{\link{quit_if}()},
\code{\link{set_system_prompt}()}

Other answer_by_prompt_wraps: 
\code{\link{answer_by_chain_of_thought}()}
}
\concept{answer_by_prompt_wraps}
\concept{pre_built_prompt_wraps}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/answer_as_json.R
\name{answer_as_json}
\alias{answer_as_json}
\title{Make LLM answer as JSON (with optional schema; structured output)}
\usage{
answer_as_json(
  prompt,
  schema = NULL,
  schema_strict = FALSE,
  schema_in_prompt_as = c("example", "schema"),
  type = c("auto", "text-based", "openai", "ollama", "openai_oo", "ollama_oo", "ellmer")
)
}
\arguments{
\item{prompt}{A single string or a \code{\link[=tidyprompt]{tidyprompt()}} object}

\item{schema}{Either a R list object which represents
a JSON schema that the response should match, or an 'ellmer' definition of
structured data (e.g., \code{\link[ellmer:type_boolean]{ellmer::type_object()}}; see \href{https://ellmer.tidyverse.org/articles/structured-data.html}{'ellmer' documentation})}

\item{schema_strict}{If TRUE, the provided schema will be strictly enforced.
This option is passed as part of the schema when using type
"openai", "ollama", or "ellmer", and when using "ollama_oo", "openai_oo", or "text-based"
it is passed to \code{\link[jsonvalidate:json_validate]{jsonvalidate::json_validate()}} when validating the response}

\item{schema_in_prompt_as}{If providing a schema and
when using type "text-based", "openai_oo", or "ollama_oo", this argument specifies
how the schema should be included in the prompt:
\itemize{
\item "example" (default): The schema will be included as an example JSON object
(tends to work best). \code{\link[=r_json_schema_to_example]{r_json_schema_to_example()}} is used to generate the example object
from the schema
\item "schema": The schema will be included as a JSON schema
}}

\item{type}{The way that JSON response should be enforced:
\itemize{
\item "auto": Automatically determine the type based on 'llm_provider$api_type'
or 'llm_provider$json_type' (if set; 'json_type' overrides 'api_type' determination).
This may not always consider model compatibility and could lead to errors;
set 'type' manually if errors occur; use 'text-based' if unsure
\item "text-based": Instruction will be added to the prompt
asking for JSON; when a schema is provided, this will also be included
in the prompt (see argument 'schema_in_prompt_as'). JSON will be parsed
from the LLM response and, when a schema is provided, it will be validated
against the schema with \code{\link[jsonvalidate:json_validate]{jsonvalidate::json_validate()}}. Feedback is sent to the
LLM when the response is not valid. This option always works, but may in some
cases may be less powerful than the other native JSON options
\item "openai" and "ollama": The response format will be set via the relevant API parameters,
making the API enforce a valid JSON response. If a schema is provided,
it will also be included in the API parameters and also be enforced by the API.
When no schema is provided, a request for JSON is added to the prompt (as required
by the APIs). Note that these JSON options may not be available for all models
of your provider; consult their documentation for more information.
If you are unsure or encounter errors, use "text-based"
\item "openai_oo" and "ollama_oo": Similar to "openai" and "ollama", but if a
schema is provided it is not included in the API parameters. Schema validation
will be done in R with \code{\link[jsonvalidate:json_validate]{jsonvalidate::json_validate()}}. This can be useful if
you want to use the API's JSON support, but their schema support is limited
\item "ellmer": A parameter will be added to the LLM provider, indicating
that the response should be structured according to the provided schema.
The native 'ellmer' \code{chat$chat_structured()} function will then be used to obtain
the response. This type only useful when using an \code{llm_provider_ellmer()} object.
When \code{type} is set to "auto" and the \code{llm_provider} is an 'ellmer' LLM provider,
this type will be automatically selected (so it should not be necessary to set
this option manually)
}
Note that the "openai" and "ollama" types may also work for other APIs with a
similar structure.
Note furthermore that the "ellmer" type is still experimental and conversion
between 'ellmer' schemas and R list schemas might contain bugs.}
}
\value{
A \code{\link[=tidyprompt]{tidyprompt()}} with an added \code{\link[=prompt_wrap]{prompt_wrap()}} which will ensure
that the LLM response is a valid JSON object. Note that the prompt wrap
will parse the JSON response and return it as an R object (usually a list)
}
\description{
This functions wraps a prompt with settings that ensure the LLM response
is a valid JSON object, optionally matching a given JSON schema
(also known as 'structured output'/'structured data'). Users
may provide either an 'ellmer' type (e.g. \code{\link[ellmer:type_boolean]{ellmer::type_object()}};
see \href{https://ellmer.tidyverse.org/articles/structured-data.html}{'ellmer' documentation}) or
a JSON schema (as R list object) to enforce structure on the response.

The function can work with all models and providers through text-based
handling, but also supports native settings for the OpenAI, Ollama,
and various 'ellmer' types of LLM providers. (See argument 'type'.)
This means that it is possible to easily switch between providers with
different levels of structured output support, while always ensuring
the response will be in the desired format.
}
\examples{
base_prompt <- "How can I solve 8x + 7 = -23?"

# This example will show how to enforce JSON format in the response,
#   with and without a schema, using the `answer_as_json()` prompt wrap.

# If you use type = 'auto', the function will automatically detect the
#   best way to enforce JSON based on the LLM provider you are using.

# `answer_as_json()` supports two ways of supplying a schema for structured output:
#   - 1) an 'ellmer' definition (e.g., `ellmer::type_object()`;
#         see https://ellmer.tidyverse.org/articles/structured-data.html)
#   - 2) a R list object representing a JSON schema
# `answer_as_json()` will convert the schema type which you supply to any
#   LLM provider; so, whether you use an ellmer LLM provider or another type,
#   you can supply either a R list object or an ellmer definition, and don't
#   have to worry about compatibility
# Supplying a schema as an ellmer definition is likely the easiest

# Below, we will show:
#  - 1) enforcing JSON with a schema; 'ellmer' definition
#  - 2) enforcing JSON with a schema; R list object
# -  3) enforcing JSON without a schema

#### Enforcing JSON with a schema (ellmer definition): ####

# Make an ellmer definition of structured output
#   For instance, a persona:
ellmer_schema <- ellmer::type_object(
  name = ellmer::type_string(),
  age = ellmer::type_integer(),
  hobbies = ellmer::type_array(ellmer::type_string())
)

\dontrun{
  # Example Ellmer LLM provider
  ellmer_openai <- llm_provider_ellmer(ellmer::chat_openai(
    model = "gpt-4.1-mini"
  ))

  # Example regular LLM provider
  tidyprompt_openai <- llm_provider_openai()$set_parameters(
    list(model = "gpt-4.1-mini")
  )

  # You can supply the ellmer definition to both types of LLM provider
  #   to generate an R list object adhering to the schema
  result_ellmer_x_ellmer <- "Create a persona" |>
    answer_as_json(ellmer_schema) |>
    send_prompt(ellmer_openai)

  result_tidyrpompt_x_ellmer <- "Create a persona" |>
    answer_as_json(ellmer_schema) |>
    send_prompt(tidyprompt_openai)
}


#### Enforcing JSON with a schema (R list object definition): ####

# Make a list representing a JSON schema,
#   which the LLM response must adhere to:
json_schema <- list(
  name = "steps_to_solve", # Required for OpenAI API
  description = NULL, # Optional for OpenAI API
  schema = list(
    type = "object",
    properties = list(
      steps = list(
        type = "array",
        items = list(
          type = "object",
          properties = list(
            explanation = list(type = "string"),
            output = list(type = "string")
          ),
          required = c("explanation", "output"),
          additionalProperties = FALSE
        )
      ),
      final_answer = list(type = "string")
    ),
    required = c("steps", "final_answer"),
    additionalProperties = FALSE
  )
  # 'strict' parameter is set as argument 'answer_as_json()'
)
# Note: when you are not using an OpenAI API, you can also pass just the
#   internal 'schema' list object to 'answer_as_json()' instead of the full
#   'json_schema' list object

# Generate example R object based on schema:
r_json_schema_to_example(json_schema)

\dontrun{
  ## Text-based with schema (works for any provider/model):
  #   - Adds request to prompt for a JSON object
  #   - Adds schema to prompt
  #   - Extracts JSON from textual response (feedback for retry if no JSON received)
  #   - Validates JSON against schema with 'jsonvalidate' package (feedback for retry if invalid)
  #   - Parses JSON to R object
  json_4 <- base_prompt |>
    answer_as_json(schema = json_schema) |>
    send_prompt(llm_provider_ollama())
  # --- Sending request to LLM provider (llama3.1:8b): ---
  # How can I solve 8x + 7 = -23?
  #
  # Your must format your response as a JSON object.
  #
  # Your JSON object should match this example JSON object:
  #   {
  #     "steps": [
  #       {
  #         "explanation": "...",
  #         "output": "..."
  #       }
  #     ],
  #     "final_answer": "..."
  #   }
  # --- Receiving response from LLM provider: ---
  # Here is the solution to the equation:
  #
  # ```
  # {
  #   "steps": [
  #     {
  #       "explanation": "First, we want to isolate the term with 'x' by
  #       subtracting 7 from both sides of the equation.",
  #       "output": "8x + 7 - 7 = -23 - 7"
  #     },
  #     {
  #       "explanation": "This simplifies to: 8x = -30",
  #       "output": "8x = -30"
  #     },
  #     {
  #       "explanation": "Next, we want to get rid of the coefficient '8' by
  #       dividing both sides of the equation by 8.",
  #       "output": "(8x) / 8 = (-30) / 8"
  #     },
  #     {
  #       "explanation": "This simplifies to: x = -3.75",
  #       "output": "x = -3.75"
  #     }
  #   ],
  #   "final_answer": "-3.75"
  # }
  # ```

  ## Ollama with schema:
  #   - Sets 'format' parameter to 'json', enforcing JSON
  #   - Adds request to prompt for a JSON object, as is recommended by the docs
  #   - Adds schema to prompt
  #   - Validates JSON against schema with 'jsonvalidate' package (feedback for retry if invalid)
  json_5 <- base_prompt |>
    answer_as_json(json_schema, type = "auto") |>
    send_prompt(llm_provider_ollama())
  # --- Sending request to LLM provider (llama3.1:8b): ---
  # How can I solve 8x + 7 = -23?
  #
  # Your must format your response as a JSON object.
  #
  # Your JSON object should match this example JSON object:
  # {
  #   "steps": [
  #     {
  #       "explanation": "...",
  #       "output": "..."
  #     }
  #   ],
  #   "final_answer": "..."
  # }
  # --- Receiving response from LLM provider: ---
  # {
  #   "steps": [
  #     {
  #       "explanation": "First, subtract 7 from both sides of the equation to
  #       isolate the term with x.",
  #       "output": "8x = -23 - 7"
  #     },
  #     {
  #       "explanation": "Simplify the right-hand side of the equation.",
  #       "output": "8x = -30"
  #     },
  #     {
  #       "explanation": "Next, divide both sides of the equation by 8 to solve for x.",
  #       "output": "x = -30 / 8"
  #     },
  #     {
  #       "explanation": "Simplify the right-hand side of the equation.",
  #       "output": "x = -3.75"
  #     }
  #   ],
  #   "final_answer": "-3.75"
  # }

  ## OpenAI with schema:
  #   - Sets 'response_format' parameter to 'json_object', enforcing JSON
  #   - Adds json_schema to the API request, API enforces JSON adhering schema
  #   - Parses JSON to R object
  json_6 <- base_prompt |>
    answer_as_json(json_schema, type = "auto") |>
    send_prompt(llm_provider_openai())
  # --- Sending request to LLM provider (gpt-4o-mini): ---
  # How can I solve 8x + 7 = -23?
  # --- Receiving response from LLM provider: ---
  # {"steps":[
  # {"explanation":"Start with the original equation.",
  # "output":"8x + 7 = -23"},
  # {"explanation":"Subtract 7 from both sides to isolate the term with x.",
  # "output":"8x + 7 - 7 = -23 - 7"},
  # {"explanation":"Simplify the left side and the right side of the equation.",
  # "output":"8x = -30"},
  # {"explanation":"Now, divide both sides by 8 to solve for x.",
  # "output":"x = -30 / 8"},
  # {"explanation":"Simplify the fraction by dividing both the numerator and the
  # denominator by 2.",
  # "output":"x = -15 / 4"}
  # ], "final_answer":"x = -15/4"}

  # You can also use the R list object schema definition with an
  #  ellmer LLM provider; `answer_as_json()` will do the conversion for you
  json_7 <- base_prompt |>
    answer_as_json(json_schema) |>
    send_prompt(ellmer_openai)
}


#### Enforcing JSON without a schema: ####

\dontrun{
  ## Text-based (works for any provider/model):
  #   Adds request to prompt for a JSON object
  #   Extracts JSON from textual response (feedback for retry if no JSON received)
  #   Parses JSON to R object
  json_1 <- base_prompt |>
    answer_as_json() |>
    send_prompt(llm_provider_ollama())
  # --- Sending request to LLM provider (llama3.1:8b): ---
  # How can I solve 8x + 7 = -23?
  #
  # Your must format your response as a JSON object.
  # --- Receiving response from LLM provider: ---
  # Here is the solution to the equation formatted as a JSON object:
  #
  # ```
  # {
  #   "equation": "8x + 7 = -23",
  #   "steps": [
  #     {
  #       "step": "Subtract 7 from both sides of the equation",
  #       "expression": "-23 - 7"
  #     },
  #     {
  #       "step": "Simplify the expression on the left side",
  #       "result": "-30"
  #     },
  #     {
  #       "step": "Divide both sides by -8 to solve for x",
  #       "expression": "-30 / -8"
  #     },
  #     {
  #       "step": "Simplify the expression on the right side",
  #       "result": "3.75"
  #     }
  #   ],
  #   "solution": {
  #     "x": 3.75
  #   }
  # }
  # ```


  ## Ollama:
  #   - Sets 'format' parameter to 'json', enforcing JSON
  #   - Adds request to prompt for a JSON object, as is recommended by the docs
  #   - Parses JSON to R object
  json_2 <- base_prompt |>
    answer_as_json(type = "auto") |>
    send_prompt(llm_provider_ollama())
  # --- Sending request to LLM provider (llama3.1:8b): ---
  # How can I solve 8x + 7 = -23?
  #
  # Your must format your response as a JSON object.
  # --- Receiving response from LLM provider: ---
  # {"steps": [
  #   "Subtract 7 from both sides to get 8x = -30",
  #   "Simplify the right side of the equation to get 8x = -30",
  #   "Divide both sides by 8 to solve for x, resulting in x = -30/8",
  #   "Simplify the fraction to find the value of x"
  # ],
  # "value_of_x": "-3.75"}


  ## OpenAI-type API without schema:
  #   - Sets 'response_format' parameter to 'json_object', enforcing JSON
  #   - Adds request to prompt for a JSON object, as is required by the API
  #   - Parses JSON to R object
  json_3 <- base_prompt |>
    answer_as_json(type = "auto") |>
    send_prompt(llm_provider_openai())
  # --- Sending request to LLM provider (gpt-4o-mini): ---
  # How can I solve 8x + 7 = -23?
  #
  # Your must format your response as a JSON object.
  # --- Receiving response from LLM provider: ---
  # {
  #   "solution_steps": [
  #     {
  #       "step": 1,
  #       "operation": "Subtract 7 from both sides",
  #       "equation": "8x + 7 - 7 = -23 - 7",
  #       "result": "8x = -30"
  #     },
  #     {
  #       "step": 2,
  #       "operation": "Divide both sides by 8",
  #       "equation": "8x / 8 = -30 / 8",
  #       "result": "x = -3.75"
  #     }
  #   ],
  #   "solution": {
  #     "x": -3.75
  #   }
  # }
}
}
\seealso{
Other pre_built_prompt_wraps: 
\code{\link{add_image}()},
\code{\link{add_text}()},
\code{\link{answer_as_boolean}()},
\code{\link{answer_as_category}()},
\code{\link{answer_as_dataframe}()},
\code{\link{answer_as_integer}()},
\code{\link{answer_as_list}()},
\code{\link{answer_as_multi_category}()},
\code{\link{answer_as_named_list}()},
\code{\link{answer_as_numeric}()},
\code{\link{answer_as_regex_match}()},
\code{\link{answer_as_text}()},
\code{\link{answer_by_chain_of_thought}()},
\code{\link{answer_by_react}()},
\code{\link{answer_using_r}()},
\code{\link{answer_using_sql}()},
\code{\link{answer_using_tools}()},
\code{\link{prompt_wrap}()},
\code{\link{quit_if}()},
\code{\link{set_system_prompt}()}

Other answer_as_prompt_wraps: 
\code{\link{answer_as_boolean}()},
\code{\link{answer_as_category}()},
\code{\link{answer_as_dataframe}()},
\code{\link{answer_as_integer}()},
\code{\link{answer_as_list}()},
\code{\link{answer_as_multi_category}()},
\code{\link{answer_as_named_list}()},
\code{\link{answer_as_numeric}()},
\code{\link{answer_as_regex_match}()},
\code{\link{answer_as_text}()}

Other json: 
\code{\link{r_json_schema_to_example}()}
}
\concept{answer_as_prompt_wraps}
\concept{json}
\concept{pre_built_prompt_wraps}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/misc_add_text.R
\name{add_text}
\alias{add_text}
\title{Add text to a tidyprompt}
\usage{
add_text(prompt, text, position = c("after", "before"), sep = "\\n\\n")
}
\arguments{
\item{prompt}{A single string or a \code{\link[=tidyprompt]{tidyprompt()}} object}

\item{text}{Text to be added to the current prompt text}

\item{position}{Where to add the text; either "after" or "before".}

\item{sep}{Separator to be used between the current prompt text and the text to be added}
}
\value{
A \code{\link[=tidyprompt]{tidyprompt()}} with an added \code{\link[=prompt_wrap]{prompt_wrap()}} which
will append the text to the end of the current prompt text
}
\description{
Add text to a prompt by adding a \code{\link[=prompt_wrap]{prompt_wrap()}} which will append the text to
the before or after the current prompt text.
}
\examples{
prompt <- "Hi there!" |>
  add_text("How is your day?")
prompt
prompt |>
  construct_prompt_text()
}
\seealso{
Other pre_built_prompt_wraps: 
\code{\link{add_image}()},
\code{\link{answer_as_boolean}()},
\code{\link{answer_as_category}()},
\code{\link{answer_as_dataframe}()},
\code{\link{answer_as_integer}()},
\code{\link{answer_as_json}()},
\code{\link{answer_as_list}()},
\code{\link{answer_as_multi_category}()},
\code{\link{answer_as_named_list}()},
\code{\link{answer_as_numeric}()},
\code{\link{answer_as_regex_match}()},
\code{\link{answer_as_text}()},
\code{\link{answer_by_chain_of_thought}()},
\code{\link{answer_by_react}()},
\code{\link{answer_using_r}()},
\code{\link{answer_using_sql}()},
\code{\link{answer_using_tools}()},
\code{\link{prompt_wrap}()},
\code{\link{quit_if}()},
\code{\link{set_system_prompt}()}

Other miscellaneous_prompt_wraps: 
\code{\link{add_image}()},
\code{\link{quit_if}()},
\code{\link{set_system_prompt}()}
}
\concept{miscellaneous_prompt_wraps}
\concept{pre_built_prompt_wraps}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/helper_other_extract_from_return_list.R
\name{extract_from_return_list}
\alias{extract_from_return_list}
\title{Function to extract a specific element from a list}
\usage{
extract_from_return_list(list, name_of_element = "response")
}
\arguments{
\item{list}{A list, typically the output from \code{\link[=send_prompt]{send_prompt()}} with \code{return_mode = "full"}}

\item{name_of_element}{A character string with the name of the element to extract from the list}
}
\value{
The extracted element from the list
}
\description{
This function is intended as a helper function for piping with output from
\code{\link[=send_prompt]{send_prompt()}} when using \code{return_mode = "full"}. It allows to
extract a specific element from the list returned by \code{\link[=send_prompt]{send_prompt()}}, which
can be useful for further piping.
}
\examples{
\dontrun{
  response <- "Hi!" |>
    send_prompt(llm_provider_ollama(), return_mode = "full") |>
    extract_from_return_list("response")
  response
  # [1] "It's nice to meet you. Is there something I can help you with,
  # or would you like to chat?"
}
}
\concept{miscellaneous_helpers}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/llm_providers.R
\name{llm_provider_ellmer}
\alias{llm_provider_ellmer}
\title{Create a new LLM provider from an \code{ellmer::chat()} object}
\usage{
llm_provider_ellmer(
  chat,
  parameters = list(stream = getOption("tidyprompt.stream", TRUE)),
  verbose = getOption("tidyprompt.verbose", TRUE)
)
}
\arguments{
\item{chat}{An \code{ellmer::chat()} object (e.g., \code{ellmer::chat_openai()})}

\item{parameters}{A named list of parameters. See 'details' for supported parameters}

\item{verbose}{A logical indicating whether the interaction with the \linkS4class{llm_provider}
should be printed to the console. Default is TRUE}
}
\value{
An \linkS4class{llm_provider} with api_type = "ellmer"
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}}

This function creates a \linkS4class{llm_provider} from an \code{ellmer::chat()} object.
This allows the user to use the various LLM providers which are supported
by the 'ellmer' R package, including respective configuration and features.

Please note that this function is experimental. This provider type may show different behavior than
other LLM providers, and may not function optimally.
}
\details{
\strong{ellmer version compatibility:} tidyprompt requires ellmer >= 0.3.0.
Some features are only fully available with ellmer >= 0.4.0; on older
versions they degrade gracefully:
\itemize{
\item Built-in tools (\code{ToolBuiltIn}, e.g. \code{claude_tool_web_search()})
require ellmer >= 0.4.0.
\item \code{type_ignore()} for ignored tool arguments requires ellmer >= 0.4.0;
on older versions, ignored arguments fall back to optional string
parameters.
}

Unlike other LLM provider classes,
most LLM provider settings need to be managed in the \code{ellmer::chat()} object
(and not in the \verb{$parameters} list). \verb{$get_chat()} and \verb{$set_chat()} may be used
to manipulate the chat object. There are however some parameters that can be set
in the \verb{$parameters} list; these are documented below.
\enumerate{
\item Streaming can be controlled through via the \verb{$parameters$stream} parameter.
If set to TRUE (default), streaming will be used if supported by the underlying
\code{ellmer::chat()} object. If the underlying \code{ellmer::chat()} object does not support streaming,
you may need to set this parameter to FALSE to avoid errors.
\item A special parameter \verb{$.ellmer_structured_type} may also be set in the \verb{$parameters} list;
this parameter is used to specify a structured output format. This should be a 'ellmer'
structured type (e.g., \code{ellmer::type_object}; see https://ellmer.tidyverse.org/articles/structured-data.html).
\code{answer_as_json()} sets this parameter to obtain structured output
(it is not recommended to set this parameter manually, but it is possible).
}
}
\examples{
# Various providers:
ollama <- llm_provider_ollama()
openai <- llm_provider_openai()
openrouter <- llm_provider_openrouter()
mistral <- llm_provider_mistral()
groq <- llm_provider_groq()
xai <- llm_provider_xai()
gemini <- llm_provider_google_gemini()

# From an `ellmer::chat()` (e.g., `ellmer::chat_openai()`, ...):
\dontrun{
ellmer <- llm_provider_ellmer(ellmer::chat_openai())
}

# Initialize with settings:
ollama <- llm_provider_ollama(
  parameters = list(
    model = "llama3.2:3b",
    stream = TRUE
  ),
  verbose = TRUE,
  url = "http://localhost:11434/api/chat"
)

# Change settings:
ollama$verbose <- FALSE
ollama$parameters$stream <- FALSE
ollama$parameters$model <- "llama3.1:8b"

\dontrun{
# Try a simple chat message with '$complete_chat()':
response <- ollama$complete_chat("Hi!")
response
# $role
# [1] "assistant"
#
# $content
# [1] "How's it going? Is there something I can help you with or would you like
# to chat?"
#
# $http
# Response [http://localhost:11434/api/chat]
# Date: 2024-11-18 14:21
# Status: 200
# Content-Type: application/json; charset=utf-8
# Size: 375 B

# Use with send_prompt():
"Hi" |>
  send_prompt(ollama)
# [1] "How's your day going so far? Is there something I can help you with or
# would you like to chat?"
}
}
\seealso{
Other llm_provider: 
\code{\link{llm_provider-class}},
\code{\link{llm_provider_google_gemini}()},
\code{\link{llm_provider_groq}()},
\code{\link{llm_provider_mistral}()},
\code{\link{llm_provider_ollama}()},
\code{\link{llm_provider_openai}()},
\code{\link{llm_provider_openrouter}()},
\code{\link{llm_provider_xai}()}
}
\concept{llm_provider}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tidyprompt.R
\name{get_chat_history}
\alias{get_chat_history}
\title{Get the chat history of a \linkS4class{tidyprompt} object}
\usage{
get_chat_history(x)
}
\arguments{
\item{x}{A \linkS4class{tidyprompt} object}
}
\value{
A dataframe containing the chat history
}
\description{
This function gets the chat history of the \linkS4class{tidyprompt} object.
The chat history is constructed from the base prompt, system prompt, and chat
history field. The returned object will be the chat history
with the system prompt as the first message with role 'system' and the
the base prompt as the last message with role 'user'.
}
\examples{
prompt <- tidyprompt("Hi!")
print(prompt)

# Add to a tidyprompt using a prompt wrap:
prompt <- tidyprompt("Hi!") |>
  add_text("How are you?")
print(prompt)

# Strings can be input for prompt wraps; therefore,
#   a call to tidyprompt() is not necessary:
prompt <- "Hi" |>
  add_text("How are you?")

# Example of adding extraction & validation with a prompt_wrap():
prompt <- "Hi" |>
  add_text("What is 5 + 5?") |>
  answer_as_integer()

\dontrun{
  # tidyprompt objects are evaluated by send_prompt(), which will
  #   handle construct the prompt text, send it to the LLM provider,
  #   and apply the extraction and validation functions from the tidyprompt object
  prompt |>
    send_prompt(llm_provider_ollama())
  # --- Sending request to LLM provider (llama3.1:8b): ---
  #   Hi
  #
  #   What is 5 + 5?
  #
  #   You must answer with only an integer (use no other characters).
  # --- Receiving response from LLM provider: ---
  #   10
  # [1] 10

  # See prompt_wrap() and send_prompt() for more details
}

# `tidyprompt` objects may be validated with these helpers:
is_tidyprompt(prompt) # Returns TRUE if input is a valid tidyprompt object

# Get base prompt text
base_prompt <- prompt$base_prompt

# Get all prompt wraps
prompt_wraps <- prompt$get_prompt_wraps()
# Alternative:
prompt_wraps <- get_prompt_wraps(prompt)

# Construct prompt text
prompt_text <- prompt$construct_prompt_text()
# Alternative:
prompt_text <- construct_prompt_text(prompt)

# Set chat history (affecting also the base prompt)
chat_history <- data.frame(
  role = c("user", "assistant", "user"),
  content = c("What is 5 + 5?", "10", "And what is 5 + 6?")
)
prompt$set_chat_history(chat_history)

# Get chat history
chat_history <- prompt$get_chat_history()
}
\seealso{
\code{\link[=chat_history]{chat_history()}}

Other tidyprompt: 
\code{\link{construct_prompt_text}()},
\code{\link{get_prompt_wraps}()},
\code{\link{is_tidyprompt}()},
\code{\link{set_chat_history}()},
\code{\link{tidyprompt}()},
\code{\link{tidyprompt-class}}
}
\concept{tidyprompt}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/answer_as_dataframe.R
\name{answer_as_dataframe}
\alias{answer_as_dataframe}
\title{Make LLM answer as a data frame via structured output}
\usage{
answer_as_dataframe(
  prompt,
  schema,
  min_rows = NULL,
  max_rows = NULL,
  schema_strict = FALSE,
  schema_in_prompt_as = c("example", "schema"),
  type = c("auto", "text-based", "openai", "ollama", "openai_oo", "ollama_oo", "ellmer")
)
}
\arguments{
\item{prompt}{A single string or a \code{\link[=tidyprompt]{tidyprompt()}} object}

\item{schema}{A JSON schema list or an 'ellmer' type definition describing a
single row, an array of rows, or a wrapper object containing a \code{rows} array.}

\item{min_rows}{(optional) Minimum number of rows required in the returned
data frame}

\item{max_rows}{(optional) Maximum number of rows allowed in the returned
data frame}

\item{schema_strict}{If TRUE, the wrapped schema will be strictly enforced.
Passed through to \code{\link[=answer_as_json]{answer_as_json()}}}

\item{schema_in_prompt_as}{Passed through to \code{\link[=answer_as_json]{answer_as_json()}} when using a
text-based JSON path}

\item{type}{Passed through to \code{\link[=answer_as_json]{answer_as_json()}} to control the structured
output backend}
}
\value{
A \code{\link[=tidyprompt]{tidyprompt()}} with an added \code{\link[=prompt_wrap]{prompt_wrap()}} which will ensure
that the LLM response is returned as a data frame.
}
\description{
This function builds on \code{\link[=answer_as_json]{answer_as_json()}} to extract a data frame from an
LLM response using structured output. The supplied \code{schema} should describe a
single row of the desired data frame, or an array of such rows. Internally,
\code{answer_as_dataframe()} standardizes the schema to a JSON object with a
\code{rows} field containing an array of row objects. This shape works well with
both text-based JSON extraction and native structured-output backends,
including 'ellmer', where arrays of objects are converted to data frames.
}
\details{
Prefer supplying an 'ellmer' row schema created with
\code{ellmer::type_object(...)} when possible. This is usually the clearest way to
describe the columns you want, and it maps cleanly to native 'ellmer'
structured output. These 'ellmer' schema definitions can also be used with
non-'ellmer' LLM providers, because 'tidyprompt' converts between 'ellmer'
schema definitions and JSON-schema representations as needed.

\code{answer_as_dataframe()} accepts the following schema shapes:
\itemize{
\item A single row schema, such as \code{ellmer::type_object(...)} or a JSON schema
object whose properties describe the columns of one row.
\item An array-of-rows schema, such as \code{ellmer::type_array(row_schema)} or a JSON
schema with \code{type = "array"} and row objects under \code{items}.
\item A wrapper object whose \strong{only} property is a \code{rows} field containing an
array of row objects (matching the shape produced internally by
\code{answer_as_dataframe()}). Schemas with additional sibling properties
alongside \code{rows} are treated as row schemas, not wrappers.
}

Regardless of which of these forms you supply, \code{answer_as_dataframe()}
normalizes it to a row-oriented structured-output schema before delegating to
\code{\link[=answer_as_json]{answer_as_json()}}.
}
\examples{
# `answer_as_dataframe()` accepts multiple schema shapes.
# Prefer an ellmer row schema when possible, because it is concise and maps
# cleanly to native ellmer structured output.
# These ellmer schema definitions also work with non-ellmer LLM providers,
# because tidyprompt converts between ellmer schemas and JSON schemas for you.

if (requireNamespace("ellmer", quietly = TRUE)) {
  person_row_schema_ellmer <- ellmer::type_object(
    name = ellmer::type_string(),
    age = ellmer::type_integer(),
    city = ellmer::type_string()
  )

  # Also accepted: an array of row objects.
  person_array_schema_ellmer <- ellmer::type_array(person_row_schema_ellmer)
}

# Also accepted: a JSON schema describing one row.
person_row_schema_json <- list(
  type = "object",
  properties = list(
    name = list(type = "string"),
    age = list(type = "integer"),
    city = list(type = "string")
  ),
  required = c("name", "age", "city"),
  additionalProperties = FALSE
)

# Also accepted: a wrapper object with a `rows` array.
person_wrapper_schema_json <- list(
  type = "object",
  properties = list(
    rows = list(
      type = "array",
      items = person_row_schema_json
    )
  ),
  required = "rows",
  additionalProperties = FALSE
)

\dontrun{
  prompt <- paste(
    "Extract the people in the following notes as a table:",
    "Alice (32, Berlin), Bob (28, Utrecht)."
  )

  # Preferred: ellmer row schema.
  # This works both with ellmer-backed providers and with regular tidyprompt
  # providers, because tidyprompt converts the schema when needed.
  if (requireNamespace("ellmer", quietly = TRUE)) {
    prompt |>
      answer_as_dataframe(person_row_schema_ellmer) |>
      send_prompt()
    #    name age    city
    # 1 Alice  32  Berlin
    # 2   Bob  28 Utrecht

    # Also works: ellmer array-of-rows schema.
    prompt |>
      answer_as_dataframe(person_array_schema_ellmer) |>
      send_prompt()
  }

  # Also works: JSON schema for one row.
  prompt |>
    answer_as_dataframe(person_row_schema_json, type = "text-based") |>
    send_prompt()

  # Also works: JSON wrapper schema with a `rows` array.
  prompt |>
    answer_as_dataframe(person_wrapper_schema_json, type = "text-based") |>
    send_prompt()
}
}
\seealso{
Other pre_built_prompt_wraps: 
\code{\link{add_image}()},
\code{\link{add_text}()},
\code{\link{answer_as_boolean}()},
\code{\link{answer_as_category}()},
\code{\link{answer_as_integer}()},
\code{\link{answer_as_json}()},
\code{\link{answer_as_list}()},
\code{\link{answer_as_multi_category}()},
\code{\link{answer_as_named_list}()},
\code{\link{answer_as_numeric}()},
\code{\link{answer_as_regex_match}()},
\code{\link{answer_as_text}()},
\code{\link{answer_by_chain_of_thought}()},
\code{\link{answer_by_react}()},
\code{\link{answer_using_r}()},
\code{\link{answer_using_sql}()},
\code{\link{answer_using_tools}()},
\code{\link{prompt_wrap}()},
\code{\link{quit_if}()},
\code{\link{set_system_prompt}()}

Other answer_as_prompt_wraps: 
\code{\link{answer_as_boolean}()},
\code{\link{answer_as_category}()},
\code{\link{answer_as_integer}()},
\code{\link{answer_as_json}()},
\code{\link{answer_as_list}()},
\code{\link{answer_as_multi_category}()},
\code{\link{answer_as_named_list}()},
\code{\link{answer_as_numeric}()},
\code{\link{answer_as_regex_match}()},
\code{\link{answer_as_text}()}
}
\concept{answer_as_prompt_wraps}
\concept{pre_built_prompt_wraps}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/chat_history.R
\name{chat_history}
\alias{chat_history}
\title{Create or validate \code{chat_history} object}
\usage{
chat_history(chat_history)
}
\arguments{
\item{chat_history}{A single string, a \code{data.frame} with 'role' and 'content' columns,
or NULL. If a \code{data.frame} is provided, it should contain 'role' and 'content' columns,
where 'role' is either 'user', 'assistant', 'system', or 'tool', and 'content' is a character string
representing a chat message}
}
\value{
A valid chat history \code{data.frame} (of class \code{chat_history})
}
\description{
This function creates and validates a \code{chat_history} object, ensuring that it matches
the expected format with 'role' and 'content' columns. It has separate methods
for \code{data.frame} and \code{character} inputs and includes a helper function to add a
system prompt to the chat history.
}
\examples{
chat <- "Hi there!" |>
  chat_history()
chat

chat_from_df <- data.frame(
  role = c("user", "assistant"),
  content = c("Hi there!", "Hello! How can I help you today?")
) |>
  chat_history()
chat_from_df

# `add_msg_to_chat_history()` may be used to add messages to a chat history
chat_from_df <- chat_from_df |>
  add_msg_to_chat_history("Calculate 2+2 for me, please!")
chat_from_df

# You can also continue conversations which originate from `send_prompt()`:
\dontrun{
  result <- "Hi there!" |>
    send_prompt(return_mode = "full")
  # --- Sending request to LLM provider (llama3.1:8b): ---
  # Hi there!
  # --- Receiving response from LLM provider: ---
  # It's nice to meet you. Is there something I can help you with, or would you
  # like to chat?

  # Access the chat history from the result:
  chat_from_send_prompt <- result$chat_history

  # Add a message to the chat history:
  chat_history_with_new_message <- chat_from_send_prompt |>
    add_msg_to_chat_history("Let's chat!")

  # The new chat history can be input for a new tidyprompt:
  prompt <- tidyprompt(chat_history_with_new_message)

  # You can also take an existing tidyprompt and add the new chat history to it;
  #   this way, you can continue a conversation using the same prompt wraps
  prompt$set_chat_history(chat_history_with_new_message)

  # send_prompt() also accepts a chat history as input:
  new_result <- chat_history_with_new_message |>
    send_prompt(return_mode = "full")

  # You can also create a persistent chat history object from
  #   a chat history data frame; see ?`persistent_chat-class`
  chat <- `persistent_chat-class`$new(llm_provider_ollama(), chat_from_send_prompt)
  chat$chat("Let's chat!")
}
}
\concept{chat_history}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/llm_break.R
\name{llm_break_soft}
\alias{llm_break_soft}
\title{Create an \code{llm_break_soft} object}
\usage{
llm_break_soft(object_to_return = NULL)
}
\arguments{
\item{object_to_return}{The object to return as the response result
from \code{\link[=send_prompt]{send_prompt()}} when this object is returned from an extraction or
validation function}
}
\value{
An list of class "llm_break_soft" containing the object to return
}
\description{
This object is used to break a extraction and validation loop defined in a \code{\link[=prompt_wrap]{prompt_wrap()}},
as evaluated by \code{\link[=send_prompt]{send_prompt()}}. When an extraction or validation function returns
this object, it will prevent any future interactions with the LLM provider for the
current prompt. Remaining extraction and validation functions will still be applied
and it will still be possible to pass these with the current response from the LLM
provider; only, no more new tries will be made if the current response is not
satisfactory.

This is useful when, e.g., the token limit for the LLM provider has been reached,
but the final response that we got may still be satisfactory. In this case,
\code{llm_break()} cannot be used, as it would instantly return the current response as the final
result, which is not what we want. Instead, \code{llm_break_soft()} can be used to
prevent any further interactions with the LLM provider, but still allow the
remaining extraction and validation functions to be applied (and have those decide
the success of the current response).
}
\examples{
# Quitting when total token count is exceeded (Google Gemini API example)
\dontrun{
  "How are you?" |>
    # Forcing multi-response via initial error, for demonstration purposes
    answer_as_integer(add_instruction_to_prompt = FALSE) |>
    # Validation function to check total token count
    prompt_wrap(validation_fn = function(response, llm_provider, http_list) {
      total_tokens <- purrr::map_dbl(
        http_list$responses,
        ~ .x$body |>
          rawToChar() |>
          jsonlite::fromJSON() |>
          purrr::pluck("usageMetadata", "totalTokenCount")
      ) |> sum()
      if (total_tokens > 50) {
        warning("Token count exceeded; preventing further interactions")
        # Using llm_break_soft() to prevent further interactions
        return(llm_break_soft(response))
      }
    }) |>
    send_prompt(llm_provider_google_gemini(), return_mode = "full")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/answer_using_sql.R
\name{answer_using_sql}
\alias{answer_using_sql}
\title{Enable LLM to draft and execute SQL queries on a database}
\usage{
answer_using_sql(
  prompt,
  add_text = paste0("You must code in SQL to answer this prompt.",
    " You must provide all SQL code between ```sql and ```.", "\\n\\n",
    "Never make assumptions about the possible values in the tables.\\n",
    "Instead, execute SQL queries to retrieve information you need."),
  conn,
  list_tables = TRUE,
  describe_tables = TRUE,
  evaluate_code = FALSE,
  output_as_tool = FALSE,
  return_mode = c("full", "code", "object", "formatted_output", "llm_answer")
)
}
\arguments{
\item{prompt}{A single string or a \code{\link[=tidyprompt]{tidyprompt()}} object}

\item{add_text}{Single string which will be added to the prompt text,
informing the LLM that they must use SQL to answer the prompt}

\item{conn}{A DBIConnection object to the SQL database}

\item{list_tables}{Logical indicating whether to list tables available in the database
in the prompt text}

\item{describe_tables}{Logical indicating whether to describe the tables available in the database
in the prompt text. If TRUE, the columns of each table will be listed}

\item{evaluate_code}{Logical indicating whether to evaluate the SQL code.
If TRUE, the SQL code will be executed on the database and the results will be returned.
Use with caution, as this allows the LLM to execute arbitrary SQL code}

\item{output_as_tool}{Logical indicating whether to return the output as a tool result.
If TRUE, the output of the SQL query will be sent back to the LLM as a tool result.
The LLM can then provide a final answer or try another query. This can
continue until the LLM provides a final answer without any SQL code}

\item{return_mode}{Character string indicating the return mode. Options are:
\itemize{
\item "full": Return a list containing the SQL code, output, and formatted output
\item "code": Return only the SQL code
\item "object": Return only the query result object
\item "formatted_output": Return the formatted output: a string detailing the SQL code
and query result object.This is identical to how the LLM would see the output when
output_as_tool is TRUE
\item "llm_answer": Return the LLM answer. If output as tool is TRUE,
the return mode will always be "llm_answer" (since the LLM uses SQL to
provide a final answer)
}}
}
\value{
A \code{\link[=tidyprompt]{tidyprompt()}} with an added \code{\link[=prompt_wrap]{prompt_wrap()}} which will ensure
that the LLM will use SQL to answer the prompt
}
\description{
Enable LLM to draft and execute SQL queries on a database
}
\examples{
\dontrun{
  # Create an in-memory SQLite database
  conn <- DBI::dbConnect(RSQLite::SQLite(), ":memory:")

  # Create a sample table of customers
  DBI::dbExecute(conn, "
  CREATE TABLE
    customers (
      id INTEGER PRIMARY KEY,
      name TEXT,
      email TEXT,
      country TEXT
    );
  ")

  # Insert some sample customer data
  DBI::dbExecute(conn, "
  INSERT INTO
    customers (name, email, country)
  VALUES
    ('Alice', 'alice@example.com', 'USA'),
    ('Bob', 'bob@example.com', 'Canada'),
    ('Charlie', 'charlie@example.com', 'UK'),
    ('Diana', 'diana@example.com', 'USA');
  ")

  # Create another sample table for orders
  DBI::dbExecute(conn, "
  CREATE TABLE orders (
    order_id INTEGER PRIMARY KEY,
    customer_id INTEGER,
    product TEXT,
    amount REAL,
    order_date TEXT,
    FOREIGN KEY(customer_id) REFERENCES customers(id)
  );
  ")

  # Insert some sample orders
  DBI::dbExecute(conn, "
  INSERT INTO
    orders (customer_id, product, amount, order_date)
  VALUES
    (1, 'Widget', 19.99, '2024-01-15'),
    (1, 'Gadget', 29.99, '2024-01-17'),
    (2, 'Widget', 19.99, '2024-02-10'),
    (3, 'SuperWidget', 49.99, '2024-03-05'),
    (4, 'Gadget', 29.99, '2024-04-01'),
    (1, 'Thingamajig', 9.99, '2024-04-02');
  ")

  # Ask LLM a question which it will answer using the SQL database:
  "Where are my customers from?" |>
    answer_using_sql(
      conn = conn,
      evaluate_code = TRUE,
      output_as_tool = TRUE
    ) |>
    send_prompt(llm_provider_openai())
  # --- Sending request to LLM provider (gpt-4o-mini): ---
  # Where are my customers from?
  #
  # You must code in SQL to answer this prompt. You must provide all SQL code
  # between ```sql and ```.
  #
  # Never make assumptions about the possible values in the tables.
  # Instead, execute SQL queries to retrieve information you need.
  #
  # These tables are available in the database:
  #   customers, orders
  #
  # Table descriptions:
  #   - customers
  # Columns: id, name, email, country
  #
  # - orders
  # Columns: order_id, customer_id, product, amount, order_date
  #
  # Your SQL query will be executed on the database. The results will be sent back
  # to you. After seeing the results, you can either provide a final answer or try
  # another SQL query. When you provide your final answer, do not include any SQL code.
  # --- Receiving response from LLM provider: ---
  # ```sql
  # SELECT DISTINCT country FROM customers;
  # ```
  # --- Sending request to LLM provider (gpt-4o-mini): ---
  # --- SQL code: ---
  # SELECT DISTINCT country FROM customers;
  #
  # --- Query results: ---
  #   country
  # 1     USA
  # 2  Canada
  # 3      UK
  # --- Receiving response from LLM provider: ---
  # Based on the query results, your customers are from the following countries:
  # USA, Canada, and UK.
  # [1] "Based on the query results, your customers are from the following countries:
  # USA, Canada, and UK."
}
}
\seealso{
Other pre_built_prompt_wraps: 
\code{\link{add_image}()},
\code{\link{add_text}()},
\code{\link{answer_as_boolean}()},
\code{\link{answer_as_category}()},
\code{\link{answer_as_dataframe}()},
\code{\link{answer_as_integer}()},
\code{\link{answer_as_json}()},
\code{\link{answer_as_list}()},
\code{\link{answer_as_multi_category}()},
\code{\link{answer_as_named_list}()},
\code{\link{answer_as_numeric}()},
\code{\link{answer_as_regex_match}()},
\code{\link{answer_as_text}()},
\code{\link{answer_by_chain_of_thought}()},
\code{\link{answer_by_react}()},
\code{\link{answer_using_r}()},
\code{\link{answer_using_tools}()},
\code{\link{prompt_wrap}()},
\code{\link{quit_if}()},
\code{\link{set_system_prompt}()}

Other answer_using_prompt_wraps: 
\code{\link{answer_using_r}()},
\code{\link{answer_using_tools}()}
}
\concept{answer_using_prompt_wraps}
\concept{pre_built_prompt_wraps}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/answer_as_category.R
\name{answer_as_category}
\alias{answer_as_category}
\title{Make LLM answer as a category}
\usage{
answer_as_category(prompt, categories, descriptions = NULL)
}
\arguments{
\item{prompt}{A single string or a \code{\link[=tidyprompt]{tidyprompt()}} object}

\item{categories}{A character vector of category names. Must not be empty
and must not contain duplicates}

\item{descriptions}{An optional character vector of descriptions,
corresponding to each category. If provided, its length must match
the length of \code{categories}. Defaults to \code{NULL}}
}
\value{
A \code{\link[=tidyprompt]{tidyprompt()}} with an added \code{\link[=prompt_wrap]{prompt_wrap()}} which
will ensure that the LLM response is the most fitting category of a text as
a character vector of length one
}
\description{
Make LLM answer as a category
}
\details{
For multiple categories, see \code{\link[=answer_as_multi_category]{answer_as_multi_category()}}.
}
\examples{
\dontrun{
  "It is sunny, that makes me happy." |>
    answer_as_category(categories = c("environment", "weather", "work")) |>
    send_prompt()
  # --- Sending request to LLM provider (llama3.1:8b): ---
  #   You need to categorize a text.
  #
  # Text:
  #   'It is sunny, that makes me happy.'
  #
  # Possible categories:
  #   1. environment
  #   2. weather
  #   3. work
  #
  # Respond with the number of the category that best describes the text.
  # (Use no other words or characters.)
  # --- Receiving response from LLM provider: ---
  #   2
  # [1] "weather"
}
}
\seealso{
Other pre_built_prompt_wraps: 
\code{\link{add_image}()},
\code{\link{add_text}()},
\code{\link{answer_as_boolean}()},
\code{\link{answer_as_dataframe}()},
\code{\link{answer_as_integer}()},
\code{\link{answer_as_json}()},
\code{\link{answer_as_list}()},
\code{\link{answer_as_multi_category}()},
\code{\link{answer_as_named_list}()},
\code{\link{answer_as_numeric}()},
\code{\link{answer_as_regex_match}()},
\code{\link{answer_as_text}()},
\code{\link{answer_by_chain_of_thought}()},
\code{\link{answer_by_react}()},
\code{\link{answer_using_r}()},
\code{\link{answer_using_sql}()},
\code{\link{answer_using_tools}()},
\code{\link{prompt_wrap}()},
\code{\link{quit_if}()},
\code{\link{set_system_prompt}()}

Other answer_as_prompt_wraps: 
\code{\link{answer_as_boolean}()},
\code{\link{answer_as_dataframe}()},
\code{\link{answer_as_integer}()},
\code{\link{answer_as_json}()},
\code{\link{answer_as_list}()},
\code{\link{answer_as_multi_category}()},
\code{\link{answer_as_named_list}()},
\code{\link{answer_as_numeric}()},
\code{\link{answer_as_regex_match}()},
\code{\link{answer_as_text}()}
}
\concept{answer_as_prompt_wraps}
\concept{pre_built_prompt_wraps}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/llm_break.R
\name{llm_break}
\alias{llm_break}
\title{Create an \code{llm_break} object}
\usage{
llm_break(object_to_return = NULL, success = FALSE)
}
\arguments{
\item{object_to_return}{The object to return as the response result
from \code{\link[=send_prompt]{send_prompt()}} when this object is returned from an extraction or
validation function}

\item{success}{A logical indicating whether the \code{\link[=send_prompt]{send_prompt()}} loop break
should nonetheless be considered as a successful completion of the
extraction and validation process.
If \code{FALSE}, the \code{object_to_return} must
be \code{NULL} (as the response result of \code{\link[=send_prompt]{send_prompt()}} will always be 'NULL'
when the evaluation was unsuccessful); if \code{FALSE}, \code{\link[=send_prompt]{send_prompt()}} will also
print a warning about the unsuccessful evaluation.
If \code{TRUE}, the \code{object_to_return}
will be returned as the response result of \code{\link[=send_prompt]{send_prompt()}} (and \code{\link[=send_prompt]{send_prompt()}})
will print no warning about unsuccessful evaluation)}
}
\value{
An list of class "llm_break" containing the object to return and
a logical indicating whether the evaluation was successful
}
\description{
This object is used to break a extraction and validation loop defined in a \code{\link[=prompt_wrap]{prompt_wrap()}}
as evaluated by \code{\link[=send_prompt]{send_prompt()}}. When an extraction or validation function returns
this object, the loop will be broken and no further extraction or validation
functions are applied; instead, \code{\link[=send_prompt]{send_prompt()}} will be able to return
the result at that point. This may be useful in scenarios where
it is determined the LLM is unable to provide a response to a prompt.
}
\examples{
# Example usage within an extraction function similar to the one in 'quit_if()':
extraction_fn <- function(x) {
  quit_detect_regex <- "NO ANSWER"

  if (grepl(quit_detect_regex, x)) {
      return(llm_break(
        object_to_return = NULL,
        success = TRUE
      ))
  }

  return(x)
}

\dontrun{
  result <- "How many months old is the cat of my uncle?" |>
    answer_as_integer() |>
    prompt_wrap(
      modify_fn = function(prompt) {
        paste0(
          prompt, "\n\n",
          "Type only 'NO ANSWER' if you do not know."
        )
      },
      extraction_fn = extraction_fn,
      type = "break"
    ) |>
    send_prompt()
  result
  # NULL
}
}
\seealso{
\code{\link[=llm_feedback]{llm_feedback()}}

Other prompt_wrap: 
\code{\link{llm_feedback}()},
\code{\link{prompt_wrap}()}

Other prompt_evaluation: 
\code{\link{llm_feedback}()},
\code{\link{send_prompt}()}
}
\concept{prompt_evaluation}
\concept{prompt_wrap}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/misc_user_verify.R
\name{user_verify}
\alias{user_verify}
\title{Have user check the result of a prompt (human-in-the-loop)}
\usage{
user_verify(prompt)
}
\arguments{
\item{prompt}{A single string or a \linkS4class{tidyprompt} object}
}
\value{
A \linkS4class{tidyprompt} with an added \code{\link[=prompt_wrap]{prompt_wrap()}} which
will add a check for the user to accept or decline the result of the prompt,
providing feedback if the result is declined
}
\description{
This function is used to have a user check the result of a prompt.
After evaluation of the prompt and applying prompt wraps,
the user is presented with the result and asked to accept or decline.
If the user declines, they are asked to provide feedback to the large
language model (LLM) so that the LLM can retry the prompt.
}
\examples{
\dontrun{
  "Tell me a fun fact about yourself!" |>
    user_verify() |>
    send_prompt()
  # --- Sending request to LLM provider (gpt-4o-mini): ---
  # Tell me a fun fact about yourself!
  # --- Receiving response from LLM provider: ---
  # I don't have personal experiences or feelings, but a fun fact about me is that
  # I can generate text in multiple languages! From English to Spanish, French, and
  # more, I'm here to help with diverse linguistic needs.
  #
  # --- Evaluation of tidyprompt resulted in:
  # [1] "I don't have personal experiences or feelings, but a fun fact about me is
  # that I can generate text in multiple languages! From English to Spanish, French,
  # and more, I'm here to help with diverse linguistic needs."
  #
  # --- Accept or decline
  # * If satisfied, type nothing
  # * If not satisfied, type feedback to the LLM
  # Type: Needs to be funnier!
  # --- Sending request to LLM provider (gpt-4o-mini): ---
  # Needs to be funnier!
  # --- Receiving response from LLM provider: ---
  # Alright, how about this: I once tried to tell a joke, but my punchline got lost
  # in translation! Now, I just stick to delivering 'byte-sized' humor!
  #
  # --- Evaluation of tidyprompt resulted in:
  # [1] "Alright, how about this: I once tried to tell a joke, but my punchline got
  # lost in translation! Now, I just stick to delivering 'byte-sized' humor!"
  #
  # --- Accept or decline
  # * If satisfied, type nothing
  # * If not satisfied, type feedback to the LLM
  # Type:
  # * Result accepted
  # [1] "Alright, how about this: I once tried to tell a joke, but my punchline got
  # lost in translation! Now, I just stick to delivering 'byte-sized' humor!"
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/misc_llm_verify.R
\name{llm_verify}
\alias{llm_verify}
\title{Have LLM check the result of a prompt (LLM-in-the-loop)}
\usage{
llm_verify(
  prompt,
  question = "Is the answer satisfactory?",
  llm_provider = NULL,
  max_words_feedback = 50
)
}
\arguments{
\item{prompt}{A single string or a \linkS4class{tidyprompt} object}

\item{question}{The question to ask the LLM to verify the result of the prompt.
The LLM will be presented the original prompt, its result, and this question.
The LLM will be asked to provide a boolean answer to this question. If TRUE,
the result of the prompt will be accepted; if FALSE, the result will be declined}

\item{llm_provider}{A \linkS4class{llm_provider} object
which will be used to verify the evaluation led to a satisfactory result.
If not provided, the same LLM provider as the prompt was originally
evaluated with will be used}

\item{max_words_feedback}{The maximum number of words allowed in the summary of
why the result was declined.
This summary is sent back to the LLM originally asked to evaluate the prompt}
}
\value{
A \linkS4class{tidyprompt} with an added \code{\link[=prompt_wrap]{prompt_wrap()}} which
will add a check for a LLM to accept or decline the result of the prompt,
providing feedback if the result is declined
}
\description{
This function will wrap a prompt with a check for a LLM to accept or decline
the result of the prompt, providing feedback if the result is declined.
The evaluating LLM will be presented with the original prompt and the result
of the prompt, and will be asked to verify if the answer is satisfactory
(using chain of thought reasoning to arrive at a boolean decision). If
the result is declined, the chain of thought responsible for the decision
will be summarized and sent back to the original LLM that was asked to
evaluate the prompt, so that it may retry the prompt.

Note that this function is experimental and, because it relies on chain of
thought reasoning by a LLM about the answer of another LLM, it may not
always provide accurate results and can increase the token cost of
evaluating a prompt.
}
\details{
The original prompt text shown to the LLM is built from
the base prompt as well as all prompt wraps that have a modify function
but do not have an extraction or validation function. This is to ensure
that no redundant validation is performed by the evaluating LLM on
instructions which have already been validated by functions in those
prompt wraps.
}
\examples{
\dontrun{
  "What is 'Enschede'?!" |>
    answer_as_text(max_words = 50) |>
    llm_verify() |>
    send_prompt()
#   --- Sending request to LLM provider (gpt-4o-mini): ---
#   What is 'Enschede'?!
#
#   You must provide a text response. The response must be at most 50 words.
#   --- Receiving response from LLM provider: ---
#   Enschede is a city in the Netherlands, located in the eastern part near the German border.
#   It is known for its vibrant culture, history, and universities, particularly the
#   University of Twente, as well as its textiles industry and beautiful parks.
#   --- Sending request to LLM provider (gpt-4o-mini): ---
#   You are given a user's prompt.
#   To answer the user's prompt, you need to think step by step to arrive at a final answer.
#
#   ----- START OF USER'S PROMPT -----
#   >>> An assistant was asked:
#
#   What is 'Enschede'?!
#
#   >>> The assistant answered:
#
#   [1] "Enschede is a city in the Netherlands, located in the eastern part near
#   the German border. It is known for its vibrant culture, history, and universities,
#   particularly the University of Twente, as well as its textiles industry and
#   beautiful parks."
#
#   >>> Is the answer satisfactory?
#
#   You must answer with only TRUE or FALSE (use no other characters).
#   ----- END OF USER'S PROMPT -----
#
#   What are the steps you would take to answer the user's prompt?
#   Describe your thought process in the following format:
#     >> step 1: <step 1 description>
#     >> step 2: <step 2 description>
#     (etc.)
#
#   When you are done, you must type:
#     FINISH[<put here your final answer to the user's prompt>]
#
#   Make sure your final answer follows the logical conclusion of your thought process.
#   --- Receiving response from LLM provider: ---
#   >> step 1: Read the user's prompt carefully to understand what is being asked
#   regarding the assistant's answer about Enschede.
#   >> step 2: Analyze the assistant's provided answer for accuracy and completeness,
#   ensuring it conveys relevant information about Enschede.
#   >> step 3: Assess whether the answer includes key aspects, such as the city's
#   location, cultural significance, educational institutions, and any notable
#   historical references or industries tied to the city.
#   >> step 4: Determine if the assistant's answer sufficiently addresses the
#   question "What is Enschede?" and provides a clear and informative overview of the city.
#   >> step 5: Conclude whether the assistant's answer meets the standards for being
#   satisfactory in terms of detail and correctness.
#
#   FINISH[TRUE]
#   [1] "Enschede is a city in the Netherlands, located in the eastern part near the
#   German border. It is known for its vibrant culture, history, and universities,
#   particularly the University of Twente, as well as its textiles industry and beautiful parks."
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/misc_quit_if.R
\name{quit_if}
\alias{quit_if}
\title{Make evaluation of a prompt stop if LLM gives a specific response}
\usage{
quit_if(
  prompt,
  quit_detect_regex = "NO ANSWER",
  instruction =
    paste0("If you think that you cannot provide a valid answer, you must type:\\n",
    "'NO ANSWER' (use no other characters)"),
  success = TRUE,
  response_result = c("null", "llm_response", "regex_match")
)
}
\arguments{
\item{prompt}{A single string or a \code{\link[=tidyprompt]{tidyprompt()}} object}

\item{quit_detect_regex}{A regular expression to detect in the LLM's
response which will cause the evaluation to stop. The default
will detect the string "NO ANSWER" in the response}

\item{instruction}{A string to be added to the prompt to instruct the LLM
how to respond if it cannot answer the prompt. The default is
"If you think that you cannot provide a valid answer, you must type: 'NO ANSWER' (use no other characters)".
This parameter can be set to \code{NULL} if no instruction is needed in the prompt}

\item{success}{A logical indicating whether the \code{\link[=send_prompt]{send_prompt()}} loop break
should nonetheless be considered as a successful completion of the
extraction and validation process. If \code{FALSE}, the \code{object_to_return} must
will always be set to NULL and thus parameter 'response_result' must also
be set to 'null'; if \code{FALSE}, \code{\link[=send_prompt]{send_prompt()}} will also print a warning
about the unsuccessful evaluation. If \code{TRUE}, the \code{object_to_return} will be
returned as the response result of \code{\link[=send_prompt]{send_prompt()}} (and \code{\link[=send_prompt]{send_prompt()}}
will print no warning about unsuccessful evaluation); parameter 'response_result'
will then determine what is returned as the response result of \code{\link[=send_prompt]{send_prompt()}}.}

\item{response_result}{A character string indicating what should be returned
when the quit_detect_regex is detected in the LLM's response. The default is
'null', which will return NULL as the response result o f \code{\link[=send_prompt]{send_prompt()}}.
Under 'llm_response', the full LLM response will be returned as the
response result of \code{\link[=send_prompt]{send_prompt()}}.
Under 'regex_match', the part of the LLM response that matches the
quit_detect_regex will be returned as the response result of \code{\link[=send_prompt]{send_prompt()}}}
}
\value{
A \code{\link[=tidyprompt]{tidyprompt()}} with an added \code{\link[=prompt_wrap]{prompt_wrap()}} which will ensure
that the evaluation will stop upon detection of the quit_detect_regex in the
LLM's response
}
\description{
This function is used to wrap a \code{\link[=tidyprompt]{tidyprompt()}} object and ensure that the
evaluation will stop if the LLM says it cannot answer the prompt. This is
useful in scenarios where it is determined the LLM is unable to provide a
response to a prompt.
}
\examples{
\dontrun{
  "What the favourite food of my cat on Thursday mornings?" |>
    quit_if() |>
    send_prompt(llm_provider_ollama())
  # --- Sending request to LLM provider (llama3.1:8b): ---
  #   What the favourite food of my cat on Thursday mornings?
  #
  #   If you think that you cannot provide a valid answer, you must type:
  #   'NO ANSWER' (use no other characters)
  # --- Receiving response from LLM provider: ---
  #   NO ANSWER
  # NULL
}
}
\seealso{
Other pre_built_prompt_wraps: 
\code{\link{add_image}()},
\code{\link{add_text}()},
\code{\link{answer_as_boolean}()},
\code{\link{answer_as_category}()},
\code{\link{answer_as_dataframe}()},
\code{\link{answer_as_integer}()},
\code{\link{answer_as_json}()},
\code{\link{answer_as_list}()},
\code{\link{answer_as_multi_category}()},
\code{\link{answer_as_named_list}()},
\code{\link{answer_as_numeric}()},
\code{\link{answer_as_regex_match}()},
\code{\link{answer_as_text}()},
\code{\link{answer_by_chain_of_thought}()},
\code{\link{answer_by_react}()},
\code{\link{answer_using_r}()},
\code{\link{answer_using_sql}()},
\code{\link{answer_using_tools}()},
\code{\link{prompt_wrap}()},
\code{\link{set_system_prompt}()}

Other miscellaneous_prompt_wraps: 
\code{\link{add_image}()},
\code{\link{add_text}()},
\code{\link{set_system_prompt}()}
}
\concept{miscellaneous_prompt_wraps}
\concept{pre_built_prompt_wraps}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/answer_using_tools.R
\name{tools_get_docs}
\alias{tools_get_docs}
\title{Extract documentation from a function}
\usage{
tools_get_docs(func, name = NULL)
}
\arguments{
\item{func}{A function object. The function should belong to a package
and have documentation available in a help file, or it should
have documentation added by \code{\link[=tools_add_docs]{tools_add_docs()}}}

\item{name}{The name of the function if already known (optional).
If not provided it will be extracted from the documentation or the
function object's name}
}
\value{
A list with documentation for the function. See \code{\link[=tools_add_docs]{tools_add_docs()}}
for more information on the contents
}
\description{
This function extracts documentation from a help file (if available,
i.e., when the function is part of a package) or from documentation added
by \code{\link[=tools_add_docs]{tools_add_docs()}}. The extracted documentation includes
the function's name, description, arguments, and return value.
This information is used to provide an LLM with information about the functions,
so that the LLM can call R functions.
}
\details{
This function will prioritize documentation added by
\code{\link[=tools_add_docs]{tools_add_docs()}} over documentation from a help file.
Thus, it is possible to override the help file documentation by adding
custom documentation
}
\examples{
# When using functions from base R or R packages,
#   documentation is automatically extracted from help files:
prompt_with_dir_function <- "What are the files in my current directory?" |>
  answer_using_tools(dir) # The 'dir' function is from base R
\dontrun{
  send_prompt(prompt_with_dir_function)
  # --- Sending request to LLM provider (llama3.1:8b): ---
  #   What are the files in my current directory?
  #   --- Receiving response from LLM provider: ---
  #   Calling function 'nm' with arguments:
  #   {
  #     "all.files": true,
  #     "full.names": false,
  #     "ignore.case": false,
  #     "include.dirs": false,
  #     "no..": false,
  #     "path": "./",
  #     "pattern": "*",
  #     "recursive": false
  #   }
  # Result:
  #   .git, .github, .gitignore, .Rbuildignore, .Rhistory, ...
  # The files in your current directory are:
  #   .git, .github, .gitignore, .Rbuildignore, .Rhistory, ...
  # [1] "The files in your current directory are:\n\n .git, .github, ..."
}

# Users may provide custom functions in two ways:
#   1) as a function object, optionally documented with `tools_get_docs()`, or
#   2) as an 'ellmer' tool definition, using `ellmer::tool()`

# Take this fake weather function as an example:
temperature_in_location <- function(
  location = c("Amsterdam", "Utrecht", "Enschede"),
  unit = c("Celcius", "Fahrenheit")
) {
  location <- match.arg(location)
  unit <- match.arg(unit)

  temperature_celcius <- switch(
    location,
    "Amsterdam" = 32.5,
    "Utrecht" = 19.8,
    "Enschede" = 22.7
  )

  if (unit == "Celcius") {
    return(temperature_celcius)
  } else {
    return(temperature_celcius * 9/5 + 32)
  }
}


# 1: `tools_add_docs()` --------------------------------------------------------

# Generate documentation for a function, based on formals & help file
docs <- tools_get_docs(temperature_in_location)

# The types get inferred from the function's formals
# However, descriptions are still missing as the function is not from a package
# We can modify the documentation object to add descriptions:
docs$description <- "Get the temperature in a location"
docs$arguments$unit$description <- "Unit in which to return the temperature"
docs$arguments$location$description <- "Location for which to return the temperature"
docs$return$description <- "The temperature in the specified location and unit"
# (See `?tools_add_docs` for more details on the structure of the documentation)

# When we are satisfied with the documentation, we can add it to the function:
temperature_in_location <- tools_add_docs(temperature_in_location, docs)

prompt_with_weather_function <-
  "What is the weather in Enschede? Give me Celcius degrees" |>
  answer_using_tools(temperature_in_location)
\dontrun{
  send_prompt(prompt_with_weather_function)
  # --- Sending request to LLM provider (llama3.1:8b): ---
  #   What is the weather in Enschede? Give me Celcius degrees
  #   --- Receiving response from LLM provider: ---
  #   Calling function 'temperature_in_location' with arguments:
  #   {
  #     "location": "Enschede",
  #     "unit": "Celcius"
  #   }
  # Result:
  #   22.7
  # The temperature in Enschede is 22.7 Celcius degrees.
  # [1] "The temperature in Enschede is 22.7 Celcius degrees."
}

# 2: `ellmer::tool()` -----------------------------------------------

# Alternatively, we can define the function as an 'ellmer' tool

temperature_in_location_ellmer <- ellmer::tool(
  temperature_in_location,
  name = "get_temperature",
  description = "Get the temperature in a location",
  arguments = list(
    location = ellmer::type_string(
      "Location for which to return the temperature", required = TRUE
    ),
    unit = ellmer::type_string(
      "Unit in which to return the temperature", required = TRUE
    )
  )
)

prompt_with_weather_function_ellmer <-
  "What is the weather in Utrecht? Give me Fahrenheit degrees" |>
  answer_using_tools(temperature_in_location_ellmer)
\dontrun{
  send_prompt(prompt_with_weather_function_ellmer)
  # ...
}

# Because `mcptools::mcp_tools()` returns a list of `ellmer:tool()` tools,
#   you can also use Model Context Protocol (MCP) server tools with
#   `answer_using_tools()`:
\dontrun{
  prompt_using_mcp_tools <- mcptools::mcp_tools()
  "Push my latest commit to GitHub" |>
    answer_using_tools(mcp_tools)
  send_prompt(prompt_using_mcp_tools)
}

# `answer_using_tools()` will automatically attempt to use the most appropriate
#   way of sending the tool to the LLM

# If you use a LLM provider of type 'ollama' or 'openai',
#   it will automatically convert the tool definition to parameters
#   appropriate for those APIs
# If you use a LLM provider of type 'ellmer', it will call the appropriate
#   ellmer function directly which will handle the tool call for various
#   providers
# Note that both tool definitions from `tools_add_docs()` and `ellmer::tool()`
#   will work with any LLM provider; `answer_using_tools()` can convert
#   the two types of tool definitions to each other when needed
\dontrun{
  ollama <- llm_provider_ollama()
  # Ollama LLM provider:
  "What is the weather in Amsterdam? Give me Fahrenheit degrees" |>
    answer_using_tools(temperature_in_location) |>
    send_prompt(ollama)

  # Ollama LLM provider also works with `ellmer::tool()` definitions:
  "What is the weather in Amsterdam? Give me Celcius degrees" |>
    answer_using_tools(temperature_in_location_ellmer) |>
    send_prompt(ollama)

  # Similar for OpenAI API:
  openai <- llm_provider_openai()
  "What is the weather in Amsterdam? Give me Celcius degrees" |>
    answer_using_tools(temperature_in_location) |>
    send_prompt(openai)
  # ...

  # Ellmer LLM provider:
  ellmer <- llm_provider_ellmer(ellmer::chat_openai())
  "What is the weather in Amsterdam? Give me Celcius degrees" |>
    answer_using_tools(temperature_in_location_ellmer) |>
    send_prompt(ellmer)

  # Also works with `tools_add_docs()` definition:
  "What is the weather in Amsterdam? Give me Celcius degrees" |>
    answer_using_tools(temperature_in_location) |>
    send_prompt(ellmer)
}
}
\seealso{
Other tools: 
\code{\link{answer_using_tools}()},
\code{\link{tools_add_docs}()}
}
\concept{tools}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/prompt_wrap.R
\name{provider_prompt_wrap}
\alias{provider_prompt_wrap}
\title{Create a provider-level prompt wrap}
\usage{
provider_prompt_wrap(
  modify_fn = NULL,
  extraction_fn = NULL,
  validation_fn = NULL,
  handler_fn = NULL,
  parameter_fn = NULL,
  type = c("unspecified", "mode", "tool", "break", "check"),
  name = NULL
)
}
\arguments{
\item{modify_fn}{A function that takes the previous prompt text (as
first argument) and returns the new prompt text}

\item{extraction_fn}{A function that takes the LLM response (as first argument)
and attempts to extract a value from it.
Upon succesful extraction, the function should return the extracted value.
If the extraction fails, the function should return a \code{\link[=llm_feedback]{llm_feedback()}} message
to initiate a retry.
A \code{\link[=llm_break]{llm_break()}} can be returned to break the extraction and validation loop,
ending \code{\link[=send_prompt]{send_prompt()}}}

\item{validation_fn}{A function that takes the (extracted) LLM response
(as first argument) and attempts to validate it.
Upon succesful validation, the function should return TRUE. If the validation
fails, the function should return a \code{\link[=llm_feedback]{llm_feedback()}} message to initiate a retry.
A \code{\link[=llm_break]{llm_break()}} can be returned to break the extraction and validation loop,
ending \code{\link[=send_prompt]{send_prompt()}}}

\item{handler_fn}{A function that takes a 'completion' object (a result
of a request to a LLM, as returned by \verb{$complete_chat()} of a \linkS4class{llm_provider}
object) as first argument and the \linkS4class{llm_provider} object as second argument.
The function should return a (modified or identical) completion object.
This can be used for advanced side effects, like logging, or native tool calling,
or keeping track of token usage. See \linkS4class{llm_provider} for more information;
handler_fn is attached to the \linkS4class{llm_provider} object that is being used.
When using an \code{\link[=llm_provider_ellmer]{llm_provider_ellmer()}}, the up-to-date ellmer_chat is synced
onto the provider before handlers run. This allows handlers to access,
for instance, the current cost of the conversation, and, for instance,
to stop the conversation if a certain budget is exceeded.
For example usage, see source code of \code{\link[=answer_using_tools]{answer_using_tools()}}}

\item{parameter_fn}{A function that takes the \linkS4class{llm_provider} object
which is being used with \code{\link[=send_prompt]{send_prompt()}} and returns a named list of parameters
to be set in the \linkS4class{llm_provider} object via its \verb{$set_parameters()} method.
This can be used to configure specific parameters of the \linkS4class{llm_provider}
object when evaluating the prompt.
For example, \code{\link[=answer_as_json]{answer_as_json()}} may set different parameters for different APIs
related to JSON output.
This function is typically only used with advanced prompt wraps that require
specific settings in the \linkS4class{llm_provider} object}

\item{type}{The type of prompt wrap. Must be one of:
\itemize{
\item "unspecified": The default type, typically used for prompt wraps
which request a specific format of the LLM response, like \code{\link[=answer_as_integer]{answer_as_integer()}}

\item "mode": For prompt wraps that change how the LLM should answer the prompt,
like \code{\link[=answer_by_chain_of_thought]{answer_by_chain_of_thought()}} or \code{\link[=answer_by_react]{answer_by_react()}}

\item "tool": For prompt wraps that enable the LLM to use tools, like \code{\link[=answer_using_tools]{answer_using_tools()}}
or \code{\link[=answer_using_r]{answer_using_r()}} when 'output_as_tool' = TRUE

\item "break": For prompt wraps that may break the extraction and validation loop,
like \code{\link[=quit_if]{quit_if()}}. These are applied before type "unspecified" as they may
instruct the LLM to not answer the prompt in the manner specified by those
prompt wraps

\item "check": For prompt wraps that apply a last check to the final answer,
after all other prompt wraps have been evaluated.
These prompt wraps may only contain a validation function, and are applied
after all other prompt wraps have been evaluated. These prompt wraps are
even applied after an earlier prompt wrap has broken the extraction and validation loop
with \code{\link[=llm_break]{llm_break()}}
}
Types are used to determine the order in which prompt wraps are applied.
When constructing the prompt text, prompt wraps are applied to the base prompt
in the following order: 'check', 'unspecified', 'break', 'mode', 'tool'.
When evaluating the LLM response and applying extraction and validation functions,
prompt wraps are applied in the reverse order: 'tool', 'mode', 'break',
'unspecified', 'check'.
Order among the same type is preserved in the order they were added to the prompt.}

\item{name}{An optional name for the prompt wrap.
This can be used to identify the prompt wrap in the \linkS4class{tidyprompt} object}
}
\value{
A \code{provider_prompt_wrap} object, to be stored on an \linkS4class{llm_provider} object
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}}
Build a provider-specific prompt wrap, to store on an \linkS4class{llm_provider} object
(with \verb{$add_prompt_wrap()}). These prompt wraps can be applied before or
after any prompt-specific prompt wraps. In this way, you can ensure that
certain prompt wraps are always applied when using a specific LLM provider.
}
\examples{
ollama <- llm_provider_ollama()

# Add a "short answer" mode (provider-level post prompt wrap)
ollama$add_prompt_wrap(
  provider_prompt_wrap(
    modify_fn = \(txt) paste0(
      txt,
      "\n\nPlease answer concisely (< 2 sentences)."
    )
  ),
  position = "post"
)

# Use as usual: wraps are applied automatically
\dontrun{
"What's a vignette in R?" |> send_prompt(ollama)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/answer_by_chain_of_thought.R
\name{answer_by_chain_of_thought}
\alias{answer_by_chain_of_thought}
\title{Set chain of thought mode for a prompt}
\usage{
answer_by_chain_of_thought(
  prompt,
  extract_from_finish_brackets = TRUE,
  extraction_lenience = TRUE
)
}
\arguments{
\item{prompt}{A single string or a \code{\link[=tidyprompt]{tidyprompt()}} object}

\item{extract_from_finish_brackets}{A logical indicating whether the final answer
should be extracted from the text inside the "FINISH\link{...}" brackets.}

\item{extraction_lenience}{A logical indcating whether the extraction function should be lenient.
If TRUE, the extraction function will attempt to extract the final answer
even if it cannot be extracted from within the brackets, by extracting
everything after the final occurence of 'FINISH' (if present). This may
be useful for smaller LLMs which may not follow the output format
as strictly}
}
\value{
A \code{\link[=tidyprompt]{tidyprompt()}} with an added \code{\link[=prompt_wrap]{prompt_wrap()}} which will ensure
that the LLM follows the chain of thought mode in answering the prompt
}
\description{
This function enables chain of thought mode for evaluation of a prompt
or a \code{\link[=tidyprompt]{tidyprompt()}}. In chain of thought mode, the large language model (LLM)
In chain of thought mode, the large language model (LLM) is asked to think
step by step to arrive at a final answer. It is hypothesized that this may
increase LLM performance at solving complex tasks. Chain of thought mode
is inspired by the method described in Wei et al. (2022).
}
\examples{
\dontrun{
  "What is 5 + 5?" |>
    answer_by_chain_of_thought() |>
    answer_as_integer() |>
    send_prompt()
  # --- Sending request to LLM provider (llama3.1:8b): ---
  # You are given a user's prompt.
  # To answer the user's prompt, you need to think step by step to arrive at a final answer.
  #
  # ----- START OF USER'S PROMPT -----
  # What is 5 + 5?
  #
  # You must answer with only an integer (use no other characters).
  # ----- END OF USER'S PROMPT -----
  #
  # What are the steps you would take to answer the user's prompt?
  # Describe your thought process in the following format:
  #   >> step 1: <step 1 description>
  #   >> step 2: <step 2 description>
  #   (etc.)
  #
  # When you are done, you must type:
  #   FINISH[<put here your final answer to the user's prompt>]
  #
  # Make sure your final answer follows the logical conclusion of your thought process.
  # --- Receiving response from LLM provider: ---
  # >> step 1: Recognize that the user's prompt is a simple arithmetic question asking for
  # the sum of two numbers.
  #
  # >> step 2: Identify the specific numbers mentioned in the prompt, which are 5 and 5.
  #
  # >> step 3: Recall basic addition facts from memory or knowledge.
  # In this case, knowing that 5 + 0 = 5, 5 + 1 = 6, 5 + 2 = 7, and so on up to 5 + 4 = 9,
  # it is clear that the next step in adding 5 to itself would be to keep the same sum.
  #
  # >> step 4: Apply the addition principle by considering the original number (5) as being
  # added to itself once more. Since 5 + 1 = 6 and the process of adding 1 increases the total
  # sum, then 5 + 2 must equal 7, thus indicating a direct progression in the sum as the additive
  # amount increases.
  #
  # >> step 5: From steps 3 and 4, conclude that since 5 added to itself (once) would be 10,
  # then when it is added again (to itself), it would continue from the last result which was
  # achieved by adding one more than the original number. Therefore, it must follow the pattern
  # established in step 3 where each additional amount increases the total sum.
  #
  # >> step 6: Now knowing that the operation of adding a second 5 to the first one results in
  # increasing the sum to the next integer after 10 (based on steps 3 and 4), determine what the
  # next integer is after 10. Knowing that integers are whole numbers without fractions, it's
  # clear that if you add another unit to 10, it becomes 11.
  #
  # FINISH[10]
  # [1] 10
}
}
\references{
Wei, J., Wang, X., Schuurmans, D., Bosma, M., Ichter, B., Xia, F., Chi, E., Le, Q., & Zhou, D. (2022).
Chain-of-Thought Prompting Elicits Reasoning in Large Language Models.
\url{doi:10.48550/arXiv.2201.11903}
}
\seealso{
Other pre_built_prompt_wraps: 
\code{\link{add_image}()},
\code{\link{add_text}()},
\code{\link{answer_as_boolean}()},
\code{\link{answer_as_category}()},
\code{\link{answer_as_dataframe}()},
\code{\link{answer_as_integer}()},
\code{\link{answer_as_json}()},
\code{\link{answer_as_list}()},
\code{\link{answer_as_multi_category}()},
\code{\link{answer_as_named_list}()},
\code{\link{answer_as_numeric}()},
\code{\link{answer_as_regex_match}()},
\code{\link{answer_as_text}()},
\code{\link{answer_by_react}()},
\code{\link{answer_using_r}()},
\code{\link{answer_using_sql}()},
\code{\link{answer_using_tools}()},
\code{\link{prompt_wrap}()},
\code{\link{quit_if}()},
\code{\link{set_system_prompt}()}

Other answer_by_prompt_wraps: 
\code{\link{answer_by_react}()}
}
\concept{answer_by_prompt_wraps}
\concept{pre_built_prompt_wraps}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/chat_history.R
\name{add_msg_to_chat_history}
\alias{add_msg_to_chat_history}
\title{Add a message to a chat history}
\usage{
add_msg_to_chat_history(
  chat_history,
  message,
  role = c("auto", "user", "assistant", "system", "tool"),
  tool_result = NULL
)
}
\arguments{
\item{chat_history}{A single string, a data.frame which is a valid chat history
(see \verb{[chat_history()]}), a list containing a valid chat history under key
'$chat_history', a \link{tidyprompt-class} object, or NULL

A \code{\link[=chat_history]{chat_history()}} object}

\item{message}{A character string representing the message to add}

\item{role}{A character string representing the role of the message sender.
One of: \itemize{
\item "auto": The function automatically determines the role. If the last message
was from the user, the role will be "assistant". If the last message was from anything
else, the role will be "user"
\item "user": The message is from the user
\item "assistant": The message is from the assistant
\item "system": The message is from the system
\item "tool": The message is from a tool (e.g., indicating the result of a function call)
}}

\item{tool_result}{A logical indicating whether the message is a tool result
(e.g., the result of a function call)}
}
\value{
A \code{\link[=chat_history]{chat_history()}} object with the message added as the last row
}
\description{
This function appends a message to a \code{\link[=chat_history]{chat_history()}} object.
The function can automatically determine the role of the message to be
added based on the last message in the chat history. The role can also be
manually specified.
}
\details{
The chat_history object may be of different types:
\itemize{
\item A single string: The function will create a new chat history object
with the string as the first message; the role of that first message will be
"user"
\item A data.frame: The function will append the message to the data.frame.
The data.frame must be a valid chat history; see \code{\link[=chat_history]{chat_history()}}
\item A list: The function will extract the chat history from the list.
The list must contain a valid chat history under the key 'chat_history'.
This may typically be the result from \code{\link[=send_prompt]{send_prompt()}} when using
'return_mode = "full"'
\item A Tidyprompt object (\linkS4class{tidyprompt}): The function will extract the chat history
from the object. This will be done by concatenating the 'system_prompt',
'chat_history', and 'base_prompt' into a chat history data.frame. Note
that the other properties of the \linkS4class{tidyprompt} object will be lost
\item NULL: The function will create a new chat history object
with no messages; the message will be the first message
}
}
\examples{
chat <- "Hi there!" |>
  chat_history()
chat

chat_from_df <- data.frame(
  role = c("user", "assistant"),
  content = c("Hi there!", "Hello! How can I help you today?")
) |>
  chat_history()
chat_from_df

# `add_msg_to_chat_history()` may be used to add messages to a chat history
chat_from_df <- chat_from_df |>
  add_msg_to_chat_history("Calculate 2+2 for me, please!")
chat_from_df

# You can also continue conversations which originate from `send_prompt()`:
\dontrun{
  result <- "Hi there!" |>
    send_prompt(return_mode = "full")
  # --- Sending request to LLM provider (llama3.1:8b): ---
  # Hi there!
  # --- Receiving response from LLM provider: ---
  # It's nice to meet you. Is there something I can help you with, or would you
  # like to chat?

  # Access the chat history from the result:
  chat_from_send_prompt <- result$chat_history

  # Add a message to the chat history:
  chat_history_with_new_message <- chat_from_send_prompt |>
    add_msg_to_chat_history("Let's chat!")

  # The new chat history can be input for a new tidyprompt:
  prompt <- tidyprompt(chat_history_with_new_message)

  # You can also take an existing tidyprompt and add the new chat history to it;
  #   this way, you can continue a conversation using the same prompt wraps
  prompt$set_chat_history(chat_history_with_new_message)

  # send_prompt() also accepts a chat history as input:
  new_result <- chat_history_with_new_message |>
    send_prompt(return_mode = "full")

  # You can also create a persistent chat history object from
  #   a chat history data frame; see ?`persistent_chat-class`
  chat <- `persistent_chat-class`$new(llm_provider_ollama(), chat_from_send_prompt)
  chat$chat("Let's chat!")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/answer_as_integer.R
\name{answer_as_integer}
\alias{answer_as_integer}
\title{Make LLM answer as an integer (between min and max)}
\usage{
answer_as_integer(
  prompt,
  min = NULL,
  max = NULL,
  add_instruction_to_prompt = TRUE
)
}
\arguments{
\item{prompt}{A single string or a \code{\link[=tidyprompt]{tidyprompt()}} object}

\item{min}{(optional) Minimum value for the integer}

\item{max}{(optional) Maximum value for the integer}

\item{add_instruction_to_prompt}{(optional) Add instruction for replying
as an integer to the prompt text. Set to FALSE for debugging if extractions/validations
are working as expected (without instruction the answer should fail the
validation function, initiating a retry)}
}
\value{
A \code{\link[=tidyprompt]{tidyprompt()}} with an added \code{\link[=prompt_wrap]{prompt_wrap()}} which
will ensure that the LLM response is an integer.
}
\description{
Make LLM answer as an integer (between min and max)
}
\examples{
\dontrun{
  "What is 5 + 5?" |>
    answer_as_integer() |>
    send_prompt()
  # --- Sending request to LLM provider (llama3.1:8b): ---
  #   What is 5 + 5?
  #
  #   You must answer with only an integer (use no other characters).
  # --- Receiving response from LLM provider: ---
  #   10
  # [1] 10
}
}
\seealso{
Other pre_built_prompt_wraps: 
\code{\link{add_image}()},
\code{\link{add_text}()},
\code{\link{answer_as_boolean}()},
\code{\link{answer_as_category}()},
\code{\link{answer_as_dataframe}()},
\code{\link{answer_as_json}()},
\code{\link{answer_as_list}()},
\code{\link{answer_as_multi_category}()},
\code{\link{answer_as_named_list}()},
\code{\link{answer_as_numeric}()},
\code{\link{answer_as_regex_match}()},
\code{\link{answer_as_text}()},
\code{\link{answer_by_chain_of_thought}()},
\code{\link{answer_by_react}()},
\code{\link{answer_using_r}()},
\code{\link{answer_using_sql}()},
\code{\link{answer_using_tools}()},
\code{\link{prompt_wrap}()},
\code{\link{quit_if}()},
\code{\link{set_system_prompt}()}

Other answer_as_prompt_wraps: 
\code{\link{answer_as_boolean}()},
\code{\link{answer_as_category}()},
\code{\link{answer_as_dataframe}()},
\code{\link{answer_as_json}()},
\code{\link{answer_as_list}()},
\code{\link{answer_as_multi_category}()},
\code{\link{answer_as_named_list}()},
\code{\link{answer_as_numeric}()},
\code{\link{answer_as_regex_match}()},
\code{\link{answer_as_text}()}
}
\concept{answer_as_prompt_wraps}
\concept{pre_built_prompt_wraps}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/helper_text_vector_list_to_string.R
\name{vector_list_to_string}
\alias{vector_list_to_string}
\title{Convert a named or unnamed list/vector to a string representation}
\usage{
vector_list_to_string(obj, how = c("inline", "expanded"))
}
\arguments{
\item{obj}{A list or vector (named or unnamed) to be converted to a string.}

\item{how}{In what way the object should be converted to a string;
either "inline" or "expanded". "inline" presents all key-value pairs
or values as a single line. "expanded" presents each key-value pair or value
on a separate line.}
}
\value{
A single string representing the list/vector.
}
\description{
Converts a named or unnamed list/vector to a string format, intended for
sending it to an LLM (or for display or logging).
}
\examples{
named_vector <- c(x = 10, y = 20, z = 30)

vector_list_to_string(named_vector, how = "inline")

vector_list_to_string(named_vector, how = "expanded")
}
\seealso{
Other text_helpers: 
\code{\link{df_to_string}()},
\code{\link{skim_with_labels_and_levels}()}
}
\concept{text_helpers}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/answer_as_key_value.R
\name{answer_as_key_value}
\alias{answer_as_key_value}
\title{Make LLM answer as a list of key-value pairs}
\usage{
answer_as_key_value(
  prompt,
  key_name = "key",
  value_name = "value",
  pair_explanation = NULL,
  n_unique_items = NULL,
  list_mode = c("bullet", "comma")
)
}
\arguments{
\item{prompt}{A single string or a \code{\link[=tidyprompt]{tidyprompt()}} object}

\item{key_name}{(optional) A name or placeholder describing the "key" part
of each pair}

\item{value_name}{(optional) A name or placeholder describing the "value" part
of each pair}

\item{pair_explanation}{(optional) Additional explanation of what a pair should
be. It should be a single string. It will be appended after the list instruction.}

\item{n_unique_items}{(optional) Number of unique key-value pairs required in the list}

\item{list_mode}{(optional) Mode of the list: "bullet" or "comma".
\itemize{
\item "bullet" mode expects pairs like:

\if{html}{\out{<div class="sourceCode">}}\preformatted{-- key1: value1
-- key2: value2
}\if{html}{\out{</div>}}
\item "comma" mode expects pairs like:

\if{html}{\out{<div class="sourceCode">}}\preformatted{1. key: value, 2. key: value, etc.
}\if{html}{\out{</div>}}
}}
}
\value{
A \code{\link[=tidyprompt]{tidyprompt()}} with an added \code{\link[=prompt_wrap]{prompt_wrap()}} which
will ensure that the LLM response is a list of key-value pairs.
}
\description{
This function is similar to \code{answer_as_list()} but instead of returning
a list of items, it instructs the LLM to return a list of key-value pairs.
}
\examples{
\dontrun{
  "What are a few capital cities around the world?" |>
    answer_as_key_value(
      key_name = "country",
      value_name = "capital"
    ) |>
    send_prompt()
  # --- Sending request to LLM provider (llama3.1:8b): ---
  # What are a few capital cities around the world?
  #
  # Respond with a list of key-value pairs, like so:
  #   -- <<country 1>>: <<capital 1>>
  #   -- <<country 2>>: <<capital 2>>
  #   etc.
  # --- Receiving response from LLM provider: ---
  # Here are a few:
  #   -- Australia: Canberra
  #   -- France: Paris
  #   -- United States: Washington D.C.
  #   -- Japan: Tokyo
  #   -- China: Beijing
  # $Australia
  # [1] "Canberra"
  #
  # $France
  # [1] "Paris"
  #
  # $`United States`
  # [1] "Washington D.C."
  #
  # $Japan
  # [1] "Tokyo"
  #
  # $China
  # [1] "Beijing"
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tidyprompt.R
\name{set_chat_history}
\alias{set_chat_history}
\title{Set the chat history of a \linkS4class{tidyprompt} object}
\usage{
set_chat_history(x, chat_history)
}
\arguments{
\item{x}{A \linkS4class{tidyprompt} object}

\item{chat_history}{A valid chat history (see \code{\link[=chat_history]{chat_history()}})}
}
\value{
The updated \linkS4class{tidyprompt} object
}
\description{
This function sets the chat history for a \linkS4class{tidyprompt} object.
The chat history will also set the base prompt and system prompt
(the last message of the chat history should be of role 'user' and
will be used as the base prompt; the first message of the chat history
may be of the role 'system' and will then be used as the system prompt).

This may be useful when one wants to change the base prompt, system prompt,
and chat history of a \linkS4class{tidyprompt} object while retaining other fields like
the list of prompt wraps.
}
\examples{
prompt <- tidyprompt("Hi!")
print(prompt)

# Add to a tidyprompt using a prompt wrap:
prompt <- tidyprompt("Hi!") |>
  add_text("How are you?")
print(prompt)

# Strings can be input for prompt wraps; therefore,
#   a call to tidyprompt() is not necessary:
prompt <- "Hi" |>
  add_text("How are you?")

# Example of adding extraction & validation with a prompt_wrap():
prompt <- "Hi" |>
  add_text("What is 5 + 5?") |>
  answer_as_integer()

\dontrun{
  # tidyprompt objects are evaluated by send_prompt(), which will
  #   handle construct the prompt text, send it to the LLM provider,
  #   and apply the extraction and validation functions from the tidyprompt object
  prompt |>
    send_prompt(llm_provider_ollama())
  # --- Sending request to LLM provider (llama3.1:8b): ---
  #   Hi
  #
  #   What is 5 + 5?
  #
  #   You must answer with only an integer (use no other characters).
  # --- Receiving response from LLM provider: ---
  #   10
  # [1] 10

  # See prompt_wrap() and send_prompt() for more details
}

# `tidyprompt` objects may be validated with these helpers:
is_tidyprompt(prompt) # Returns TRUE if input is a valid tidyprompt object

# Get base prompt text
base_prompt <- prompt$base_prompt

# Get all prompt wraps
prompt_wraps <- prompt$get_prompt_wraps()
# Alternative:
prompt_wraps <- get_prompt_wraps(prompt)

# Construct prompt text
prompt_text <- prompt$construct_prompt_text()
# Alternative:
prompt_text <- construct_prompt_text(prompt)

# Set chat history (affecting also the base prompt)
chat_history <- data.frame(
  role = c("user", "assistant", "user"),
  content = c("What is 5 + 5?", "10", "And what is 5 + 6?")
)
prompt$set_chat_history(chat_history)

# Get chat history
chat_history <- prompt$get_chat_history()
}
\seealso{
\code{\link[=chat_history]{chat_history()}}

Other tidyprompt: 
\code{\link{construct_prompt_text}()},
\code{\link{get_chat_history}()},
\code{\link{get_prompt_wraps}()},
\code{\link{is_tidyprompt}()},
\code{\link{tidyprompt}()},
\code{\link{tidyprompt-class}}
}
\concept{tidyprompt}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/llm_provider.R
\name{llm_provider-class}
\alias{llm_provider-class}
\title{LlmProvider R6 Class}
\description{
This class provides a structure for creating \linkS4class{llm_provider}
objects with different implementations of \verb{$complete_chat()}.
Using this class, you can create an \linkS4class{llm_provider} object that interacts
with different LLM providers, such Ollama, OpenAI, or other custom providers.
}
\examples{
# Example creation of a llm_provider-class object:
llm_provider_openai <- function(
    parameters = list(
      model = "gpt-4o-mini",
      stream = getOption("tidyprompt.stream", TRUE)
    ),
    verbose = getOption("tidyprompt.verbose", TRUE),
    url = "https://api.openai.com/v1/chat/completions",
    api_key = Sys.getenv("OPENAI_API_KEY")
) {
  complete_chat <- function(chat_history) {
    headers <- c(
      "Content-Type" = "application/json",
      "Authorization" = paste("Bearer", self$api_key)
    )

    body <- list(
      messages = lapply(seq_len(nrow(chat_history)), function(i) {
        list(role = chat_history$role[i], content = chat_history$content[i])
      })
    )

    for (name in names(self$parameters))
      body[[name]] <- self$parameters[[name]]

    request <- httr2::request(self$url) |>
      httr2::req_body_json(body) |>
      httr2::req_headers(!!!headers)

    request_llm_provider(
      chat_history,
      request,
      stream = self$parameters$stream,
      verbose = self$verbose,
      api_type = self$api_type
    )
  }

  return(`llm_provider-class`$new(
    complete_chat_function = complete_chat,
    parameters = parameters,
    verbose = verbose,
    url = url,
    api_key = api_key,
    api_type = "openai"
  ))
}

llm_provider <- llm_provider_openai()

\dontrun{
  llm_provider$complete_chat("Hi!")
  # --- Sending request to LLM provider (gpt-4o-mini): ---
  # Hi!
  # --- Receiving response from LLM provider: ---
  # Hello! How can I assist you today?
}
}
\seealso{
Other llm_provider: 
\code{\link{llm_provider_ellmer}()},
\code{\link{llm_provider_google_gemini}()},
\code{\link{llm_provider_groq}()},
\code{\link{llm_provider_mistral}()},
\code{\link{llm_provider_ollama}()},
\code{\link{llm_provider_openai}()},
\code{\link{llm_provider_openrouter}()},
\code{\link{llm_provider_xai}()}
}
\concept{llm_provider}
\section{Public fields}{
\if{html}{\out{<div class="r6-fields">}}
\describe{
\item{\code{parameters}}{A named list of parameters to configure the \linkS4class{llm_provider}.
Parameters may be appended to the request body when interacting with the
LLM provider API}

\item{\code{verbose}}{A logical indicating whether interaction with the LLM provider should be
printed to the console}

\item{\code{url}}{The URL to the LLM provider API endpoint for chat completion}

\item{\code{api_key}}{The API key to use for authentication with the LLM provider API}

\item{\code{api_type}}{The type of API to use (e.g., "openai", "ollama", "ellmer").
This is used to determine certain specific behaviors for different APIs,
for instance, as is done in the \code{\link[=answer_as_json]{answer_as_json()}} function}

\item{\code{json_type}}{The type of JSON mode to use (e.g., 'auto', 'openai', 'ollama', 'ellmer', or 'text-based').
Using 'auto' or having this field not set, the api_type field will be used to
determine the JSON mode during the \code{\link[=answer_as_json]{answer_as_json()}} function. If this field
is set, this will override the api_type field for JSON mode determination.
(Note: this determination only happens when the 'type' argument in
\code{\link[=answer_as_json]{answer_as_json()}} is also set to 'auto'.)}

\item{\code{tool_type}}{The type of tool use mode to use (e.g., 'auto', 'openai', 'ollama', 'ellmer', or 'text-based').
Using 'auto' or having this field not set, the api_type field will be used to
determine the tool use mode during the \code{\link[=answer_using_tools]{answer_using_tools()}} function. If this field
is set, this will override the api_type field for tool use mode determination
(Note: this determination only happens when the 'type' argument in
\code{\link[=answer_using_tools]{answer_using_tools()}} is also set to 'auto'.)}

\item{\code{handler_fns}}{A list of functions that will be called after the completion of a chat.
See \verb{$add_handler_fn()}}

\item{\code{stream_callback}}{Optional callback function for streaming tokens/chunks. If set, this function will be
called with arguments \verb{(chunk, meta)} where \code{chunk} is the latest
streamed text chunk and \code{meta} is a list with fields including \code{llm_provider},
\code{chat_history}, and \code{partial_response} to provide more context about the current
prompt that is being replied to. You may use this to implement custom streaming behaviour;
see \verb{vignette(}"streaming_shiny_ipc", "tidyprompt")` for an example of how this
function is used to stream a non-blocking async LLM response to a Shiny app using the
'ipc' package.}

\item{\code{pre_prompt_wraps}}{A list of prompt wraps that will be applied to any prompt evaluated
by this \linkS4class{llm_provider} object, before any prompt-specific
prompt wraps are applied. See \verb{$add_prompt_wrap()}.
This can be used to set default behavior for all prompts
evaluated by this \linkS4class{llm_provider} object.}

\item{\code{post_prompt_wraps}}{A list of prompt wraps that will be applied to any prompt evaluated
by this \linkS4class{llm_provider} object, after any prompt-specific
prompt wraps are applied. See \verb{$add_prompt_wrap()}.
This can be used to set default behavior for all prompts
evaluated by this \linkS4class{llm_provider} object.}
}
\if{html}{\out{</div>}}
}
\section{Methods}{
\subsection{Public methods}{
\itemize{
\item \href{#method-LlmProvider-new}{\code{llm_provider-class$new()}}
\item \href{#method-LlmProvider-set_parameters}{\code{llm_provider-class$set_parameters()}}
\item \href{#method-LlmProvider-complete_chat}{\code{llm_provider-class$complete_chat()}}
\item \href{#method-LlmProvider-add_handler_fn}{\code{llm_provider-class$add_handler_fn()}}
\item \href{#method-LlmProvider-set_handler_fns}{\code{llm_provider-class$set_handler_fns()}}
\item \href{#method-LlmProvider-add_prompt_wrap}{\code{llm_provider-class$add_prompt_wrap()}}
\item \href{#method-LlmProvider-apply_prompt_wraps}{\code{llm_provider-class$apply_prompt_wraps()}}
\item \href{#method-LlmProvider-clone}{\code{llm_provider-class$clone()}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-LlmProvider-new"></a>}}
\if{latex}{\out{\hypertarget{method-LlmProvider-new}{}}}
\subsection{Method \code{new()}}{
Create a new \linkS4class{llm_provider} object
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{llm_provider-class$new(
  complete_chat_function,
  parameters = list(),
  verbose = TRUE,
  url = NULL,
  api_key = NULL,
  api_type = "unspecified"
)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{complete_chat_function}}{Function that will be called by the \linkS4class{llm_provider} to complete a chat.
This function should take a list containing at least '$chat_history'
(a data frame with 'role' and 'content' columns) and return a response
object, which contains:
\itemize{
\item 'completed': A dataframe with 'role' and 'content' columns,
containing the completed chat history

\item 'http': A list containing a list 'requests' and a list 'responses',
containing the HTTP requests and responses made during the chat completion
}}

\item{\code{parameters}}{A named list of parameters to configure the \linkS4class{llm_provider}.
These parameters may be appended to the request body when interacting with
the LLM provider.
For example, the \code{model} parameter may often be required.
The 'stream' parameter may be used to indicate that the API should stream.
Parameters should not include the chat_history, or 'api_key' or 'url', which
are handled separately by the \linkS4class{llm_provider} and '$complete_chat()'.
Parameters should also not be set when they are handled by prompt wraps}

\item{\code{verbose}}{A logical indicating whether interaction with the LLM
provider should be printed to the console}

\item{\code{url}}{The URL to the LLM provider API endpoint for chat completion
(typically required, but may be left NULL in some cases, for instance
when creating a fake LLM provider)}

\item{\code{api_key}}{The API key to use for authentication with the LLM
provider API (optional, not required for, for instance, Ollama)}

\item{\code{api_type}}{The type of API to use (e.g., "openai", "ollama").
This is used to determine certain specific behaviors for different APIs
(see for example the \code{\link[=answer_as_json]{answer_as_json()}} function)}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
A new \linkS4class{llm_provider} R6 object
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-LlmProvider-set_parameters"></a>}}
\if{latex}{\out{\hypertarget{method-LlmProvider-set_parameters}{}}}
\subsection{Method \code{set_parameters()}}{
Helper function to set the parameters of the \linkS4class{llm_provider}
object.
This function appends new parameters to the existing parameters
list.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{llm_provider-class$set_parameters(new_parameters)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{new_parameters}}{A named list of new parameters to append to the
existing parameters list}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
The modified \linkS4class{llm_provider} object
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-LlmProvider-complete_chat"></a>}}
\if{latex}{\out{\hypertarget{method-LlmProvider-complete_chat}{}}}
\subsection{Method \code{complete_chat()}}{
Sends a chat history (see \code{\link[=chat_history]{chat_history()}}
for details) to the LLM provider using the configured \verb{$complete_chat()}.
This function is typically called by \code{\link[=send_prompt]{send_prompt()}} to interact with the LLM
provider, but it can also be called directly.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{llm_provider-class$complete_chat(input)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{input}}{A string, a data frame which is a valid chat history
(see \code{\link[=chat_history]{chat_history()}}), or a list containing a valid chat history under key
'$chat_history'}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
The response from the LLM provider
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-LlmProvider-add_handler_fn"></a>}}
\if{latex}{\out{\hypertarget{method-LlmProvider-add_handler_fn}{}}}
\subsection{Method \code{add_handler_fn()}}{
Helper function to add a handler function to the
\linkS4class{llm_provider} object.
Handler functions are called after the completion of a chat and can be
used to modify the response before it is returned by the \linkS4class{llm_provider}.
Each handler function should take the response object
as input (first argument) as well as 'self' (the \linkS4class{llm_provider}
object) and return a modified response object.
The functions will be called in the order they are added to the list.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{llm_provider-class$add_handler_fn(handler_fn)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{handler_fn}}{A function that takes the response object plus
'self' (the \linkS4class{llm_provider} object) as input and
returns a modified response object}
}
\if{html}{\out{</div>}}
}
\subsection{Details}{
If a handler function returns a list with a 'break' field set to \code{TRUE},
the chat completion will be interrupted and the response will be returned
at that point.
If a handler function returns a list with a 'done' field set to \code{FALSE},
the handler functions will continue to be called in a loop until the 'done'
field is not set to \code{FALSE}.
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-LlmProvider-set_handler_fns"></a>}}
\if{latex}{\out{\hypertarget{method-LlmProvider-set_handler_fns}{}}}
\subsection{Method \code{set_handler_fns()}}{
Helper function to set the handler functions of the
\linkS4class{llm_provider} object.
This function replaces the existing
handler functions list with a new list of handler functions.
See \verb{$add_handler_fn()} for more information
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{llm_provider-class$set_handler_fns(handler_fns)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{handler_fns}}{A list of handler functions to set}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-LlmProvider-add_prompt_wrap"></a>}}
\if{latex}{\out{\hypertarget{method-LlmProvider-add_prompt_wrap}{}}}
\subsection{Method \code{add_prompt_wrap()}}{
Add a provider-level prompt wrap template to be applied to all prompts.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{llm_provider-class$add_prompt_wrap(prompt_wrap, position = c("pre", "post"))}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{prompt_wrap}}{A list created by \code{\link[=provider_prompt_wrap]{provider_prompt_wrap()}}}

\item{\code{position}}{One of "pre" or "post" (applied before/after prompt-specific wraps)}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-LlmProvider-apply_prompt_wraps"></a>}}
\if{latex}{\out{\hypertarget{method-LlmProvider-apply_prompt_wraps}{}}}
\subsection{Method \code{apply_prompt_wraps()}}{
Apply all provider-level wraps to a prompt (character or tidyprompt)
and return a tidyprompt with wraps attached.
This is typically called inside \code{send_prompt()} before evaluation of
the prompt.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{llm_provider-class$apply_prompt_wraps(prompt)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{prompt}}{A string, a chat history, a list containing
a chat history under key '$chat_history', or a \linkS4class{tidyprompt} object}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-LlmProvider-clone"></a>}}
\if{latex}{\out{\hypertarget{method-LlmProvider-clone}{}}}
\subsection{Method \code{clone()}}{
The objects of this class are cloneable with this method.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{llm_provider-class$clone(deep = FALSE)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{deep}}{Whether to make a deep clone.}
}
\if{html}{\out{</div>}}
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/helper_text_skim_with_labels_and_levels.R
\name{skim_with_labels_and_levels}
\alias{skim_with_labels_and_levels}
\title{Skim a dataframe and include labels and levels}
\usage{
skim_with_labels_and_levels(data)
}
\arguments{
\item{data}{A \code{data.frame} to be skimmed}
}
\value{
A \code{data.frame} with variable names, labels, levels, and a skim summary
}
\description{
This function takes a \code{data.frame} and returns a skim summary with variable names,
labels, and levels for categorical variables. It is a wrapper around the \code{\link[skimr:skim]{skimr::skim()}}
function.
}
\examples{
if (requireNamespace("skimr", quietly = TRUE)) {
  # First add some labels to 'mtcars':
  mtcars$car <- rownames(mtcars)
  mtcars$car <- factor(mtcars$car, levels = rownames(mtcars))
  attr(mtcars$car, "label") <- "Name of the car"

  # Then skim the data:
  mtcars |>
    skim_with_labels_and_levels()
}
}
\seealso{
Other text_helpers: 
\code{\link{df_to_string}()},
\code{\link{vector_list_to_string}()}
}
\concept{text_helpers}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/chat_history.R
\name{chat_history.default}
\alias{chat_history.default}
\title{Default method for \code{chat_history()}}
\usage{
\method{chat_history}{default}(chat_history)
}
\arguments{
\item{chat_history}{Object which is not \code{character} or \code{data.frame}}
}
\value{
No return value; an error is thrown
}
\description{
Calls error which indicates that the input was not a \code{character} or \code{data.frame}.
}
\details{
When input is a \code{character} or \code{data.frame}, the appropriate method will be called
(see `\code{\link[=chat_history.character]{chat_history.character()}} and \code{\link[=chat_history.data.frame]{chat_history.data.frame()}}).
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/helper_other_r_json_schema.R
\name{r_json_schema_to_example}
\alias{r_json_schema_to_example}
\title{Generate an example object from a JSON schema}
\usage{
r_json_schema_to_example(schema)
}
\arguments{
\item{schema}{A list (R object) representing a JSON schema}
}
\value{
A list (R object) which matches the JSON schema definition
}
\description{
This function generates an example JSON object from a JSON schema.
This is used when enforcing a JSON schema through text-based handling
(requiring an example to be added to the prompt text).
}
\examples{
base_prompt <- "How can I solve 8x + 7 = -23?"

# This example will show how to enforce JSON format in the response,
#   with and without a schema, using the `answer_as_json()` prompt wrap.

# If you use type = 'auto', the function will automatically detect the
#   best way to enforce JSON based on the LLM provider you are using.

# `answer_as_json()` supports two ways of supplying a schema for structured output:
#   - 1) an 'ellmer' definition (e.g., `ellmer::type_object()`;
#         see https://ellmer.tidyverse.org/articles/structured-data.html)
#   - 2) a R list object representing a JSON schema
# `answer_as_json()` will convert the schema type which you supply to any
#   LLM provider; so, whether you use an ellmer LLM provider or another type,
#   you can supply either a R list object or an ellmer definition, and don't
#   have to worry about compatibility
# Supplying a schema as an ellmer definition is likely the easiest

# Below, we will show:
#  - 1) enforcing JSON with a schema; 'ellmer' definition
#  - 2) enforcing JSON with a schema; R list object
# -  3) enforcing JSON without a schema

#### Enforcing JSON with a schema (ellmer definition): ####

# Make an ellmer definition of structured output
#   For instance, a persona:
ellmer_schema <- ellmer::type_object(
  name = ellmer::type_string(),
  age = ellmer::type_integer(),
  hobbies = ellmer::type_array(ellmer::type_string())
)

\dontrun{
  # Example Ellmer LLM provider
  ellmer_openai <- llm_provider_ellmer(ellmer::chat_openai(
    model = "gpt-4.1-mini"
  ))

  # Example regular LLM provider
  tidyprompt_openai <- llm_provider_openai()$set_parameters(
    list(model = "gpt-4.1-mini")
  )

  # You can supply the ellmer definition to both types of LLM provider
  #   to generate an R list object adhering to the schema
  result_ellmer_x_ellmer <- "Create a persona" |>
    answer_as_json(ellmer_schema) |>
    send_prompt(ellmer_openai)

  result_tidyrpompt_x_ellmer <- "Create a persona" |>
    answer_as_json(ellmer_schema) |>
    send_prompt(tidyprompt_openai)
}


#### Enforcing JSON with a schema (R list object definition): ####

# Make a list representing a JSON schema,
#   which the LLM response must adhere to:
json_schema <- list(
  name = "steps_to_solve", # Required for OpenAI API
  description = NULL, # Optional for OpenAI API
  schema = list(
    type = "object",
    properties = list(
      steps = list(
        type = "array",
        items = list(
          type = "object",
          properties = list(
            explanation = list(type = "string"),
            output = list(type = "string")
          ),
          required = c("explanation", "output"),
          additionalProperties = FALSE
        )
      ),
      final_answer = list(type = "string")
    ),
    required = c("steps", "final_answer"),
    additionalProperties = FALSE
  )
  # 'strict' parameter is set as argument 'answer_as_json()'
)
# Note: when you are not using an OpenAI API, you can also pass just the
#   internal 'schema' list object to 'answer_as_json()' instead of the full
#   'json_schema' list object

# Generate example R object based on schema:
r_json_schema_to_example(json_schema)

\dontrun{
  ## Text-based with schema (works for any provider/model):
  #   - Adds request to prompt for a JSON object
  #   - Adds schema to prompt
  #   - Extracts JSON from textual response (feedback for retry if no JSON received)
  #   - Validates JSON against schema with 'jsonvalidate' package (feedback for retry if invalid)
  #   - Parses JSON to R object
  json_4 <- base_prompt |>
    answer_as_json(schema = json_schema) |>
    send_prompt(llm_provider_ollama())
  # --- Sending request to LLM provider (llama3.1:8b): ---
  # How can I solve 8x + 7 = -23?
  #
  # Your must format your response as a JSON object.
  #
  # Your JSON object should match this example JSON object:
  #   {
  #     "steps": [
  #       {
  #         "explanation": "...",
  #         "output": "..."
  #       }
  #     ],
  #     "final_answer": "..."
  #   }
  # --- Receiving response from LLM provider: ---
  # Here is the solution to the equation:
  #
  # ```
  # {
  #   "steps": [
  #     {
  #       "explanation": "First, we want to isolate the term with 'x' by
  #       subtracting 7 from both sides of the equation.",
  #       "output": "8x + 7 - 7 = -23 - 7"
  #     },
  #     {
  #       "explanation": "This simplifies to: 8x = -30",
  #       "output": "8x = -30"
  #     },
  #     {
  #       "explanation": "Next, we want to get rid of the coefficient '8' by
  #       dividing both sides of the equation by 8.",
  #       "output": "(8x) / 8 = (-30) / 8"
  #     },
  #     {
  #       "explanation": "This simplifies to: x = -3.75",
  #       "output": "x = -3.75"
  #     }
  #   ],
  #   "final_answer": "-3.75"
  # }
  # ```

  ## Ollama with schema:
  #   - Sets 'format' parameter to 'json', enforcing JSON
  #   - Adds request to prompt for a JSON object, as is recommended by the docs
  #   - Adds schema to prompt
  #   - Validates JSON against schema with 'jsonvalidate' package (feedback for retry if invalid)
  json_5 <- base_prompt |>
    answer_as_json(json_schema, type = "auto") |>
    send_prompt(llm_provider_ollama())
  # --- Sending request to LLM provider (llama3.1:8b): ---
  # How can I solve 8x + 7 = -23?
  #
  # Your must format your response as a JSON object.
  #
  # Your JSON object should match this example JSON object:
  # {
  #   "steps": [
  #     {
  #       "explanation": "...",
  #       "output": "..."
  #     }
  #   ],
  #   "final_answer": "..."
  # }
  # --- Receiving response from LLM provider: ---
  # {
  #   "steps": [
  #     {
  #       "explanation": "First, subtract 7 from both sides of the equation to
  #       isolate the term with x.",
  #       "output": "8x = -23 - 7"
  #     },
  #     {
  #       "explanation": "Simplify the right-hand side of the equation.",
  #       "output": "8x = -30"
  #     },
  #     {
  #       "explanation": "Next, divide both sides of the equation by 8 to solve for x.",
  #       "output": "x = -30 / 8"
  #     },
  #     {
  #       "explanation": "Simplify the right-hand side of the equation.",
  #       "output": "x = -3.75"
  #     }
  #   ],
  #   "final_answer": "-3.75"
  # }

  ## OpenAI with schema:
  #   - Sets 'response_format' parameter to 'json_object', enforcing JSON
  #   - Adds json_schema to the API request, API enforces JSON adhering schema
  #   - Parses JSON to R object
  json_6 <- base_prompt |>
    answer_as_json(json_schema, type = "auto") |>
    send_prompt(llm_provider_openai())
  # --- Sending request to LLM provider (gpt-4o-mini): ---
  # How can I solve 8x + 7 = -23?
  # --- Receiving response from LLM provider: ---
  # {"steps":[
  # {"explanation":"Start with the original equation.",
  # "output":"8x + 7 = -23"},
  # {"explanation":"Subtract 7 from both sides to isolate the term with x.",
  # "output":"8x + 7 - 7 = -23 - 7"},
  # {"explanation":"Simplify the left side and the right side of the equation.",
  # "output":"8x = -30"},
  # {"explanation":"Now, divide both sides by 8 to solve for x.",
  # "output":"x = -30 / 8"},
  # {"explanation":"Simplify the fraction by dividing both the numerator and the
  # denominator by 2.",
  # "output":"x = -15 / 4"}
  # ], "final_answer":"x = -15/4"}

  # You can also use the R list object schema definition with an
  #  ellmer LLM provider; `answer_as_json()` will do the conversion for you
  json_7 <- base_prompt |>
    answer_as_json(json_schema) |>
    send_prompt(ellmer_openai)
}


#### Enforcing JSON without a schema: ####

\dontrun{
  ## Text-based (works for any provider/model):
  #   Adds request to prompt for a JSON object
  #   Extracts JSON from textual response (feedback for retry if no JSON received)
  #   Parses JSON to R object
  json_1 <- base_prompt |>
    answer_as_json() |>
    send_prompt(llm_provider_ollama())
  # --- Sending request to LLM provider (llama3.1:8b): ---
  # How can I solve 8x + 7 = -23?
  #
  # Your must format your response as a JSON object.
  # --- Receiving response from LLM provider: ---
  # Here is the solution to the equation formatted as a JSON object:
  #
  # ```
  # {
  #   "equation": "8x + 7 = -23",
  #   "steps": [
  #     {
  #       "step": "Subtract 7 from both sides of the equation",
  #       "expression": "-23 - 7"
  #     },
  #     {
  #       "step": "Simplify the expression on the left side",
  #       "result": "-30"
  #     },
  #     {
  #       "step": "Divide both sides by -8 to solve for x",
  #       "expression": "-30 / -8"
  #     },
  #     {
  #       "step": "Simplify the expression on the right side",
  #       "result": "3.75"
  #     }
  #   ],
  #   "solution": {
  #     "x": 3.75
  #   }
  # }
  # ```


  ## Ollama:
  #   - Sets 'format' parameter to 'json', enforcing JSON
  #   - Adds request to prompt for a JSON object, as is recommended by the docs
  #   - Parses JSON to R object
  json_2 <- base_prompt |>
    answer_as_json(type = "auto") |>
    send_prompt(llm_provider_ollama())
  # --- Sending request to LLM provider (llama3.1:8b): ---
  # How can I solve 8x + 7 = -23?
  #
  # Your must format your response as a JSON object.
  # --- Receiving response from LLM provider: ---
  # {"steps": [
  #   "Subtract 7 from both sides to get 8x = -30",
  #   "Simplify the right side of the equation to get 8x = -30",
  #   "Divide both sides by 8 to solve for x, resulting in x = -30/8",
  #   "Simplify the fraction to find the value of x"
  # ],
  # "value_of_x": "-3.75"}


  ## OpenAI-type API without schema:
  #   - Sets 'response_format' parameter to 'json_object', enforcing JSON
  #   - Adds request to prompt for a JSON object, as is required by the API
  #   - Parses JSON to R object
  json_3 <- base_prompt |>
    answer_as_json(type = "auto") |>
    send_prompt(llm_provider_openai())
  # --- Sending request to LLM provider (gpt-4o-mini): ---
  # How can I solve 8x + 7 = -23?
  #
  # Your must format your response as a JSON object.
  # --- Receiving response from LLM provider: ---
  # {
  #   "solution_steps": [
  #     {
  #       "step": 1,
  #       "operation": "Subtract 7 from both sides",
  #       "equation": "8x + 7 - 7 = -23 - 7",
  #       "result": "8x = -30"
  #     },
  #     {
  #       "step": 2,
  #       "operation": "Divide both sides by 8",
  #       "equation": "8x / 8 = -30 / 8",
  #       "result": "x = -3.75"
  #     }
  #   ],
  #   "solution": {
  #     "x": -3.75
  #   }
  # }
}
}
\seealso{
Other json: 
\code{\link{answer_as_json}()}
}
\concept{json}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/answer_as_text.R
\name{answer_as_text}
\alias{answer_as_text}
\title{Make LLM answer as a constrained text response}
\usage{
answer_as_text(
  prompt,
  max_words = NULL,
  max_characters = NULL,
  add_instruction_to_prompt = TRUE
)
}
\arguments{
\item{prompt}{A single string or a \code{\link[=tidyprompt]{tidyprompt()}} object}

\item{max_words}{(optional) Maximum number of words allowed in the response.
If specified, responses exceeding this limit will fail validation}

\item{max_characters}{(optional) Maximum number of characters allowed in the response.
If specified, responses exceeding this limit will fail validation}

\item{add_instruction_to_prompt}{(optional) Add instruction for replying
within the constraints to the prompt text. Set to FALSE for debugging if
extractions/validations are working as expected (without instruction the
answer should fail the validation function, initiating a retry)}
}
\value{
A \code{\link[=tidyprompt]{tidyprompt()}} with an added \code{\link[=prompt_wrap]{prompt_wrap()}} which
will ensure that the LLM response conforms to the specified constraints
}
\description{
Make LLM answer as a constrained text response
}
\examples{
\dontrun{
  "What is a large language model?" |>
    answer_as_text(max_words = 10) |>
    send_prompt()
  # --- Sending request to LLM provider (llama3.1:8b): ---
  # What is a large language model?
  #
  # You must provide a text response. The response must be at most 10 words.
  # --- Receiving response from LLM provider: ---
  # A type of AI that processes and generates human-like text.
  # [1] "A type of AI that processes and generates human-like text."
}
}
\seealso{
Other pre_built_prompt_wraps: 
\code{\link{add_image}()},
\code{\link{add_text}()},
\code{\link{answer_as_boolean}()},
\code{\link{answer_as_category}()},
\code{\link{answer_as_dataframe}()},
\code{\link{answer_as_integer}()},
\code{\link{answer_as_json}()},
\code{\link{answer_as_list}()},
\code{\link{answer_as_multi_category}()},
\code{\link{answer_as_named_list}()},
\code{\link{answer_as_numeric}()},
\code{\link{answer_as_regex_match}()},
\code{\link{answer_by_chain_of_thought}()},
\code{\link{answer_by_react}()},
\code{\link{answer_using_r}()},
\code{\link{answer_using_sql}()},
\code{\link{answer_using_tools}()},
\code{\link{prompt_wrap}()},
\code{\link{quit_if}()},
\code{\link{set_system_prompt}()}

Other answer_as_prompt_wraps: 
\code{\link{answer_as_boolean}()},
\code{\link{answer_as_category}()},
\code{\link{answer_as_dataframe}()},
\code{\link{answer_as_integer}()},
\code{\link{answer_as_json}()},
\code{\link{answer_as_list}()},
\code{\link{answer_as_multi_category}()},
\code{\link{answer_as_named_list}()},
\code{\link{answer_as_numeric}()},
\code{\link{answer_as_regex_match}()}
}
\concept{answer_as_prompt_wraps}
\concept{pre_built_prompt_wraps}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tidyprompt.R
\name{is_tidyprompt}
\alias{is_tidyprompt}
\title{Check if object is a \linkS4class{tidyprompt} object}
\usage{
is_tidyprompt(x)
}
\arguments{
\item{x}{An object to check}
}
\value{
TRUE if the object is a valid \linkS4class{tidyprompt} object, otherwise FALSE
}
\description{
Check if object is a \linkS4class{tidyprompt} object
}
\examples{
prompt <- tidyprompt("Hi!")
print(prompt)

# Add to a tidyprompt using a prompt wrap:
prompt <- tidyprompt("Hi!") |>
  add_text("How are you?")
print(prompt)

# Strings can be input for prompt wraps; therefore,
#   a call to tidyprompt() is not necessary:
prompt <- "Hi" |>
  add_text("How are you?")

# Example of adding extraction & validation with a prompt_wrap():
prompt <- "Hi" |>
  add_text("What is 5 + 5?") |>
  answer_as_integer()

\dontrun{
  # tidyprompt objects are evaluated by send_prompt(), which will
  #   handle construct the prompt text, send it to the LLM provider,
  #   and apply the extraction and validation functions from the tidyprompt object
  prompt |>
    send_prompt(llm_provider_ollama())
  # --- Sending request to LLM provider (llama3.1:8b): ---
  #   Hi
  #
  #   What is 5 + 5?
  #
  #   You must answer with only an integer (use no other characters).
  # --- Receiving response from LLM provider: ---
  #   10
  # [1] 10

  # See prompt_wrap() and send_prompt() for more details
}

# `tidyprompt` objects may be validated with these helpers:
is_tidyprompt(prompt) # Returns TRUE if input is a valid tidyprompt object

# Get base prompt text
base_prompt <- prompt$base_prompt

# Get all prompt wraps
prompt_wraps <- prompt$get_prompt_wraps()
# Alternative:
prompt_wraps <- get_prompt_wraps(prompt)

# Construct prompt text
prompt_text <- prompt$construct_prompt_text()
# Alternative:
prompt_text <- construct_prompt_text(prompt)

# Set chat history (affecting also the base prompt)
chat_history <- data.frame(
  role = c("user", "assistant", "user"),
  content = c("What is 5 + 5?", "10", "And what is 5 + 6?")
)
prompt$set_chat_history(chat_history)

# Get chat history
chat_history <- prompt$get_chat_history()
}
\seealso{
Other tidyprompt: 
\code{\link{construct_prompt_text}()},
\code{\link{get_chat_history}()},
\code{\link{get_prompt_wraps}()},
\code{\link{set_chat_history}()},
\code{\link{tidyprompt}()},
\code{\link{tidyprompt-class}}
}
\concept{tidyprompt}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/answer_using_tools.R
\name{answer_using_tools}
\alias{answer_using_tools}
\title{Enable LLM to call R functions (and/or MCP server tools)}
\usage{
answer_using_tools(
  prompt,
  tools = list(),
  type = c("auto", "openai", "ollama", "ellmer", "text-based")
)
}
\arguments{
\item{prompt}{A single string or a \code{\link[=tidyprompt]{tidyprompt()}} object}

\item{tools}{An R function, an 'ellmer' tool definition (from \code{\link[ellmer:tool]{ellmer::tool()}}),
or a list of either, which the LLM will be able to call. If an R function is passed
which has been documented in a help file (e.g., because it is part of a package),
the documentation will be parsed from the help file. If it is a custom function,
documentation should be added with \code{\link[=tools_add_docs]{tools_add_docs()}} or with \code{\link[ellmer:tool]{ellmer::tool()}}.
Note that you can also provide Model Context Protocol (MCP) tools from MCP servers
as returned from \code{\link[mcptools:client]{mcptools::mcp_tools()}}.
Provider-specific 'ellmer' built-in tools are only supported with
'ellmer'-backed providers.}

\item{type}{(optional) The way that tool calling should be enabled.
"auto" will automatically determine the type based on \code{llm_provider$api_type}
or 'llm_provider$tool_type' (if set; 'tool_type' overrides 'api_type' determination)
(note that this may not consider model compatibility, and could lead to errors;
set 'type' manually if errors occur). "openai" and "ollama" will set the
relevant API parameters. "ellmer" will register the tool in the 'ellmer' chat
object of the LLM provider; note that this will only work for an \code{\link[=llm_provider_ellmer]{llm_provider_ellmer()}}
("auto" will always set the type to "ellmer" if you are using an 'ellmer' LLM provider).
"text-based" will provide function definitions in the prompt, extract function
calls from the LLM response, and call the
functions, providing the results back via \code{\link[=llm_feedback]{llm_feedback()}}. "text-based"
always works, but may be inefficient for APIs that support tool calling
natively. Note that when using "openai", "ollama", or "ellmer", tool calls are not counted
as interactions by \code{\link[=send_prompt]{send_prompt()}} and may continue indefinitely unless restricted
by other means}
}
\value{
A \code{\link[=tidyprompt]{tidyprompt()}} with an added \code{\link[=prompt_wrap]{prompt_wrap()}} which
will allow the LLM to call the given R functions when evaluating
the prompt with \code{\link[=send_prompt]{send_prompt()}}
}
\description{
This function adds the ability for the a LLM to call R functions.
Users can specify a list of functions that the LLM can call, and the
prompt will be modified to include information, as well as an
accompanying extraction function to call the functions (handled by
\code{\link[=send_prompt]{send_prompt()}}). Documentation for the functions is extracted from
the help file (if available), or from documentation added by
\code{\link[=tools_add_docs]{tools_add_docs()}}. Users can also provide an 'ellmer' tool definition
(see \code{\link[ellmer:tool]{ellmer::tool()}}; \href{https://ellmer.tidyverse.org/articles/tool-calling.html}{'ellmer' documentation}).
Model Context Protocol (MCP) tools from MCP servers, as returned from
\code{\link[mcptools:client]{mcptools::mcp_tools()}}, may also be used. Regular \code{\link[ellmer:tool]{ellmer::tool()}}
definitions are converted as needed and can work with both 'ellmer' and
regular LLM providers. Provider-specific 'ellmer' built-in tools
(i.e., 'ToolBuiltIn' objects) only work with 'ellmer'-backed providers.
}
\details{
Note that conversion between 'tidyprompt' and 'ellmer' tool definitions
is experimntal and might contain bugs.
Provider-specific 'ellmer' built-in tools are only supported when using an
\code{\link[=llm_provider_ellmer]{llm_provider_ellmer()}} object (or an 'ellmer' chat passed directly to
\code{\link[=send_prompt]{send_prompt()}}).
}
\examples{
# When using functions from base R or R packages,
#   documentation is automatically extracted from help files:
prompt_with_dir_function <- "What are the files in my current directory?" |>
  answer_using_tools(dir) # The 'dir' function is from base R
\dontrun{
  send_prompt(prompt_with_dir_function)
  # --- Sending request to LLM provider (llama3.1:8b): ---
  #   What are the files in my current directory?
  #   --- Receiving response from LLM provider: ---
  #   Calling function 'nm' with arguments:
  #   {
  #     "all.files": true,
  #     "full.names": false,
  #     "ignore.case": false,
  #     "include.dirs": false,
  #     "no..": false,
  #     "path": "./",
  #     "pattern": "*",
  #     "recursive": false
  #   }
  # Result:
  #   .git, .github, .gitignore, .Rbuildignore, .Rhistory, ...
  # The files in your current directory are:
  #   .git, .github, .gitignore, .Rbuildignore, .Rhistory, ...
  # [1] "The files in your current directory are:\n\n .git, .github, ..."
}

# Users may provide custom functions in two ways:
#   1) as a function object, optionally documented with `tools_get_docs()`, or
#   2) as an 'ellmer' tool definition, using `ellmer::tool()`

# Take this fake weather function as an example:
temperature_in_location <- function(
  location = c("Amsterdam", "Utrecht", "Enschede"),
  unit = c("Celcius", "Fahrenheit")
) {
  location <- match.arg(location)
  unit <- match.arg(unit)

  temperature_celcius <- switch(
    location,
    "Amsterdam" = 32.5,
    "Utrecht" = 19.8,
    "Enschede" = 22.7
  )

  if (unit == "Celcius") {
    return(temperature_celcius)
  } else {
    return(temperature_celcius * 9/5 + 32)
  }
}


# 1: `tools_add_docs()` --------------------------------------------------------

# Generate documentation for a function, based on formals & help file
docs <- tools_get_docs(temperature_in_location)

# The types get inferred from the function's formals
# However, descriptions are still missing as the function is not from a package
# We can modify the documentation object to add descriptions:
docs$description <- "Get the temperature in a location"
docs$arguments$unit$description <- "Unit in which to return the temperature"
docs$arguments$location$description <- "Location for which to return the temperature"
docs$return$description <- "The temperature in the specified location and unit"
# (See `?tools_add_docs` for more details on the structure of the documentation)

# When we are satisfied with the documentation, we can add it to the function:
temperature_in_location <- tools_add_docs(temperature_in_location, docs)

prompt_with_weather_function <-
  "What is the weather in Enschede? Give me Celcius degrees" |>
  answer_using_tools(temperature_in_location)
\dontrun{
  send_prompt(prompt_with_weather_function)
  # --- Sending request to LLM provider (llama3.1:8b): ---
  #   What is the weather in Enschede? Give me Celcius degrees
  #   --- Receiving response from LLM provider: ---
  #   Calling function 'temperature_in_location' with arguments:
  #   {
  #     "location": "Enschede",
  #     "unit": "Celcius"
  #   }
  # Result:
  #   22.7
  # The temperature in Enschede is 22.7 Celcius degrees.
  # [1] "The temperature in Enschede is 22.7 Celcius degrees."
}

# 2: `ellmer::tool()` -----------------------------------------------

# Alternatively, we can define the function as an 'ellmer' tool

temperature_in_location_ellmer <- ellmer::tool(
  temperature_in_location,
  name = "get_temperature",
  description = "Get the temperature in a location",
  arguments = list(
    location = ellmer::type_string(
      "Location for which to return the temperature", required = TRUE
    ),
    unit = ellmer::type_string(
      "Unit in which to return the temperature", required = TRUE
    )
  )
)

prompt_with_weather_function_ellmer <-
  "What is the weather in Utrecht? Give me Fahrenheit degrees" |>
  answer_using_tools(temperature_in_location_ellmer)
\dontrun{
  send_prompt(prompt_with_weather_function_ellmer)
  # ...
}

# Because `mcptools::mcp_tools()` returns a list of `ellmer:tool()` tools,
#   you can also use Model Context Protocol (MCP) server tools with
#   `answer_using_tools()`:
\dontrun{
  prompt_using_mcp_tools <- mcptools::mcp_tools()
  "Push my latest commit to GitHub" |>
    answer_using_tools(mcp_tools)
  send_prompt(prompt_using_mcp_tools)
}

# `answer_using_tools()` will automatically attempt to use the most appropriate
#   way of sending the tool to the LLM

# If you use a LLM provider of type 'ollama' or 'openai',
#   it will automatically convert the tool definition to parameters
#   appropriate for those APIs
# If you use a LLM provider of type 'ellmer', it will call the appropriate
#   ellmer function directly which will handle the tool call for various
#   providers
# Note that both tool definitions from `tools_add_docs()` and `ellmer::tool()`
#   will work with any LLM provider; `answer_using_tools()` can convert
#   the two types of tool definitions to each other when needed
\dontrun{
  ollama <- llm_provider_ollama()
  # Ollama LLM provider:
  "What is the weather in Amsterdam? Give me Fahrenheit degrees" |>
    answer_using_tools(temperature_in_location) |>
    send_prompt(ollama)

  # Ollama LLM provider also works with `ellmer::tool()` definitions:
  "What is the weather in Amsterdam? Give me Celcius degrees" |>
    answer_using_tools(temperature_in_location_ellmer) |>
    send_prompt(ollama)

  # Similar for OpenAI API:
  openai <- llm_provider_openai()
  "What is the weather in Amsterdam? Give me Celcius degrees" |>
    answer_using_tools(temperature_in_location) |>
    send_prompt(openai)
  # ...

  # Ellmer LLM provider:
  ellmer <- llm_provider_ellmer(ellmer::chat_openai())
  "What is the weather in Amsterdam? Give me Celcius degrees" |>
    answer_using_tools(temperature_in_location_ellmer) |>
    send_prompt(ellmer)

  # Also works with `tools_add_docs()` definition:
  "What is the weather in Amsterdam? Give me Celcius degrees" |>
    answer_using_tools(temperature_in_location) |>
    send_prompt(ellmer)
}
}
\seealso{
\code{\link[=answer_using_r]{answer_using_r()}} \code{\link[=tools_get_docs]{tools_get_docs()}}

Other pre_built_prompt_wraps: 
\code{\link{add_image}()},
\code{\link{add_text}()},
\code{\link{answer_as_boolean}()},
\code{\link{answer_as_category}()},
\code{\link{answer_as_dataframe}()},
\code{\link{answer_as_integer}()},
\code{\link{answer_as_json}()},
\code{\link{answer_as_list}()},
\code{\link{answer_as_multi_category}()},
\code{\link{answer_as_named_list}()},
\code{\link{answer_as_numeric}()},
\code{\link{answer_as_regex_match}()},
\code{\link{answer_as_text}()},
\code{\link{answer_by_chain_of_thought}()},
\code{\link{answer_by_react}()},
\code{\link{answer_using_r}()},
\code{\link{answer_using_sql}()},
\code{\link{prompt_wrap}()},
\code{\link{quit_if}()},
\code{\link{set_system_prompt}()}

Other answer_using_prompt_wraps: 
\code{\link{answer_using_r}()},
\code{\link{answer_using_sql}()}

Other tools: 
\code{\link{tools_add_docs}()},
\code{\link{tools_get_docs}()}
}
\concept{answer_using_prompt_wraps}
\concept{pre_built_prompt_wraps}
\concept{tools}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/llm_providers.R
\name{llm_provider_openai}
\alias{llm_provider_openai}
\title{Create a new OpenAI LLM provider}
\usage{
llm_provider_openai(
  parameters = list(model = "gpt-4o-mini", stream = getOption("tidyprompt.stream", TRUE)),
  verbose = getOption("tidyprompt.verbose", TRUE),
  url = "https://api.openai.com/v1/responses",
  api_key = Sys.getenv("OPENAI_API_KEY")
)
}
\arguments{
\item{parameters}{A named list of parameters. Currently the following parameters are required:
\itemize{
\item model: The name of the model to use
\item api_key: The API key to use for authentication with the OpenAI API. This
should be a project API key (not a user API key)
\item url: The URL to the OpenAI API (may also be an alternative endpoint
that provides a similar API.)
\item stream: A logical indicating whether the API should stream responses
}

Additional parameters are appended to the request body; see the OpenAI API
documentation for more information: https://platform.openai.com/docs/api-reference/chat}

\item{verbose}{A logical indicating whether the interaction with the LLM provider
should be printed to the console. Default is TRUE.}

\item{url}{The URL to the OpenAI API endpoint for chat completion
(typically: "https://api.openai.com/v1/chat/completions"
or "https://api.openai.com/v1/responses"; both the Chat Completions API and
the Responses API are supported; the Responses API is more modern)}

\item{api_key}{The API key to use for authentication with the OpenAI API}
}
\value{
A new \linkS4class{llm_provider} object for use of the OpenAI API
}
\description{
This function creates a new \linkS4class{llm_provider} object that interacts with the Open AI API.
Supports both the Chat Completions API (/v1/chat/completions) and the Responses API (/v1/responses).
}
\examples{
# Various providers:
ollama <- llm_provider_ollama()
openai <- llm_provider_openai()
openrouter <- llm_provider_openrouter()
mistral <- llm_provider_mistral()
groq <- llm_provider_groq()
xai <- llm_provider_xai()
gemini <- llm_provider_google_gemini()

# From an `ellmer::chat()` (e.g., `ellmer::chat_openai()`, ...):
\dontrun{
ellmer <- llm_provider_ellmer(ellmer::chat_openai())
}

# Initialize with settings:
ollama <- llm_provider_ollama(
  parameters = list(
    model = "llama3.2:3b",
    stream = TRUE
  ),
  verbose = TRUE,
  url = "http://localhost:11434/api/chat"
)

# Change settings:
ollama$verbose <- FALSE
ollama$parameters$stream <- FALSE
ollama$parameters$model <- "llama3.1:8b"

\dontrun{
# Try a simple chat message with '$complete_chat()':
response <- ollama$complete_chat("Hi!")
response
# $role
# [1] "assistant"
#
# $content
# [1] "How's it going? Is there something I can help you with or would you like
# to chat?"
#
# $http
# Response [http://localhost:11434/api/chat]
# Date: 2024-11-18 14:21
# Status: 200
# Content-Type: application/json; charset=utf-8
# Size: 375 B

# Use with send_prompt():
"Hi" |>
  send_prompt(ollama)
# [1] "How's your day going so far? Is there something I can help you with or
# would you like to chat?"
}
}
\seealso{
Other llm_provider: 
\code{\link{llm_provider-class}},
\code{\link{llm_provider_ellmer}()},
\code{\link{llm_provider_google_gemini}()},
\code{\link{llm_provider_groq}()},
\code{\link{llm_provider_mistral}()},
\code{\link{llm_provider_ollama}()},
\code{\link{llm_provider_openrouter}()},
\code{\link{llm_provider_xai}()}
}
\concept{llm_provider}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/answer_as_regex_match.R
\name{answer_as_regex_match}
\alias{answer_as_regex_match}
\title{Make LLM answer match a specific regex}
\usage{
answer_as_regex_match(prompt, regex, mode = c("full_match", "extract_matches"))
}
\arguments{
\item{prompt}{A single string or a \code{\link[=tidyprompt]{tidyprompt()}} object}

\item{regex}{A character string specifying the regular expression the response must match}

\item{mode}{A character string specifying the mode of the regex match. Options are
"exact_match" (default) and "extract_all_matches".
Under "full_match", the full LLM response must match the regex. If
it does not, the LLM will be sent feedback to retry. The full LLM response
will be returned if the regex is matched.
Under "extract_matches", all matches of the regex in the LLM response will be returned
(if present). If the regex is not matched at all, the LLM will be sent feedback to retry.
If there is at least one match, the matches will be returned as a character vector}
}
\value{
A \code{\link[=tidyprompt]{tidyprompt()}} with an added \code{\link[=prompt_wrap]{prompt_wrap()}} which will ensure
that the LLM response matches the specified regex
}
\description{
Make LLM answer match a specific regex
}
\examples{
\dontrun{
  "What would be a suitable e-mail address for cupcake company?" |>
    answer_as_regex_match("^[a-zA-Z0-9._\%+-]+@[a-zA-Z0-9.-]+\\\\.[a-zA-Z]{2,}$") |>
    send_prompt(llm_provider_ollama())
  # --- Sending request to LLM provider (llama3.1:8b): ---
  #   What would be a suitable e-mail address for cupcake company?
  #
  #   You must answer with a response that matches this regex format:
  #     ^[a-zA-Z0-9._\%+-]+@[a-zA-Z0-9.-]+\.[a-zA-Z]{2,}$
  #     (use no other characters)
  # --- Receiving response from LLM provider: ---
  #   sweet.treats.cupcakes@gmail.com
  # [1] "sweet.treats.cupcakes@gmail.com"

  "What would be a suitable e-mail address for cupcake company?" |>
    add_text("Give three ideas.") |>
    answer_as_regex_match(
      "[a-zA-Z0-9._\%+-]+@[a-zA-Z0-9.-]+\\\\.[a-zA-Z]{2,}",
      mode = "extract_matches"
    ) |>
    send_prompt(llm_provider_ollama())
  # --- Sending request to LLM provider (llama3.1:8b): ---
  #   What would be a suitable e-mail address for cupcake company?
  #
  #   Give three ideas.
  #
  #   You must answer with a response that matches this regex format:
  #     [a-zA-Z0-9._\%+-]+@[a-zA-Z0-9.-]+\.[a-zA-Z]{2,}
  # --- Receiving response from LLM provider: ---
  #   Here are three potential email addresses for a cupcake company:
  #
  #   1. sweettreats.cupcakes@yummail.com
  #   2. cupcakes.and.love@flourpower.net
  #   3. thecupcakery@gmail.com
  # [1] "sweettreats.cupcakes@yummail.com" "cupcakes.and.love@flourpower.net"
  # "thecupcakery@gmail.com"
}
}
\seealso{
Other pre_built_prompt_wraps: 
\code{\link{add_image}()},
\code{\link{add_text}()},
\code{\link{answer_as_boolean}()},
\code{\link{answer_as_category}()},
\code{\link{answer_as_dataframe}()},
\code{\link{answer_as_integer}()},
\code{\link{answer_as_json}()},
\code{\link{answer_as_list}()},
\code{\link{answer_as_multi_category}()},
\code{\link{answer_as_named_list}()},
\code{\link{answer_as_numeric}()},
\code{\link{answer_as_text}()},
\code{\link{answer_by_chain_of_thought}()},
\code{\link{answer_by_react}()},
\code{\link{answer_using_r}()},
\code{\link{answer_using_sql}()},
\code{\link{answer_using_tools}()},
\code{\link{prompt_wrap}()},
\code{\link{quit_if}()},
\code{\link{set_system_prompt}()}

Other answer_as_prompt_wraps: 
\code{\link{answer_as_boolean}()},
\code{\link{answer_as_category}()},
\code{\link{answer_as_dataframe}()},
\code{\link{answer_as_integer}()},
\code{\link{answer_as_json}()},
\code{\link{answer_as_list}()},
\code{\link{answer_as_multi_category}()},
\code{\link{answer_as_named_list}()},
\code{\link{answer_as_numeric}()},
\code{\link{answer_as_text}()}
}
\concept{answer_as_prompt_wraps}
\concept{pre_built_prompt_wraps}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/llm_providers.R
\name{llm_provider_ollama}
\alias{llm_provider_ollama}
\title{Create a new Ollama LLM provider}
\usage{
llm_provider_ollama(
  parameters = list(model = "llama3.1:8b", stream = getOption("tidyprompt.stream", TRUE)),
  verbose = getOption("tidyprompt.verbose", TRUE),
  url = "http://localhost:11434/api/chat",
  num_ctx = NULL
)
}
\arguments{
\item{parameters}{A named list of parameters. Currently the following parameters are required:
\itemize{
\item model: The name of the model to use
\item stream: A logical indicating whether the API should stream responses
}

Additional parameters may be passed by adding them to the parameters list;
these parameters will be passed to the Ollama API via the body of the POST request.

Note that various Ollama options need to be set in a list named 'options' within
the parameters list (e.g., context window size is represented in $parameters$options$num_ctx).
For ease of configuration, the 'set_option' and 'set_options' functions are available
(e.g., \verb{$set_option("num_ctx", 1024)} or \verb{$set_options(list(num_ctx = 1024, temperature = 1))}).}

\item{verbose}{A logical indicating whether the interaction with the LLM provider
should be printed to the console}

\item{url}{The URL to the Ollama API endpoint for chat completion
(typically: "http://localhost:11434/api/chat")}

\item{num_ctx}{The context window size to use.
When NULL, the default context window size will be used. This is a function
argument for convenience, and will be passed to '$parameters$options$num_ctx'}
}
\value{
A new \linkS4class{llm_provider} object for use of the Ollama API
}
\description{
This function creates a new \linkS4class{llm_provider} object that interacts with the Ollama API.
}
\examples{
# Various providers:
ollama <- llm_provider_ollama()
openai <- llm_provider_openai()
openrouter <- llm_provider_openrouter()
mistral <- llm_provider_mistral()
groq <- llm_provider_groq()
xai <- llm_provider_xai()
gemini <- llm_provider_google_gemini()

# From an `ellmer::chat()` (e.g., `ellmer::chat_openai()`, ...):
\dontrun{
ellmer <- llm_provider_ellmer(ellmer::chat_openai())
}

# Initialize with settings:
ollama <- llm_provider_ollama(
  parameters = list(
    model = "llama3.2:3b",
    stream = TRUE
  ),
  verbose = TRUE,
  url = "http://localhost:11434/api/chat"
)

# Change settings:
ollama$verbose <- FALSE
ollama$parameters$stream <- FALSE
ollama$parameters$model <- "llama3.1:8b"

\dontrun{
# Try a simple chat message with '$complete_chat()':
response <- ollama$complete_chat("Hi!")
response
# $role
# [1] "assistant"
#
# $content
# [1] "How's it going? Is there something I can help you with or would you like
# to chat?"
#
# $http
# Response [http://localhost:11434/api/chat]
# Date: 2024-11-18 14:21
# Status: 200
# Content-Type: application/json; charset=utf-8
# Size: 375 B

# Use with send_prompt():
"Hi" |>
  send_prompt(ollama)
# [1] "How's your day going so far? Is there something I can help you with or
# would you like to chat?"
}
}
\seealso{
Other llm_provider: 
\code{\link{llm_provider-class}},
\code{\link{llm_provider_ellmer}()},
\code{\link{llm_provider_google_gemini}()},
\code{\link{llm_provider_groq}()},
\code{\link{llm_provider_mistral}()},
\code{\link{llm_provider_openai}()},
\code{\link{llm_provider_openrouter}()},
\code{\link{llm_provider_xai}()}
}
\concept{llm_provider}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tidyprompt.R
\name{set_system_prompt}
\alias{set_system_prompt}
\title{Set system prompt of a \linkS4class{tidyprompt} object}
\usage{
set_system_prompt(prompt, system_prompt)
}
\arguments{
\item{prompt}{A single string or a \code{\link[=tidyprompt]{tidyprompt()}} object}

\item{system_prompt}{A single character string representing the system prompt}
}
\value{
A \code{\link[=tidyprompt]{tidyprompt()}} with the system prompt set
}
\description{
Set the system prompt for a prompt. The system prompt will be added
as a message with role 'system' at the start of the chat history when
this prompt is evaluated by \code{\link[=send_prompt]{send_prompt()}}.
}
\details{
The system prompt will be stored in the \code{\link[=tidyprompt]{tidyprompt()}} object
as '$system_prompt'.
}
\examples{
prompt <- "Hi there!" |>
  set_system_prompt("You are an assistant who always answers in very short poems.")
prompt$system_prompt

\dontrun{
  prompt |>
    send_prompt(llm_provider_ollama())
  # --- Sending request to LLM provider (llama3.1:8b): ---
  #   Hi there!
  # --- Receiving response from LLM provider: ---
  #   Hello to you, I say,
  #   Welcome here, come what may!
  #   How can I assist today?
  # [1] "Hello to you, I say,\nWelcome here, come what may!\nHow can I assist today?"
}
}
\seealso{
Other pre_built_prompt_wraps: 
\code{\link{add_image}()},
\code{\link{add_text}()},
\code{\link{answer_as_boolean}()},
\code{\link{answer_as_category}()},
\code{\link{answer_as_dataframe}()},
\code{\link{answer_as_integer}()},
\code{\link{answer_as_json}()},
\code{\link{answer_as_list}()},
\code{\link{answer_as_multi_category}()},
\code{\link{answer_as_named_list}()},
\code{\link{answer_as_numeric}()},
\code{\link{answer_as_regex_match}()},
\code{\link{answer_as_text}()},
\code{\link{answer_by_chain_of_thought}()},
\code{\link{answer_by_react}()},
\code{\link{answer_using_r}()},
\code{\link{answer_using_sql}()},
\code{\link{answer_using_tools}()},
\code{\link{prompt_wrap}()},
\code{\link{quit_if}()}

Other miscellaneous_prompt_wraps: 
\code{\link{add_image}()},
\code{\link{add_text}()},
\code{\link{quit_if}()}
}
\concept{miscellaneous_prompt_wraps}
\concept{pre_built_prompt_wraps}
